/* menu.h -- Curses menu parameters

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	12 Jun 2006: superfluous declarations removed; remaining ones reordered
	 2 Jun 2006: eliminate dup'd declarations and smoothexit
	15 Oct 2005: add rm_origp pointer to REFMRUN structure
	21 Jan 2004: license message inserted
	20 Nov 2003: add REFMRUN structure type
	20 Sep 2003: function declarations removed from here
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	21 Apr 1994: renaming polydet.h->xgen.h, adding vsort.h
	 7 Dec 1993: initial X-GEN version

   Purpose: ifdef and typedefs for curses routines
   Cross References: setupcur.c, refcur.c, ...

********************************************************************/
#ifndef MENU_H
#define MENU_H	1

#ifdef CTRL
#undef CTRL
#endif
#define	CTRL(x)	(x&037)
/*
	Some preprocessor defines and typedefs associated with the
	use of the CURSES package in creating screen menus.
*/
/* defines for the signed-character values of the function and arrow keys */
#define		K_PF1		-128
#define		K_PF2		-127
#define		K_PF3		-126
#define		K_PF4		-125
#define		K_KP0		-124
#define		K_KP5		-119
#define		K_PERIOD	-111
#define		K_UP		-110
#define		K_DOWN		-109
#define		K_LEFT		-108
#define		K_RIGHT		-107
#define		K_HELP		-89
#define		K_DO		-88
#define		K_INSERT_HERE	-72
#define		K_REMOVE	-71
#define		K_PREV_SCREEN	-69
#define		K_NEXT_SCREEN	-68

#define		NSYMBOLS	36
#define		ESCAPE		033
	/* values of the data-type variable we_type */
#define		WCH	1	/* character */
#define		WSH	2	/* short */
#define		WLG	4	/* long */
#define		WUCH	17	/* unsigned char */
#define		WUSH	18	/* unsigned short */
#define		WULG	20	/* unsigned long */
#define		WFL	36	/* floating point */
#define		WDB	40	/* double-precision float */
#define		WST	65	/* string or character pointer */
#define		WPT	128	/* no variable: just a point-and-shoot spot */

	/* values of the window-entry flag variable */
#define		WWNO	1	/* white space illegal within string */
#define		WCAP	2	/* entry must be all caps */
#define		WLCA	4	/* entry must be all lowercase */
#define		WSKP	8	/* passive entry: don't allow changes */
#define		WFIL	16	/* file-type string variable */
#define		WFLG	32	/* denotes a flag-style param to a program */
#define		WBOO	64	/* boolean variable */
#define		WFBO	96	/* boolean flag */
#define		WNBO	128	/* boolean, reverse sense (yes=0,no=1) */
#define		WFNB	160	/* boolean flag, reverse sense */
#define		WARG	256	/* non-flag-style program argument */
#define		WNPG	512	/* name of a program */
#define		WNPR	520	/* name of a program, or'd with WSKP */
#define		WTYT	1024	/* space is tight: don't put in "| " between*/
#define		WTCP	1026	/* space tight; all caps (WTYT | WCAP) */
#define		WTYS	1032	/* passive entry, tight space (WTYT | WSKP)*/

typedef struct {	/* structure type for window-entry templates */
		char	*wt_label; /* string containing the prefixed label */
		char	*wt_string; 	/* string containing the entry itself */
		char	*wt_post;	/* string with postfixed label */
		char	wt_delim;	/* one-char delimiter before wt_string*/
		u_char	wt_x0;		/* column where wt_string begins */
		u_char	wt_y;		/* row where the entry lives */
		u_char	wt_strlen;	/* length of wt_string */
		u_char	wt_lablen;	/* length of wt_label */
		u_short	wt_flags;	/* flags assoc. with window entry */
		u_char	wt_afterdot;	/* digits to right of decimal if float*/
		u_char	wt_type;	/* type of variable in entry */
	} WINDOW_TEMPLATE;		

typedef struct {	/* structure type assoc. with a window entry */
		char	*we_label; /* string containing the prefixed label */
		char	*we_string; /* string containing the entry itself */
		char	*we_post;	/* string with postfixed label */
		char	we_delim;	/* one-char delimiter before we_string*/
		u_char	we_x0;		/* column where we_string begins */
		u_char	we_y;		/* row where the entry lives */
		u_char	we_strlen;	/* length of we_string */
		u_char	we_lablen;	/* length of we_label */
		u_short	we_flags;	/* flags assoc. with window entry */
		u_char	we_afterdot;	/* digits to right of decimal if float*/
		u_char	we_type;	/* type of variable in entry */
		u_char	we_prevelt;	/* element before this one */
		u_char	we_nextelt;	/* element after this one */
		u_char	we_prevrow;	/* element just above this one */
		u_char	we_nextrow;	/* element just below this one */
		u_char	we_nrstart;	/* beginning of next row */
		u_char	we_trstart;	/* beginning of this row */
		float	we_wvfloat;	/* floating-point value of element */
		long	we_wvlong;	/* long integer value of element */
		} WINDOW_ENTRY;

typedef struct {	/* overall menu structure */
		short		wnparams, wfirstactive, wlastactive;
		WINDOW_ENTRY	*wen;	/* array of window entry structures */
	      } MENUDATA;
		
/* a little tidbit to make sure `(w)addch(x) takes the right argument: */
#ifdef xctype
#undef xctype
#endif
#ifdef ultrix	/* ancient version of Curses--`addch(x)' takes char as arg */
#define	xctype	char
#endif
#ifdef __alpha
#define	xctype	unsigned int
#endif
#ifdef alliant
#define xctype	char
#endif
#ifdef sun
#define	xctype	int
#endif
#ifdef linux
#define	xctype	chtype
#endif
#ifdef VMS /* ancient version of Curses--`addch(x)' takes char as arg */
#define	xctype	char
extern int	smg$insert_line(), smg$read_keystroke();
extern int	smg$set_physical_cursor();
#endif
#ifdef _H_STANDARDS
#define xctype	int
#endif
#ifdef sgi
#define xctype	int
#endif
#ifndef xctype
#define xctype int
#endif
 
typedef struct { /* structure type associated with curses-based refinement */
		char		*rm_headstr;
		MENUDATA	*rm_currp, *rm_prevp, *rm_tempp;
		MENUDATA	*rm_statwp, *rm_origp;
		WINDOW		*rm_mwp, *rm_statp;
		} REFMRUN;

/* function prototypes */

extern int edcenmain(int argc, char *argv[]);
extern int effecx(WINDOW_ENTRY *wp);
extern int ewmmain(int argc, char *argv[]);
extern int getans(MENUDATA *mp, MENUDATA *olp, WINDOW *mwp, int par, char *hst);
extern int refmain(int argc, char *argv[]);
extern int rejmain(int argc, char *argv[]);
extern int setmore(int diagnostic);
extern int setupmain(int argc, char *argv[]);
extern int tgetent(char *bp, const char *name);
extern int winsline(WINDOW *wp);
extern char wget1ch(WINDOW *wp);
extern sichar getescseq(WINDOW *mwp);
extern sichar ordinary_key(sichar c);
extern void construct_links(MENUDATA *mp);
extern void disp_param(MENUDATA *mp, WINDOW *mwp, int parno);
extern void menu_setup(MENUDATA *mp, WINDOW *mwp, char *hdrstr, int menufl);
extern void menufree(MENUDATA *oldmenu);
extern void put_pars(FILE *fp, MENUDATA *mp, int where);
extern void restore_value(MENUDATA *curmp, MENUDATA *orp, int parno);
extern void setmzap(void);
extern void stuff_value(WINDOW_ENTRY *wep);
extern void update_par(MENUDATA *mp, WINDOW *mwp, int par);
extern void valtostr(WINDOW_ENTRY *wep);
extern MENUDATA *newmenu(unsigned nparams, WINDOW_TEMPLATE *window_init);

#endif /* MENU_H */

/* end of menu.h */
