/* refmain.c -- top-level routines for XGEN's parameter refinement

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Nov 2006: derivedglobs -> makeq21, mat2vecs -> rp2vecs;;
	 update showbravais,artocel,showbravchoice calls.
	12 Jun 2006: guesstoxp replaces startvals and some local code;
		usebest moved to ../xgrefine/refutil.c
	15 Jan 2006: bugfixes
	 7 Jan 2006: eliminate static global "lstatp";
		this allows elimination of some onerous bookkeeping
	21 Nov 2005: incorporate routines from refmenu.c into here
	26 Oct 2005: replace str2xgbuf() and flush_xgbuf() calls with
		calls to xgbuf()
	15 Oct 2005: eliminate static globals; origp now part of REFMRUN struct
	 4 Oct 2005: incorporate refcur.c & autoindex.c into here;housekeeping
	28 Jun 2005: remap_distortion() call in command() updated
	22 Jun 2005: init_menu(), refcur_init(), crefexit(), killit() static
	10 Jan 2005: default gamma2 value is 0.0010, not 0.0026
	 6 Sep 2004: cosmetic changes
	27 Aug 2004: routines used only here are declared as static; cosmetics
	 4 Jun 2004: zero the REFINERUN structure in refmain(), not refinit();
		eliminate svchange() routine;
		get_rmserrs always recalculates diffraction-vectors
	28 May 2004: diagnostic in bravmenu() reflects figure-of-merit concept
	30 Mar 2004: set rfp->rfr_fpl to NULL after closing it in command()
		on 'E' and 'Q' commands
	16 Feb 2004: add missing arguments in two calls to gripe()
	24 Jan 2004: nfiel[] made static local in journal_parchange()
	21 Jan 2004: license message inserted;
		merge routines from refmlib.c back into here
	13 Jan 2004: incorporate routines from minpar.c into here
	 8 Jan 2004: REFMRUN, REFPARS, REFINERUN structure elements replace
		global variables; some extra function args required
	 3 Jan 2004: replace global variables with elements of REFPARS
		structures from function arguments
	22 Dec 2003: incorporate nonlinear-refinement codes so that
		the signal-catching stuff works correctly without global vbls
	13 Dec 2003: eliminate failure as variable from command();
		bugfix converting from radians to degrees in partgl()
	 8 Dec 2003: eliminate framdeg as a variable
	24 Nov 2003: replace globals associated with curses with elements
		of the REFMRUN structures
	 8 Nov 2003: convert lca from a vector to a pointer
	 5 Nov 2003: minor reformatting in bravmenu()
	19 Sep 2003: initialize array of structs according to ansi C braces;
		 explicit typecasting of lc_latchar when we print it
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	24 Aug 2003: display lattice type & centering
		using routines glattice(), gcentering()
	14 Aug 2003: convert origp to a static global;
		eliminate unused MENUDATA, REFPARS structures
		use local array ldext[2] to hold the starting and
		ending difference vectors so we don't carry diffv0 and dvlast
		track return value from globtxs in all cases
	13 Aug 2003: remove declaration of lcacmp()
	20 Mar 2003: rocking_ref call takes extra argument for diagnostics
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h;
		material in refmenu.h imported into this file;
		replace line[] with mline[] to minimize overlap with globals
	15 Jul 2002: failure in rocking_ref() as called from menu is non-fatal
	25 Mar 2002: removed MSI attachments
	 2 Sep 1994: slight change in output information
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: toplevel routines for refining crystal and detector params.
	It does setup operations and on a cycle-by-cycle basis it calls
	routines found in other files. The routines associated with
	nonlinear refinement are also found here.
   Cross references: ../xgrefine/(*).c
*************************************************************************/
#define		NEED_UNISTD	1
#define		NEED_TIME	1
#define		NEED_CURSES	1
#include	"refidefs.h"
#include	"menu.h"
#define		MENU_HEIGHT	6	/* # of solutions to view at a time */
#define		FABULOUS	0.08	/* very good residual (unsquared) */
#define		DEFAULT_BASISANGLE	50
#define		NSHRINKS	4	/*# of times to shrink deltas*/
#define		SHRINKDEL	(0.5)	/* shrink factor per cycle */
#define		ITER		5	/* # of iterations to perform */
	/* macros for manipulations of certain MENUDATA entries */
#define		MPCH(x)		(parp->wen[x].we_string[0])
#define		MFL(x)		(parp->wen[x].we_wvfloat)
#define		SFL(x)		(swp->wen[x].we_wvfloat)
#define		SLG(x)		(swp->wen[x].we_wvlong)

extern void abort(void);
extern int refmain(int argc, char *argv[]);
extern int endwin(void);

static REFMRUN *managermp(int opno)
{ /*	Manages a REFMRUN structure */
	int		nentries, sentries;
	static REFMRUN	*rmp;
	static char	headmsg[] = "rmp->rm_headstr in managermp";
	static char	rmpmsg[] = "rmp in managermp";
	static WINDOW_TEMPLATE	rftemp[] = {
	{ "Refining? a","Y","",':',12, 1, 1,11,WTCP,0,WCH},
	{ "b","Y","",':',20,1,1,1,WTCP,0,WCH},
	{"c","Y","",':',28,1,1,1,WTCP,0,WCH}, 
	{"alpha","N","",':',36,1,1,5,WTCP,0,WCH},
	{"beta","N","",':',44,1,1,4,WTCP,0,WCH},
	{"gamma","N","",':',52,1,1,5,WTCP,0,WCH},
	{"omega","Y","",':',60,1,1,5,WTCP,0,WCH},
	{"chi","Y","",':',68,1,1,3,WTCP,0,WCH},
	{"phi","Y","",':',76,1,1,3,WTCP,0,WCH},
	{"Value"," 80.000","",':', 6, 2,7,5,WTYT,3,WFL},
	{"","88.000","",' ',13,2,8,0,WTYT,3,WFL},
	{"","95.000","",' ',21,2,8,0,WTYT,3,WFL},
	{"","90.000","",' ',29,2,8,0,WTYT,3,WFL},
	{"","90.000","",' ',37,2,8,0,WTYT,3,WFL},
	{"","90.000","",' ',45,2,8,0,WTYT,3,WFL},
	{""," 0.000","",' ',53,2,8,0,WTYT,3,WFL},
	{""," 0.000","",' ',61,2,8,0,WTYT,3,WFL},
	{""," 0.000","",' ',69,2,8,0,WTYT,3,WFL},
	{"Shifts","  0.000","",':', 6, 3,7,6,WTYS,4,WFL},
	{ "","0.","",' ',13, 3, 8,0,WTYS,4,WFL},
	{"","0.","",' ',21, 3,8,0,WTYS,4,WFL},
	{"","0.","",' ',29, 3, 8,0,WTYS,4,WFL},
	{"","0.","",' ',37, 3,8,0,WTYS,4,WFL},
	{"","0.","",' ',45, 3, 8,0,WTYS,4,WFL},
	{"","0.","",' ',53, 3,8,0,WTYS,4,WFL},
	{"","0.","",' ',61, 3, 8,0,WTYS,4,WFL},
	{"","0.","",' ',69, 3,8,0,WTYS,4,WFL},
	{"plateD","N","",':',12,4,1,6,WTCP,0,WCH},
	{"Xcen","Y","",':',20,4,1,4,WTCP,0,WCH},
	{"Ycen","Y","",':',28,4,1,4,WTCP,0,WCH},
	{"tilt","N","",':',36,4,1,4,WTCP,0,WCH},
	{"swing","N","",':',44,4,1,5,WTCP,0,WCH},
	{"gam0","Y","",':',52,4,1,4,WTCP,0,WCH},
	{"gam1","Y","",':',60,4,1,4,WTCP,0,WCH},
	{"gam2","N","",':',68,4,1,4,WTCP,0,WCH},
	{"Value","10.0000","",':', 6, 5, 7, 5, WTYT, 4,WFL},
	{"","0.0","",' ',14,5,7,0,WTYT, 4,WFL},
	{"","0.0","",' ',22,5,7,0,WTYT, 4,WFL},
	{"","0.0","",' ',30,5,7,0,WTYT, 4,WFL},
	{"","0.0","",' ',38,5,7,0,WTYT, 3,WFL},
	{"","1.0","",' ',46,5,7,0,WTYT, 3,WFL},
	{"","1.0","",' ',54,5,7,0,WTYT, 3,WFL},
	{"",".0010","",' ',62, 5, 7, 0, 0, 4,WFL},
	{"Shifts","0.0","",':', 6, 6,7,6,WTYS,4,WFL},
	{"","0.0","",' ',14, 6,7,0,WTYS,4,WFL},
	{"","0.0","",' ',22, 6,7,0,WTYS,4,WFL},
	{"","0.0","",' ',30, 6,7,0,WTYS,4,WFL},
	{"","0.0","",' ',38, 6,7,0,WTYS,4,WFL},
	{"","0.0","",' ',46, 6,7,0,WTYS,4,WFL},
	{"","0.0","",' ',54, 6,7,0,WTYS,4,WFL},
	{"","0.0","",' ',62, 6,7,0,WTYS,4,WFL},
	{"Max Errors Allowed in:        Resolution",
		"","",' ',46,7,0,42,WTYS,0,WST},
	{" Wave- Stepsiz Crystal   Space","","",' ',78,7,0,30,WTYS,0,WST},
	{"index X(pix)  Y(pix) phi(fm)     Min     Max",
		"","",' ',46,8,0,44,WSKP,0,WST},
	{"length degrees  System   Group","","",' ',78,8,0,30,WTYS,0,WST},
	{"","0.2000","",' ', 0,9,6,0,WTYT,4,WFL},
	{""," 5.000","",' ', 7,9,6,0,WTYT,4,WFL},
	{""," 5.000","",' ',15,9,6,0,WTYT,4,WFL},
	{""," 5.000","",' ',23,9,6,0,WTYT,4,WFL},
	{"","80.000","",' ',30,9,7,0,WTYT,4,WFL},
	{""," 0.800","",' ',38,9,7,0,WTYT,4,WFL},
	{"","1.5418","",' ',46,9,7,0,WTYT,4,WFL},
	{"","0.2500","",' ',54,9,7,0,WTYT,4,WFL},
	{"","     3","",' ',62,9,7,0,WTYT,0,WSH},
	{"","    19","",' ',70,9,7,0,WTYT,0,WSH},
	{"True start omega","   0.00","",':',17,10,8,16,WTYT,3,WFL},
	{"chi","   0.00","",':', 31,10,8,3,WTYT,3,WFL},
	{"phi","   0.00","",':', 44,10,8,3,WTYT,3,WFL},
	{"Mainbeam",  "255","", ':',62,10, 7, 8,WTYS,1,WFL},
	{"",          "255","", ' ',70,10, 7, 0,WTYS,1,WFL},
	{"Command(IRULPHAMWEQGSC,0-9)", "R","",':',31,11,1,30,WCAP,0,WCH},
	{"{ TYPE ^X to do it, ^Y for help }",
		"","",' ',78,11, 0,38,WTYS,0,WST } };

	static WINDOW_TEMPLATE	statemp[] = {
	{ "-----Root-Mean-Squared Errors In","","",':',36, 0, 0,32,WTYS,0,WST},
	{ "--------    <#frms    # of    # of","","",
		' ',71, 0, 0,35,WTYS,0,WST},
	{  "Index      X       Y     phi   Gamma", "","",
		' ',39, 1, 0,37,WTYS,0,WST},
	{ "Profile  moved>   refls indices","","",' ',72, 1, 0,33,WTYS,0,WST},
	{"","0.0500","",' ', 0,2,6,0,WTYS,4,WFL},
	{"","0.2500","",' ', 7,2,6,0,WTYS,4,WFL},
	{"","0.2500","",' ',14,2,7,0,WTYS,4,WFL},
	{"","0.1500","",' ',22,2,7,0,WTYS,4,WFL},
	{"","0.6000","",' ',30,2,7,0,WTYS,4,WFL},
	{"","1.5000","",' ',38,2,7,0,WTYS,4,WFL},
	{"","0.6000","",' ',46,2,7,0,WTYS,4,WFL},
	{"","   200","",' ',54,2,7,0,WTYS,0,WSH},
	{"","   600","",' ',62,2,7,0,WTYS,0,WSH},
	{"","0.0500","",' ', 0,3,6,0,WTYS,4,WFL},
	{"","0.2500","",' ', 7,3,6,0,WTYS,4,WFL},
	{"","0.2500","",' ',14,3,7,0,WTYS,4,WFL},
	{"","0.1500","",' ',22,3,7,0,WTYS,4,WFL},
	{"","   200","",' ',54,3,7,0,WTYS,0,WSH},
	{"","   600","",' ',62,3,7,0,WTYS,0,WSH},
	{"D","14.00","",'>', 8,4,5,1,WTYS,2,WFL},
	{"D", "8.00","",'>',16,4,5,1,WTYS,2,WFL},
	{"D"," 5.00","",'>',24,4,5,1,WTYS,2,WFL},
	{"D", "4.00","",'>',32,4,5,1,WTYS,2,WFL},
	{"D"," 3.20","",'>',40,4,5,1,WTYS,2,WFL},
	{"D", "2.80","",'>',48,4,5,1,WTYS,2,WFL},
	{"D"," 2.60","",'>',56,4,5,1,WTYS,2,WFL},
	{"D"," 2.50"," Shells",'>',64,4,5,1,WTYS,2,WFL},
	{"", "0.040","",' ', 6,5,7,0,WTYS,4,WFL},
	{"", "0.040","",' ',14,5,7,0,WTYS,4,WFL},
	{"", "0.040","",' ',22,5,7,0,WTYS,4,WFL},
	{"", "0.040","",' ',30,5,7,0,WTYS,4,WFL},
	{"", "0.040","",' ',38,5,7,0,WTYS,4,WFL},
	{"", "0.040","",' ',46,5,7,0,WTYS,4,WFL},
	{"", "0.040","",' ',54,5,7,0,WTYS,4,WFL},
	{"", "0.040"," Err(h)",' ',62,5,7,0,WTYS,4,WFL},
	{"", " 48","/",' ', 6, 6,3,7,WTYS,0,WSH},
	{"", "144","", ' ',10, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',14, 6,3,0,WTYS,0,WSH},
	{"", "144","", ' ',18, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',22, 6,3,0,WTYS,0,WSH},
	{"", "144","", ' ',26, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',30, 6,3,0,WTYS,0,WSH},
	{"", "144","", ' ',34, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',38, 6,3,0,WTYS,0,WSH},
	{"", "144","", ' ',42, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',46, 6,3,0,WTYS,0,WSH},
	{"", "144","", ' ',50, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',54, 6,3,0,WTYS,0,WSH},
	{"", "144","", ' ',58, 6,3,0,WTYS,0,WSH},
	{"", " 48","/",' ',62, 6,3,0,WTYS,0,WSH},
	{"", "144"," ind/ref",' ',66, 6,3,0,WTYS,0,WSH},
	{"Index err(h)>",
		" 0.000 0.050 0.100 0.150 0.200 0.250 0.300 0.350 0.400 0.450",
		"", ' ',15, 7,60,13,WTYS,0,WST},
	{"h","150","",' ',16, 8,5,10,WTYS,0,WSH},
	{""," 80","",' ',22, 8,5,0,WTYS,0,WSH},
	{"", " 20","",' ',28, 8,5, 0,WTYS,0,WSH},
	{""," 10","",' ',34, 8,5,0,WTYS,0,WSH},
	{"", "  6","",' ',40, 8,5, 0,WTYS,0,WSH},
	{"","  3","",' ',46, 8,5,0,WTYS,0,WSH},
	{"", "  0","",' ',52, 8,5, 0,WTYS,0,WSH},
	{"", "  0","",' ',58, 8,5,0,WTYS,0,WSH},
	{"", "  0","",' ',64, 8,5, 0,WTYS,0,WSH},
	{"", "  0","",' ',70, 8,5,0,WTYS,0,WSH},
	{"k","150","",' ',16, 9,5,10,WTYS,0,WSH},
	{"", " 80","",' ',22, 9,5,0,WTYS,0,WSH},
	{"", " 20","",' ',28, 9,5, 0,WTYS,0,WSH},
	{"", " 10","",' ',34, 9,5,0,WTYS,0,WSH},
	{"", "  6","",' ',40, 9,5, 0,WTYS,0,WSH},
	{"", "  3","",' ',46, 9,5,0,WTYS,0,WSH},
	{"", "  0","",' ',52, 9,5, 0,WTYS,0,WSH},
	{"", "  0","",' ',58, 9,5,0,WTYS,0,WSH},
	{"", "  0","",' ',64, 9,5, 0,WTYS,0,WSH},
	{"", "  0","",' ',70, 9,5,0,WTYS,0,WSH},
	{"l","150","",' ',16,10,5,10,WTYS,0,WSH},
	{"", " 80","",' ',22,10,5,0,WTYS,0,WSH},
	{"", " 20","",' ',28,10,5, 0,WTYS,0,WSH},
	{"", " 10","",' ',34,10,5,0,WTYS,0,WSH},
	{"", "  6","",' ',40,10,5, 0,WTYS,0,WSH},
	{"", "  3","",' ',46,10,5,0,WTYS,0,WSH},
	{"", "  0","",' ',52,10,5, 0,WTYS,0,WSH},
	{"", "  0","",' ',58,10,5,0,WTYS,0,WSH},
	{"", "  0","",' ',64,10,5, 0,WTYS,0,WSH},
	{"", "  0","",' ',70,10,5,0,WTYS,0,WSH} } ;

	if (!opno) /* eliminate memory and other resources */
	  {
		if (NULL != rmp) (void)managermp(01774);
		memzap((AVOID **)&rmp, rmpmsg);	return NULL;
	  }
	if (!(opno & 01)) /* even-numbered args: selective eliminations */
	  {
		if (NULL != rmp)
		  {
			if (opno & 04)
			 memzap((AVOID **)&(rmp->rm_headstr), headmsg);
			if (opno & 010)
			  {
				menufree(rmp->rm_currp); rmp->rm_currp = NULL;
			  }
			if (opno & 020)
			  {
				menufree(rmp->rm_prevp); rmp->rm_prevp = NULL;
			  }
			if (opno & 040)
			  {
				menufree(rmp->rm_tempp); rmp->rm_tempp = NULL;
			  }
			if (opno & 0100)
			  {
				menufree(rmp->rm_statwp); rmp->rm_statwp = NULL;
			  }
			if (opno & 0200)
			  {
				menufree(rmp->rm_origp); rmp->rm_origp = NULL;
			  }
			if ((opno & 0400) && (NULL != rmp->rm_mwp))
			  {
				(void)delwin(rmp->rm_mwp); rmp->rm_mwp = NULL;
			  }
			if ((opno & 01000) && (NULL != rmp->rm_statp))
			  {
				(void)delwin(rmp->rm_statp);
				rmp->rm_statp = NULL;
			  }
		  }
		return rmp;
	  }
	if (opno == 03) return rmp; /* unconditional return */
	if (NULL == rmp) rmp = (REFMRUN *)callop(1, sizeof (REFMRUN), rmpmsg);
	nentries = (sizeof (rftemp)) / (sizeof (WINDOW_TEMPLATE));
	sentries = (sizeof (statemp)) / (sizeof (WINDOW_TEMPLATE));
	if ((opno & 04) && (NULL == rmp->rm_headstr))
	  {
		if (NULL == (rmp->rm_headstr = (char *)mallop(80, headmsg)))
			return managermp(0);
	  }
	if ((opno & 010) && (NULL == rmp->rm_currp)) /* current MENUDATA */
		rmp->rm_currp = newmenu(nentries, rftemp);
	if ((opno & 020) && (NULL == rmp->rm_prevp)) /* previous MENUDATA */
		rmp->rm_prevp = newmenu(nentries, rftemp);
	if ((opno & 040) && (NULL == rmp->rm_tempp)) /* temporary MENUDATA */
		rmp->rm_tempp = newmenu(nentries, rftemp);
	if ((opno & 0100) && (NULL == rmp->rm_statwp)) /* stats MENUDATA */
		rmp->rm_statwp = newmenu(sentries, statemp);
	if ((opno & 0200) && (NULL == rmp->rm_origp))
		rmp->rm_origp = newmenu(nentries, rftemp);
	if ((opno & 0400) && (NULL == rmp->rm_mwp))
	  {
		initscr();	crmode();	noecho();	nonl();
		rmp->rm_mwp = newwin(12, 80, 0, 0);
		clear();	wrefresh(stdscr);
	  }
	if ((opno & 01000) && (NULL == rmp->rm_statp))
		rmp->rm_statp = newwin(11, 80, 12, 0); /* help menu */
	return rmp;
}	

static double menu_number(char *s, REFMRUN *rmp)
{ /* This displays a query and gets back a floating-point response */
	register int	i;
	int		herex, herey;
	char		c, mline[40];

	wclrtoeol(rmp->rm_statp);	/* so line is clean to begin with */
	wprintw(rmp->rm_statp,"%s: ", s);	wrefresh(rmp->rm_statp);
	for (i = 0; '\0' != (c = wget1ch(rmp->rm_statp));
	 i++, wrefresh(rmp->rm_statp)) /* loop, looking for valid characters */
	   {
		if ('.' == c || ('0' <= c && c <= '9')) /* legal character */
		  {
			mline[i] = c;	waddch(rmp->rm_statp, (xctype)(c));
		  }
		else if ('\b' == c || 0177 == c) /* erase last char */
		  {
			if (!i) i--; /* zeroth char: just start over */
			else	/* > 0th char: logical and physical erase */
			  {
				i -= 2; getyx(rmp->rm_statp,herey,herex);
				wmove(rmp->rm_statp,herey,herex-1);
				waddch(rmp->rm_statp, (xctype)(' '));
				wmove(rmp->rm_statp,herey,herex-1);
			  }
		  }
		else if ('\n' == c || '\r' == c) break;	/* quit updating */
		/* any other character we just ignore... */
	   }
	mline[i] = '\0';	wrefresh(rmp->rm_statp);
	return atof(mline);	/* answer should be a double */
}

static int ask_solution(int is_at_end, REFMRUN *rmp, REFINERUN *rfp)
{ /* This asks the user for a solution choice.  The parameter `is_at_end'
	determines what we do if the user hits a carriage return; if
	is_at_end == 0, we just continue going through the list; otherwise
	we restore the original parameters and return. */

	register int	i;
	char		c, mline[120];
	int		desired_soln, herex, herey, stated_soln;
	REFPARS		*xp;

	desired_soln = -1;	/* default is no solution at all */
	if (is_at_end) mvwaddstr(rmp->rm_mwp,0,0,
	 "Enter # of preferred solution (<CR> for no solution):     ");
	else	mvwaddstr(rmp->rm_mwp,0,0,
	 "Enter # of preferred solution (<CR> to continue looking): ");
	wrefresh(rmp->rm_mwp);
	xp = rfp->rfr_xp;	/* shorthand for current REFPARS structure */
	for (i = 0; '\0' != (c = wget1ch(rmp->rm_mwp));
	 i++, wrefresh(rmp->rm_mwp)) /* loop through characters of answer */
	   {
		if ('0' <= c && c <= '9') /* legal character */
		  {
			mline[i] = c;	waddch(rmp->rm_mwp, (xctype)(c));
		  }
		else if ('\b' == c || 0177 == c) /* erase last char */
		  {
			if (!i) i--; /* zeroth char: just start over */
			else	/* > 0th char: logical and physical erase */
			  {
				i -= 2; getyx(rmp->rm_mwp,herey,herex);
				wmove(rmp->rm_mwp, herey, herex-1);
				waddch(rmp->rm_mwp, (xctype)(' '));
				wmove(rmp->rm_mwp, herey, herex-1);
			  }
		  }
		else break;	/* anything else means quit updating */
	   }
	mline[i] = '\0';
	if ('0' <= mline[0] && mline[0] <= '9' && /* legal numerical answer */
	 0 <= (stated_soln = atoi(mline)) &&
	 stated_soln < (int)rfp->rfr_ndecent)
	  {
		desired_soln = stated_soln; /* remember unit cell&orientation*/
		guesstoxp(rfp->rfr_gu0 + desired_soln, xp);
		fprintf(rfp->rfr_fpl, "Solution %d chosen\n",desired_soln);
		mvwaddstr(rmp->rm_mwp,0,0,
		 "Successful completion of autoindexing operation          \n");
		wclrtoeol(rmp->rm_mwp);
		wprintw(rmp->rm_mwp,"Solution %d chosen",desired_soln);
		wrefresh(rmp->rm_mwp);
	  }
	else {
		if (is_at_end) /* have we seen all the solutions? */
		  {
			guesstoxp(rfp->rfr_gu0 + (NGUESSES - 1), rfp->rfr_xp);
			wprintw(rmp->rm_mwp, "No solution chosen\n");
			wrefresh(rmp->rm_mwp);
			fprintf(rfp->rfr_fpl, "No solution chosen\n");
		  }
		else {
			wmove(rmp->rm_mwp, 4, 0); wrefresh(rmp->rm_mwp); }
	     }
	return desired_soln; /* return the index of the desired solution */
}

static void guess_query(REFMRUN *rmp, REFINERUN *rfp)
{ /*	This asks the user if he wants any of the auto-indexing solutions */
	register GUESS	*gp;
	unsigned int	i;

	wclear(rmp->rm_mwp);	wrefresh(rmp->rm_mwp);
	mvwaddstr(rmp->rm_mwp,2,0,"Possible solutions, ranked by residual:\n");
	wprintw(rmp->rm_mwp, "  #      a      b      c  alpha   beta  gamma");
	wprintw(rmp->rm_mwp, "   omega     chi     phi residual\n");
	for (i = 0, gp = rfp->rfr_gu0; i < rfp->rfr_ndecent; i++, gp++)
	   {
		(void)winsline(rmp->rm_mwp); /* insert a blank line here */
		wprintw(rmp->rm_mwp,
		 "%3d%7.2f%7.2f%7.2f%7.2f%7.2f%7.2f%8.3f%8.3f%8.3f %8.5f\r",
			i, gp->gabc[0], gp->gabc[1], gp->gabc[2],
			gp->gang[0], gp->gang[1], gp->gang[2],
			gp->gomega[0],gp->gomega[1],gp->gomega[2],gp->gresid);
		wrefresh(rmp->rm_mwp);
		if (!((i+1) % MENU_HEIGHT) && (i < (rfp->rfr_ndecent-1)) &&
			(-1 != ask_solution(0, rmp, rfp))) return;
	   }
	(void)ask_solution(1, rmp, rfp); /* ask for an answer */
}

static void outresid(double initr[], REFMRUN *rmp, REFINERUN *rfp)
{ /* This prints some in-progress diagnostics to the CURSES screen */
	register int	i, j, m;
	double		*bp;
	MENUDATA	*ltp, *lsp, *lpp;

	bp = &(rfp->rfr_bestsofar[0]);	wmove(rmp->rm_mwp, 0, 0);
	wprintw(rmp->rm_mwp,
 "Nonlinear residual calculation %4d: best residual so far = %12.7f\n",
		(int)(rfp->rfr_nrescalc), sqrt(rfp->rfr_bestres));
	ltp = rmp->rm_tempp;	lsp = rmp->rm_statwp;	lpp = rmp->rm_prevp;
	for (j = 0; j < THESEPARS; j++, bp++)
	   {
		if (j < NPAR0)
		  {
			m = j + NPAR0;	i = m + NPAR0;
		  }
		else {
			m = j + 2 * NPAR0 + NPAR1;	i = m + NPAR1;
		     }
		ltp->wen[m].we_wvfloat = *bp;
		ltp->wen[i].we_wvfloat = *bp - lpp->wen[m].we_wvfloat;
		valtostr(&(ltp->wen[m]));	valtostr(&(ltp->wen[i]));
		disp_param(ltp, rmp->rm_mwp, m);
		disp_param(ltp, rmp->rm_mwp, i);
	   }
	lsp->wen[13].we_wvfloat = sqrt(rfp->rfr_besth);
	lsp->wen[14].we_wvfloat = sqrt(rfp->rfr_bestx * initr[1] / initr[0]);
	lsp->wen[15].we_wvfloat = sqrt(rfp->rfr_besty * initr[2] / initr[0]);
	lsp->wen[16].we_wvfloat = sqrt(rfp->rfr_bestphi * initr[3] / initr[0]);
	lsp->wen[17].we_wvlong = rfp->rfr_bestnr;
	lsp->wen[18].we_wvlong = rfp->rfr_bestnr * 3;
	for (j = 13; j <= 18; j++)
	   {
		valtostr(&(lsp->wen[j])); disp_param(lsp, rmp->rm_statp, j);
	   }
	wrefresh(rmp->rm_statp);	wmove(rmp->rm_mwp, 0, 0);
	wrefresh(rmp->rm_mwp);
}

static void end_menu(void)
{ /* This is called just before exiting or returning to the caller. */
	REFINERUN	*rfp;
	REFMRUN		*rmp;

	signal(SIGINT, SIG_DFL);
	if (NULL != (rmp = managermp(3)))
	  {
		if (NULL != rmp->rm_statp)
		  {
			wmove(rmp->rm_statp, 10, 0);	/* move to the bottom*/
			wclrtoeol(rmp->rm_statp);	/* clean up that line*/
			wrefresh(rmp->rm_statp);
		  }
	  }
	if (NULL != (rfp = managerfp(3, 0)))
	  {
		rfp->rfr_flags &= RR_QUITNONLIN; /* no more nonlinear */
		rfp->rfr_flags |= RR_QUITREFINE; /* we're ready to quit */
		if (NULL != rfp->rfr_fpl) (void)fclose(rfp->rfr_fpl);
		rfp->rfr_fpl = NULL;
	  }
	endwin();	/* end the curses stuff altogether */
}

static void gripe(char *s, int code, REFMRUN *rmp, REFINERUN *rfp)
{ /* this prints an error message in two useful places and then
		either exits or returns depending on `code'. */
	if (NULL != rmp->rm_mwp)
	  {
		wmove(rmp->rm_mwp,0,0);
		wclrtoeol(rmp->rm_mwp); /* clear top row */
		waddstr(rmp->rm_mwp, s); /* print error message on the screen*/
	  }
	if (NULL == rfp->rfr_fpl) rfp->rfr_fpl = propen("xlog");
	if (NULL != rfp->rfr_fpl) fputs(s, rfp->rfr_fpl); /* print to logfile*/
	switch (code)
	      {
		case	0:	/* code for abort-cycle warnings */
		  if (NULL != rmp->rm_mwp)
		    {
			waddstr(rmp->rm_mwp,"; refinement cycle aborted");
			wrefresh(rmp->rm_mwp);
		    }
		  if (NULL != rfp->rfr_fpl)
			fputs("; refinement cycle aborted\n", rfp->rfr_fpl);
		  return;
		case	1:	/* code for uncommented exit */
		  if (NULL != rmp->rm_mwp) waddstr(rmp->rm_mwp,"; exiting");
		  if (NULL != rfp->rfr_fpl) fputs("; exiting\n", rfp->rfr_fpl);
		  end_menu();
		case	2:	/* code for abort-calculation warnings */
		  if (NULL != rmp->rm_mwp)
		    {
			waddstr(rmp->rm_mwp,"; no calculation performed");
			wrefresh(rmp->rm_mwp);
		    }
		  if (NULL != rfp->rfr_fpl)
			fputs("; no calculation performed\n", rfp->rfr_fpl);
		  return;
		default :	/* error codes (usually < 0) */
		  if (NULL !=rmp->rm_mwp)
			wprintw(rmp->rm_mwp,", error code %d; exiting", code);
		  if (NULL != rfp->rfr_fpl)
		    fprintf(rfp->rfr_fpl,", error code %d; exiting\n", code);
		  end_menu();
	      }
}

static void partgl(MENUDATA *parp, REFMRUN *rmp, REFINERUN *rfp)
{ /* This creates global variables from a MENUDATA structure */
	register int	i, j, m;
	int		retval;
	float		*pshp;
	REFPARS		*xp;
	WINDOW_ENTRY	*wep;

	xp = rfp->rfr_xp;
	for (i = parp->wnparams, wep = &(parp->wen[0]); i--; wep++)
		stuff_value(wep);
	for (i = NPAR0; i--; ) xp->xisref[i] = MPCH(i);
	for (i = 0, j = NPAR0, m = NPAR0 + 3; i < 3; i++, j++, m++)
	   {
		xp->xabc[i] = MFL(j);	xp->xang[i] = MFL(m);	}
	xp->xdelom  = MFL(15); xp->xdelchi = MFL(16); xp->xdelphi = MFL(17);
	pshp = rfp->rfr_pshift;
	for (i = 0, j = 2 * NPAR0; i < NPAR0; i++, j++) *pshp++ = MFL(j);
	for (i = NPAR0, j = 3 * NPAR0; i < (NPAR0 + NPAR1); i++, j++)
		xp->xisref[i] = MPCH(j);
	xp->xplated = MFL(35);	xp->xx0 = MFL(36);	xp->xy0 = MFL(37);
	xp->xtilt = MFL(38) * RADC;	xp->xswing = MFL(39) * RADC;
	xp->xgam[0] = MFL(40) * fabs(xp->xfrmsize);
	xp->xgam[1] = MFL(41) * fabs(xp->xfrmsize);	xp->xgam[2] = MFL(42);
	for (i = NPAR0, j = 3 * NPAR0 + 2 * NPAR1; i < NPAR0 + NPAR1; i++, j++)
		*pshp++ = MFL(j);
	xp->xerrind = MFL(55);	xp->xerrx = MFL(56);
	xp->xerry = MFL(57);	xp->xerrphi = MFL(58);
	xp->xresmin = MFL(59);	xp->xresmax = MFL(60);
	xp->xlambda = MFL(61);	xp->xfrmsize = MFL(62) * RADC;
	xp->xisys = parp->wen[63].we_wvlong;
	xp->xispgp = parp->wen[64].we_wvlong;
	for (i = 65; i < 68; i++) xp->xstart[i-65] = MFL(i);
	xp->xpltd2 = xp->xplated * xp->xplated;	makeq21(xp);
	xp->xcmd = parp->wen[70].we_string[0];
	if ((retval = unitcellx(rfp)))
	  {
		fprintf(rfp->rfr_fpl,
		 "unit_cell failure in partgl: retval = %d\n", retval);
		gripe("Cannot define unit cell from matrix", 1, rmp, rfp);
	  }
}

static void end_incr(REFMRUN *rmp, REFINERUN *rfp)
{ /* wrap up a cycle of incremental refinement */
	register int	i;
	MENUDATA	*lcp, *lmp;

	lcp = rmp->rm_currp;	lmp = rmp->rm_tempp;
	for (i = lcp->wnparams; i--; ) restore_value(lcp, lmp, i);
	partgl(lcp, rmp, rfp);	get_rmserrs("incr. linear lsq", rfp);
	rfp->rfr_jtj = NULL;
}

static void finishnl(REFMRUN *rmp, REFINERUN *rfp, char rtype)
{ /* neatly close off a run of nonlinear refinement from Curses */
	register int	i, j, m;
	register double	*xnow;
	MENUDATA	*lcp, *lpp;
	static char	outmsg[] = "nonlinear,code  ";

	lcp = rmp->rm_currp;	lpp = rmp->rm_prevp;
	for (j = 0, xnow = rfp->rfr_xnow0; j < THESEPARS; j++, xnow++)
	   {
		if (j < NPAR0)
		  {
			m = j + NPAR0;	i = m + NPAR0;
		  }
		else {
			m = j + 2 * NPAR0 + NPAR1;	i = m + NPAR1;
		     }
		lcp->wen[m].we_wvfloat = *xnow;
		lcp->wen[i].we_wvfloat = rfp->rfr_pshift[j] =
			*xnow - lpp->wen[m].we_wvfloat;
		valtostr(&(lcp->wen[m]));	valtostr(&(lcp->wen[i]));
	   }
	for (j = THESEPARS, m = THESEPARS + 34; j < ALLPARS; j++, m++)
	   {
		rfp->rfr_pshift[j] = lcp->wen[m].we_wvfloat = 0.;
		valtostr(&(lcp->wen[m]));
	   }
	partgl(rmp->rm_currp, rmp, rfp);
	outmsg[15] = rtype;	get_rmserrs(outmsg, rfp);
}

static void minpar(char reftype, REFMRUN *rmp, REFINERUN *rfp)
{ /*	Optimization routine for refinement of parameters.
	The argument `reftype' determines what kind of residual to get back:
		'P' = spindle-angle only;
		'H' = integerness only;
		'A' = weighted sum of spindle, integerness, X, Y.
	  '0' - '9' = mixture of 'P' & 'H', varying in weight:  0 == all 'P',
							 	9 == all 'H' */
	register int	i;
	int		doshrink, nk;
	double		fold, iwt, xywt, *dfp, initr[4];
	MENUDATA	*lcp;
	WINDOW_ENTRY	*wep;
/* initialize some local variables out of the refinement globals */
	get_rmserrs("before nonlinear", rfp);	lcp = rmp->rm_currp;
	for (i = 0, dfp = &(rfp->rfr_xnow0[0]), wep = &(lcp->wen[NPAR0]);
	  i < NPAR0; i++, dfp++, wep++) *dfp = wep->we_wvfloat;
	for (wep = &(lcp->wen[3 * NPAR0 + NPAR1]);
	  i < THESEPARS; i++, dfp++, wep++) *dfp = wep->we_wvfloat;
	/* set up for refinement and get the initial residual */
	if ((reftype == '0') || (reftype == 'P')) iwt = 0.;
	else if ((reftype == '9') || (reftype == 'H')) iwt = 1.;
	else if (reftype == 'A') iwt = 0.45;
	else if (('1' <= reftype) && (reftype <= '9'))
	  {
		i = reftype - '0';	fold = i; iwt = 0.11 * fold;
	  }
	else iwt = 0.5;	/* shouldn't ever happen! */
	xywt = (reftype == 'A') ? 0.05 : 0.;
	fold = rfp->rfr_fnow = initmin(iwt, xywt, initr, reftype, rfp);
	/* set up ^C signal routine so that ^Cs cause nonlinear
	 refinement to terminate with the best residual so far used */
	signal(SIGINT, (sig_t)qnonlin1);
/* adjust the parameters, finding the minimum difference between predicted
    and observed values.  The overall loop is entered nk times.  */
	outresid(initr, rmp, rfp);	/* report initial residual to CURSES */
	for (doshrink = 0, nk = ITER; nk > 0; )
	   {
		if (minph1(iwt, xywt, reftype, doshrink, rfp))
		  {
			usebest(rfp);	finishnl(rmp, rfp, reftype); return;
		  }
		if (rfp->rfr_fnow < fold)
		  {
			if (minph2(iwt, xywt, reftype, rfp))
			  {
				usebest(rfp);	finishnl(rmp, rfp, reftype);
				return;
			  }
			fold = rfp->rfr_fnow;	outresid(initr, rmp, rfp);
			doshrink = 0;
		  }
		/* residual is as good as we can get it for current values
		 of deltas.  Shrink them and loop through again. */
		else {
			nk--;	doshrink = 1;	outresid(initr, rmp, rfp);
		     }
	   }
	/* we've gone through the shrink loop as often as permitted, so
	 we're finished.  Reconvert the `xnow' variables to globals. */
	finishnl(rmp, rfp, reftype);	return;
}

static void statglob(MENUDATA *swp, REFINERUN *rfp)
{ /*	This saves the statistical information into a MENUDATA structure. */
	register int	i, j, m;
	WINDOW_ENTRY	*wep;

	SFL(4) = rfp->rfr_indrmse;	SFL(5) = rfp->rfr_xrmse;
	SFL(6) = rfp->rfr_yrmse;	SFL(7) = rfp->rfr_phirmse;
	SFL(8) = rfp->rfr_gamrmse;	SFL(9) = rfp->rfr_profrmse;
	SFL(10) = rfp->rfr_avedelfm;	SLG(11) = rfp->rfr_nreftot;
	SLG(12) = 3 * rfp->rfr_nreftot;
	SFL(13) = rfp->rfr_indurmse;	SFL(14) = rfp->rfr_xurmse;
	SFL(15) = rfp->rfr_yurmse;	SFL(16) = rfp->rfr_phiurmse;
	SLG(17) = rfp->rfr_nrefok;	SLG(18) = rfp->rfr_nindok;
	for (i = 0, j = 35; i < NSHELLS; i++, j += 2)
	   {
		SFL(19 + i) = rfp->rfr_shldmin[i];
		SFL(27 + i) = rfp->rfr_shlrmse[i];
		SLG(j) = rfp->rfr_shlnref[i];
		SLG(j + 1) = rfp->rfr_shlnind[i];
	   }
	for (i = 0; i < CATEGS; i++) for (j = 0, m = 52 + i; j < 3;
	 j++, m += 10) SLG(m) = rfp->rfr_ncateg[j][i];
	for (i = 0, wep = &(swp->wen[0]); i < 82; i++, wep++) valtostr(wep);
}

static void curse_peaks(int nnot, REFMRUN *rmp, REFINERUN *rfp)
{ /*	This prints some diagnostics to the CURSES menu when
	 the user reads in some centroids. */
	MENUDATA	*lcmp, *lpmp;

	wmove(rmp->rm_statp, 0, 0);	wclrtoeol(rmp->rm_statp);
	if ((NULL == (rfp->rfr_cname)) || ('\0' == *(rfp->rfr_cname)))
		wprintw(rmp->rm_statp, "No CENTROIDS file available\n");
	else {
		wprintw(rmp->rm_statp,
			"Reading in reflections from %s\n", rfp->rfr_cname);
		wmove(rmp->rm_statp, 1, 0);	wclrtoeol(rmp->rm_statp);
		wprintw(rmp->rm_statp,
		 "%6d usable and %6d unusable peaks found and sorted",
			(int)(rfp->rfr_npeaks), nnot);
	      }
	wrefresh(rmp->rm_statp);
	get_rmserrs("", rfp);/* compute rms errors in index, X, Y, phi */
	statglob(rmp->rm_statwp, rfp);	/* convert globals to `statwp' struct*/
	lcmp = rmp->rm_currp;	lpmp = rmp->rm_prevp;
	lpmp->wen[35].we_wvfloat = lcmp->wen[35].we_wvfloat; /* undo damage */
}

static void give_up(char *message, REFMRUN *rmp, REFINERUN *rfp)
{ /* This gets called when incremental lsq refinement fails */
	register int	i;
	REFREF	*lrp;

	partgl(rmp->rm_currp, rmp, rfp);	/* restore parameters */
	for (lrp = rfp->rfr_rpp0 + rfp->rfr_firstok, i = 0;
	 i < rfp->rfr_inbounds; i++, lrp++)
	   if (lrp->rflags < BADXYPHI) /* restore scatt vectors */
	     {
		rfp->rfr_rpp = lrp;	scattv(rfp);	refgonio(rfp);
	     }
	for (i = 0; i < ALLPARS; i++) rfp->rfr_pshift[i] = 0; /* zero shifts */
	gripe(message, 0, rmp, rfp);
}

static void dels_to_menu(REFMRUN *rmp, REFINERUN *rfp)
{ /*	This takes the parameter updates and sticks 'em in the visible menu */
	register int	i, j, m;
	float		*pshp;
	REFPARS		*xp;
	MENUDATA	*lmp;

	lmp = rmp->rm_tempp;	xp = rfp->rfr_xp;
		/* tone down huge changes & apply changes*/
	for (i = 0, pshp = rfp->rfr_pshift; i < ALLPARS - 3; i++, pshp++)
	   if (xp->xisref[i] == 'Y')
	     {
		if (i < NPAR0)
		  {
			j = i + NPAR0;	m = j + NPAR0;
		  }
		else {
			j = i + 2 * NPAR0 + NPAR1; m = j + NPAR1;
		     }
		lmp->wen[m].we_wvfloat = *pshp;	valtostr(&(lmp->wen[m]));
		lmp->wen[j].we_wvfloat += *pshp; valtostr(&(lmp->wen[j]));
	     }
	partgl(lmp, rmp, rfp); /* correct the global variables based on that */
}

static int incrlin(char resid_type, REFMRUN *rmp, REFINERUN *rfp)
{ /*	This sets up & performs a cycle of incremental least-squares refinement.
	In this version, both crystal and detector parameters are refined. */
	register int	i, j;
	int		nchanged;
	REFPARS		*xp;
	MENUDATA	*lcmp;

	xp = rfp->rfr_xp;
	if (EOK != make_incrvecs(rfp))
	  {
		give_up("memory problems for least-squares", rmp, rfp);
		return 0;
	  }
	for (lcmp = rmp->rm_currp, i = 0; i < lcmp->wnparams; i++)
		restore_value(rmp->rm_tempp, lcmp, i);	/* save parameters */
	ztos(xp->xstart, rfp->rfr_zsa);	/* construct matrix to rotate from
			(0,0,0) to the run's start position */
	if (incr_cycle(resid_type, rfp))
	  {
		give_up("not enough refinable refls", rmp, rfp); return 0;
	  }
	apply_dels(rfp);	/* get the revised params */
	dels_to_menu(rmp, rfp);	/* effect those changes ! */
	if (newresid(resid_type, &nchanged, rfp) < rfp->rfr_btb) /* better? */
	  {
		end_incr(rmp, rfp);	return 1; /* done with the cycle */
	  }
	else { /* no, it didn't get better: try some tricks to help out */
		partgl(lcmp, rmp, rfp); /* restore original params */
		if (nchanged)	/* no indices changed--try smaller deltas */
		  {
			if (incr_cycle(resid_type, rfp)) /* not enough refls */
			  {
				give_up("requested changes too large",rmp,rfp);
				return 0;
			  }
			/* try recomputing residual without the changed refls*/
			apply_dels(rfp); /* revised params */
			dels_to_menu(rmp, rfp); /* changes */
			if (newresid(resid_type, &nchanged,rfp) < rfp->rfr_btb)
			  { /* quit if new resid is < the best so far */
				end_incr(rmp, rfp);	return 1;
			  } /* quit! */
			else	partgl(lcmp, rmp, rfp);	/*restore orig params*/
		  }
		/* next trick: use proportionally smaller deltas */
		for (j = NSHRINKS; j--; ) /*shrink deltas several times*/
		   {
			for (i = rfp->rfr_npref; i--; )
				rfp->rfr_delt[i] *= SHRINKDEL;
			apply_dels(rfp);	/* revised params */
			dels_to_menu(rmp, rfp); /* changes */
			if (newresid(resid_type, &nchanged,rfp) < rfp->rfr_btb)
			  {
				end_incr(rmp, rfp);	return 1;
			  } /* quit! */
			partgl(lcmp, rmp, rfp);	/* restore orig params  */
		   }
		/*any time we drop out of this loop, the resid didn't improve*/
		give_up("Residual got worse after cycle!", rmp, rfp);
		return 0;
	     }
}

static void show_upars(FILE *fp, char *s, REFMRUN *rmp, REFINERUN *rfp)
{ /* This summarizes a cycle of refinement a la the menu to a file */
	REFPARS		*xp;
	MENUDATA	*lcp, *lsp;

	fprintf(fp,"Summary of parameters and performance %s\n", s);
	xp = rfp->rfr_xp;
	lcp = rmp->rm_currp;	put_pars(fp, lcp, 1);
	lsp = rmp->rm_statwp;	put_pars(fp, lsp, 1);
	pr3m(fp, xp->xareal, "Real-space unit-cell matrix"); xeflush(fp);
}

static void spotpar(WINDOW *win, WINDOW_ENTRY *wep)
{ /*	This places the cursor on a specific parameter of the window `win'.
	If the window entry `wep' is of type WST, we move to the beginning of
	that parameter; otherwise we move to its end.  Then we refresh `win'.*/

	wmove(win, (int)wep->we_y, effecx(wep));	wrefresh(win);
}

static void gltpar(MENUDATA *parp, REFINERUN *rfp)
{	/* this creates a MENUDATA structure from the global variables */
	register int	i, j, m;
	float		xymb[2];
	REFPARS		*xp;
	WINDOW_ENTRY	*wep;

	xp = rfp->rfr_xp;	/* remember the current REFPARS structure */
	for (i = NPAR0; i--; ) MPCH(i) = xp->xisref[i];
	for (i = 0, j = NPAR0, m = NPAR0 + 3; i < 3; i++, j++, m++)
	   {
		MFL(j) = xp->xabc[i];	MFL(m) = xp->xang[i];
	   }
	MFL(15) = xp->xdelom;	MFL(16) = xp->xdelchi;	MFL(17) = xp->xdelphi;
	for (i = 0, j = 2 * NPAR0; i < NPAR0; i++, j++)
		MFL(j) = rfp->rfr_pshift[i];
	for (i = NPAR0, j = 3 * NPAR0; i < (NPAR0 + NPAR1); i++, j++)
		MPCH(j) = xp->xisref[i];
	MFL(35) = xp->xplated;	MFL(36) = xp->xx0;	MFL(37) = xp->xy0;
	MFL(38) = xp->xtilt / RADC;		MFL(39) = xp->xswing / RADC;
	MFL(40) = xp->xgam[0] / fabs(xp->xfrmsize);
	MFL(41) = xp->xgam[1] / fabs(xp->xfrmsize);	MFL(42) = xp->xgam[2];
	for (i = NPAR0, j = 3 * NPAR0 + 2 * NPAR1; i < NPAR0 + NPAR1; i++, j++)
		MFL(j) = rfp->rfr_pshift[i];
	MFL(55) = xp->xerrind;	MFL(56) = xp->xerrx;
	MFL(57) = xp->xerry;	MFL(58) = xp->xerrphi;
	MFL(59) = xp->xresmin;	MFL(60) = xp->xresmax;
	MFL(61) = xp->xlambda;	MFL(62) = xp->xfrmsize / RADC;
	parp->wen[63].we_wvlong = xp->xisys;
	parp->wen[64].we_wvlong = xp->xispgp;
	for (i = 65; i < 68; i++) MFL(i) = xp->xstart[i - 65];
	getmainbeam(xp, xymb);	MFL(68) = xymb[0]; MFL(69) = xymb[1];
	parp->wen[70].we_string[0] = xp->xcmd;
	for (i = parp->wnparams, wep = &(parp->wen[0]); i--; wep++)
		valtostr(wep);
}

static void journal_command(char thisc)
{ /*	This records a command in the X-GEN journal buffer */
	static char	ljbuf[2];

	if (('a' <= thisc) && (thisc <= 'z')) thisc -= 'a' - 'A';
	/* special cases have been handled elsewhere */
	if ((thisc != 'I') && (thisc != 'B') && (thisc != 'T'))
	  {
		ljbuf[0] = thisc; ljbuf[1] = '\0'; (void)xgbuf(ljbuf, 1);
	  }
}

static void journal_parchange(int parn, char *oldst, REFMRUN *rmp)
{ /*	This records a change in parameter `par' */
	int		vbl;
	MENUDATA	*lcp, *lpp;
	WINDOW_ENTRY	*wep, *wepp;
	static char	ljbuf[86];
	static char *nfiel[] = {"A", "B", "C", "ALPHA", "BETA", "GAMMA",
				"OMEGA", "CHI", "PHI",
				"PLATED", "XCEN", "YCEN", "TILT", "SWING",
				"GAMMA0", "GAMMA1", "GAMMA2",
				"ERRORH", "ERRORX", "ERRORY", "ERRORPHI",
				"RESMIN", "RESMAX", "LAMBDA", "STEPSIZE",
				"SYSTEM", "SPACEGROUP",
				"STARTOMEGA", "STARTCHI", "STARTPHI"} ;

	lcp = rmp->rm_currp;	wep = lcp->wen + parn;
	lpp = rmp->rm_prevp;	wepp = lpp->wen + parn;
	if (strcmp(oldst, wep->we_string) || strcmp(oldst, wepp->we_string))
	  {
		if ((parn < 0) || (parn >= 68)) return;
		if (parn < 9) vbl = parn;
		else if (parn < 18) vbl = parn - 9;
		else if (parn < 35) vbl = parn - 18;
		else if (parn < 43) vbl = parn - 26;
		else if (parn < 51) vbl = parn - 34;
		else vbl = parn - 38;
		if (wep->we_type == WCH)
		 sprintf(ljbuf, "J %s %c", nfiel[vbl], wep->we_string[0]);
		else	sprintf(ljbuf, "D %s %s", nfiel[vbl], wep->we_string);
		(void)xgbuf(ljbuf, 1);
	  }
}

static void help_setup(REFMRUN *rmp)
{ /*	This sets up the help display */
	register int	i;
	static char	*helpstr[] = {
 "                       Meanings of the Control Keys:",
"<RIGHT>: next value   <CR>: next line    f1: perform command @: restore orig-",
"<DOWN>:  down 1 line  ^:    go to top    f2: print help           inal value",
"<LEFT>:  previous val $:    go to bottom f3: undo delta      #: restore all",
"<UP>:    prev line    ^L:   redraw menu  f4: undo all deltas    orig values",
"?:       define param %:    redraw stats <BS>,<DEL>: erase previous character",
"COMMANDS:I: autoindex; R: rocking-curve ref'mnt;  U: update;  T: transform;",
"L: linear refine on h; P: nonlinear ref on phi;   H: nonlinear ref on h;",
"A: nonlinear,all vbls; M: remap detector spline;  W: write params to disc;",
"E: write params&exit;  Q: quit without writing;   G: reread params from disc;",
"S: 3-refl'n indexing;  C: closed-form refine;   0-9: nonlinear ref, 0=P,9=H."
 } ;

	wmove(rmp->rm_statp, 0, 0);
	for (i = 0; i < 11; i++)
	   {
		wmove(rmp->rm_statp, i, 0);	wclrtoeol(rmp->rm_statp);
		waddstr(rmp->rm_statp, helpstr[i]);
	   }
	wrefresh(rmp->rm_statp);
}

static void define_par(int parno, REFMRUN *rmp)
{	/* this supplies a short definition for any parameter */
	int		msgno;
	MENUDATA	*lcp;
	static char	*pardefs[] = {
"a: Real-space unit cell dimension a in Angstroms",
"b: Real-space unit cell dimension b in Angstroms",
"c: Real-space unit cell dimension c in Angstroms",
"alpha: Real-space unit cell angle alpha in degrees",
"beta: Real-space unit cell angle beta in degrees",
"gamma: Real-space unit cell angle gamma in degrees",
"omega: pseudogoniostat angle corresponding to vertical Euler angle,in degrees",
"chi: pseudogoniostat angle corresponding to tilt of an axis, in degrees",
"phi: pseudogoniostat angle corresponding to spindle offset, in degrees",
"plateD: distance from the crystal to the detector in centimeters",
"xcen: x coordinate (in cm) of undiffracted beam at swing == 0",
"ycen: y coordinate (in cm) of undiffracted beam at swing == 0",
"tilt: rotation angle of detector in its own plane, in degrees",
"swing: angle betw. detector normal & undiffracted beam direction in degrees",
"Gam0: first rocking-curve parameter, in frames",
"Gam1: 2nd rocking curve parameter, in frames",
"Gam2: beam-divergence rocking curve param in radians",
"max index error: largest value of |h-[h]| allowed for inclusion in refinement",
"max X(cm) error: largest error in X in cm allowed for inclusion in refinement",
"max Y(cm) error: largest error in Y in cm allowed for inclusion in refinement",
"max phi(frame) error: largest error in phi in frames allowed for inclusion",
"Resolution Min: largest D spacing used during refinement",
"Resolution Max: smallest D spacing used during refinement",
"Wavelength: Mean X-ray wavelength used, in Angstroms",
"Stepsize: angular distance from one frame to the next, in degrees",
"System: 1=tricl.;2=monocl.;3=orthorh.;4=tetrag.;5=cubic;6=hexag.;7=rhomb.",
"Space Group: International Tables number designating spacegroup (1-230)",
"a direction: principal & 2ndry directions of real-space a (e.g.+X or -Z+Y)",
"b direction: principal & 2ndry directions of real-space b (e.g.+X or -Z+Y)",
"c direction: principal & 2ndry directions of real-space c (e.g.+X or -Z+Y)",
"command: operation to be performed on next `f1' (see help for choices)"
 } ;

	wmove(rmp->rm_statp, 10, 0);	/* go to bottom line */
	wclrtoeol(rmp->rm_statp);	/* blank that line */
	lcp = rmp->rm_currp;
	if (parno < 3 * NPAR0) msgno = parno % NPAR0;
	else if (parno < 3 * ALLPARS)
		msgno = NPAR0 + (parno - 3 * NPAR0) % NPAR1;
	else if (parno < lcp->wlastactive)
		msgno = parno - 2 * ALLPARS - 4;
	else	msgno = parno - 2 * ALLPARS - 6;
	waddstr(rmp->rm_statp, pardefs[msgno]);
	wrefresh(rmp->rm_statp);	/*	display it now */
	spotpar(rmp->rm_mwp, lcp->wen + parno);
}

static void sys_chk(int sys, REFMRUN *rmp)
{ /* this destroys refineability of params which can't be refined.
		It doesn't add refineability to anything. */
	register int	i;
	MENUDATA	*parp;
	WINDOW_ENTRY	*wep;

	parp = rmp->rm_currp;
	switch (sys)
	      {
		case	2:	MPCH(3) = MPCH(5) = 'N';
				MFL(12) = MFL(14) = 90.;	break;
		case	5:	MPCH(2) = 'N';
				MFL(11) = MFL(9);	/* falls through */
		case	4:	case	6:	MPCH(1) = 'N';
				MFL(10) = MFL(9);
		case	3:	for (i = 3; i < 6; i++)
				   {
					MPCH(i) = 'N'; MFL(i+9) = 90;
				   }
				if (sys == 6) MFL(14) = 120.;	break;
		case	7:	for (i = 1; i < 5; i++) MPCH(i) = 'N';
				MFL(11) = MFL(10) = MFL(9);
				MFL(12) = MFL(13) = MFL(14);	break;
		case	1:	default:	return;
	      }
	for (i = 9, wep = &(parp->wen[9]); i <= 50; i++, wep++)
	   if (wep->we_type == WFL)
	     {
		valtostr(wep);	disp_param(parp, rmp->rm_mwp, i);
	     }
}

static void menu_shifts(REFMRUN *rmp, REFINERUN *rfp)
{ /*	This compares the current parameters with previous cycle's MENUDATA
	structure and determines how much things have changed thereby. */
	int		i;
	float		*pshp;
	REFPARS		*xp;
	MENUDATA	*lpp;
	WINDOW_ENTRY	*wep;

	lpp = rmp->rm_prevp;	pshp = rfp->rfr_pshift;	xp = rfp->rfr_xp;
	for (i = 0, wep = &(lpp->wen[NPAR0]); i < 3; i++, wep++, pshp++)
		*pshp = xp->xabc[i] - wep->we_wvfloat;
	for (; i < 6; i++, wep++, pshp++)
		*pshp = xp->xang[i - 3] - wep->we_wvfloat;
	*pshp++ = xp->xdelom  - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xdelchi - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xdelphi - wep->we_wvfloat;	wep++;
	wep = &(lpp->wen[3 * NPAR0 + NPAR1]);
	*pshp++ = xp->xplated - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xx0 - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xy0 - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xtilt / RADC - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xswing / RADC - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xgam[0] / fabs(xp->xfrmsize) - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xgam[1] / fabs(xp->xfrmsize) - wep->we_wvfloat;	wep++;
	*pshp++ = xp->xgam[2] - wep->we_wvfloat;	wep++;
}

static void askconvert(REFMRUN *rmp, REFINERUN *rfp)
{ /*	Implement the cell transformation algorithm in the curses context.
	The list of legal cell transformations is given in curses format */
	int		convtype, i, j;
	char		incha;
	static char	ljbuf[4];
	static char	*chelp[] = {
 "a: primitive a <b <c   ; b: permute (abc)->(bca); c: permute (abc)->(cab);",
 "d: swap (abc)->(bac)   ; e: swap (abc)->(cba)   ; f: swap (abc)->(acb)   ;",
 "g: supplement alfa,beta; h: supplement alfa,gama; i: supplement beta,gama;",
 "j: use $REINDEX matrix ;" };

	for (i = 0, j = 2; i < 4; i++, j++)
	    {
		wmove(rmp->rm_statp, j, 0);	wclrtoeol(rmp->rm_statp);
		waddstr(rmp->rm_statp, chelp[i]);
	    }
	wmove(rmp->rm_statp, 1, 0);	wclrtoeol(rmp->rm_statp);
	wrefresh(rmp->rm_statp);
	wprintw(rmp->rm_statp, "Convert unit cell using the codes a-z: ");
	wmove(rmp->rm_statp, 1, 39);	wrefresh(rmp->rm_statp);
	incha = wget1ch(rmp->rm_statp);
	if (('a' <= incha) && (incha <= 'z')) incha -= 'a' - 'A';
	if ((incha < 'A') || (incha > 'Z')) incha = 'A';
	waddch(rmp->rm_statp, (xctype)incha);
	convtype = 1 + (int)(incha - 'A');
	sprintf(ljbuf, "T %c", (char)convtype);
	(void)xgbuf(ljbuf, 1);
	i = cellconvert(rfp->rfr_xp, convtype);
	wmove(rmp->rm_statp, 2, 0);	wclrtoeol(rmp->rm_statp);
	wprintw(rmp->rm_statp,
	 "Converting cell according to code %2d was %s\n",
	 convtype, (i? "unsuccessful" : "successful"));
	wrefresh(rmp->rm_statp);
}

static void bravmenu(REFMRUN *rmp, REFINERUN *rfp)
{ /* This implements the Bravais lattice calculations the curses context. */
	int		chara, choicel, i, j, listno, nlfound, wline;
	char		incha;
	float		cel[9];
	double		lar[3][3];
	WINDOW		*tewinp;
	LATTICESCORE	*lca0, *lcap;
	REFPARS		*xp;
	static char	ljbuf[16];

	for (i = 0; i < 11; i++)
	    {
		wmove(rmp->rm_mwp, i, 0);	wclrtoeol(rmp->rm_mwp);
		wmove(rmp->rm_statp, i, 0);	wclrtoeol(rmp->rm_statp);
	    }
	wrefresh(rmp->rm_mwp);	wrefresh(rmp->rm_statp);
	wmove(rmp->rm_statp, 1, 0);	wclrtoeol(rmp->rm_statp);
	wrefresh(rmp->rm_statp);
	for (xp = rfp->rfr_xp, i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		lar[i][j] = xp->xareal[i][j] * xp->xlambda;
	wmove(rmp->rm_mwp, 0, 0);	wprintw(rmp->rm_mwp,
		"Results of calculations using the 44 characters:");
	wmove(rmp->rm_mwp, 1, 0);	wprintw(rmp->rm_mwp, " in latt\n");
	wmove(rmp->rm_mwp, 2, 0);
	wprintw(rmp->rm_mwp, "dex char   score       a       b       c");
	wmove(rmp->rm_mwp, 3, 0);
	wprintw(rmp->rm_mwp, "  alpha   beta  gamma lattice symm cen\n");
	wrefresh(rmp->rm_mwp);
	if ((NULL == (lca0 = allbravais(lar, &nlfound))) || (nlfound < 16))
		return;
	showbravais(rfp->rfr_fpl, xp->xareal, xp->xstart); /* to logfile */
	for (i = 0, lcap = lca0; i < 45; i++, lcap++)
	   if (lcap->lc_choice >= 0)
	     {
		listno = lcap->lc_choice;
		chara = lcap->lc_latchar;
		artocel(xp->xstart, lcap->lc_cel, cel);
		if (listno < 9)
		  {
			tewinp = rmp->rm_mwp; wline = listno + 3;
		  }
		else {
			tewinp = rmp->rm_statp; wline = listno - 9;
		     }
		wmove(tewinp, wline, 0);
		wprintw(tewinp,
	 "  %c %4d %7.3f %7.2f %7.2f %7.2f %6.1f %6.1f %6.1f %s %s",
		 (char)(listno + 'a'), chara, 100. * lcap->lc_score,
		 cel[0], cel[1], cel[2], cel[3], cel[4], cel[5],
		 glattice(chara), gcentering(chara));
	     }
	wmove(rmp->rm_statp, 10, 0);
	wrefresh(rmp->rm_mwp);	wrefresh(rmp->rm_statp);
	if (i >= 45) choicel = 0;
	wmove(rmp->rm_statp, 8, 0);
	wprintw(rmp->rm_statp, "Enter index value of desired lattice (a-p): ");
	wrefresh(rmp->rm_statp);	wmove(rmp->rm_statp, 8, 46);
	wrefresh(rmp->rm_statp);	incha = wget1ch(rmp->rm_statp);
	if (('a' <= incha) && (incha <= 'z')) incha -= 'a' - 'A';
	if ((incha < 'A') || (incha > 'P')) incha = 'A';
	waddch(rmp->rm_statp, (xctype)incha);
	choicel = (int)(incha - 'A');
	if ((choicel < 0) || (choicel > 15)) choicel = 0;
	wmove(rmp->rm_statp, 2, 0);	wclrtoeol(rmp->rm_statp);
	wrefresh(rmp->rm_statp);
	wmove(rmp->rm_statp, 9, 0);
	wprintw(rmp->rm_statp, "Should this symmetry be fully imposed? [n] ");
	wrefresh(rmp->rm_statp);	wmove(rmp->rm_statp, 9, 44);
	wrefresh(rmp->rm_statp);
	incha = wget1ch(rmp->rm_statp);
	if (('a' <= incha) && (incha <= 'z')) incha -= 'a' - 'A';
	if ((incha < 'A') || (incha > 'Z')) incha = 'N';
	chara = lca0->lc_latchar;
	for (i = 0, lcap = lca0; i < 45; i++, lcap++)
	 {
	   if (choicel == lcap->lc_choice) /* determine which choice this is*/
	     {
		chara = lcap->lc_latchar;
		sprintf(ljbuf, "F %d", (int)(-chara));	/* journal it */
		(void)xgbuf(ljbuf, 1);	break;
	     }
	 }
	if (i >= 45) return;
	fprintf(rfp->rfr_fpl,
	 "Bravais choice %c chosen: raw score %9.5f, figure of merit %9.5f\n",
		'a' + choicel, 100. * lcap->lc_score, lcap->lc_fom);
	show_bravchoice(rfp->rfr_fpl, xp->xstart, lcap);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xp->xareal[i][j] = lcap->lc_cel[i][j] / xp->xlambda;
	if ('Y'== incha)	/* impose symmetry if requested */
	  {
		impose_symmetry(rfp->rfr_xp, chara);
		wmove(rmp->rm_mwp, 3, 0);	wclrtoeol(rmp->rm_mwp);
		wprintw(rmp->rm_mwp,
	 "imposing symmetry associated with xtal system %1d, spacegroup %3d\n",
		 (int)(xp->xisys), (int)(xp->xispgp));
		wrefresh(rmp->rm_mwp);
	  }
	(void)rp2vecs(xp);
	for (i = 0; i < 3; i++)
	   {
		wmove(rmp->rm_mwp, i, 0);	wclrtoeol(rmp->rm_mwp);
	   }
	wprintw(rmp->rm_mwp, "%s %s lattice chosen",
		glattice(chara), gcentering(chara));
	wrefresh(rmp->rm_mwp);
	sleep(3);
}

static void autoindex(REFMRUN *rmp, REFINERUN *rfp)
{ /*	Controlling-routine for curses-based auto-indexing */
	register int	i;
	int		saveisys;
	REFPARS		*xp;
	double		basangle, errv[2];
	double		err_allowed, errspec;
	time_t		tla;
	static char	ljbuf[32];

	rfp->rfr_flags |= RR_CELLLIMITS;
	xp = rfp->rfr_xp; /* remember the REFPARS structure */
	err_allowed = 0.2;
	tla = NOW;	strcpy(rfp->rfr_rundate, ctime(&tla));
	wmove(rmp->rm_statp, 0, 0);	wclrtobot(rmp->rm_statp);
	wprintw(rmp->rm_statp,"Auto-Indexing Attempted %s",
		rfp->rfr_rundate);
	saveisys = xp->xisys;	xp->xisys = 1;	/* autoindex in P1 */
	errspec = menu_number(
	 "Enter fractional error in cell lengths allowed", rmp);
	if (DFUZZ < errspec && errspec <= 1) err_allowed = errspec;
	if (err_allowed >= (1. - DFUZZ))
	  {
		err_allowed = 1.;	rfp->rfr_flags &= ~RR_CELLLIMITS;
	  }
	if (rfp->rfr_flags & RR_CELLLIMITS) for (i = 3; i--; )
	  {
		rfp->rfr_amin[i] = xp->xabc[i] * (1. - err_allowed);
		rfp->rfr_amax[i] = xp->xabc[i] * (1. / (1. - err_allowed));
		errv[0] = err_allowed * 100.;
	  }
	else errv[0] = 90.;
	errspec = menu_number(
		"\nEnter fractional error in cell angles  allowed", rmp);
	if (DFUZZ < errspec && errspec <= 1) err_allowed = errspec;
	else err_allowed = 0.1;
	errv[1] = err_allowed * 100.;
	if (rfp->rfr_flags & RR_CELLLIMITS) for (i = 3; i--; )
	  {
		rfp->rfr_angmin[i] = xp->xang[i] - 90. * err_allowed;
		if (rfp->rfr_angmin[i] < 0) rfp->rfr_angmin[i] = 0;
		rfp->rfr_angmax[i] = xp->xang[i] + 90. * err_allowed;
		if (rfp->rfr_angmax[i] > 180) rfp->rfr_angmax[i] = 180;
	  }
	basangle = DEFAULT_BASISANGLE;
	basangle = menu_number
	 ("\nMin Angle Between Basis Vectors (default=50 deg)", rmp);
	if ((basangle < DFUZZ) || (basangle > 89.9))
		basangle = DEFAULT_BASISANGLE;
	sprintf(ljbuf, "I %7.4lf %7.4lf %7.4lf\n", errv[0], errv[1], basangle);
	(void)xgbuf(ljbuf, 1); /* restore the curses environment */
	wclear(rmp->rm_mwp);	wclear(rmp->rm_statp);
	wrefresh(rmp->rm_mwp);	wrefresh(rmp->rm_statp);
	clear();	refresh();	nocbreak();	nl();
	xeflush(stdout);
	(void)indhkl(errv, basangle, rfp); /* calculate residuals */
	cbreak();	nonl(); /* clean up curses yet again */
	guess_query(rmp, rfp);	/* see if user wants one of these solutions */
	signal(SIGINT, (sig_t)end_menu);
	xp->xisys = saveisys;	/* restore original `isys' value */
	rfp->rfr_flags |= RR_NEEDHEADER; /* we need a header after all this */
}

static void refcommand(char comd, REFMRUN *rmp, REFINERUN *rfp)
{ /* This interprets the command itself; it's called on an `f1' */
	register int	i, j;
	int		nnot, success;
	float		oldga[3];
	VSPLINE		*lsp;
	REFPARS		*xp;
	MENUDATA	*lcp, *lpp;
	WINDOW_ENTRY	*wep;

	wmove(rmp->rm_mwp, 0, 0);	wclrtoeol(rmp->rm_mwp);
	rfp->rfr_tl0 = NOW;
	strcpy(rfp->rfr_rundate, ctime(&(rfp->rfr_tl0)));
	rfp->rfr_rundate[24] = '\0';
	sprintf(rmp->rm_headstr, "XGEN refinement performed %s",
		rfp->rfr_rundate);
	/* enforce the crystal-system requirements */
	lcp = rmp->rm_currp;	lpp = rmp->rm_prevp;
	sys_chk((int)(lcp->wen[63].we_wvlong), rmp);
	partgl(lcp, rmp, rfp);		/* convert params to globals */
	/* see if any detector parameters have changed; set flag accordingly */
	for (i = 0; i < ALLPARS; i++) /* clear all the delta values */
	   {
		j = (i < NPAR0 ? i + 2 * NPAR0 : i + 2 * ALLPARS);
		rfp->rfr_pshift[i] = 0;	wep = lcp->wen + j;
		wep->we_wvfloat = 0.;	valtostr(lcp->wen + j);
	   }
	for (i = 0; i < lpp->wnparams; i++) restore_value(lpp, lcp, i);
	if ((comd != 'W') && (comd != 'E') && (comd != 'Q') && (comd != 'S'))
	  {
		/* read in the refls if they're not there yet*/
		if (!(rfp->rfr_npeaks))
		  {
			if (-1 == (nnot = gpeaks(rfp))) /* read 'em in */
			  gripe("cannot rewind centroids to read refls in",
				2, rmp, rfp);
			curse_peaks(nnot, rmp, rfp); /* tell window about em */
		  }
			/* set up the stats window */
		menu_setup(rmp->rm_statwp, rmp->rm_statp, "", 0);
	  }
	success = 1;	/* most commands are `successful' */
	lsp = rfp->rfr_spl;	xp = rfp->rfr_xp;
	switch (comd)
	   {
		case	'A':	/* nonlinear refinement, using all resids */
		case	'P':	/* nonlinear refinement, using Phi residual */
		case	'H':	/* nonlinear refinement on integerness */
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			minpar(comd, rmp, rfp);	break;
		case	'B':	/* Bravais lattice calculation */
			bravmenu(rmp, rfp);	break;
		case	'C':	/* closed-form linear refinement */
			get_rmserrs("before closed-form", rfp);
			if (ref3p(rfp)) gripe(
			 "cannot perform closed-form refinement", 0, rmp, rfp);
			else {
				menu_shifts(rmp, rfp);
				get_rmserrs("after closed-form", rfp);
			     }
			break;
		case	'E':	/* write and exit */
			if (globtxs(3, rfp)) gripe(
		 "Warning: Failed to recalc params on exit\n", 0, rmp, rfp);
			show_upars(rfp->rfr_fpl,
			 "upon writing out parameter file", rmp, rfp);
			end_menu();	rfp->rfr_fpl = NULL;	return;
		case	'G':	/* re-read params from file and do "U" */
			success = !readpars(rfp);	break;
		case	'I':	/* automatic indexing */
			autoindex(rmp, rfp);	wmove(rmp->rm_mwp, 1, 0);
			wclrtobot(rmp->rm_mwp); get_rmserrs("", rfp);	break;
		case	'K':	/* linear refinement on scanning angle */
			success = incrlin(comd, rmp, rfp);	break;
		case	'L':	/* incremental linear refinement */
			success = incrlin(comd, rmp, rfp);	break;
		case	'M':	/*modify mapping from pix to cm & cm to pix*/
			get_rmserrs("before remapping", rfp); /*initial stats*/
			wmove(rmp->rm_mwp, 0, 0);	wclrtobot(rmp->rm_mwp);
			wrefresh(rmp->rm_mwp);
			nocbreak();	nl();	xeflush(stdout);
			if (remap_distortion(rfp))
			  {
				cbreak();	nonl();
				wmove(rmp->rm_mwp, 12, 0);
				wprintw(rmp->rm_mwp,
					"Failure in remapping positions");
			  }
			else {
				cbreak();	nonl();
				wmove(rmp->rm_mwp, 11, 0);
				wprintw(rmp->rm_mwp,
 " %4d fiducial points remapped; rms shift = %7.3f, max shift = %7.3f\n",
				 lsp->v_ntrue,lsp->v_rmsshift,lsp->v_maxshift);
				wrefresh(rmp->rm_mwp); waddstr(rmp->rm_mwp,
		 "Adjustment of the spatial distortion mapping complete");
			     }
			wrefresh(rmp->rm_mwp);	sleep(3); /* so we can see it*/
			wmove(rmp->rm_mwp, 0, 0); wclrtobot(rmp->rm_mwp);
			wrefresh(rmp->rm_mwp);	break;
		case	'O':	/* output refl list to logfile */
			showpeaks(rfp->rfr_fpl,"during refinement", rfp);
			return;
		case	'Q':	/* don't write, but exit */
			end_menu();	rfp->rfr_fpl = NULL;	return;
		case	'R':	/* rocking-curve refinement */
			wep = &(lcp->wen[40]);
			for (i = 0; i < 3; i++, wep++)
				oldga[i] = wep->we_wvfloat;
			/* July 02: failure in rocking-refinement non-fatal */
			success = !rocking_ref(oldga, 1, rfp);
			if (success)
			  {
				wep = &(lcp->wen[48]);
				for (i = 14; i < 17; i++, wep++)
				   if (xp->xisref[i]  == 'Y')
					wep->we_wvfloat = rfp->rfr_pshift[i];
			  }
			else success = 1;	/* clear error */
			break;
		case	'S':	/* specify refls by h, k, l, X, Y, phi */
			if (get_hkl(rfp)) /* failure: ignore results */
			  {
				gripe(
	"Error in attempt to obtain matrix from REFERENCE file", 0, rmp, rfp);
				return;
			  }
			menu_shifts(rmp, rfp); /* compare globals to MENUDATA*/
			/* (re-)read reflections file */
			if (-1 == (nnot = gpeaks(rfp))) /* read 'em in */
			  gripe("cannot rewind centroids to read refls in",
				1, rmp, rfp);
			curse_peaks(nnot, rmp, rfp); /* tell window about em */
			show_upars(rfp->rfr_fpl,
			 "before any refinement", rmp, rfp);
			get_rmserrs("after rereading file", rfp); success = 1;
			break;
		case	'T':	/* transform unit cell */
			askconvert(rmp, rfp);	break;
		case	'U': /* no refinement: just update stats */
			get_rmserrs("", rfp);	break;
		case	'W':	/* write out parameters but don't exit*/
			if (globtxs(3, rfp)) gripe(
		 "Warning: Failed to recalc params on exit\n", 0, rmp, rfp);
			return;
		case	'X':	case	'Y':	case	'Z': /*xyomega resid */
			success = incrlin(comd, rmp, rfp); break;
		default:	return;	/* don't do anything else */
	   }
	gltpar(lcp, rfp);	/* convert globals to current-parameter list */
	statglob(rmp->rm_statwp, rfp);	/* ... do same thing with the stats */
	menu_setup(rmp->rm_statwp, rmp->rm_statp, "", 1); /* stats display */
	if (success) /*on failed refinements we leave the diagnostic showing*/
	  {
		wmove(rmp->rm_mwp, 0, 0);
		wclrtoeol(rmp->rm_mwp); /*erase diagnostics in mwp*/
		menu_setup(lcp, rmp->rm_mwp, rmp->rm_headstr, 0); /* display*/
	  }
	else {
		for (i = 0; i < ALLPARS; i++)
		   {
			j = (i < NPAR0 ? i + 2 * NPAR0 : i + 2 * (ALLPARS));
			rfp->rfr_pshift[i] = lcp->wen[j].we_wvfloat = 0.;
			valtostr(lcp->wen +j); disp_param(lcp, rmp->rm_mwp, j);
		   }
	     }
	putc((char)07, stdout);	/* ring the bell */
}

static void killit(char *s)
{ /* this exits gracefully, but with a core dump and a message*/
	REFINERUN	*rfp;
	REFMRUN		*rmp;

	if (NULL != (rfp = managerfp(3, 0)))
	  {
		if (NULL == rfp->rfr_fpl)
		  {
			fprintf(rfp->rfr_fpl,"\n\n%s: exiting\n", s);
			xeflush(rfp->rfr_fpl);	rfp->rfr_fpl = NULL;
		  }
		rfp = managerfp(0, 0);
	  }
	if (NULL != (rmp = managermp(3)))
	  {
		if (NULL != rmp->rm_statp)
		  {
			wmove(rmp->rm_statp, 10, 0); wclrtoeol(rmp->rm_statp);
			wprintw(rmp->rm_statp,"%s: exiting", s);
			wrefresh(rmp->rm_statp);
		  }
		rmp = managermp(0);
	  }
	endwin();
#ifdef VMS
	exit(EBAD);
#else
	abort();
#endif
}

static void bus(void)
{
	killit("Bus Error");
}

static void fpe(void)
{
	killit("Floating-point Exception");
}

static void illi(void)
{
	killit("Illegal instruction");
}

static void segv(void)
{
	killit("Segmentation Violation");
}

static void catchsigs(void)
{ /* smooth handling of the common interrupting signals */
	signal(SIGINT, (sig_t)end_menu);	signal(SIGILL, (sig_t)illi);
	signal(SIGBUS, (sig_t)bus);		signal(SIGSEGV, (sig_t)segv);
	signal(SIGFPE, (sig_t)fpe);
}

static int refcur_init(REFINERUN *rfp)
{ /* This initializes the CURSES-related stuff for `refine' */
	int		i, par;
	MENUDATA	*lcp;
	WINDOW_ENTRY	*wep;
	REFMRUN		*rmp;

	if (NULL == (rmp = managermp(01777))) return EALLOC;
	for (i = 0, wep = &((rmp->rm_origp)->wen[0]);
	 i < 3 * (NPAR0 + NPAR1); i++, wep++) valtostr(wep);
#ifndef VMS	/* set up symbols for menus */
	if (verror(setmore(0),
	 "set up keystroke equivalences for this terminal"))
	  {
		if ((NULL != rfp->rfr_fpl) && (stdout != rfp->rfr_fpl))
			(void)fclose(rfp->rfr_fpl);
		rfp->rfr_fpl = NULL;	return EOPENERR;
	  }
#endif
	gltpar(rmp->rm_origp, rfp); /* copy globals to "MENUDATA" structure*/
	statglob(rmp->rm_statwp, rfp);	/* ... same idea but for the stats */
	catchsigs();
	sprintf(rmp->rm_headstr, "XGEN V.%s Processing: Run of refine on %s",
		XGEN_VERSION, rfp->rfr_rundate);
	menu_setup(rmp->rm_origp, rmp->rm_mwp, rmp->rm_headstr, 0);
	menu_setup(rmp->rm_statwp, rmp->rm_statp, "", 0);
	help_setup(rmp);
	for (par = 0; par < (rmp->rm_origp)->wnparams; par++)
	   {
		restore_value(rmp->rm_currp, rmp->rm_origp, par);
		restore_value(rmp->rm_prevp, rmp->rm_origp, par);
	   }
	lcp = rmp->rm_currp;	wep = &(lcp->wen[lcp->wlastactive]);
	wmove(rmp->rm_mwp, (int)wep->we_y, (int)wep->we_x0);
	wrefresh(rmp->rm_mwp);	return EOK;
}

static int crefexit(int code)
{ /* this exits from `refmain' cleanly after freeing memory allocated in
   `setmore' and various other places*/
	setmzap();	(void)managermp(0);	(void)managerfp(0, 0);
	globalkill();	return code;
}

/*ARGSUSED*/
int refmain(int argc, char *argv[])
{ /*	This sets up the menu for refinement.  It fetches the current set of
	parameters from the parameter file and then runs the menu on them.
	Whenever the user types an F1, an actual command will be executed--
	rocking curve refinement, linear parameter refinement,
	nonlinear parameter refinement, write-to-file, or exit. */
	int		i, newpar, par, refla;
	char		thiscmd, locjb[80];
	MENUDATA	*lcp;
	WINDOW_ENTRY	*wep;
	REFINERUN	*rfp;
	REFMRUN		*rmp;
	static pid_t	currpid;

	if (0 == (int)currpid) currpid = getpid(); /* remember process ID */
	(void)flagproc(argc, argv, (char *)NULL);
	sprintf(locjb, "refine1 << EOF_%d\n", (int)currpid);
	if (EOK != xgbuf(locjb, 1)) return EBAD;	refla = 0;
	if (EOK != (i = refinit(refla))) /* initial file I/O and memory */
	  {
		fprintf(stderr, "Failure in starting `refine': code %d\n", i);
		return i;
	  }
	if (NULL == (rfp = managerfp(7, 0))) return crefexit(EALLOC);
	if (EOK != (i = refcur_init(rfp))) /*start up the CURSES menus*/
	  {
		if ((NULL != rfp->rfr_fpl) && (stdout != rfp->rfr_fpl))
			(void)fclose(rfp->rfr_fpl);
		fprintf(stderr,
		 "Failure in starting curses info for `refine': code %d\n", i);
		return crefexit(i);
	  }
	/* main loop through user's keystrokes: */
	rmp = managermp(3);
	for (lcp = rmp->rm_currp, par = lcp->wlastactive; par >= 0; )
	   {
		if (rfp->rfr_flags & RR_QUITREFINE) break;
		newpar = getans(lcp, rmp->rm_prevp, rmp->rm_mwp, par,
			rmp->rm_headstr);
		if (rfp->rfr_flags & RR_QUITREFINE) break;
		wep = lcp->wen + par; thiscmd = lcp->wen[70].we_string[0];
		if (K_PF1 == newpar)
		  {
			refcommand(thiscmd, rmp, rfp);
			journal_command(thiscmd);
			if (rfp->rfr_flags & RR_QUITREFINE) break;
		  }
		else if (K_PF2 == newpar)
		  {
			help_setup(rmp); spotpar(rmp->rm_mwp, wep);
		  }
		else if (K_HELP == newpar) define_par(par, rmp);
		else if (K_INSERT_HERE == newpar) /*restore orig value of par*/
		  {
			restore_value(lcp, rmp->rm_origp, par);
			if ((par < 3 * ALLPARS) && (wep->we_type == WFL))
			  {
				i = par + (par < 3 * NPAR0 ? NPAR0 : NPAR1);
				wep = lcp->wen + i;	wep->we_wvfloat = 0.;
				valtostr(wep); disp_param(lcp, rmp->rm_mwp, i);
			  }
			disp_param(lcp, rmp->rm_mwp, par);
			wrefresh(rmp->rm_mwp);
		  }
		else if (K_PREV_SCREEN == newpar) /* restore all */
		  {
			for (newpar = 0; newpar < (rmp->rm_origp)->wnparams;
			 newpar++)
			   {
				restore_value(lcp, rmp->rm_origp, newpar);
				disp_param(lcp, rmp->rm_mwp, newpar);
			   }
			par = lcp->wlastactive; wep = &(lcp->wen[par]);
			wmove(rmp->rm_mwp, (int)wep->we_y, (int)wep->we_x0);
			wrefresh(rmp->rm_mwp);
		  }
		else	/* ordinary-entry case */
		  {
			wep = lcp->wen + par;
			strcpy(locjb, wep->we_string);
			update_par(lcp, rmp->rm_mwp, par);
			journal_parchange(par, locjb, rmp);
			par = newpar;
			wep = lcp->wen + par;	spotpar(rmp->rm_mwp, wep);
		  }
	   }
	sprintf(locjb, "EOF_%d\n", (int)currpid);
	(void)xgbuf(locjb, 7); /* write out to journal buffer and flush it */
	end_refinement();
	return crefexit(EOK);
}

/* end of refmain.c */
