/* calibdefs.h -- structure types for X-GEN calibration

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	 7 Jun 2006: eliminate luycmp, which is now static
	12 May 2006: localize a bunch of stuff in calibmain and ucautils
	28 Oct 2005: add additional variables
	16 Aug 2005: #defines for DEFRASTER and CF_* moved here
	21 Jan 2004: license message inserted
	28 Dec 2003: add more elements to CALIBRUN structure
	23 Nov 2003: converted global variables from ../prerefine/calglob.c

   Purpose: structure type for calibration, i.e.,
	the mapping from pixels to cm on the detector face and vice versa.
   Cross references: ../prerefine/cal*.c

*************************************************************************/
#ifndef CALIBDEFS_H
#define CALIBDEFS_H	1
#include	"xgen.h"
#define		DEFRASTER	(0.508)	/* dist.between fiducial points in cm*/
	/* bits in `runflags' for calibration algorithms */
#define	CF_HEXAGGRID	1	/* hexagonal, not rectangular, grid */
#define	CF_NOEXTRAP	2	/* don't extrapolate outside centroids */
#define	CF_POLARODD	4	/* handles polarity of hexagonal grid */
#define	CF_READCEN	010	/* read in centroids from ascii file */
#define	CF_USEBORDER	020	/* only use data inside AMASK region */
#define	CF_VERBOSE	040	/* verbose logfile and stdout output */
#define	CF_WRITECEN	0100	/* write centroids to ascii file */
#define	CF_INDEXED	0200	/* ascii centroid I/O includes indexes */
#define	CF_ASCIICEN	(CF_READCEN | CF_WRITECEN | CF_INDEXED)

 typedef struct {
		char	ar_known;
		short	ar_npts, ar_nobs;
		double	ar_pe2, ar_pe1, ar_pe0, ar_resid;
		} PARABOLA;

 typedef struct {
		xglong		car_cencol, car_cenrow, car_dhwidth;
		xglong		car_dhheight, car_flags, car_imflags;
		xglong		car_imht, car_imwid, car_linear, car_maxrows;
		xglong		car_maxcols, car_minlim, car_nmbox;
		xglong		car_specht, car_specwid;
		xgulong		car_ntroids;
		time_t		car_tl0, car_tl1;
		char		*car_cdname;
		double		car_beamx, car_beamy, car_imposespacing;
		double		car_idarkfrac, car_mincordist, car_minradsq;
		double		car_overcut, car_scalxy, car_sigmult;
		double		car_startspacing, car_usedrx, car_usedry;
		FILE		*car_fpl;
		VSPLINE		*car_spl;
		MAIL		*car_heads;
		CTROID		*car_ctr, *car_ctr0, *car_ctend, *car_g0;
		CTROID		*car_mid, *car_next, *car_prev, *car_unindex0;
		PARABOLA	*car_rowpara, *car_colpara;
		} CALIBRUN;

/* function prototypes */

extern int calfill(CALIBRUN *cap);
extern int calibmain(int argc, char *argv[]);
extern int ext_troids(int *nindexed, CALIBRUN *cap);
extern int outindcen(VSPLINE *lsp);
extern int recenter(VSPLINE *lsp, double bex, double bey);
extern void calindex(CALIBRUN *cap);
extern void list_troids(FILE *fp, CTROID *ctb, unsigned ntr, char *str);

#endif /* CALIBDEFS_H */

/* end of calibdefs.h */
