/* displays.h -- definitions and prototypes for displays

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	27 Nov 2006: eliminate di_q21[3][3] from DISP_ATTRS structure
	22 Jun 2006: move DISP_ATTRS and XG_Point structure typedefs here;
		#defines and function prototypes carry over with those
	12 Jun 2006: reorder declarations in a rational way
	20 Nov 2005: initial version
 Changes to xgen.h relevant to the current contents of this file:
	10 Sep 2005: X-GEN 5.5.5: histogram-binned display colorings;
		DISP_ATTRS structure typedef modified accordingly
	 3 Sep 2004: X-GEN 5.4.0: add di_maxw and di_maxh to DISP_ATTRS
	16 Aug 2004: di_lar and di_lpcor elements added to DISP_ATTRS struct;
		byte counts for DISP_ATTRS structure included in comments
	 5 Jan 2004: additional elements in DISP_ATTRS structure type
	13 Sep 2003: reorganize DISP_* defines in a simple pattern
	25 Aug 2003: add DISP_ATTRS and XG_Point structure typedefs
	18 Aug 2003: add DISP_* #defines for display properties

   Purpose: primarily to handle function prototypes for display functions
   Cross References: ../displays/ *.c, ../motif/ *.c

********************************************************************/
#ifndef DISPLAYS_H
#define DISPLAYS_H	1

#include	"xgen.h"
	/* bits in the display-flag variable */
#define	DISPCALIB_NEED	01
#define	DISPCALIB_IN	02
#define	DISPCALIB_CONT	04
#define	DISPCALIB_CALC	(DISPCALIB_IN | DISPCALIB_CONT)
#define	DISPBORDER_NEED	010
#define	DISPBORDER_IN	020
#define	DISPBORDER_CONT	040
#define	DISPBORDER_CALC	(DISPBORDER_IN | DISPBORDER_CONT)
#define	DISPHKL_NEED	0100
#define	DISPHKL_IN	0200
#define	DISPHKL_CONT	0400
#define	DISPHKL_CALC	(DISPHKL_IN | DISPHKL_CONT)
#define	DISPCALCEN_NEED	01000
#define	DISPCALCEN_IN	02000
#define	DISPCALCEN_CONT	04000
#define	DISPCALCEN_CALC	(DISPCALCEN_IN | DISPCALCEN_CONT)
#define	DISPCEN_NEED	010000
#define	DISPCEN_IN	020000
#define	DISPCEN_CONT	040000
#define	DISPCEN_CALC	(DISPCEN_IN | DISPCEN_CONT)
#define	DISPRESOL_NEED	0100000
#define	DISPRESOL_IN	0200000
#define	DISPRESOL_CONT	0400000
#define	DISPRESOL_CALC	(DISPRESOL_IN | DISPRESOL_CONT)
#define	DISPROWCOL_NEED	01000000
#define	DISPROWCOL_IN	02000000
#define	DISPROWCOL_CONT	04000000
#define	DISPROWCOL_CALC	(DISPROWCOL_IN | DISPROWCOL_CONT)
/* other display flags -- in hex so we don't type too many zeroes */
#define	DISPVERBOSE1	0x200000
#define	DISPVERBOSE2	0x400000
#define	DISPVERBOSEANY	0x600000

typedef struct { /* attributes of a display: bytecount assumes 4-byte ptrs */
		xglong	di_disflags, di_distype, di_imflags;	/* 12 bytes */
		short	di_wintype, di_wid, di_ht, di_nranges;	/*  8 bytes */
		short	di_widown, di_wiup, di_htdown, di_htup;	/*  8 bytes*/
		short	di_imwid, di_imht, di_boxh, di_nbperp;	/*  8 bytes*/
		short	di_magw, di_magh;			/*  4 bytes */
		short	di_minw, di_minh, di_roco, di_fmno;	/*  8 bytes */
		short	di_maxw, di_maxh, di_xylim[4];		/* 12 bytes */
		short	di_currx, di_curry, di_disx, di_disy;	/*  8 bytes */
		xglong	di_maxv, di_nmbox, di_darkfla;		/* 12 bytes */
		double	di_lpcor, di_xsca, di_ysca, di_valmul;	/* 32 bytes */
		float	di_lar[3][3], *di_limi;			/* 40 bytes */
		ino_t	di_curino;				/*  4 bytes */
		xglong	*di_range, *di_regb, *di_hist;		/* 12 bytes */
		MAIL	*di_head;				/*  4 bytes */
		REFPARS	*di_xp;					/*  4 bytes */
		VSPLINE	*di_spl;				/*  4 bytes */
		FRAMEINDEX	*di_fmind;			/*  4 bytes */
		CRYSTAL	*di_xhdp;				/*  4 bytes */
		DATARUN	*di_rhdp;				/*  4 bytes */
		char	*di_imaname;				/*  4 bytes */
		char	*di_imm, *di_magm, *di_rotm;		/* 12 bytes */
		} DISP_ATTRS;				/* Total: 244 bytes */

typedef struct { /* This mimics an XPoint structure in a non-X environment */
		short	xgx, xgy;
		} XG_Point;

typedef struct {	/* structure for manipulating UREFLS file displays*/
		short	uro_hkl[3];
		double	uro_ii, uro_sig, uro_xob, uro_yob, uro_omob;
		double	uro_stl, uro_delx, uro_dely, uro_delz;
		} UROBJECT;

typedef struct { /* structure for handling groups of UREFLS file displays */
		int		urw_hlim[6];
		UROBJECT	*urw_val, *urw_hv;
		} URWORK;

typedef struct {
		short	mro_hkl[3];
		double	mro_ii, mro_sig, mro_xob, mro_yob;
		double	mro_omob, mro_iscal, mro_ibar, mro_stl;
		} MROBJECT;

extern int *calib_contours(DISP_ATTRS *dispatt, int *ncaptp);
extern int check_uh(char *uname, short *minh, short *maxh);
extern int dget_hkl(DISP_ATTRS *dispatt, double zhkl[4]);
extern int exact_inten(short cuhkl[], double *iv);
extern int exact_mrxy(DISP_ATTRS *dispatt, short *cuh, double *iv);
extern int exact_xy(DISP_ATTRS *dispatt, short *cuh, double iv[]);
extern int exam_mref(char *mname, int whichstat, DISP_ATTRS *dispatt);
extern int exam_urefl(char *uname, int whichstat, DISP_ATTRS *dispatt);
extern int glocave(DISP_ATTRS *dispatt, int  *nregvp);
extern int gpars(DISP_ATTRS *dispatt);
extern int grab_image(int darkfla, DISP_ATTRS *dispatt);
extern int imhisto(DISP_ATTRS *dispatt, int force);
extern int locba(xglong *locof, int yce, int xce, int nrey, int nrex);
extern int reflstats(DISP_ATTRS *dispatt, double *istat) ;
extern int scatt_uh(char *uname, int uhva[], DISP_ATTRS *dispatt);
extern int scatt_uref(char *uname, int whichd, int whichi, DISP_ATTRS *dispa);
extern void calc_ranges(DISP_ATTRS *dispatt);
extern void disrot(int rotco, int direc, int imlim, int *imco);
extern void dorotate(DISP_ATTRS *dispatt, char *im00, char *im0r);
extern void expand_sdisplay(DISP_ATTRS *dispatt, int bound[]);
extern void fdisrot(int rotco, int direc, int imlim, double *imco);
extern void itoggle(int *ival, int thismask);
extern void rot2l(DISP_ATTRS *dispatt);
extern double nudge(double fv);
extern double quickstl(DISP_ATTRS *dispatt, int mx, int myy, int fmnu);
extern CTROID *get_calcen(int *ncalcep);
extern MROBJECT *managemrv(int opno);
extern TROID *gettroids(DISP_ATTRS *dispatt, int fmnu, int *ncenp);
extern URWORK *manageuwp(int opno, int nobj);
extern XG_Point *calibdisp(int *calp, DISP_ATTRS *datt, int ncpts, int *nptp);
extern XG_Point *grespts(DISP_ATTRS *dispatt, int *resinfo);
extern XG_Point *mkborderolay(DISP_ATTRS *datt, XG_Point *xgxp, int  *nptp);
extern XG_Point *rowcolpts(DISP_ATTRS *dispatt, int *nrcptp);

#endif /* DISPLAYS_H */

/* end of displays.h */
