/* integdefs.h -- #defines and typedefs for XGEN's integration function

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	19 Dec 2006: eliminate unused INRWORK structure typedef
	23 Jun 2006: include xgmodel.h so we can get to those structures
	12 Jun 2006: reorder declarations alphabetically
	 5 Jun 2006: delete xy2h,intr_integ,precenter, which are static now
	 8 Aug 2005: add IR_REGIONADJUST flag bit
	28 Jun 2005: add igr_overxy,igr_remaplim elements to INTEGRUN
	10 Jun 2005: delete igr_upbgnd, add igr_oversc elements in INTEGRUN
	17 Jan 2004: ra_{nrpred,nrref,rmx,rmy,rmz} added to REGIONATT type
	 9 Sep 2004: eliminate unneeded several igr_* elements
	18 Aug 2004: more elements in INTEGRUN to obviate static globals
	12 Aug 2004: add new elements to INTEGRUN structure for flexibility
	28 Jun 2004: put plotting variables into INTEGRUN structure type
	25 Mar 2004: express IR_* in hex rather than octal;
		add IR_LONGJUMP flag for proper program transfers
	21 Jan 2004: license message inserted
	 2 Jan 2004: add several more elements to INTEGRUN structure type
	20 Nov 2003: add INTEGRUN structure type
	20 Sep 2003: remove function declarations from here
	 5 Jan 2003: reorganized source-file header
	10 Nov 2002: reorganize REGIONATT structure: it's more general now.
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	31 May 1995: `pinrefls' and `pupdated' are ints, not longs
	12 Apr 1995: replace s_char type with sichar
	 2 Mar 1995: get rid of BLOTREF structure type
	 2 Mar 1995: add IR_PREDBACK and IR_PLANEBG;
		structure types BG_REFL, BGPLANE defined
	 5 May 1994: source-file header cleaned up
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: #defines and typedefs for X-GEN's integration functionality
   Cross References: int*.c

********************************************************************/

#ifndef INTEGDEFS_H
#define INTEGDEFS_H	1

#define		NEED_SIGNAL	1
#include	"xgmodel.h"
#include	<setjmp.h>

#define		BGDSCALE	16.	/* scale factor between bgnd & data */
#define		NFATES		257	/* # of possible fates for a refl */
#define		MAXDEV_RATIO	(4 * BGDSCALE)	/*max(BGnew-BGold)/sig(BG)*/
#define		BGOLDFAC	(BGDSCALE - 1.)
#define		PREDICTED	29	/* flag to get_limits */
#define		OBSERVED	47	/* ditto */
#define		NXBBINS		16	/* # of bgnd bins in X */
#define		NYBBINS		8	/* # of bgnd bins in Y */
#define		TOO_MANY_UNRELIABLE	3	/* can't use overflowed refls*/
#define		BGVARSCALE	0.1	/* var(B) currently 0.1 * Background */
/* Possible values for INTEGREF->flags: */
#define		I_USABLE	0	/* refl is well-behaved */
#define		I_BADMODEL	01	/* refl's model profile pathological */
#define		I_XYEDGE	02	/* refl is near edge of detector */
#define		I_BADPHISHAPE	04	/* refl is too broad or twinned */
#define		I_RANGEEDGE	010	/* too close to 0th or last frame*/
#define		I_OVERLAP	020	/* refl overlaps a lot with another */
#define		I_TOOWEAK	040	/* total intensity < 1 */
#define		I_PEAKONEDGE	0100	/* peak is on edge of xy range*/
#define		I_MOVETOOFAR	0200	/* obs. centroid too far from pred.*/
#define		I_ALREADYOUT	0400	/* we've already processed the refl */
#define		I_NOPIXTOCM	01000	/* unable to convert pix to cm */
#define		I_TINYSIGMA	02000	/* sigma too small to use in fitting*/
#define		I_NUNRELIABLE	04000	/* too many overflowed pixels */
#define		I_BADPLANEBG	010000	/* planar-background failed */
#define		I_FATAL		014732	/* fatal flags */
/* Possible values for runflags: */
#define		IR_JAGGED	0x1	/* don't smooth backgrounds */
#define		IR_UPD_MODEL	0x4	/* update model after strong refl */
#define		IR_LOSSCORR	0x8	/* correct for dead-time losses */
#define		IR_YCORREL	0x10	/* data are correlated along vertical*/
#define		IR_LONGJUMP	0x20	/* when finished, jump to integmain */
#define		IR_USEOLDBG	0x40	/* use the original background est. */
#define		IR_PLANEBGND	0x800	/* use (X,Y) plane backgrounds */
#define		IR_IGNOREEXPOS	0x1000	/* don't rescale on exposure time */
#define		IR_ICERINGS	0x4000	/* skip ranges of icerings */
#define		IR_VERBOSE1	0x8000	/* somewhat verbose */
#define		IR_VERBOSE2	0x10000	/* very verbose */
#define		IR_VERBOSETEMP	0x20000	/* temporarily verbose */
#define		IR_REGIONADJUST	0x40000 /* adjust (X,Y) in regions */
#define		IR_VERBOSEANY	(IR_VERBOSE1 | IR_VERBOSE2 | IR_VERBOSETEMP)
/* default values sometimes overriden by user flags */
#define		DEF_INTPASSES	3	/*max # passes thru main summing loop*/
/* flag values for blotting out reflections. */
#define		L_RECT_BORDER	65	/* pixel on outside rectang. border */
#define		L_ELL_BORDER	66	/* pixel on outside elliptical border */
#define		L_OVERLAP	68	/* pixel contained in (another) refl */

#define		NRESTYPES	11	/* number prediction of fates, below */

/* defines for the various fates of an attempt at prediction */
#define		XYRANGE		0	/* X,Y out of range */
#define		FIRSTPHIOK	1	/* first phi value  predicted */
#define		SECPHIOK	2	/* 2nd phi value predicted */
#define		BOTHPHIOK	3	/* both phi values predicted */
#define		PHIRANGE	4	/* phi outside (mu1->mu2) */
#define		BADY2		5	/* nonsense value of "y2" */
#define		ALONGBEAM	6	/* scatt vector along beam */
#define		TOOLOWRES	7	/* scatt vector too small */
#define		TOOHIRES	8	/* scatt vector too large */
#define		BADHKL		9	/* illegal indices in spacegroup */
#define		SZRANGE		10	/* z component of s out of range */
#define		CORCMON		(1.8945e04)
#define		CORCMIR		(2.e04)
#define		MONOC2		0.8945

/* structure type for general manipulations of the 12 detector regions */
typedef struct {
		u_char		ra_nmodels, ra_bx, ra_by, ra_bz;
		u_char		*ra_incl;	/* binary mask */
		xglong		ra_nrpred, ra_nrref;
		double		ra_offx, ra_offy, ra_offphi, ra_xc, ra_yc;
		double		ra_rmx, ra_rmy, ra_rmz;
		double		ra_sigx, ra_sigy, ra_sigz, ra_meangam;
		} REGIONATT;

/* structure type for log-file information associated with each frame */
typedef struct {
		double	lossfrac, fcorr, meandelb;
		double	iis_in, iis_ios, iis_d2, iis_i000, iis_b;
		double	iis_rmx, iis_rmy, iis_rmz;
		double	iis_lni, iis_ssq;
		int	pinrefls, pupdated;
		short	predrefl, outrefl, outstrong, maskupd;
		short	readsecs, bgsecs, sumsecs;
		} FRAMELOG;

/* structure type for omega-dependent updating of regional backgrounds */
typedef struct {
		double	r_ratio, r_newrat;
		int	r_npix;
		} REGION_BG;

typedef struct { /* structure type for one frame of planar background */
		short	bgp_p0;
		sichar	bgp_p1, bgp_p2;
		} BG_PLANE;	/* four bytes */

typedef struct { /* structure type for all of planar background */
		double		bg_sc;		/* 4 bytes */
		short		bgx_sc, bgy_sc;	/* 4 bytes */
		BG_PLANE	bg_pl[8];	/* 8 * 4 bytes */
		} BGREFL;	/* total: 40 bytes */

typedef struct {
		short		igr_xmi, igr_ymi, igr_xma, igr_yma;
		short		igr_zmi, igr_zma, igr_nrverbose;
		short		*igr_trwp, **igr_trlb;
		int		igr_idarkfrac, igr_imfl;
		int		igr_jzmin, igr_zmax, igr_zmin;
		int		*igr_ytopb, *igr_ybotb;
		void		*igr_wintv;
		u_char		*igr_curinbg, *igr_binc;
		xglong		igr_bgout, igr_bunch, igr_flags, igr_initend;
		xglong		igr_minunrely, igr_nactive, igr_nbginit;
		xglong		igr_nmbox, igr_totalrefls, igr_xbwidth;
		xglong		igr_xcen, igr_ybwidth, igr_ycen;
		xglong		igr_zhalfw, igr_zwidth, *igr_reflfates;
		xgulong		*igr_upwp;
		time_t		igr_tl0, igr_tl1;
		float		*igr_bm, *igr_bs, *igr_bv, *igr_kbg, *igr_kbgd;
		float		*igr_km, *igr_ks, *igr_kv, *igr_tq;
		float		igr_r0tos[3][3];
		double		igr_aduperph, igr_avexp100, igr_cost;
		double		igr_dmaxspec, igr_dminspec, igr_e1c, igr_e2c;
		double		igr_exc, igr_fphipr, igr_fracout, igr_fxpr;
		double		igr_fypr, igr_gamco, igr_iosmin, igr_minmod;
		double		igr_minsoi, igr_miny2sq, igr_missf;
		double		igr_newinflu, igr_overphot, igr_oversc;
		double		igr_overxy, igr_polfrac, igr_polnormrot;
		double		igr_remaplim, igr_roughi, igr_roughv;
		double		igr_sint, igr_vlp, igr_xscal, igr_yscal;
		double		igr_indlim[6], igr_kaf[6], igr_loch[3];
		double		igr_meanbg[2], igr_plbg[8][3], igr_tailc[2];
		jmp_buf		igr_jbuf;
		FILE		*igr_fpbgnd, *igr_fpmask, *igr_fpo, *igr_fpl;
		MAIL		*igr_heads;
		FRAMEINDEX	*igr_fmi;
		REFPARS		*igr_xp;
		VSPLINE		*igr_spl;
		OBSREF		*igr_reflp, *igr_rfbase, *igr_rflast;
		OBSREF		*igr_rfmin, *igr_rfmax, *igr_rfend;
		REGIONATT	*igr_bmask;
		FRAMELOG	*igr_logst, *igr_thislog;
		BOUNDARY	*igr_boundp;
		MASK_3D		*igr_wmask;
		REGIONPROFILE	*igr_rgb;
		REGION_BG	*igr_region0;
		} INTEGRUN;

/* function prototypes */
 
extern int after_images(INTEGRUN *igp);
extern int analumain(int argc, char *argv[]);
extern int bg_from_disk(INTEGRUN *igp);
extern int getrefls(int workframe, INTEGRUN *igp);
extern int init_bg(INTEGRUN *igp);
extern int integ_setup(int argc, char *argv[], INTEGRUN *igp);
extern int integmain(int argc, char *argv[]);
extern int integplot(int opno, int fmno, FILE *fp, INTEGRUN *igp);
extern int intpredini(INTEGRUN *igp);
extern int istestrefl(INTEGRUN *igp);
extern int mkintbord(INTEGRUN *igp);
extern int mkmodel(int regi, INTEGRUN *igp);
extern int orderly_exit(char *exitstring);
extern int pre_image(int framen, int preflags, INTEGRUN *igp);
extern int pred_bunch(int framemin, int nframes, int testonly, INTEGRUN *igp);
extern int process_image(int framen, INTEGRUN *igp);
extern int put_bgimage(MAIL *scrhead, INTEGRUN *igp);
extern int regionmem(INTEGRUN *igp);
extern int remap_intdistort(INTEGRUN *igp, int preflags);
extern int sharpmodel(int datatype, double riv, INTEGRUN *igp);
extern int trefmain(int argc, char *argv[]);
extern int upd_masks(INTEGRUN *igp);
extern short ssqcalc(INTEGRUN *igp);
extern void disp_bglp(INTEGRUN *igp);
extern void dispprofile(int dco,int jzm,double scv,float *ve,INTEGRUN *igp);
extern void framestats(int fmno, INTEGRUN *igp);
extern void get_limits(OBSREF *rp, int whichpart, INTEGRUN *igp);
extern void init3dvec(INTEGRUN *igp);
extern void note_badmodel(FILE *fp, double *devn, INTEGRUN *igp);
extern void pr_header_summing(FILE *fp);
extern void summ_integ(FILE *fp, INTEGRUN *igp);
extern void upd_bg(int fmno, INTEGRUN *igp);
extern void upd_xtal(int fmno, INTEGRUN *igp);
extern double ev2lp(double evp[], INTEGRUN *igp);
extern double losscorr(MAIL *mp, INTEGRUN *igp);
extern double varlp(INTEGRUN *igp);
extern double zerf(double zx);
extern INTEGRUN *manageigp(int opno, size_t nelts);
extern OBSREF *rforder(double zphi, INTEGRUN *igp);

#endif /* INTEGDEFS_H */

/* end of integdefs.h */
