/* refidefs.h -- definitions and typedefs for XGEN's "refine" function

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	27 Dec 2006: ice-ring-related defines; declare legald here
	19 Dec 2006: WORKOB typedef moved from here to ../xgrefine/refrefine.c
	28 Nov 2006: replace rfr_ra[3],rfr_ran[3] with xp->xast,xp->xanst;
	 remove rfr_q21 and rfr_q21t elements from REFINERUN structure type
	10 Jun 2006: refjournal,ref1args,ref1log,updateorigv deleted:now static
	24 May 2006: reorg of autoindexing allows us to omit many routines
	19 May 2006: fewer args to get_hkl, getinthkl; delete local routines 
	19 Mar 2006: add EC_USECURSES bit to ec_flags
	 5 Mar 2006: in REFREF structure, rename "frame" element to "framem"
	21 Oct 2004: rfr_minrmsref element in REFINERUN structure type;
		RR_ASKBRAVAIS bit in rfr_flags
	20 Aug 2004: RR_ASKSPGRP flag set up; all RR_ flags expressed in hex
	12 Aug 2004: add EE_* #defines & typedefs tfor editing centroids
	11 Aug 2004: add rfr_a0[] element to REFINERUN structure
	15 Jun 2004: RR_NORECENTER bit in runflags added
	 4 Jun 2004: GUESS and DIFFVECTOR structures moved from xgen.h;
		RR_CHANGEDDET is now irrelevant; RR_FLIPSTEP added
	29 May 2004: add rfr_lc0 element in REFINERUN structure
	21 Jan 2004: license message inserted
	 2 Jan 2004: add a few more elements to REFINERUN structure type
	25 Nov 2003: add REFINERUN structure type
	 5 Nov 2003: LATTICESCORE structure type and a few other items
		moved to xgen.h
	20 Sep 2003: remove function declarations from here
	18 Aug 2003: add typedef for GHKL to this header
		and include the MIN_EUNORMSQ #define
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	 2 Mar 1995: cross-references added
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: #defines and typedefs for refinement in X-GEN
   Cross References: ../xgrefine/refiglob.c

********************************************************************/
#ifndef REFIDEFS_H
#define REFIDEFS_H	1

#define		NEED_SIGNAL	1
#define		NEED_CTYPE	1
#include	"xgen.h"
#define		CATEGS		10
#define		CATEGSIZE	(0.5 / (float)CATEGS)
#define		NPELTS		12
#define		NPAR0		9	/* # of params in group 0 */
#define		NPAR1		8	/* # of params in group 1 */
#define		NPAR2		10	/* # of elements in group 2 */
#define		NSHORTS		2	/* # of `short' (not float) params */
#define		ALLPARS		(NPAR0 + NPAR1)
#define		NPAR2_FL	(NPAR2 - NSHORTS)
#define		ALLFLOATS	(ALLPARS + NPAR2_FL)
#define		NPARAMS		(ALLFLOATS + NSHORTS)
#define		THESEPARS	(ALLPARS-3) /* all refinables except gammas */
#define		NSHELLS		8	/* # of resolution shells to use*/
#define		NGUESSES	1000	/* # of GUESS structures to make */
#define		INDEXMAX	9	/* max index to try in autoindexing*/
#define		MAX_MULTIP	4	/* largest multipl. of axis to check */
#define		MAXG		5000	/* max (ax,ay,az) vectors kept */
#define		MIN_EUNORMSQ	0.0025	/* if Eucl.norm of matrix < this,==0 */
#define		MINY2SQ		0.9e-4	/* default minimum s(y) value */
#define		RANDOM	0.19 		/* very bad residual for one index */
/* values of the flag variable in the REFREF structure */
#define		OKREFL		0
#define		DTOOHIGH	01	/* D spacing > "resmin" */
#define		DTOOLOW		02	/* D spacing < "resmax" */
#define		BADINTEGER	04	/* indices too far from integers */
#define		BADXYPHI	010	/* calc X,Y,phi too far from obs'vd*/
#define		BADINDICES	020	/* indices illegal in this spacegrp */
#define		BADSCATTVEC	040	/* can't compute phi value */
					/* usually too close to rotation axis*/
#define		NOCALIB		0100	/* x,y outside calibration range */
#define		CHANGEDINDEX	0200	/* indices changed during cycle */
#define		INICERING	0400	/* refl is in an icering */
/* bits in the run-flag variable */
#define		RR_ISOLDTROID	0x1	/* old `troidinf' file type */
#define		RR_NEEDHEADER	0x2	/* display will require a header */
#define		RR_CHANGEDDET	0x4	/* det.params have changed:unused now*/
#define		RR_QUITNONLIN	0x8	/* quit nonlinear refinement soon */
#define		RR_CELLLIMITS	0x10	/* specified unit cell limits */
#define		RR_QUITTRYING	0x20	/* autoindex: don't keep looking */
#define		RR_QUITREFINE	0x40	/* refine itself: exit from "refine"*/
#define		RR_FLIPSTEP	0x80	/* sign of stepsize should be flipped*/
#define		RR_NORECENTER	0x100	/* no recentering after autoindex */
#define		RR_ASKSPGRP	0x200	/* ask user to specify spacegroup */
#define		RR_ASKBRAVAIS	0x400	/* ask user to confirm Bravais choice*/
#define		RR_FORCEREAD	0x800	/* force input from CENTROIDS */
#define		RR_EXCLUDEICE	0x1000	/* don't index on refls in icerings */
	/* error values for filtering out bad reflections: */
#define		EE_GOODREF	0	/* good reflection */
#define		EE_BADCM2PIX	1	/* cm-to-pixel conversion failed */
#define		EE_XISQZERO	2	/* xisq = sx^2 + sy^2 is zero */
#define		EE_SZSQZERO	3	/* szsquared is zero */
#define		EE_LORENTZ	4	/* Lorentz (|sy|) too small */
#define		EE_HERROR	5	/* h too far from integer */
#define		EE_KERROR	6	/* k too far from integer */
#define		EE_LERROR	7	/* l too far from integer */
#define		EE_XERROR	8	/* |Xob-Xpred| > Xerror */
#define		EE_YERROR	9	/* |Yob-Ypred| > Yerror */
#define		EE_ZERROR	10	/* |Zob-Zpred| > Zerror */
#define		EE_TOOLOWRES	11	/* D spacing > resmin */
#define		EE_TOOHIGHRES	12	/* D spacing < resmax */
#define		EE_XYOFFDETEC	13	/* (X,Y) outside detector area */
#define		EE_BADPIX2CM	14	/* pixel-to-cm conversion failed */
#define		EE_TOOWEAK	15	/* intensity < Imin */
#define		EE_OUTOFRANGE	16	/* frame number outside frame limits */
#define		EE_NONMONOTONE	17	/* profile not monotonically up/down */
#define		EE_TOOSHARP	18	/* profile is too sharp */
#define		EE_CLOSESPACE	19	/* this refl too close to another */
#define		EE_INICERING	20	/* refl is in an icering */
	/* bits in the `ec_flags' variable */
#define		EC_COMPLEMENT	01	/* save bad ones, not good ones */
#define		EC_EVERYREFL	02	/* check every reflection */
#define		EC_AUTODEL	04	/* automatically delete suspects */
#define		EC_DONE		010	/* we're finished */
#define		EC_CHECKHXYO	020	/* check errors in h,X,Y,omega */
#define		EC_USECURSES	040	/* "Curses" routines being used */
#define		EC_ICERINGS	0100	/* suppress output if refl in icering */

typedef struct {
		float	xto,yto,frame0;		 		/* 12 */
		short	obkmin,nobk;				/*  4 */
		float	totinten;				/*  4 */
		float	obparts[NPELTS];			/* 48 */
		float	dspac,wt,maxdevh;			/* 12 */
		float	rsl[3], rs[3], rsi[3], rh[3], rhi[3];	/* 60 */
		float	framem, rsp, xtcm, ytcm, xtc, ytc, framc; /* 28 */
		int	rflags;					/*  4 */
		} REFREF;			/* total:	  172 bytes*/

typedef struct {
		float	g_av[3], g_rf[3];
		float	g_le, g_re, g_lerf, g_rerf, g_lam;
		short	g_hv[3];
		} GHKL;

typedef struct { /* guessed unit cell solution for auto-indexing */
		float	gabc[3], gang[3], gomega[3], gresid, gfom;
		int	ghs[3][3];
		} GUESS;

typedef struct { /* diffraction vector structure--used in auto-indexing */
		float	sh[3], ih[3], ss[3];
		float	dwt,dnsq;
		} DIFFVECTOR;
 
typedef struct {
		char		*rfr_cname;
		char		rfr_rundate[26];
		short		rfr_spsys, rfr_spcen, rfr_spspgrp, rfr_autoind;
		xglong		rfr_firstok, rfr_lastok, rfr_inbounds;
		xglong		rfr_nrescalc, rfr_nreftot, rfr_nrefok;
		xglong		rfr_nindok, rfr_bestnr, rfr_npref, rfr_flags;
		xglong		rfr_menulist[14], rfr_ncateg[3][CATEGS];
		xglong		rfr_shlnref[NSHELLS], rfr_shlnind[NSHELLS];
		xgulong		rfr_ndecent, rfr_ndiffvecs, rfr_npeaks;
		xgulong		rfr_maxorig, rfr_minorig;
		time_t		rfr_tl0, rfr_tl1;
		float		rfr_gamrmse, rfr_profrmse, rfr_phirmse;
		float		rfr_xrmse, rfr_yrmse, rfr_indrmse;
		float		rfr_minrmsref, rfr_avedelfm, rfr_phiurmse;
		float		rfr_xurmse, rfr_yurmse, rfr_indurmse;
		float		rfr_xin, rfr_yin, rfr_rni, rfr_rni2;
		float		rfr_amin[3], rfr_amax[3], rfr_a0[3];
		float		rfr_angmin[3], rfr_angmax[3];
		float		rfr_rasq[3], rfr_ss[3];
		float		rfr_shlrmse[NSHELLS], rfr_shldmin[NSHELLS];
		float		rfr_pshift[ALLPARS];
		float		rfr_q3[3][3], rfr_sinv[3][3];
		float		rfr_sref[3][3], rfr_dhc[3][3];
		float		rfr_zsa[3][3], rfr_ov[2][3];
		double		rfr_besth, rfr_bestx, rfr_besty, rfr_bestphi;
		double		rfr_fnow, rfr_bestres, rfr_btb;
		double		*rfr_jtj, *rfr_jtb, *rfr_delt;
		double		rfr_bestsofar[THESEPARS], rfr_xnow0[THESEPARS];
		FILE		*rfr_fpl;
		VSPLINE		*rfr_spl;
		GUESS		*rfr_gu0;
		REFREF		*rfr_rpp, *rfr_rpp0;
		REFPARS		*rfr_xp, *rfr_fmprp;
		} REFINERUN;

typedef struct { /* working run structure for editing centroids */
		int		ec_flags, ec_nrin, ec_nrout, ec_nrdel;
		int		ec_fmmin, ec_fmmax;
		float		ec_mininten, ec_maxnonmono, ec_minsecpeak;
		float		ec_resmin, ec_resmax, ec_maxreld;
		FILE		*ec_fpi, *ec_fpo, *ec_fpl;
		REFINERUN	*ec_rfp;
		} EDITCENRUN;

typedef struct {
		int		vw_overhmax, vw_autoverbose, vw_vecheader;
		int		vw_ndused[MAX_MULTIP], vw_totpro[7];
		float		vw_minnsq, vw_maxnsq, vw_maxovsq;
		float		vw_nmpena, vw_nmpenb;
		float		vw_spena, vw_spenb, vw_spend;
		float		vw_ualim[6];
		FILE		*vw_fpl;
		GHKL		vw_gh[MAXG + 2], *vw_gl0, *vw_glc, *vw_glen;
		DIFFVECTOR	*vw_dvlim[2];
		} VECWORK;

/* function prototypes */

extern REFINERUN *managerfp(int opno, size_t ntroids);
extern double cossqv(float *s0, float *s1);
extern double gcloser(REFREF *rp, REFINERUN *rfp);
extern double getinthkl(REFINERUN *rfp);
extern double initmin(double iw, double xyw, double *ires, char rety, REFINERUN *rfp);
extern double newresid(char residtype, int *nchanged, REFINERUN *rfp);
extern int autoind1(double erv[], REFINERUN *rfp);
extern int diffm3(float *spacings, double mib, REFINERUN *rfp,VECWORK *vwp);
extern int get_hkl(REFINERUN *rfp);
extern int geth(REFINERUN *rfp, short excluc);
extern int globtxs(int op, REFINERUN *rfp);
extern int gpeaks(REFINERUN *rfp);
extern int impose_solution(int desired_soln, REFINERUN *rfp);
extern int impose_special(REFINERUN *rfp);
extern int incr1p(sichar refcomd, int nlpa, REFINERUN *rfp);
extern int incr_cycle(char resid_type, REFINERUN *rfp);
extern int indhkl(double percerr[], double basis, REFINERUN *rfp);
extern int legald(REFINERUN *rfp, REFREF *lrp);
extern int make_incrvecs(REFINERUN *rfp);
extern int minph1(double iw, double xyw, char rft, int dosh, REFINERUN *rfp);
extern int minph2(double indwt, double xywt, char reftype, REFINERUN *rfp);
extern int readpars(REFINERUN *rfp);
extern int ref1main(int argc, char *argv[]);
extern int ref3p(REFINERUN *rfp);
extern int refallmain(int argc, char *argv[]);
extern int refco(int ro, int ac, char *av[], int ia, double *fl, REFINERUN *rp);
extern int refinit(int refla);
extern int refl_coords(REFINERUN *rfp);
extern int refl_edit(EDITCENRUN *ecp);
extern int reflinfo(REFINERUN *rfp);
extern int remap_distortion(REFINERUN *rfp);
extern int rocking_ref(float oldgam[], int report, REFINERUN *rfp);
extern int s2xyp(REFREF *rrp, REFINERUN *rfp);
extern int unitcellx(REFINERUN *rfp);
extern void apply_dels(REFINERUN *rfp);
extern void crossproduct(float *s0, float *s1, float *norm);
extern void end_refinement(void);
extern void filterhead(FILE *fp, int withparams);
extern void get_rmserrs(char *s, REFINERUN *rfp);
extern void gethfacs(REFINERUN *rfp);
extern void getmainbeam(REFPARS *xp, float *xymp);
extern void guesstoxp(GUESS *gp, REFPARS *xp);
extern void impose_symmetry(REFPARS *xp, int lsym);
extern void list_commands(void);
extern void log_ecendel(FILE *fp, REFREF *rp, int reason);
extern void matrices(REFINERUN *rfp);
extern void minpar1(char reftype, REFINERUN *rfp);
extern void qnonlin1(void);
extern void refgonio(REFINERUN *rfp);
extern void scalars(REFINERUN *rfp);
extern void scattv(REFINERUN *rfp);
extern void showpeaks(FILE *fp, char *s, REFINERUN *rfp);
extern void sum_curparam(FILE *fp, char *s, REFINERUN *rfp);
extern void sum_edcenresults(FILE *fp, EDITCENRUN *ecp);
extern void usebest(REFINERUN *rfp);
extern void xptoguess(REFPARS *xp, GUESS *gp);

#endif /* REFIDEFS_H */

/* end of refidefs.h */
