/* spotdefs.h -- #define's and typedefs for `spots' functionality

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	 8 Jun 2007: enable external specification of stepsize between images
	19 Apr 2007: float->double conversions
	21 Mar 2007: spr_maxs added to SPOTRUN structure
	24 Feb 2007: add spr_lsq to SPOTRUN structure
	20 Jul 2006: add in P_FORCEOMEGA and P_FORCEPHI flag activities
	23 Jun 2006: include xgmodel.h so we can get to those structures
	 5 Jun 2006: several routines deleted as static; sift_peaks added
			flag values PA_* moved from peaks.c
	15 May 2006: peaks.c & peakproc.c merged; some routines now static
	 4 Mar 2006: add P_CONTIG* spr_flags; renumber pflags P_* as hex.;
		add several other spr_* elements
	 2 Jan 2006: reorder prototypes
	28 Sep 2005: add P_QUITQUIETLY bit to spr_flags
	16 Aug 2005: eliminate unnecessary definition SPOT_FRAME
	 2 Sep 2004: unneeded spr_medv element deleted from SPOTRUN struct
	21 May 2004: add more P_*BORDER flag bits to the spr_flags options
	21 Jan 2004: license message inserted
	 6 Jan 2004: additional elements in SPOTRUN structures;
		 flag value bits P_USEBORDER, ... defined here
	25 Nov 2003: add SPOTRUN structure type
	20 Sep 2003: remove function declarations from here
	 5 Jan 2003: reorganized source-file header
	10 Nov 2002: eliminate unnecessary definition SPOT_FRAMEHALF
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	21 Jul 1995: remove SHOEBOX structure type;
		replace P_ALREADYOUT with P_LOCALREADERR
	 2 Mar 1995: cross-reference added
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: major rewrite:
		different ways of finding peaks and backgrounds
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: #defines and typedefs for `spots' in X-GEN
   Cross References: ../prerefine/sp*.c, ../prerefine/peak*.c

********************************************************************/

#ifndef SPOTDEFS_H
#define SPOTDEFS_H	1
#include	"xgmodel.h"

#define		HALFPWIDTH	3
#define		PARTIAL_OVERLAP_RATIO	2.5
#define		SAMEPEAK_DISTSQ	32	/* radial size of spot squared */
#define		HUGEPHI		16000	/* enormous phi flag value */
#define		HUGEX		999.	/* flag value of "X" for weak refls */
#define		MININTEN	1	/* minimum intensity worth recording*/
#define		USELESS		2	/* useful pixels are 2<=x,y<=509*/
#define		BOXHEIGHT	32	/* Y height of   "     "       "    */
#define		BOXWIDTH	32	/* X width of region within detector*/
/* Possible values for PEAK->pflags: */
#define		P_USABLE	0	/* PEAK is well-behaved */
#define		P_XYDIMPLED	01	/* PEAK has a dimple in (X,Y) */
#define		P_XYEDGE	02	/* PEAK is near edge of detector */
#define		P_BADPHISHAPE	04	/* PEAK is too broad or twinned */
#define		P_RANGEEDGE	0x8	/* too close to 0th or last frame*/
#define		P_OVERLAP	0x10	/* peak overlaps a lot with another */
#define		P_TOOWEAK	0x20	/* total intensity < 1 */
#define		P_MOVETOOFAR	0x40	/* PEAK moved too far on 2nd pass */
#define		P_WASMERGED	0x80	/* PEAK has been merged into another */
#define		P_LOCALREADERR	0x100	/* can't read an image for a refl */
#define		P_TOOSHARP	0x200	/* all intensity in too few pixels */
#define		P_BELOWBG	0x400	/* I < bgnd for numerous pixels */
#define		P_NOLORENTZ	0x800	/* can't compute Lorentz factor */
#define		P_BIGLORENTZ	0x1000	/* large Lorentz factor */
#define		P_SOMELORENTZ	0x1800	/* bad or large Lorentz factor */
#define		P_FORCEPHI	0x2000	/* force motor in motion to be phi*/
#define		P_FORCEOMEGA	0x4000	/* force motor in motion to be omega */
#define		P_EXTERNSTEP	0x8000	/* stepsize is imposed by a flag */
	/* bits in spr_flags: */
#define		P_USEBORDER	1	/* check the border */
#define		P_VERBOSE1	2	/* moderately verbose */
#define		P_VERBOSE2	4	/* excruciatingly verbose */
#define		P_VERBOSEANY	(P_VERBOSE1 | P_VERBOSE2)
#define		P_FRAMEOFFSETS	010	/* foffb-style offsets */
#define		P_MAYBEBORDER	020
#define		P_MUSTDOBORDER	040
#define		P_DOBORDER	(P_MAYBEBORDER | P_MUSTDOBORDER)
#define		P_QUITQUIETLY	0100	/* don't wait for images to appear */
#define		P_CONTIGLEFT	0200	/* 1st image contiguous to neighbors */
#define		P_CONTIGRIGHT	0400	/* final image contig. to neighbors */
#define		P_CONTIG	0600	/* or'ing of those */
/* flag values for the `peakass' vector.  peakass[x * spp->spr_deth + y] can be
 the index of a specific reflection, i.e., it can range between 0 and the
 last peak in the list.  But it can also take on these flag values */
#define		PA_ANYFLAG	65500	/* any flag value is >= this */
#define		PA_REFLCORNER	65514	/* on the corner of a refl's 15x15 */
#define		PA_TEMP_OVERLAP	65516	/* overlapping another spot */
#define		PA_PERM_FLAG	65520	/* all values > this are permanent*/
#define		PA_UNASSIGNED	65522	/* pixel unassigned */
#define		PA_RECTBORDER	65524	/* x or y < 2 or > height-3) */
#define		PA_ELLBORDER	65526	/* outside elliptical active area*/

typedef struct {
		int		spr_badi, spr_beamx, spr_continue_refls;
		int		spr_deth, spr_detw, spr_etype;
		int		spr_firstrange, spr_flags, spr_ifrm, spr_imfl;
		int		spr_lowpx, spr_lz2, spr_maxpeaks;
		int		spr_mnbright, spr_nboxonx, spr_nboxony;
		int		spr_nfmsofar, spr_nmbox, spr_npts;
		int		spr_outpx, spr_passno, spr_sgz;
		int		spr_smallheight, spr_smallwidth;
		int		spr_smlhalf, spr_totalpeaks;
		int		spr_valshrink, spr_zma, spr_zmi;
		int		*spr_curbox, *spr_shoe;
		int		spr_ignorelow[NREGIONS];
		int		spr_ignorehigh[NREGIONS];
		sichar		*spr_sic;
		u_short		*spr_peakass, spr_cor[4];
		time_t		spr_tl0, spr_tl1;
		double		spr_aduperph, spr_ax0, spr_ay0;
		double		spr_badf, spr_efrm, spr_fbgnd, spr_framrad;
		double		spr_iosmin, spr_locint, spr_maxbelow;
		double		spr_maxmove, spr_maxlor, spr_maxs, spr_medbg;
		double		spr_mincenter, spr_minint, spr_nearphi;
		double		spr_nearxy, spr_nsigma, spr_outom, spr_overphot;
		double		spr_partoverlap, spr_phitoxscale, spr_plated;
		double		spr_pltd2, spr_qvalu, spr_rfrm, spr_samepeaksq;
		double		spr_spind, spr_swing, spr_tilt;
		double		spr_lsq[5], spr_q21[3][3];
		float		*spr_regbg, *spr_rprop, *spr_thisi0;
		FILE		*spr_fpmask, *spr_fpo, *spr_fpl;
		MAIL		*spr_heads;
		VSPLINE		*spr_spl;
		FRAMEINDEX	*spr_fmind;
		BOUNDARY	*spr_bdr;
		PEAK		*spr_peakp, *spr_pkbase;
		PEAK		*spr_pklast, *spr_pkend, *spr_pk0infm;
		MASK_3D		*spr_wmask;
		REGIONPROFILE	*spr_rgb;
		} SPOTRUN; /* as in "See Spot run" */

/* function prototypes */

extern int add2model(PEAK *tpk, int region, SPOTRUN *spp);
extern int eftable(int opno, SPOTRUN *spp);
extern int init_models(SPOTRUN *spp);
extern int fill_shoe(PEAK *tpk, SPOTRUN *spp);
extern int get_peaks(int fmno, SPOTRUN *spp);
extern int mkcentroid(PEAK *tpk, int *regp, SPOTRUN *spp);
extern int peakskill(int irv, SPOTRUN *spp);
extern int sift_peaks(SPOTRUN *spp);
extern int spotmain(int argc, char *argv[]);
extern void anal_peaks(SPOTRUN *spp, int fail[]);
extern void default_model(float *te0, SPOTRUN *spp);
extern double radialdist(PEAK *pk0, PEAK *pk1, SPOTRUN *spp);
 
#endif /* SPOTDEFS_H */

/* end of spotdefs.h */
