/* unixxdefs.h -- Unix-specific parameters for XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	29 May 2007: <endian.h> consulted under Linux in setting LITTLEND
	27 Apr 2007: #define of "i386" consulted for setting LITTLEND
	 5 Sep 2006: explicit handling of IRIX and IRIX64: typedef for sig_t
	 6 May 2005: add EHELP code
	20 Apr 2005: error values now resemble those in ASM standard, sort of
	21 Jan 2004: license message inserted
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	10 Feb 2002: (?) change handling of atof declaration
	15 Sep 1997: LITTLEND defined as true under Linux
	24 Aug 1995: get rid of IBM oddity for `curses' includes
	12 Apr 1995: replace s_char type with sichar and clean up its use
	 2 Mar 1995: cross-reference added
	 5 May 1994: OSF-1 definition of LITTLEND added
	21 Apr 1994: renaming polydet.h->xgen.h, adding vsort.h
	 7 Dec 1993: initial X-GEN version
	18 Jan 1988: include sys/types.h in this system.

   Purpose: preprocessor defines and typedefs used in XGEN that are
	specific to Unix implementations.
   Cross References: xgen.h

********************************************************************/

#ifndef UNIXXDEFS_H
#define UNIXXDEFS_H	1

#include	<stdio.h>
#include	<math.h>
#include	<sys/types.h>	/* XENGEN 2.1 we include this under Unix.*/
/*
	This material has been pulled out of xgen.h to make it more portable.
	Much of the material that is required comes from system include files,
	and the need for it is governed by the defines NEED_ERRNO,
	NEED_CURSES, NEED_SIGNAL, and so on.
	The return values EBAD, EOK, etc. are given appropriate values.
*/
#ifdef	NEED_STAT
#include	<sys/stat.h>
#endif
#ifdef	NEED_TIME
#include	<time.h>
#endif
	/* includes responding to defines in the invoking code: */
#ifdef	NEED_ERRNO
#include	<errno.h>
#ifdef alliant
extern int	errno;
#endif
#endif

#ifdef	NEED_CURSES
#include	<curses.h>
#endif

#ifdef	NEED_SIGNAL
#include	<signal.h>
#ifdef SunOS	/* This typedef is absent on SunOS somehow */
typedef	void (*sig_t)(int);	/* type of signal function */
#endif
#ifdef sgi
typedef void (*sig_t)(int);     /* type of signal function */
#endif
#endif

#ifdef	NEED_CTYPE
#include	<ctype.h>
#endif

#ifdef	NEED_ERRNO
#ifndef	EPERM
extern int	errno;
#endif
#endif

#ifdef NEED_UNISTD
#include	<unistd.h>
#endif
	/* any system-wide definitions we want specifically for non-VMS
	versions should go here */
#define		EOK		0	/* normal exit status */
#define		EBAD		1	/* error (EPERM) */
#define		EALLOC		12	/* memory allocation error (ENOMEM) */
#define		EREADERR	5	/* error in reading (EIO) */
#define		EWRITERR	14	/* error in writing (EFAULT) */
#define		EOPENERR	2	/* error in opening a file (ENOENT) */
#define		ECLOSERR	23	/* error in closing a file (ENFILE) */
#define		ESEEKERR	29	/* error in seeking (ESPIPE) */
#define		ESINGULAR	33	/* singular matrix (EDOM) */
#define		EBADVALUE	22	/* bad arg to a function (EINVAL) */
#define		EHELP		38	/* help/version called (ENOSYS) */
#define		ETIMEOUT	110	/* read timed out (ETIMEDOUT) */
	/* any Unix-specific type declarations go here */
 /* for now, we'll declare atof here except for Irix, which gets it right */
#ifndef irix
extern double	atof(const char *str);
#endif
#ifdef _H_STANDARDS
extern int	unlink(char *str);
#endif
extern unsigned	sleep(unsigned int secs);
#ifdef LITTLEND
#undef	LITTLEND	/* we'll re-define this for some systems below. */
#endif
	/* declarations peculiar to specific Unix systems */
#ifdef _HPUX_SOURCE
typedef char		sichar;
#else
typedef signed char	sichar;
#endif
#ifdef ultrix
#define	LITTLEND	1
#endif
#ifdef __alpha
#define	LITTLEND	1
#endif
#ifdef i386 
#define	LITTLEND	1
#endif
#ifdef linux
#include	<endian.h>
#if __BYTE_ORDER==__LITTLE_ENDIAN
#define	LITTLEND	1
#endif
#endif
/* alliant won't let you declare a function's formal argument as `void' */
#ifdef alliant
typedef char	AVOID;
#else
typedef void	AVOID;
#endif
 
#endif /* UNIXXDEFS_H */

/* end of unixxdefs.h */
