/* vmsxdefs.h -- VMS-specific typedefs and defines for XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	21 Jan 2004: license message inserted
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments

   Purpose: to provide preprocessor defines and typedefs for XGEN
	in its VMS-specific implementations.
   Cross References: unixxdefs.h

********************************************************************/

#ifndef VMSXDEFS_H
#define VMSXDEFS_H	1

#ifdef VAX11C	/* these constructions are for old VMS systems */
#include	stdio
#include	math
#else		/* this is for DEC C -- Alpha, etc. */
#include	<stdio.h>
#include	<math.h>
#endif
/*
	This material has been pulled out of xdefs.h to make it more portable.
	Much of the material that is required comes from system include files,
	and the need for it is governed by the defines NEED_ERRNO,
	NEED_CURSES, NEED_SIGNAL, and so on.
	The symbols EBAD and EOK are given values appropriate to VMS contexts.
*/
#ifdef	NEED_STAT
#ifdef VAX11C
#include	stat
#else
#include	<stat.h>
#endif
#else
typedef unsigned	off_t;	/* otherwise defined in stat.h */
#endif

#ifdef	NEED_ERRNO
#ifdef VAX11C
#include	errno
#else
#include	<errno.h>
#endif
#endif

#ifdef	NEED_SSDEF
#ifdef VAX11C
#include	ssdef
#else
#include	<ssdef.h>
#endif
#endif

#ifdef	NEED_CURSES
#ifdef VAX11C
#include	curses
#else
#include	<curses.h>
extern void	decc$gwprintw();
extern void	decc$gwscanw();
extern void	decc$gprintw();
extern void	decc$gscanw();
#endif
#ifndef mvwaddstr
# define mvwaddstr(win,y,x,str) (wmove(win,y,x)==ERR)?ERR:waddstr(win,str)
# define mvaddstr(y,x,str)      mvwaddstr (stdscr, y, x, str)
#endif
#endif

#ifdef	NEED_SIGNAL
#ifdef VAX11C
#include	signal
#else
#include	<signal.h>
#endif
#ifdef	SIG_DFL		/* correct minor boo-boos in VMS C declarations */
#undef	SIG_DFL
#endif
#define	SIG_DFL	( (void (*)())0 )
#ifdef	SIG_IGN
#undef	SIG_IGN
#endif
#define	SIG_IGN	( (void (*)())1 )
#else
int	sleep();	/* otherwise defined in signal.h */
#endif

#ifdef	NEED_CTYPE
#ifdef VAX11C
#include	ctype
#else
#include	<ctype.h>
#endif
#endif

#ifdef	NEED_DESCRIP
#ifdef VAX11C
#include	descrip
#else
#include	<descrip.h>
extern int	lib$create_dir(), lib$set_logical();
extern int	lib$set_symbol(), lib$spawn(), lib$sys_trnlog();
#endif
#endif

#ifdef	NEED_VSORT
/* HEK 1/5/94 */
/* #include "../incl/vsort.h" */
#include "vsort.h"
#endif
/* any system-wide definitions we want specifically for the VMS
	version should go here */
#define		EOK		1	/* normal exit status on VMS */
#define		EBAD		20	/* operator-defined error */
#define		LITTLEND	1	/* VMS puts the little-end first */
/* HEK 1/4/94 Preliminary numbers from Andy Howard */
#define		EALLOC		292	/* memory allocation error */
#define		EREADERR	9412	/* error in reading */
#define		EWRITERR	9420
#define		EOPENERR	156	/* error in opening a file */
#define		ECLOSERR	1705	/* error in closing a file */
#define		ESEEKERR	156	/* error in seeking within a file */
#define		ESINGULAR	20	/* singular matrix */
#define		EBADVALUE	20	/* bad arg to a function */
#define		ETIMEOUT	158	/* attempt to read image timed out */
#define         AVOID           void    /* null argument */
/* any VMS-specific type declarations go here */
extern double	atof();	/* vms C v.2.3 and later don't put this in MATH.H */
	/* typedefs that Unix does in sys/types.h */
typedef unsigned char	u_char;
typedef unsigned short	u_short;
typedef char		sichar;
#ifndef NEED_MOTIF
typedef unsigned long	time_t;
extern time_t	time();	 /* conflicts with X under VMS; works fine otherwise. */
#endif
typedef unsigned int	mode_t;
 
#endif /* VMSXDEFS_H */

/* end of vmsxdefs.h */
