/* xgen.h -- preprocessor defines, typedefs, function delarations for X-GEN.

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
 Changes for the X-GEN package as a whole:
	27 Jun 2007: 5.8.3: bugfix in vecmat; DFUZZ invented; `imageform' is
	  more effective; unnecessary calls to `exit' replaced with returns
	 6 Jun 2007: 5.8.2: support for 2-segment and 4-segment detectors;
	  better support for separating CALCENTROIDS and merging USPLINEs
	20 Apr 2007: massive conversion of floats to doubles;
	 style of curly-bracket usage reworked throughout package
	24 Mar 2007: 5.8.1: better plane-backgrounds in integration;
		quick ascii output via xgvariable
	27 Dec 2006: 5.8.0: better remapping functionalities;
		more ways to exclude ice-rings in refinement
	31 Oct 2006: 5.7.9: ./configure and Makefile.in set up;
	 more system portability established.
	 8 Sep 2006: 5.7.8: support for (old) SGI gcc compilers
	12 Aug 2006: 5.7.7: handle byteswapped formats without
		recourse to swap-specific format defines
	 2 Aug 2006: 5.7.6: improved support for Quantum 315;
	  handle inconsistencies in image header information better;
	  better handling of help routines
	 3 Jun 2006: 5.7.5: many routines now static: not declared in xgen.h;
	 general reorganization of parameter-file dependencies & Makefile
	15 Apr 2006: 5.7.4: more diagnostics in strategy; spacepname added
	 7 Mar 2006: 5.7.3: Mar IP filetypes augmented and systematized;
		Mar headers improved; plor element added to PEAK
	 3 Jan 2006: 5.7.1: qsorts and setjmps cleaned up
	 5 Dec 2005: 5.7.0: inline function declarations & function prototypes;
		numerous bugfixes; effective framenumbers available
	 9 Nov 2005: 5.6.0: major housekeeping: most static globals eliminated;
		numerous sign-checks inserted
	16 Aug 2005: 5.5.4: superior handling of (X,Y) errors in integration;
		move specialized #defines from here to other places
	28 Jun 2005: 5.5.3: handle 64-bit machines correctly by
	 redefining xglong, xgulong, reorganizing structures appropriately
	22 Jun 2005: 5.5.2: housekeeping, more static routines;
		full-featured single Makefile
	15 Jun 2005: 5.5.1: background calcs in integration extensively redone
	24 May 2005: X-GEN 5.5.0: only five executables:
	 xgen, xgentools, pmotif, pdisplay, imageform.
	20 Apr 2005: X-GEN 5.4.8: reorganize Makefiles and directories
	14 Jan 2005: X-GEN 5.4.6: gamma2 = 0.0010; remove simulate_click;
	 maybe_image -> imagename and imagelike:  marccd 300mm added
	27 Dec 2004: X-GEN 5.4.5: handle --help and --version args;
	 memory improvements in xgnow; cosmetics at end of integration
	23 Oct 2004: X-GEN 5.4.4: `ready' knows about more programs;
	 help docs upgraded; doc suffix is .txt; GUI labels <= 18 characters
	20 Oct 2004: X-GEN 5.4.3: pmotif augmented; make_ftab bugfix
	15 Oct 2004: X-GEN 5.4.2: pmotif fixed and augmented
	 9 Oct 2004: X-GEN 5.4.1: major changes to pdisplay
	 1 Sep 2004: X-GEN 5.3.0: stats redesigned; global variables removed;
		editcen and filterc rendered functional
	 2 Jul 2004: software version set to 5.1.0
	 9 Feb 2004: software version set to 5.0.0
 Changes for xgen.h specifically:
	25 Jun 2007: DFUZZ defined as double-precision error (epsilon) value
	19 Apr 2007: float->double conversions
	22 Mar 2007: xgvariable() added for ascii quick output
	21 Dec 2006: add least-squares coefficients to VSPLINE typedef;
	 declare modspline, polyadd, polymap
	28 Nov 2006: add new elements to REFPARS structure;
	 add rp2vecs routine; adjust calls to mat2vecs, makeq21, etc.
	 5 Sep 2006: dummy variable in exclucode declaration is now "sys"
	23 Jun 2006: xgimage.h,xgmodel.h,xgpostint.h created, contents
	 removed from here; other typedefs moved to xgenpred.h
	22 Jun 2006: move DISP_ATTRS and XG_Point typedefs to displays.h
	13 Jun 2006: SCANREF and OUT_OBSSTYLE typedefs moved to reddefs.h
	14 May 2005: system routines read and write declared as size_t
	13 Aug 2004: run_fmmin and run_fmmax elements added to DATARUN struct
	21 Jun 2004: move #define for CMSLOP here
	15 Jun 2004: imageflag bit for gzipping is now correctly abbreviated
	 4 Jun 2004: move GUESS and DIFFVECTOR structures to refidefs.h
	28 May 2004: add lc_fom element in LATTICESCORE structure;
		add SHEL_HKL structure type to this file
	25 Mar 2004: add DISPVERBOSE* flags -- how chatty should displays be?
	05 Mar 2004: add #define for CNEUTRAL
	08 Feb 2004: ANNOUNCE_BOX flag for detector reading added;
	 numerous #defines reordered so they're a bit easier to find
	21 Jan 2004: license message inserted
	18 Jan 2004: cosmetic changes to several structure types
	15 Jan 2004: DUMPRUN image-dumping structure type
	 3 Jan 2004: XGNAME structure element added
	 2 Jan 2004: MAXDEFAULTS, LABELLENGTH changed a little;
		DATARUN structure has aduperph,etc. added to it
	 5 Nov 2003: LATTICESCORE typedef and a few other things moved
		here from refidefs.h
	 5 Jan 2003: reorganized source-file header
	20 Sep 2002: reordering of some defines and comments;
		incorporate setup/pmotif defines (e.g. MAXLABELS) here;
		incorporate CF_* bits in runflags from calibration here
	10 Jun 2002: (?) move XMSHOW and PMENUSTUFF structure types here
	25 Mar 2002: removed MSI attachments
	31 Aug 1995: correct spelling on 2 comments
	 7 Jul 1995: typedef for pointer-equivalent integer added
	 6 Jul 1995: add FMT_DIPX000 and _DIPXKSWAP;
		new flags for image-reading;
		specify xglong and xgulong as 32-bit on irix6.4
	20 May 1995: declare ctime, even on IBM; declare shcalloc, shfree
		eventcount is an xglong, not an xgulong
	12 Apr 1995: replace s_char type with sichar
	 2 Mar 1995: add FRAMEOFF structure type
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 4 Nov 1994: add structure elements for detector type and X-GEN
		version # to REFPARS and CRYSTAL structures
	 5 Oct 1994: add XG_* preprocessor constants and
		xgread, xgwrite definitions
	30 Sep 1994: DATARUN structure element `run_stepreal'
		replaces old `run_fwhmave'
	 5 May 1994: definition and use of "xglong" and "xgulong" added
	21 Apr 1994: initial X-GEN version
	29 Apr 1993: flexibly-sized VSPLINE structures replace
		SPLINE structures.
	17 Nov 1992: remove deadwood and conform to ANSI C a bit better.
	15 Mar 1990: pull out Unix-specific and VMS-specific stuff
		into other files.
	24 Nov 1986: incorporate 1024-byte image header as an option.
	21 Sep 1985: incorporate data reduction materials herein.
	xx xxx 1983: created in the dim past.

   Purpose: Providing #defines, typedefs, and function declarations
	for the XGEN crystallographic data processing package.
   Cross References: unixxdefs.h, vmsxdefs.h

********************************************************************/
#ifndef XGEN_H
#define XGEN_H	1
#ifdef VMS
#include	"vmsxdefs.h"
#else
#include	"unixxdefs.h"
#endif
#ifdef ANCIENT_16 /* xglong is always 32 bit */
#define xgulong	unsigned long
#define xglong	long
#else
#define	xgulong unsigned int
#define xglong	int
#endif
/* parameters */
#define		ANNOUNCE_BOX	040000	/* print framename in 1st read_box */
#define		BIGSPOTHT	15	/* height of rectangular refl box */
#define		BIGSPOTWID	15	/* width of rectangular refl box */
#define		DEF_LAMBDA	1.5418	/* default wavelen (Cu Kal--mean 1&2)*/
#define		DETECTOR_EDGE	8	/* ok data are this far from edge */
#define		GAUSSIAN	1	/* fit profile to 3-D Gaussian */
#define		LORENZIAN	0	/* fit profile to 3-D Lorenzian */
#define		MASKRESCALE	(99.)	/* model-profile rescaler */
#define		MAXMODIFIERS	48	/* max # of modifiable entries in menu*/
#define		MAXOB		127	/* max # of obs of a merged refl*/
#define		FUZZ		(0.119e-6)	/* hardware FUZZ value */
#define		DFUZZ		(0.220e-15)	/* hardware double FUZZ value */
#define		MAXFMCHARS	180	/*max # of chars in a frame filename*/
#define		MEANPIXPERCM	(47.5)	/* <pixels per centimeter> in X */
#define		NREGIONS	12	/* # of 3-D profile detector regions */
#define 	PFRAMES		7	/* max # frames summed */
#define		XT_HOLDER	36	/* # unused bytes in CRYSTAL struct*/
#ifdef PI
#undef PI
#endif
#define		PI		3.1415926536
#define		R3		(1./3.)
#define		T3		(2./3.)
#define		PROCESSING	020	/* code for data-processing-type pgms*/
#define		REDUCTION	040	/* code for data-reduction-type pgms */
#define		SIZREF		20	/*size of REF part of TROID structure*/
#define		TROIDWIDTH	11	/* # of profile elements in a TROID */
/* parameters which are defined in terms of other parameters */
#define		PIO2		(PI / 2.)
#define		PIX2		(PI * 2.)
#define		RADC		(PI / 180.)
 /* modify the line below this very carefully! we parse it in the software */
#define		XGEN_VERSION	"5.8.3"	/* current version of XGEN */
 /* return value from routines that read detector images; others in xgimage.h*/
#define		FRAME_OK	0	/* no errors in reading in frame */
	/* datatypes for normal X-GEN input/output operations.  Each datatype
	below corresponds to the typedef structure whose name follows the XG_*/
#define	XG_FLOAT	0
#define	XG_CALPOINT	1
#define	XG_REFPARS	2
#define	XG_DATARUN	3
#define	XG_CRYSTAL	4
#define	XG_OBSREF	5
#define	XG_ABBROB	6
#define	XG_MASK_3D	7
#define	XG_PREDICTREST	8	/* offset from start of predict structure */
	/* all datatypes ABOVE this point are indivisible; those below this
	point contain a variable number of items that are read or written
	separately */
#define	XG_VSPLINE	10
#define	XG_TROID	11
#define	XG_MERGEDREF	12
#define	XG_PREDREF	13
	/* numeric codes indicating whether we're ready to do certain tasks */
#define	CANNOT		0
#define	CAN		1
#define	SHOULD		2
#define	NEEDNOT		3
#define	COMMAND		4
#define	NEUTRAL		5
#define	NCOLORS		6
#define	CNEUTRAL	(NEUTRAL + NCOLORS)
	/* flag bits for calls to `read_image' */
#define		ANNOUNCE	2	/* summarize frame contents to stderr*/
#define		JUST_HEADER	4	/* just read the header in */
#define		RETURN_IF_CONVERTED	010	/* don't continue if file
							is in VMS format */
#define		WAIT_IF_ABSENT	020	/* if frame is absent, don't quit;
					instead, just wait 20 minutes */
#define		GET_RUNHEADER	0100	/* fish out DATARUN header info */
#define		GZIP_AFTER_CLOSE	0200	/* gzip frame after closure*/
#define		DARK_CURRENT	0400	/* subtract dark-current image */
#define		GRIPE_IF_SHORT	01000	/* complain if frame is too short */
#define		JUST_HEADPFLOWS	02000	/* just read header&overflows */
#define		LEAVE_OPEN	010000	/* don't close image file yet */
	/* macros */
#ifdef	NEED_TIME
#define		NOW	(time((time_t *)((xglong)0)))
#endif

/* structure definitions */
/* this makes qsort happy when we call it */
 typedef int (*QPE)(const void *, const void *);

 typedef struct { /* remember excluded resolution ranges (e.g. icerings)*/
		int		icw_nranges;
		double		*icw_ramin, *icw_ramax;
		} ICEWORK;

	/* Overflow-table element, as used in the data reduction software */
 typedef struct {
		u_short	value, xcoord, ycoord;
		} VOFLOW;

	/* 16-bit overflow */
 typedef struct {
		xgulong	olvalue;
		u_short	olxcoord, olycoord;
		} VOFL16;

 typedef struct {		/* structure type for a data run: 256 bytes */
		char	run_code[8];	/* 8-character code name of run */
		char	run_notes[64];	/* notes on the run */
		short	run_id;		/* integer identifier of the run */
		short	run_volxray;	/* X-ray voltage in kV */
		short	run_curxray;	/* X-ray current in ma */
		short	run_chvol;	/* chamber voltage in kV */
		short	run_stepsize;	/* stepsize,units of 1/96 deg*/
		short	run_nframes;	/* number of frames in run */
		short	run_motor;	/* motor being moved during run */
		short	run_nshifts;	/* number of shifts in the run */
		short	run_nchams;	/* # of chambers being used */
		short	run_temp;	/* temperature during run */
		short	run_width;	/*horizontal dimension of detector(s)*/
		short	run_height;	/*vertical dimension of detector(s)*/
		xglong	run_ninteg;	/* # of refls successfully integrated*/
		xglong	run_nexam;	/* # of refls examined during run */
		float	run_lambda;	/* Xray wavelength in Angstroms */
		float	run_startpos[3]; /* start-of-run goniostat position */
		float	run_recip[3][3]; /* reciprocal-space unit cell matrix:
						columns are a*, b*, c* */
		float	run_fmmin;	/* first image that produced data */
		/* float run_refval; reference value of motor in motion*/
		float	run_expos;	/* exposure time per frame */
		float	run_fmmax;	/* last image that produced data */
		/* float run_bgiexpo;	 expos time/frame in background run */
		float	run_d0pos[3]; /* coords of center of detector #0 */
		float	run_d0ang[3]; /* angular attitude of detector #0 */
		float	run_d1pos[3]; /* coords of center of detector #1 */
		float	run_aduperph;	/* # of adu per photon */
		float	run_eperadu;	/* # of electrons that produce 1 adu */
		float	run_eperph;	/* # of electrons per photon */
		/* float	run_d1ang[3]; angular attitude of detector #1*/
		float	run_wtsum;	/* sum of weights in the next 3 aves */
		float	run_intave;	/* average value of I in run */
		float	run_sigave;	/* average value of sigma in run */
		float	run_iosave;	/* <I/sigma> */
		float	run_delomave;	/* <delta(omega)>*/
		float	run_rffu;	/* reserved for future use */
		/* float	run_rmsomave; */ /* RMS error in omega */
		float	run_ssqave;	/* <square of scattering vector length*/
		float	run_stepreal;	/* true stepsize of run in degrees */
		float	run_reslim;	/* resolution limit in Angstroms */
		float	run_ssqmax;	/* largest squared scattvector */
		/* The element shown below is NOT read and written */
		VOFLOW	*run_oflowend;	/* pointer to end of overflow list */
		} DATARUN;

 typedef struct {		/* structure type for each crystal: 256 bytes */
		char	xt_code[16];	/* 16-character code name of crystal */
		char	xt_desc[48];	/* description of crystal */
		char	xt_notes[64];	/* notes on crystal itself */
		char	xt_spaceg[8];	/* spacegroup */
		char	xt_device[8];	/* name of xray device */
		short	xt_intl;	/* International Tables' spacegroup #*/
		short	xt_isys;	/* spacegroup system number */
		short	xt_shhold[2];	/* extra "short" variables */
		float	xt_ar[3];	/* unit cell lengths in Angstroms */
		float	xt_ang[3];	/* unit cell angles in degrees */
		float	xt_stdpos[3];	/* goniostat position where c* is along
						+Z and b* mostly +Y, +Z */
		float	xt_lambda;	/* X-ray wavelength in Angstroms */
		float	xt_reslim;	/* overall resolution limit, Angstrom*/
		float	xt_ssqmax;	/* largest squared scattvector length*/
		xglong	xt_ninteg;	/* total # of observations integrated*/
		xglong	xt_nrefls;	/* total # non-Bijvoet-merged refls */
		xglong	xt_nmerged;	/* total # of Bijvoet-merged refls */
		short	xt_nruns;	/* total # of data collection runs */
		short	xt_nshifts;	/* total # of shifts */
		short	xt_dettype;	/* detector type code */
		short	xt_version;	/* X-GEN version code */
		char	xt_rffu[XT_HOLDER];	/* space holders */
		} CRYSTAL;

/*	Mailbox structure...what the collection routine looks at	*/
 typedef struct {
		xgulong	format;		/* mapped or raw format */
		xgulong ptrspace0[4];	/* space for input pointers */
		xgulong	xdimen;		/*  width in pixels of each detector */
		xgulong	ydimen;		/* height in pixels of each detector */
		xgulong	quadrant;	/* 0-4 indicate quadrants*/
		xglong	eventcount;	/* # of events collected */
		xgulong	oflowcount;	/* # of overflows */
		xgulong	zmax;		/* pixel with highest count*/
		u_short	zmaxx;		/* x coord of zmax */
		u_short	zmaxy;		/* y coord of zmax*/
		xgulong	zdisp;		/* display level param */
		xgulong	ontime;		/* # of counts on time */
		xgulong	late;		/* # of counts late */
		xgulong	datetime[2];	/* date&time space */
		xgulong	accumsecs;	/* accum.expos.time,secs */
		xgulong	elap100ths;	/* elapsed time/frame,100ths*/
		xgulong	cellsize;	/* nominal size of a pixel */
		xgulong	nonzpix;	/* # of pixels with > 0 cts */
		xgulong	over64k;	/* # of pixels with >64k cts*/
		xglong	frameno;	/* frame number */
		xgulong ptrspace1;	/* space for another pointer */
		short	count0;		/* counter for various things */
		short	count1;		/* another counter */
		u_short	improps;	/* image properties */
		short	imtype;		/* cardinal image type */
		xglong	count2;		/* another counter (32 bits here) */
		xglong	count3;		/* and another counter */
		char	filename[16];	/* output file name */
		/* the portions beyond this point are not read & written */
		u_char	*data8;		/* ptr to 8-bit data */
		u_short	*data16;	/* pointer to 16-bit data */
		VOFLOW	*overflow;	/* ptr to start of oflow list*/
		VOFL16	*oflow16;	/* ptr to list of 16-bit overflows */
		DATARUN	*runheader;	/* pointer to run header */
	} MAIL;

 typedef struct {	/* structure describing an element in the frame index*/
		short	fmno;			/* frame number */
		char	fmname[MAXFMCHARS];	/* name of frame file */
	       } FRAMEENTRY;

 typedef struct { /* structure describing the entire frame index */
		xglong		fi_totfms, fi_firstfm, fi_lastfm, fi_flags;
		xglong		fi_firstfmd, fi_firstfmm;
		xglong		fi_lastfmd, fi_lastfmm;
		FRAMEENTRY	*fi_fmi, *fi_fme;
		} FRAMEINDEX;

 typedef struct { /* (X,Y) calibration point structure; totals   32 bytes:*/
		float	c_ptocx, c_ptocy, c_ctopx, c_ctopy;	/*    16 */
		float	c_qec, c_frfu[2];			/*    12 */
		u_char	c_ift, c_set, c_cset, c_ucrfu;		/*     4 */
		} CALPOINT;

 typedef struct { float   fmdx, fmdy;     int     fmfm; } FRAMEOFF;

 typedef struct { /* complete calibration structure;	totals 512 bytes:*/
		short		v_detw, v_deth, v_ncol, v_nrow;	 /*    8 */
		short		v_ntrue, v_midc, v_midr, v_nreg; /*    8 */
		float		v_axc, v_ayc, v_rxc, v_ryc;	 /*   16 */
		float		v_xrcm, v_yrcm, v_xrpx, v_yrpx;	 /*   16 */
		float		v_x0cm, v_y0cm, v_x0px, v_y0px;	 /*   16 */
		float		v_rmsshift, v_maxshift;		 /*    8 */
		float		v_maxx, v_maxy;			 /*    8 */
		double		v_lax[21], v_lay[21];		 /*  336 */
		double		v_ldx[6], v_ldy[6];		 /*   96 */
		/* These doubles and pointers are not read and written */
		double		v_dex[6], v_dey[6];		 /*   96 */
		CALPOINT	*v_cpt;				 /*    4 */
		FRAMEOFF	*v_foffb;			 /*    4 */
		} VSPLINE;	/* total (not counting last stuff):  512 */

 typedef struct {
		float	xlambda,xrec[3][3],xx0,xy0;		/* 48 */
		float	xplated,xpltd2,xareal[3][3],xtilt;	/* 48 */
		float	xctilt,xstilt,xfrmsize,xgam[3];		/* 24 */
		float	xswing,xssw,xcsw;			/* 12 */
			/* end of supplied parameters: 132 bytes */
		float	xabc[3],xang[3];			/* 24 */
		float	xerrind,xerrx,xerry,xerrphi;		/* 16 */
		float	xresmin,xresmax;			/*  8 */
		char	xisref[20];				/* 20 */
		char	xdir[3][5], xcmd;			/* 16 */
		float	xdelchi,xdelphi,xdelom;			/* 12 */
		short	xisys,xispgp;				/*  4 */
		/* end of XENGEN 1.0- XENGEN 2.0 params: 100 more. */
		float	xstart[3];				/* 12 */
		char	xmoveax, xtemp, xrunno, xchamno;	/*  4 */
		short	xdetwid, xdetht, xdtype, xversion;	/*  8 */
		/* end of portion read and written.
		 These additional parameters are not written or read. */
		float	xast[3], xanst[3], xq21[3][3];		/* 60 */
		} REFPARS;			/* total:	  316 bytes*/

 typedef struct {					/* # of bytes:*/
		float	xto,yto,frame0;				/* 12 */
		short	obkmin,nobk;				/*  4 */
		float	totinten;				/*  4 */
		float	obparts[TROIDWIDTH];			/* 44 */
		} TROID;				/* total:  68 bytes */

 typedef struct {					/* # of bytes:*/
		double	pxcen, pycen, pphicen;			/* 24 */
		double	pinten, pbgnd, psigma, plor;		/* 32 */
		int	pflags, pncells;			/*  8 */
		TROID	ptroi;					/* 68 */
		} PEAK;					/* total: 132 bytes */

 typedef struct { /* structure type for a reflection observation:72 bytes*/
		short	obs_hkl[3];	/* indices of the reflection */
		short	obs_lpa;	/* Lorenz/polarization/absorption
						 	factor * 10000 */ 
		short	obs_ssq;	/*scatt vector length squared * 10**4*/
		short	obs_xpred;	/* predicted x position in pixels */
		short	obs_ypred;	/* predicted y position in pixels */
		short	obs_region;	/* detector partition (0-11) */
		float	obs_ii;		/* integrated intensity */
		float	obs_sigma;	/* e.s.d. of integrated intensity */
		float	obs_iunfit;	/* non-shapefit intensity */
		float	obs_sunfit;	/* non-shapefit e.s.d. */
		float	obs_ompred;	/* predicted scanning angle position */
		float	obs_omob;	/* observed scanning angle position */
		short	obs_gof;	/* goodness-of-fit parameter */
		short	obs_shift;	/* "shift" or "bin" value of obs */
		short	obs_gama;	/* rocking-curve correction */
		short	obs_xob;	/* observed x position in pixels */
		short	obs_yob;	/* observed y position in pixels */
		short	obs_widob;	/* full width at half-max of profile */
		short	obs_bgnd;	/* average background under refl */
		short	obs_bgunfit;	/* background, un-shape-fit */
		short	obs_ncells;	/* number of cells used in summing */
		short	obs_reflno;	/* reflection number (now unused)*/
		char	obs_lowfrm;	/* index of first frame integrated */
		char	obs_hifrm;	/* index of last frame integrated */
		u_char	obs_chamno;	/* chamber number */
		u_char	obs_runno;	/* integer identifier of run */
		short	obs_frmint[PFRAMES];/* background-corrected intensity
					   readings in individual "frames" */
		short	obs_flags;	/* flags set during processing */
		} OBSREF;

 typedef struct { /* defines the (12) regions on the detector face */
		int	toptri, bottri, l1tri, leftri, l2tri, rightri, l3tri;
		} BOUNDARY;

 typedef struct {
		double	cxcen, cycen, cinten, ccutoff;
		short	ux, uy;
		} CTROID;	/* calibration "Troids" structure */

 typedef struct { /* Bravais lattice scoring structure type */
		xglong	lc_latchar, lc_choice;
		double	lc_score, lc_fom;
		float	lc_cel[3][3];
		} LATTICESCORE;

 typedef struct { /* structures for entries in menus (Curses, Motif, ... ) */
		char	*xs_label;
		char	*xs_defau;
		} XMSHOW;

 typedef struct { /* overal descriptor for a menu entry */
	int	pr_nfile, pr_nbool, pr_nsw, pr_nparam, pr_flags;
	char	*pr_name, *pr_title, *pr_label;
	XMSHOW	pr_show[MAXMODIFIERS];
		} PMENUSTUFF;

 typedef struct {	/* structure for various shared names */
		char	xmm_misc[512];	/* temporary memory stores */
		char	xmm_idna[128];	/* holds project identifier's name */
			/* WORK, DATA, SCRATCH directory names */
		char	xmm_workna[128], xmm_datana[128], xmm_scrtna[128];
		char	xmm_listna[128];	/* holds `xgensave.lst' */
		char	xmm_pname[128], xmm_ptitle[256];
		char	xmm_miscb[2560], xmm_miscc[2048];
		} XGNAME;

 typedef struct { /* structure for SHELX-style file */
		int	x_hkl[3]; 
		double	x_inten, x_sigma;
		short	x_batch;
		} SHEL_HKL;

 typedef union {	/* union for proper pointer and integer alignment */
		unsigned int	ui;
		u_char		*ucpt;
		u_short		*uspt;
		void		*vpt;
		void		**vdpt;
		} PAU;

/* system-level declarations, appropriately latched */
#ifndef _H_STANDARDS
extern void	sync(void);
extern ssize_t	read(int fd, void *buf, size_t nbytes);
extern ssize_t	write(int fd, const void *buf, size_t nbytes);
extern int	close(int fd);
extern int	access(const char *path, int mode);
extern int	link(const char *name1, const char *name2);
extern off_t	lseek(int fd, off_t offs, int whence);
#ifndef	__stdlib_h
extern void	exit(int status), free(void *ptr);
extern void	qsort(void *base, size_t nmemb, size_t size,
		 int (*compar)(const void *, const void *));
extern void	*calloc(size_t count, size_t size);
extern void	*malloc(size_t count);
extern void	*realloc(void *ptr, size_t size);
extern int	abs(int j), atoi(const char *str);
extern char	*getenv(const char *name);
#endif
#ifndef X11
extern int	creat(const char *path, mode_t mode);
extern int	open(const char *path, int flags);
#endif
#endif
#include	<string.h>

/* (deep breath here) function prototypes */

extern BOUNDARY *make_boundaries(int beamx, int dwidth, int dheight);
extern FILE *plotsetup(int ndep, int plflags, int *errp, char *finame, char *pltitle, char *xlabel, char *ylabel, char *legend[]);
extern FILE *propen(char *name);
extern FILE *xgopen(char *name, char *mode);
extern FRAMEINDEX *ftables(int opno);
extern ICEWORK *manageice(int opno);
extern LATTICESCORE *allbravais(double lareal[3][3], int *nlattp);
extern LATTICESCORE *managelca(int opno);
extern PMENUSTUFF **pmall(int *nprogp);
extern REFPARS *safepars(int *rtypep);
extern VSPLINE *cleanspline(VSPLINE *oldspl, int *resultp);
extern VSPLINE *gspline(int opno, int *errtp);
extern XGNAME *getproject(char *projnam, int fromcu);
extern XGNAME *sharename(int whichop);
extern char *filepart(char *name, int extension_code);
extern char *findpath(char *str);
extern char *findsw(int aqc, char *aqv[], char swf);
extern char *gcenfi(int cenind);
extern char *gcentering(int latva);
extern char *get_xgdoc(int gxflags);
extern char *glatfi(int latind);
extern char *glatfi4(int latind);
extern char *glatt4(int latva);
extern char *glattice(int latva);
extern char *grab_template(int fromcu);
extern char *indexname(int frmnum);
extern char *intltolc(short thisispgp);
extern char *journalname(void);
extern char *lineend(FILE *fp, int doprint);
extern char *smartdefa(char *title, int killem, int *ctypep);
extern char *spacegname(short thisispgp);
extern char *spacepname(short thisispgp);
extern char *squeezeall(char *str);
extern char *trans_envname(char *str);
extern char *trname(char *name);
extern char *xgprogname(char *inname, int nflag);
extern char **flagproc(int ac, char *av[], char *choices);
extern char **str2acav(char *st, int *acp);
extern double ddeter(double xi[3][3]);
extern double ddotprod(double *v1, double *v2, int n);
extern double dflimits(double z, double low, double high);
extern double dimscale(short indimen);
extern double merge_frames(MAIL *headp, int *nom_bgnd);
extern double mycbrt(double rx);
extern double needlambda(void);
extern double newtcube(double root, double a2, double a1, double a0);
extern double pirange(double x);
extern double *remap_setup(VSPLINE *losp, double *flp);
extern float deter(float xi[3][3]);
extern float dotprod(float *v1, float *v2, int n);
extern float flimits(float z, float low, float high);
extern float gen_centroid(float *vd, float *vv, int nx, int ny, int nz, double *vc, double *vob);
extern float mymedian(float *flb, int npt);
extern float to360(float rv);
extern ino_t record_inode(char *nam);
extern int allnumeric(char *s);
extern int basic_names(int fromcu);
extern int calinv(FILE *fp, VSPLINE *lspl, int calverbose);
extern int cellconvert(REFPARS *xp, int convtype);
extern int centercode(int latva);
extern int centolim(int xce, int yce, int zce, int xyzl[]);
extern int check_directories(int ntocheck, int maketoo);
extern int check_sys(short oldspaceg, short lsys);
extern int cm2pix(float xcm, float ycm, float *xp, float *yp, int fno);
extern int cm2pixq(float xcm, float ycm, int lquad, float *xp, float *yp, int fno);
extern int comparemid(char *fnam, int fde, int sizfil, int verbo);
extern int compressiontype(char *nam, int *lolenp);
extern int create_direc(char *dirname);
extern int detdimen(int dimcode);
extern int detregion(float xva, float yva, VSPLINE *lspl);
extern int dettype(int dimen[]);
extern int dir_access(char *name, int mode);
extern int dinv3(double xi[3][3], double xinv[3][3]);
extern int eigvec(double m[3][3], double elam, double *evec);
extern int envtoname(char *namein, char *nameout);
extern int envtopol(double *polfp, double *polnp);
extern int execlognames(int fromcu);
extern int expandit(char *nam, int ftyp);
extern int find_dimension(int *widp, int *htp, int imagefirst, char *finame);
extern int findbool(int aqc, char *aqv[], char fla);
extern int fmtoupr(xglong refrunflag, REFPARS *rprp);
extern int foffset(int opno);
extern int get_3dprof(FRAMEINDEX *lfmi, int imfla, int *xyzl, MAIL *bgmp);
extern int get_nominal(char *str, char **eprefix, char **otemplate);
extern int getorthog(double x[3][3]);
extern int getregion(short ix, short iy, /* float iphi, */ BOUNDARY *bdp);
extern int grab_base(void);
extern int grab_interval(void);
extern int hklcmp(u_char *h0, u_char *h1);
extern int image_error(FILE *fp, char *name, int error_code, int disposition);
extern int imagelike(char *diren, char *filen);
extern int imagename(char *filen);
extern int imedian(int *ilb, int npt);
extern int imvalue(MAIL *headp, int ix, int iy);
extern int inv3(float xi[3][3], float xinv[3][3]);
extern int inv_solve(double *ls, double *rs, double *p, int n);
extern int iscentric(u_char *phkl, short spaceg);
extern int isinring(double dspac);
extern int ismarip(int informat);
extern int keephkl(short *lh, short excluco);
extern int lattcode(int latva);
extern int locsysc(char *stri, int shouldwait);
extern int managekill(int inglob);
extern int mat2vecs(float rmat[3][3], float *ra, float *ran);
extern int match_inode(FRAMEINDEX **dfmpi, char *finame);
extern int modspline(VSPLINE *losp, FILE *fp, int outflag);
extern int more_logicals(int newonly);
extern int mydircreate(char *name);
extern int newidname(char *idnam, int idnflags);
extern int ofycmp(VOFLOW *of0, VOFLOW *of1);
extern int parallel(float *v0, float *v1, int n);
extern int pix2cm(float xp, float yp, float *xcm, float *ycm, int fno);
extern int plotadd(FILE *fplo, int ndep, float inde, float *depv);
extern int plotfinish(FILE *fplo, float markv);
extern int plotnow(char *pfn);
extern int polymap(VSPLINE *losp, FILE *fp, int reflags);
extern int position_code(float xpix, float ypix);
extern int process_total(FILE *fp, double tcts, double qva, int inw, int inh, int bfl);
extern int qerror(AVOID *pvalue, char *errstr);
extern int read_box(char *name, int flags, MAIL *mailp, int xylim[], int *xybu);
extern int read_image(char *name, int flags, MAIL *mailp);
extern int ready(char *app);
extern int reducedcell(float are[3][3]);
extern int refparin(REFPARS *xp, int *nfoundp);
extern int refparout(REFPARS *xp);
extern int reg2hab(int ixpx, int iypx, double *ov);
extern int remap_process(VSPLINE *losp, double *adel, FILE *fp, int dowrite);
extern int remap_sums(VSPLINE *losp, double *xyv, double *adel);
extern int rp2vecs(REFPARS *rpp);
extern int set_logical(char *lname, char *newvalue);
extern int shorten(char *myname, char *logdir);
extern int shrinkit(char *nam);
extern int spatial(int owidth, int oheight, int verbosity, MAIL *thd);
extern int startspline(int fromcur);
extern int strcend(char *st1, char *st2);
extern int timecvt(xgulong dt[], char *ist, char *ost, time_t *tl, int ity, int ott);
extern int ucell(float abc[], float abg[], float ast[], float angst[], short isys);
extern int unset_logical(char *lname);
extern int vecmat(float *len, float *ang, float mat[3][3]);
extern int verror(int value, char *errstr);
extern int writelognames(int fromcu);
extern int xeflush(FILE *fp);
extern int xgaccess(char *nam, int mode, int verbosity);
extern int xgbuf(char *str, int bflag);
extern int xgcomargs(int arc, char *arv[], char *xgve[], int (*xgaps[])(int ac, char *av[]));
extern int xgdelete(char *name);
extern int xgenrename(char *inpname);
extern int xgfork(char **comdve, int shouldw);
extern int xgjournal(int ac, char *av[]);
extern int xgread(AVOID *base, int xgtype, FILE *fp);
extern int xgstrend(char *bigstr, char *suffix, int cutit);
extern int xgsuffix(char *stri);
extern int xgvariable(char *suffx, float *valu, int nvalues);
extern int xgwrite(AVOID *ibase, int xgtype, FILE *fp);
extern short exclucode(short spgrp, short sys);
extern short irodown(double x);
extern short iroup(double x);
extern short max2(short a, short b);
extern short min2(short a, short b);
extern short spgtocen(short spg);
extern short spgtosys(short spg);
extern short sswab(short sh);
extern u_char **activmask(int wid, int ht, int opno);
extern u_short uswab(u_short ush);
extern void adjust_to_fm(float *rxp, float *ryp, int fno, int direction);
extern void artocel(float *stang, float are[3][3], float cell[]);
extern void cleanorthog(double q[3][3]);
extern void condpr(FILE *fp, float fval, int nspace, int nafter, int forspace);
extern void cpymem(AVOID *mo, AVOID *mi, unsigned nb);
extern void describe_rings(FILE *fp);
extern void disp_3dprof(FILE *fp, int xyzl[]);
extern void dmm3(double c[3][3], double a[3][3], double b[3][3]);
extern void dmm31(double *y, double a[3][3], double *x);
extern void dmm3t(double c[3][3], double a[3][3], double b[3][3]);
extern void flcvt(u_char *z, int direction);
extern void float_cvt(u_char *z);
extern void get_calibparams(int dty, double *xspap, double *yspap, int *linp);
extern void globalkill(void);
extern void head_results(FILE *fp);
extern void headbravais(FILE *fp);
extern void header(FILE *fp, char progtype, char **iolist, char *ftypelist);
extern void helpapp(char *pname);
extern void hunpack(xglong phkl, short uhkl[]);
extern void inv_add(double *ls, double *rs, double *av, double b, int n);
extern void inv_init(double *ls, double *rs, int n);
extern void kill_mail(MAIL **headdp, int *nmboxp);
extern void kill3d(void);
extern void lowercaseify(char *str);
extern void make_upper(char *name);
extern void makeq21(REFPARS *xp);
extern void memzap(AVOID **base, char *str);
extern void mm3(float c[3][3], float a[3][3], float b[3][3]);
extern void mm31(float *y, float a[3][3], float *x);
extern void mm3t(float c[3][3], float a[3][3], float b[3][3]);
extern void mmt3(float c[3][3], float a[3][3], float b[3][3]);
extern void mmt31(float *y, float a[3][3], float *x);
extern void mpack(short hup[], u_char hp[], short intlno);
extern void munpack(u_char *hp, short *hup, short intl);
extern void mygetwd(char *bufr);
extern void nomoredv(void);
extern void polyadd(VSPLINE *losp, double *wve);
extern void pr3m(FILE *fp, float x[3][3], char *s);
extern void printshel(FILE *fp, SHEL_HKL *shp);
extern void prtroid(FILE *fp, TROID *trp);
extern void prurefl(FILE *fp, OBSREF *urp, double *scal);
extern void pspline(FILE *fp, VSPLINE *splip);
extern void qtoangs(double q[3][3], double *rph);
extern void recen_vec(float *v, double *delv, float *wve, int nx, int ny, int nz);
extern void rot9r(float rea[3][3], float *rfi);
extern void show_bravchoice(FILE *fp, float *stang, LATTICESCORE *lcap);
extern void show_params(FILE *fp, REFPARS *rpp, int nbr);
extern void showbravais(FILE *fp, float remat[3][3], float *stang);
extern void shrink(int npixels, int dewidth, int deheight);
extern void sum_row(FILE *fp, int row, int iwid);
extern void swap(AVOID *b0, AVOID *b1, unsigned dwidth);
extern void symtr(u_char sp[], short scode);
extern void transfer_base(char *stri);
extern void transfer_interval(char *stri);
extern void transfer_template(char *stri);
extern void versionapp(char *pname);
extern void writeindex(FRAMEINDEX *fmix, int fromcu);
extern void zeromem(AVOID *base, size_t nbytes);
extern void ztos(float *rusr, float zs[3][3]);
extern void *callop(size_t nitems, size_t sizitem, char *str);
extern void *mallop(size_t nbytes, char *str);
extern void *reallop(AVOID *oldptr, size_t nbytes, char *str);
extern xglong hpack(short *lh);
extern xglong lswawb(xglong lv, int direc);
extern xgulong swawb(xgulong ulv);
 
#endif /* XGEN_H */

/* end of xgen.h */
