/* xgenpred.h -- preprocessor defines and typedefs for reflection prediction

	Copyright (c) 2004, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	28 Nov 2006: delete po_q21, po_rabc, po_rang elements from PREDWORK
	23 Jun 2006: PREDREF structure typedef moved here
	 3 Jan 2006: add po_dminspec, po_dmaxspec elements to structure
	17 Jan 2005: eliminate po_npts from structure; not being used.
	 9 Sep 2004: eliminate po_{nranges,deltamu,mu1,mu2} from structure
	31 Aug 2004: created, in part from ../nogui/predict.c

   Purpose: Providing #defines, typedefs, and function declarations
	for predicting reflections
   Cross References: 

********************************************************************/
#ifndef XGENPRED_H
#define XGENPRED_H	1

#include	"xgen.h"
#define		NRESTYPES	11	/* number of fates, below... */
/* defines for the various fates of an attempt at prediction */
#define		PR_XYRANGE	0	/* X,Y out of range */
#define		PR_FIRSTPHIOK	1	/* first  phi value  predicted */
#define		PR_SECPHIOK	2	/* second phi value  predicted */
#define		PR_BOTHPHIOK	3	/* both   phi values predicted */
#define		PR_PHIRANGE	4	/* phi outside (mu1->mu2) */
#define		PR_BADY2	5	/* nonsense value of "y2" */
#define		PR_ALONGBEAM	6	/* scatt vector along beam */
#define		PR_TOOLOWRES	7	/* scatt vector too small */
#define		PR_TOOHIRES	8	/* scatt vector too large */
#define		PR_BADHKL	9	/* illegal indices in spacegroup */
#define		PR_SZRANGE	10	/* z component of s out of range */

typedef struct {	/* structure describing a predicted reflection */
		short	pfmno; /* integer frame number of center of refl */
		float	pxt,pyt,pphi; /* predicted (X,Y,omega) centroid */
		xglong	pindx;	/* compressed index value for reflection */
		float	pcor,pgama; /* LP correction and predicted width */
		} PREDREF;

typedef struct { /* prediction operational structure */
		short	po_excluc, po_lmin, po_lmax;		/*  6 bytes */
		short	po_ih[3], po_ph[3];			/* 12 bytes */
		short	po_nresult[NRESTYPES+1];		/* 24 bytes */
		xglong	po_nfound, po_nchecked;			/*  8 bytes */
		xglong	po_totfound, po_totchecked;		/*  8 bytes */
		time_t	po_tl0, po_tl1;				/*  8 bytes */
		double	po_al, po_hslop, po_hsq, po_miny2sq;	/* 32 bytes */
		double	po_maxxcm, po_maxycm;			/* 16 bytes */
		double	po_minxcm, po_minycm;			/* 16 bytes */
		double	po_mumin, po_mumax, po_polfrac;		/* 24 bytes */
		double	po_polnormrot, po_rangecen, po_rhomax;	/* 24 bytes */
		double	po_smasq, po_smisq, po_szmin, po_szmax;	/* 24 bytes */
		double	po_y2co, po_yzco, po_z2co;		/* 24 bytes */
		double	po_dminspec, po_dmaxspec;		/* 16 bytes */
		float	po_orec[3][3], po_p[2][3][3], po_ps[7];	/*136 bytes */
		REFPARS	*po_xp;					/*  4 bytes */
		VSPLINE	*po_spl;				/*  4 bytes */
		PREDREF	*po_prp, *po_prp0;			/*  8 bytes */
	 } PREDWORK; /* total size: 310 bytes */

/* function prototypes */

extern int dphicmp(PREDREF *pp0, PREDREF *pp1);
extern int pfindref(PREDWORK *pwp);
extern int prloop(PREDWORK *pwp, int prverbose);
extern int prpar(PREDWORK *pwp);
extern void prpred(FILE *fp, PREDREF *prp);
extern PREDWORK	*manageprw(int opno);

#endif /* XGENPRED_H */

/* end of xgenpred.h */
