/* xgpostint.h -- typedefs & declarations for reduction, scaling, rejection

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	21 Apr 2007: QOB structure elements added to facilitate fast calcs
	19 Apr 2007: float->double conversions
	20 Dec 2006: eliminate RFACTOR typedef; SPACEGSTAT typedef moved away
	 6 Dec 2006: reorganize MUREFL and add routines muinterp() and outmu()
	27 Nov 2006: remove os_q21 from OUTOBSSTYLE, sta_rabc,rang from STATRUN
	20 Oct 2006: add decls for rejswitch, autoreject; delete rejstart
	22 Jun 2006: merged reddefs,scdefs,rejdefs,statdefs.h, some of xgen.h
 Previous incarnations:
	13 Jun 2006: reddefs.h created from redurefl.c and ../xglib/xgenmisc.c
	 6 Jun 2006: end_stats deleted: it's now static
	 6 Jun 2006: numerous routines now static with creation of sccycles()
	16 Jan 2006: fix SUMSTRONG_ABG construct
	19 Oct 2005: add more SCALERUN elements to eliminate static globals
	17 Sep 2004: additional elements in SCALERUN to handle quick summaries
	17 Sep 2004: increase max number of resolution bins from 5 to 10
	27 Aug 2004: mm_vecrei[] and mm_matrei[][] added to MAKEMURUN struct
	13 Aug 2004: use OUT_OBSSTYLE structure to pass args around;
		SCANREF structure type used for old data format
	 8 Aug 2004: add mm_fpsca and mm_fpasca file pointers to MAKEMURUN
	20 Jun 2004: Boolean flag bits reorganized and expressed in hex;
			default is now SC_MERGEBJ | SC_ALLOFDET
	29 May 2004: makemu flags and structure types here, not in makemu.c;
		add rj_qfrac and rj_qrval elements to REJECTRUN structure
	27 May 2004: move SPACEGSTAT structure and spaceg flags to statdefs.h;
			NSRFACS replaces NRFACS
	21 Jan 2004: license message inserted
	 6 Jan 2004: SCALERUN,REJECTRUN,STATRUN structure types enlarged
	25 Nov 2003: add SCALERUN,REJECTRUN,STATRUN structure types
	20 Sep 2003: remove function declarations from here
	 5 Jan 2003: reorganized source-file header
	10 Nov 2002: added elements sh_nrscaled and sh_nobscaled to SHLSTAT
	10 Oct 2002: add rp_ar0 element in RUNWORK structure for "spaceg"
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	31 May 1995: w_omxy and del_omxy are xgulongs, not u_longs
	31 May 1995: s_nob, rf_not, anr rf_nref are ints, not longs
	31 May 1995: w_n,w_niosig, sh_n*, r_n* structure elements
		are ints, not longs
	 2 Mar 1995: cross-references added to scdefs.h,rejdefs.h,statdefs.h
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	14 Nov 1994: add runflags bits for
		SC_SHRINKIFUP, SC_NONLINEAR, SC_SWITCH
	 4 Nov 1994: add SC_FLATVARIANCE option so sigmas don't have to
		depend on R-factor of scalegroup;
		otherwise variances are augmented by (I * sigma(g))^2
	23 Sep 1994: add new bits to runflags
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN versions of scdefs.h,rejdefs.h,statdefs.h
   Purpose: typedefs, includes, defines for post-integration steps
   Cross references: ../utilities/redu{refl,tils}.c, ../postprocess/[a-z]*.c
*/
#ifndef XGPOSTINT_H
#define XGPOSTINT_H	1
#define		NEED_UNISTD	1
#define		NEED_TIME	1
#define		NEED_DESCRIP	1
#define		NEED_SSDEF	1
#include	"xgen.h"
	/* miscellaneous definitions */
#define		MACHINE_FAC	(0.625e-3)
#define		MINSIGI		0.01	/* minimum scaled sigma(I) value*/
#define		MINVAR		(MINSIGI * MINSIGI)
#define		NFUNCS		9	/* # of operations the user can do */
#define		MESSAGETIME	2	/*# of seconds between "working" msgs*/
#define		NXBINS		100	/* # of bins in x for x,y displays */
#define		NYBINS		50	/* # of bins in y for x,y displays */
#define		CONV_RESC	(0.1e-5)	/* rescale convergence factor*/
#define		DVCSCALE	(3.4e-4)	/* scale fac for sph coeffs*/
#define		MINOSHELL	1000	/* min.obs per wilson shell */
#define		MEDIUMRES	(0.04)	/* Medium resolution 1/(d*d) */
#define		MAXWSHELLS	30	/* # of Wilson-plot shells in stats */
#define		MAXRSHELLS	10	/* max # of R-factor shells in stats*/
#define		NOBMAX		7	/* Max. recorded obs/ref in stats*/
#define		NISRANGE	7	/* # of I/sigma bins in stats*/
#define		NYSRANGE	7	/* # of <I>/sigma bins in stats*/
#define		NYRANGE		9	/* # of <I> bins in stats */
#define		NSRFACS		5	/* No. of calc'd R-factors in stats */
#define		M_EVEN		0376	/* mask to set a "char" even */
#define		MAXRUNS		63	/* max # of data runs in a dataset */
#define		MAX_RUN_INDEX	160	/* max index # for a data run */
#define		MAXVECPARS	16	/* max # of vector parameters */
#define		REFBYTES	16	/* # of bytes in MERGEDREF before obs*/
	/* flags for reasons for reflection rejection during data reduction */
#define		LAUE_BAD	1	/* 1st reason for rejection */
#define		SCREW_BAD	2	/* 2nd reason for rejection */
#define		STL_TOO_HIGH	3	/* 3th reason for rejection */
#define		STL_TOO_LOW	4	/* 4th reason for rejection */
#define		IOSIGMA_BAD	5	/* 5th reason for rejection */
#define		NOT_INTEGER	6	/* 6th reason for rejection */
#define		BAD_X_Y		7	/* 7th reason for rejection */
#define		FIT_NE_SUM	8	/* 8th reason for rejection */
#define		INDEX_CHANGED	9	/* 9th reason for rejection */
#define		NONSENSE_HKL	10	/* 10th reason for rejection */
#define		DEL_XYZ		11	/* 11th reason for rejection */
#define		ICERING		12	/* 12th reason for rejection */
#define		BADTYPES	13	/* total # of reasons for rejection */
	/* values of the `ru_inttype' flag: */
#define		SIMPLE_SUMMATION	1
#define		PROFILE_FITTED		2
#define		BEST_I_OVER_SIGMA	4
#define		SCAN_FORMAT		010
#define		ABG_EITHER		020
#define		SUMMED_IF_STRONG	0100
#define		ABG_SUM			(ABG_EITHER | SIMPLE_SUMMATION)
#define		ABG_FIT			(ABG_EITHER | PROFILE_FITTED)
#define		SUMSTRONG_ABG		(SUMMED_IF_STRONG | ABG_EITHER)
	/* bits in the `oob->os_rflags' flag */
#define		RD_DOSORT		1	/* do sort & merge the output*/
#define		RD_BADFOUND		2	/* already found >0 bad obs */
#define		RD_POLARIZED		4	/* need to unpolarize Lp's */
#define		RD_IOSCUT		010	/* make I/sigma cut */
#define		RD_MAKE_CENTROIDS	020	/* CENTROIDS,not MULTIREF out*/
#define		RD_DEL_BIG_IDIFF	040	/* delete big |Ifit-Isum| */
#define		RD_SCALE_SUMFIRST	0100	/* rescale summation first */
#define		RD_REINDEX		0200	/* reindex with 3x4 matrix */
#define		RD_RECOMPUTE		0400	/* get hkl's from UPARAMS */
#define		RD_ANGLE_RESTRICTION	01000	/* only output spec'd frames */
#define		RD_PRINTALL		02000	/* print ALL deletions */
#define		RD_KILL_NOT_INT		04000	/* delete if h-[h] big */
#define		RD_KILL_IF_REINDEXED	010000	/* delete if reindexed */
#define		RD_KILL_DELXYZ		020000	/* del.if |Xobs-Xpred| big */
#define		RD_ZOPTION		014400	/* combination for `z' option*/
#define		RD_KILL_ICERING		040000	/* del if refl in ice ring */
#define		RD_SYNCHRO_V		0100000	/* vertical-omega synchrotron*/
#define		RD_SYNCHRO_H		0200000	/* horiz-omega synchrotron*/
#define		RD_SYNCHRO		0300000	/* either synchrotron */
#define		RD_STLCUT		0400000	/* sin(theta) cutoff */
#define		RD_KEEPDUPS		01000000 /* don't eliminate dup obs. */
	/* values for the "w_flag" entry in the WORKINGOB structure */
#define		W_NOT_SCALED	0
#define		W_UNMARKED	0
#define		W_ORDINARY_OB	1
#define		W_ISBIJVOET	2
#define		W_ISCENTRIC	4
#define		W_1ST_OF_TYPE	010	/* on if it's the 1st primary, or the
					1st Bijvoet, or the 1st centric*/
#define		W_HANDMARKED	020
#define		W_AUTOMARKED	040
#define		W_PREMARKED	0100
#define		W_ISMARKED	0160	/* OR'ing of the marking types */
	/* values for the bits in the `runflags' variable */
#define		RJ_SEPBJ	01	/* separate Bijvoets */
#define		RJ_MACHINESIG	02	/* tack on machine sigma */
#define		RJ_RESUME	04	/* start up run partway thru multiref*/
#define		RJ_BATCH	010	/* don't set up for interactive stuff*/
#define		RJ_ALLCALC	020	/* do R-factor on all reflections */
#define		RJ_VERBOSE	040	/* more stuff in logfile */
#define		RJ_PRESERVEOLD	0100	/* preserve input deletions */
#define		RJ_HUMANAMBIG	0200	/* human fixes ambiguous cases */
#define		RJ_HUMANBAD	0400	/* human fixes all bad observations */
#define		RJ_HUMANALL	01000	/* human looks at ALL observations */
#define		RJ_HUMANANY	01600	/* or'ing of last three phenomena */
	/* flags related to reflections with exactly two observations: */
#define		RJ_2_SKIP	02000	/* don't examine them at all */
#define		RJ_2_DISPLAY	04000	/* user examines all such refls */
#define		RJ_2_MINSIG	010000	/* keep ob with minimum sigma */
#define		RJ_2_KILL	020000	/* delete both observations */
#define		RJ_2_MAXI	040000	/* keep ob with maximum intensity */
#define		RJ_2_MAXIOS	0100000	/* keep ob with max I/sigma */
#define		RJ_2_CHOICES	0176000	/* or'ing of all those bits */
	/* bits in the run-condition flag value for `makemu' */
#define		MUF_OUTPUT_F	01	/* write out F's, not I's. */
#define		MUF_NONEGI	02	/* set I<0 to 0 */
#define		MUF_PRESERVE	04	/* preserve sin(theta) dependence */
#define		MUF_ABSORDER	010	/* output refls in absolute order */
#define		MUF_STLORDER	020	/* output refls in sin theta order*/
#define		MUF_REINDEX	040	/* refls to be reindexed on output*/
#define		MUF_REORDER	030	/* any re-sorting possibility */
#define		MUF_VAROFMEAN	0100	/* alternate formula used for
						variance of mean */
#define		MUF_MULOBSERVED	0200	/* only write multiply-obs. refls */
#define		MUF_XPLORSEP	0400	/* output in X-PLOR format */
#define		MUF_XPLORMER	01000	/* X-PLOR format with no Bijvoet sep */
#define		NRFACS	7	/* # of different R-factors computed */
#define		SELECTIVITY	3	/* min redundancy for skipping refls*/
	/* flags associated with spherical-harmonic scaling */
#define		SC_DIFFRACTION	0100	/* scale with diffraction vectors */
#define		SC_EXITBEAM	0200	/* scale with exit-beam vectors */
#define		SC_VECREFINED	0300	/* do some vector-dependent scaling*/
#define		SC_VECONLY	04000	/* flag: only return the 3-vector */
	/* masks for Boolean flags in scaling */
#define		SC_ALLOBS	0x1	/* use all obs, even in unscaled shifts
						(default off) */
#define		SC_KEEPCYC	0x2	/* keep cycling after convergence
						(default off) */
#define		SC_MERGEBJ	0x4	/* merge Bijvoets (default on) */
#define		SC_WILSON	0x8	/* initialize scales by
					   Wilson statistics (default off) */
#define		SC_VERBOSE	0x10	/* verbose printouts (default off) */
#define		SC_ALLOFDET	0x20	/* merge top & bottom of detector */
#define		SC_FLATVARIANCE	0x40	/* don't augment variances by
					 (I*sigma(g))^2 */
#define		SC_DIFF_NOS2	0x80	/* special case:
					diff.vector scaling but ssq term==0*/
#define		SC_NONLINEAR	0x100	/* nonlinear scaling! */
#define		SC_SWITCH	0x200	/* switch to nonlinear if R goes up */
#define		SC_INPUTSCALE	0x400	/* valid input scale factors */
#define		SC_DEFAULT	(SC_MERGEBJ | SC_ALLOFDET)	/* default */
/* default values of numerical command-line input values */
#define		SC_IMIN		0	/* min intensity for use = 0 */
#define		SC_NOBSMIN	2	/* min # observations for use = 2 */
#define		SC_MAXCYCLES	9	/* max # of scaling cycles = 9 */
#define		SC_CONVCRIT	100	/* default convergence criterion; i.e.,
	if |g(new)-g(old)| < 0.001 for all scales, we've converged) */
#define		SC_NPARAMS	1	/* default # of params per shift */
	/* bit values for the `runflags' variable */
#define		ST_SEPBJ	01	/* separate Bijvoets for scaling */
#define		ST_ISSCALED	02	/* scale the data */
#define		ST_OUTMISS	04	/* output MISSING refls list */
#define		ST_DONE		010	/* we're finished reading */
	/* bits relevant to spaceg in the sta_flags variable */
#define	SP_VERBOSE	020	/* verbose output */
#define	SP_OUTMREF	040	/* output a new multiref */
#define	SP_CONDOUTMREF	0400	/* output multiref only if same point group */
	/* macros */
#define		LSH(x,y)	(((xgulong)x) << y)

	/* structure typedefs */

 typedef struct {		/* abbreviated observation structure:16 bytes*/
		short	abo_shift;		/* "shift" or "bin" number*/
		short	abo_omob;		/* omega observed */
		u_char	abo_chamno;	/* chamber number */
		u_char	abo_runno;	/* integer identifier of run */
		u_char	abo_xob;	/* observed chamber x */
		u_char	abo_yob;	/* observed chamber y */
		float	abo_ii;			/* integrated intensity */
		float	abo_sigma;		/* sigma */
		} ABBROB;

 typedef struct {			/* structure for merged reflection */
			/* 16+MAXOB*16=16+127*16=128*16=2048 bytes */
		u_char	mr_hkl[4];	/* hkl of reflection */
		u_short	mr_nob;		/* # of obs total */
		u_char	mr_nobpri; /* # of obs. of primary refl.*/
		u_char	mr_nobbij; /* # of obs. of Bijvoet refl.*/
		u_short	mr_ibar; /* averaged, scaled intensity */
		u_short	mr_isig; /* esd of averaged, scaled intensity*/
		float		mr_stl;		/* sin theta over lambda */
		ABBROB	mr_ob[MAXOB];	/* array of observation structures */
		} MERGEDREF;

 typedef struct {		/* structure for one-ob "merged" reflection */
				/* 16+16=32 bytes */
		u_char	fo_hkl[4];	/* hkl of reflection */
		u_short	fo_nob;		/* # of obs total */
		u_char	fo_nobpri;	/* # of obs. of primary refl.*/
		u_char	fo_nobbij;	/* # of obs. of Bijvoet refl.*/
		u_short	fo_ibar; /* averaged, scaled intensity */
		u_short	fo_isig; /* esd of averaged, scaled intensity*/
		float		fo_stl;		/* sin theta over lambda */
		ABBROB		fo_ob;		/* observation structure */
		} FASTOB;

 typedef struct { /* a structure used in handling shifts (scaling gropus) */
		short 		shval;		/* the shift value itself */
		u_char	chamno;		/* chamber number */
		u_char	runno;		/* run number */
		float		shg;		/* scale factor (g) for shift */
		float		sha;		/* coeff of sin(theta)/lambda*/
		float		shb;		/* Debye-Waller factor */
		float		shvar;		/* var(scale)/(scale*scale) */
		short		shomc;		/* omega centroid of group */
		short		shnbr;		/* neighbor flags */
		} SHIFT;

 typedef struct {	/* diffraction or output-vector scaling coeffs */
		int	v_ms;	/* min scaling group associated with these*/
		float	v_coef[MAXVECPARS]; /* the coefficients themselves*/
		char	v_ptype;	/* type of scaling it goes with */
		} VECPAR;


 typedef struct { /* structure type for printed MULIST reflection: 44 bytes */
		short	mu_hkl[3];	/* signed hkl values */
		short	mu_rffu;	/* in case we need it later */
		float	mu_stl;		/* sin(theta)/lambda */
		float	mu_f, mu_sig;	/* <F>, sigma(<F>) */
		float	mu_fplus, mu_sigplus;	/* <F>,etc for primary obs */
		float	mu_fminus, mu_sigminus;	/* <F>,etc for Bijvoet obs */
		sichar	mu_flag;	/* flag */
		sichar	mu_extra[3];	/* in case we need them */
		u_char	mu_uhkl[4];	/* unsigned char version of indices */
	       } MUREFL;

 typedef struct {
		u_char		del_hkl[4];	/* packed hkl of reflection */
		xgulong		del_omxy;	/* packed omega, X, Y */
		short		del_shift;	/* shift value */
		u_char		del_chamno, del_runno;
		} DELETEDOB;

 typedef struct { /* Structure for output from old BUDDHA-style data */
		int	rhkl;					/*  4 bytes */
		float	rxcen, rycen, rzcen;			/* 12 bytes */
		short	rnspot, rnback;				/*  4 bytes */
		short	rback1, rback2, rspotz1, rspotz2;	/*  8 bytes */
		short	rback3, rback4;				/*  4 bytes */
		short	rmaskno, roffset;			/*  4 bytes */
		float	rsumspot, rvar_spot, rsumback;		/* 12 bytes */
		float	rvar_back, rfidlp;			/*  8 bytes */
		} SCANREF;				/* Total:  56 bytes */

 typedef struct { /* keeps track of output styles derived from OBSREFs */
		short	os_delxlim, os_delylim, os_ncbx, os_ncby, os_ncperbin;
		short	os_ncbins, os_minfm, os_maxfm, *os_cenpbin;
		xglong	os_outobs, os_rflags;
		float	os_delomlim, os_xdel, os_ydel, os_xscal, os_yscal;
		VSPLINE	*os_spl;
		} OUT_OBSSTYLE;

 typedef struct { /* operational parameters pased among functions */
		int		ru_hto2, ru_htrat, ru_inttype, ru_maxeven;
		int		ru_nchanged, ru_nconverted, ru_nduptoss;
		int		ru_npairs, ru_nread, ru_outrefls;
		int		ru_representative, ru_shoffset, ru_widrat;
		int		ru_nbad[BADTYPES];
		time_t		ru_tl0, ru_tl1;
		short		ru_excluc, ru_vecrei[3];
		float		ru_delfrm, ru_delilim, ru_fmpergroup, ru_ifit;
		float		ru_ioscut, ru_isum, ru_lamco, ru_lordel;
		float		ru_machineerr, ru_maxinten, ru_maxpr, ru_minpr;
		float		ru_minsigoi, ru_maxslegal, ru_minslegal;
		float		ru_nomwidth, ru_omegainterval, ru_rmingiven;
		float		ru_rmaxgiven, ru_sigfit, ru_sigsum;
		float		ru_stlioscut, ru_strongcut, ru_sumscale;
		float		ru_sumscsq, ru_trigc;
		float		ru_iose[5], ru_sloc[3], ru_matrei[3][3];
		FILE		*ru_fpi, *ru_fpl, *ru_fpo;
		FASTOB		*ru_orefb, *ru_orefend;
		REFPARS		*ru_rpp;
		CRYSTAL		*ru_xhp;
		DATARUN		*ru_rhp;
		OUT_OBSSTYLE	*ru_oob;
	 } REDUWORK;

 typedef struct {	/*structure type for calculating diffraction vectors */
		int	rp_shadd;
		float	rp_d, rp_dsq, rp_x0, rp_y0;	/* detector params */
		float	rp_delfm;	/* angular stepsize in radians */
		float	rp_q21[3][3];	/* tilt&swing matrix */
		float	rp_qz[3][3];	/* orientation-zeroing matrix */
		float	rp_areal[3][3];	/* realspace unit cell matrix*/
		float	rp_ar0[3][3];	/* realspace matrix in primitive cell*/
		} RUNWORK;

 typedef struct {	/* structure for working observation */
		short		w_shift;	/* shift */
		short		w_omob;		/* omega observed */
		short		w_xob;		/* x observed */
		short		w_yob;		/* y observed */
		float		w_isc;		/* scaled intensity */
		float		w_sigsc;	/* scaled sigma */
		float		w_gofp;		/* g.o.f.with this ob deleted*/
		float		w_dev;		/* (I-<I>)/ sigma */
		float		w_worst;	/* worst dev with this ob del.*/
		float		w_e0, w_e1, w_e2;	/* summations with this
						observation deleted */
		xgulong		w_omxy;		/* packed omega,x,y */
		u_char		w_flag;		/* flag: see above */
		u_char		w_rffu;		/* reserved for future use */
		char		w_chamno;	/* chamber number */
		char		w_runno;	/* run number */
		} WORKINGOB;

 typedef struct {
		int		rj_xbinwid, rj_ybinwid, rj_nobused, rj_nobaf;
		int		rj_flags, rj_scflags, rj_nvpars, rj_nscales;
		int		rj_htrat, rj_hto2, rj_widrat;
		int		rj_tl0, rj_tl1;
		unsigned	rj_nobmarked, rj_onobused, rj_onobout;
		short		*rj_ndshift;
		u_char		*rj_ncp0, rj_shkl[4];
		float		rj_e0, rj_e1, rj_e2;
		float		rj_ibar, rj_sigysq, rj_sig, rj_gof, rj_worst;
		float		rj_ibaraf, rj_sigaf;
		float		rj_bestgofp, rj_gofaf, rj_worstaf;
		float		rj_rfacin, rj_rfacout, rj_rfdenin, rj_rfdenout;
		double		rj_qval, rj_qval2, rj_qfrac, rj_qrval;
		FILE		*rj_fpmi, *rj_fpcheck, *rj_fpl, *rj_fpmo;
		CRYSTAL		*rj_xhdp;
		DATARUN		*rj_rhdp;
		SHIFT		*rj_shlp;
		VECPAR		*rj_vshp;
		RUNWORK		*rj_rub;
		WORKINGOB	*rj_wobs, *rj_wobend, *rj_lastdis;
		DELETEDOB	*rj_drefs, *rj_drefend, *rj_drefp;
		MERGEDREF	*rj_mrp;
		short		rj_lhkl[3];
		char		rj_somedecisions;
		} REJECTRUN;

 typedef struct {	/* structure type for info on a shell of data */
		int	sh_nob, sh_nref, sh_index;
		float	sh_int, sh_scint, sh_iratio, sh_resmin, sh_resmax;
		float	sh_gof;
		} SHELL;

 typedef struct {
		char		*mm_mrname, *mm_cnsname, *mm_shelname;
		int		mm_flags;
		unsigned	mm_nscales;
		short		mm_vecrei[3];
		time_t		mm_tl0;
		float		mm_oversc, mm_oversq, mm_scut, mm_specres;
		float		mm_matrei[3][3];
		FILE		*mm_fpl, *mm_fpmi, *mm_fpo, *mm_fpx;
		FILE		*mm_fpc, *mm_fpshel, *mm_fpsca, *mm_fpasca;
		CRYSTAL		mm_xhed;
		MUREFL		*mm_mur0, *mm_mup;
		SHIFT		*mm_shlp;
		VECPAR		*mm_vshp;
		RUNWORK		*mm_rub;
		SHELL		*mm_shlb;
		} MAKEMURUN;

 typedef struct {	/* structure for "quick" temporary observations */
		unsigned short	qo_shift; /* index of shift */
		unsigned short	qo_dvgrp; /* index of diffraction-vector group*/
		float		qo_sigma; /* e.s.d. of intensity */
		float		qo_inten; /* intensity value */
		short		*qo_sc; /* spherical-harmonic coefficients */
		short		qo_omob; /* remember omega of current ob */
		/* added Apr 2007. By convention the actual applied
		 scale factor is the product of the next two. */
		double		qo_lscal;	/* local scale factor */
		double		qo_lharm;	/* spherical-harmonic factor */
		double		qo_iwt;		/* weight factor */
		} QOB;

 typedef struct {
		unsigned	qr_nob;		/* # of obs of this refl */
		float		qr_stl;		/* sin(theta) / lambda */ 
		} QREF;

 typedef struct {
		int	s_nob;	/* # of observations in a shift */
		float	s_res;	/* average resolution of refls */
		float	s_r[7];	/* sums used to compute wilson stats */
		float	s_i000;	/* extrapolated intensity of (0,0,0) */
		float	s_b;	/* Debye-Waller factor */
		float	s_sigi0;	/* sigma (I000) */
		float	s_sigb;		/* sigma (B) */
		float	s_sk;	/* overall scale factor */
		} SHSTAT;

 typedef struct {
		double	y2, iy, i2, y2s2, y2s4, iys2, y2s, y2s3, iys;
		double	newg, newa, newb, rfac, mvar[6];
		int	nobsh;
		} CYCLESHIFT;

 typedef struct {
		short		sc_convcrit, sc_freq, sc_imin;
		short		sc_maxcycles, sc_norders, sc_nparpshift;
		short		sc_nrepmin, sc_overredun;
		xglong		sc_acenobu, sc_acenrefu, sc_cenobu;
		xglong		sc_cenrefu, sc_cycno, sc_flags, sc_ngt0;
		xglong		sc_nk, sc_nkbest, sc_nobin, sc_nobput;
		xglong		sc_nrefin, sc_nrefput, sc_nscales;
		xgulong		sc_ndvgroups, sc_noutpar;
		size_t		sc_totpar;
		time_t		sc_tl0, sc_tl1;
		u_char		sc_lasthkl[4];
		double		sc_appshift, sc_gsca, sc_gsum, sc_gsumden;
		double		sc_lastf, sc_maxdi, sc_maxqf, sc_meanqf;
		double		sc_minqf, sc_minshiftv, sc_ovrf, sc_ovrfden;
		double		sc_qfden, sc_r1, sc_r3, sc_resq4, sc_rmsde;
		double		sc_bar, sc_barsq, sc_scrnow, sc_scrold;
		double		*sc_best0, *sc_del0, *sc_max0;
		double		*sc_min0, *sc_new0, *sc_now0, *sc_prev0;
		double		*sc_s0, *sc_savea, *sc_saveb, *sc_saveg;
		double		*sc_savesh;
		double		sc_maxco[10], sc_maxp[10];
		double		sc_minco[10], sc_minp[10], sc_pout[4];
		double		sc_rfd[NRFACS], sc_rfn[NRFACS];
		double		*sc_avec, **sc_rb, **sc_yb;
		FILE		*sc_fpmi, *sc_fpl;
		CRYSTAL		*sc_xhdp;
		DATARUN		*sc_rhdp;
		SHIFT		*sc_shlp;
		VECPAR		*sc_vshp;
		SHSTAT		*sc_shwl, sc_alls;
		RUNWORK		*sc_rub;
		QOB		*sc_obbase, *sc_oblist;
		QREF		*sc_refbase, *sc_reflist;
		CYCLESHIFT	*sc_g;
		} SCALERUN;

 typedef struct {	/* structure used in getting Wilson statistics */
		int	w_n, w_niosig[NISRANGE+1];
		float	w_dmin, w_resol, w_ibar, w_iovsig;
	       } WILSON;

 typedef struct {	/* structure used in examining a resolution shell */
		int	sh_nposs, sh_ncoll, sh_nob, sh_nrscaled, sh_nobscaled;
		int	sh_nquan[NOBMAX+2], sh_nyosig[NYSRANGE+1];
		float	sh_dmin, sh_resol, sh_ybar, sh_yovsig;
		float	sh_rnum[NSRFACS+1], sh_rden[NSRFACS+1];
	       } SHLSTAT;

 typedef struct {	/* structure used in examining a data run */
		char	r_code[8];		/* code name for run */
		short	r_id;			/* integer identifier for run */
		int	r_nexam;		/* # of observations examined */
		int	r_ninteg;		/* # of obs. integrated */
		int	r_nfound;		/* # of obs. actually found */
		int	r_nscaled;		/* # of obs actually scaled */
		int	r_nabsent;		/* ninteg-nfound */
		int	r_nflagged;		/* # of obs. flagged */
		int	r_nrefs;		/* # of unique reflections*/
		int	r_nshifts;		/* # of shifts specified */
		float	r_i000;			/* Y intercept on Wilson plot */
		float	r_b;			/* Debye-Waller factor */
		float	r_rnum[NSRFACS+1];	/* R-factor numerators */
		float	r_rden[NSRFACS+1];	/* R-factor denominators */
		WILSON	r_wil[MAXWSHELLS+1]; 	/* Wilson-shell structures */
		WILSON	r_allw;			/* overall Wilson structure */
	       } RUNSTAT;

 typedef struct { /* structure type for handling statistics calculations */
		char	sta_runindex[MAX_RUN_INDEX];
		u_char	**sta_checks, sta_hklmax[4];
		short	sta_mobmerge, sta_mobnomerge, sta_ir, sta_iw;
		short	sta_nrshells, sta_nwshells, sta_spgrp, sta_rffu;
		xglong	sta_nobill, sta_nrobd, sta_nobobd, sta_nobnohead;
		xglong	sta_nobnegi, sta_allmark, sta_nrill, sta_flags;
		xglong	sta_nroutorder, sta_nrmerge, sta_nrnomerge;
		xglong	sta_nvpars, sta_nscales, sta_nlines;
		float	sta_resstate, sta_res, sta_origlim, sta_resmax;
		float	sta_wresb[MAXWSHELLS+1], sta_rresb[MAXRSHELLS+1];
		float	sta_hklco[6], sta_ismin[NISRANGE+1];
		float	sta_ysmin[NYSRANGE+1], sta_yymin[NYRANGE+1];
		FILE	*sta_fpmi, *sta_fpmis, *sta_fpl;
		REFPARS	*sta_xp;
		CRYSTAL	*sta_xhdp;
		DATARUN	*sta_rhdp;
		SHIFT	*sta_shlp;
		VECPAR	*sta_vshp;
		RUNWORK	*sta_rub;
		SHLSTAT	sta_allshl;
		SHLSTAT	sta_shl[MAXRSHELLS+1], sta_yrange[NYRANGE+1];
		RUNSTAT	sta_allruns, *sta_runs;
		WILSON	sta_runrw[MAXRSHELLS+1];
		} STATRUN;

	/* function prototypes */
extern int abptox(ABBROB *abp);
extern int abptoy(ABBROB *abp);
extern int apply_deletions(void);
extern int autoreject(void);
extern int dihcalc(MERGEDREF *mrp);
extern int dihistmain(int argc, char *argv[]);
extern int do_cycle(SCALERUN *spi);
extern int eliminate_dups(MERGEDREF *mrp);
extern int endmu(int ina, MAKEMURUN *mmup);
extern int endstat(int irv, STATRUN *star);
extern int find_shift(SHIFT *ish, xgulong m, int nscal);
extern int get_omcen(FILE *fpm, SHIFT *ish, int nscal, int isquad);
extern int headmem(CRYSTAL **xhdp, DATARUN **rhdp, int ndruns);
extern int input_means(MERGEDREF *mrp);
extern int keep_dimen(DATARUN *drp);
extern int makemumain(int argc, char *argv[]);
extern int makemuops(MAKEMURUN *mmup);
extern int mufiles(MAKEMURUN *mmup);
extern int muinterp(char *mustr, MUREFL *mup);
extern int murcmp(MUREFL *mu0, MUREFL *mu1);
extern int out_refls(REDUWORK *rwp);
extern int outmu(FILE *fp, MUREFL *mup);
extern int print_deletions(void);
extern int red_setup(int argc, char *argv[], REDUWORK *rwp);
extern int reducemain(int argc, char *argv[]);
extern int refl_process(MERGEDREF *mrp);
extern int refvec(ABBROB *aobp, RUNWORK *rup, int vecr, int nvp, double ss[]);
extern int rejdie(int irv);
extern int rejgref(void);
extern int rejsetup(int argc, char *argv[]);
extern int rejswitch(int argc, char *argv[]);
extern int save_refl(void);
extern int scalemmain(int argc, char *argv[]);
extern int scalemmain(int argc, char *argv[]);
extern int scalemmain(int argc, char *argv[]);
extern int sccycles(SCALERUN *spi);
extern int scend(int irv, SCALERUN *spi);
extern int scmain(int argc, char *argv[]);
extern int screw_ok(short ihkl[], short ispgp);
extern int spacegmain(int argc, char *argv[]);
extern int spacegops(STATRUN *star);
extern int spacegstart(STATRUN *star);
extern int statmain(int argc, char *argv[]);
extern int statops(STATRUN *star);
extern int statstart(STATRUN *star);
extern int stlcmp(MUREFL *mu0, MUREFL *mu1);
extern int terpalt(short tom, SHIFT *shv);
extern int try_reindex(short vecre[], float matre[3][3]);
extern int wobcmp(WORKINGOB *wp0, WORKINGOB *wp1);
extern void add_shift(SHIFT *ish, int indx, int nscal);
extern void calc_sh(SHLSTAT *shp, STATRUN *star);
extern void dihfinal(void);
extern void dihinit(void);
extern void dihshow(FILE *fp, int nru);
extern void last_calcs(void);
extern void neighbors(SHIFT *ish, unsigned nsca, int isquad);
extern void pr_dr(FILE *fp, DATARUN *drp);
extern void pr_merref(FILE *fp, MERGEDREF *mrp, short spgp);
extern void pr_xt(FILE *fp, CRYSTAL *xp);
extern void redo_ob(WORKINGOB *wobp);
extern void sccalc_kill(SCALERUN *spi);
extern void scminim(int state, SCALERUN *spi);
extern void splitmr(MERGEDREF *mrpin, MERGEDREF *mrppri, MERGEDREF *mrpbij);
extern void use_deleted_sums(WORKINGOB *wobmis);
extern void workrun(DATARUN *drp, RUNWORK *rup, CRYSTAL *xhp, int shadd);
extern char *anyob_display(u_char obtype);
extern char *marker_display(u_char marker);
extern double qfromfrac(double delfrac);
extern double qfromrval(double rtarget);
extern double stlval(short shkl[], REFPARS *rp);
extern double terpscale(short tom, SHIFT *shv, double st);
extern REJECTRUN *managerjp(int opno);
extern SHIFT *get_shifts(SHIFT *ish, VECPAR *vsp, int *nship);
extern SHIFT *init_shift(int ntosave);
extern WORKINGOB *get_deleted_sums(void);

#endif /* XGPOSTINT_H */

/* end of xgpostint.h */
