/* dispparse.c -- callbacks for pdisplay associated with the character parser

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	18 Oct 2006: showuref: fix calling-sequence bug; improve error recovery
	23 Jun 2006: image memory moved from the XG_X11PROPS structure to
		the DISP_ATTRS structure so things are better localized
	21 Jan 2006: indexname args changed; di_maxw,h consistent with dispx.c;
		correct bad widno input values in showuref,showmref;
		store_image takes no arg; uarv used in dispinteg
	 7 Jan 2006: old uref, mref moved here and become show[mu]ref();
		mkuscatt moved here
	17 Jun 2005: in cmdparse(): get_colorscheme() called with just 1 arg
	10 Nov 2004: prevent crashes in cmdparse() with invalid colorscheme
	17 Oct 2004: *widq replaces wid in cmdparse()
	 3 Oct 2004: determine size of display based on call to DisplayHeight
	27 Aug 2004: routines used only here are declared as static; cosmetics
	18 Aug 2004: make *igp local; it doesn't need to be global
	15 Jun 2004: imageflag bit for gzipping is now correctly abbreviated
	25 Mar 2004: dispinteg() written: for integration within pdisplay;
	 reorganized this file; rewrote parts of cmdparse();
	 corrected bug in displays of backgrnd or CALDATA
	21 Jan 2004: license message inserted
	 9 Jan 2004: indexname() requires arg with FRAMEINDEX pointer
		in switch_images(); we must use the right MAIL structure too
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	18 Sep 2003: created from dispx.c
 Earlier changes to dispx.c relevant to this file:
	18 Sep 2003: numerous routines moved to other files
	25 Aug 2003: reduce reliance on static globals
	12 Aug 2003: add new overlays for resolution circles &
		row+column sums; change displays of predictions to
		reflections with large gamma values cover more images
	 6 Aug 2003: change Motif global variables to static globals
	25 Mar 2003: major rewrite to enable input from stdin
	 7 Dec 1993: initial X-GEN version
	15 Jun 1992: created, based on Mike Carson's display code.
	
   Purpose: provides callbacks obtainable from dispx.c
	that enable command-line or file inputs as well as button-pushes.
   Cross references: ./disp*.c, ../displays/disp*.c
*************************************************************************/
#define	NEED_ERRNO	1
#define	NEED_MOTIF	1
#include	"xgmdefs.h"
#include	"integdefs.h"

static void switch_images(XG_X11PROPS *xxg, int fno, char *newnam)
{ /*	This does a quick re-draw after changing images */
	char		*result;
	int		lfm, wido, htdo, wiup, htup, darkfla, loscreen;
	xgulong		maxwid, maxht;
	DISP_ATTRS	*lattr, *magat;
	MAIL		*lhd;

	lattr = xxg->gp_at;	magat = xxg->gp_bl;
	if (0 <= fno)
	  {
		lfm = fno;
		if ((NULL == (result = indexname(lfm))) ||
		 ('\0' == *result))
		  {
			fprintf(stderr,
			 "Can't establish image name for frame %d\n", lfm);
			return;
		  }
		lattr->di_fmno = lfm;
	  }
	else result = newnam;
	if (0 != xgaccess(result, 4, 1))
	  {
		fprintf(stderr, "Can't open image %s\n", result); return;
	  }
	if (lattr->di_disflags & DISPHKL_CONT)
		lattr->di_disflags &= ~DISPHKL_CONT;
	magat->di_imaname = lattr->di_imaname = result;
	darkfla = lattr->di_imflags & ~0xFFBF;
	loscreen = DefaultScreen(xxg->gp_di);
	maxwid = DisplayWidth (xxg->gp_di, loscreen);
	maxht  = DisplayHeight(xxg->gp_di, loscreen);
	if (maxht < maxwid - 280)
		lattr->di_maxw = lattr->di_maxh = maxht;
	else	lattr->di_maxw = lattr->di_maxh = maxwid - 280;
	if (FRAME_OK != grab_image(darkfla, lattr)) return;
	attr2locals();	(void)store_image();	filereport(xxg);
	get_colorscheme(xxg);	/* understand color scheme */
	show_iscales(xxg); /* max&multiplier scales */
	lhd = lattr->di_head + lattr->di_fmno % lattr->di_nmbox;
	wido = lattr->di_widown;	wiup = lattr->di_wiup;
	htdo = lattr->di_htdown;	htup = lattr->di_htup;
	lattr->di_disx = ((int)lhd->zmaxx) * wiup / wido;
	lattr->di_disy = ((int)lhd->zmaxy) * htup / htdo;
	redraw();	init_inten(xxg);	/* display I(zmaxx, zmaxy) */
}

void showuref(Widget widget)
{ /*	This sets up a display of the contents of a UREFLS file.
	It may be called directly from one of the pushbuttons in the UREFLS
	menu or from the "Do it!" button of the scatter-plot menu. */
	char		*result;
	int		dispac, lditype, thisdisplay, widno;
	Arg		lda[1];
	Widget		*wp;
	Boolean		setnow;
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lmain;

	if (NULL == (xgp = managex11p(3))) return;
	result = trname("UREFLS");	lmain = xgp->gp_at;
	/* we explicitly examine the UREFLS file to see if it's available
	 before we mess with any of the motif features: thus, if it's
	 missing, we don't get sidetracked. */
	if (0 != xgaccess(result, 4, 1))
	  {
		fprintf(stderr, "Error opening %s as UREFLS input\n", result);
		return;
	  }
	dispac = ((int)lmain->di_wid) * ((int)lmain->di_ht) *
		((int)(lmain->di_nbperp));
	if (widget == xgp->gp_urefcan)
	  { /* stop displaying scatterplot box */
		XtUnmanageChild(xgp->gp_urefsc);
		filereport(xgp);	redraw();	return;
	  }
	if (widget == xgp->gp_urefg) widno = 0; /* scatterplot: "Do it!"*/
	else {
		for (widno = FIRSTUREFLS; widno < FIRSTMREF; widno++)
			if (widget == xgp->gp_dpush[widno]) break;
		if (widno >= FIRSTMREF) /* bad input: do something innocuous */
		  {
			widno = FIRSTUREFLS + 1; widget = xgp->gp_dpush[widno];
		  }
	     }
	thisdisplay = widno + 1 - FIRSTUREFLS;
	 /* scatter plots are always recalculated; for other kinds of map
	  we look to see if the appropriate display is already available,
	  and if it is, we don't recalculate it. */
	lditype = lmain->di_distype;
	if ((thisdisplay == 1 - FIRSTUREFLS) || (lditype != thisdisplay))
	  {
		printf("Examining urefls file %s\n", result);
		if (NULL == lmain->di_rotm) /* make sure memory available */
		  {
			if (NULL == (xgp = managex11p(2 * dispac))) return;
			lmain->di_rotm = lmain->di_imm + dispac;
		  }
		/* scatter-plot cases: err(omega) vs. omega, etc. */
		if (thisdisplay == 1 - FIRSTUREFLS)
		  { /* figure out what the independent variable is */
			for (xgp->gp_ivn = 0, wp = &(xgp->gp_urefi[0]);
			 xgp->gp_ivn < NUIND; xgp->gp_ivn++, wp++)
			   {
				XtSetArg(lda[0], XmNset, &setnow);
				XtGetValues(*wp, lda, 1); if (setnow) break;
			   }
			if (xgp->gp_ivn == NUIND) xgp->gp_ivn = 1;
			/* figure out what the dependent variable is */
			for (xgp->gp_dvn = 0, wp = &(xgp->gp_urefd[0]);
			 xgp->gp_dvn < NUDEP; (xgp->gp_dvn)++, wp++)
			   {
				XtSetArg(lda[0], XmNset, &setnow);
				XtGetValues(*wp, lda, 1); if (setnow) break;
			   }
			if (xgp->gp_dvn == NUDEP) xgp->gp_dvn = 0;
			/* stop displaying this box */
			XtUnmanageChild(xgp->gp_urefsc);
			if (0 >= (xgp->gp_nobinu = scatt_uref(result,
			 xgp->gp_dvn, xgp->gp_ivn, xgp->gp_at)))
			  {
				fprintf(stderr, "scatt_uref failure code %d\n",
					xgp->gp_nobinu); return;
			  }
		  }
		/* plots of (a statistic) vs. (X,Y): */
		else if (0 >= (xgp->gp_nobinu = exam_urefl(result,
		 thisdisplay, xgp->gp_at))) return;
			/*convert from simple integer values to Pixel values*/
		printf("Converting urefls values to colors...");
		valtocolors(xgp);
	  }
		/* now display it */
	if (thisdisplay == 1 - FIRSTUREFLS)
		printf("Displaying scatter-plot of UREFLS stats\n");
	else	printf("Displaying UREFLS statistics vs (X,Y) map\n");
	(xgp->gp_bl)->di_distype = lmain->di_distype = thisdisplay;
	redraw();	filereport(xgp);
}

void mkuscatt(void)
{
/* This sets up a BulletinBoard containing a dialog for doing scatter plots
 of statistics in a UREFLS file.  There are two RadioBoxes--one for
 independent variables, and one for dependent variables.  There's also a "go"
 button, which does what you think it does */
	int		i, miw, ng;
	Widget		*wp;
	XmString	locstr;
	XG_X11PROPS	*xgp;
	static Arg	lda[4];
	static Widget	ureflab, urefradd, urefradi;

	if (NULL == (xgp = managex11p(3))) return;
	miw = (xgp->gp_at)->di_minw;
	if (NULL == xgp->gp_urefsc)
	  {
		ng = 0;
		XtSetArg(lda[ng], XmNx, miw+ 18);		ng++;
		XtSetArg(lda[ng], XmNy, 80);			ng++;
		XtSetArg(lda[ng], XmNwidth, 300);		ng++;
		XtSetArg(lda[ng], XmNheight, 210);		ng++;
		xgp->gp_urefsc = XmCreateBulletinBoard(xgp->gp_dimain,
			"uref_scatt", lda, ng);
	  }
	if (NULL == xgp->gp_urefsc) return;
	XtManageChild(xgp->gp_urefsc);
	if (NULL == ureflab)
	  {
		ng = 0;
		XtSetArg(lda[ng], XmNx, 10);	ng++;
		XtSetArg(lda[ng], XmNy,  0);	ng++;
		locstr = strstufstr(
 "UREFLS Scatter plot\nIndependent Variable    Dependent Variable","");
		XtSetArg(lda[ng], XmNlabelString, locstr);	ng++;
		ureflab = XmCreateLabelGadget(xgp->gp_urefsc,
			"uref_lab", lda, ng);
		XmStringFree(locstr);
	  }
	if (NULL == ureflab) return;
	XtManageChild(ureflab);
	if (NULL == xgp->gp_urefg)
	  {
		ng = 0;
		XtSetArg(lda[ng], XmNx, 110);			ng++;
		XtSetArg(lda[ng], XmNy, 182);			ng++;
		locstr = strstufset("Do it!", xgp->gp_chpol);
		XtSetArg(lda[ng], XmNlabelString, locstr);	ng++;
		xgp->gp_urefg = XmCreatePushButtonGadget(xgp->gp_urefsc,
		 "uref_go", lda, ng);
		XtAddCallback(xgp->gp_urefg, XmNactivateCallback,
			(XtCallbackProc)showuref, XpNU);
		XmStringFree(locstr);
	  }
	XtManageChild(xgp->gp_urefg);
	if (NULL == xgp->gp_urefcan)
	  {
		ng = 0;
		XtSetArg(lda[ng], XmNx,  40);			ng++;
		XtSetArg(lda[ng], XmNy, 182);			ng++;
		locstr = strstufset("Cancel", xgp->gp_chpol);
		XtSetArg(lda[ng], XmNlabelString, locstr);	ng++;
		xgp->gp_urefcan = XmCreatePushButtonGadget(xgp->gp_urefsc,
			"uref_can", lda,ng);
		XtAddCallback(xgp->gp_urefcan, XmNactivateCallback,
			(XtCallbackProc)showuref, XpNU);
		XmStringFree(locstr);
	  }
	XtManageChild(xgp->gp_urefcan);
	if (NULL == urefradi)
	  {
		ng = 0;
		XtSetArg(lda[ng], XmNx, 10);	ng++;
		XtSetArg(lda[ng], XmNy, 50);	ng++;
		XtSetArg(lda[ng], XmNradioAlwaysOne, True);	ng++;
		urefradi = XmCreateRadioBox(xgp->gp_urefsc,
			"uref_radi", lda, ng);
		XtManageChild(urefradi);
		for (i = 0, wp = &(xgp->gp_urefi[0]); i < NUIND; i++, wp++)
		  {
			ng = 0;
			XtSetArg(lda[ng], XmNset,
				(i == xgp->gp_ivn ? True : False)); ng++;
			locstr = strstufset(varlabel(i, 0), xgp->gp_chpol);
			XtSetArg(lda[ng], XmNlabelString, locstr);	ng++;
			*wp = XmCreateToggleButtonGadget(urefradi,
			 varlabel(i, 0), lda, ng);
			XmStringFree(locstr);
		  }
		XtManageChildren(xgp->gp_urefi, NUIND);
	  }
	if (NULL == urefradd)
	  {
		ng = 0;
		XtSetArg(lda[ng], XmNx, 160);	ng++;
		XtSetArg(lda[ng], XmNy, 50);	ng++;
		XtSetArg(lda[ng], XmNradioAlwaysOne, True);	ng++;
		urefradd = XmCreateRadioBox(xgp->gp_urefsc,
			"uref_radi", lda, ng);
		XtManageChild(urefradd);
		for (i = 0, wp = &(xgp->gp_urefd[0]); i < NUDEP; i++, wp++)
		  {
			ng = 0;
			XtSetArg(lda[ng], XmNset,
			 (i == xgp->gp_dvn? True:False)); ng++;
			locstr = strstufset(varlabel(i, 1), xgp->gp_chpol);
			XtSetArg(lda[ng], XmNlabelString, locstr);	ng++;
			*wp = XmCreateToggleButtonGadget(urefradd,
			 varlabel(i, 1), lda, ng);
			XmStringFree(locstr);
		  }
		XtManageChildren(xgp->gp_urefd, NUDEP);
	  }
}

void showmref(Widget widget)
{ /*	This sets up a display of the contents of a MULTIREF file */
	char		*result;
	int		dispac, lditype, thisdisplay, widno, y;
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lmain;

	if (NULL == (xgp = managex11p(3))) return;
	result = trname("MULTIREF");	lmain = xgp->gp_at;
	if (0 != xgaccess(result, 4, 1))
	  {
		fprintf(stderr, "Error opening %s as MULTIREF input\n",result);
		return;
	  }
	for (widno = FIRSTMREF; widno < FIRSTOVERL; widno++)
		 if (widget == xgp->gp_dpush[widno]) break;
	if (widno >= FIRSTOVERL) /* default (shouldn't be needed) */
	  {
		widno = FIRSTMREF + 1;	widget = xgp->gp_dpush[widno];
	  }
	if (widno == FIRSTOVERL - 2)
	  {
		if (0 != xgaccess(trname("SCALEIN"), 4, 1))
		  {
			fprintf(stderr,
			 "Can't compute <dI/sig> without scaling parameters\n");
			return;
		  }
	  }
	thisdisplay = widno - FIRSTUREFLS;
	dispac = lmain->di_wid * lmain->di_ht * lmain->di_nbperp;
	lditype = lmain->di_distype;
	if (lditype != thisdisplay)
	  {
		printf("Examining multiref file %s\n", result);
		if (NULL == lmain->di_rotm) /* make sure memory available */
		  {
			if (NULL == (xgp = managex11p(2 * dispac))) return;
			lmain->di_rotm = lmain->di_imm + dispac;
		  }
		if ((y = exam_mref(result, thisdisplay, xgp->gp_at)))
		  {
			printf("Error %d in displaying multiref\n",y);
			return; /* get the statistics */
		  }
			/*convert from simple integer values to Pixel values*/
		valtocolors(xgp);
	  }
		/* now display it */
	(xgp->gp_bl)->di_distype = lmain->di_distype = thisdisplay;
	redraw();	filereport(xgp);
}

void cmdparse(char *stri)
{ /*	This parses a command either from a file or from a pre-cooked string */
	int		nw, newframe, oldframe, indwidi, depwidi;
	int		i, newv, oldv, licn, previcn, newicn, iq, ldifl;
	xglong		oldma, lma;
	char		*cpi;
	char		verb[50], obj0[120], obj1[40], lverb[50];
	double		newmax, newmul;
	static Arg	dag[10];
	XG_X11PROPS	*xxg;
	DISP_ATTRS	*lattr, *magat;
	Widget		*widp, *widq;
	Boolean		newb;

	if (NULL == (xxg = managex11p(3))) return;
	licn = xxg->gp_icn;	lattr = xxg->gp_at;	magat = xxg->gp_bl;
	ldifl = lattr->di_disflags;	oldframe = lattr->di_fmno; 
	/* parse this minimally */
	if (0 >= (nw = sscanf(stri, "%s %s %s", verb, obj0, obj1)))
	  {
		fprintf(stderr, "No parsable information in command %s\n",
			stri);	return;
	  }
	/* copy and lowercaseify the command. We'll need the version
	 that isn't lowercaseified if we don't recognize any command here */
	strcpy(lverb, verb);	lowercaseify(lverb);
	oldma = lma = lattr->di_maxv;
	if (!strncmp(lverb, "di", 2)) /* display new image */
	  {
		cpi = obj0;
		if (('-' == *cpi) || (('0' <= *cpi) && (*cpi <= '9')))
		  {
			newframe = atoi(cpi);
			if (newframe != oldframe)
				switch_images(xxg, newframe, (char *)NULL);
			else printf("already displaying image %d\n",
				newframe);
		  }
		else if (strcmp(cpi, lattr->di_imaname))
			switch_images(xxg, -1, cpi);
		else printf("already displaying %s\n", lattr->di_imaname);
		return;
	  }
	else if (nw > 1)
	  {
		lowercaseify(obj0); /* obj0 should be lowercase now */
		if (nw > 2) lowercaseify(obj1); /* obj1 also! */
	  }
	if (!strncmp(lverb, "ex", 2)) doexit(); /* quit */
	else if (!strncmp(lverb, "qu", 2)) doexit(); /* quit */
	else if (!strncmp(lverb, "ne", 2)) /* next image */
	  {
		newframe = oldframe + 1;
		switch_images(xxg, newframe, (char *)NULL);
	  }
	else if (!strncmp(lverb, "pr", 2)) /* previous image */
	  {
		newframe = oldframe - 1;
		switch_images(xxg, newframe, (char *)NULL);
	  }
	else if (!strncmp(lverb, "up", 2)) /* display-contrast up */
	  {
		lma = (u_short)((double)lma * 1.2 + 0.49999);
		if (lma <= oldma) lma = oldma + 1;
		lattr->di_maxv = lma;
		calc_ranges(lattr); (void)store_image(); redraw();
	  }
	else if (!strncmp(lverb, "do", 2)) /* display-contrast down */
	  {
		lma = (u_short)((double)lma * 0.8 + 0.49999);
		if (lma >= oldma) lma = oldma - 1;
		lattr->di_maxv = lma;
		calc_ranges(lattr); (void)store_image(); redraw();
	 }
	else if (!strncmp(lverb, "re", 2)) redraw(); /* clean up image */
		/* display the calibration ("brass-plate") image */
	else if ((!strncmp(lverb, "br", 2)) || (!strncmp(lverb, "ca", 2)))
	{
		if ((NULL == (cpi = trname("caldata"))) || ('\0' == *cpi))
		  {
			printf("Can't determine brass-plate name\n"); return;
		  }
		else if (strcmp(cpi, lattr->di_imaname))
			switch_images(xxg, -1, cpi);
		else printf("already displaying %s\n", lattr->di_imaname);
	  }
	else if (!strncmp(lverb, "bg", 2)) /* background */
	  {
		if ((NULL == (cpi = trname("bgimage"))) || ('\0' == *cpi))
		  {
			printf("Can't determine background name\n"); return;
		  }
		else if (strcmp(cpi, lattr->di_imaname))
			switch_images(xxg, -1, cpi);
		else printf("already displaying %s\n", lattr->di_imaname);
	  }
	else if (!strncmp(lverb, "ma", 2)) /* inten for max color */
	  {
		cpi = obj0;
		if (0. >= (newmax = atof(cpi))) printf(
			"Unrecognized max-color value (%s)\n", cpi);
		else {
			i = newmax + 0.49999;
			XtSetArg(dag[0], XmNvalue, i);
			XtSetValues(xxg->gp_iscal[0], dag, 1);
			XtManageChild(xxg->gp_iscal[0]);
			printf("New max-color = %8.2f\n", newmax); chgdisp();
		     }
	  }
	else if (!strncmp(lverb, "mu", 2)) /* multiplier */
	  {
		cpi = obj0;
		if (0. >= (newmul = atof(cpi))) printf(
			"Unrecognized max-color value (%s)\n", cpi);
		else {
			i = newmul * 1000. + 0.49999;
			XtSetArg(dag[0], XmNvalue, i);
			XtSetValues(xxg->gp_iscal[1], dag, 1);
			XtManageChild(xxg->gp_iscal[1]);
			printf("New color-multiplier: %8.4f\n",newmul);
			chgdisp();
		     }
	  }
		/* color scheme for displayed images and other stuff */
	else if ((!strncmp(lverb, "cs", 2)) || (!strncmp(lverb, "co", 2)))
	  {
		cpi = obj0;	previcn = licn;	newicn = -2;
		if (!strncmp(cpi, "bw", 2)) newicn = 0;
		else if (!strncmp(cpi, "sp", 2)) newicn = 1;
		else if (!strncmp(cpi, "wb", 2)) newicn = 2;
		else if ((!strncmp(cpi, "sr", 2)) ||
			(!strncmp(cpi, "wr", 2))) newicn = 3;
		else printf("Unrecognized color scheme: %s\n", cpi);
		if (previcn == newicn) printf(
		 "Specified color scheme %s is already active\n", cpi);
		else if (-2 != newicn)
		    {
			XtSetArg(dag[0], XmNset, False);
			XtSetValues(xxg->gp_color[previcn], dag, 1);
			XtManageChild(xxg->gp_color[previcn]);
			XtSetArg(dag[0], XmNset, True);
			XtSetValues(xxg->gp_color[newicn], dag, 1);
			XtManageChild(xxg->gp_color[newicn]);
			get_colorscheme(xxg);
			xxg->gp_icn = newicn;
		     }
	  }			
	else if (!strncmp(lverb, "ro", 2)) /* rotation convention 0 - 7 */
	  {
		cpi = obj0;	i = atoi(cpi);
		if ((0 > i) || (i > 7))
		 printf("Illegal rotation convention %d\n", i);
		if (i == lattr->di_roco)
			printf("Image already in rotation %d\n", i);
		else {
			lattr->di_roco = magat->di_roco = i; chgdisp();
		     }
	  }
	else if (!strncmp(lverb, "ur", 2)) /* UREFLS displays */
	  {
		cpi = obj0;	widq = (Widget *)0;
		widp = &(xxg->gp_dpush[FIRSTUREFLS]);
		if (!strncmp(cpi, "in", 2)) widq = widp;
		else if (!strncmp(cpi, "io", 2)) widq = widp + 1;
		else if ((!strncmp(cpi, "eo", 2)) || (!strncmp(cpi, "ep", 2)))
			widq = widp + 2;
		else if (!strncmp(cpi, "ex", 2)) widq = widp + 3;
		else if (!strncmp(cpi, "ey", 2)) widq = widp + 4;
		if ((*widp < *widq) && (*widq <= *(widp + 4))) showuref(*widq);
		else printf("UREFLS choice %s unimplemented\n", cpi);
	  }
	else if (!strncmp(lverb, "mr", 2)) /* MULTIREF displays */
	  {
		cpi = obj0;	widq = (Widget *)0;
		widp = &(xxg->gp_dpush[FIRSTMREF]);
		if (!strncmp(cpi, "in", 2)) widq = widp;
		else if (!strncmp(cpi, "io", 2)) widq = widp + 1;
		else if (!strncmp(cpi, "di", 2)) widq = widp + 2;
		if ((*widp <= *widq) && (*widq <= *(widp + 2))) showmref(*widq);
		else printf("MULTIREF choice %s unimplemented\n", cpi);
	  }
	/* UREFLS Scatterplots */
	else if ((!strncmp(lverb, "us", 2)) || (!strncmp(lverb, "sc", 2)))
	  { /* determine the independent variable */
		cpi = obj0;	indwidi = depwidi = -1;
		if (*cpi == 's') indwidi = 0;
		else if ((*cpi == 'o') || (*cpi == 'p')) indwidi = 1;
		else if (*cpi == 'x') indwidi = 2;
		else if (*cpi == 'y') indwidi = 3;
		cpi = obj1;
		if ((!strncmp(cpi, "eo", 2)) || (!strncmp(cpi, "ep", 2)) ||
		 (!strncmp(cpi, "do", 2)) || (!strncmp(cpi, "dp", 2)))
			depwidi = 0;
		if ((!strncmp(cpi, "ex", 2)) ||
		 (!strncmp(cpi, "dx", 2))) depwidi = 1;
		if ((!strncmp(cpi, "ey", 2)) ||
		 (!strncmp(cpi, "dy", 2))) depwidi = 2;
		if (!strncmp(cpi, "in", 2)) depwidi = 3;
		if (!strncmp(cpi, "io", 2)) depwidi = 4;
		/* if both variables are legal, set up the display */
		if ((indwidi >= 0) && (depwidi >= 0))
		  {
			mkuscatt(); /* create widgets if needed */
			XtSetArg(dag[0], XmNset, True);
			XtSetValues(xxg->gp_urefi[indwidi], dag, 1);
			for (iq = 0; iq < 4; iq++)
			   if (iq != indwidi)
			     {
				XtSetArg(dag[0], XmNset, False);
				XtSetValues(xxg->gp_urefi[iq], dag, 1);
			     }
			XtSetArg(dag[0], XmNset, True);
			XtSetValues(xxg->gp_urefd[depwidi], dag, 1);
			for (iq = 0; iq < 5; iq++)
			   if (iq != depwidi)
			     {
				XtSetArg(dag[0], XmNset, False);
				XtSetValues(xxg->gp_urefd[iq], dag, 1);
			     }
			/* ... and do it! */
			showuref(xxg->gp_urefg);
		  }
		else printf("Unrecognized scatterplot %s\n", cpi);
	  }
	/* turning on or off the overlays */
	else if ((!strncmp(lverb, "ov", 2)) || (!strncmp(lverb, "no", 2)))
	  {
		cpi = obj0;	widq = (Widget *)0;	
		newv = (*lverb == 'n') ? 0 : 1;
		widp = &(xxg->gp_dpush[FIRSTOVERL]);
		if (!strncmp(cpi, "ca", 2))
		  {
			oldv = (ldifl & DISPCALIB_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPCALIB_NEED;
			else ldifl &= ~DISPCALIB_NEED; widq = widp;
		  }
		else if (!strncmp(cpi, "pr", 2))
		  {
			oldv = (ldifl & DISPHKL_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPHKL_NEED;
			else ldifl &= ~DISPHKL_NEED; widq = widp + 1;
		 }
		else if ((!strncmp(cpi, "bo", 2)) || (!strncmp(cpi, "am", 2)))
		  {
			oldv = (ldifl & DISPBORDER_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPBORDER_NEED;
			else ldifl &= ~DISPBORDER_NEED; widq = widp + 2;
		  }
		else if (!strncmp(cpi, "ce", 2))
		  {
			oldv = (ldifl & DISPCEN_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPCEN_NEED;
			else ldifl &= ~DISPCEN_NEED;	widq = widp + 3;
		  }
		else if (!strncmp(cpi, "cc", 2))
		  {
			oldv = (ldifl & DISPCALCEN_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPCALCEN_NEED;
			else ldifl &= ~DISPCALCEN_NEED;	widq = widp + 4;
		  }
		else if (!strncmp(cpi, "re", 2))
		  {
			oldv = (ldifl & DISPRESOL_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPRESOL_NEED;
			else ldifl &= ~DISPRESOL_NEED;	widq = widp + 5;
		  }
		else 	/* nominally default: really "rc" ... */
		  {
			oldv = (ldifl & DISPROWCOL_NEED) ? 1 : 0;
			if (newv) ldifl |= DISPROWCOL_NEED;
			else ldifl &= ~DISPROWCOL_NEED; widq = widp + 6;
		  }
		lattr->di_disflags = magat->di_disflags = ldifl;
		if ((Widget *)0 != widq)
		  {
			if (oldv != newv) /* display new version */
			  {
				newb = (newv) ? True: False;
				XtSetArg(dag[0], XmNset, newb);
				XtSetValues(*widq, dag, 1);
				XtManageChild(*widq);
				draw_it();	/* update display */
				show_region(xxg);
			  }
			else printf(" overlay for %s already %d\n", cpi, oldv);
		  }
		else printf("Unrecognized overlay\n");
	  }
	/* if the user simply specifies a name or a number,
	 we figure that means he just wants to display that name or number */
	else {
		cpi = verb;
		if (('-' == *cpi) || (('0' <= *cpi) && (*cpi <= '9')))
		  {
			newframe = atoi(cpi);
			if (newframe != oldframe)
				switch_images(xxg, newframe, (char *)NULL);
			else printf("already displaying image %d\n",
				newframe);
		  }
		else if (strcmp(cpi, lattr->di_imaname))
			switch_images(xxg, -1, cpi);
		else printf("already displaying %s\n", lattr->di_imaname);
	      }

}

int dispinteg(char *callstr)
{ /*	A version of the `integmain' mainline; it also sets up display calls */
	int		arc, fmno, delfm, bgalready, origflags;
	char		**arv, **uarv, dispstr[40];
	INTEGRUN	*igp;

	if (NULL == (arv = str2acav(callstr, &arc))) return EALLOC;
	igp = manageigp(7, 0); /* to remember the proper INTEGRUN structure */
	igp->igr_flags &= ~(IR_LONGJUMP); /* clear long-jump flag */
	/* process args and set up program run */
	arc--;	uarv = arv + 1;	/* to correct a quirk in str2acav */
	cmdparse("no pr");
	if (EOK != integ_setup(arc, uarv, igp))
		return orderly_exit("integration setup failed");
	/* do a preliminary integration unless forbidden */
	/* don't delete frames on this first pass */
	origflags = igp->igr_imfl;	igp->igr_imfl &= ~GZIP_AFTER_CLOSE;
	cmdparse("ov pr");
	for (fmno = igp->igr_zmi, delfm = 0; fmno <= igp->igr_initend;
	 fmno++, delfm++)
	   { /* once per "igp->igr_bunch" frames, re-refine & re-predict */
		if (EOK != pre_image(fmno, 0, igp))
		  return orderly_exit("Problem in setting up image");
		if (!delfm) cmdparse("bg");
		if (process_image(fmno, igp))
		  return orderly_exit("Problem in processing image");
		dispstr[0] = '\0';	sprintf(dispstr, "di %d", fmno);
		cmdparse(dispstr);
	   }
	if (igp->igr_initend >= igp->igr_zmi)
	  {
		if (NULL != igp->igr_fpo)
			(void)fseek(igp->igr_fpo, 0L, 0); /* to restart*/
		bgalready = 1;
	  }
	else bgalready = 0;
	igp->igr_imfl = origflags;
	for (fmno = igp->igr_zmi, delfm = 0; fmno <= igp->igr_zma;
	 fmno++, delfm++)
	   { /* once per "igp->igr_bunch" frames, re-refine & re-predict */
		if (EOK != pre_image(fmno, bgalready, igp))
		  return orderly_exit("Problem in setting up image");
		if (process_image(fmno, igp))
		  return orderly_exit("Problem in processing image");
		dispstr[0] = '\0';	sprintf(dispstr, "di %d", fmno);
		cmdparse(dispstr);
	   }
	/* print stats and timings for later frames; final output; shut down */
	fmno = after_images(igp);
	cmdparse("ur io");	return fmno;
}

void fdproc(char *strbuf, int *sourp, pid_t *inpid)
{ /* This acts as a parser for data supplied from file descriptor *sourp. */
	int		nr;

	zeromem((void *)strbuf, 480);	errno = 0;
	if (1 < (nr = read(*sourp, strbuf, 500))) cmdparse(strbuf);
	else if (errno) printf(
	 "No data available from file descriptor %d: errno = %d\n",
	 *sourp, errno);
}

/* end of dispparse.c */
