/* dispx.c -- X-windows displays of XGEN results

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	18 Oct 2006: bugfix mkuhkl: lda array enlarged to 16.
	 8 Sep 2006: SGI-specific stuff in graphics() now depends on #ifdef sgi
	23 Jun 2006: image memory moved from the XG_X11PROPS structure to
		the DISP_ATTRS structure so things are better localized
	12 Jun 2006: gdisnam,varlabel moved to dispxutil
	21 Jan 2006: managex11p, title_mainwin, xgprocess  cleaned up;
	 store_image takes no args; separate gp_dpush from gp_spush;
	 create gp_poldmenu only if needed
	 4 Jan 2006: move Motif function prototypes to xgmdefs.h;
		move mkuscatt, [mu]mref to xgenparse.c;
		managex11p implemented; longjmp handled better
	10 Sep 2005: force histogram-calc when di_valmul changes to < 1
	22 Jun 2005: title_mainwin() and dispscale() now static
	17 Jun 2005: xgprocess(): getproject() takes fromcur arg.
	 9 Jan 2005: eliminate simulate_click() since it doesn't really work
	19 Nov 2004: make simulate_click() go to the right window;
		call it only on the outset, not later as well.
	22 Oct 2004: attempt to initialize display-window slidebars
	20 Oct 2004: environment variable COLORSCHEME used for display
	 9 Oct 2004: major rewrite for X-GEN 5.4.1: bugfixes,
		sliders on drawing area, flexible image window size,
		image-specific label on the outermost frame
	 3 Oct 2004: color specifications based on NRANGES==32 and
		NEXTRAS==6 rather than hardwired; update_uh() written
		and called where appropriate so scatterplot labels work;
		display size based on call to DisplayHeight()
	27 Aug 2004: routines used only here are declared as static;
	 cosmetics; remove call to zapscales --handled elsewhere
	28 Mar 2004: start di_disflags as DISPVERBOSE1;
	 set up fdbuf memory before initializing the toolkit
	 5 Mar 2004: move Process button from menubar to the side;
		reorganize XmStrings used for urefls buttons;
		be careful about freeing XmString memory
	 1 Mar 2004: zero out the pmotif push-button widgets right
		after we start; otherwise the pmotif top menu fails
	29 Feb 2004: substantial simplification of xgprocess(),
		since the calls to getproject() and pmpopulate()
		don't really require any protection from memory messups
	26 Feb 2004: eliminate specific calls to programs in xgprocess();
		everything is subsumed in the pmotif-like menu
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	31 Jan 2004: invoke pmotif rather than xgnow with the topmost
		processing button
	24 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted
	 6 Jan 2004: globals replaced with elements of DISP_ATTRS structs;
		calls to indexname() require FRAMEINDEX arg;
		explicit management of progname replaced with calls
		to routines hide_oldpname, restore_oldpname, etc.;
		graphics() has call simplified through DISP_ATTRS arg
	13 Dec 2003: eliminate unused variable `lditype' from chgframe()
		and static global `inpid'
	21 Sep 2003: rename getattrs to getxgattrs--
		prevents collision with curses routine
	19 Sep 2003: default value of "result" in chgfrm() is defined;
		fudge in call of XtOpenDisplay to get Irix to shut up
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: eliminate a few xgaccess() calls because they
		are latent in the xgopen() calls anyway
	15 Sep 2003: numerous routines moved to other files
	 4 Sep 2003: enable uncompression of compressed X-GEN files
	25 Aug 2003: move numerous routines to ../displays/disp*.c
		and to ./dispxutil.c; eliminate globals from pdispmain.c;
		reduce reliance on static globals
	16 Aug 2003: major workover to reduce reliance on global variables,
		at the cost of some extra elements in the limits[] array
		and a few extra arguments to some functions
	12 Aug 2003: add new overlays for resolution circles &
		row+column sums; change displays of predictions to
		reflections with large gamma values cover more images
	11 Aug 2003: subtract local sliding-averaged background from
		displayed intensity values
	 6 Aug 2003: change Motif global variables to static globals
	 1 May 2003: add additional data-processing functions;
		improve handling of data in blowups
	25 Mar 2003: major rewrite to enable input from stdin
		and to allow eight different image orientations
	14 Mar 2003: corrected bug in one call to strstufstr()
	 5 Jan 2003: reorganized source-file header
	25 Aug 2002: (?) redo color schemes: add black-and-white, inverses.
		add routine get_colorscheme() to accomplish this
	20 Jul 2002: chgdisp() decleared as void, not static void (big deal)
	15 Jul 2002: in show_predict, make main-window boxsize
		sensitive to the density of the prediction-box pattern
	25 Mar 2002: removed MSI attachments
	15 Oct 1998: in calc_ranges, make ranges xglongs rather than u_shorts.
		Also: scale depicted intensities by the merged values
		rather than the individual-pixel values.
	27 Oct 1994: renamed global variable "display"
		to static global "mydispla"
	 5 May 1994: source-file header, POLYDET->XGEN,
		"xgulong" construction used
	16 Mar 1994: minor bugfixes
	14 Jan 1994: new memory management and user interface tools
	27 Dec 1993: in store_image(), rescale tval by the area
		(in detector pixels squared0 of a display pixel
	 7 Dec 1993: initial X-GEN version
	15 Jun 1992: created, based on Mike Carson's display code.
	
   Purpose: displays detector images and other objects
	directly on an X window display governed under Motif.
   Cross references: ./disp*.c, ../displays/disp*.c

*************************************************************************/

#define	NEED_ERRNO	1
#include	"xgmdefs.h"
#include	<setjmp.h>

#define		I_VS_HKL	(NUREFLBS-2)	/* I vs hkl code */
#define		IOS_VS_HKL	(NUREFLBS-1)	/* I/sigma vs hkl code */
#define		MAGWIDTH	168	/* # pixels across blowup window */
#define		MAGHEIGHT	168	/* # pixels down   blowup window */
#define		MAXFLAGVAL	077

XG_X11PROPS *managex11p(opno)
{ /* manage a static XG_X11PROPS structure */
	DISP_ATTRS		*ldi;
	static XG_X11PROPS	*xgp;
	static char		imdmsg[] = "ldi->di_imm in managex11p";
	static char		magdmsg[] = "ldi->di_magm in managex11p";
	static char		limimsg[] = "ldi->di_limi in managex11p";
	static char		fdbmsg[] = "xgp->gp_fdbuf in managex11p";
	static char		xgpmsg[] = "xgp in managex11p";

	if (!opno) /* delete memory for the whole thing */
	  {
		if (NULL != xgp) (void)managex11p(0170);
		memzap((AVOID **)&xgp, xgpmsg);	return NULL;
	  }
	if (opno > MAXFLAGVAL) /* big values are memory measurements for imda*/
	  {
		xgp = managex11p(010);	/* delete old di_imm values */	
		if (NULL == (ldi = xgp->gp_at)) return managex11p(0);
		if (NULL == (ldi->di_imm = mallop((size_t)opno, imdmsg)))
			return managex11p(0);
		xgp->gp_imdasize = opno;
		zeromem((AVOID *)(ldi->di_imm), (size_t)opno);
		if ((NULL == ldi->di_limi) && (NULL == (ldi->di_limi =
		 (float *)callop(NALLOCCOL, sizeof (float), limimsg))))
			return managex11p(0);
		return xgp;
	  }
		/* small even numbers: selective memory kills */
	if (!(opno & 01))
	  { 
	    if ((NULL != xgp) && (NULL != (ldi = xgp->gp_at)))
	      {
		if (opno & 010) memzap((AVOID **)&(ldi->di_imm),  imdmsg);
		if (opno & 020) memzap((AVOID **)&(ldi->di_magm), magdmsg);
		if (opno & 040) memzap((AVOID **)&(xgp->gp_fdbuf), fdbmsg);
		if (opno & 0100) memzap((AVOID **)&(ldi->di_limi), limimsg);
	      }
	    return xgp;
	  }
	if (opno == 1) xgp = managex11p(0); /* force kill of old stuff */
	if (opno == 3) return xgp;	/* unconditional return */
	if (NULL == xgp)
	  {
		if (NULL == (xgp = (XG_X11PROPS *)callop(1,
		 sizeof (XG_X11PROPS), xgpmsg))) return NULL;
	  }
	if ((opno & 040) && (NULL == xgp->gp_fdbuf))
	  {
		if (NULL == (xgp->gp_fdbuf = mallop(500, fdbmsg)))
			return managex11p(0);
		zeromem((AVOID *)(xgp->gp_fdbuf), 500);
	  }
		/* we'll get more specific later... */
	return xgp;
}

void doexit(void)
{ /*	This is a callback to tell the program to quit */
	int		i;
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*ldi;
	jmp_buf		lojbu;

	if (NULL != (xgp = managex11p(3)))
	  {
		if (NULL != (ldi = xgp->gp_at)) (void)managex11p(0170);
		for (i = 0; i < 9; i++) if (NULL != xgp->gp_uhstr[i])
			XmStringFree(xgp->gp_uhstr[i]);
		XCloseDisplay(xgp->gp_di);
		(void)manageuwp(0, 0);	/* eliminate UREFLS memory */
		(void)managemrv(0);	/* eliminate MULTIREF memory */
		cpymem((AVOID *)lojbu, xgp->gp_jbuf, sizeof (jmp_buf));
		xgp = managex11p(0);	longjmp(lojbu, 1);
	  }
	else exit(EOK);
}

void draw_it(void)
{ /*	This actually outputs the current detector image to the window.
	In the X implementation the image is in the XImage format. */
	int		ldih, lditype, ldiw, result;
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lattr;
	static XImage	*polimage;	/* primary display image */

	if (NULL == (xgp = managex11p(3))) return;
	lattr = xgp->gp_at;	lditype = lattr->di_distype;
	ldih = lattr->di_ht;	ldiw = lattr->di_wid;
	get_colorscheme(xgp); /* understand colorscheme */
	if (NULL != xgp->gp_filesel) XtUnmanageChild(xgp->gp_filesel);
	if (NULL != xgp->gp_urefsc) XtUnmanageChild(xgp->gp_urefsc);
	XFlush(xgp->gp_di);	/* to catch up on things */
	if (lditype) printf("Displaying %s %s versus %s...",
	 lditype >= NUREFLBS ?  " multiref" : (lditype ? " urefls" : ""),
	 lditype < 0 ? varlabel(xgp->gp_dvn, 1) : gdisnam(lditype),
	 lditype < 0 ? varlabel(xgp->gp_ivn, 0) : (lditype >= I_VS_HKL) ?
		"hkl" : "(X,Y)");
	if (0 >= lattr->di_nbperp) getbitmapv(xgp);
	if (NULL == polimage)
	  {
		if (NULL == (polimage = XCreateImage(xgp->gp_di, xgp->gp_disv,
		 xgp->gp_dep, xgp->gp_imf, xgp->gp_imo, lattr->di_rotm,
		 ldiw, ldih, xgp->gp_bmp, xgp->gp_bpl)))
		  {
			fprintf(stderr, "Unable to create image!\n"); return;
		  }
	  }
#ifdef linux
	polimage->byte_order = ImageByteOrder(xgp->gp_di);
#else
#ifdef LITTLEND
	polimage->byte_order = LSBFirst;
#else
	polimage->byte_order = MSBFirst;
#endif
#endif
	if ((Window)0 == xgp->gp_wi)
	  {
		fprintf(stderr,
		 "Image window does not exist yet: can't call XPutImage\n");
		return;
	  }
	result = XPutImage(xgp->gp_di, xgp->gp_wi, xgp->gp_igc, polimage,
		0, 0, 0, 0, polimage->width, polimage->height);
	if (result == BadDrawable) fprintf(stderr, "Bad drawable\n");
	else if (result == BadGC) fprintf(stderr, "Bad GC\n");
	else if (result == BadMatch) fprintf(stderr, "Bad Match\n");
	else if (result == BadValue) fprintf(stderr, "Bad Value\n");
	/* superimpose the appropriate overlays on the (X,Y) displays */
	if ((lditype >= 0) && (lditype != I_VS_HKL) &&
		(lditype != IOS_VS_HKL)) do_superpositions(xgp, 0);
	if (lditype < 0) draw_axes(xgp, 0);
	else if ((lditype == I_VS_HKL) || (lditype == IOS_VS_HKL))
		hkl_axes(xgp, 0);
	XMapWindow(xgp->gp_di, xgp->gp_wi);	/* let it all hang out */
	XFlush(xgp->gp_di);	/* to catch up on things */
	if (lditype) printf("Done!\n");
}

static void title_mainwin(XG_X11PROPS *xxp)
{ /*	This puts an informative title on the display window */
	int			nametype, nar;
	char			*lfname;
	DISP_ATTRS		*lattr;
	Arg			lda[1];
	static size_t		mystrlen, oldstrlen;
	static char		*overstr;
	static XmString		locstr;
	static char		overmsg[] = "overstr in title_mainwin";

	if (NULL == xxp) /* special case to kill memory for overstr */
	  {
		memzap((AVOID **)&overstr, overmsg);	return;
	  }
	lattr = xxp->gp_at;	nametype = lattr->di_distype;
	if (!nametype) lfname = lattr->di_imaname;
	else if (nametype >= NUREFLBS) lfname = trname("multiref");
	else lfname = trname("urefls");
	mystrlen = 32 + strlen(lfname) + strlen(XGEN_VERSION);
	if (mystrlen < 68) mystrlen = 68;
	if ((NULL == overstr) || (mystrlen > oldstrlen))
	  {
		memzap((AVOID **)&overstr, overmsg);
		if (NULL == (overstr = (char *)mallop(mystrlen, overmsg)))
			return;
		oldstrlen = mystrlen;
	  }
	zeromem((AVOID *)overstr, mystrlen);
	sprintf(overstr, "X-GEN V.%s Display of %s", XGEN_VERSION, lfname);
	locstr = strstufstr(overstr, "");
	nar = 0;	XtSetArg(lda[nar], XmNdialogTitle, locstr);	nar++;
	XtSetValues(xxp->gp_dimain, lda, nar);	XtManageChild(xxp->gp_dimain);
	XmStringFree(locstr);
}

void redraw(void)
{	/* This redraws various parts of the image when they're overlapped */
	XG_X11PROPS	*xgp;

	draw_it();
	if (NULL != (xgp = managex11p(3)))
	  {
		show_region(xgp);	show_table(xgp); title_mainwin(xgp);
	  }
}

void chgdisp(void)
{	/* This re-does the values of the max and the multiplier for display*/
	int		lmax, lmul;
	double		newvm, oldvm;
	Arg		lda[1];
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lmain, *locd;

	if (NULL == (xgp = managex11p(3))) return;
	XtSetArg(lda[0], XmNvalue, &lmax);
	XtGetValues(xgp->gp_iscal[0], lda, 1);
	XtSetArg(lda[0], XmNvalue, &lmul);
	XtGetValues(xgp->gp_iscal[1], lda, 1);
	lmain = xgp->gp_at;
	oldvm = lmain->di_valmul;	locd = xgp->gp_bl;
	newvm = locd->di_valmul = lmain->di_valmul =
		((double)lmul) / 1000.;
		/* if the multiplier has changed, force a new histogram */
	if ((newvm < 1.) && (fabs(oldvm - newvm) > DFUZZ))
		(void)imhisto(xgp->gp_at, 1);
	locd->di_maxv = lmain->di_maxv = lmax * locd->di_valmul;
	locd->di_distype = lmain->di_distype;	locd->di_xp = lmain->di_xp;
	calc_ranges(lmain);
	if (!(lmain->di_distype)) (void)store_image();
	redraw();
}

void attr2locals(void)
{ /* This takes some information out of `xgp->gp_at' and stores it elsewhere */
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lmain;

	if (NULL == (xgp = managex11p(3))) return;
	lmain = xgp->gp_at;
	xgp->gp_ndcolors = lmain->di_nranges + NEXTRAS;
	lmain->di_magw = MAGWIDTH;
	lmain->di_magh = MAGHEIGHT;
	cpymem((AVOID *)xgp->gp_bl, (AVOID *)xgp->gp_at, sizeof (DISP_ATTRS));
	lmain->di_wintype = 0;	(xgp->gp_bl)->di_wintype = 1;
}

static void dispscale(XG_X11PROPS *xxp)
{ /* This gets the scale of the image right */
	DISP_ATTRS	*lattr;
	int		loscreen;
	static xgulong	maxht, maxwid, oldmaxh, oldmaxw;

	lattr = xxp->gp_at;
	oldmaxw = lattr->di_maxw;	oldmaxh = lattr->di_maxh;
	loscreen = DefaultScreen(xxp->gp_di);
	maxwid = DisplayWidth (xxp->gp_di, loscreen);
	maxht  = DisplayHeight(xxp->gp_di, loscreen);
	/* convention: we allow the di_maxw and di_maxh values to be
	 slightly larger than the display limits, because we figure
	 we can adjust that with the scrollbars, i.e. the data in
	 the top and bottom of the image is usually uninteresting anyway */
	if (maxht < maxwid - 280)
		lattr->di_maxw = lattr->di_maxh = maxht;
	else	lattr->di_maxw = lattr->di_maxh = maxwid - 280;
	if ((maxwid != oldmaxw) || (maxht != oldmaxh))
	  {
		(void)grab_image(lattr->di_darkfla, lattr);
		printf("Displaying %4d*%4d images on a %4d*%4d display\n",
			lattr->di_imwid, lattr->di_imht,
			(int)maxwid, (int)maxht);
		if ((lattr->di_htdown > 1) || (lattr->di_widown > 1))
		 printf(
		  "Each display pixel will correspond to a %dx%d image box\n",
		  	lattr->di_widown, lattr->di_htdown);
		else if ((lattr->di_htup > 1) || (lattr->di_wiup > 1))
		 printf(
		  "Each image pixel will be duplicated into a %dx%d box\n",
			lattr->di_wiup, lattr->di_htup);
		else printf(
		 "Each display pixel corresponds 1-to-1 with an image pixel\n");
	  }
}

/*ARGSUSED*/
void chgframe(Widget widget, char *tag, XmAnyCallbackStruct *callback_data)
{ /*	This tells the program to alter its display of a detector image,
	by changing the frame being displayed or the color table values.
	thereby repainting the image in a different color range. */
	int		htdo, htup, lfm, n, newnum, oldfrm, widno, wido, wiup;
	char		*result;
	xglong		lma, oldma;
	MAIL		*lhd;
	FRAMEINDEX	*lfmi;
	DISP_ATTRS	*lmain;
	XG_X11PROPS	*xgp;
	Arg		lda[1];
	static Widget	frmform;
	static XmString	picked;

	/* this gets called from standard buttons 1-8, from the file
	 selection box "OK" button, or from the frame scale buttons.
	 Figure out which one we're on.  If we're on
	 none of the standard buttons and not on the frame scale buttons,
	 we're on a file selection box. */
	if (NULL == (xgp = managex11p(3))) return;
	lmain = xgp->gp_at;
	for (widno = 1; widno < FIRSTUREFLS - 1; widno++)
		if (widget == xgp->gp_dpush[widno]) break;
	if (widget == xgp->gp_frsca[1]) widno = FIRSTUREFLS;
	if (widget == xgp->gp_frsca[2]) widno = FIRSTUREFLS + 1;
	if ((xgp->gp_lastfwid == xgp->gp_dpush[FIRSTUREFLS-1]) ||
	 (widno != FIRSTUREFLS-1))
		(xgp->gp_bl)->di_distype = lmain->di_distype = 0;
	oldma = lma = lmain->di_maxv;	/* save the old maximum value */
	oldfrm = lmain->di_fmno;	/* save the old frame number */
	get_colorscheme(xgp);	/* understand color scheme */
	if (NULL == lmain->di_fmind) lmain->di_fmind = ftables(7);
	lfmi = lmain->di_fmind;
	result = lmain->di_imaname;	/* default is image name */
	switch (widno)
	      {
		case	1:	case	2: /* increase or decrease max value */
			lma = (widno == 1 ?
			 (u_short)((double)lma * 1.2 + 0.49999) :
			 (u_short)((double)lma * 0.8 + 0.49999));
			if ((widno == 1) && (lma <= oldma)) lma = oldma + 1;
			else if ((widno== 2) && (lma>= oldma)) lma = oldma - 1;
			lmain->di_maxv = lma;	calc_ranges(lmain);
			(void)store_image();
			/* this case drops through now to the "redraw" case*/
		case	3:	/* redraw screen */
			redraw();	xgp->gp_lastfwid = widget;	return;
		case	4:	case	5: /* next or previous frame */
			if (lmain->di_fmno > -1) /*is frmindex defined?*/
			  {
				if (widno == 4) (lmain->di_fmno)++;
				else (lmain->di_fmno)--;
				if (NULL != lfmi)
				  {
					if (lmain->di_fmno < lfmi->fi_firstfm)
					 lmain->di_fmno = lfmi->fi_firstfm;
					if (lmain->di_fmno > lfmi->fi_lastfm)
					 lmain->di_fmno = lfmi->fi_lastfm;
				  }
			  }
			lfm = lmain->di_fmno;
			if ((-1 < lfm) && (lfm != oldfrm))
			  {
				result = indexname(lfm);	break;
			  }
			else if (lmain->di_fmno <= -1)
				fprintf(stderr, "Index not in use\n");
			else if ((lmain->di_fmno != oldfrm) &&
			 (lmain->di_fmno == lfmi->fi_firstfm))
				fprintf(stderr, "at start of frame list!\n");
			else if ((lmain->di_fmno != oldfrm) &&
			 (lmain->di_fmno == lfmi->fi_lastfm))
				fprintf(stderr, "at end of frame list!\n");
				/* no change in display */
			xgp->gp_lastfwid = widget; return;
		case	6:	/* use CALDATA */
			if ((NULL == (result = getenv("CALDATA"))) ||
			 ('\0' == *result)) result = "CALDATA";	break;
		case	7:	/* use BGIMAGE */
			if ((NULL == (result = getenv("BGIMAGE"))) ||
			 ('\0' == *result)) result = "BGIMAGE";	break;
		case	8:	/* pick frame from scale-bar */
			if (lmain->di_fmno > -1) /*is frame index defined?*/
			  {
				if (NULL == frmform)
					frmform = setupfrmscale(xgp);
				else {
					XtSetArg(lda[0], XmNvalue,
						(int)(lmain->di_fmno));
					n = 1;
					XtSetValues(xgp->gp_frsca[0], lda, n);
					XtManageChildren(xgp->gp_frsca, 3);
					XtManageChild(frmform);
				     }
			  }
			else printf("No frame index defined\n");
			/* no immediate change */
			xgp->gp_lastfwid = widget; return;
		case	9:	/* choose file from selection box */
			XtUnmanageChild(xgp->gp_filesel); /* stop displaying */
			if (callback_data->reason == XmCR_OK)
			  {
				XtSetArg(lda[0],XmNdirSpec,(XtArgVal)(&picked));
				XtGetValues(widget, lda, 1);
				result = dextract_first(picked);	break;
			  }
			else {
				draw_it();	/* restore previous drawing */
				xgp->gp_lastfwid = widget;
				return;	/* don't re-define the frame */
			     }
		case	10:	/* from frame scale "OK" button */
			lfm = lmain->di_fmno;
			XtSetArg(lda[0], XmNvalue, &lfm);
			XtGetValues(xgp->gp_frsca[0], lda, 1);
			lmain->di_fmno = (xgp->gp_bl)->di_fmno = lfm;
			/* the rest of this case falls through to the next */
		case	11:	/* from frame scale "Cancel" button */
			XtUnmanageChild(frmform);
			lfm = lmain->di_fmno;
			if ((-1 < lfm) && (lfm != oldfrm))
			  {
				result = indexname(lfm);	break;
			  }
			draw_it();	/* redraw the image where the box was*/
			xgp->gp_lastfwid = widget;
			return;	/* no change in frame # => no new display */
	     }
	/* if indeed we called this routine from the file selection box,
	 we need to figure out what kind of selection box we had previously.
	 If it was the UREFLS menu, all we're doing is setting the environment
	 variable `UREFLS' equal to the chosen filename */
	if (xgp->gp_lastfwid == xgp->gp_dpush[FIRSTMREF - 1])
	  {
		if (1 == (newnum = set_logical("UREFLS", result))) printf(
			"Setting UREFLS to %s\n", result);
		else if (2 == newnum)
			printf("UREFLS already is %s; unchanged\n", result);
		else fprintf(stderr, "Unable to re-define UREFLS\n");	return;
	  }
	/* similarly if we called this from the MULTIREF menu ... */
	else if (xgp->gp_lastfwid == xgp->gp_dpush[FIRSTOVERL-1])
	  {
		if (1 == (newnum = set_logical("MULTIREF", result))) printf(
			"Setting MULTIREF to %s\n", result);
		else if (2 == newnum)
			printf("MULTIREF already is %s; unchanged\n", result);
		else fprintf(stderr, "Unable to re-define MULTIREF\n"); return;
	  }
	 /* if we get to here, we have a new data image to consider */
	if (0 != xgaccess(result, 4, 1))
	  {
		fprintf(stderr, "Can't open %s\n", result); redraw();	return;
	  }
	if (lmain->di_disflags & DISPHKL_CONT)
		lmain->di_disflags &= ~DISPHKL_CONT;
	if (oldfrm == lmain->di_fmno) /* search for framenum if needed */
	  {
		if (-1 < (newnum = match_inode(&(lmain->di_fmind),
		 lmain->di_imaname)))
		  {
			printf("Frame number now set to %d\n",newnum);
			lmain->di_fmno = newnum;
		  }
	  }
	lmain->di_imaname = result;
	lmain->di_minw = 0;	/* flag to make sure dispscale functions */
	dispscale(xgp);
	attr2locals();	(void)store_image();	filereport(xgp);
	show_iscales(xgp); /* max&multiplier scales */
	lhd = lmain->di_head + lmain->di_fmno % lmain->di_nmbox;
	wiup = lmain->di_wiup;	wido = lmain->di_widown;
	htup = lmain->di_htup;	htdo = lmain->di_htdown;
	lmain->di_disx = ((int)lhd->zmaxx) * wiup / wido;
	lmain->di_disy = ((int)lhd->zmaxy) * htup / htdo;
	redraw();	init_inten(xgp);	/* display I(zmaxx, zmaxy) */
}

static void update_uh(void)
{ /* This updates the strings that fit into the UREFLS hkl buttons */
	int		j;
	XG_X11PROPS	*xgp;
	static char	*uhchars[] = {
	 "Constant h", "Constant k", "Constant l",
	 "Intensity", "Inten/sigma", "Profile-fit", "Summation",
	 "X, Y, Omega", "Local coords" } ;

	if (NULL == (xgp = managex11p(3))) return;
	for (j = 0; j < 9; j++)
	   {
		if (NULL != xgp->gp_uhstr[j]) XmStringFree(xgp->gp_uhstr[j]);
		xgp->gp_uhstr[j] = strstufstr(uhchars[j], "");
	   }
}

static void newfile(Widget widget)
{ /*	This callback results in creation (or re-management) of a file
	selection box widget to let the user choose a new file to display */
	int		ng, nranges;
	char		*primarydir;
	XG_X11PROPS	*xgp;
	static Arg	lda[8];
	static XmString	locstr, locstr2;

		/* invoked from detector-image category */
	if (NULL == (xgp = managex11p(3))) return;
	if (widget == xgp->gp_dpush[FIRSTUREFLS - 1])
	  {
		if (NULL == (primarydir = getenv("DATA"))) primarydir = ".";
	  }
	else 		/* invoked from UREFLS or MULTIREF */
	  {
		if (NULL == (primarydir = getenv("WORK"))) primarydir = ".";
	  }
	locstr = strstufstr(primarydir, "");
	nranges = (xgp->gp_at)->di_nranges;
	ng = 0;	XtSetArg(lda[ng], XmNdirectory, locstr);	ng++;
	if (NULL == xgp->gp_filesel)
	  {
		XtSetArg(lda[ng], XmNbackground, xgp->gp_dbcol[0]);	ng++;
		XtSetArg(lda[ng], XmNforeground, xgp->gp_dbcol[nranges]); ng++;
		locstr2 = strstufstr("Use it", "");
		XtSetArg(lda[ng], XmNokLabelString, locstr2);	ng++;
		XtSetArg(lda[ng], XmNx, 10);			ng++;
		XtSetArg(lda[ng], XmNy, 10);			ng++;
		XtSetArg(lda[ng], XmNwidth, 512);		ng++;
		XtSetArg(lda[ng], XmNheight, 512);		ng++;
		xgp->gp_filesel = XmCreateFileSelectionBox(xgp->gp_dimain,
			"fselect", lda, ng);
		XtAddCallback(xgp->gp_filesel, XmNokCallback,
			(XtCallbackProc)chgframe, XpNU);
		XtAddCallback(xgp->gp_filesel, XmNcancelCallback,
			(XtCallbackProc)chgframe, XpNU);
		XmStringFree(locstr2);
	  }
	else	XtSetValues(xgp->gp_filesel, lda, ng);
	XtManageChild(xgp->gp_filesel);
		/* remember what kind of file selection this is */
	xgp->gp_lastfwid = widget;
	XmStringFree(locstr);
}

static void xgprocess(Widget widget)
{ /*	data processing capabilities. Note that we've collapsed this down
	to an invocation of the guts of the `pmotif' functionality.
	In doing so we've greatly reduced the risk of screwing up the internals
	of the display program while running the processing efforts. */
	int		retv;
	XG_X11PROPS	*xgp;
	XGNAME		*locna;

	(void)xgprogname("pdisplay", 4);	(void)xgprogname(NULL, 2);
	(void)managekill(1);	/* protect some globals from deletion */
	/* create & populate XG_11PROPS structure to be used in processing */
	if (NULL == (xgp = managex11p(3)))  return;
	/* establish the logical names we need if necessary */
	if (NULL == (locna = getproject(NULL, 0)))
	  {
		printf("Unable to define the project run ID\n"); return;
	  }
	if (EOK != (retv = pmpopulate(xgp->gp_dtoplev)))
	  {
		printf("Failure %d in call to pmpopulate\n", retv); return;
	  }
	(void)xeflush(stdout);	(void)xgprogname(NULL, 3);
	(void)managekill(0);
}

static void overlay(Widget widget)
{ /*	This toggles on or off the various overlays;
	then it displays the result. */
	int		dioverlay, overno;
	XG_X11PROPS	*xgp;
	Widget		*widp;

	if (NULL == (xgp = managex11p(3))) return;
	overno = FIRSTOVERL;	widp = &(xgp->gp_dpush[overno]);
	dioverlay = (xgp->gp_at)->di_disflags;
	if (widget == *widp++) itoggle(&dioverlay, DISPCALIB_NEED);
	else if (widget == *widp++) itoggle(&dioverlay, DISPHKL_NEED);
	else if (widget == *widp++) itoggle(&dioverlay, DISPBORDER_NEED);
	else if (widget == *widp++) itoggle(&dioverlay, DISPCEN_NEED);
	else if (widget == *widp++) itoggle(&dioverlay, DISPCALCEN_NEED);
	else if (widget == *widp++) itoggle(&dioverlay, DISPRESOL_NEED);
	else if (widget == *widp++) itoggle(&dioverlay, DISPROWCOL_NEED);
	(xgp->gp_at)->di_disflags = (xgp->gp_bl)->di_disflags = dioverlay;
		/* add in whatever overlays the user has asked for */
	draw_it();	/* update the display accordingly */
	show_region(xgp);	/* update overlays in magnified view */
}

static void uhdo(Widget widget, char *tag)
{ /*	This is called whenever any of the buttons on the urefls hkl
	menu is pushed*/
	int		j, lditype, uhv[4];
	Arg		lda[1];
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lmain;
	static sichar	prev_display;	/* for cancellations */

	if (NULL == (xgp = managex11p(3))) return;
	lmain = xgp->gp_at;
	for (j = 1; j < NURHBUTTS; j++) if (widget == xgp->gp_urhbu[j]) break;
	lditype = lmain->di_distype;	update_uh();
	switch (j)
	      {
		case	1:	/* toggle integration type (fit vs. summed) */
			if (xgp->gp_uhty & 01) xgp->gp_uhty -= 01;
			else xgp->gp_uhty |= 01;
			XtSetArg(lda[0], XmNlabelString,
				xgp->gp_uhstr[5 + (xgp->gp_uhty & 01)]);
			XtSetValues(widget, lda, 1);	break;
		case	2:	/* calculation type (XYomega vs local scatt) */
			if (xgp->gp_uhty & 02) xgp->gp_uhty -= 02;
			else xgp->gp_uhty |= 02;
			XtSetArg(lda[0], XmNlabelString,
			 xgp->gp_uhstr[7 + ((xgp->gp_uhty & 02) >> 1)]);
			XtSetValues(widget, lda, 1);	break;
		case	3:	/* colored variable type (Inten vs inten/sig)*/
			if (xgp->gp_uhty & 04) xgp->gp_uhty -= 04;
			else xgp->gp_uhty |= 04;
			XtSetArg(lda[0], XmNlabelString,
				xgp->gp_uhstr[3 + ((xgp->gp_uhty & 04) >> 2)]);
			XtSetValues(widget, lda, 1);	break;
		case	4:	/* re-set the value displayed */
			XtSetArg(lda[0], XmNvalue, &(xgp->gp_uhva));
			XtGetValues(widget, lda, 1);	break;
		case	5:	/* define which index is constant (h,k,l) */
			xgp->gp_uhco = (xgp->gp_uhco + 1) % 3;
			XtSetArg(lda[0], XmNlabelString,
				xgp->gp_uhstr[xgp->gp_uhco]);
			XtSetValues(widget, lda, 1);	break;
		case	6:	/* calc and do the display and leave menu up */
			uhv[0] = xgp->gp_uhco;	uhv[1] = xgp->gp_uhva;
			uhv[2] = xgp->gp_uhty;
			xgp->gp_nobinu = scatt_uh(tag, uhv, xgp->gp_at);
			if (xgp->gp_nobinu > 0)
			  { /* un-cover the text ports */
				valtocolors(xgp);
				XtUnmanageChild(xgp->gp_urhform);
				XtManageChildren(xgp->gp_tport, NTPORTS);
				prev_display = lditype;
				(xgp->gp_bl)->di_distype =
				 lmain->di_distype = lditype =
				 (xgp->gp_uhty & 04) ?  IOS_VS_HKL : I_VS_HKL;
				printf( "Displaying hkl vs UREFLS %s values\n",
					gdisnam(lditype));
				lmain->di_disx = lmain->di_wid /2;
				lmain->di_disy = lmain->di_ht / 2;
				redraw();	/* hkl display, etc. */
				filereport(xgp);	/* UREFLS file info */
			  }
			else printf(" Error %d from scatt_uh\n",
				xgp->gp_nobinu);
			/* this case drops through to the last one...  */
		case	7:	/* cancel this menu */
			if (j == 7) (xgp->gp_bl)->di_distype =
			 lmain->di_distype = lditype = prev_display;
			XtUnmanageChild(xgp->gp_urhform);
			XtManageChild(xgp->gp_poldmenu);
			XtManageChildren(xgp->gp_tport, NTPORTS);
			show_region(xgp); XtManageChild(xgp->gp_magwid); break;
	     }
	for (j = 0; j < 9; j++)
	   if (NULL != xgp->gp_uhstr[j])
	     {
		XmStringFree(xgp->gp_uhstr[j]);	xgp->gp_uhstr[j] = NULL;
	     }
}

static void mkuhkl(void)
{ /*	This creates a dialogue related to (hkl) plots of UREFLS files. */
	int		curval, j, ng, nranges, scwid;
	short		deltah, maximh, minimh, miw;
	char		*uname;
	XG_X11PROPS	*xgp;
	static Arg	lda[16];
	static char	*bls[] = { "hkl vs. I or I/sigma plots",
	 "integtype", "calctype", "vartype",
	 "hkl_scale", "whichindex", "Do it!", "Cancel" } ;
	static int	uhbx[] = { 10, 10, 10, 10, 10,110,110,110 } ;
	static int	uhby[] = {  5, 25, 45, 65, 95, 30, 57, 84 } ;
	static XmString	locstr;
	Widget	*wp;

	if (NULL == (xgp = managex11p(3))) return;
	if ((NULL == (uname = trname("UREFLS"))) || (0 != xgaccess(uname,4,1)))
	  {
		fprintf(stderr, "Can't open %s\n", uname); return;
	  }
	XtUnmanageChild(xgp->gp_poldmenu);	XtUnmanageChild(xgp->gp_magwid);
	/* first create the appropriate widgets if they don't exist already */
	miw = (xgp->gp_at)->di_minw;	update_uh();
	nranges = (xgp->gp_at)->di_nranges;
	if (NULL == xgp->gp_urhform)
	  {
		ng = 0;	XtSetArg(lda[ng], XmNx, miw + 8);	ng++;
		XtSetArg(lda[ng], XmNy, 20);		ng++;
		XtSetArg(lda[ng], XmNwidth, 205);	ng++;
		XtSetArg(lda[ng], XmNheight, 410);	ng++;
		XtSetArg(lda[ng], XmNbackground, xgp->gp_dbcol[nranges + 4]);
			ng++;
		XtSetArg(lda[ng], XmNforeground, xgp->gp_dbcol[0]);	ng++;
		xgp->gp_urhform = XmCreateBulletinBoard(xgp->gp_dimain,
		 "uref_hkl", lda, ng);
		locstr = strstufset(bls[0], xgp->gp_chpo2);
		ng = 0;	XtSetArg(lda[0], XmNlabelString, locstr); ng++;	
		XtSetArg(lda[ng], XmNx, uhbx[0]);	ng++;
		XtSetArg(lda[ng], XmNy, uhby[0]);	ng++;
		xgp->gp_urhbu[0] = XmCreateLabelGadget(xgp->gp_urhform,
		 "urh_label", lda, ng);
		XmStringFree(locstr);
		XtSetArg(lda[0], XmNindicatorOn, False);
		for (j = 1, wp = &(xgp->gp_urhbu[1]); j < 4; j++, wp++)
		   {
			XtSetArg(lda[1], XmNx, uhbx[j]);
			XtSetArg(lda[2], XmNy, uhby[j]);
			*wp = XmCreateToggleButtonGadget( xgp->gp_urhform,
				bls[j], lda, 3);
			XtAddCallback(*wp, XmNvalueChangedCallback,
				(XtCallbackProc)uhdo,XpNU);
		   }
			/* create the scale itself */
		printf("Checking ranges of h,k,and l...");
		if (0 >= (xgp->gp_nobinu = check_uh(uname, &minimh, &maximh)))
		  {
			printf("Warning: Error %2d from check_uh\n",
				xgp->gp_nobinu);
		  }
		/* if we plan to click on points on this plot, we're going
		 to need some statistics provided by the following routine */
		if (0 >= (xgp->gp_nobinu = exam_urefl(uname, 1, xgp->gp_at)))
		  {
			printf("Warning: error %d from exam_urefl\n",
				xgp->gp_nobinu);
		  }
		if ((deltah = maximh - minimh) > 138) scwid = 20 + deltah;
		else if (deltah > 68) scwid = 20 + 2 * deltah;
		else scwid = 20 + 4 * deltah;
		curval = (maximh >= 0 && minimh <= 0 ? 0 :
			(maximh + minimh) / 2);
		ng = 0; XtSetArg(lda[ng], XmNminimum, (int)minimh); ng++;
		XtSetArg(lda[ng], XmNmaximum, (int)maximh);	ng++;
		XtSetArg(lda[ng], XmNx, uhbx[j]);		ng++;
		XtSetArg(lda[ng], XmNy, uhby[j]);		ng++;
		XtSetArg(lda[ng], XmNscaleHeight, scwid);	ng++;
		XtSetArg(lda[ng], XmNscaleWidth, 20);		ng++;
		XtSetArg(lda[ng], XmNorientation, XmVERTICAL);	ng++;
		XtSetArg(lda[ng], XmNbackground, xgp->gp_dbcol[0]);	ng++;
		XtSetArg(lda[ng], XmNforeground, xgp->gp_dbcol[nranges]); ng++;
		XtSetArg(lda[ng], XmNshowValue, True);		ng++;
		locstr = strstufstr("index\nvalue", "");
		XtSetArg(lda[ng], XmNtitleString, locstr);	ng++;
		XtSetArg(lda[ng], XmNvalue, curval);		ng++;
		XtSetArg(lda[ng], XmNfontList, xgp->gp_dflis);		ng++;
		*wp = XmCreateScale(xgp->gp_urhform, bls[4], lda, ng);
		XtAddCallback(*wp, XmNvalueChangedCallback,
			(XtCallbackProc)uhdo, XpNU);
		XmStringFree(locstr);
		for (j = 5, wp = &(xgp->gp_urhbu[5]); j < 8; j++, wp++)
		   {
			XtSetArg(lda[0], XmNx, uhbx[j]);
			XtSetArg(lda[1], XmNy, uhby[j]);
			*wp = XmCreatePushButtonGadget(xgp->gp_urhform,
				bls[j],lda,2);
			XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)uhdo, (XtPointer)uname);
		   }
	  }
	XtSetArg(lda[0], XmNlabelString,
		xgp->gp_uhstr[5 + (xgp->gp_uhty & 01)]);
	XtSetValues(xgp->gp_urhbu[1], lda, 1);
	XtSetArg(lda[0], XmNlabelString,
		xgp->gp_uhstr[7 + ((xgp->gp_uhty & 02) >> 1)]);
	XtSetValues(xgp->gp_urhbu[2], lda, 1);
	XtSetArg(lda[0], XmNlabelString,
		xgp->gp_uhstr[3 + ((xgp->gp_uhty & 04) >> 2)]);
	XtSetValues(xgp->gp_urhbu[3], lda, 1);
	XtSetArg(lda[0], XmNvalue, xgp->gp_uhva);
	XtSetValues(xgp->gp_urhbu[4], lda, 1);
	XtSetArg(lda[0], XmNlabelString, xgp->gp_uhstr[xgp->gp_uhco]);
	XtSetValues(xgp->gp_urhbu[5], lda, 1);
	XtManageChildren(xgp->gp_urhbu, NURHBUTTS);
	XtManageChild(xgp->gp_urhform);
}

void make_buttons(void)
{ /*	This draws and labels the active buttons on the screen */
	char		blab[50];
	int		ib, igreen, imidgray, iyellow, miw, narg, nranges;
	Widget		*wp, *wm;
	XG_X11PROPS	*xgp;
	static Arg	lda[6];
	static XmString	locstr;
	static Widget	colorradi, cslabel;
	static char	*ml[] = { "Frames", "Urefls", "Multirefs","Overlays"} ;
	static char	*pl[]= { "Exit    ",
	 "Up Display", "Down Display", "Redraw", "Next Frame", "Prev Frame",
	 "Brass-plate", "Background", "Frame #s", "Files...",
	 "Intensity", "I/sigma", "Err(omega)", "Err(X)", "Err(Y)",
	 "Scatterplot", "hkl plots...", "Files...",
	 "Intensity", "I/sigma", "delI/s", "Files...",
	 "Calibrate", "Predict  ", "Border   ",
	 "Centroids", "Cal.Centroids", "Resol.Circles", "Row/Col Sums",
	 "Process" };
	static char		*costr[] =
	 { "Black->White", "Spectral    ", "White->Black", "RSpectral   " };

	if (NULL == (xgp = managex11p(3))) return;
	/* if needed, create menu bar in which toplevel cascades will live */
	nranges = (xgp->gp_at)->di_nranges;
	miw = (xgp->gp_at)->di_minw;
	if (NULL == xgp->gp_poldmenu)
	  {
		narg = 0;	XtSetArg(lda[narg], XmNx, miw + 8);	narg++;
		XtSetArg(lda[narg], XmNy,  0);			narg++;
		XtSetArg(lda[narg], XmNbackground, xgp->gp_dbcol[nranges + 2]);
			narg++;
		XtSetArg(lda[narg], XmNforeground, xgp->gp_dbcol[nranges + 6]);
			narg++;
		xgp->gp_poldmenu = XmCreateMenuBar(xgp->gp_dimain,
		 "pold_menu",lda,narg);	XtManageChild(xgp->gp_poldmenu);
	  }
	iyellow = (int)((double)nranges * 0.75);
	imidgray = nranges + NEXTRAS + nranges / 2;
	igreen = nranges / 2;
	 /* create the pulldown menus and cascade buttons. */
	for (ib = 0, wm = &(xgp->gp_toppu[0]), wp = &(xgp->gp_topbu[0]);
	  ib < NMENUS; ib++, wm++, wp++)
	   {
		strcpy(blab, ml[ib]);
		if (NULL == *wm)
		  {
			*wm = XmCreatePulldownMenu(xgp->gp_poldmenu,
			 blab, NULL, 0);
			 /* we must re-draw things after this menu vanishes: */
			XtAddCallback(*wm, XmNunmapCallback,
				(XtCallbackProc)redraw, XpNU);
		  }
		else XtManageChild(*wm);
		if (NULL == *wp)
		  {
			strcat(blab, "_button");	narg = 0;
			XtSetArg(lda[narg], XmNbackground,
			 xgp->gp_dbcol[imidgray]); narg++;
			XtSetArg(lda[narg], XmNforeground, xgp->gp_dbcol[0]);
			narg++;	locstr = strstufset(ml[ib], xgp->gp_chpo2);
			XtSetArg(lda[narg], XmNlabelString, locstr);	narg++;
			XtSetArg(lda[narg], XmNfontList, xgp->gp_dflis); narg++;
			XtSetArg(lda[narg], XmNsubMenuId, *wm);	narg++;
			*wp = XmCreateCascadeButton(xgp->gp_poldmenu,
			 blab, lda, narg);
			XtAddCallback(*wp, XmNactivateCallback,
			 (XtCallbackProc)xgprocess, XpNU);
			XtManageChild(*wp);	XmStringFree(locstr);
		  }
		else XtManageChild(*wp);
	   }
	 /* create the pushbuttons and toggles within those menus */
	for (ib = 0, wp = &(xgp->gp_dpush[0]); ib <= NPUSHES; ib++, wp++)
	   {
		if (NULL != *wp) continue; /* skip it if it's already OK */
		narg = 0;
		locstr = strstufset(pl[ib],
		 (ib ? xgp->gp_chpol : xgp->gp_chpo2));
		XtSetArg(lda[narg], XmNlabelString, locstr);	narg++;
		XtSetArg(lda[narg], XmNfontList, xgp->gp_dflis); narg++;
		strncpy(blab, pl[ib], 7); blab[7] = '\0';
		strcat(blab, "butt");
		if (!ib)	/* push button for "Exit" */
		  {
			XtSetArg(lda[narg], XmNx, miw + 288);	narg++;
			XtSetArg(lda[narg], XmNy, 492);		narg++;
			XtSetArg(lda[narg], XmNbackground,
				xgp->gp_dbcol[iyellow]); narg++;
			XtSetArg(lda[narg], XmNforeground,
				xgp->gp_dbcol[nranges]); narg++;
			*wp = XmCreatePushButton(xgp->gp_dimain, blab,lda,narg);
			XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)doexit, XpNU);
		  }
		else if (ib < FIRSTUREFLS) /* pushbuttons in frames menu */
		  {
			XtSetArg(lda[narg], XmNbackground,
				xgp->gp_dbcol[nranges + 4]); narg++;
			XtSetArg(lda[narg], XmNforeground,
				xgp->gp_dbcol[0]); narg++;
			*wp = XmCreatePushButton(xgp->gp_toppu[0],
			 blab, lda, narg);
			if (ib < FIRSTUREFLS - 1)
			 XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)chgframe,XpNU);
			else XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)newfile, XpNU);
		  }
		else if (ib < FIRSTMREF) /* pushbuttons in urefls menu */
		  {
			XtSetArg(lda[narg], XmNbackground,
				xgp->gp_dbcol[iyellow]); narg++;
			XtSetArg(lda[narg], XmNforeground,
				xgp->gp_dbcol[0]); narg++;
			strcat(blab, "ur");	/* unique naming */
			*wp = XmCreatePushButton(xgp->gp_toppu[1],
			 blab, lda, narg);
			if (ib < FIRSTMREF - 3)
			  XtAddCallback(*wp,XmNactivateCallback,
				(XtCallbackProc)showuref, XpNU);
			else if (ib == FIRSTMREF - 3)
			 XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)mkuscatt,XpNU);
			else if (ib == FIRSTMREF - 2)
			 XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)mkuhkl, XpNU);
			else XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)newfile, XpNU);
		  }
		else if (ib < FIRSTOVERL) /* pushbuttons in multiref menu */
		  {
			XtSetArg(lda[narg], XmNbackground,
				xgp->gp_dbcol[iyellow]); narg++;
			XtSetArg(lda[narg], XmNforeground,
				xgp->gp_dbcol[0]); narg++;
			strcat(blab, "mr"); /* unique naming */
			*wp = XmCreatePushButton(xgp->gp_toppu[2],
			 blab, lda, narg);
			if (ib < FIRSTOVERL -1)
			 XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)showmref, XpNU);
			else XtAddCallback(*wp, XmNactivateCallback,
				(XtCallbackProc)newfile, XpNU);
		  }
		else if (ib < NPUSHES)	/* toggle buttons in overlay menu */
		  {
			XtSetArg(lda[narg], XmNbackground,
				xgp->gp_dbcol[igreen]); narg++;
			XtSetArg(lda[narg], XmNforeground,
				xgp->gp_dbcol[nranges]); narg++;
			*wp = XmCreateToggleButton(xgp->gp_toppu[3],
			 blab, lda,narg);
			XtAddCallback(*wp, XmNvalueChangedCallback,
				(XtCallbackProc)overlay, XpNU);
		 }
		else {	/* Process button */
			XtSetArg(lda[narg], XmNx, miw + 220);	narg++;
			XtSetArg(lda[narg], XmNy, 492);		narg++;
			XtSetArg(lda[narg], XmNbackground,
				xgp->gp_dbcol[igreen]); narg++;
			XtSetArg(lda[narg], XmNforeground,
				xgp->gp_dbcol[nranges]); narg++;
			*wp = XmCreatePushButton(xgp->gp_dimain, blab,lda,narg);
			XtAddCallback(*wp, XmNactivateCallback,
			 (XtCallbackProc)xgprocess, XpNU);
		     }
		XmStringFree(locstr);
 	   }
		/* manage 'em all at once */
	XtManageChildren(xgp->gp_dpush, 1);	/* manage the exit button */
	XtManageChildren(&(xgp->gp_dpush[1]), NFRAMEBS); /* frame buttons */
	XtManageChildren(&(xgp->gp_dpush[FIRSTUREFLS]), NUREFLBS);
	XtManageChildren(&(xgp->gp_dpush[FIRSTMREF]), NMREFBS);
	XtManageChildren(&(xgp->gp_dpush[FIRSTOVERL]), NOVERBS);
	XtManageChildren(&(xgp->gp_dpush[NPUSHES]), 1);
	/* color-scheme management buttons */
	if (NULL == cslabel)
	  {
		narg = 0;
		XtSetArg(lda[narg], XmNx, miw + 228);	narg++;
		XtSetArg(lda[narg], XmNy,  340);		narg++;
		locstr = strstufstr("Color scheme:", "");
		XtSetArg(lda[narg], XmNlabelString, locstr);	narg++;
		cslabel = XmCreateLabelGadget(xgp->gp_dimain,
		 "cs_lab", lda, narg);
		XtManageChild(cslabel);	XmStringFree(locstr);
	  }
	else XtManageChild(cslabel);
	if (NULL == colorradi)
	  {
		narg = 0;	XtSetArg(lda[narg], XmNx, miw + 228);	narg++;
		XtSetArg(lda[narg], XmNy, 362);	narg++;
		XtSetArg(lda[narg], XmNradioAlwaysOne, True);	narg++;
		colorradi = XmCreateRadioBox(xgp->gp_dimain,
		 "color_radi", lda, narg);
	  }
	XtManageChild(colorradi);
	for (ib = 0, wp = &(xgp->gp_color[0]); ib < NSCHEMES; ib++, wp++)
	  if (NULL == *wp)
	    {
		narg = 0; XtSetArg(lda[narg], XmNset, (ib == xgp->gp_icn) ?
			True : False); narg++;
		locstr = strstufset(costr[ib], xgp->gp_chpol);
		XtSetArg(lda[narg], XmNlabelString, locstr);	narg++;
		*wp = XmCreateToggleButtonGadget(colorradi, costr[ib],
			lda, narg);
		XtAddCallback(*wp, XmNvalueChangedCallback,
			(XtCallbackProc)redraw, XpNU);
		XmStringFree(locstr);
	    }
	XtManageChildren(xgp->gp_color, NSCHEMES);
}

static Widget startbars(Widget *scrop, int hoff, int voff)
{ /*	This establishes callbacks for the scrollbars and sets the initial
	values of the scrollbars. The latter doesn't seem to work properly.
	The routine returns the widget associated with the scrolling. */
	int		narg;
	Arg		lda[2];
	static Widget	horbar, verbar;

	narg = 0;	XtSetArg(lda[narg], XmNhorizontalScrollBar, &horbar);
	narg++;		XtSetArg(lda[narg], XmNverticalScrollBar, &verbar);
	narg++;		XtGetValues(*scrop, lda, narg);
	if (XtCallbackHasSome != XtHasCallbacks(horbar,
	 XmNvalueChangedCallback)) XtAddCallback(horbar,
		XmNvalueChangedCallback, (XtCallbackProc)redraw, XpNU);
	if (XtCallbackHasSome != XtHasCallbacks(verbar,
	 XmNvalueChangedCallback)) XtAddCallback(verbar,
		XmNvalueChangedCallback, (XtCallbackProc)redraw, XpNU);
	if (XtCallbackHasSome != XtHasCallbacks(horbar, XmNdragCallback))
	 XtAddCallback(horbar, XmNdragCallback, (XtCallbackProc)redraw, XpNU);
	if (XtCallbackHasSome != XtHasCallbacks(verbar, XmNdragCallback))
	 XtAddCallback(verbar, XmNdragCallback, (XtCallbackProc)redraw, XpNU);
	narg = 0;	XtSetArg(lda[narg], XmNvalue, 0);
	narg++;	XtSetValues(horbar, lda, narg);
	narg = 0;	XtSetArg(lda[narg], XmNvalue, 0);
	narg++;	XtSetValues(verbar, lda, narg);
	XtManageChild(horbar);	XtManageChild(verbar);
	narg = 0;	XtSetArg(lda[narg], XmNvalue, hoff);
	narg++;	XtSetValues(horbar, lda, narg);
	narg = 0;	XtSetArg(lda[narg], XmNvalue, voff);
	narg++;	XtSetValues(verbar, lda, narg);
	XtManageChild(verbar); XtManageChild(horbar); XtManageChild(*scrop);
	return *scrop;
}

static void make_graphboxes(Colormap tma, Widget *scrop)
{ /*	This defines a couple of widgets associated with the actual
	detector-image viewing areas and associated windows for them */
	int		ldih, ldiw, miw, narg, nranges;
	Arg		lda[8];
	XG_X11PROPS	*xgp;
	DISP_ATTRS	*lmain;
	PAU		gbu;
	static int	horoff, veroff;
	static Font	helvn12;
	static char	imtitle[] = "imagedraw";
	static char	magtitle[] = "magnifydraw";
	static char	scrtitle[] = "scrolledim";
	static xgulong	eventmask =
		 ExposureMask | ButtonMotionMask | PointerMotionHintMask |
		 PointerMotionMask | ButtonPressMask | OwnerGrabButtonMask;

	if (NULL == (xgp = managex11p(3))) return;
	lmain = xgp->gp_at;
	miw = lmain->di_minw;	ldih = lmain->di_ht;
	nranges = lmain->di_nranges;
	ldiw = lmain->di_wid;
	horoff = (ldiw < lmain->di_minw) ?
		0 : 16 + (ldiw - lmain->di_minw) / 2;
	veroff = (ldih < lmain->di_minw ) ?
		0 : 16 + (ldih - lmain->di_minw ) / 2;
	if ((horoff != 0) || (veroff != 0)) printf(
	 "Sliders will be created, initially offset by %3dx%3d pixels\n",
		horoff, veroff);
	if (NULL == *scrop)
	  {
		narg = 0;	XtSetArg(lda[narg], XmNx, 5);	narg++;
		XtSetArg(lda[narg], XmNy, 5);	narg++;
		XtSetArg(lda[narg], XmNscrollingPolicy, XmAUTOMATIC);	narg++;
/* XtSetArg(lda[narg], XmNscrollBarDisplayPolicy, XmAS_NEEDED); narg++; */
		XtSetArg(lda[narg], XmNwidth, lmain->di_minw);	narg++;
		XtSetArg(lda[narg], XmNheight, lmain->di_minh);	narg++;
		*scrop = XmCreateScrolledWindow(xgp->gp_dimain,
		 scrtitle, lda, narg);
		XtManageChild(*scrop);
	  }
	if (NULL == xgp->gp_imwid)
	  {
		narg = 0;	XtSetArg(lda[narg], XmNx, 0);	narg++;
		XtSetArg(lda[narg], XmNy, 0);	narg++;
		XtSetArg(lda[narg], XmNbackground,
			xgp->gp_dbcol[0]);	narg++;
		XtSetArg(lda[narg], XmNwidth, ldiw);	narg++;
		XtSetArg(lda[narg], XmNheight, ldih);	narg++;
		XtSetArg(lda[narg], XmNcolormap, tma);	narg++;
		XtSetArg(lda[narg], XmNresizePolicy, XmRESIZE_NONE);	narg++;
		xgp->gp_imwid = XmCreateDrawingArea(*scrop, imtitle, lda, narg);
		XtAddEventHandler(xgp->gp_imwid, eventmask, FALSE,
			(XtEventHandler)display_pixel, XpNU);
	  }
	XtManageChild(xgp->gp_imwid);
	if (NULL == xgp->gp_magwid)
	  {
		narg = 0; XtSetArg(lda[narg], XmNx, miw + 12); /* move over */
		narg++; XtSetArg(lda[narg], XmNy, 38); /* allow for menubar */
		narg++; XtSetArg(lda[narg], XmNbackground,
		 xgp->gp_dbcol[0]);
		narg++; XtSetArg(lda[narg], XmNwidth, 170);
		narg++; XtSetArg(lda[narg], XmNheight,170); narg++;
		xgp->gp_magwid = XmCreateDrawingArea(xgp->gp_dimain,
			magtitle, lda ,narg);
		XtAddEventHandler(xgp->gp_magwid, eventmask, FALSE,
			(XtEventHandler)display_pixel, XpNU);
		XtManageChild(xgp->gp_magwid);
	  }
	xgp->gp_wi = XtWindow(xgp->gp_imwid);		/* define the window */
	xgp->gp_maw = XtWindow(xgp->gp_magwid);
	*scrop = startbars(scrop, horoff, veroff);
	/* the image GC itself is the default; the overlay GC is intended
	to change, and it starts out with the background color being black
	and the foreground being white. */
	if ((Window)0 != xgp->gp_wi)
	  {
		if (NULL == xgp->gp_igc) xgp->gp_igc = XCreateGC(xgp->gp_di,
		 xgp->gp_wi, 0, (XGCValues *)NULL);
		if (NULL == xgp->gp_ogc) xgp->gp_ogc = XCreateGC(xgp->gp_di,
		 xgp->gp_wi, 0, (XGCValues *)NULL);
		XSetBackground(xgp->gp_di, xgp->gp_ogc, xgp->gp_dbcol[0]);
		XSetForeground(xgp->gp_di, xgp->gp_ogc, xgp->gp_dbcol[nranges]);
		if (None == helvn12)
		  {
			if (None == (helvn12 = XLoadFont(xgp->gp_di,
			 "*Helvetica-Bold-R*120*")))
			 fprintf(stderr,"Warning: no text on plots\n");
			switch (XSetFont(xgp->gp_di, xgp->gp_ogc, helvn12))
			 {
			   case	BadAlloc: fprintf(stderr,
			    "Trouble allocating font for overlays\n"); break;
			   case	BadFont: fprintf(stderr,
			    "helvn12 = 0x%x is a bad font\n", (int)helvn12);
				break;
			   case	BadGC:	gbu.vpt = (void *)(xgp->gp_ogc);
				fprintf(stderr,
			    "xgp->gp_ogc = 0x%x is badGC for fontlery\n",
				gbu.ui);	break;
			   default : break;
			 }
		  }
	  }
}

int dispkill(int irv)
{ /*	Free up memory used in the display program */
	XG_X11PROPS	*xgp;

	if (NULL != (xgp = managex11p(3)))
	  {
		memzap((AVOID **)&(xgp->gp_at), "xgp->gp_at in pdisplay");
	  }
	title_mainwin(NULL);	(void)managekill(0);
	(void)reflstats(NULL, NULL);	return irv;
}

int graphics(int argc, char **argv, DISP_ATTRS *mainatt)
{ /*	this sets up the graphics calls for the display program `pdisplay'. */
	int			htdo, htup, ldhi, mih, miw, newinp;
	int			nranges, result, thmin, twmin, wido, wiup;
	char			*cp, wintitle[80], colorind[20];
	MAIL			*lhd;
	DISP_ATTRS		*lmain;
	static XG_X11PROPS	*xgp;
	Arg			lda[8];
	Widget			scro;
	static char		*ndfon2= "*New Century Schoolbook-Bold-I*140*";
	static XmString		locstr;
	static XFontStruct	*dfon1, *dfon2;
	static XtAppContext  	dapp_ctxt;
	static XmFontList	dflis1;
	static XmFontListEntry	flentry1, flentry2;
	static Colormap		tmap;

	/* set up a jump from before the X stuff begins.  This allows
	the "Exit" button to jump here after it closes the X display. */
	if (NULL == (xgp = managex11p(07))) return dispkill(EALLOC);
	if (setjmp(xgp->gp_jbuf))
	  {
		(void)dispkill(EOK);	(void)managex11p(0); return EOK;
	  }
	if (NULL == xgp->gp_fdbuf) xgp = managex11p(053);
	lmain = xgp->gp_at = mainatt;
	xgp->gp_bl = xgp->gp_at + 1;
	lmain->di_disflags = DISPVERBOSE1; /* default verbosity */
	/* deliberately silly values for maxv and valmul so that
	 grab_image() will recalculate them. */
	lmain->di_maxv = -1;	lmain->di_valmul = -1.;
	lmain->di_distype = 0;	/* default is image display */
	lmain->di_maxw = lmain->di_maxh = 752;
	lmain->di_nranges = nranges = NRANGES;
	xgp->gp_uhco = 2;	/* default display is constant l */
	xgp->gp_uhva = 0;	xgp->gp_uhty = 2;
	(void)grab_image(lmain->di_darkfla, xgp->gp_at);
	attr2locals();
	newinp = 0;	/* looking for input from stdin */
	XtToolkitInitialize();	/* get the X tools initialized */
	dapp_ctxt = XtCreateApplicationContext();
	(void)XtAppAddInput(dapp_ctxt, newinp, (XtPointer)XtInputReadMask,
		 (XtInputCallbackProc)fdproc, (XtPointer)(xgp->gp_fdbuf));
/* before attempting to open the display, we look to see if the environment
 variable "DISPLAY" is defined.  If not, we try to set it to :0 */
	if ((NULL == (cp = getenv("DISPLAY"))) || ('\0' == *cp))
	  {
		if (1 > set_logical("DISPLAY", ":0"))
		  {
			fprintf(stderr,
		"Unable to define the environment variable DISPLAY as `:0'\n");
			return EBADVALUE;
		  }
	  }
	if (NULL == (xgp->gp_di = XtOpenDisplay(dapp_ctxt,(String)NULL,argv[0],
#ifdef sgi
		"dispimclass", NULL, 0, (int *)&argc, argv)))
#else
		"dispimclass", NULL, 0, (XGCAR *)&argc, argv)))
#endif
          {
		fprintf(stderr, "%s:  Can't open display\n", argv[0]);
		return EOPENERR;
	  }
	getbitmapv(xgp);
	dispscale(xgp);	/* get the scale of the image right */
	/* experiment */
	lmain->di_minw = ((lmain->di_maxw - 120) > lmain->di_wid) ?
		lmain->di_wid + 40 : lmain->di_maxw - 80;
	if (lmain->di_minw > lmain->di_wid + 8)
		lmain->di_minw = lmain->di_wid + 8;
	lmain->di_minh = ((lmain->di_maxh - 120) > lmain->di_ht) ?
		lmain->di_ht + 40 : lmain->di_maxh - 80;
	if (lmain->di_minh > lmain->di_ht + 8)
		lmain->di_minh = lmain->di_ht + 8;
	miw = lmain->di_minw;	mih = lmain->di_minh;
	wiup = lmain->di_wiup;	wido = lmain->di_widown;
	htup = lmain->di_htup;	htdo = lmain->di_htdown;
	ldhi = lmain->di_ht;
	/*	Define the fonts used in the internal character sets. */
	xgp->gp_chpol = "pol1";
	xgp->gp_chpo2 = "pol2";
	xgp->gp_uhty = 2;	/* type of (hkl) intensity display */
	xgp->gp_dflis = dflis1 = NULL;
	xgp->gp_iscal[0] = xgp->gp_iscal[1] = NULL;
	if ((dfon1 = XLoadQueryFont(xgp->gp_di, "*Helvetica-Bold-R*120*")))
	  {
		flentry1 = XmFontListEntryCreate(xgp->gp_chpol,
			XmFONT_IS_FONT, (XtPointer)dfon1);
		dflis1 = XmFontListAppendEntry(NULL, flentry1);
	  }
	if ((dfon2 = XLoadQueryFont(xgp->gp_di, ndfon2)))
	  {
		flentry2 = XmFontListEntryCreate(xgp->gp_chpo2,
			XmFONT_IS_FONT, (XtPointer)dfon2);
		xgp->gp_dflis = XmFontListAppendEntry(dflis1, flentry2);
	  }
	if (NULL == xgp->gp_dtoplev)
	  {
		XtSetArg(lda[0], XmNallowShellResize, False);
		XtSetArg(lda[1], XmNx, 50);	XtSetArg(lda[2], XmNy, 50);
		xgp->gp_dtoplev = XtAppCreateShell((String)argv[0],
		 (String)NULL, applicationShellWidgetClass, xgp->gp_di, lda, 3);
	  }
    /* Call XmCreateForm to fetch and create the widgets.  */
	/* temporary window title */
	if (NULL == xgp->gp_dimain)
	  {
		sprintf(wintitle, "XGEN V.%s Image Display", XGEN_VERSION);
		locstr = strstufstr(wintitle, "");
		XtSetArg(lda[0], XmNdialogTitle, locstr);
		twmin = miw + 376;	if (twmin < 800) twmin = 800;
		XtSetArg(lda[1], XmNwidth, twmin);
		thmin = mih + 8;	if (thmin < 640) thmin = 640;
		if (ldhi < 632) XtSetArg(lda[2], XmNheight, thmin);
		else XtSetArg(lda[2], XmNheight, ldhi + 8);
		xgp->gp_dimain = XmCreateForm(xgp->gp_dtoplev,
		 "dispim_main", lda, 3);
	  }
	if (NULL != (cp = getenv("COLORSCHEME"))) /* colorscheme specified? */
	  {
		zeromem(colorind, 20);	strncpy(colorind, cp, 18);
		make_upper(colorind);
		if (!strncmp("BLA", colorind, 3)) xgp->gp_icn = 0;
		else if (!strncmp("SPE", colorind, 3)) xgp->gp_icn = 1;
		else if (!strncmp("WHI", colorind, 3)) xgp->gp_icn = 2;
		else if (!strncmp("RSP", colorind, 3)) xgp->gp_icn = 3;
		else xgp->gp_icn = 1;
	  }
	else xgp->gp_icn = 1;	/* default color scheme is Spectral */
	/* define color map(s) and colors to use */
	start_dcolors(xgp, &tmap);
	XtSetArg(lda[0], XmNbackground, xgp->gp_dbcol[nranges + 4]);
	XtSetValues(xgp->gp_dimain, lda, 1);
	XtManageChild(xgp->gp_dimain);
	XtRealizeWidget(xgp->gp_dtoplev); /* so managed hierarchy displays */
	XmStringFree(locstr);
	make_buttons();		/* create the controlling buttons */
	scro = NULL;	make_graphboxes(tmap, &scro);
	xgp->gp_col = xgp->gp_dbcol[(xgp->gp_at)->di_nranges];
	make_textports(xgp); /* information boxes */
	if (EOK != (result = store_image())) return result;
	show_iscales(xgp); /* max&multiplier scales */
	lhd = lmain->di_head + lmain->di_fmno % lmain->di_nmbox;
	lmain->di_roco = 0;	/* default image rotation convention */
	lmain->di_disx = ((int)lhd->zmaxx) * wiup / wido;
	lmain->di_disy = ((int)lhd->zmaxy) * htup / htdo;
	/* Simulating a mouse-click would fit here if I could get it to work:
	simulate_click(horbar);	simulate_click(verbar); */
	draw_it();	/* show the image to start with */
	show_table(xgp);
	/*summarize contents of file */
	filereport(xgp);
	show_region(xgp);
	init_inten(xgp); /* display I(zmaxx, zmaxy) */
	title_mainwin(xgp);
	XtAppMainLoop(dapp_ctxt); /* loop and process user-created events */
	/* UNREACHABLE */
	return EOK;
}

/* end of dispx.c */
