/* dispxmag.c -- utilities for X-windows displays of magnified results

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Jun 2006: image memory moved from the XG_X11PROPS structure to
		the DISP_ATTRS structure so things are better localized
	12 Jun 2006: show_actual,show_stats,showrefmr moved here
	 so they're static; show_stats reformatted
	22 Jan 2006: indexname call updated
	 7 Jan 2006: getxgattrs replaced with direct determination;
		show_mref replaced by showrefmr()
	11 Nov 2005: replace newdump(drp) with dumpimage(drp, 1)
	 5 Oct 2005: housekeeping
	 9 Sep 2005: blowupraw() & threedblow() regional bg values from di_regb
	20 Oct 2004: make_ftab called unconditionally from display_pixel()
	 9 Oct 2004: show_stats reorganized
	 8 Oct 2004: in blowupraw(), handle (Count-Background) < 0 correctly;
		bugfix in create_blowup() in call to hkl_axes();
		shorter label on "Cursor:" output
	27 Aug 2004: routines used only here are declared as static;cosmetics
	31 Jan 2004: skip limit checks in threedblow() before calling read_box
		because they're done internally. This means the call
		will work even if the header isn't populated in advance.
	30 Jan 2004: background-subtract the 3D-profile images
	21 Jan 2004: license message inserted
	 6 Jan 2004: globals replaced with elements of DISP_ATTRS structs;
		calls to indexname() require FRAMEINDEX arg
	13 Dec 2003: eliminate 4 unused variables from blowupraw()
		plus dioverl from show_region() and actual from display_pixel()
	21 Sep 2003: rename getattrs to getxgattrs--
		prevents collision with curses routine
	19 Sep 2003: explicitly null out the one-byte and two-byte
		versions when nbperp == 4, etc.
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: correct bug in Linux version for magimage->byteorder
	15 Sep 2003: created out of dispx.c
 Relevant changes made to dispx.c:
	16 Aug 2003: major workover to reduce reliance on global variables,
		at the cost of some extra elements in the limits[] array
		and a few extra arguments to some functions
	12 Aug 2003: add new overlays for resolution circles &
		row+column sums; change displays of predictions to
		reflections with large gamma values cover more images
	11 Aug 2003: subtract local sliding-averaged background from
		displayed intensity values
	 6 Aug 2003: change Motif global variables to static globals
	 1 May 2003: add additional data-processing functions;
		improve handling of data in blowups
	25 Mar 2003: major rewrite to enable input from stdin
		and to allow eight different image orientations
	14 Mar 2003: corrected bug in one call to strstufstr()
	 5 Jan 2003: reorganized source-file header
	20 Jul 2002: digital dumps now with newdump(), not dumpimage()
	25 Mar 2002: removed MSI attachments
	 5 May 1994: source-file header, POLYDET->XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: utilities for blown-up displays of detector images and other
	objects directly on an X window display governed under Motif.
	Unlike ../displays/dispcalc.c, these routines ARE Motif-specific,
		but they do not depend very much on X-GEN itself.
   Cross references: ./disp*.c, ../displays/disp*.c
*************************************************************************/
#define		NEED_MOTIF	1
#include	"xgmdefs.h"
#include	"xgmodel.h"
#define		NUREFLBS	8	/* # of urefls-related buttons */
#define		MIN_SIZE	(5000)	/* min size of printing tank */

static void blowupraw(XG_X11PROPS *xxg, int *dili)
{ /*	This sets up a blowup region from raw data. */
	int		dx, dy, htdo, htup, imht, imwid, jrange, lnrx, lnry;
	int		magh, magw, nbperp, ndcolors, nranges, roco;
	int		x, xmn, y, ymn, tval, tcoi, wido, wiup, imcoo[4];
	char		*magd, *ocp;
	short		*osp;
	xglong		*locof, *olp;
	DISP_ATTRS	*dispatt, *maina;
	Pixel		tcl;
	MAIL		*lhd;
	PAU		mau;

	dispatt = xxg->gp_bl;		nbperp = dispatt->di_nbperp;
	htdo = dispatt->di_htdown;	htup = dispatt->di_htup;
	wido = dispatt->di_widown;	wiup = dispatt->di_wiup;
	magw = dispatt->di_magw;	magh = dispatt->di_magh;
	roco = dispatt->di_roco;	maina = xxg->gp_at;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	imwid = dispatt->di_imwid;	imht = dispatt->di_imht;
	nranges = dispatt->di_nranges;	ndcolors = nranges + 6;
	lnrx = xxg->gp_nregx;		lnry = xxg->gp_nregy;
	magd = maina->di_magm;		locof = dispatt->di_regb;
	xmn = dili[0];	ymn = dili[1];
	if (nbperp == 1)
	  {
		ocp = magd;	osp = (short *)0;	olp = (xglong *)0;
	  }
	else if (nbperp == 2)
	  {
		mau.ucpt = (u_char *)magd;	osp = (short *)(mau.uspt);
		ocp = (char *)0;		olp = (xglong *)0;
	  }
	else {
		mau.ucpt = (u_char *)magd;	olp = (xglong *)(mau.vpt);
		ocp = (char *)0;		osp = (short *)0;
	     }
	for (dy = 0, y = ymn * htdo * 8; dy < magh; dy++, y += htdo)
	   {
		imcoo[1] = y / (8 * htup);
		for (dx = 0, x = xmn * wido * 8; dx < magw; dx++, x += wido)
		   {
			imcoo[0] = x / (8 * wiup);
			disrot(roco, 0, imwid, imcoo);
			if ((0 <= imcoo[2]) && (0 <= imcoo[3]) &&
			 (imcoo[2] < imwid) && (imcoo[3] < imht))
				tval = imvalue(lhd, imcoo[2], imcoo[3]) -
				 locba(locof, imcoo[3], imcoo[2], lnry, lnrx);
			else	tval = 0;
			if (tval < 1) jrange = 0;
			else for (jrange = 0; jrange < nranges - 1; jrange++)
				if (tval < dispatt->di_range[jrange]) break;
			tcoi = xxg->gp_coi[jrange];
			tcl = xxg->gp_dbcol[tcoi];
			if (nbperp == 1) *ocp++ = tcl;
			else if (nbperp == 2) *osp++ = tcl;
			else	*olp++ = tcl;
		   } /* end of loop through dx */
	   } /* end of loop through dy */
}

static int threedblow(XG_X11PROPS *xxg, int zcurr, int duwidth, int duheight)
{ /*	This establishes a component (one image's worth)
	of a three-dimensional blowup */
	register int	x, y;
	int		boxflags, curba, delfm, delta, halfheight, halfwidth;
	int		jrange, l1, l2, l3, lnbp, mah, maw, nranges, retv;
	int		roco, tcoi, treg, totx, toty, tval, xcent, xdel, xdi;
	int		xmn, xmx, ycent, ydel, ydi, ymn, ymx, zeta;
	int		*ip, *ipo, *ipt, imcoo[4], xyli[4];
	char		*magd, *cp, *ocp;
	short		*osp;
	xglong		*olp;
	MAIL		*lhd;
	DISP_ATTRS	*lattr;
	PAU		thu;
	static int	*ipz;
	static size_t	curisize, nowsize;

	lattr = xxg->gp_at;
	if (NULL == (cp = indexname(zcurr)))
	  {
		printf("Cannot find filename for image number %d\n", zcurr);
		return EOPENERR;
	  }
	xcent = lattr->di_currx;	ycent = lattr->di_curry;
	lnbp = lattr->di_nbperp;	nranges = lattr->di_nranges;
/* establish limits of box to be dumped. Sometimes when we reach this spot,
 the header associated with the image we want isn't actually populated yet.
 So in those instances we need to get the relevant information from
 somewhere else. */
	lhd = lattr->di_head + (zcurr % lattr->di_nmbox);
	halfwidth = (duwidth > 1 ? duwidth / 2 : 1);
	halfheight = (duheight > 1 ? duheight / 2 : 1);
	xmn = xcent - halfwidth;	xmx = xcent + halfwidth;
	ymn = ycent - halfheight;	ymx = ycent + halfheight;
	maw = lattr->di_magw;		mah = lattr->di_magh;
/* reserve memory for the box of data if needed. We make it double-sized
 so that we can rotate it into the proper orientation */
	toty = ymx + 1 - ymn;	totx = xmx + 1 - xmn;	nowsize = totx * toty;
	xyli[0] = xmn;	xyli[1] = xmx;	xyli[2] = ymn;	xyli[3] = ymx;
	if ((NULL != ipz) && (curisize < nowsize))
		memzap((AVOID **)&ipz, "ipz in threedblow");
	if (NULL == ipz)
	  {
		curisize = (nowsize < MIN_SIZE) ? MIN_SIZE : nowsize;
		if (NULL == (ipz = (int *)callop(2 * curisize, sizeof (int),
		 "ipz in threedblow"))) return EALLOC;
	  }
	ipt = ipz + curisize;	boxflags = GET_RUNHEADER | ANNOUNCE_BOX;
	if (FRAME_OK != (retv = read_box(cp, boxflags, lhd, xyli,ipz)))
	  {
		fprintf(stderr, "read_box for image %s returned %d\n",
			cp, retv);
		return retv;
	  }
	/* subtract away the local background value and
	 rotate the image according to the current rotation convention */
	roco = lattr->di_roco;
	treg = xxg->gp_nregx * (xxg->gp_nregy * ycent / lhd->ydimen) +
		xxg->gp_nregx * xcent / lhd->xdimen;
	if (treg < xxg->gp_nregx * xxg->gp_nregy)
		curba = *(lattr->di_regb + treg); /* regional bgnd estimate */
	else	curba = 0;	/* we're confused, so subtract zero */
	for (imcoo[1] = y = 0, ip = ipz; y < duheight; y++, (imcoo[1])++)
	   {
		for (imcoo[0] = x = 0; x < duwidth; x++, ip++, (imcoo[0])++)
		   {
			disrot(roco, 1, duwidth, imcoo);
			ipo = ipt + imcoo[3] * duwidth + imcoo[2];
			*ipo = *ip - curba;
			if (*ipo < 0) *ipo = 0; /* don't allow negatives */
		   }
	   }
	/* now we use this information. First we figure out which box
	 out of the image we're actually populating here. This code
	 is specific to producing seven images, arranged thus:
	.____________________________________.
	|                                    |
	|      **********  ***********       |
	|      **********  ***********       |
	|      **********  ***********       |
	|      **********  ***********       |
	|      **********  ***********       |
	|                                    |
	| **********  **********  ********** |
	| **********  **********  ********** |
	| **********  **********  ********** |
	| **********  **********  ********** |
	| **********  **********  ********** |
	|                                    |
	|      **********  ***********       |
	|      **********  ***********       |
	|      **********  ***********       |
	|      **********  ***********       |
	|      **********  ***********       |
	|____________________________________|
	*/
	delta = duwidth / 2;	zeta = 3;
	delfm = zcurr + 3 - lattr->di_fmno;
	if ((delfm < 0) || (delfm > 6))
	  {
		printf(
		 "Can't use data from image %d = %s: delfm=%d out of range\n",
			zcurr, cp, delfm);
		return EBADVALUE;
	  }
	switch (delfm)
	  {
		case	0:	default:
			xdel = delta + zeta;	ydel = zeta;	break;
		case	1:	xdel = delta + 2 * zeta + duwidth;
				ydel = zeta;	break;
		case	2:	xdel = zeta;
				ydel = duheight + 2 * zeta;	break;
		case	3:	xdel = 2 * zeta + duwidth;
				ydel = duheight + 2 * zeta;	break;
		case	4:	xdel = 3 * zeta + 2 * duwidth;
				ydel = duheight + 2 * zeta;	break;
		case	5:	xdel = delta + zeta;
				ydel = 2 * duheight + 3 * zeta;	break;
		case	6:	xdel = delta + 2 * zeta + duwidth;
				ydel = 2 * duheight + 3 * zeta;	break;
	  }
	/* now we populate the appropriate magnification-image. We always
	 populate a tiny 2x2 box */
	l1 = 1;	l2 = mah;	l3 = mah + 1;	magd = lattr->di_magm;
	for (y = 0, ydi = ydel, ip = ipt; y < duheight; y++, ydi++)
	 for (x = 0, xdi = xdel; x < duwidth; x++, xdi++, ip++)
	  {
		tval = *ip;
		if (lnbp == 1)
		  {
			ocp = magd + 2 * (maw * ydi + xdi);
			if (tval < 1) *ocp = xxg->gp_dbcol[xxg->gp_ndcolors];
			else {
				for (jrange = 0; jrange < nranges-1; jrange++)
				   if (tval < lattr->di_range[jrange])
					break;
				tcoi = xxg->gp_coi[jrange];
				*ocp = xxg->gp_dbcol[tcoi];
			     }
			*(ocp + l3) = *(ocp + l2) = *(ocp + l1) = *ocp;
		  }
  		else if (lnbp == 2)
		  {
			thu.ucpt = (u_char *)magd;
			osp = ((short *)(thu.uspt)) + 2 * (maw * ydi + xdi);
			if (tval < 1) *osp = xxg->gp_dbcol[xxg->gp_ndcolors];
			else {
				for (jrange = 0; jrange < nranges-1; jrange++)
				   if (tval < lattr->di_range[jrange])
					break;
				tcoi = xxg->gp_coi[jrange];
				*osp = xxg->gp_dbcol[tcoi];
			     }
			*(osp + l3) = *(osp + l2) = *(osp + l1) = *osp;
		  }
  		else if (lnbp == 4)
		  {
			thu.ucpt = (u_char *)magd;
			olp = ((xglong *)(thu.vpt)) + 2 * (maw * ydi + xdi);
			if (tval < 1) *olp = xxg->gp_dbcol[xxg->gp_ndcolors];
			else {
				for (jrange = 0; jrange < nranges-1; jrange++)
				 if (tval < lattr->di_range[jrange])
					break;
				tcoi = xxg->gp_coi[jrange];
				*olp = xxg->gp_dbcol[tcoi];
			     }
			*(olp + l3) = *(olp + l2) = *(olp + l1) = *olp;
		  }
	    } /* end of loops through y, x values */
	return retv;
}

static void make_blowupwin(XG_X11PROPS *xxg)
{ /*	This creates and displays the blowup window */
	int		magh, magw, result;
	static XImage	*magimage;
	DISP_ATTRS	*lattr;

	lattr = xxg->gp_at;	magw = lattr->di_magw;	magh = lattr->di_magh;
	if (NULL == magimage)
	  {
		if (NULL == (magimage = XCreateImage(xxg->gp_di,
		 xxg->gp_disv, xxg->gp_dep, xxg->gp_imf, xxg->gp_imo,
		 lattr->di_magm, magw, magh, xxg->gp_bmp, xxg->gp_bpl)))
		  {
			fprintf(stderr, "Unable to create image!\n"); return;
		  }
	  }
#ifdef linux
	magimage->byte_order = ImageByteOrder(xxg->gp_di);
#else
#ifdef LITTLEND
	magimage->byte_order = LSBFirst;
#else
	magimage->byte_order = MSBFirst;
#endif
#endif
	if ((Window)0 == xxg->gp_maw)
	  {
		fprintf(stderr, "Image window does not exist yet\n");
		return;
	  }
	result = XPutImage(xxg->gp_di, xxg->gp_maw, xxg->gp_igc, magimage,
	 0, 0, 0, 0, magimage->width, magimage->height);
	if (result == BadDrawable) fprintf(stderr, "Bad drawable\n");
	else if (result == BadGC) fprintf(stderr, "Bad GC\n");
	else if (result == BadMatch) fprintf(stderr, "Bad Match\n");
	else if (result == BadValue) fprintf(stderr, "Bad Value\n");
}

static void create_blowup(XG_X11PROPS *xxg, int bou[])
{ /*	This creates an 8-fold magnified picture of the data in the rectangle
	bounded by (bou[0],bou[1]) and (bou[2],bou[3]) and displays it.
	If we're displaying actual image data, we go back to the original raw
	data to do this; if we're displaying statistics, we just blow them up*/
	int		lditype, lnbp, masize;
	DISP_ATTRS	*lattr;

	lattr = xxg->gp_at;	lditype = lattr->di_distype;
	if (NULL == lattr->di_rotm) return; /* image must be set up first */
	if (0 >= (lnbp = lattr->di_nbperp)) return; /*display info needed 1st*/
	masize = lattr->di_magw * lattr->di_magh * lnbp;
	if (NULL == lattr->di_magm)
	  {
		if (NULL == (lattr->di_magm = (char *)mallop(masize,
		 "lattr->di_magm in create_blowup"))) return;
	  }
	zeromem((void *)(lattr->di_magm), (size_t)masize);
	/* for displays of statistics, we just blow up the ordinary display:*/
	if (lditype) expand_sdisplay(lattr, bou);
	/* whereas for displays of detector data, we actually blow up the
	 raw data in the appropriate spot.  If the detector dimensions are
	 small enough there's no difference, but if the detector width or
	 height is > 800 we need more information to do this. As above, we
	 have to do this differently depending on the # of bytes/pixel */
	else blowupraw(xxg, bou);
	make_blowupwin(xxg);	 /* create the blowup window if needed */
	if (lditype < 0) draw_axes(xxg, 1);
	else if ((lditype == NUREFLBS - 2) || (lditype == NUREFLBS - 1))
		hkl_axes(xxg, 1);
}

static void show_hkl(XG_X11PROPS *xxg, double *phkl)
{ /*	This sticks an estimate of (h,k,l) into an appropriate place */
	char		htext[42];
	DISP_ATTRS	*dispatt;

	dispatt = xxg->gp_at;
	if (dispatt->di_distype) htext[0] = '\0';
	else {
		if (phkl[3] < DFUZZ) phkl[3] = 999.;
		else phkl[3] = 0.5 / phkl[3];
		sprintf(htext, "hkl,D = %5.1lf %5.1lf %5.1lf %6.2lf",
			phkl[0], phkl[1], phkl[2], phkl[3]);
	     }
	replace_string(xxg->gp_tport[1], htext, xxg->gp_chpol);
}

static void show_stats(XG_X11PROPS *xxg, double *istat)
{ /*	This summarizes the reflection statistics found in `reflstats'
	onto the window */
	char		stbuf[120];
	double		cyc;
	DISP_ATTRS	*dispatt;
	static char	stl1[] = "<Count> = ";
	static char	stl2[] = "Median,<Bg>= ";
	static char	stl3[] = "I, sigma: ";

	dispatt = xxg->gp_at;
	zeromem((AVOID *)stbuf, 120);
	if (!dispatt->di_distype)
	  {
		cyc = (double)(dispatt->di_imht - 1) - istat[6];
		zeromem((AVOID *)stbuf, 120);
		sprintf(stbuf, "Near %5d %5d:\n<Xc,Yc>=%5.1f %5.1f\n",
		 dispatt->di_currx, dispatt->di_curry, istat[5], cyc);
		sprintf(&(stbuf[39]),
		 "%s%5.1f\n%s%5.1f %5.1f\n%s%6.0f %5.1f\n",
			stl1, istat[0], stl2, istat[1],
			istat[2], stl3, istat[3], istat[4]);
	   }
	replace_string(xxg->gp_tport[0], stbuf, xxg->gp_chpol);
}

void show_region(XG_X11PROPS *xxg)
{ /*	This displays a (roughly) 21x21 region around the specified point.
	It computes the intensity and the 2D centroid of the spot found there.
	Then if it can it computes the hkl of the spot and does some simple
	intensity statistics on the resulting spot */
	int		htdo, htup, ldih, lditype, ldiw, result, wido, wiup;
	int		xc, xmn, xmx, yc, ymn, ymx, dibou[4], imcoo[4];
	double		istat[4];
	DISP_ATTRS	*lattr, *magtr;

	lattr = xxg->gp_at;	xc = lattr->di_disx;	yc = lattr->di_disy;
	magtr = xxg->gp_bl;
	ldiw = lattr->di_wid;	ldih = lattr->di_ht;
	magtr->di_distype = lditype = lattr->di_distype;
		/* figure out the boundaries of the 21x21 */
	xmn = xc - 10;	if (xmn < 0) xmn = 0;
	xmx = xc + 10;	if (xmx >= ldiw) xmx = ldiw - 1;
	ymn = yc - 10;	if (ymn < 0) ymn = 0;
	ymx = yc + 10;	if (ymx >= ldih) ymx = ldih-1;
	wiup = lattr->di_wiup;	wido = lattr->di_widown;
	htup = lattr->di_htup;	htdo = lattr->di_htdown;
	magtr->di_disx = lattr->di_disx = xc;
	magtr->di_currx = lattr->di_currx = xc * wido / wiup;
	magtr->di_disy = lattr->di_disy = yc;
	magtr->di_curry = lattr->di_curry = yc * htdo / htup;
	magtr->di_xylim[0] = lattr->di_xylim[0] = (xmn * wido) / wiup;
	magtr->di_xylim[1] = lattr->di_xylim[1] = (ymn * htdo) / htup;
	magtr->di_xylim[2] = lattr->di_xylim[2] = (xmx * wido) / wiup;
	magtr->di_xylim[3] = lattr->di_xylim[3] = (ymx * htdo) / htup;
	xxg->gp_col = xxg->gp_dbcol[lattr->di_nranges];
	dibou[0] = xmn; dibou[1] = ymn; dibou[2] = xmx;	dibou[3] = ymx;
	create_blowup(xxg, dibou);	/* make 8-fold magnified picture */
	/* superimpose the appropriate overlays on the (X,Y) displays */
	if ((lditype >= 0) && (lditype != NUREFLBS - 2) &&
		(lditype != NUREFLBS - 1)) do_superpositions(xxg, 1);
	/* don't bother with the hkl and statistics if we're looking
	 at a non-image picture */
	if (lditype) return;
	/* determine the actual detector coordinates associated with
	 these screen coordinates and rotate into the proper orientation */
	imcoo[0] = xc / wiup;	imcoo[0] *= wido;
	imcoo[1] = yc / htup;	imcoo[1] *= htdo;
	disrot((int)(lattr->di_roco), 0, (int)(lattr->di_imwid), imcoo);
	lattr->di_currx = imcoo[2];
	lattr->di_curry = imcoo[3];
	/* compute the hkl of the pixel we're centered on */
	if (dget_hkl(lattr, istat)) show_hkl(xxg, istat);
	/* do some reflection statistics on the result */
	if (0 == (result = reflstats(lattr, istat))) show_stats(xxg, istat);
	else fprintf(stderr, "Failed to do statistics near %4d %4d; code=%d\n",
		imcoo[2], (int)(lattr->di_imht - 1 - imcoo[3]), result);
}

static char myform(int ispo, double extre)
{	/* this returns a value between '0' and '4', depending
	 on how big the extremum is and whether ispo is true */
	int	iv;
	iv = (extre < 10. ? 4 : (extre < 100. ? 3 :
		(extre < 1000. ? 2 : (extre < 10000. ? 1 : 0))));
	if ((!ispo) && (iv > 0)) iv--;
	return '0' + iv;
}

static void show_actual(XG_X11PROPS *xxg, short *cuh, double *ive)
{ /*	This displays the contents of cuh and ive meaningfully */
	char	texb[240];

	zeromem((AVOID *)texb, 240);
	sprintf(texb, "hkl,D=%4d %4d %4d %6.1f",
	 cuh[0], cuh[1], cuh[2], ive[5]);
	replace_string(xxg->gp_tport[1], texb, xxg->gp_chpol);
	sprintf(texb,
 "I=%7.0f+/-%6.1f @\n%6.1f,%6.1f,%6.1f\nError(X,Y,omega) =\n%5.1f,%5.1f,%5.1f",
		ive[0], ive[1], ive[2], ive[3], ive[4],
	 	ive[6], ive[7], ive[8]);
	replace_string(xxg->gp_tport[0], texb, xxg->gp_chpol);
}

static void showrefmr(XG_X11PROPS *xxg, short *cuh, double *ive)
{ /*	This displays the contents of multiref cuh and ive meaningfully */
	char	texb[240], *cp;
	double	tstl;

	zeromem((AVOID *)texb, 240);	tstl = fabs(ive[7]);
	if (tstl > DFUZZ) sprintf(texb, "hkl,D=%4d %4d %4d %6.1f",
	 cuh[0], cuh[1], cuh[2], 0.5 / tstl);
	replace_string(xxg->gp_tport[1], texb, xxg->gp_chpol);
	sprintf(texb, "I=%7.0lf+/-%6.1lf @\n%6.1lf,%6.1lf,%6.1lf",
		ive[0], ive[1], ive[2], ive[3], ive[4]);
	if (ive[5] > -1.e5)
	  {
		cp = texb + strlen(texb);
		sprintf(cp, "\nIsc, <I> = %8.2lf,%8.2lf", ive[5], ive[6]);
	  }
	if (ive[7] < 0.) /* tell us about deleted observations */
	  {
		cp = texb + strlen(texb);
		sprintf(cp, "\nFlagged for deletion");
	  }
	replace_string(xxg->gp_tport[0], texb, xxg->gp_chpol);
}

void display_pixel(Widget widget)
{ /*	This slaps up the actual count value observed in the pixel at which
	the cursor is currently pointing.  If the mousebutton is pressed,
	we also display a 21x21 region around the current position, measure
	the hkl of the current point, and do some statistics on the count
	values in the neighborhood of the spot. */
	char		*cp, xytext[40], htext[40];
	int		detx, dety, dih, diw, duwidth, duheight, fra;
	int		htdo, htup, ishkld, ispos, j, lditype, ldvn;
	int		livn, lnr, lwi, maxfm, minfm, origgx, origgy;
	int		roco, rootx, rooty, tcoi, wido, wiup, xys, imcoo[4];
	short		*sp, curhkl[3];
	xglong		*ip;
	unsigned int	keys_buttons;
	double		dv, extremum, highli, iv, lowli, other, ivec[9];
	Window		childwin, rwin, windo;
	XG_X11PROPS	*xxp, *xxq;
	DISP_ATTRS	*dispatt, *maina;
	FRAMEINDEX	*lfmi;
	MAIL		*lhd;
	PAU		dpu;
	static int	gwin_x, gwin_y, prev_x, prev_y;
	static char	sfmt[] = "%7.1f<%s<%7.1f";
	static char	scfmt[] = "%s=%7.2f; %s=%7.2f";
	static DUMPRUN	duru;

	if (NULL == (xxp = managex11p(3))) return;
	dispatt = (widget == xxp->gp_imwid) ? xxp->gp_at : xxp->gp_bl;
	maina = xxp->gp_at;
	livn = xxp->gp_ivn;		ldvn = xxp->gp_dvn;
	roco = dispatt->di_roco;	lwi = dispatt->di_imwid;
	if (dispatt->di_wintype)
	  {
		diw = dispatt->di_magw;	dih = dispatt->di_magh;
		windo = xxp->gp_maw;
	  }
	else {
		diw = dispatt->di_wid;	dih = dispatt->di_ht;
		windo = xxp->gp_wi;
	     }
	wiup = dispatt->di_wiup;	wido = dispatt->di_widown;
	htup = dispatt->di_htup;	htdo = dispatt->di_htdown;
	lditype = dispatt->di_distype;	lnr = dispatt->di_nranges;
	ishkld = (lditype == NUREFLBS - 2) || (lditype == NUREFLBS - 1);
		/* if the pointer is in the magnified window rather than
		 the unmag window, we need to convert the (x,y) value */
	cp = (char *)0;	sp = (short *)0;	ip = (xglong *)0;
	if (dispatt->di_wintype)
	  {
		if (0 == XQueryPointer(xxp->gp_di, windo, &rwin, &childwin,
		 &rootx, &rooty, &gwin_x, &gwin_y, &keys_buttons)) return;
		if ((gwin_x < 0) || (gwin_x >= dispatt->di_magw) ||
		  (gwin_y < 0) || (gwin_y >= dispatt->di_magh)) return;
		origgx = gwin_x;	origgy = gwin_y;
		if ((ishkld) || (lditype < 0))
		  {
			gwin_x = imcoo[0] = gwin_x / 8 + maina->di_disx - 10;
			gwin_y = imcoo[1] = gwin_y / 8 + maina->di_disy - 10;
		  }
		else {
			imcoo[0] = (gwin_x + 8 * maina->di_disx - 80) *
			wido / (8 * wiup);
			imcoo[1] = (gwin_y + 8 * maina->di_disy - 80) *
			 htdo / (8 * htup);
			disrot(roco, 0, lwi, imcoo);
			gwin_x = imcoo[2] * wiup / wido;
			gwin_y = imcoo[3] * htup / htdo;
		     }
		switch (dispatt->di_nbperp)
		  {
			case	1:	cp = maina->di_magm;	break;
			case	2:	dpu.ucpt = (u_char *)(maina->di_magm);
					sp = (short *)(dpu.uspt);	break;
			default:	dpu.ucpt = (u_char *)(maina->di_magm);
					ip = (xglong *)(dpu.vpt);	break;
		  }
	  }
	else /* for the main window, we just use the value we see */
	  {
		if (0 == XQueryPointer(xxp->gp_di, windo, &rwin,&childwin,
		 &rootx, &rooty, &gwin_x, &gwin_y, &keys_buttons)) return;
		/* convert from screen coordinates to detector coordinates
		 and rotate into the proper orientation */
		origgx = gwin_x;	origgy = gwin_y;
		imcoo[0] = gwin_x * wido / wiup;
		imcoo[1] = gwin_y * htdo / htup;
		disrot(roco, 0, lwi, imcoo);
		switch (dispatt->di_nbperp)
		  {	
			case	1:	cp = maina->di_rotm;	break;
			case	2:	dpu.ucpt = (u_char *)(maina->di_rotm);
					sp = (short *)(dpu.uspt);	break;
			default:	dpu.ucpt = (u_char *)(maina->di_rotm);
					ip = (xglong *)(dpu.vpt);	break;
		  }
	  }
	dispatt->di_currx = maina->di_currx = detx = imcoo[2];
	dispatt->di_curry = maina->di_curry = dety = imcoo[3];
	if ((detx < 0) || (detx >= dispatt->di_imwid) ||
	  (dety < 0) || (dety >= dispatt->di_imht)) return;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	if ((detx != prev_x) || (dety != prev_y))
	  {
		if (lditype < 0)
		  {
			iv = dispatt->di_limi[lnr + 2] +
				gwin_x * dispatt->di_limi[lnr + 4];
			scfmt[6] = myform((iv >= 0.), fabs(iv));
			dv = dispatt->di_limi[lnr + 5] +
			 (dih - 1 - gwin_y) * dispatt->di_limi[lnr + 7];
			scfmt[16] = myform((dv >= 0.), fabs(dv));
			sprintf(xytext, scfmt, varlabel(livn, 0),
			 iv, varlabel(ldvn, 1), dv);
		  } /* end of "if (lditype < 0)" clause */
		else if (lditype)
		  {
			if (ishkld)
			  {
				gwin_x -= gwin_x % 4;	gwin_y -= gwin_y % 4;
				find_indices(xxp, gwin_x, gwin_y, curhkl);
				sprintf(htext, "h, k, l = %3d,%3d,%3d",
				 curhkl[0], curhkl[1], curhkl[2]);
			  }
			else sprintf(xytext, "[%4d,%4d]:", detx, dety);
			switch (dispatt->di_nbperp)
			  {
				case	1:
				 cp += diw * origgy + origgx;
				 for (j = 0; j < lnr; j++)
				    {
					tcoi = xxp->gp_coi[j];
					if (*cp ==
					 (char)(xxp->gp_dbcol[tcoi])) break;
				    }
				 break;
				case	2:
				 sp += diw * origgy + origgx;
				 for (j = 0; j < lnr; j++)
				    {
					tcoi = xxp->gp_coi[j];
					if (*sp ==
					 (short)(xxp->gp_dbcol[tcoi])) break;
				    }
				 break;
				case	4:	default:
				 ip += diw * origgy + origgx;
				 for (j = 0; j < lnr; j++)
				    {
					tcoi = xxp->gp_coi[j];
					if (*ip ==
					 (int)(xxp->gp_dbcol[tcoi])) break;
				    }
				 break;
			  }
			if (j)
			  {
				lowli = dispatt->di_limi[j - 1];
				highli = dispatt->di_limi[j];
				extremum = fabs(lowli);	ispos = lowli >= 0.;
				other = fabs(highli);
				if (other > extremum)
				  {
					extremum = other; ispos = highli >= 0.;
				  }
				sfmt[3] = sfmt[12] = myform(ispos, extremum);
				xys = (ishkld) ? 0 : 12;
				sprintf(&(xytext[xys]), sfmt, lowli,
				 dispatt->di_distype < 0 ? gdisnam(2):
				 gdisnam(lditype), highli);
			  }
			else if (ishkld) strcpy(xytext, "no points");
			else strcat(xytext, "no points");
		  } /* end of "else if (dispatt->di_distype" clause */
		else sprintf(xytext, "Cursor: I[%4d,%4d]=%5d",
			 detx, dety, imvalue(lhd, detx, dety));
		replace_string(xxp->gp_tport[2], xytext, xxp->gp_chpol);
		if (ishkld) replace_string(xxp->gp_tport[1], htext,
			xxp->gp_chpol);
		prev_x = detx;	prev_y = dety;
	  }
	if (keys_buttons & Button1Mask)
	  {
		if (dispatt->di_wintype)
		  {
			imcoo[0] = (origgx + 8 * maina->di_disx - 80) *
			 wido / (8 * wiup);
			imcoo[1] = (origgy + 8 * maina->di_disy - 80) *
			 htdo / (8 * htup);
			disrot(roco, 0, lwi, imcoo);
			detx = imcoo[2];	dety = imcoo[3];
			maina->di_disx = dispatt->di_disx = detx * wiup / wido;
			maina->di_disy = dispatt->di_disy = dety * htup / htdo;
		  }
		else {
			dispatt->di_disx = gwin_x; dispatt->di_disy = gwin_y;
		     }
		if (ishkld)
		  {
			gwin_x -= gwin_x % 4;	gwin_y -= gwin_y % 4;
			find_indices(xxp, gwin_x, gwin_y, curhkl);
			if (0 < exact_inten(curhkl, ivec))
				show_actual(xxp, curhkl, ivec);
		  }
		else if (lditype < NUREFLBS)
		  {
			if (0 < exact_xy(dispatt, curhkl, ivec))
				show_actual(xxp, curhkl, ivec);
			else {
				replace_string(xxp->gp_tport[0], " ",
					xxp->gp_chpol);
				replace_string(xxp->gp_tport[1], " ",
					xxp->gp_chpol);
			     }
		  }
		else if (lditype < (NUREFLBS + 3))
		  {
			if (0 < exact_mrxy(dispatt, curhkl, ivec))
				showrefmr(xxp, curhkl, ivec);
			else {
				replace_string(xxp->gp_tport[0], " ",
					xxp->gp_chpol);
				replace_string(xxp->gp_tport[1], " ",
					xxp->gp_chpol);
			     }
		  }
		else printf(
 "No special display associated with mouse-click when display type = %d\n",
		 lditype);
		if (NULL == (xxq = managex11p(3))) return;
		show_region(xxq);
		if (keys_buttons != Button1Mask)
		 printf("actual keys_buttons value = 0x%x\n", keys_buttons);
	  }
	if (keys_buttons & Button2Mask)
	  { /* user has asked for an (X,Y) dump:
		 first construct the structure that we'll feed to `dumpiamge' */
		duru.dui_fpd = xgopen("udump", "a");
		duru.dui_wid = BIGSPOTWID;	duru.dui_ht = BIGSPOTHT;
		duru.dui_heads = dispatt->di_head;
		duru.dui_nmbox = dispatt->di_nmbox;
		duru.dui_xce = detx;	duru.dui_yce = dety;
		duru.dui_fmi = dispatt->di_fmind; /* need frame index */
		duru.dui_imname = dispatt->di_imaname;
		dumpimage(&duru, 1);	/* dump the counts in the pixels */
		if (NULL != duru.dui_fpd) (void)fclose(duru.dui_fpd);
		duru.dui_fpd = NULL; /* close and null the output file */
		if (keys_buttons != Button2Mask)
		 printf("actual keys_buttons value = 0x%x\n", keys_buttons);
	  }
	if (keys_buttons & Button3Mask)
	  {
		minfm = dispatt->di_fmno - 3;	maxfm = dispatt->di_fmno + 3;
		duwidth = BIGSPOTWID;	duheight = BIGSPOTHT;
		lfmi = dispatt->di_fmind = ftables(3);
		if (NULL == lfmi) return;
		if (lfmi->fi_firstfm >= minfm) minfm = lfmi->fi_firstfm;
		if (lfmi->fi_lastfm  <= maxfm) maxfm = lfmi->fi_lastfm;
		zeromem((void *)(maina->di_magm), (int)(dispatt->di_magw *
			dispatt->di_magh * dispatt->di_nbperp));
		for (fra = minfm; fra <= maxfm; fra++)
		   {
			if (EOK != (j = threedblow(xxp, fra, 21, 21)))
			  {
				printf("threedblow returned %d for image %d\n",
					j, fra);
				break;
			  }
		   }
		make_blowupwin(xxp);	/* create & display blowup */
		if (keys_buttons != Button3Mask) printf(
			"actual keys_buttons value = 0x%x\n", keys_buttons);
	  }
	if (keys_buttons & Button4Mask)
	  {
		printf(" Button4Mask activated");
		if (keys_buttons != Button4Mask) printf(
			"; actual keys_buttons value = 0x%x\n", keys_buttons);
		else printf("\n");
	  }
	if (keys_buttons & Button5Mask)
	  {
		printf(" Button5Mask activated");
		if (keys_buttons != Button5Mask) printf(
			"; actual keys_buttons value = 0x%x\n", keys_buttons);
		else printf("\n");
	  }
	if (keys_buttons & AnyModifier)
		printf("Modifier: keys_buttons = 0%o\n", keys_buttons);
}

/* end of dispxmag.c */
