/* pdispmain.c -- main routines for `pdisplay'

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Jun 2006: displays.h included
	22 Jan 2006: indexname call updated
	23 May 2005: flag xgjournal() --help and --version calls correctly
	11 May 2005: check return status from xgjournal
	 6 Sep 2004: cosmetic changes
	28 Aug 2004: routine pdispargs(), used only here, is declared as static
	27 Feb 2004: call to mallop() for lstr in pdispargs is correctly labeled
	21 Jan 2004: license message inserted
	 6 Jan 2004: globals replaced with elements of DISP_ATTRS structure;
		update calls to make_ftab appropriately;
		handle array of MAIL structures carefully;
		call to graphics() simplified by redesign of DISP_ATTRS struct
	13 Dec 2003: handle return from xgnowmain() properly in pdispargs()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	25 Aug 2003: eliminate global variables;
		postpone call to grab_image() until graphics() call
	16 Aug 2003: reduce reliance on global variables
		at the cost of a few additional arguments to functions
	 1 May 2003: run the equivalent of `xgnow' as we start up
	 5 Jan 2003: reorganized source-file header
	10 Nov 2002: remove redundant extern declaration of darkflag
	25 Mar 2002: removed MSI attachments
	27 Oct 1994: renamed global variable "display"
		to static global "mydispla"
	 5 May 1994: source-file header, POLYDET->XGEN,
		"xgulong" construction used
	16 Mar 1994: minor bugfixes
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: displays detector images and other objects
	directly on an X window display governed under Motif.
   Cross references: ./disp*.c, ../displays/disp*.c
*************************************************************************/
#include	"displays.h"
extern int dispkill(int irv);
extern int xgnowmain(int ac, char *av[]);
extern int dispcmain(int argc, char *argv[]);
extern int graphics(int argc, char **argv, DISP_ATTRS *mainatt);

static int pdispargs(int argc, char *argv[], DISP_ATTRS **maindp)
{ /*	argument handler for `pdisplay'.
	Syntax:		pdisplay [framenum] [filename]	*/
	int			iarg, lac, retv;
	DISP_ATTRS		*lattp;
	static int		fmnumb, darkflag;
	static char		*darkname, *imagname, **dcp, *lstr, **lqav;
	static FRAMEINDEX	*lfmi;
	static int		idarkfrac = 6400;
	static char		lstrmsg[] = "lstr in pdispargs";

	if (EHELP == xgjournal(argc, argv)) return EHELP; /* journal effort */
	darkflag = 0;	fmnumb = -1;
	iarg = argc - 1;	dcp = argv + 1;
	if (NULL == (*maindp = (DISP_ATTRS *)callop(2, sizeof (DISP_ATTRS),
	 "*maindp in pdisplay"))) return dispkill(EALLOC);
	lattp = *maindp;	lattp->di_nmbox = PFRAMES + 1;
	if (NULL == (lattp->di_head = (MAIL *)callop((size_t)(lattp->di_nmbox),
		sizeof (MAIL), "lattp->di_head in pdispargs"))) return EALLOC;
	while ((iarg) && (*dcp) && (**dcp == '-')) /*flag-style argument */
	  {
		if ((*((*dcp) + 1) == 's') || (*((*dcp) + 1) == 'S'))
		  {
			darkflag = DARK_CURRENT;
			if ('\0' != *((*dcp) + 2))
				idarkfrac = 8000. * atof((*dcp) + 2);
		  }
		else fprintf(stdout, "Flag argument %s ignored\n", *dcp);
		iarg--;	dcp++;
	  }
	if (idarkfrac > 16000) idarkfrac = 16000;
	if (darkflag == DARK_CURRENT)
	  {
		darkflag |= (idarkfrac << 16); darkname = trname("DARKIMAGE");
		printf(
 "%7.4lf of the dark-current in %s will be subtracted from each image\n",
			((double)idarkfrac) / 8000., darkname);
	  }
	if (iarg && !allnumeric(*dcp)) imagname = trname(*dcp);
	if (NULL == (lfmi = ftables(3)))
	  { /* If the user has specified a filename, we try to deal with it.
		But in that instance we warn the user that there isn't
		necessarily a frame index available */
		if (NULL != imagname)
			fprintf(stdout, "Warning: no frame index available\n");
		/* on the other hand, if a filename hasn't been specified,
		 we do the equivalent of invoking `xgnow'. Then we endeavor
		 to open the image index and display the first frame */
		else {
			lstr = (char *)mallop(8, lstrmsg);
			zeromem((AVOID *)lstr, 8);
			(void)xgprogname(NULL, 2);
			strcpy(lstr, "xgnow");
			if (NULL == (lqav = str2acav(lstr, &lac)))
				return dispkill(EALLOC);
			memzap((AVOID **)&lstr, lstrmsg);
			/* get environment variables and check for success */
			if (EOK != (retv = xgnowmain(lac, lqav)))
			  {
				fprintf(stderr, "Failure %d from xgnow\n",
					retv);	return retv;
			  }
			if (NULL == (lfmi = ftables(3))) /*can we make it?*/
			  {
				fprintf(stderr, "Can't find images\n");
				return EOPENERR;
			  }
			fmnumb = iarg ? atoi(*dcp) : lfmi->fi_firstfm;
			if (fmnumb < lfmi->fi_firstfm)
				fmnumb = lfmi->fi_firstfm;
			if (fmnumb > lfmi->fi_lastfm) fmnumb = lfmi->fi_lastfm;
			imagname = indexname(fmnumb);
		     }
	  }
	else {
		if (NULL == imagname)
		  {
			fmnumb = iarg ? atoi(*dcp) : lfmi->fi_firstfm;
			if (fmnumb < lfmi->fi_firstfm)
				fmnumb = lfmi->fi_firstfm;
			if (fmnumb > lfmi->fi_lastfm) fmnumb = lfmi->fi_lastfm;
			imagname = indexname(fmnumb);
		  }
		else if (-1 < (fmnumb = match_inode(&lfmi, imagname)))
			printf("Frame number set to %d\n",fmnumb);
	     }
	(*maindp)->di_fmind = lfmi; /* we remember this index in all cases */
	if (qerror((AVOID *)imagname, "find requested frame file"))
		return EOPENERR;
	lattp->di_fmno = fmnumb;	lattp->di_darkfla = darkflag;
	lattp->di_imaname = imagname;	return EOK;
}

int dispcmain(int argc, char *argv[])
{ /* The mainline for pdisplay. Not a lot happens here */
	int		result;
	DISP_ATTRS	*maina;

	if (EOK != (result = pdispargs(argc, argv, &maina)))
		return (EHELP == result) ? result : dispkill(result);
	return dispkill(graphics(argc, argv, maina));
}

/* end of pdispmain.c */
