/* pmmain.c -- top-level routine for XGEN's motif-based user interface

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	 8 Sep 2006: SGI-specific stuff in graphics() now depends on #ifdef sgi
	 3 Jan 2006: set_logical can return >= 1.
	 5 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	12 Oct 2004: blank out the lodisp structure before using it in pmmain()
	 6 Sep 2004: cosmetic changes
	16 Aug 2004: eliminate setjmp stuff (I hope that will work!)
	29 Feb 2004: getproject() takes second argument, which here is zero.
	26 Feb 2004: put the routines that do most of the work in pmutils.c,
		and the next-to-top-level routines into pmmain.c
 Changes to pmotif.c and its predecessors prior to that:
	24 Feb 2004: incorporate XG_X11PROPS constructs
	24 Jan 2004: numerous static globals converted to local variables
	21 Jan 2004: license message inserted
	 2 Jan 2004: globals replaced by elements of FRAMEINDEX and XGNAME
		structure pointers accessed as function args
	17 Oct 2003: merged pmmain.c, pmprogs.c, pmutil.c into one file.
	25 Aug 2003: #include xgmdefs.h rather than pmdefs.h;
		xgmdefs.h includes some items usable in pdisplay as well
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 1 Dec 1994: globals formerly found in xgenglob.c get initialized here
	14 Jan 1994: initial X-GEN version

   Purpose: mainline (almost) for setting up and using a Motif interface to
	the functionalities of X-GEN.
   Cross references: xgmdefs.h
*************************************************************************/
#include	"xgmdefs.h"
/* #include	<setjmp.h> */

int pmmain(int argc, char *argv[])
{ /* pmotif's mainline */
	int			n;
	char			*cp;
	XG_X11PROPS		*pm11p;
	DISP_ATTRS		lodisp, *dispat;
	static XGNAME		*namb;
	static Widget		toplevel;
	static XmFontList	flist1;
	static XFontStruct	*font1, *font2;
	static XtAppContext 	app_context;
	static XmFontListEntry	flentry1, flentry2;
	static Arg		ag[1];
	static char		*nfont1 ="*New Century Schoolbook-Bold-R*120*";
	static char		*nfont2 ="*New Century Schoolbook-Bold-I*140*";

	/* if ((n = setjmp(pmjbuf))) return (n == 1) ? EOK : n; */
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* initialize a few variables. We define memory for an XG_X11PROPS
	 structure first */
	zeromem((void *)&lodisp, sizeof (DISP_ATTRS));
	if (NULL == (pm11p = get_pm11())) return EALLOC;
	pm11p->gp_at = &lodisp;	dispat = pm11p->gp_at; dispat->di_fmind = NULL;
	pm11p->gp_bl = NULL;	/* flag showing which version we're using */
	/* define the initial project name and some environment variables */
	if (NULL == (namb = getproject(((argc < 2) ? NULL : argv[1]), 0)))
		return EALLOC;
	dispat->di_fmind = ftables(7); /* get 1st & last frames in index*/
	/* now we start dealing with X11 objects */
	XtToolkitInitialize();	/* get the X tools initialized */
	app_context = XtCreateApplicationContext();
/* before attempting to open the display, we look to see if the environment
 variable "DISPLAY" is defined.  If not, we try to set it to :0 */
	if ((NULL == (cp = getenv("DISPLAY"))) || ('\0' == *cp))
	  {
		if (1 > set_logical("DISPLAY", ":0"))
		  {
			fprintf(stdout,
		"Unable to define the environment variable DISPLAY as `:0'\n");
			return EBAD;
		  }
	  }
	if (NULL == (pm11p->gp_di = XtOpenDisplay(app_context, NULL, argv[0],
#ifdef sgi
		"pmotifclass", NULL, 0, (int *)&argc, argv)))
#else
		"pmotifclass", NULL, 0, (XGCAR *)&argc, argv)))
#endif
          {
		fprintf(stdout, "%s:  Can't open display\n", argv[0]);
		return EBAD;
	  }
	/*	Define the fonts used in the internal character sets. */
	pm11p->gp_chpol = "pol1";	pm11p->gp_chpo2 = "pol2";
	pm11p->gp_dflis = flist1 = NULL;
	if ((font1 = XLoadQueryFont(pm11p->gp_di, nfont1)))
	  {
		flentry1 = XmFontListEntryCreate(pm11p->gp_chpol,
			XmFONT_IS_FONT, (XtPointer)font1);
		flist1 = XmFontListAppendEntry(NULL, flentry1);
	  }
	if ((font2 = XLoadQueryFont(pm11p->gp_di, nfont2)))
	  {
		flentry2 = XmFontListEntryCreate(pm11p->gp_chpo2,
			XmFONT_IS_FONT, (XtPointer)font2);
		pm11p->gp_dflis = XmFontListAppendEntry(flist1, flentry2);
	  }
	XtSetArg(ag[0], XmNallowShellResize, True);	n = 1;
	toplevel = XtAppCreateShell(argv[0], NULL, applicationShellWidgetClass,
                              pm11p->gp_di, ag, n);
	/* having created the toplevel shell, we populate it with
	 the functional widgets that the user can see */
	if (EOK != pmpopulate(toplevel)) return EALLOC;
	XtAppMainLoop(app_context); /* loop and process user-created events */
	/*UNREACHABLE*/
	return (0);
}

/* end of pmmain.c */
