/* pmmenus.c -- utilities for XGEN's motif-based user interface

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	12 Jun 2006: move update_cmd_line around and localize it (non-static)
	20 Jan 2006: replace gp_dbcol with gp_progcol;
	 destroy many widgets in killprogmenus; correct indexname calls;
	 conditional creation of widgets, especially in update_cmd_line
	 7 Jan 2006: function prototypes moved to xgmdefs.h
	17 Oct 2005: squash all static globals into structure type PMMWORK
	 5 Oct 2005: housekeeping
	16 Aug 2005: MAXDEFAULTS, etc. moved from ../include/xgen.h
	28 Jun 2005: compare *widp to NULL, not to 0, for portability
	17 Jun 2005: pmpopulate(): basic_names() takes fromcur argument
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	20 Oct 2004: 33 programs supported
	15 Oct 2004: 31 programs supported; "dial down the center" set up
	12 Oct 2004: locstr is XmString in all instances; various strings
	 now derived from namebuf tank; various argument checks inserted;
	 in stick_string(), make sure the labels don't have spaces in them
	27 Aug 2004: routines used only here are declared as static; cosmetics
	16 Aug 2004: created from the messier parts of pmutils.c
 Changes to pmutils.c prior to that date:
	 3 Apr 2004: numerous typecasts because a Strings isn't an XmString
	29 Mar 2004: killprogmenus() and whatshown() written;
		called from prevmenu() and forktoIt()
	 7 Mar 2004: cosmetic changes
	 6 Mar 2004: delete memory for String variables after using them
	 5 Mar 2004: move the processing button from the menubar to the side;
		shrink height of entries a little so they fit better;
		move program-selection menu to the right of the display and
		lower it when dealing with a specific app;
		handle asynchronous calls more carefully
	 4 Mar 2004: replace gspline() calls with cleanspline() calls;
		restore spline in exitbutt_activate();
		reposition switch & param menus
	29 Feb 2004: arg to gspline() in forktoIt needs to be an address
	26 Feb 2004: put the routines that do most of the work in pmmenus.c,
		and the next-to-top-level routines into pmmain.c
 Changes to pmotif.c prior to that:
	24 Feb 2004: incorporate XG_X11PROPS constructs
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	31 Jan 2004: include "spaceg" as one of the programs we can run
	24 Jan 2004: numerous static globals converted to local variables
	21 Jan 2004: license message inserted
	 5 Jan 2004: handle global-variable memzaps via setkillglobs()
	 2 Jan 2004: globals replaced by elements of FRAMEINDEX and XGNAME
		structure pointers accessed as function args
	13 Dec 2003: moderate rewrite: toplevel menu has more options,
		we rely on pmall call more, button-positions semi-automated
	 7 Dec 2003: move defaults[] and labels[] here as static globals
	25 Nov 2003: eliminate many global variables by using calls to pmall()
	17 Oct 2003: merged pmmain.c, pmprogs.c, pmutil.c into one file.
 Changes made to pmmain.c, pmprogs.c, pmutil.c prior to that:
	15 Oct 2003: move exitbutt_activate() from pmprogs.c to pmmain.c
		so that we don't have to keep the jmpbuf stuff there
	21 Sep 2003: XtOpenDisplay call has variant for IRIX
	20 Sep 2003: in button_color(), set an explicit default for ftype
	19 Sep 2003: in redo_maincolors(), fna0 explicitly set to NULL
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() calls with xgaccess() calls;
		that allows for gunzipping as we pass through
	25 Aug 2003: #include xgmdefs.h rather than pmdefs.h;
		xgmdefs.h includes some items usable in pdisplay as well
	24 Aug 2003: rename variable now called `bmorph' to reduce confusion
	22 Aug 2003: modify use of glistname() and the listname variable
		according to the changes in xgenviron.c;
		stick the setup-menu labels in downstream of listname,
		rather than downstream of scrtname.
	17 Aug 2003: major reorganization: pmnames.c now part of pmprogs.c
		and we've reduced our dependence on global variables
		at the cost of a few additional arguments to functions;
	13 Aug 2003: fped is now static global, not global global
	10 Jul 2003: search for X-GEN documentation by call to get_xgdoc()
	 9 Jul 2003: determine colors of program buttons by calls to ready()
	 5 Jan 2003: reorganized source-file header;
		setkillglob(1) just before forking so that
		the wrong memory doesn't get released
	15 Nov 2002: extern declarations made here rather than in pmext.h
		line[] replaced with nabuf[] to prevent name overlap;
		nabuf[] memory grabbed with mnbuf() call
		filewids[], progbutwids[], etc. and position variables
		(progdim[0], progdim[1], etc.) moved from pmglob.c;
	 1 Nov 2002: rename construct_command to makecmd
	15 Apr 2002: (?) leave the colors fixed for all buttons except 1-4.
	15 Apr 2002: (?) reformat and redefine frame-range menu
		to deal more effectively with arbitrary frame-name templates
		(?) in forktoIt(), replace call to system()
		with a proper fork & exec or even a series of calls to routines
	25 Mar 2002: removed MSI attachments
	 1 Dec 1994: globals formerly found in xgenglob.c get initialized here
	16 Nov 1994: more_logicals now takes an argument
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few variables moved around
	14 Jan 1994: initial X-GEN version

   Purpose: setting up and using a Motif interface to the
	functionalities of X-GEN.
   Cross references: xgmdefs.h
*************************************************************************/
#include	"xgmdefs.h"
#ifdef VMS
#define		FILEFORMAT1	"%s%s.%s"	/* VMS format */
#define		READWRITE	4	/* don't try to require write access */
#define		CMDSUFFIX	"COM"	/* standard suffix for command file */
#else
#define		FILEFORMAT1	"%s/%s.%s"	/* Unix format */
#define		READWRITE	6	/* do require write access */
#define		CMDSUFFIX	"cmd"	/* standard suffix for command file */
#endif

#define	CALL2(x,y)	XtAddCallback(*widp,XmNactivateCallback,(XtCallbackProc)x,(XtPointer)y)

#define		NPPROGS		33
#define		NPFILES		9
#define 	CANMOVELEFT	360	/* ok to put command below booleans */
#define		MAXPFILES	18	
#define		MAXPBOOLS	24
#define		MAXPSWITCH	16
#define		MAXPPARAM	16
#define		MAXCOLS 	36	
#define		MAXLABELS	50	/* max# of labels on user interfaces*/
#define		MAXDEFAULTS	50	/* max# of default vals in GUIs */
#define		LABELLENGTH	45	/* max length of labels in GUIs */
#define		VALUELENGTH	30	/* max length of values in GUIs */

 typedef struct {
		int		pm_entryy, pm_labely, pm_nbool, pm_nfb, pm_nfbs;
		int		pm_nfiles, pm_nparam, pm_nswitch, pm_poscoo[8];
		char		*pm_strs;
		char		*pm_ubu, *pm_labels[MAXLABELS], *pm_endtitle;
		char		*pm_defaults[MAXDEFAULTS], *pm_stlb[10];
		XGNAME		*pm_namb;
		XG_X11PROPS	*pm_p11p;
		Widget		pm_boolform, pm_commandwid;
		Widget		pm_fileform, pm_ourmain, pm_paramform;
		Widget		pm_proglabel, pm_setxbb;
		Widget		pm_filew[MAXPFILES], pm_boolw[MAXPBOOLS];
		Widget		pm_labelw[MAXLABELS], pm_progbutw[NPPROGS];
		Arg		pm_ag[12];
		} PMMWORK;

extern void update_cmd_line(void);

static PMMWORK *managepmm(int opno)
{ /*	Manage the static global "pmmp" */
	size_t		totchars;
	static PMMWORK	*pmmp;
	static char	pmpmsg[] = "pmmp in managepmm";
	static char	strmsg[] = "pmmp->pmm_strs in managepmm";

	if (!opno)
	  {
		if (NULL != pmmp)
		  {
			sharename(0);	/* get rid of pm_namb */
			(void)managex11p(0);	/* get rid of pmmp->p11p */
			memzap((AVOID **)&(pmmp->pm_strs), strmsg);
			memzap((AVOID **)&pmmp, pmpmsg);
		  }
	  }
	if (opno == 1) pmmp = managepmm(0);
	if (opno == 7) return pmmp;	/* don't try to create stuff */
	if (NULL == pmmp) /* create pmmp if needed */
	  {
		if (NULL == (pmmp = (PMMWORK *)callop(1, sizeof (PMMWORK),
		 pmpmsg))) return NULL;
	  }
	if ((NULL == pmmp->pm_p11p) && (NULL ==
	 (pmmp->pm_p11p = managex11p(073)))) return managepmm(0);
	if (NULL == (pmmp->pm_namb = sharename(7))) return managepmm(0);
	if (NULL == pmmp->pm_strs)
	  {
		totchars = (MAXLABELS + 4) * (LABELLENGTH + 4) +
		 (MAXDEFAULTS + 4) * (VALUELENGTH + 4);
		if (NULL == (pmmp->pm_strs = (char *)
		 mallop(totchars, strmsg))) return managepmm(0);
		zeromem((AVOID *)(pmmp->pm_strs), totchars);
	  }
	return pmmp;
}

static void stexit(void)	/* returns from setup menu */
{
	PMMWORK	*pmmp;

	if (NULL == (pmmp = managepmm(7))) return;
	XtUnmanageChild(pmmp->pm_setxbb);	redo_maincolors();
}

static void setupbutton(int position)
{ /*	This creates one of the buttons derived from the `setup' menu,
	provided that it isn't already there.  If it is already there, it
	just manages it and goes away */
	Widget		*widp;
	XmString	locstr;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	static char	butlabel[16];
	static int	sbx[] = {  5,150,295,440, 85,255,375 } ;
	static int	sby[] = {305,305,305,305,355,355,355 } ;
	static char	*sbla[] = { "Write Logical\nName File",
	"Execute Logical\nName File", "Write Out\nFrame Index",
	"Toplevel\nMenu", "Edit Logical\nName File", "Edit\nFrame Index",
	"Help" } ;

	if (NULL == (pmmp = managepmm(3))) return;
	lx11p = pmmp->pm_p11p;
	if ((position < 0) || (position > 7)) return;
	widp = &(lx11p->gp_spush[position]);
	if (NULL != *widp)
	  {
		button_color(position);	XtManageChild(*widp);	return;
	  }
	XtSetArg(pmmp->pm_ag[0], XmNx, sbx[position]);
	XtSetArg(pmmp->pm_ag[1], XmNy, sby[position]);
	locstr = XmStringCreateLtoR(sbla[position], lx11p->gp_chpo2);
	XtSetArg(pmmp->pm_ag[2], XmNlabelString, locstr);
	XtSetArg(pmmp->pm_ag[3], XmNwidth, 135);
	XtSetArg(pmmp->pm_ag[4], XmNheight, 50);
	XtSetArg(pmmp->pm_ag[5], XmNrecomputeSize, False);
	XtSetArg(pmmp->pm_ag[6], XmNbackground, lx11p->gp_progcol[SHOULD]);
	XtSetArg(pmmp->pm_ag[7], XmNforeground,
		lx11p->gp_progcol[NCOLORS + SHOULD]);
	XtSetArg(pmmp->pm_ag[8], XmNfontList, lx11p->gp_dflis);
	butlabel[0] = '\0';
	strncpy(butlabel, sbla[position], 6);	butlabel[6] = '\0';
	strcat(butlabel, "_button");
	*widp = XmCreatePushButton(pmmp->pm_setxbb, butlabel, pmmp->pm_ag, 9);
	switch(position)
	   {
		case	0:	CALL2(writelognames, 0);
				CALL2(all_colors, NULL);	break;
		case	1:	CALL2(execlognames, 0);
				CALL2(all_colors, NULL);	break;
		case	2:	CALL2(pwriteind, 0);
				CALL2(all_colors, NULL);	break;
		case	3:	CALL2(stexit, NULL);		break;
		case	4:	CALL2(editwork, CMDSUFFIX);	break;
		case	5:	CALL2(editwork, "frm");		break;
		case	6:	CALL2(help_cb, "pmotif");	break;
	   }
	all_colors();	XtManageChild(*widp);	XmStringFree(locstr);
	zeromem((void *)butlabel, strlen(butlabel));
}

static void setupvalue(Widget widget, char *tag)
{ /*	Takes appropriate action when we change an entry in the setup menu*/
	int		slot;
	String		result;
	XGNAME		*lxgn;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	DISP_ATTRS	*dispat;

	if (NULL == (pmmp = managepmm(3))) return;
	lxgn = pmmp->pm_namb;	lx11p = pmmp->pm_p11p;
		/* record the color and the string in the entry */
	XtSetArg(pmmp->pm_ag[0], XmNvalue, &result);
	XtGetValues(widget, pmmp->pm_ag, 1);
		/* figure out which entry this is... */
	for (slot = 0; slot <= 8; slot++)
		if (!strncmp(tag, pmmp->pm_stlb[slot], 3)) break;
	dispat = lx11p->gp_at;
	switch (slot)	/* set logicals or frame numbers appropriately */
	  {
		case	0:	(void)newidname(result, 0); /*changes id name*/
			break;
		case	1:	/* change WORK directory */
		 strcpy(lxgn->xmm_workna, result);
		 (void)set_logical("WORK", lxgn->xmm_workna); break;
		case	2:	/* change DATA directory */
		 strcpy(lxgn->xmm_datana, result);
		 (void)set_logical("DATA", lxgn->xmm_datana); break;
		case	3:	/* change SCRT directory */
		 strcpy(lxgn->xmm_scrtna, result);
		 (void)set_logical("SCRT", lxgn->xmm_scrtna);	break;
		case	4:	/* change XGDOC directory */
		 (void)set_logical("XGDOC", result);	break;
		case	5:	transfer_template(result);	break;
		case	6:	transfer_base(result);		break;
		case	7:	transfer_interval(result);	break;
		case	8:	if (NULL == dispat->di_fmind)
					dispat->di_fmind = ftables(7);
				(dispat->di_fmind)->fi_firstfm = atoi(result);
				break;
		case	9:	if (NULL == dispat->di_fmind)
					dispat->di_fmind = ftables(7);
				(dispat->di_fmind)->fi_lastfm = atoi(result);
				break;
	  }
	slot_color(slot);	/* re-color the slot if necessary */
	if (slot < 3)
	  {
		button_color(0); button_color(1); button_color(4);
	  }
	else if (slot > 4)
	  {
		button_color(2);	button_color(5);
	  }
}

static void setuptext(int position)
{ /*	This sets up one of the text-labelled questions within the setup menu;
	some of these are file name designations and the others are frame
	numbers. If the appropriate widget already exists we just manage it
	and go away. */
	int		bmorph, thisv, thisx, thisy;
	char		*cp, *cpe;
	Widget		*widp;
	XmString	locstr;
	DISP_ATTRS	*dispat;
	PMMWORK		*pmmp;
	XGNAME		*lxgn;
	XG_X11PROPS	*lx11p;
	static char	fixedv[6], setuplabel[16], setupprompt[16];
	static Widget	setxlab[NTEXTS];
	static int	filex = 170;	/* poscoo[0] + poscoo[2] */
	static char	*value = (char *)0;
	static int	stw[] = { 10, 30, 30, 30,  30, 30, 7, 7, 6, 7 } ;

	if ((position < 0) || (position >= NTEXTS)) return; /* verify position*/
	if (NULL == (pmmp = managepmm(3))) return;
	lx11p = pmmp->pm_p11p;	lxgn = pmmp->pm_namb;
	*setuplabel = '\0';	widp = &(setxlab[position]);
	thisx = pmmp->pm_poscoo[0]; thisy = pmmp->pm_poscoo[1];
	if (NULL == *widp)
	  {
		locstr = XmStringCreate(pmmp->pm_stlb[position],
			lx11p->gp_chpol);
		XtSetArg(pmmp->pm_ag[0], XmNlabelString, locstr);
		XtSetArg(pmmp->pm_ag[1], XmNfontList, lx11p->gp_dflis);
		if (position < 7) thisy = pmmp->pm_poscoo[1] +
			position * pmmp->pm_poscoo[6];
		else if (position == 7)
		  {
			thisx = pmmp->pm_poscoo[0] + 230;
			thisy = pmmp->pm_poscoo[1] + 6 * pmmp->pm_poscoo[6];
		  }
		else if (position == 8)
			thisy = pmmp->pm_poscoo[1] + 7 * pmmp->pm_poscoo[6];
		else	/* in principle this is general, but it's really == 9*/
		  {
			thisx = pmmp->pm_poscoo[0] + 230;
			thisy = pmmp->pm_poscoo[1] + 7 * pmmp->pm_poscoo[6];
		  }
		XtSetArg(pmmp->pm_ag[2], XmNx, thisx);
		XtSetArg(pmmp->pm_ag[3], XmNy, thisy);
		strncpy(setuplabel, pmmp->pm_stlb[position], 6);
		setuplabel[6] = '\0';
		strcat(setuplabel, "_label");
		*widp = XmCreateLabelGadget(pmmp->pm_setxbb, setuplabel,
			pmmp->pm_ag, 4);
		XmStringFree(locstr);
		zeromem((void *)setuplabel, strlen(setuplabel));
	  }
	XtManageChild(*widp);
	dispat = lx11p->gp_at;
	widp = &(lx11p->gp_spush[7 + position]);	*setupprompt = '\0';
	if (NULL == *widp)
	  { /* if we're dealing with names and RUN_ID isn't defined yet,
		we try to pick it out from the name of the WORK directory */
		if (position < 4)
		  {
			if (NULL == (value = getenv("RUN_ID")) ||'\0' ==*value)
			  {
				value = getenv("WORK");
				if (NULL != value && '\0' != *value)
				  {
					cpe = value + strlen(value);
					for (cp = cpe-1; cp > value; cp--)
						if (*cp == '/') break;
					if (cp == (cp-1) || cp == value)
						value = "testxtal";
					else value = cp + 1;
				  }
				/* if we're still stuck, use `testxtal' */
				else value = "testxtal";
			  }
			(void)newidname(value, 0);
		  }
		switch (position)
		  {
			case	0:	break;	/* id name: set it all up */
			case	1:	default:
					value = getenv("WORK");	break;
			case	2:	value = getenv("DATA");	break;
			case	3:	value = getenv("SCRT");	break;
			case	4:	value = get_xgdoc(0);	break;
			case	5:	value = grab_template(0); break;
			case	6:	thisv = grab_base();
					if (0 >= thisv) thisv = 10;
				sprintf(fixedv, "%3d", thisv);	break;
			case	7:	thisv = grab_interval();
					if (0 >= thisv) thisv = 1;
				sprintf(fixedv, "%3d", thisv);	break;
			case	8: if (NULL == dispat->di_fmind)
					dispat->di_fmind = ftables(7);
				sprintf(fixedv, "%3d",
				 (int)((dispat->di_fmind)->fi_firstfm));
				break;
			case	9: if (NULL == dispat->di_fmind)
					dispat->di_fmind = ftables(7);
				sprintf(fixedv, "%3d",
				 (int)((dispat->di_fmind)->fi_lastfm));
				break;
		  }
		if (position > 5) value = fixedv;
		if (NULL == value) value = "";
		XtSetArg( pmmp->pm_ag[0], XmNvalue, value);
		XtSetArg( pmmp->pm_ag[1], XmNcolumns, stw[position]);
		if (position < 7)
		  {
			thisx = filex;
			thisy = pmmp->pm_poscoo[5] +
				position * pmmp->pm_poscoo[6];
		  }
		else if (position == 7)
		  {
			thisx = filex + 160;
			thisy = pmmp->pm_poscoo[5] + 6 * pmmp->pm_poscoo[6];
		  }
		else if (position == 8)
		  {
			thisx = filex;
			thisy = pmmp->pm_poscoo[5] + 7 * pmmp->pm_poscoo[6];
		  }
		else	/* again, this is really position == 9 */
		  {
			thisx = filex + 160;
			thisy = pmmp->pm_poscoo[5] + 7 * pmmp->pm_poscoo[6];
		  }
		XtSetArg( pmmp->pm_ag[2], XmNx, thisx);
		XtSetArg( pmmp->pm_ag[3], XmNy, thisy);
		bmorph = NEEDNOT; /* nominal colorcode; changed in slot_color*/
		XtSetArg( pmmp->pm_ag[4], XmNbackground,
			lx11p->gp_progcol[bmorph]);
		XtSetArg( pmmp->pm_ag[5], XmNforeground,
			lx11p->gp_progcol[NCOLORS+bmorph]);
		strncpy(setupprompt, pmmp->pm_stlb[position], 6);
		setupprompt[6] = '\0';
		strcat(setupprompt, "_prompt");
		*widp = XmCreateTextField(pmmp->pm_setxbb,
			(String)setupprompt, pmmp->pm_ag, 6);
		XtAddCallback(*widp, XmNlosingFocusCallback,
		 (XtCallbackProc)setupvalue, pmmp->pm_stlb[position]);
		XtAddCallback(*widp, XmNactivateCallback,
		 (XtCallbackProc)setupvalue, pmmp->pm_stlb[position]);
		zeromem((void *)setupprompt, strlen(setupprompt));
	  }
	XtManageChild(*widp);	slot_color(position);	/* color it properly */
	/* re-do the button colors if needed */
	if (position < 3)
		for (bmorph = 0; bmorph <= 5; bmorph++) button_color(bmorph);
	else if (position == 3)
	  {
		button_color(0); button_color(1);
	  }
	else if (position > 4)
	  {
		button_color(2); button_color(5);
	  }
}

static void setxgenenv(void)
{ /*	Establishes the Motif window associated with the X-GEN environment */
	int		i;
	Boolean		is3digit, isbase36, isoldfmt;
	Widget		*widp;
	XmString	locstr;
	DISP_ATTRS	*dispat;
	FRAMEINDEX	*lofm;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	static Widget	setxtoggle[4];

	if (NULL == (pmmp = managepmm(3))) return;
	lx11p = pmmp->pm_p11p;
	if (NULL == pmmp->pm_setxbb) /* create environ window from scratch */
	  {
		XtSetArg(pmmp->pm_ag[0], XmNx, pmmp->pm_poscoo[3]);
		XtSetArg(pmmp->pm_ag[1], XmNy, pmmp->pm_poscoo[4]);
		XtSetArg(pmmp->pm_ag[2], XmNbackground,
			lx11p->gp_progcol[NEUTRAL]);
		XtSetArg(pmmp->pm_ag[3], XmNforeground,
			lx11p->gp_progcol[CNEUTRAL]);
		XtSetArg(pmmp->pm_ag[4], XmNwidth, 680);
		XtSetArg(pmmp->pm_ag[5], XmNheight,525);
		XtSetArg(pmmp->pm_ag[6], XmNresizePolicy, XmRESIZE_NONE);
		XtSetArg(pmmp->pm_ag[7], XmNlabelFontList, lx11p->gp_dflis);
		XtSetArg(pmmp->pm_ag[8], XmNtextFontList, lx11p->gp_dflis);
		XtSetArg(pmmp->pm_ag[9], XmNbuttonFontList, lx11p->gp_dflis);
		pmmp->pm_setxbb = XmCreateBulletinBoard(lx11p->gp_dimain,
			"setup_menu", pmmp->pm_ag, 10);
	  }
	dispat = lx11p->gp_at;	lofm = dispat->di_fmind;
	if ((NULL == lofm) || (999 == lofm->fi_lastfm)) /* set frame limitss*/
		lofm = dispat->di_fmind = ftables(7);
	for (i = 0; i < NTEXTS; i++) setuptext(i); /* establish text widgets */
	is3digit = lofm->fi_flags & 01;
	isbase36 = (lofm->fi_flags >> 1) & 01;	isoldfmt = lofm->fi_flags >> 2;
	if (NULL == *(widp = &(setxtoggle[0])))
	  {
		XtSetArg( pmmp->pm_ag[0], XmNset, is3digit);
		XtSetArg( pmmp->pm_ag[1], XmNx, 5);
		XtSetArg( pmmp->pm_ag[2], XmNy, 285);
		locstr = XmStringCreate("3-digit frame numbers",
			lx11p->gp_chpol);
		XtSetArg( pmmp->pm_ag[3], XmNlabelString, (XtArgVal)locstr);
		*widp = XmCreateToggleButtonGadget(pmmp->pm_setxbb,
			"threedigitbutton", pmmp->pm_ag, 4);
		XtAddCallback(*widp, XmNvalueChangedCallback,
		 (XtCallbackProc)index_booleans, "digits");
		XmStringFree(locstr);
	  }
	XtManageChild(*widp);
	if (NULL == *(++widp))
	  {
		XtSetArg( pmmp->pm_ag[0], XmNset, isbase36);
		XtSetArg( pmmp->pm_ag[1], XmNx, 205);
		XtSetArg( pmmp->pm_ag[2], XmNy, 285);
		locstr = XmStringCreate("base-36 frame numbers",
			lx11p->gp_chpol);
		XtSetArg( pmmp->pm_ag[3], XmNlabelString, (XtArgVal)locstr);
		*widp = XmCreateToggleButtonGadget(pmmp->pm_setxbb,
			"base36button", pmmp->pm_ag, 4);
		XtAddCallback(*widp, XmNvalueChangedCallback,
		 (XtCallbackProc)index_booleans, "base");
		XmStringFree(locstr);
	  }
	XtManageChild(*widp);
	if (NULL == *(++widp))
	  {
		XtSetArg( pmmp->pm_ag[0], XmNset, isoldfmt);
		XtSetArg( pmmp->pm_ag[1], XmNx, 405);
		XtSetArg( pmmp->pm_ag[2], XmNy, 285);
		locstr = XmStringCreate("Old DEG.MN format",
			lx11p->gp_chpol);
		XtSetArg(pmmp->pm_ag[3], XmNlabelString, (XtArgVal)locstr);
		*widp = XmCreateToggleButtonGadget(pmmp->pm_setxbb,
			"degminbutton", pmmp->pm_ag, 4);
		XtAddCallback(*widp, XmNvalueChangedCallback,
		 (XtCallbackProc)index_booleans, "convention");
		XmStringFree(locstr);
	  }
	XtManageChild(*widp);	for (i = 0; i < 7; i++) setupbutton(i);
	XtManageChild(pmmp->pm_setxbb);	XtRealizeWidget(pmmp->pm_setxbb);
	XRaiseWindow(lx11p->gp_di, XtWindow(pmmp->pm_setxbb));
}

static void killprogmenus(void)
{ /*	Unmanage all the program-related menus */
	int		i;
	PMMWORK		*pmmp;
	Widget		*widp;

	if (NULL == (pmmp = managepmm(7))) return;
	for (i = 0, widp = &(pmmp->pm_filew[0]); i < MAXPFILES; i++, widp++)
	   if (NULL != *widp)
	     {
		XtDestroyWidget(*widp);	*widp = NULL;
	     }
	for (i = 0, widp = &(pmmp->pm_boolw[0]); i < MAXPBOOLS; i++, widp++)
	   if (NULL != *widp)
	     {
		XtDestroyWidget(*widp);	*widp = NULL;
	     }
	for (i = 0, widp = &(pmmp->pm_labelw[0]); i < MAXLABELS; i++, widp++)
	   if (NULL != *widp)
	     {
		XtDestroyWidget(*widp);	*widp = NULL;
	     }
	if ((Widget)0 != pmmp->pm_fileform)
	  {
		XtUnmanageChild(pmmp->pm_fileform);
		XtDestroyWidget(pmmp->pm_fileform); pmmp->pm_fileform = NULL;
	  }
	if ((Widget)0 != pmmp->pm_paramform)
	  {
		XtUnmanageChild(pmmp->pm_paramform);
		XtDestroyWidget(pmmp->pm_paramform); pmmp->pm_paramform = NULL;
	  }
	if ((Widget)0 != pmmp->pm_boolform)
	  {
		XtUnmanageChild(pmmp->pm_boolform);
		XtDestroyWidget(pmmp->pm_boolform); pmmp->pm_boolform = NULL;
	  }
	if ((Widget)0 != pmmp->pm_proglabel)
	  {
		XtUnmanageChild(pmmp->pm_proglabel);
		XtDestroyWidget(pmmp->pm_proglabel); pmmp->pm_proglabel = NULL;
	  }
	if ((Widget)0 != pmmp->pm_commandwid)
	  {
#ifndef sun
		XtSetArg(pmmp->pm_ag[0], XmNtextString, ANULL);
		XtSetValues(pmmp->pm_commandwid, pmmp->pm_ag, 1);
#endif
		XtUnmanageChild(pmmp->pm_commandwid);
		XtDestroyWidget(pmmp->pm_commandwid);
		pmmp->pm_commandwid = NULL;
	  }
	if (NULL != (pmmp->pm_p11p)->gp_bl) /* unmanage program widget too */
	  {
		XtUnmanageChild(pmmp->pm_ourmain); redraw();
	  }
}

static void prevmenu(void)
{ /*	Exit smoothly out of a program-level menu and drop back to main */
	Window	locwin;
	PMMWORK	*pmmp;

	if (NULL == (pmmp = managepmm(7))) return;
	killprogmenus();
	XtManageChild(pmmp->pm_ourmain);
	if (BadWindow != (locwin = XtWindow(pmmp->pm_ourmain)))
		XRaiseWindow((pmmp->pm_p11p)->gp_di, locwin);
}

/*ARGSUSED*/
static void forktoIt(Widget widget, int tag)
{ /*	This calls `locsysc', which either forks & execs or simply
	executes the appropriate X-GEN command */
	int		ntoshow, oldht, olddisflags, oldfmno, oldwid, result;
	char			*value, cmd[60];
	static XmString		contents;
	static PMMWORK		*pmmp;
	static DISP_ATTRS	*dispat;
	VSPLINE			*lospl;
	XG_X11PROPS		*lx11p;
	static int		n = 1;

	if (NULL == (pmmp = managepmm(3))) return;
	lx11p = pmmp->pm_p11p;
	if ((Widget)0 == pmmp->pm_commandwid)
	  {
		printf("Cannot determine the contents of the command\n");
		return;
	  }
	dispat = lx11p->gp_at;	dispat->di_fmind = ftables(3);
	XtSetArg(pmmp->pm_ag[0], XmNtextString, &contents);
	XtGetValues(pmmp->pm_commandwid, pmmp->pm_ag, n);
	value = extract_first_xms_segment(contents);
	strcpy(pmmp->pm_ubu, value);
	if (NULL != contents) XmStringFree(contents);
	/* remember the spline values if we have them */
	if (NULL != (lospl = dispat->di_spl =
		cleanspline(dispat->di_spl, &result)))
	  {
		oldwid = lospl->v_detw;	oldht = lospl->v_deth;
	  }
	else {
		printf("Warning: spline unavailable before fork\n");
		oldwid = oldht = -1;
	     }
	oldfmno = dispat->di_fmno;
	(void)managekill(1); /* so we don't kill memory inappropriately */
	/* move the menus back if we're doing this from pdisplay */
	if (NULL != lx11p->gp_bl)
	  {
		killprogmenus();
		if (0 < (ntoshow = whatshown(pmmp->pm_ubu, cmd, 0)))
			cmdparse(cmd);
		XtManageChildren(lx11p->gp_tport, NTPORTS);
	  }
	olddisflags = dispat->di_disflags;
	dispat->di_disflags &= ~(DISPVERBOSEANY); /* reduce verbosity for now*/
	if (NULL == lx11p->gp_bl) /* from motif we always hop */
		result = locsysc(pmmp->pm_ubu, !tag);
	else if ((!tag) && (!strncmp(pmmp->pm_ubu, "integrate", 9)))
		result = dispinteg(pmmp->pm_ubu); /*integration is special */
	else	result = locsysc(pmmp->pm_ubu, !tag);	/* fork to RPC */
	/* flush output to stdout immediately and make sure it's still being
	 line-buffered (Curses applications mess this up on some systems) */
	dispat->di_disflags = olddisflags; /* restore display flags */
	(void)xeflush(stdout);	(void)setlinebuf(stdout);
	if (!tag) printf(" Run of (%s) resulted in status %d\n",
		pmmp->pm_ubu, result);
	else {
		printf("Command (%s) running in the ", pmmp->pm_ubu);
		printf("background; launch returned %d.\n", result);
		printf("Diagnostic output is being written to ./xgen.out\n");
	     }
	/* restore the display and the FRAMEINDEX,
	 which may have been chingered */
	(void)ftables(0);
	dispat->di_fmind = ftables(3);	dispat->di_fmno = oldfmno;
	dispat->di_imaname = indexname(oldfmno);
	dispat->di_spl = cleanspline((lx11p->gp_at)->di_spl, &result);
	if ((NULL != dispat->di_spl) &&
	 ((dispat->di_spl != lospl) || (lospl->v_detw !=
	 oldwid) || (lospl->v_deth != oldht) || (EOK != result)))
	  {
		dispat->di_spl = gspline(0, &result);
	  }
	if (NULL == dispat->di_spl)
	  {
		dispat->di_spl = cleanspline((VSPLINE *)0, &result);
		if (EOK != result)
		  {
			printf("Note: spline not correctly read in\n");
			dispat->di_spl = NULL;
		  }
	  }
	if (dispat != lx11p->gp_at) lx11p->gp_at = dispat;
	if (NULL == lx11p->gp_bl) prevmenu();	/* called from pmotif */
	else {
		if (0 < (ntoshow = whatshown(pmmp->pm_ubu, cmd, 1)))
			cmdparse(cmd);
		XtManageChild(pmmp->pm_ourmain);
	     }
	redo_maincolors();
}

static void make_bool(int position)
{ /*	Position and display a Boolean variable */
	int		i;
	char		cs, *cp, boolname[16], btitle[33];
	Boolean		isset;
	XmString	locstr;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	Widget		*widp;

	if (NULL == (pmmp = managepmm(7))) return;
	lx11p = pmmp->pm_p11p;
	pmmp->pm_entryy = pmmp->pm_poscoo[7] * (position - pmmp->pm_nfiles) + 4;
	isset = (((cs = pmmp->pm_defaults[position][1]) == 'y') || (cs == 'Y'));
	widp = &(pmmp->pm_boolw[position - pmmp->pm_nfiles]);
	XtSetArg( pmmp->pm_ag[0], XmNset, isset);
	if (NULL == *widp) /* create the widget if needed */
	  {
		XtSetArg( pmmp->pm_ag[1], XmNx, 2);
		XtSetArg( pmmp->pm_ag[2], XmNy, pmmp->pm_entryy);
		strcpy(btitle, pmmp->pm_labels[position]);
		i = strlen(btitle);
		for (cp = &(btitle[i]); i < 27; cp++, i++) *cp = ' ';
		strcpy(&(btitle[27]), pmmp->pm_endtitle);
		btitle[30] = pmmp->pm_defaults[position][0];
		locstr = XmStringCreate(btitle, lx11p->gp_chpol);
		XtSetArg( pmmp->pm_ag[3], XmNlabelString, (XtArgVal)locstr);
		XtSetArg( pmmp->pm_ag[4], XmNbackground,
		 lx11p->gp_progcol[NEUTRAL]);
		XtSetArg( pmmp->pm_ag[5], XmNforeground,
		 lx11p->gp_progcol[CNEUTRAL]);
		XtSetArg( pmmp->pm_ag[6], XmNfontList, lx11p->gp_dflis);
		strncpy(boolname, pmmp->pm_labels[position], 6);
		boolname[6] = '\0';
		strcat(boolname, "button");
		*widp = XmCreateToggleButton(pmmp->pm_boolform, boolname,
			pmmp->pm_ag, 7);
		XtAddCallback(*widp, XmNvalueChangedCallback,
		 (XtCallbackProc)update_cmd_line, NULL);
		XtManageChild(*widp);	XmStringFree(locstr);
	  }
	else {
		XtSetValues(*widp, pmmp->pm_ag, 1);	XtManageChild(*widp);
	     }
}

static void stick_file(int position)
{ /*	Put a filename into the proper place on a menu */
	char		*logname, *value;
	short		ncols;
	Widget		*widp;
	XmString	locstr;
	PMMWORK		*pmmp;
	XGNAME		*lxgn;
	XG_X11PROPS	*lx11p;
	static char	tfilename[80], tflabname[64], tcp[256];

	if ((position < 0) || (position >= MAXLABELS)) return;
	if (NULL == (pmmp = managepmm(3))) return;
	lx11p = pmmp->pm_p11p;	lxgn = pmmp->pm_namb;
	if (NULL == (pmmp = managepmm(3))) return;
	pmmp->pm_labely = pmmp->pm_poscoo[1] + (pmmp->pm_poscoo[6]) * position;
	logname = pmmp->pm_defaults[position];
	locstr = XmStringCreate(pmmp->pm_labels[position], lx11p->gp_chpol);
	widp = &(pmmp->pm_labelw[position]);
	XtSetArg(pmmp->pm_ag[0], XmNlabelString, (XtArgVal)locstr);
	XtSetArg(pmmp->pm_ag[1], XmNx, pmmp->pm_poscoo[0]);
	XtSetArg(pmmp->pm_ag[2], XmNy, pmmp->pm_labely);
	XtSetArg(pmmp->pm_ag[3], XmNbackground,
	 lx11p->gp_progcol[NEUTRAL]);
	XtSetArg(pmmp->pm_ag[4], XmNforeground,
	 lx11p->gp_progcol[CNEUTRAL]);
	XtSetArg(pmmp->pm_ag[5], XmNfontList, lx11p->gp_dflis);
	tflabname[0] = '\0';	strcpy(tflabname, logname);
	strcat(tflabname, "_label");
	*widp = XmCreateLabel(pmmp->pm_fileform,
		(String)tflabname, pmmp->pm_ag, 6);
	XtManageChild(*widp);	XmStringFree(locstr);
	ncols = MAXCOLS;	*tcp = '\0';
	if (NULL != (value = getenv(logname))) /* found the logical name */
	  { /* try to abbreviate it for display purposes.  If we
		succeed in abbreviating it, re-assign the logical name */
		strcpy(tcp, value);
#ifndef VMS
		if (!shorten(tcp, "SCRT"))
		  {
			if (!shorten(tcp, "DATA"))
				(void)shorten(tcp, "WORK");
		  }
#endif
	  }
	else *tcp = '\0';
	XtSetArg( pmmp->pm_ag[0], XmNvalue, tcp);
	widp = &(pmmp->pm_filew[position]);
	if (NULL == *widp)
	  {
		pmmp->pm_entryy = pmmp->pm_poscoo[5] +
		 (pmmp->pm_poscoo[6]) * position;
		XtSetArg( pmmp->pm_ag[1], XmNcolumns, ncols);
		XtSetArg( pmmp->pm_ag[2], XmNx, pmmp->pm_poscoo[2]);
		XtSetArg( pmmp->pm_ag[3], XmNy, pmmp->pm_entryy);
		XtSetArg( pmmp->pm_ag[4], XmNbackground,
		 lx11p->gp_progcol[SHOULD]);
		XtSetArg( pmmp->pm_ag[5], XmNforeground,
			lx11p->gp_progcol[NCOLORS+SHOULD]);
		*tfilename = '\0';
		strcpy(tfilename, logname);	strcat(tfilename, "_prompt");
		*widp = XmCreateTextField(pmmp->pm_fileform,
			(String)tfilename, pmmp->pm_ag, 6);
		XtAddCallback(*widp, XmNlosingFocusCallback,
		 (XtCallbackProc)add_logical,(caddr_t)logname);
		XtAddCallback(*widp, XmNlosingFocusCallback,
		 (XtCallbackProc)update_cmd_line, NULL);
		XtAddCallback(*widp, XmNactivateCallback,
		 (XtCallbackProc)add_logical,(caddr_t)logname);
		XtAddCallback(*widp, XmNactivateCallback,
		 (XtCallbackProc)update_cmd_line, NULL);
	  }
	else	XtSetValues(*widp, pmmp->pm_ag, 1);
	XtManageChild(*widp);
}

static void stick_string(int position)
{ /*	This handles one of the "switch" or "parameter" arguments */
	int		i, poff, poff2;
	char		*cp, *labelgname, *qant, *stitle, *textname;
	Widget		*widp;
	XmString	locstr;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	XGNAME		*lxgn;
	static int	vlabelw = 230;

	if ((position < 0) || (position >= MAXLABELS)) return;
	if (NULL == (pmmp = managepmm(3))) return;
	lxgn = pmmp->pm_namb;	lx11p = pmmp->pm_p11p;
	labelgname = &(lxgn->xmm_miscc[1280]);	*labelgname = '\0';
	stitle = &(lxgn->xmm_miscc[1408]);	*stitle = '\0';
	pmmp->pm_labely = pmmp->pm_poscoo[1] +
		pmmp->pm_poscoo[6] * (position - pmmp->pm_nfb);
	if (position < pmmp->pm_nfbs) /* switch label includes mnemonic */
	  {
		strcpy(stitle, pmmp->pm_labels[position]);
		i = strlen(stitle);
		for (cp = &(stitle[i]); i < 27; i++, cp++) *cp = ' ';
		strcpy(&stitle[27], pmmp->pm_endtitle);
		stitle[30] = pmmp->pm_defaults[position][0];
		locstr = XmStringCreate(stitle, lx11p->gp_chpol);
		XtSetArg(pmmp->pm_ag[0], XmNlabelString, locstr);
	  }
	else {
		locstr = XmStringCreate(pmmp->pm_labels[position],
			lx11p->gp_chpol);
		XtSetArg(pmmp->pm_ag[0], XmNlabelString, (XtArgVal)locstr);
	     }
	widp = &(pmmp->pm_labelw[position]);
	if (NULL == *widp)
	  {
		XtSetArg(pmmp->pm_ag[1], XmNx, 2);
		XtSetArg( pmmp->pm_ag[2], XmNy, pmmp->pm_labely);
		XtSetArg(pmmp->pm_ag[3], XmNbackground,
		 lx11p->gp_progcol[NEUTRAL]);
		XtSetArg(pmmp->pm_ag[4], XmNforeground,
		 lx11p->gp_progcol[CNEUTRAL]);
		XtSetArg(pmmp->pm_ag[5], XmNfontList, lx11p->gp_dflis);
		strncpy(labelgname, pmmp->pm_labels[position], 6);
		labelgname[6] = '\0';
		strcat(labelgname, "_label");
		*widp = XmCreateLabel(pmmp->pm_paramform, labelgname,
		 pmmp->pm_ag, 6);
		XtManageChild(*widp);	XmStringFree(locstr);
	  }
	else {
		XtSetValues(*widp, pmmp->pm_ag, 1); XtManageChild(*widp);
	     }
	pmmp->pm_entryy = pmmp->pm_poscoo[5] +
		pmmp->pm_poscoo[6] * (position - pmmp->pm_nfb);
	qant = &(pmmp->pm_defaults[position][position < pmmp->pm_nfbs]);
	if ('\0' == *qant) XtSetArg( pmmp->pm_ag[0], XmNvalue, "");
	else XtSetArg( pmmp->pm_ag[0], XmNvalue, qant);
	poff = position + NPUSHES + 1;	poff2 = poff + MAXPSWITCH;
	widp = (position < pmmp->pm_nfbs) ?
	 &(lx11p->gp_dpush[poff - pmmp->pm_nfb]) :
		&(lx11p->gp_dpush[poff2 - pmmp->pm_nfbs]);
	XtSetArg( pmmp->pm_ag[1], XmNcolumns,12);
	XtSetArg( pmmp->pm_ag[2], XmNx, vlabelw);
	XtSetArg( pmmp->pm_ag[3], XmNy, pmmp->pm_entryy);
	XtSetArg( pmmp->pm_ag[4], XmNbackground,
	 lx11p->gp_progcol[NEEDNOT]);
	XtSetArg( pmmp->pm_ag[5], XmNforeground,
		lx11p->gp_progcol[NCOLORS + NEEDNOT]);
	textname = &(lxgn->xmm_miscc[1344]);
	strncpy(textname, pmmp->pm_labels[position], 6);
	textname[6] = '\0';
	strcat(textname, "_prompt");
	for (cp = textname; '\0' != *cp; cp++)
		if ((' ' == *cp) || ('\t' == *cp)) *cp = '_';
	if (NULL != (*widp = XmCreateTextField(pmmp->pm_paramform,
	 textname, pmmp->pm_ag, 6)))
	  {
		XtAddCallback(*widp, XmNlosingFocusCallback,
		 (XtCallbackProc)update_cmd_line,NULL);
		XtAddCallback(*widp, XmNactivateCallback,
		 (XtCallbackProc)update_cmd_line,NULL);
		XtManageChild(*widp);
	  }
}

void update_cmd_line(void)
{ /*	Update the contents of the command line in a program menu */
	char		blabel, ftype, leaveablank, work[80];
	int		accv, bmorph, i, ncsofar, newx, newy;
	Widget		*widp;
	Boolean		isset, normally;
	String		svalue;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	XGNAME		*lxgn;
	Widget		cwi, butwid[5];
	static XmString	locsta[5];
	static int	oldx = 0;
	static int 	oldy = 0;
	static char	*locchar[] = { "Constructed Command",
	 "RUN &\nWait", "RUN &\nDon't Wait",  "Previous\nMenu", "Help"} ;
	static Arg	agc[] = {
	 { XmNselectionLabelString, ANULL },	{ XmNokLabelString, ANULL },
	 { XmNapplyLabelString, ANULL },	{ XmNcancelLabelString,ANULL },
	 { XmNhelpLabelString, ANULL },		{ XmNbackground, (XtArgVal) 0 },
	 { XmNdialogType, XmDIALOG_PROMPT },	{ XmNtextColumns,(XtArgVal)50 },
	 { XmNmarginHeight, (XtArgVal) 2 }, 	{ XmNmarginWidth,(XtArgVal)2 },
	 { XmNx, (XtArgVal) 450 },		{ XmNy, (XtArgVal) 440 },
	 { XmNlabelFontList, ANULL },		{ XmNtextFontList, ANULL },
	 { XmNbuttonFontList, ANULL } } ;

	if (NULL == (pmmp = managepmm(7))) return;
	lx11p = pmmp->pm_p11p;	lxgn = pmmp->pm_namb;
	if ((Widget)0 == pmmp->pm_fileform) return; /* fileform must be there*/
	for (i = 0; i < 5; i++)
	   {
		if (!i) locsta[i] =
			XmStringCreate(locchar[i], lx11p->gp_chpol);
		else	locsta[i] =
			XmStringCreateLtoR(locchar[i],lx11p->gp_chpo2);
		agc[i].value = (XtArgVal)(locsta[i]);
	   }
	agc[5].value = (XtArgVal)(lx11p->gp_progcol[NEUTRAL]);
	newx = ((pmmp->pm_nfiles * pmmp->pm_poscoo[6]) +
		(pmmp->pm_nbool * pmmp->pm_poscoo[7]) < CANMOVELEFT) ?
		140 : 450;
	agc[10].value = (XtArgVal)newx;
	if (NULL != lx11p->gp_bl) agc[10].value -= 100;
	newy = (pmmp->pm_nparam + pmmp->pm_nswitch + 2) * pmmp->pm_poscoo[6];
	if ((newx == 140) || (newy < 340)) newy = 340;
	agc[11].value = (XtArgVal)newy;
	agc[14].value = agc[13].value = agc[12].value =
		(XtArgVal)lx11p->gp_dflis;
	if ((Widget)0 != pmmp->pm_commandwid)
	  {
		if (True == XtIsRealized(pmmp->pm_commandwid))
			XtUnrealizeWidget(pmmp->pm_commandwid);
		XtDestroyWidget(pmmp->pm_commandwid);
	  }
#ifdef VMS
	accv = 0;
#else
	accv = 2;
#endif
	pmmp->pm_commandwid = cwi = XmCreateSelectionBox(
	 lx11p->gp_dimain, "commandwidget", agc, 15);
	oldx = newx;	oldy = newy;
	XtAddCallback(cwi, XmNokCallback,
	 (XtCallbackProc)forktoIt,(caddr_t)(0));
	XtAddCallback(cwi, XmNapplyCallback,
	 (XtCallbackProc)forktoIt, (caddr_t)(1));
	XtAddCallback(cwi, XmNcancelCallback, (XtCallbackProc)prevmenu, NULL);
	XtAddCallback(cwi, XmNhelpCallback,
	 (XtCallbackProc)help_cb, (caddr_t)(lxgn->xmm_pname));
	butwid[4] = XmSelectionBoxGetChild(cwi, XmDIALOG_TEXT);
	for (i = 0; i < 5; i++) XmStringFree(locsta[i]);
	if (NULL != butwid[4])
	  {
		XtSetArg(pmmp->pm_ag[0], XmNbackground,
			lx11p->gp_progcol[COMMAND]);
		XtSetArg(pmmp->pm_ag[1], XmNforeground,
			lx11p->gp_progcol[NCOLORS+COMMAND]);
		XtSetValues(butwid[4], pmmp->pm_ag, 2);
	  }
	strcpy(pmmp->pm_ubu, lxgn->xmm_pname);
	ncsofar = strlen(pmmp->pm_ubu);
	for (i = 0, widp = &(pmmp->pm_filew[0]);
	 i < pmmp->pm_nfiles; i++,widp++)
	   { /* ftype is 1 if it's an output file; 0 otherwise */
		ftype = !strncmp(pmmp->pm_labels[i], "Output", 6);
		if (NULL == *widp) stick_file(i);
		XtSetArg(pmmp->pm_ag[0], XmNvalue, &svalue);
		XtGetValues(*widp, pmmp->pm_ag, 1);
		if (*svalue)
		  {
#ifdef VMS
			strcpy(work, svalue);	/* no translation under vms */
#else
			(void)envtoname(svalue, work); /*translate&get color*/
#endif
			if (ftype) bmorph = (0 == access(work, accv)) ?
				COMMAND : SHOULD;
			else bmorph = (0 == access(work, accv)) ?
				NEEDNOT : CANNOT;
		  }
		else bmorph = CANNOT;
		XtSetArg(pmmp->pm_ag[0], XmNbackground,
		 lx11p->gp_progcol[bmorph]);
		XtSetArg(pmmp->pm_ag[1], XmNforeground,
		 lx11p->gp_progcol[NCOLORS+bmorph]);
		XtSetValues(*widp, pmmp->pm_ag, 2);
	   }
	for (i = 0, leaveablank = 1, widp = &(pmmp->pm_boolw[0]);
	  i < pmmp->pm_nbool; i++, widp++)
	   {
		XtSetArg(pmmp->pm_ag[0], XmNset, &isset);
		if (NULL == *widp) make_bool(pmmp->pm_nfiles + i);
		XtGetValues(*widp, pmmp->pm_ag, 1);
		normally = (pmmp->pm_defaults[i + pmmp->pm_nfiles][1] <= 'Z');
		blabel = pmmp->pm_defaults[i + pmmp->pm_nfiles][0];
		if (isset ^ normally)
		  {
			if (leaveablank)
			  {
				strcat(pmmp->pm_ubu, " -"); ncsofar += 2;
			  }
			pmmp->pm_ubu[ncsofar++] = blabel;
			pmmp->pm_ubu[ncsofar] = '\0';
			leaveablank = 0;
		  }
	   }
	for (i = 0, widp = &(lx11p->gp_dpush[NPUSHES + 1]);
	 i < pmmp->pm_nswitch; i++, widp++)
	   {
		if (NULL == *widp) stick_string(i + pmmp->pm_nfiles +
		 pmmp->pm_nbool);
		XtSetArg(pmmp->pm_ag[0], XmNvalue, &svalue);
		XtGetValues(*widp, pmmp->pm_ag, 1);
		if (*svalue)
		  {
			if (!leaveablank) sprintf(work, "%c%s",
				pmmp->pm_defaults[i+pmmp->pm_nfb][0], svalue);
			else sprintf(work, " -%c%s",
				pmmp->pm_defaults[i+pmmp->pm_nfb][0],svalue);
			strcat(pmmp->pm_ubu, work);	leaveablank = 1;
		  }
	   }
	for (i = 0, widp = &(lx11p->gp_dpush[NPUSHES + MAXPSWITCH + 1]);
	 i < pmmp->pm_nparam; i++, widp++)
	   {
		if (NULL == *widp) stick_string(i + pmmp->pm_nfiles +
			pmmp->pm_nbool + pmmp->pm_nswitch);
		XtSetArg(pmmp->pm_ag[0], XmNvalue, &svalue);
		XtGetValues(*widp, pmmp->pm_ag, 1);
		if (*svalue)
		  {
			strcat(pmmp->pm_ubu, " ");
			strcat(pmmp->pm_ubu, svalue);
		  }
	   }
	locsta[0] = XmStringCreate(pmmp->pm_ubu, "");
	XtSetArg(pmmp->pm_ag[0], XmNtextString, locsta[0]);
	XtSetValues(cwi, pmmp->pm_ag, 1);
	butwid[0] = XmSelectionBoxGetChild(cwi, XmDIALOG_OK_BUTTON);
	butwid[1] = XmSelectionBoxGetChild(cwi, XmDIALOG_APPLY_BUTTON);
	butwid[2] = XmSelectionBoxGetChild(cwi, XmDIALOG_CANCEL_BUTTON);
	butwid[3] = XmSelectionBoxGetChild(cwi, XmDIALOG_HELP_BUTTON);
	XtSetArg(pmmp->pm_ag[0], XmNwidth, 30);
	XtSetArg(pmmp->pm_ag[1], XmNmappedWhenManaged, True);
	XtSetValues(butwid[0], pmmp->pm_ag, 2);
	XtSetValues(butwid[1], pmmp->pm_ag, 2);
	XtSetValues(butwid[2], pmmp->pm_ag, 2);
	XtSetValues(butwid[3], pmmp->pm_ag, 2);
	XtManageChild(pmmp->pm_commandwid);
	XtManageChildren(butwid, 5);
	XmStringFree(locsta[0]);
}

/*ARGSUSED*/
static void makecmd(Widget widget, PMENUSTUFF *tagst)
{ /*	This produces the command-related widget and manages it.*/
	char		put, *cpi, *cpo, *newcom, *valstr;
	char		**azv, **dcp, **lazv;
	int		azc, bheight, fheight, i, j, lfirst, llast;
	int		position, rheight, thisx, ldim[2];
	Cardinal	narg;
	XmString	locstr;
	DISP_ATTRS	*dispat;
	XMSHOW		*xmp;
	FRAMEINDEX	*lofm;
	PMMWORK		*pmmp;
	XGNAME		*lxgn;
	XG_X11PROPS	*lx11p;
	static int	currtype = 0;
	static int	progdim[] = { 420, 6, 415, 280 };

	/* first, copy the information in the tagged structure into
	the appropriate global variables */
	if (NULL == (pmmp = managepmm(3))) return;
	lxgn = pmmp->pm_namb;	lx11p = pmmp->pm_p11p;
	strcpy(lxgn->xmm_pname, tagst->pr_name);
	for (cpi = lxgn->xmm_pname, cpo = lxgn->xmm_ptitle;
	 '\0' != *cpi; cpi++,cpo++)
		*cpo = ('a' <= *cpi && *cpi <= 'z' ? *cpi - ('a' - 'A') :*cpi);
	*cpo = '\0';	strcat(lxgn->xmm_ptitle, ":\n");
	strcat(lxgn->xmm_ptitle, tagst->pr_title);
	pmmp->pm_nfiles = tagst->pr_nfile;
	pmmp->pm_nbool = tagst->pr_nbool;
	pmmp->pm_nswitch = tagst->pr_nsw;
	pmmp->pm_nparam = tagst->pr_nparam;
	pmmp->pm_nfb = pmmp->pm_nfiles + pmmp->pm_nbool;
	pmmp->pm_nfbs = pmmp->pm_nfb + pmmp->pm_nswitch;
	for (i = 0, xmp = &(tagst->pr_show[0]);
	 i < pmmp->pm_nfbs + pmmp->pm_nparam; i++, xmp++)
	   {
		if ((NULL == xmp) || (NULL == xmp->xs_label))
		  {
			fprintf(stderr,
			 "Only %4d of the labels and defaults defined\n", i);
			break;
		  }
		strcpy(pmmp->pm_labels[i], xmp->xs_label);
		strcpy(pmmp->pm_defaults[i], xmp->xs_defau);
	   }
	dispat = lx11p->gp_at;	lofm = dispat->di_fmind;
	if (tagst->pr_flags & 07)	/* special values for frame limits */
	  {
		if ((NULL == lofm) || (lofm->fi_lastfm == 999))
		  lofm = dispat->di_fmind = ftables(7); /* find them */
		if (tagst->pr_flags & 01)	/* 1st numeric arg= start fm*/
		  {
			lfirst = lofm->fi_firstfm; llast = lofm->fi_lastfm;
			if ((0 <= lfirst) && (lfirst <= 999))
			  sprintf(pmmp->pm_defaults[pmmp->pm_nfbs],
				"%d", lfirst);
			if ((tagst->pr_flags & 02) &&
			 (0 <= llast) && (llast <= 999))
			   sprintf(pmmp->pm_defaults[pmmp->pm_nfbs+1],
				"%d", llast);
		  }
		else if (tagst->pr_flags & 04)
		  { /* special case: pr_flags & 04: we construct an argument
		    that looks like "-a<startfm>,<endfm>, e.g.  "-a23,47" */
			if ((0 <= lofm->fi_firstfm && lofm->fi_firstfm <= 999)
			 && (0 <= lofm->fi_lastfm && lofm->fi_lastfm <= 999))
			   sprintf(pmmp->pm_defaults[pmmp->pm_nfb], "a%d,%d",
			    (int)(lofm->fi_firstfm), (int)(lofm->fi_lastfm));
		  }
	  }
	/* now set the context-sensitive pmmp->pm_defaults */
	if (0 == currtype) currtype = dettype(ldim);
	if (NULL == (newcom = smartdefa(lxgn->xmm_pname,
	 0, &currtype))) newcom = lxgn->xmm_pname;
	/* convert command to argc,argv-type setup */
	/* first we test memory allocation */
	lazv = str2acav(newcom, &azc);
	azc--;	azv = lazv + 1;	/* to correct peculiarity in str2acav */
	xmp = &(tagst->pr_show[pmmp->pm_nfiles]);
	position = pmmp->pm_nfiles;
	for (i = 0; i < pmmp->pm_nbool; i++, xmp++, position++)
	   {
		if (findbool(azc, azv, xmp->xs_defau[0])) /* Boolean found */
			put = (xmp->xs_defau[1] == 'Y') ? 'Y' : 'y';
		else	put = (xmp->xs_defau[1] == 'N') ? 'N' : 'n';
		pmmp->pm_defaults[position][1] = put;
		pmmp->pm_defaults[position][2] = '\0';
	   }
	for (i = 0; i < pmmp->pm_nswitch; i++, xmp++, position++)
	   {
		if (NULL != (valstr = findsw(azc, azv, xmp->xs_defau[0])))
			strcpy(&(pmmp->pm_defaults[position][1]), valstr);
		else pmmp->pm_defaults[position][1] = '\0';
	   }
	/* handle parameters: do so by moving past the
	 Booleans and switches and recording the remaining parameters */
	for (dcp = azv + 1, j = 1; j < azc; j++, dcp++)
	   if (**dcp != '-') break;
	for (i = 0; i < pmmp->pm_nparam; i++, xmp++, position++)
	   {
		if ((j < azc) && ('\0' != **dcp))
		  {
			strcpy(pmmp->pm_defaults[position], *dcp);
			j++;	dcp++;
		  }
		else {
			for ( ; i < pmmp->pm_nparam; i++, position++)
			   {
				pmmp->pm_defaults[position][0] = '\0';
			   }
			break;
		     }
	   }
	/* in previous versions we killed the strings here. No more... */
	narg = 0;	thisx = progdim[0];	if (lx11p->gp_bl) thisx -= 100;
	XtSetArg(pmmp->pm_ag[narg], XmNx, thisx);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNy, progdim[1]);			narg++;
	locstr = XmStringCreateLtoR(lxgn->xmm_ptitle, lx11p->gp_chpo2);
	XtSetArg(pmmp->pm_ag[narg], XmNlabelString, locstr);		narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNbackground,
	 lx11p->gp_progcol[NEUTRAL]); narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNforeground,
	 lx11p->gp_progcol[CNEUTRAL]); narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNfontList, lx11p->gp_dflis);			narg++;
	if (NULL != pmmp->pm_proglabel) XtDestroyWidget(pmmp->pm_proglabel);
	pmmp->pm_proglabel = XmCreateLabel(lx11p->gp_dimain,
		"program_label", pmmp->pm_ag, narg);
	XtManageChild(pmmp->pm_proglabel);
	/* create three separate bulletin boards--one for the files,
	one for the booleans, and one for the switches and run-time
	 (non-switch-style) parameters */
	fheight = (pmmp->pm_nfiles + 1) * pmmp->pm_poscoo[6];
	bheight = pmmp->pm_nbool * pmmp->pm_poscoo[7] + 6;
	rheight = (pmmp->pm_nparam + pmmp->pm_nswitch + 1) * pmmp->pm_poscoo[6];
	thisx = pmmp->pm_poscoo[3];	if (NULL != lx11p->gp_bl) thisx -= 100;
	narg = 0;	XtSetArg(pmmp->pm_ag[narg], XmNx, thisx);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNy, pmmp->pm_poscoo[4]);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNwidth, progdim[2]);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNheight, fheight);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNresizePolicy, XmRESIZE_NONE); narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNbackground,
	 lx11p->gp_progcol[NEUTRAL]); narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNforeground,
	 lx11p->gp_progcol[CNEUTRAL]); narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNlabelFontList, lx11p->gp_dflis);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNtextFontList, lx11p->gp_dflis);	narg++;
	XtSetArg(pmmp->pm_ag[narg], XmNbuttonFontList,lx11p->gp_dflis);	narg++;
	if (NULL != pmmp->pm_fileform) XtDestroyWidget(pmmp->pm_fileform);
	pmmp->pm_fileform = XmCreateBulletinBoard(lx11p->gp_dimain,
	 "file_menu", pmmp->pm_ag,narg);
	XtManageChild(pmmp->pm_fileform);
	if (pmmp->pm_nbool) /* These retain several arguments used above*/
	  {
		narg = 0; XtSetArg(pmmp->pm_ag[narg], XmNx, thisx);
		narg++;
		XtSetArg(pmmp->pm_ag[narg], XmNy, pmmp->pm_poscoo[4] + fheight);
		narg++;
		XtSetArg(pmmp->pm_ag[narg], XmNwidth, progdim[3]);	narg++;
		XtSetArg(pmmp->pm_ag[narg], XmNheight, bheight);	narg++;
		if (NULL != pmmp->pm_boolform)
			XtDestroyWidget(pmmp->pm_boolform);
		pmmp->pm_boolform = XmCreateBulletinBoard(
		 lx11p->gp_dimain, "bool_menu", pmmp->pm_ag, 10);
		XtManageChild(pmmp->pm_boolform);
	  }
	if (pmmp->pm_nswitch || pmmp->pm_nparam) /* 2nd menu is right of 1st */
	  {
		XtSetArg(pmmp->pm_ag[0], XmNx, thisx + progdim[2]);
		XtSetArg(pmmp->pm_ag[1], XmNy, pmmp->pm_poscoo[4]);
		XtSetArg(pmmp->pm_ag[2], XmNwidth, progdim[2] - 30);
		XtSetArg(pmmp->pm_ag[3], XmNheight, rheight);
		if (NULL != pmmp->pm_paramform)	
			XtDestroyWidget(pmmp->pm_paramform);
		pmmp->pm_paramform = XmCreateBulletinBoard(
		 lx11p->gp_dimain, "param_menu", pmmp->pm_ag, 10);
		XtManageChild(pmmp->pm_paramform);
	  }
	for (i = j = 0; i < pmmp->pm_nfiles; i++, j++) stick_file(i);
	for (i = 0; i < pmmp->pm_nbool; i++, j++) make_bool(j);
	for (i = 0; i < pmmp->pm_nswitch; i++, j++) stick_string(j);
	for (i = 0; i < pmmp->pm_nparam; i++, j++) stick_string(j);
	update_cmd_line();
	/* it's less confusing if we move the program-choice menu back */
	XLowerWindow(lx11p->gp_di, XtWindow(pmmp->pm_ourmain));
	/* ... but then we need to move the command widget up */
	if ((Widget)0 != pmmp->pm_commandwid)
	 XRaiseWindow(lx11p->gp_di, XtWindow(pmmp->pm_commandwid));
}

void redo_maincolors(void)
{ /*	This resets the background colors of all the buttons in the menu
	according to the presence or absence of some critical files. The
	X location of each button is set to 0 for buttons we don't display */
	int			i, lprgy, prev, prognum, shown;
	char			*bname, *fna, *fna0, *lastc, *lastoc, *loname;
	Widget			*wp;
	XmString		locstr;
	Window			thiswin;
	PMMWORK			*pmmp;
	XGNAME			*lxgn;
	XG_X11PROPS		*lx11p;
	static char		isthere[NPFILES], pcolor[NPPROGS];
	static char		*shortname;
	static int		nprogs;
	static PMENUSTUFF	*tmenu, **pmlist;
	static char		*fxgen[NPFILES] = {
	 "XGENCMD",	"AMASK",	"USPLINE",	"CENTROIDS", "WMASK",
	 "UPARAMS",	"UREFLS",	"MULTIREF",	"SCALEIN" } ;
	static short		prgx[NPPROGS] = {
	 10, 125,  10, 125,  78,  10, 125,  10, 125,  10, 125,  72, 10, 125,
	 62,  10, 125, 10, 125,  10, 125,  72,  10, 125,  72,  10, 125,  10,
	125, 10,  125, 10, 125 } ; /* currently we're displaying all of them */

	if (NULL == (pmmp = managepmm(3))) return;
	lxgn = pmmp->pm_namb;	lx11p = pmmp->pm_p11p;
	for (fna0 = NULL, i = 0; i < NPFILES; i++)
	   {
		fna = getenv(fxgen[i]);	errno = 0;	if (!i) fna0 = fna;
		isthere[i] = (NULL != fna) && (0 == xgaccess(fna, 4, 1));
	   }
	if ((NULL == fna0) || ('\0' == *fna0)) fna0 = "./newxtal.cmd";
	shortname = filepart(fna0, 2);
	if (NULL == (pmlist = pmall(&nprogs)))
	  {
		fprintf(stderr, "Unable to create program-specific menus\n");
		return;
	  }
	/* The basic command file ".cmd" or ".COM" must exist to do anything
	else. Also, the RUN_ID should match the prefix of the command file. */
	if ((NULL != shortname) && ('\0' != *shortname))
		lastc = shortname + strlen(shortname) - 1;
	else	lastc = NULL;
	if ((NULL != (bname = lxgn->xmm_idna)) && ('\0' != *bname))
		lastoc = lxgn->xmm_idna + strlen(lxgn->xmm_idna) - 1;
	else	lastoc = NULL;
	if ((NULL != lastc) && (NULL != lastoc) &&
	 (*lastc == '.') && (*lastoc != '.')) *lastc = '\0';
	if ((NULL != (fna = getenv("RUN_ID"))) && ('\0' != *fna) &&
	 (!strcmp(fna, lxgn->xmm_idna)) &&
	 (isthere[0]) && !strcmp(shortname, lxgn->xmm_idna))
	  {
		for (prognum = 0; prognum < nprogs; prognum++)
		   {
			tmenu = *(pmlist + prognum);
			pcolor[prognum] = ready(tmenu->pr_name);
		   }
	  }
	else {
		if (NULL != getenv("ALLOC_VERBOSE"))
		  {
			fprintf(stderr, "Everything is gray because ");
			if ((fna == NULL) || (*fna == '\0')) fprintf(stderr,
			 "we don't know the run ID\n");
			else if (strcmp(fna, lxgn->xmm_idna)) fprintf(stderr,
			 "fna is (%s); idname is (%s)\n", fna, lxgn->xmm_idna);
			else if (!isthere[0]) fprintf(stderr,
			 "isthere[0] is zero\n");
			else if (strcmp(shortname, lxgn->xmm_idna))
			 fprintf(stderr, "shortname is (%s), idname is (%s)\n",
				shortname, lxgn->xmm_idna);
			else fprintf(stderr, "of some unfathomable problem\n");
		  }
		pcolor[0] = SHOULD;
		for (prognum = 1; prognum < nprogs; prognum++)
			pcolor[prognum] = CANNOT;
	     }
	bname = &(lxgn->xmm_miscc[1408]);	*bname = '\0';
	if (NULL == pmmp->pm_progbutw[0])
	  {
		lprgy = 40;
		for (prev = -1, shown = prognum = 0; prognum < nprogs;
		 prognum++)
		   if (prgx[prognum] > 0)
		     {
			XtSetArg(pmmp->pm_ag[0], XmNx, prgx[prognum]);
			if ((prognum) && (prgx[prognum] <= prgx[prev]))
			  {
				if (((prgx[prognum] < 50) && (prgx[prev] < 90))
				 || (prgx[prognum] > 50)) lprgy += 24;
				else lprgy += 30;
			  }
			XtSetArg(pmmp->pm_ag[1], XmNy, lprgy);
			tmenu = *(pmlist + prognum); loname = tmenu->pr_name;
			locstr = XmStringCreateLtoR(loname, lx11p->gp_chpol);
			XtSetArg(pmmp->pm_ag[2], XmNlabelString,
			 (XtArgVal)locstr);
			XtSetArg(pmmp->pm_ag[3],XmNbackground,
			 lx11p->gp_progcol[(int)(pcolor[prognum])]);
			XtSetArg(pmmp->pm_ag[4], XmNforeground,
			 lx11p->gp_progcol[NCOLORS+(int)(pcolor[prognum])]);
			XtSetArg(pmmp->pm_ag[5], XmNfontList, lx11p->gp_dflis);
			strcpy(bname, loname);	strcat(bname, "_button");
			pmmp->pm_progbutw[shown++] = XmCreatePushButton(
			 pmmp->pm_ourmain, (String)bname, pmmp->pm_ag, 5);
			XmStringFree(locstr);
			zeromem((void *)bname, strlen(bname));
			prev = prognum;
		   }
		wp = &(pmmp->pm_progbutw[0]);
		XtAddCallback(*wp, XmNactivateCallback,
		 (XtCallbackProc)setxgenenv, NULL); wp++;
		for (prognum = 1; prognum < nprogs; prognum++)
		   if (prgx[prognum] > 0)
		     {
			tmenu = *(pmlist + prognum);
			XtAddCallback(*wp, XmNactivateCallback,
			 (XtCallbackProc)makecmd, (XtPointer)tmenu);
			wp++;
		     }
		XtManageChildren(pmmp->pm_progbutw, shown);
	   }
	else for (prognum = 0, wp = &(pmmp->pm_progbutw[0]); prognum < nprogs;
	 prognum++)
	  if (prgx[prognum] > 0)
	    {
		XtSetArg(pmmp->pm_ag[0], XmNbackground,
			lx11p->gp_progcol[(int)(pcolor[prognum])]);
		XtSetArg(pmmp->pm_ag[1], XmNforeground,
		 lx11p->gp_progcol[NCOLORS+(int)(pcolor[prognum])]);
		XtSetValues(*wp, pmmp->pm_ag, 2);	XtManageChild(*wp);
		wp++;
	    }
	if (BadWindow == (thiswin = XtWindow(pmmp->pm_ourmain))) printf(
	 "Couldn't determine the window associated with the program menu\n");
	else XRaiseWindow(lx11p->gp_di, thiswin);
}

XG_X11PROPS *get_pm11(void)
{ /*	Stub to tell the outside world about the static global called
	 (pmmp->pm_p11) here*/
	PMMWORK	*pmmp;

	return (NULL == (pmmp = managepmm(3))) ? NULL : pmmp->pm_p11p;
}

static void exitbutt_activate(void)
{ /*	Orderly return to the end of these windows*/
	int	result;
	VSPLINE	*lospl;
	PMMWORK	*pmmp;
	XG_X11PROPS	*lx11p;

	(void)xeflush(stdout);
	if (NULL == (pmmp = managepmm(3))) return;	lx11p = pmmp->pm_p11p;
	if (NULL == lx11p->gp_bl)
	  {
		backtomain(lx11p); /* invoked from pmotif */
		exit(EOK);	/* to really get out */
	  }
	else {
		prevmenu();	/* in case we exit with the prog window open */
		XtUnmanageChild(pmmp->pm_ourmain); /*pdisplay*/
		lospl = cleanspline((VSPLINE *)0, &result);
		pmmp->pm_progbutw[0] = NULL;	redraw();
	     }
}

int pmpopulate(Widget toplev)
{ /*	fetch and create the program-management widgets */
	int		i, miw, n, sperr;
	char		*cp;
	PMMWORK		*pmmp;
	XG_X11PROPS	*lx11p;
	XGNAME		*lxgn;
	static XmString	locstr;
	static Widget	exitwid, polabel;
	static Arg	agf[] = {
		{ XmNx, (XtArgVal) 10 }, { XmNy, (XtArgVal) 2 },
		{ XmNlabelString, ANULL } } ;
	static Arg	age[] = {
	 { XmNx, (XtArgVal) 75 }, { XmNy, (XtArgVal) 554 },
	 { XmNlabelString, ANULL }, { XmNfontList, ANULL } } ;
	static int	poscoo[] = {10, 20, 160, 160, 46, 12, 30, 24 };
	static char	*stlb[] = { "Run Identifier", "Directories: Work: ",
	 "Raw Data", "Scratch", "Helpfiles", "Frame-name template",
	 "Frame-# Base", "Interval", "Frame Range: from", " to " } ;
	static char	endtitle[] = "( -v)";

	if (NULL == (pmmp = managepmm(3))) return EALLOC;
	lx11p = pmmp->pm_p11p;	lxgn = pmmp->pm_namb;
	for (i = 0; i < 8; i++) pmmp->pm_poscoo[i] = poscoo[i];
	for (i = 0; i < 10; i++) pmmp->pm_stlb[i] = stlb[i];
	pmmp->pm_endtitle = endtitle;
	/* The program-jumping widget is set up differently depending
	 on whence we're calling it */
	n = 0;
	if (lx11p->gp_bl)
	  {
		start_colors();	/* define the colors we'll use for the menus */
		if (NULL == pmmp->pm_ourmain)
		  {
			miw = (lx11p->gp_at)->di_minw;	n = 0;
			XtSetArg(pmmp->pm_ag[n], XmNx, miw + 16);	n++;
			XtSetArg(pmmp->pm_ag[n], XmNy, 40);	n++;
			XtSetArg(pmmp->pm_ag[n], XmNwidth, 200);	n++;
			XtSetArg(pmmp->pm_ag[n], XmNheight, 680);	n++;
			XtSetArg(pmmp->pm_ag[n], XmNbackground,
				lx11p->gp_progcol[NEUTRAL]); n++;
			XtSetArg(pmmp->pm_ag[n], XmNforeground,
				lx11p->gp_progcol[CNEUTRAL]); n++;
			pmmp->pm_ourmain = XmCreateBulletinBoard(
			 lx11p->gp_dimain, "progs_main", pmmp->pm_ag, n);
		  }
	  }
	else if (NULL == pmmp->pm_ourmain)
	  {
		locstr = XmStringCreate("XGEN", "");
		XtSetArg(pmmp->pm_ag[n], XmNdialogTitle,(XtArgVal)locstr); n++;
		pmmp->pm_ourmain = XmCreateForm(toplev, "pmotif_main",
		 pmmp->pm_ag, n);
		lx11p->gp_dimain = pmmp->pm_ourmain;
		start_colors();	/* define the colors we'll use */
		XmStringFree(locstr);
	   }
	if (NULL == polabel) /* overall label */
	  {
		locstr = XmStringCreateLtoR(
		 "Press appropriate button\nto set up a XGEN program",
			lx11p->gp_chpol);
		agf[2].value = (XtArgVal)locstr;
		polabel = XmCreateLabelGadget(pmmp->pm_ourmain,
		 "xgenlabel", agf, 3);
		XmStringFree(locstr);
	  }
	if (NULL == exitwid) /* exit button */
	  {
		locstr = XmStringCreate("Exit", lx11p->gp_chpo2);
		age[2].value = (XtArgVal)locstr;
		age[3].value = (XtArgVal)lx11p->gp_dflis;
		exitwid = XmCreatePushButtonGadget(pmmp->pm_ourmain,
		 "exit_button", age, 4);
		XtAddCallback(exitwid, XmNactivateCallback,
		 (XtCallbackProc)exitbutt_activate, NULL);
	  }
	XtManageChild(pmmp->pm_ourmain); /* manage visible toplevel window*/
	XtManageChild(polabel);		/* manage the label */
	XtManageChild(exitwid);		/* manage the exit button */
	XtRealizeWidget(toplev);	/* so the managed hierarchy displays */
	XmStringFree(locstr);
	if (0 >= basic_names(0)) printf("Warning: no ID name known\n");
	if ('\0' != *(lxgn->xmm_idna))
	  {
		(void)set_logical("RUN_ID",lxgn->xmm_idna);
		printf("Setting RUN_ID to %s\n", getenv("RUN_ID"));
	  }
	else printf("Warning: lxgn doesn't contain an id name\n");
	pmmp->pm_ubu = lxgn->xmm_misc;  /* general-purpose buffer */
	cp = pmmp->pm_strs;
	for (i = 0; i < MAXLABELS; i++, cp += LABELLENGTH)
		pmmp->pm_labels[i] = cp;
	for (i = 0; i < MAXDEFAULTS; i++, cp += VALUELENGTH)
		pmmp->pm_defaults[i] = cp;
	/* attempt to start the pixel-to-centimeter conversions */
	(lx11p->gp_at)->di_spl = cleanspline((VSPLINE *)0, &sperr);
	/* update color maps according to contents of critical directories */
	redo_maincolors();	return EOK;
}

/* end of pmmenus.c */
