/* pmutil.c -- global-free utilities for XGEN's motif-based user interface

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	21 Jan 2006: gp_dpush, gp_dbcol replaced by gp_spush, gp_spush
	 6 Jan 2006: prototypes in xgmdefs.h; set_logical can return >= 1
	 6 Oct 2005: housekeeping
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	18 Oct 2004: help menus use constant-width (Courier) font
		and are derived from files with suffix .txt rather than .doc
	12 Oct 2004: locstr is XmString throughout; cosmetic changes
	27 Aug 2004: routines used only here are declared as static; cosmetics
	16 Aug 2004: created from simple routines in pmutils.c
 Changes to pmutils.c and its predecessors that are relevant here:
	 3 Apr 2004: numerous typecasts because a String isn't an XmString
	29 Mar 2004: killprogmenus() and whatshown() written;
		called from prevmenu() and forktoIt()
	 6 Mar 2004: delete memory for String variables after using them
	 1 Mar 2004: use get_xgdoc to get the directory for
		the help files in the setup menu
	25 Aug 2003: #include xgmdefs.h rather than pmdefs.h;
		xgmdefs.h includes some items usable in pdisplay as well

   Purpose: setting up nearly static-global-free Motif stuff
   Cross references: xgmdefs.h
*************************************************************************/
#include	"xgmdefs.h"
#ifdef VMS
#define		FILEFORMAT1	"%s%s.%s"	/* VMS format */
#define		READWRITE	4	/* don't try to require write access */
#define		CMDSUFFIX	"COM"	/* standard suffix for command file */
#else
#define		FILEFORMAT1	"%s/%s.%s"	/* Unix format */
#define		READWRITE	6	/* do require write access */
#define		CMDSUFFIX	"cmd"	/* standard suffix for command file */
#endif
#define	DOCSUFFIX	"txt"
#define	CALL2(x,y)	XtAddCallback(*widp,XmNactivateCallback,(XtCallbackProc)x,(XtPointer)y)

 typedef struct	{
		char	*pmu_etext;
		Widget	pmu_ewid, pmu_exwid, pmu_quited;
		} PMUWORK;

static PMUWORK *managepmu(int opno)
{ /*	manages a static variable *pmup according to the operation opno */
	static PMUWORK	*pmup;
	static char	etmsg[] = "pmup->pmu_etext in managepmu";
	static char	pmumsg[] = "pmup in managepmu";

	if (opno == 0)
	  {
		if (NULL != pmup)
		  {
			memzap((AVOID **)&(pmup->pmu_etext), etmsg);
			XtDestroyWidget(pmup->pmu_ewid);
			XtDestroyWidget(pmup->pmu_exwid);
			XtDestroyWidget(pmup->pmu_quited);
			memzap((AVOID **)&pmup, pmumsg);
		  }
		return NULL;
	  }
	if (opno == 1) pmup = managepmu(0);
	if (NULL == pmup)
	  {
		if (NULL == (pmup = (PMUWORK *)callop(1, sizeof (PMUWORK),
		 pmumsg))) return NULL;
	  }
	return pmup;	/* we'll get fancier later */
}

void pwriteind(void)
{ /*	Argument-free wrapper for writeindex */
	XG_X11PROPS	*locxgp;
	DISP_ATTRS	*dispat;

	locxgp = get_pm11();	dispat = locxgp->gp_at;
	if (NULL == dispat->di_fmind) dispat->di_fmind = ftables(7);
	writeindex(dispat->di_fmind, 0);
}

void slot_color(int slotno)
{ /*	This updates the color of one of the directories defined by a logical
	name on the setup menu. slotno is 0 for WORK, 1 for DATA, 2 for SCRT.*/
	char		*result;
	int		newindex;
	XG_X11PROPS	*locxgp;
	DISP_ATTRS	*dispat;
	Pixel		thiscolor, newcolor;
	Widget		*widp;
	Arg		lag[3];
	static char	*dirn[] = { NULL, "WORK", "DATA", "SCRT", "XGDOC" } ;
	static unsigned	accval[] = { 0, 6, 4, 6, 4 } ;

	if (slotno < 1 || slotno >= 5) return;	/* unchanging */
		/* we can't recolor the widget unless it exists...*/
	locxgp = get_pm11();	dispat = locxgp->gp_at;
	if (NULL == *(widp = &(locxgp->gp_spush[7 + slotno]))) return;
	XtSetArg(lag[0], XmNbackground, &thiscolor);
	XtGetValues(*widp, lag, 1);	/* figure out current color */
	if (slotno <= 4)
	  {
		if ((NULL != (result = getenv(dirn[slotno]))) &&
		 ('\0' != *result))
		  {
			errno = 0;	newindex = (dir_access(result,
				(int)(accval[slotno])) ?  CANNOT: NEEDNOT);
		  }
		else	newindex = CANNOT;
	  }
	else if (slotno == 6 || slotno == 8)
	  {
		if (NULL == dispat->di_fmind) newindex = CANNOT;
		else newindex = ((dispat->di_fmind)->fi_flags & 04) ?
			NEEDNOT : CANNOT;
	  }
	else newindex = CAN;
	newcolor = locxgp->gp_progcol[newindex];
	if (thiscolor != newcolor)
	  {
		XtSetArg(lag[0],XmNbackground,newcolor);
		XtSetArg(lag[1],XmNforeground,
		 locxgp->gp_progcol[NCOLORS+newindex]);
		XtSetValues(*widp, lag, 2);	XtManageChild(*widp);
	  }
}

void index_booleans(Widget widget, char *tag)
{ /*	Geometric determinations associated with a Boolean variable */
	int		bitval;
	Boolean		isset;
	Arg		lag[1];
	XG_X11PROPS	*locxgp;
	DISP_ATTRS	*dispat;
	FRAMEINDEX	*lofm;

	XtSetArg(lag[0], XmNset, &isset);	XtGetValues(widget, lag, 1);
	locxgp = get_pm11();	dispat = locxgp->gp_at;
	if (NULL == dispat->di_fmind) dispat->di_fmind = ftables(7);
	lofm = dispat->di_fmind;
	bitval = (*tag == 'd') ? 1 : ((*tag == 'b') ? 2 : 4);
	if (isset) lofm->fi_flags |= bitval;	else lofm->fi_flags &= ~bitval;
	if (bitval == 4)
	  {
		slot_color(6);	slot_color(8);
	  }
}

void button_color(int position)
{ /*	This modifies the color of a button according to the existence of
	some appropriate files or logical names. */
	int		bmorph, ftype, nlogi;
	char		*svalue;
	XGNAME		*namb;
	Pixel		curcolor, newcolor;
	Widget		*widp;
	Arg		lag[2];
	XG_X11PROPS	*locxgp;

	locxgp = get_pm11();
	widp = &(locxgp->gp_spush[position]);	if (NULL == *widp) return;
	bmorph = CANNOT;	ftype = 1;	/* default values */
	namb = sharename(7);
	switch (position)
	      {
		case	0:	default:
			svalue = getenv("XGENCMD");		break;
		case	1:	svalue = getenv("RUN_ID");
			if ('\0' == *(namb->xmm_idna)) bmorph = CANNOT;
			else {
				nlogi = more_logicals(1);
				bmorph = ((nlogi < 27) || (NULL == svalue)
				 || ('\0' == *svalue) || strcmp(
				 namb->xmm_idna, svalue) ?  SHOULD : COMMAND);
			     }
			break;
		case	2:	svalue = getenv("FRAMES");	break;
		case	3:	return;
		case	4:	svalue = getenv("XGENCMD"); ftype = 0; break;
		case	5:	svalue = getenv("FRAMES"); ftype = 0; break;
		case	6:	return;
	      }
	if (1 != position)
	  {
		if (ftype)
		  {
			if ((svalue) && ('\0' != *svalue) &&
			 (0 == xgaccess(svalue, 4, 1))) bmorph = COMMAND;
			else bmorph = SHOULD;
		  }
		else {
			if ((svalue) && ('\0' != *svalue) && (0 ==
			 xgaccess(svalue, READWRITE, 1))) bmorph = NEEDNOT;
			else bmorph = CANNOT;
		     }
	  }
	newcolor = locxgp->gp_progcol[bmorph];
	XtSetArg(lag[0], XmNbackground, &curcolor); XtGetValues(*widp, lag, 1);
	if (curcolor != newcolor)	/* different: change it */
	  {
		XtSetArg(lag[0], XmNbackground, newcolor);
		XtSetArg(lag[1], XmNforeground,
			locxgp->gp_progcol[NCOLORS+bmorph]);
		XtSetValues(*widp, lag, 2);
	  }
}

void all_colors(void)
{ /*	Update colors on all buttons & logical-name slots in setup menu. */
	register int	i;

	for (i = 1; i < 9; i++) slot_color(i);
	for (i = 0; i < 7; i++) button_color(i);
	redo_maincolors();	/* even the colors of the program buttons! */
}

static void quiteditor(void)
{ /* Orderly departure from this vale of tears */
	PMUWORK	*pmup;

	if (NULL == (pmup = managepmu(3))) return;
	memzap((AVOID **)&(pmup->pmu_etext), "etext in quiteditor");
	XtUnmanageChild(pmup->pmu_quited);
	XtUnmanageChild(XtParent(pmup->pmu_quited));
}

/*ARGSUSED*/
static void exiteditor(Widget widget, char *tag)
{ /*	Smoothly exit out of a text editor after writing buffer out */
	int		nchars;
	char		*cpe;
	String		result;
	Arg		lag[1];
	PMUWORK		*pmup;
	static FILE	*fped;

	if (NULL == (pmup = managepmu(3))) return;
	if ((NULL != tag) && ('\0' != *tag)) /* write out contents of file */
	  {
		XtSetArg(lag[0], XmNvalue, &result);
		XtGetValues(pmup->pmu_ewid, lag, 1);
		if (NULL == result || '\0' == *result) fprintf(stdout,
			"No output to %s: text damaged\n", tag);
		else if (NULL != (fped = fopen(tag, "w"))) /* open output */
		  {
			for (cpe = result; '\0' != *cpe; ) cpe++;
			nchars = cpe - result;	/* see how much to write */
			printf("writing %d characters to %s; ", nchars, tag);
			if (1 != fwrite((void *)result, (size_t)nchars,1,fped))
			  fprintf(stdout,
				"Warning: error in writing to output file\n");
			(void)fclose(fped);	/* close the output file */
			fped = NULL;	printf("done\n");
		  }
		else fprintf(stdout, "Error opening %s as output\n", tag);
	  }
	else if (NULL != tag)
	  fprintf(stdout, "Faulty output filename (%s); no output\n", tag);
	else	fprintf(stdout, "Output filename undefined; no output\n");
	XtUnmanageChild(pmup->pmu_exwid); quiteditor();
}

static void pmedit(char *finame, Boolean canwrite)
{ /*	A simple text editor routine */
#ifdef VMS
	char		*cp;
	int		nhere;
#endif
	int		nchars, nread;
	Cardinal	narg;
	Window		thiswindow;
	XmString	locstr;
	Arg		lag[9];
	Widget		scrwid;
	XGNAME		*namb;
	XG_X11PROPS	*locxgp;
	PMUWORK		*pmup;
	static int	helpcoo[] = { 180, 66, 600, 480, 550, 20, 40, 40 };
	static char	*constf;
	static FILE	*fped;
	static XFontStruct	*fontc;
	static char	*nfontc ="*courier-bold-r*120*";
	static XmFontListEntry	constle;
	static XmFontList	constl;

	namb = sharename(7);
	strcpy(namb->xmm_misc, finame);	errno = 0;
	locxgp = get_pm11();
	/* create a fontlist for constant-width characters */
	if ((char *)0 == constf)
	  {
		constf = "constant";	fontc = (XFontStruct *)0;
		if ((fontc = XLoadQueryFont(locxgp->gp_di, nfontc)))
		  {
			constle = XmFontListEntryCreate(constf,
			 XmFONT_IS_FONT, (XtPointer)fontc);
			constl = XmFontListAppendEntry(NULL, constle);
		  }
		if (((XFontStruct *)0 == fontc) || ((XmFontList)0 == constl))
		  {
			fprintf(stderr, "Error creating font for help\n");
			return;
		  }
	  }
	if (NULL == (fped = fopen(namb->xmm_misc, "r")))
	  {
		fprintf(stdout,
#ifdef VMS
	 "Can't open %s; errno = %d, %d\n", namb->xmm_misc, errno, vaxc$errno);
#else
	 "Can't open %s; errno = %d\n", namb->xmm_misc, errno);
#endif
		return;
	  }
	if (-1 == fseek(fped, 0L, 2))
	  {
		fprintf(stdout, "Can't seek to end of the file %s\n",
			namb->xmm_misc);
		(void)fclose(fped);	return;
	  }
	if (NULL == (pmup = managepmu(3))) return;
	if (NULL == (pmup->pmu_etext = (char *)mallop((unsigned)((nchars =
	 (int)ftell(fped)) + 2), "pmup->pmu_etext in pmedit")))
	  {
		(void)fclose(fped);	return;
	  }
	if (-1 == fseek(fped, 0L, 0))
	  {
		fprintf(stdout, "Can't rewind %s\n", namb->xmm_misc);
		(void)fclose(fped);	fped = NULL;	return;
	  }
#ifdef VMS
	for (cp = pmup->pmu_etext, nread = 0; NULL != fgets(cp, 120, fped); )
	   {
		nhere = strlen(cp);	cp += nhere;	nread += nhere;
	   }
	/* Under VMS some characters never come out. */
	if ((double)nread * 1.3 < nchars)
	  {
		fprintf(stdout,
		 "Tried to read about %d characters from %s; only got %d\n",
			nchars, nread);	return;
	  }
	nhere = nchars - nread;	nchars -= nhere;
	zeromem((AVOID *)cp, nhere);	/* nulls after what we read */
#else
	if (1 != (nread = fread((void *)pmup->pmu_etext,
	 (size_t)nchars, 1, fped)))
	  {
		fprintf(stdout,
	 "Error in fread from %s: got %d, should have gotten 1\n",
			namb->xmm_misc, nread);	return;
	  }
	*(pmup->pmu_etext + nchars) = '\0'; /* explicit null so no mixed files*/
#endif
	printf("%sing %s; %d characters\n",
		(canwrite ? "Edit" : "Scann"), namb->xmm_misc, nchars);
	narg = 0;	XtSetArg(lag[narg], XmNvalue, pmup->pmu_etext);	narg++;
	XtSetArg(lag[narg], XmNeditable, canwrite);	narg++;
	if (NULL == pmup->pmu_ewid)
	  {
		XtSetArg(lag[narg], XmNx, helpcoo[0]);	narg++;
		XtSetArg(lag[narg], XmNy, helpcoo[1]);	narg++;
		XtSetArg(lag[narg], XmNheight, helpcoo[3]);	narg++;
		XtSetArg(lag[narg], XmNwidth, helpcoo[2]);	narg++;
		XtSetArg(lag[narg], XmNeditMode, XmMULTI_LINE_EDIT);	narg++;
		XtSetArg(lag[narg], XmNbackground, locxgp->gp_progcol[CANNOT]);
			narg++;
		XtSetArg(lag[narg], XmNforeground,
			locxgp->gp_progcol[NCOLORS+CANNOT]);	narg++;
		XtSetArg(lag[narg], XmNfontList, constl);	narg++;
		pmup->pmu_ewid = XmCreateScrolledText(locxgp->gp_dimain,
			(String)"XGENedit", lag, narg);
	  }
	else XtSetValues(pmup->pmu_ewid, lag, narg);
	scrwid = XtParent(pmup->pmu_ewid);	XtManageChild(pmup->pmu_ewid);
	if (NULL == pmup->pmu_quited)
	  {
		narg = 0;	XtSetArg(lag[narg], XmNx, helpcoo[4]); narg++;
		XtSetArg(lag[narg], XmNy, helpcoo[5]);	narg++;
		XtSetArg(lag[narg], XmNwidth, helpcoo[6]);	narg++;
		XtSetArg(lag[narg], XmNheight, helpcoo[7]);	narg++;
		locstr = XmStringCreate("quit", "");
		XtSetArg(lag[narg], XmNlabelString, (XtArgVal)locstr);	narg++;
		XtSetArg(lag[narg], XmNfontList, locxgp->gp_dflis);	narg++;
		pmup->pmu_quited = XmCreatePushButton(scrwid,
		 "quitedbutton", lag, narg);
		XtAddCallback(pmup->pmu_quited, XmNactivateCallback,
		 (XtCallbackProc)quiteditor,
		 (canwrite ? (XtPointer)(namb->xmm_misc) : NULL));
		XmStringFree(locstr);
	  }
	XtManageChild(pmup->pmu_quited);
	if (canwrite)
	  {
		if (NULL == pmup->pmu_exwid)
		  {
			narg = 0; XtSetArg(lag[narg], XmNx, helpcoo[4]); narg++;
			XtSetArg(lag[narg], XmNy, helpcoo[5] + 50);	narg++;
			XtSetArg(lag[narg], XmNwidth, helpcoo[6]);	narg++;
			XtSetArg(lag[narg], XmNheight, helpcoo[7]);	narg++;
			locstr = XmStringCreate("exit", "");
			XtSetArg(lag[narg], XmNlabelString, (XtArgVal)locstr);
			narg++;
			XtSetArg(lag[narg], XmNfontList, locxgp->gp_dflis);
			narg++;
			pmup->pmu_exwid = XmCreatePushButton(scrwid,
				"pmup->pmu_exwidbutton", lag, narg);
			XtAddCallback(pmup->pmu_exwid, XmNactivateCallback,
			 (XtCallbackProc)exiteditor,
			 (XtPointer)(namb->xmm_misc));
			XmStringFree(locstr);
		  }
		XtManageChild(pmup->pmu_exwid);
	  }
	else if (NULL != pmup->pmu_exwid) XtUnmanageChild(pmup->pmu_exwid);
	XtManageChild(scrwid);	thiswindow = XtWindow(scrwid);
	XRaiseWindow(locxgp->gp_di, thiswindow);
}

/*ARGSUSED*/
void editwork(Widget widget, char *tag)
{ /*	This edits the file whose name is <workname>/<idname>.%s, where %s
	can be "cmd", "frm", "xlg", etc. and is picked up from callback tag. */
	static char	*lwname;
	static XGNAME	*namb;

	namb = sharename(7);	lwname = namb->xmm_miscc; *lwname = '\0';
	if ('\0' == *(namb->xmm_workna))
		strcpy(namb->xmm_workna, getenv("WORK"));
	sprintf(lwname, FILEFORMAT1, namb->xmm_workna, namb->xmm_idna, tag);
	pmedit(lwname, True); /* set up edit window for command file */
}

/*ARGSUSED*/
void help_cb(Widget widget, char *tag)
{ /*	Set up read-only edit window for examining help */
	char	*result, helpname[80];

	if (NULL == (result = get_xgdoc(0)))
	  {
		printf("Cannot provide help: XGDOC undefined\n"); return;
	  }
	sprintf(helpname, FILEFORMAT1, result, tag, DOCSUFFIX);
	pmedit(helpname, False);
}

char *extract_first_xms_segment(XmString cs)
{ /*	Get normal string from XmString */
	XmStringContext		context;
	XmStringCharSet		charset;
	XmStringDirection	direction;
	Boolean			separator;
	char			*primitive_string;

	XmStringInitContext(&context,cs);
	XmStringGetNextSegment(context,&primitive_string,
	   &charset,&direction,&separator);
	XmStringFreeContext(context);	return(primitive_string);
}

int whatshown(char *instr, char *cstr, int whenask)
{ /*	This suggests a string, cstr, that might get sent to
	the display parser based on the contents of instr.
	It also turns on and off certain display overlays directly.
	The results may depend slightly on whether this is called before
	or after the command is executed: that's embodied in the variable
	`whenask', which is 0 before the command and 1 afterward.
	This is a temporary version of this routine; a more sophisticated
	version that really picks apart the instr string will come later.
	The routine returns the number of strings returned (0 or 1). */
	int		endf, initf, midf, nret;
	char		*tname;
	XG_X11PROPS	*locxgp;
	DISP_ATTRS	*dispat;
	FRAMEINDEX	*lofm;

	locxgp = get_pm11();	dispat = locxgp->gp_at;
	lofm = dispat->di_fmind;
	cstr[0] = '\0';
	if ((NULL != dispat) && (NULL != lofm))
	  {
		initf = lofm->fi_firstfm; endf = lofm->fi_lastfm;
		midf = (initf + endf) / 2;
	  }
	else {
		initf = endf = midf = -1;
		printf("\nframindex undefined, evidently\n");
	     }
	nret = 1; /* by default, we do return a command */
	if (!strncmp(instr, "calibrate", 9))
	  {
		tname = trname("uspline");
		if (0 == xgaccess(tname, 4, 1))
			dispat->di_disflags |= DISPCALIB_NEED;
		tname = trname("caldata");
		if (0 == xgaccess(tname, 4, 1)) strcpy(cstr, "br");
		else if (initf >= 0) sprintf(cstr, "di %d", initf);
		else nret--;
	  }
	else if (!strncmp(instr, "border", 6))
	  {
		tname = trname("amask");
		if (0 == xgaccess(tname, 4, 1))
			dispat->di_disflags |= DISPBORDER_NEED;
		if (initf >= 0) sprintf(cstr, "di %d", initf);
		else nret--;
	  }
	else if (!strncmp(instr, "spots", 5))
	  {
		tname = trname("centroids");
		if (0 == xgaccess(tname, 4, 1))
			dispat->di_disflags |= DISPCEN_NEED;
		if (initf >= 0) sprintf(cstr, "di %d", initf);
		else nret--;
	  }
	else if (!strncmp(instr, "refine", 6))
	  {
		tname = trname("uparams");
		if (0 == xgaccess(tname, 4, 1))
				dispat->di_disflags |= DISPHKL_NEED;
		if (!whenask)
		  {
			tname = trname("centroids");
			if (0 == xgaccess(tname, 4, 1))
				dispat->di_disflags |= DISPCEN_NEED;
		  }
		else dispat->di_disflags &= ~DISPCEN_NEED;
		if (midf >= 0) sprintf(cstr, "di %d", midf);
		else nret--;
	  }
	else if (!strncmp(instr, "integrate", 9))
	  {
		tname = trname("uparams");
		if (0 == xgaccess(tname, 4, 1))
			dispat->di_disflags |= DISPHKL_NEED;
		dispat->di_disflags &= ~(DISPCEN_NEED | DISPCALIB_NEED);
		if (initf >= 0) sprintf(cstr, "di %d", initf);
		else nret--;
	  }
	else if ((!strncmp(instr, "reduce", 6)) ||
	 (!strncmp(instr, "pickuref", 8)))
	  {
		dispat->di_disflags &=
			~(DISPBORDER_NEED | DISPCALIB_NEED | DISPCEN_NEED);
		if (whenask)
		  {
			dispat->di_disflags &= ~DISPHKL_NEED;
			tname = trname("urefls");
			if (0 == xgaccess(tname,  4, 1))
				strcpy(cstr, "ur io");
			else nret--;
		  }
		else {
			tname = trname("uparams");
			if (0 == xgaccess(tname, 4, 1))
				dispat->di_disflags |= DISPHKL_NEED;
			if (initf >= 0) sprintf(cstr, "di %d", initf);
			else nret--;
		     }
	  }
	else if (!strncmp(instr, "scalek", 6))
	  {
		dispat->di_disflags &= ~(DISPBORDER_NEED | DISPCALIB_NEED |
			DISPHKL_NEED);
		if (whenask)
		  {
			tname = trname("scalein");
			if (0 == xgaccess("scalein", 4, 1))
				strcpy(cstr, "mr di");
			else nret--;
		  }
		else {
			tname = trname("multiref");
			if (0 == xgaccess(tname, 4, 1))
				strcpy(cstr, "mr io");
			else nret--;
		     }
	  }
	else if ((!strncmp(instr, "reject", 6)) ||
	 (!strncmp(instr, "stats", 5)) || (!strncmp(instr, "makemu", 6)))
	  {
		dispat->di_disflags &= ~(DISPBORDER_NEED | DISPCALIB_NEED |
			DISPHKL_NEED);
		tname = trname("scalein");
		if (0 == xgaccess(tname, 4, 1)) strcpy(cstr, "mr di");
		else nret--;
	  }
	else nret = 0;	/* ill-defined cases don't get covered yet */
	return nret;
}

void add_logical(Widget widget, char *tag)
{ /*	This picks the "XmNvalue" field out of a widget and sets 
	a logical name equal to that value. The logical name is in the tag. */
	char	adtbuf[80];
	int	slout;
	String	adlbuf;
	Arg	lag[1];

	if (NULL != tag || '\0' != *tag)
	  {
		XtSetArg(lag[0], XmNvalue, &adlbuf);
		XtGetValues(widget, lag, 1);
#ifdef VMS
		strcpy(adtbuf, adlbuf);
#else
		(void)envtoname(adlbuf, adtbuf); /* translate out shorthand */
#endif
		if (1 > (slout = set_logical(tag, adtbuf))) fprintf(stdout,
		 "Failure in setting %s to %s\n", tag, adtbuf);
		else if (slout == 2) fprintf(stdout,
		 "%s was already defined as %s; no change\n", tag, adtbuf);
	  }
	else if (NULL != tag)
	 fprintf(stdout, "add_logical: faulty tag starts octal %o,%o,%o\n",
		*tag, *(tag+1), *(tag+2));
	else fprintf(stdout, "add_logical received NULL tag\n");
}

void start_colors(void)
{/*	This establishes background and foreground colors */
	int		i, status;
	Colormap	tmap;
	XColor		screen[11], exact, *scrp;
	Pixel		*bgq, *fgq;
	Arg		lag[1];
	XG_X11PROPS	*locxgp;
	static int	n = 1;
	static char	*cn[] =
	 {	"White", "Black", "Red", "Limegreen", "Forestgreen", "Pink", 
		"Yellow", "VioletRed", "DimGray", "SkyBlue", "Khaki" } ;

	locxgp = get_pm11();	XtSetArg(lag[0], XmNcolormap, &tmap);
	XtGetValues(locxgp->gp_dimain, lag, n);
	for (i = 0, scrp = &(screen[0]); i < 11; i++, scrp++)
	   {
		status = XAllocNamedColor(locxgp->gp_di,
		 tmap, cn[i], scrp, &exact);
		if (BadAccess == status || BadColor == status ||
		 BadName == status || BadValue == status) fprintf(stdout,
		  "Warning: the color %s may not appear properly\n", cn[i]);
	   }
	bgq = &(locxgp->gp_progcol[0]);	fgq = bgq + NCOLORS;
	*(bgq + CANNOT) = screen[8].pixel;  *(fgq + CANNOT) = screen[0].pixel;
	*(bgq + CAN)    = screen[3].pixel;  *(fgq + CAN)    = screen[2].pixel;
	*(bgq + SHOULD) = screen[6].pixel;  *(fgq + SHOULD) = screen[7].pixel;
	*(bgq + NEEDNOT) = screen[9].pixel; *(fgq + NEEDNOT) = screen[1].pixel;
	*(bgq + COMMAND) = screen[10].pixel;*(fgq + COMMAND) = screen[2].pixel;
	*(bgq + NEUTRAL) = screen[5].pixel; *(fgq + NEUTRAL) = screen[1].pixel;
}

void backtomain(XG_X11PROPS *pm11p) /* return to pmmain cleanly */
{
	XFlush(pm11p->gp_di);	XCloseDisplay(pm11p->gp_di);
	(void)managekill(0);	/* longjmp(pmjbuf, 1); */
}

/* end of pmutil.c */
