/* xgmdefs.h -- definitions and typedefs for the motif-based user interface

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	22 Jun 2006: replace gp_{rotda,imda,magda,limits} elements with
	 components of the DISP_ATTRS structure; color designations moved here
	12 Jun 2006: reorder function declarations
	21 Jan 2006: add gp_progcol[], gp_imdasize, gp_spush[] to struct;
		store_image has no args now; toggleon gone; NPCTOT defined
	 4 Jan 2006: Motif function prototypes moved here
	 9 Sep 2005: eliminate gp_locof element of XG_X11PROPS structure type
	 5 Oct 2004: only 4 text ports, not 5
	 3 Oct 2004: 32 data ranges rather than 8; NEXTRAS defined;
		XG_X11PROPS structure type adjusted accordingly
	 4 Mar 2004: move the processing button from the menubar to the side
	29 Feb 2004: increase # of push buttons to enable all menus to work
	28 Feb 2004: eliminate #defines associated with pulldown processing
		menu in pdisplay, since we do that all within pmutils now
	21 Jan 2004: license message inserted
	15 Sep 2003: regularize and organize the XG_X11PROPS structure
		and move most of the #defines from dispx.c
	28 Aug 2003: incorporate XG_X11PROPS structure
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: move NTEXTS from pmglob.c and pmext.h to here.
	 1 Nov 2002: move MAXPFILES, etc. to pmprogs.c
	10 Jun 2002: move XMSHOW, PMENUSTUFF typedefs to ../include/xgen.h
	25 Mar 2002: removed MSI attachments
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: initial X-GEN version

   Purpose: definitions and typedefs for pmotif and pdisplay
   Cross References: pmmain.c, etc.
********************************************************************/
#ifndef XGMDEFS_H
#define XGMDEFS_H	1

#define		NEED_MOTIF	1
#include	"displays.h"
#include	<setjmp.h>

#ifdef	VMS
#include	"decw$include:intrinsic.h"
#include	"decw$include:stringdefs.h"
#include	"decw$include:shell.h"
#include	"decw$include:xmp.h"
#include	errno
#else
#include	<Xm/Xm.h>
#include	<errno.h>
#include	<pwd.h>
#endif

#define		ANULL		((XtArgVal)NULL)
#define		NTEXTS		10
#define		NTPORTS		4	/* # of places to put text info */
#define		NSCHEMES	4	/* # of color schemes */
#define		NMENUS		4	/* # of menus at top of the screen */
#define		NFRAMEBS	9	/* # of frame-related buttons */
#define		NUREFLBS	8	/* # of urefls-related buttons */
#define		NMREFBS		4	/* # of multiref-related buttons */
#define		NOVERBS		7	/* # of overlay-related buttons */
#define		NRANGES		32	/* # of intensity ranges in displays */
#define		NEXTRAS		6	/* extra colors besides intensities */
#define		NDCOLORS	(NRANGES + NEXTRAS)	/* # of colors used */
#define		NALLOCCOL	(NDCOLORS + NRANGES)	/* allocated colrs*/
#define		FIRSTUREFLS	(1+NFRAMEBS)	/* index of 1st urefls button*/
#define		FIRSTMREF	(FIRSTUREFLS+NUREFLBS) /* index of 1st mref */
#define		FIRSTOVERL	(FIRSTMREF + NMREFBS) /* index of 1st overlay*/
#define		NPUSHES		(FIRSTOVERL + NOVERBS)	/* total # of buttons*/
#define		NUIND		4	/* # of Urefls independent vbls */
#define		NUDEP		5	/* # of Urefls dependent vbls */
#define		NURHBUTTS	8	/* # of Urefls hkl buttons */
#define		NPCTOT		(2 * NCOLORS)	/* # of program-code colors */
/* Macro to make picky compilers happy */
#define		XpNU	((XtPointer) NULL)

#ifdef machine_irix64
typedef int		XGCAR;
#else
#ifdef machine_indy
typedef int		XGCAR;
#else
#ifdef sgi
typedef Cardinal	XGCAR;
#else
typedef int		XGCAR;
#endif
#endif
#endif

/* convenience typedef for certain kinds of X-GEN X-Windows displays */
typedef struct {
		DISP_ATTRS	*gp_at, *gp_bl;
		Window		gp_wi, gp_maw;
		Display		*gp_di;
		GC		gp_ogc, gp_igc;
		Pixel		gp_col, gp_dbcol[NALLOCCOL], gp_progcol[NPCTOT];
		int		gp_ivn, gp_dvn, gp_uhco, gp_uhva, gp_uhty;
		int		gp_nobinu, gp_nregx, gp_nregy;
		int		gp_icn, gp_ndcolors, gp_imdasize;
		int		gp_bmp, gp_bpl, gp_imo, gp_imf;
		unsigned	gp_dep;
		char		*gp_fdbuf;
		xglong		gp_coi[NRANGES];
		jmp_buf		gp_jbuf;
		Visual		*gp_disv;
		Widget		gp_color[NSCHEMES], gp_iscal[2];
		Widget		gp_dpush[NPUSHES + 40], gp_spush[20];
		Widget		gp_frsca[3], gp_urefi[NUIND], gp_urefd[NUDEP];
		Widget		gp_topbu[NMENUS], gp_toppu[NMENUS];
		Widget		gp_urhbu[NURHBUTTS], gp_tport[NTPORTS];
		Widget		gp_dimain, gp_dtoplev, gp_filesel;
		Widget		gp_imwid, gp_lastfwid, gp_magwid, gp_poldmenu;
		Widget		gp_urefcan, gp_urefg, gp_urefsc, gp_urhform;
		XmFontList	gp_dflis;
		XmString	gp_uhstr[9];
		char		*gp_chpol, *gp_chpo2;
		} XG_X11PROPS;

/* function prototypes for Motif */
extern Widget	XmCreateBulletinBoard(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateBulletinBoard(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateCascadeButton(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateDrawingArea(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateFileSelectionBox(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateForm(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateLabelGadget(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateLabel(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateMenuBar(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreatePulldownMenu(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreatePushButtonGadget(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreatePushButton(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateRadioBox(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateScale(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateScrolledText(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateScrolledWindow(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateSelectionBox(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateTextField(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateToggleButtonGadget(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmCreateToggleButton(Widget parent, char *name,
			ArgList args, Cardinal argct);
extern Widget	XmSelectionBoxGetChild(Widget sb, u_char which);

/* function prototypes from within X-GEN */
 
extern int count65k(MAIL *headp);
extern int diamond(XG_X11PROPS *olayp, int ismag);
extern int dispcmain(int argc, char *argv[]);
extern int dispinteg(char *callstr);
extern int dispkill(int irv);
extern int graphics(int argc, char **argv, DISP_ATTRS *mainatt);
extern int main(int argc, char *argv[]);
extern int main(int argc, char *argv[]);
extern int pmmain(int argc, char *argv[]);
extern int pmpopulate(Widget toplev);
extern int store_image(void);
extern int whatshown(char *instr, char *cstr, int whenask);
extern void add_logical(Widget widget, char *tag);
extern void all_colors(void);
extern void attr2locals(void);
extern void backtomain(XG_X11PROPS *pm11p);
extern void button_color(int position);
extern void chgdisp(void);
extern void chgframe(Widget wid, char *tag, XmAnyCallbackStruct *cbdata);
extern void cmdparse(char *stri);
extern void display_pixel(Widget widget);
extern void do_superpositions(XG_X11PROPS *xxg, int ismag);
extern void doexit(void);
extern void draw_axes(XG_X11PROPS *olayp, int ismag);
extern void draw_it(void);
extern void editwork(Widget widget, char *tag);
extern void fdproc(char *strbuf, int *sourp, pid_t *inpid);
extern void filereport(XG_X11PROPS *xxg);
extern void find_indices(XG_X11PROPS *xxp, int xi, int yi, short *ohkl);
extern void get_colorscheme(XG_X11PROPS *xxg);
extern void getbitmapv(XG_X11PROPS *xxg);
extern void help_cb(Widget widget, char *tag);
extern void hkl_axes(XG_X11PROPS *olayp, int ismag);
extern void index_booleans(Widget widget, char *tag);
extern void init_inten(XG_X11PROPS *xxg);
extern void make_buttons(void);
extern void make_textports(XG_X11PROPS *xxg);
extern void mkuscatt(void);
extern void pwriteind(void);
extern void redo_maincolors(void);
extern void redraw(void);
extern void replace_string(Widget wid, char *newstr, XmStringCharSet chaset);
extern void show_iscales(XG_X11PROPS *xxg);
extern void show_region(XG_X11PROPS *xxg);
extern void show_table(XG_X11PROPS *xxp);
extern void showmref(Widget widget);
extern void showuref(Widget widget);
extern void slot_color(int slotno);
extern void start_colors(void);
extern void start_dcolors(XG_X11PROPS *xxg, Colormap *tma);
extern void valtocolors(XG_X11PROPS *xxg);
extern char *dextract_first(XmString cs);
extern char *extract_first_xms_segment(XmString cs);
extern char *gdisnam(int dispno);
extern char *varlabel(int varno, int isdepend);
extern Widget setupfrmscale(XG_X11PROPS *xxg);
extern XG_X11PROPS *get_pm11(void);
extern XG_X11PROPS *managex11p(int opno);
extern XmString strstufset(char *str, char *chaset);
extern XmString strstufstr(char *str, char *chaset_str);

#endif /* XGMDEFS_H */

/* end of xgmdefs.h */
