/* misctools.c -- miscellaneous, seldom-used tools for X-GEN

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	20 Jun 2007: return rather than exit in tnamemain and matcomparmain
	 9 Jun 2007: binucamain written to convert a USPLINE for unbinned
	 data to a USPLINE for binned data or vice versa
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: tinvmain, tpolymain added: tinvmain calls lu_inv
	22 Dec 2006: add tsplinemain and spreaducamain functionalities
	 5 Dec 2006: add tnamemain(): routine to print International Tables
	 numbers and corresponding spacegroup strings
	28 Nov 2006: bravaismain: update calls to showbravais,artocel;
		sie2umain: replace mat2vecs call with rp2vecs call
	26 Jun 2006: include xgpostint.h so we can deal with MERGEDREF stuff
	24 Jun 2006: include xgmodel.h so we can manipulate 3-D profiles here
	16 May 2006: merge {bravais,matcompare,revframe,sie2u,uv_convert}.c
	 6 Oct 2005: housekeeping
	29 Jun 2005: mrcvt(): read and write exactly 256 bytes per DATARUN
	28 Jun 2005: ucacvt(): read and write exactly sizvspline bytes;
	13 May 2005: rename from uconvert to uv_convert
	13 May 2005: change bravais() to a standard X-GEN callable routine
	10 Jan 2005: sie2u: change gamma2 default val from 0.0026 to 0.0010
	 3 Dec 2004: bravais() created out of specialized parts of
	 6 Sep 2004: cosmetic changes to bravais() and matcompare
	 4 Jun 2004: sie2u: fmtoupr() called with an argument (see refutil.c)
	 9 Feb 2004: sie2u: xp->xversion in default_pars() set to 500
	25 Jan 2004: uv_convert: convert most of the static globals
		into locals; fix memory-allocation bug in cvtuca()
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: replace extern global variables with
		static globals and calls to currpname
	27 Nov 2003: sort Bravais choices _after_ finding rhombohedral H
		cell as well as beforehand so it appears in the right order
	 8 Nov 2003: convert lca from vector to pointer in bravais();
		reserve memory for it in allbravais()
	 3 Nov 2003: bravais: created from parts of ./xgenmisc.c and
		../xgrefine/refxtal.c
	19 Sep 2003: sie2u: two ampersands in comparison in readp4p()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: uv_convert: replace access() with xgaccess() calls;
		that enables gunzipping the inputs on the fly
	24 Aug 2003: manipulate and display lattice type & centering
		using routines glattice, gcentering
	25 Feb 1998: initial version of matcompare
	 5 Jul 1997: initial version of sie2u
	16 Mar 1994: a few global variables moved around in revframe
	14 Jan 1994: initial X-GEN version of revframe
	 7 Dec 1993: initial X-GEN version of uv_convert
	29 Apr 1993: uv_convert: new format of calibration spline file used.
	15 Jan 1991: uv_convert: bug fixes and added VMS capabilities
	28 Sep 1990: initial version of uv_convert,
		based on ucacvt.c, uprcvt.c, urfcvt.c, and wmaskcvt.c.
   Purpose: various X-GEN tools, including:
	 XGEN files, converting floating-point & integer formats appropriately.
   Cross references: ../xglib/{xgbravais,xgenio,xgenmisc,setupio,ftable,
		xtal}.c, ../xgrefine/ref*.c
*************************************************************************/
#define	NEED_ERRNO	1
#include	"xgmodel.h"
#include	"xgpostint.h"
#define		TYPICAL_WIDTH	10.
#define		NTYPES	5

 typedef struct {
		char	**uv_realn, **uv_fnames;
		int	uv_fdi, uv_fdo;
		size_t	uv_ffound;
		} UVWORK;

extern int binucamain(int argc, char *argv[]);
extern int bravaismain(int argc, char *argv[]);
extern int matcomparmain(int argc, char *argv[]);
extern int revfmain(int argc, char *argv[]);
extern int sie2umain(int argc, char *argv[]);
extern int spreaducamain(int argc, char *argv[]);
extern int tinvmain(int argc, char *argv[]);
extern int tnamemain(int argc, char *argv[]);
extern int tpolymain(int argc, char *argv[]);
extern int tsplinemain(int argc, char *argv[]);
extern int uconvertmain(int argc, char *argv[]);

extern int lu_inv(double *ain, double *aout, int n);

int tpolymain(int argc, char *argv[])
{ /* This does a detector remapping via polynomial fits. */
	int	i, ier, nitems, nusable;
	char	inl[242];
	double	wvec[5];
	FILE	*fpm;
	VSPLINE	*losp;

	if (NULL == (fpm = xgopen("xyerror", "r")))
	  {
		fprintf(stderr, "Error opening xyerror file\n");
		return EOPENERR;
	  }
	if (NULL == (losp = gspline(5, &ier)))
	  {
		fprintf(stderr, "Error %d inputting spline\n", ier);
		return ier;
	  }
	nusable = 0;
	inv_init(losp->v_lax, losp->v_ldx, 6); /* zero the vectors */
	inv_init(losp->v_lay, losp->v_ldy, 6);
	while (NULL != fgets(inl, 240, fpm))
	     {
		nitems = sscanf(inl, "%lf %lf %lf %lf %lf",
		 &(wvec[0]), &(wvec[2]), &(wvec[1]), &(wvec[3]), &(wvec[4]));
		if (nitems < 5) wvec[4] = 1.;
		polyadd(losp, wvec);
		nusable++;
	     }
	fprintf(stdout, " %6d reflections input\n",nusable);
	if (nusable < 12) return EBAD; /* need >= 12 refls */
	/* now do the actual polynomial remapping and recalc errors */
	i = polymap(losp, stdout, 032);
	fprintf(stdout, "polynomial mapping returned %d\n", i);
	return i;
}

int tinvmain(int argc, char *argv[])
{ /* This tests the matrix inversion facility in lu_inv */
	int		i, ier, j, jq, mdim, mdimsq, ni;
	char		inl[242];
	double		fi, maxoff, maxon, prod, sqfi, fip;
	double		*flp, im[256], iminv[256];
	FILE		*fpm;

	if (NULL == (fpm = xgopen("testmat", "r")))
	  {
		fprintf(stderr, "Error opening testmat file\n");
		return EOPENERR;
	  }
	zeromem((AVOID *)im, 256 * sizeof (double));
	/* read in the matrix elements, up to 16 at a time */
	for (i = 0; NULL != fgets(inl, 240, fpm); )
	   {
		ni = sscanf(inl,
	 " %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
		 &(im[i]), &(im[i+1]), &(im[i+2]), &(im[i+3]), 
		 &(im[i+4]), &(im[i+5]), &(im[i+6]), &(im[i+7]), 
		 &(im[i+8]), &(im[i+9]), &(im[i+10]), &(im[i+11]), 
		 &(im[i+12]), &(im[i+13]), &(im[i+14]), &(im[i+15]));
		i += ni;
		if (i > 255)
		  {
			fprintf(stderr, "Elements beyond 255 ignored\n");
			break;
		  }
	   }
	/* make sure the total number of elements is a perfect square */
	fi = i;	sqfi = sqrt(fi);	fip = sqfi * sqfi;
	mdim = sqfi;	mdimsq = mdim * mdim;
	if (i != mdimsq)
	  {
		fprintf(stderr, " %d elements isn't square: only %d used\n",
		 i, mdimsq);
		return EBAD;
	  }
	fprintf(stdout, " Input matrix:\n");
	for (flp = im, i = 0; i < mdim; i++)
	   {
		for (j = 0; j < mdim; j++, flp++)
			fprintf(stdout, " %7.3f", *flp);
		fprintf(stdout, "\n");
	   }
	if (EOK != (ier = lu_inv(im, iminv, mdim)))
	  {
		fprintf(stdout, "lu_inv failed with error %d\n", ier);
		return EBAD;
	  }
	fprintf(stdout, "Output matrix:\n");
	for (flp = iminv, i = 0; i < mdim; i++)
	   {
		for (j = 0; j < mdim; j++, flp++)
			fprintf(stdout, " %7.3f", *flp);
		fprintf(stdout, "\n");
	   }
	/* check that the product is the identity matrix */
	for (maxon = maxoff = i = 0; i < mdim; i++)
	   {
		for (j = 0; j < mdim; j++)
		   {
			prod = 0.;
			for (jq = 0; jq < mdim; jq++)
			   {
				prod += im[i * mdim +jq] * iminv[jq * mdim +j];
			   }
			if (i != j)
			  {
				if (fabs(prod) > fabs(maxoff))
					maxoff = prod;
			  }
			else {
				if (fabs(prod - 1.) > fabs(maxon))
					maxon = prod - 1.;
			     }
		   }
	   }
	fprintf(stdout,
	 "biggest deviations from identity: %13.5e off, %13.5e on diagonal\n",
		maxoff, maxon);
	return EOK;
}

int binucamain(int argc, char *argv[])
{ /* This reads in a USPLINE associated with unbinned detector data and writes
  it out in a form that is appropriate for a 2x2 binned detector system. */
	char		**dcp, **rflags;
	int		fmnu, ier, j, npoints;
	float		rescale;
	VSPLINE		*lspl;
	CALPOINT	*calp;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*fte;
	FRAMEOFF	*foffp;
	static char	*fnames[] = {"USPLINE", NULL } ;
	static char	ftypes[] = { 3, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	rflags = flagproc(argc, argv, "GS");
	rescale = 2.;
	if ((NULL != rflags) && (NULL != *rflags))
	  {
		for (dcp = rflags; NULL != dcp && NULL != *dcp; dcp++)
		   switch (**dcp)
		     {
			case	'G':	/* increase the size by 2x */
				rescale = 2.;	break;
			case	'S':	default:
				rescale = 0.5;	break;
		     }
	  }
	header(stdout, PROCESSING, fnames, ftypes);
	fprintf(stdout, "Rescaling a USPLINE file by a factor of %8.2f\n",
		rescale);
	if ((NULL == (lspl = gspline(5, &ier))) || (0 != ier))
	  {
		fprintf(stderr, "Error %d in reading input spline\n", ier);
		return EREADERR;
	  }
	npoints = lspl->v_ncol * lspl->v_nrow;
	/* changes to VSPLINE structure itself */
	lspl->v_detw *= rescale;	lspl->v_deth *= rescale;
	lspl->v_maxx = lspl->v_detw + 10;
	lspl->v_maxy = lspl->v_deth + 10;
	lspl->v_axc *= rescale;	lspl->v_ayc *= rescale;
	lspl->v_rxc *= rescale;	lspl->v_ryc *= rescale;
	lspl->v_xrpx *= rescale;	lspl->v_yrpx *= rescale;
	lspl->v_x0px *= rescale;	lspl->v_y0px *= rescale;
	lspl->v_rmsshift *= rescale;	lspl->v_maxshift *= rescale;
	/* This is dangerous to count on, but theoretically correct.
	The v_lax values are already scaled to between -1 and +1 */
	for (j = 0; j < 6; j++)
	   {
		lspl->v_ldx[j] *= rescale; lspl->v_ldy[j] *= rescale;
		lspl->v_dex[j] *= rescale; lspl->v_dey[j] *= rescale;
	   }
	/* rescale the frame offsets if there are any */
	if ((NULL != lspl->v_foffb) && (NULL != (lfmi = ftables(3))))
	  {
		for (fte = lfmi->fi_fmi, fmnu = 0; fte <= lfmi->fi_fme &&
		 fte->fmno > 0 && fte->fmname[0] != '\0'; fte++) fmnu++;
		for (j = 0, foffp = lspl->v_foffb; j < fmnu; j++, foffp++)
		   {
			foffp->fmdx *= rescale;	foffp->fmdy *= rescale;
		   }
	  }
	/* rescale all the ptocx and ptocy values,
	 but not the ctopx and ctopy values */
	for (j = 0, calp = lspl->v_cpt; j <= npoints; j++, calp++)
	   {
		calp->c_ptocx *= rescale;	calp->c_ptocy *= rescale;
	   }
	/* having rescaled everything, we output the spline structure */
	if ((NULL == (lspl = gspline(9, &ier))) || (EOK != ier))
	  {
		fprintf(stderr, "Error %d outputting USPLINE\n", ier);
		return EWRITERR;
	  }
	return EOK;
}

int spreaducamain(int argc, char *argv[])
{
 /* This reads in twelve coefficients and modifies all the 
  fiducial-point positions (u,v) in a VSPLINE file according to
	u' = a + b*x + c*y + d*x*x + e*x*y + f*y*y
	v' = g + h*x + i*y + j*x*x + k*x*y + l*y*y
   where a-l are the twelve coefficients and x=(u/W)-1/2, y=(v/H)-1/2
	are rescalings of the (u,v) values.
	The coefficients are in a file called SPREADUCA. */
	int		i, ier, ix, iy, nitems;
	char		inl[242];
	double		maxshift, ox, oy, rht, rwid;
	double		scx, scy, shift, sumrms, pr[2][6];
	FILE		*fpsp;
	VSPLINE		*lspl;
	CALPOINT	*cpt;

	if (NULL == (fpsp = xgopen("spreaduca", "r")))
	  {
		fprintf(stderr, "Error opening SPREADUCA file\n");
		return EOPENERR;
	  }
	for (i = 0; i < 2; i++)
	   {
		if (NULL == fgets(inl, 240, fpsp))
		  {
			fprintf(stderr,
			 "Failure to read line %d from SPREADUCA\n", i);
			return EREADERR;
		  }
		nitems = sscanf(inl, "%lf %lf %lf %lf %lf %lf",
		 &pr[i][0], &pr[i][1], &pr[i][2], 
		 &pr[i][3], &pr[i][4], &pr[i][5]);
		if (nitems < 6)
		  {
			fprintf(stderr,
			 "Only %d items on line %d of SPREADUCA\n", nitems, i);
			return EREADERR;
		  }
	   }
	if ((NULL == (lspl = gspline(5, &ier))) || (0 != ier))
	  {
		fprintf(stderr, "Error %d in reading input spline\n", ier);
		return EREADERR;
	  }
	rwid = lspl->v_detw;	rht = lspl->v_deth;	maxshift = 0.;
	for (sumrms = iy = 0, cpt = lspl->v_cpt; iy < lspl->v_nrow; iy++)
	   {
		for (ix = 0; ix < lspl->v_ncol; ix++, cpt++)
		   {
			scx = cpt->c_ptocx / rwid - 0.5;
			scy = cpt->c_ptocy / rht - 0.5;
			ox = pr[0][0] + pr[0][1] * scx + pr[0][2] * scy +
			 pr[0][3] * scx * scx + pr[0][4] * scx * scy +
			 pr[0][5] * scy * scy;
			oy = pr[1][0] + pr[1][1] * scx + pr[1][2] * scy +
			 pr[1][3] * scx * scx + pr[1][4] * scx * scy +
			 pr[1][5] * scy * scy;
			cpt->c_ptocx += ox;	cpt->c_ptocy += oy;
			shift = ox * ox + oy * oy;
			if (shift > maxshift) maxshift = shift;
			sumrms += shift;
		   }
	   }
	lspl->v_ntrue = lspl->v_nrow * lspl->v_ncol;
	lspl->v_rmsshift = sqrt(sumrms / (double)lspl->v_ntrue);
	lspl->v_maxshift = sqrt(maxshift);
	return modspline(lspl, stdout, 030);
}

int tsplinemain(int argc, char *argv[])
{ /* Tests the capabilities of the splines */
	int		ier, ix, iy, ngood;
	double		dex, dey, ormse, prevx, prevy, rme, ux, uy;
	float		cux, cuy, oux, ouy;
	VSPLINE		*lspl;
	CALPOINT	*cpt;

	if ((NULL == (lspl = gspline(5, &ier))) || (0 != ier))
	  {
		fprintf(stderr, "Error %d in reading input spline\n", ier);
		return EREADERR;
	  }
	prevx = prevy = 0;
	for (ormse = ngood = iy = 0, cpt = lspl->v_cpt; iy < lspl->v_nrow; iy++)
	   {
		for (ix = 0; ix < lspl->v_ncol; ix++, cpt++)
		   {
			ux = cpt->c_ptocx + 0.25 * lspl->v_xrpx;
			uy = cpt->c_ptocy - 0.25 * lspl->v_yrpx;
			if (0 != pix2cm((float)ux, (float)uy, &cux, &cuy, 0))
			  {
				if (0 != cm2pix(cux, cuy, &oux, &ouy, 0))
				  {
					ngood++;
					dex = oux - ux;	dey = ouy - uy;
					rme = dex * dex + dey * dey;
					ormse += rme; rme = sqrt(rme);
					if (ix) fprintf(stdout,
	 " %3d %3d %7.1f %7.1f %7.3f %7.3f %7.4f %7.4f %7.1f %7.1f %9.5f\n",
				 	 ix, iy, ux, uy, cux, cuy,
					 prevx - cux, prevy - cuy,
					 oux, ouy, rme);
					else fprintf(stdout,
	 " %3d %3d %7.1f %7.1f %7.3f %7.3f %7.1f %7.1f %7.1f %7.1f %9.5f\n",
				 	 ix, iy, ux, uy, cux, cuy,
					 0., 0., oux, ouy, rme);
				  }
				prevx = cux;	prevy = cuy;
			  }
		   }
	   }
	ormse = (ngood) ? sqrt(ormse / ((double)ngood)) : 0.;
	fprintf(stdout, " %6d successful inversions: rmserr = %13.4e\n",
		ngood, ormse);
	return EOK;
}

int tnamemain(int argc, char *argv[])
{
	short	intl;

	for (intl = 0; intl < 230; intl++)
		printf("spacegname(%3d) = %s\n",
			intl, spacegname(intl));
	return EOK;
}

/***************************************
	bravais: determine and display Bravais lattice values
	based on the contents of a UPARAMS file
***************************************/
int bravaismain(int argc, char *argv[])
{ /* The mainline for calculating and displaying the Bravais choices */
	register int		i, j;
	int			chara, choicel, nbr, nlatts;
	char			iline[30];
	float			cel[9], pareal[3][3];
	double			lareal[3][3];
	LATTICESCORE		*lcap;
	static REFPARS		rx;
	static LATTICESCORE	*lca;
	static char		*fnames[] = { "uparams", NULL } ;
	static char		ftypes[] = { 1, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* open and read the input files */
	if (EOK != (i = refparin(&rx, &nbr))) return i;
			/* summarize what we're doing to stdout */
	header(stdout, PROCESSING, fnames, ftypes);
	/* extract the unit cell matrix in the appropriate format. */
	if (rx.xlambda < FUZZ) rx.xlambda = 1.5418;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		lareal[i][j] = pareal[i][j] = rx.xareal[i][j] * rx.xlambda;
	pr3m(stdout, pareal, "input real-space unit cell matrix");
	/* compute the relevant scores and unit cells */
	if (NULL == (lca = allbravais(lareal, &nlatts))) return EALLOC;
	if (nlatts < 15) return EBADVALUE;
	showbravais(stdout, rx.xareal, rx.xstart);
	fprintf(stdout, "Enter index value of desired lattice (0-15): ");
	while (NULL == fgets(iline, 25, stdin)) ;
	sscanf(iline, "%d", &choicel);
	if ((0 <= choicel) && (choicel <= 15))
	  {
		for (i = 0, lcap = lca; i < 45; i++, lcap++)
		   if (choicel == lcap->lc_choice)
		     {
			artocel(rx.xstart, lcap->lc_cel, cel);
			chara = lcap->lc_latchar;
			printf(
 "%s %s lattice chosen; %7.2f %7.2f %7.2f %6.1f %6.1f %6.1f\n",
			 glattice(chara), gcentering(chara),
			 cel[0], cel[1], cel[2], cel[3],cel[4],cel[5]);
			return choicel;
		     }
	  }
	if ((0 > choicel) || (choicel > 15)) choicel = -1;
	printf("resulting numerical choice is %d\n", (choicel));
	globalkill();	return EOK;
}

/* end of bravais functionality */

/*******************************************
	matcompare: determine rotation matrix between 2 UPARAMS unit cells
*******************************************/

int matcomparmain(int argc, char *argv[])
{ /*	This reads a pair of input UPARAMS files and compares them.
	It does so by multiplying the inverse of one `areal' matrix by
	the other `areal' matrix and analyzing the product. For the
	moment I'll do this by calculating the "cell dimensions"
	(really eigenvalues, I hope)
	of the product and the "orientation matrix" associated with it.  */
	int	i, j, k;
	char	*cpname, *fn0, *fn1;
	double	dotp, len[3], lang[3];
	float	ar1i[3][3], resmat[3][3];
	FILE	*fpi, *fplog;
	REFPARS	rfpa0, rfpa1;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	cpname = xgprogname(NULL, 0);
	if (argc < 3)
	  {
		fprintf(stderr, "Usage: %s file0.upr file1.upr\n", cpname);
		return EOK;
	  }
	if (NULL == (fn0 = trname(argv[1])))
	  {
		fprintf(stderr, "%s: cannot recognize %s as input\n",
			cpname, argv[1]);
		return EBAD;
	  }
	if (NULL == (fpi = xgopen(fn0, "r")))
	  {
		fprintf(stderr, "%s: cannot open %s as input\n",
			cpname, fn0);
		return EOPENERR;
	  }
	if (256 < xgread(&rfpa0, XG_REFPARS, fpi))
	  {
		fprintf(stderr, "%s: cannot read %s as input\n",
			cpname, fn0);
		return EREADERR;
	  }
	pr3m(stdout, rfpa0.xareal, "areal for matrix 0");
	(void)fclose(fpi);	fpi = NULL;
	if (NULL == (fn1 = trname(argv[2])))
	  {
		fprintf(stderr, "%s: cannot recognize %s as input\n",
			cpname, argv[2]);
		return EBAD;
	  }
	if (NULL == (fpi = xgopen(fn1, "r")))
	  {
		fprintf(stderr, "%s: cannot open %s as input\n",
			cpname, fn1);
		return EOPENERR;
	  }
	if (256 < xgread(&rfpa1, XG_REFPARS, fpi))
	  {
		fprintf(stderr, "%s: cannot read %s as input\n",
			cpname, fn1);
		return EREADERR;
	  }
	pr3m(stdout, rfpa1.xareal, "areal for matrix 1");
	(void)fclose(fpi);	fpi = NULL;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	if (0 != inv3(rfpa1.xareal, ar1i))
	  {
		fprintf(stderr,
			"%s: cannot invert second cell-matrix\n", cpname);
		return EBADVALUE;
	  }
	mm3(resmat, ar1i, rfpa0.xareal);
	pr3m(stdout, resmat, "product matrix");
	pr3m(fplog, resmat, "product matrix");
	for (i = 3; i--; )	/* lengths: */
	   {
		for (dotp = 0, j = 3; j--; ) /* length^2 = norm^2 of row */
			dotp += resmat[i][j] * resmat[i][j];
		len[i] = sqrt(dotp);
	   }
	for (i = 0; i < 3; i++)	/* angles: from dot products */
	   {
		j = (i == 2 ? 0 : (i+1));	k = 3 - i - j;
		dotp = (resmat[j][0] * resmat[k][0] +
		 resmat[j][1] * resmat[k][1] + resmat[j][2] * resmat[k][2]);
		lang[i] = (fabs(dotp) > DFUZZ) ?
		  acos(dotp / (len[j] * len[k])) / RADC : 90.;
	   }
	fprintf(stdout,
	 "lengths, angles: %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n",
	 len[0], len[1], len[2], lang[0], lang[1], lang[2]);
	fprintf(fplog ,
	 "lengths, angles: %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f\n",
	 len[0], len[1], len[2], lang[0], lang[1], lang[2]);
	/* we'll save the angles for the next version */
	if (stdout != fplog) (void)fclose(fplog);
	fplog = NULL;	globalkill();	return EOK;
}

/* end of matcompare functionality */

/***********************************
	revframe: reverse the order of the file names in a FRAMES index
	It reads FRAMES & writes it back out with the order of names reversed,
	but with the order of index numbers left unchanged.
	Thus if the original file is
		    1/usr/xengen/garbagea/garbagea.0001
		    2/usr/xengen/garbagea/garbagea.0002
			...
		  479/usr/xengen/garbagea/garbagea.0479
		  480/usr/xengen/garbagea/garbagea.0480
	then the output file will be
		    1/usr/xengen/garbagea/garbagea.0480
		    2/usr/xengen/garbagea/garbagea.0479
			...
		  479/usr/xengen/garbagea/garbagea.0001
		  480/usr/xengen/garbagea/garbagea.0002
************************************/

int revfmain(int argc, char *argv[])
{	/* this reads in the `frames' file and writes it back out
	with the first frame labeled as the last and vice versa.
	The frames must be numbered contiguously for this to work. */

	int	firstframe, frame, lastframe, nframes, oframe;
	char	filename[80], rvline[90];
	FILE	*fpfin, *fpfout;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpfin = xgopen("frames", "r")),
		"open input frames file")) return EOPENERR;
	if (NULL == fgets(rvline, 80, fpfin))
	  {
		(void)fclose(fpfin);	return EOK;
	  }
	(void)sscanf(rvline, " %d", &firstframe);
	for (nframes = 1; NULL != fgets(rvline, 80, fpfin); ) nframes++;
	if (verror(1 + fseek(fpfin, 0L, 0), "rewind index file"))
	  {
		(void)fclose(fpfin);	return ESEEKERR;
	  }
	lastframe = nframes + firstframe - 1;
	if (qerror((AVOID *)(fpfout = xgopen("frames", "w")),
		"open output frames file")) return EOPENERR;
	for (nframes = 0; NULL != fgets(rvline, 80, fpfin); nframes++)
	   {
		(void)sscanf(rvline, " %d%s", &frame, filename);
		oframe = lastframe - nframes;
		fprintf(fpfout, " %4d%s\n", oframe, filename);
	   }
	(void)fclose(fpfin);	(void)fclose(fpfout);	return EOK;
}

/* end of revframe functionality */

/**********************************
	sie2u: convert Siemens .p4p format to X-GEN UPARAMS format
	 and (if appropriate) an X-GEN CENTROIDS file
***********************************/

static void default_pars(REFPARS *xp)
{ /* This sticks default parameters in a REFPARS structure */
	int		i;
	static char	defstr[] = "YYYNNNYYYNYYNNYYYNNN";
	static char	adir0[] = " X  ";
	static char	adir1[] = " Y  ";
	static char	adir2[] = " Z  ";

	xp->xlambda = DEF_LAMBDA;	xp->xx0 = xp->xy0 = 0.;
	xp->xplated = 10;	xp->xtilt = 0.;
	xp->xisys = xp->xispgp = 1;	xp->xcmd = 'R';
	xp->xfrmsize = 0.25 * RADC;
	xp->xgam[0] = xp->xgam[1] = 1. / xp->xfrmsize;	xp->xgam[2] = 0.0010;
	xp->xswing = 0.;
	xp->xerrind = 0.2;	xp->xerrx = xp->xerry = xp->xerrphi = 5.;
	xp->xresmin = 99.;	xp->xresmax = 0.;
	for (i = 0; i < 20; i++) xp->xisref[i] = defstr[i];
	xp->xdelchi = xp->xdelom = xp->xdelchi = 0.;
	for (i = 0; i < 3; i++)
	   {
		xp->xabc[i] = 100.;	xp->xang[i] = 90.;
		xp->xrec[i][i] = 0.01;	xp->xstart[i] = 0.;
	   }
	for (i = 0; i < 5; i++)
	   {
		xp->xdir[0][i] = adir0[i];
		xp->xdir[1][i] = adir1[i];
		xp->xdir[2][i] = adir2[i];
	   }
	xp->xrunno = xp->xchamno = xp->xtemp = 0;
	xp->xmoveax = 2;	xp->xdtype = 4099;
	xp->xversion = 500;	xp->xdetwid = xp->xdetht = 512;
}

static int readp4p(REFPARS *xp, FILE **fpop)
{ /*	This creates and fills a reciprocal-space unit cell matrix
	by reading a p4p file. It returns 0 if successful, 1 if not. */
	char		inl[258], *cp;
	int		failco, i, ini0, ncentroids, nitems, retval, rowsfound;
	double		approxwid, approxcen, iq0, iq1, iq2, iq3, iq4, iq5;
	double		xn0, xn1, xn2;
	FILE		*fppi;
	REFPARS		*rpfm;
	static TROID	trb;

	default_pars(xp);	/* establish semi-reasonable defaults */
	rpfm = xp + 1;	/* put image parameters in xp + 1 */
	if (EOK == (i = fmtoupr((xglong)0, rpfm)))
	  {
		cpymem((void *)xp, (void *)(xp + 1), sizeof (REFPARS));
		printf(
 " from images: plated, swing, stepsize = %8.2f, %8.2f, %8.2f\n",
		 xp->xplated, xp->xswing / RADC, xp->xfrmsize / RADC);
	  }
	else printf(
 " Warning: no parameter information derivable from images; error %d\n", i);
	if (NULL == (fppi = xgopen("p4p", "r")))
	  {
		fprintf(stderr, "Error opening p4p input\n");
		return EOPENERR;
	  }
	retval = EREADERR; /* ... until we've read what we need */
	rowsfound = failco = ncentroids = 0;
	while (NULL != fgets(inl, 254, fppi))
	     {
		for (cp = inl; ('\0' != *cp && '\n' != *cp); cp++)
		   if (('a' <= *cp) && (*cp <= 'z'))
			*cp -= 'a' - 'A';
		if (!strncmp(inl, "ORT1", 4))
		  {
			nitems = sscanf(&(inl[5]), " %le %le %le ",
			 &iq0, &iq1, &iq2);
			if (nitems >= 3)
			  {
				xp->xrec[0][0] = iq0; xp->xrec[1][0] = iq1;
				xp->xrec[2][0] = -iq2; 
			  }
			rowsfound |= 1;
		  }
		else if (!strncmp(inl, "ORT2", 4))
		  {
			nitems = sscanf(&(inl[5]), " %le %le %le ",
			 &iq0, &iq1, &iq2);
			if (nitems >= 3)
			  {	
				xp->xrec[0][1] = iq0; xp->xrec[1][1] = iq1;
				xp->xrec[2][1] = -iq2; 
			  }
			rowsfound |= 2;
		  }
		else if (!strncmp(inl, "ORT3", 4))
		  {
			nitems = sscanf(&(inl[5]), " %le %le %le ",
			 &iq0, &iq1, &iq2);
			if (nitems >= 3)
			  {
				xp->xrec[0][2] = -iq0; xp->xrec[1][2] = -iq1;
				xp->xrec[2][2] = iq2; 
			  }
			rowsfound |= 4;
		  }
		else if ((!strncmp(inl, "CELL", 4)) && strncmp(inl,
		 "CELLSD", 6))
		  {
			nitems = sscanf(&(inl[8]), " %lf %lf %lf %lf %lf %lf",
				&iq0, &iq1, &iq2, &iq3, &iq4, &iq5);
			if (nitems >= 3)
			  {
				xp->xabc[0] = iq0; xp->xabc[1] = iq1;
				xp->xabc[2] = iq2;
				if (nitems >= 6)
				  {
					xp->xang[0] = iq3; xp->xang[1] = iq4;
					xp->xang[2] = iq5;
				  }
			  }
		  }
		else if (!strncmp(inl, "SOURCE", 6))
		  {
			nitems = sscanf(&(inl[11]), "%lf ", &iq0);
			xp->xlambda = iq0;
		  }
		else if (!strncmp(inl, "ADPAR", 5))
		  {
			nitems = sscanf(&(inl[6]), "%le %le %le %d",
			 &iq0, &iq1, &iq2, &ini0);
			if (nitems >= 4)
			  {
				xp->xplated = iq2;
				if (ini0 < 64) ini0 = 64;
				xp->xdetwid = xp->xdetht = ini0;
				approxwid = TYPICAL_WIDTH / (double)ini0;
				approxcen = (double)(ini0 / 2);
				xp->xx0 = -(iq0 - approxcen) * approxwid;
				xp->xy0 = -(iq1 - approxcen) * approxwid;
			  }
		  }
		else if (!strncmp(inl, "REF05 ACHS", 10))
		  {
		    if ((NULL == *fpop) && !failco)
		      {
			if (NULL == (*fpop = xgopen("centroids", "w")))
			  {
				fprintf(stderr,
				 "Warning: failure to open centroids\n");
				failco = 1;
			  }
			trb.nobk = 5;
			trb.obparts[0] = trb.obparts[4] = 0.05;
			trb.obparts[1] = trb.obparts[3] = 0.25;
			trb.obparts[2] = 0.4;
		     }
		   if (NULL != *fpop)
		     {
			nitems = sscanf(&(inl[31]), " %lf %lf %lf %lf %lf %lf",
				&iq0, &iq1, &iq2, &iq3, &iq4, &iq5);
			if (nitems >= 6)
			  {
				trb.xto = iq3;	trb.yto = iq4;
				trb.frame0 = iq0; trb.totinten = iq5;
				trb.obkmin = iq0 - 1.501;
				if (0 >= xgwrite(&trb, XG_TROID, *fpop))
				  {
					fprintf(stderr,
					"Warning: reflection not output\n");
				  }
				else ncentroids++;
			  }
		     }
		  }
	     }
	(void)fclose(fppi); fppi =  NULL;
	if (NULL != *fpop)
	  {
		(void)fclose(*fpop);
		printf(" %6d centroids found and output\n", ncentroids);
	  }
	*fpop = NULL;
	if (rowsfound == 07)
	  {	printf("all three rows found\n");
		printf("determinant of recip sp matrix = %12.5e\n",
			deter(xp->xrec));
		for (xn2 = xn1 = xn0 = 0., i = 0; i < 3; i++)
		   {
			xn2 += xp->xrec[2][i] * xp->xrec[2][i];
			xn1 += xp->xrec[1][i] * xp->xrec[1][i];
			xn0 += xp->xrec[0][i] * xp->xrec[0][i];
		   }
		xn2 = sqrt(xn2); xn1 = sqrt(xn1); xn0 = sqrt(xn0);
		iq0 = acos(dotprod(xp->xrec[1], xp->xrec[2], 3) /
		 (xn1 * xn2)) / RADC;
		iq1 = acos(dotprod(xp->xrec[2], xp->xrec[0], 3) /
		 (xn1 * xn2)) / RADC;
		iq2 = acos(dotprod(xp->xrec[0], xp->xrec[1], 3) /
		 (xn0 * xn0)) / RADC;
		printf("angles between columns: %12.5e %12.5e, %12.5e\n",
		 iq0, iq1, iq2);
		retval = EOK;
	  }
	else {	printf("fewer than three rows found\n"); retval = EBADVALUE; }
	return retval;
}

int sie2umain(int argc, char *argv[])
{ /*	In this mainline, we handle the arguments and then process each file */
	int		i, j, retval;
	static char	*flist[] = {"p4p", "uparams", NULL};
	static char	ftypes[] = { 1, 2, 0};
	static FILE	*fpo;
	static REFPARS	*xp;
	static char	xpmsg[] = "xp in sie2u";

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	header(stdout, PROCESSING, flist, ftypes);
	if (NULL == (xp = (REFPARS *)callop(2, sizeof (REFPARS), xpmsg)))
		return EALLOC;
	if (EOK != (retval = readp4p(xp, &fpo)))
	  {
		fprintf(stderr, "Failure %d in readp4p\n", retval);
		return retval;
	  }
	if (EOK != (i = inv3(xp->xrec, xp->xareal)))
	  {
		fprintf(stderr,"Failure %d in inverting recip-sp matrix\n", i);
		return EBADVALUE;
	  }
	if (FUZZ >= xp->xlambda) xp->xlambda = DEF_LAMBDA;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xp->xareal[i][j] /= xp->xlambda;
	if (EOK != (retval = rp2vecs(xp)))
	  {
		fprintf(stderr, "Failure %d in rp2vecs\n", retval);
		return retval;
	  }
	if (EOK != (i = refparout(xp))) return i;
	memzap((AVOID **)xp, xpmsg);	globalkill();	return EOK;
}

/* end of sie2u functionality */

/**********************************
	uv_convert: convert XGEN files from one format to another.
	On a Unix machine, this converts files from VMS format to Unix format.
	On a VMS machine, it converts files from Unix format to VMS format.
	In general this involves:
	 converting VMS floating-point numbers to Unix floating-point numbers
		or Unix floating-point numbers to VMS floating-point numbers;
	 swapping words and bytes in 4-byte integers;
	 swapping bytes in 2-byte integers.
	It is capable of handling USPLINE, UPARAMS, WMASK, and UREFLS files.
	Currently this only works on non-DEC Unix machines; when I get some more
	time on a Decstation 3100 or similar machine I'll generalize some more.
*************************************************************************/
static int cvread(AVOID *base, size_t nbytes, char *datatype, UVWORK *uvp)
{ /*	Read routine with built-in error message and (in the VMS version)
	the capability of reading past block-boundaries. */
	int	nread;

#ifdef VMS
	int		nthis, readtry;
	unsigned	nleft;
	char		*root;

	errno = nread = 0;
	for (readtry = 0; (readtry < 100) && (nbytes > nread); readtry++)
	   {
		nleft = nbytes - nread;	root = ((char *)base) + nread;
		if ((nthis = read(uvp->uv_fdi, root, nleft)) < 0)
		  {
			fprintf(stderr,
		 "\tError reading %3d bytes of %s data: errno = %d\n",
			 (int)nleft, datatype, errno);
			break;	 /* give up on error */
		  }
		else {	/* ... remainder in next record(s): */
			errno = 0; nread += nthis;
		     }
	   }
	if ((readtry >= 100) || (nread < nbytes)) fprintf(stderr,
	 "\tTried to read %3d bytes of %s data, got %3d; errno = %d\n",
		(int)nbytes, datatype, nread, errno);
#else
	errno = 0;
	if ((int)(nbytes) != (nread = read(uvp->uv_fdi, base, nbytes)))
	 fprintf(stderr,
	 "\tTried to read %3d bytes of %s data, got %3d; errno = %d\n",
		(int)nbytes, datatype, nread, errno);
#endif
	return nread;
}

static int cvwrite(AVOID *base, size_t nbytes, char *datatype, UVWORK *uvp)
{ /*	Write routine with built-in error message */
	int	nwrite;

	errno = 0;
	if ((int)(nbytes) != (nwrite = write(uvp->uv_fdo, base, nbytes)))
	 fprintf(stderr,
	 "\tTried to write %3d bytes of %s data, got %3d; errno = %d\n",
		(int)nbytes, datatype, nwrite, errno);
	return nwrite;
}

static void uprcvt(UVWORK *uvp)
{ /*	This effects a VMS-to-Unix or Unix-to VMS conversion for a UPARAMS
	file.  It works on both the old and the new formats of UPARAMS files.*/
	u_char	*cp, *endfloat;
	off_t	ntoread;
	REFPARS	rp;

		/* see which size (232 or 256 bytes) we had to start */
	if (-1 == (ntoread = lseek(uvp->uv_fdi, (off_t)0L, 2))) return;
		/* ... and then rewind it */
	if (-1 == lseek(uvp->uv_fdi, (off_t)0L, 0)) return;
	if (-1 == cvread((AVOID *)(&rp), (size_t)ntoread, "UPARAMS", uvp))
		return;
	endfloat = (u_char *)(&(rp.xresmax));
	for (cp = (u_char *)(&rp.xlambda); cp <= endfloat; cp += sizeof (float))
		float_cvt(cp);
	endfloat = (u_char *)(&(rp.xdelom));
	for (cp = (u_char *)(&(rp.xdelchi)); cp <= endfloat;
		cp += sizeof (float)) float_cvt(cp);
	rp.xisys = sswab(rp.xisys);
	rp.xispgp = sswab(rp.xispgp);
	if (ntoread == sizeof (REFPARS))	/* longer format */
	  {
		endfloat = (u_char *)&(rp.xmoveax);
		for (cp = (u_char *)&(rp.xstart[0]); cp < endfloat;
			cp += sizeof (float)) float_cvt(cp);
	  }
	(void)cvwrite((AVOID *)(&rp), ntoread, "UPARAMS", uvp);
	return;
}

static void ucacvt(UVWORK *uvp)
{ /*	This effects a VMS-to-Unix conversion for a USPLINE file */
	int		i;
	u_char		*cp, *endfloat;
	short		*shp;
	size_t		ntoread;
	VSPLINE		*sgsp;
	CALPOINT	*calp;
	static size_t	sizvspline = 512;

	if (NULL == (sgsp = (VSPLINE *)callop(1, sizeof (VSPLINE),
	 "sgsp in ucacvt"))) return;
	if (-1 == cvread((AVOID *)sgsp, sizvspline, "USPLINE", uvp)) return;
	for (shp = &(sgsp->v_detw); shp <= &(sgsp->v_nreg); shp++)
		*shp = sswab(*shp);
	endfloat = (u_char *)(&(sgsp->v_maxy));
	for (cp = (u_char *)(&(sgsp->v_axc)); cp <= endfloat; cp++)
		float_cvt(cp);
	(void)cvwrite((AVOID *)sgsp, sizvspline, "USPLINE", uvp);
	ntoread = sgsp->v_nrow * sgsp->v_ncol;
	if (NULL == (sgsp->v_cpt = (CALPOINT *)callop(ntoread,
	 sizeof (CALPOINT), "sgsp->v_cpt in ucacvt")))
	  {
		memzap((AVOID **)&sgsp, "sgsp in ucacvt");	return;
	  }
	if (-1 == cvread((AVOID *)(sgsp->v_cpt),
	 ntoread * sizeof (CALPOINT), "USPLINE", uvp)) return;
	for (calp = sgsp->v_cpt, i = 0; i < (int)ntoread; i++, calp++)
	   {
		endfloat = (u_char *)(&(calp->c_frfu[2]));
		for (cp = (u_char *)(&(calp->c_ptocx)); cp <= endfloat; cp++)
			float_cvt(cp);
	   }
	(void)cvwrite((AVOID *)(sgsp->v_cpt),
	 ntoread * sizeof (CALPOINT), "USPLINE", uvp);
	memzap((AVOID **)&(sgsp->v_cpt), "sgsp->v_cpt in ucacvt");
	memzap((AVOID **)&sgsp, "sgsp in ucacvt");
	return;
}

static void urfcvt(UVWORK *uvp)
{ /*	This effects conversion of a VMS UREFLS file to a Unix UREFLS file */
	int		nref;
	u_char		*cp, *cpe;
	size_t		ntoread, iswritten;
	short		*sp, *spe;
	OBSREF		urf;

	ntoread = sizeof (OBSREF);
	for (iswritten = nref = 0; ((int)ntoread ==
	 cvread((AVOID *)&urf, ntoread, "OBSREF", uvp)); nref++)
	   {
		spe = (short *)&(urf.obs_region);
		for (sp = (short *)&(urf.obs_hkl[0]);
		 sp <= spe; sp++) *sp = sswab(*sp);
		cpe = (u_char *)(&(urf.obs_omob));
		for (cp = (u_char *)(&(urf.obs_ii)); cp <= cpe;
			cp += sizeof (float)) float_cvt(cp);
		spe = (short *)&(urf.obs_reflno);
		for (sp = (short *)&(urf.obs_gof); sp <= spe; sp++)
			*sp = sswab(*sp);
		spe = &(urf.obs_flags);
		for (sp = (short *)&(urf.obs_frmint[0]); sp <= spe; sp++)
			*sp = sswab(*sp);
		if (ntoread != (iswritten = cvwrite((AVOID *)&urf,
			ntoread, "UREFLS", uvp))) return;
	   }
	printf("\t\t%6d reflections translated\n", nref);
	if (!nref) fprintf(stderr, 
 "\tWarning: tried to read %3d bytes of reflection data, got %3d; errno = %d\n",
		(int)ntoread, (int)iswritten, errno);
}

static void wmaskcvt(UVWORK *uvp)
{ /*	This effects conversion of a VMS WMASK file to a Unix WMASK file */
	register int	ifm, region;
	size_t		ntoread, ntowrite;
	MASK_3D		wm;
	PHIMASK		*phip;
	static char	modmsg[] = "model profile";

	ntoread = sizeof (PHIMASK) - 3;
	ntowrite = sizeof (MASK_3D);
	for (region = 0; region < NREGIONS; region++)
	   {
		for (ifm = 0, phip = &(wm.phielt[0]); ifm < PFRAMES;
		  ifm++, phip++)
		   {
			(void)cvread((AVOID *)phip, ntoread, modmsg, uvp);
			float_cvt((u_char *)(&(phip->isqsum)));
			float_cvt((u_char *)(&(phip->maxf)));
			phip->mnobs = lswawb(phip->mnobs, 1);
		   }
		if ((int)ntowrite != cvwrite((AVOID *)(&wm), ntowrite,
			modmsg, uvp)) return;
	   }
	(void)close(uvp->uv_fdi); (void)close(uvp->uv_fdo);	return;
}

static void cvt_xhd(CRYSTAL *xhp)
{ /*	This converts a CRYSTAL structure from VMS format to Unix format */
	u_char		*cp, *cpe;

	xhp->xt_intl = uswab((u_short)xhp->xt_intl);
	xhp->xt_isys = uswab((u_short)xhp->xt_isys);
	cpe = (u_char *)(&(xhp->xt_ssqmax));
	for (cp = (u_char *)(&(xhp->xt_ar[0])); cp <= cpe;
		cp += sizeof (float)) float_cvt(cp);
	xhp->xt_ninteg = lswawb(xhp->xt_ninteg, 1);
	xhp->xt_nrefls = lswawb(xhp->xt_nrefls, 1);
	xhp->xt_nmerged = lswawb(xhp->xt_nmerged, 1);
	xhp->xt_nruns = sswab(xhp->xt_nruns);
	xhp->xt_nshifts = sswab(xhp->xt_nshifts);
}

static void cvt_rhd(DATARUN *rhp)
{ /*	This converts a DATARUN header from VMS format to Unix format */
	u_char		*cp, *cpe;
	short		*sp, *spe;

	spe = (short *)(&(rhp->run_height));
	for (sp = (short *)(&(rhp->run_id)); sp <= spe; sp++)
		*sp = sswab(*sp);
	rhp->run_ninteg = lswawb(rhp->run_ninteg, 1);
	rhp->run_nexam = lswawb(rhp->run_nexam, 1);
	cpe = (u_char *)(&(rhp->run_ssqmax));
	for (cp = (u_char *)(&(rhp->run_lambda)); cp <= cpe;
		cp += sizeof (float)) float_cvt(cp);
}

static void cvt_mr(MERGEDREF *mrp)
{ /*	This converts the first 16 bytes of a MERGEDREF structure */
	mrp->mr_nob = uswab(mrp->mr_nob);
	mrp->mr_ibar = uswab(mrp->mr_ibar);
	mrp->mr_isig = uswab(mrp->mr_isig);
	float_cvt((u_char *)(&(mrp->mr_stl)));
}

static void cvt_abbrob(ABBROB *aobp)
{ /*	This converts an abbreviated-observation structure from VMS to Unix */
	aobp->abo_shift = sswab(aobp->abo_shift);
	aobp->abo_omob = sswab(aobp->abo_omob);
	float_cvt((u_char *)(&(aobp->abo_ii)));
	float_cvt((u_char *)(&(aobp->abo_sigma)));
}

static void mrcvt(UVWORK *uvp)
{ /*	This converts a MULTIREF from VMS to Unix or vice versa */
	int		iob, irun, nrefl, nob, nobs;
	size_t		ntoread, nreadob;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;
	ABBROB		*aobp;

/* Read the CRYSTAL header from the input multiref, convert it, and write it */
	ntoread = sizeof (CRYSTAL);
	(void)cvread((AVOID *)(&xhd), ntoread, "multiref crystal header", uvp);
	cvt_xhd(&xhd);
	if ((int)ntoread != cvwrite((AVOID *)(&xhd), ntoread,
		"multiref crystal header", uvp)) return;
	/* Read all the DATARUN headers, convert them, and write them out */
	for (irun = 0; irun < xhd.xt_nruns; irun++)
	   {
		ntoread = 256;
		(void)cvread((AVOID *)(&rhd), ntoread,
			"multiref datarun header", uvp);
		cvt_rhd(&rhd);
		if ((int)ntoread != cvwrite((AVOID *)(&rhd), ntoread,
			"multiref datarun header", uvp)) return;
	  }
	printf(
	 "\t 1 CRYSTAL header and %6d DATARUN headers read and written\n",
		xhd.xt_nruns);
/* Read through the reflections till we don't have any more.
	Convert and write each reflection, read in the observations, 
	convert them, and write them out */
	ntoread = REFBYTES;	nreadob = sizeof (ABBROB);
	for (nrefl = nob = 0; ((int)ntoread == read(uvp->uv_fdi, (void *)(&mr),
	 ntoread)); nrefl++, nob += nobs)
	    {
		cvt_mr(&mr);	nobs = mr.mr_nob;
		if ((int)ntoread != cvwrite((AVOID *)(&mr), ntoread,
			"multiref reflection", uvp)) return;
		for (iob = 0, aobp = &(mr.mr_ob[0]); iob < nobs; iob++, aobp++)
		   {
			if ((int)nreadob != cvread((AVOID *)aobp, nreadob,
				"multiref observation", uvp)) return;
			cvt_abbrob(aobp);
			if ((int)nreadob != cvwrite((AVOID *)aobp,
			 nreadob, "multiref observation", uvp)) return;
		   }
	     }
	printf(
	 "\t%6d observations of %6d reflections read and written\n",
		nob, nrefl);
}

static int chk_setup(int argc, char *argv[], UVWORK *uvp)
{ /*	This looks to see which files the user wants converted and
	whether they're really there. */
	int		i, thislen;
	char		isfound, *cp, conv[NTYPES], thisreal[80];
	static char	ftypes[] = { 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 0 };
	static char	*logn[NTYPES] =
		{ "USPLINE", "UPARAMS", "UREFLS", "WMASK", "MULTIREF" } ;
	static char	*logo[NTYPES] =
		{ "USPOUT", "UPAROUT", "UREFOUT", "WMASKOUT", "MREFOUT" } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* see which kinds of files the user wants converted */
	for (i = 0; i < NTYPES; i++) conv[i] = 0;
	for (argc--, argv++ ; argc--; argv++)
	   {
		isfound = 0;
		for (cp = *argv; '\0' != *cp; cp++)
			if ('a' <= *cp && *cp <= 'z') *cp -= ('a' - 'A');
		for (i = 0; i < NTYPES; i++)
		   if (!strcmp(*argv, logn[i]))
		     {
			conv[i] = isfound = 1;	break;
		     }
		if (!isfound) fprintf(stderr, "Argument %s ignored\n", *argv);
	   }
	for (i = uvp->uv_ffound = 0; i < NTYPES; i++)
		if (conv[i]) uvp->uv_ffound += 2;
	if (NULL == (AVOID *)(uvp->uv_realn = (char **)callop(2 *
	 uvp->uv_ffound, sizeof (char *), "realn in chk_setup"))) return EALLOC;
	uvp->uv_fnames = uvp->uv_realn + uvp->uv_ffound;
	/* make sure the requested files are really there */
	for (i = uvp->uv_ffound = 0; i < NTYPES; i++)
	   {
		if (conv[i])
		  {
			strcpy(thisreal, trname(logn[i]));
			if (0 != xgaccess(thisreal, 4, 1))
			  {
				fprintf(stderr,
	 "Warning: Filetype %s unreadable; no conversion done\n", logn[i]);
				conv[i] = 0;
			   }
			else
			 {
				thislen = strlen(thisreal) + 2;
				if (NULL == (AVOID *)
				 (uvp->uv_realn[uvp->uv_ffound] =
				 (char *)mallop(thislen,
				 "realn[ff] in chk_setup"))) return EALLOC;
				strcpy(uvp->uv_realn[uvp->uv_ffound],thisreal);
				if (NULL == (AVOID *)
				 (uvp->uv_fnames[uvp->uv_ffound] =
				 (char *)mallop(strlen(logn[i]) + 2,
				 "fnames[ff] in chk_setup"))) return EALLOC;
				strcpy(uvp->uv_fnames[(uvp->uv_ffound)++],
					logn[i]);
				strcpy(thisreal, trname(logo[i]));
				if (NULL == (AVOID *)(
				 uvp->uv_realn[uvp->uv_ffound] =
				 (char *)mallop(thislen,
				 "realn[ff + 1] in chk_setup"))) return EALLOC;
				strcpy(uvp->uv_realn[uvp->uv_ffound],thisreal);
				if (NULL == (AVOID *)(
				 uvp->uv_fnames[uvp->uv_ffound] =
				 (char *)mallop(strlen(logo[i]) + 2,
				 "fnames[ff+1] in chk_setup"))) return EALLOC;
				strcpy(
				 uvp->uv_fnames[(uvp->uv_ffound)++], logo[i]);
			 }
		  }
	   }
	ftypes[uvp->uv_ffound] = 0;	uvp->uv_fnames[uvp->uv_ffound] = NULL;
	header(stdout, REDUCTION, uvp->uv_fnames, ftypes);
	return EOK;
}

static void ucfree(UVWORK *uvp)
{ /*	Memory-releasing routine for uconvert */
	int	i, nf;
	char	**dcp;
	static char	ucmsg[] = "*dcp in uconvert";

	if (uvp->uv_ffound)
	  {
		nf = 2 * uvp->uv_ffound;
		for (i = 0, dcp = uvp->uv_realn; i < nf; i++)
			memzap((AVOID **)&(*dcp), ucmsg);
		for (i = 0, dcp = uvp->uv_fnames; i < nf; i++)
			memzap((AVOID **)&(*dcp), ucmsg);
	  }
	memzap((AVOID **)&(uvp->uv_realn), "realn in uconvert");
	memzap((AVOID **)&(uvp->uv_fnames), "fnames in uconvert");
	globalkill();
}

int uconvertmain(int argc, char *argv[])
{ /*	In this mainline, we handle the arguments and then process each file*/
	int		nc;
	char		*cpn;
	static UVWORK	uvs;

	if ((nc = chk_setup(argc, argv, &uvs)))
	  {
		ucfree(&uvs);	return nc;
	  }
	for (nc = 0; nc < (int)(uvs.uv_ffound); nc += 2)
	   { /* establish the real file names to be used */
		cpn = uvs.uv_fnames[nc];	errno = 0;
			/* open the input and output files */
		if (0 > (uvs.uv_fdi = open(uvs.uv_realn[nc], 0)))
		 fprintf(stderr, "Cannot open %s for  input; errno = %d\n",
		 uvs.uv_realn[nc], errno);
		else if (0 > (uvs.uv_fdo = creat(uvs.uv_realn[nc+1], 0664)))
		 fprintf(stderr, "Cannot open %s for output; errno = %d\n",
		 uvs.uv_realn[nc+1],errno);
			/* figure out which kind of file we're being asked to
			convert and do it */
		else if (!strcmp(cpn, "USPLINE")) ucacvt(&uvs);
		else if (!strcmp(cpn, "UPARAMS")) uprcvt(&uvs);
		else if (!strcmp(cpn, "UREFLS")) urfcvt(&uvs);
		else if (!strcmp(cpn, "WMASK")) wmaskcvt(&uvs);
		else if (!strcmp(cpn, "MULTIREF")) mrcvt(&uvs);
			/* close the input and output files */
		errno = 0;
		if ((0 <= uvs.uv_fdi) && (0 > close(uvs.uv_fdi)))
		 fprintf(stderr, "Warning: error closing %s; errno = %d\n",
		 uvs.uv_realn[nc], errno);
		errno = 0; if ((0 <= uvs.uv_fdo) && (0 > close(uvs.uv_fdo)))
		 fprintf(stderr, "Warning: error closing %s; errno = %d\n",
		 uvs.uv_realn[nc+1] ,errno);
	  }
	ucfree(&uvs);	return 0;
}

/* end of uv_convert functionality */

/* end of misctools.c */
