/* mutools -- various tools associated with MULISTs

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Dec 2006: use muinterp to read in MUREFLs in these routines;
		use outmu for printing them out
	24 Jun 2006: include xgpostint.h; that's where the structures live
	16 May 2006: merge abs_order, expand_mu, mubij_fmt, mukb_2fmt to here
	31 Dec 2005: don't qsort if fewer than 2 elements
	 6 Oct 2005: housekeeping
	13 May 2005: add underscores to absorder, mukb2fmt, mubij2fmt
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 6 Sep 2004: cosmetic change
	28 Aug 2004: routines used only here are declared as static;cosmetics
	25 Jan 2004: mudata is now a static local, not a static global
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Dec 2003: explicit declaration for strcpy on some machine types
	21 Nov 2003: include ctype in an attempt to get -Wall to shut up
	 5 Nov 2003: variable formerly called dspace is now dspac.
	23 Sep 2003: remove declaration of strcpy()--sometimes it's a macro
	19 Sep 2003: in main loop of mukb_2fmt, always set fnat and fder
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	13 Aug 2003: correct spelling of globalkill()
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	17 Jan 1994: Remove superfluous function declarations,
		which IBMRIOS dislikes.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN versions of these functionalities
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version
	15 Mar 1992: add -b option to mubij2fmt.
	15 Mar 1992: initial code, reverse engineered from `mubij2fmt'
	 9 Jan 1992: initial code for mubij2fmt, based on old mukb2fmt
	
   Purpose: various manipulations of MULISTs.
	Specifically:
	 abs_order reads an ordinary mulist and writes one out which has been
		sorted on actual ("absolute") hkl order.  If the "-m" flag is
		given, the indices are fiddled with according to
			h' = l,  l' = h.
		The convention on monoclinics follows SCATT,
		i.e., k and l is positive and h can be negative.
	expand_mu reads a MULIST and expands it to its P1 equivalent, i.e.
		all the symmetry equivalents in a hemisphere are included.
	mubij_2fmt reformats a MULIST of Bijvoet deltaF's into a formatted
		DORGBIN file
	mukb_2fmt converts a DORGBIN-style file output by mukb into
		one that contains delta F's
   Cross references: ../postprocess/makemu.c, ../postprocess/mukb.c
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_CTYPE	1
#include	"xgpostint.h"
#ifdef VMS
#include	"vsort.h"
#endif
#include	<stdlib.h>
#define		DEF_SPACEG	19

extern int absordmain(int argc, char *argv[]);
extern int expandmumain(int argc, char *argv[]);
extern int mubijmain(int argc, char *argv[]);
extern int mu2fmtmain(int argc, char *argv[]);

/*************************************
 The first routine enables reordering of MULIST records
 into actual ("absolute") hkl order.
*************************************/
int absordmain(int argc, char *argv[])
{ /* The main line for converting a MULIST into absolute (HKL) order */
	register MUREFL	*mup, *muend;
	int		twiddle;
	unsigned	nrefls, nelts;
	short		ts;
	char		line[90], cflag;
	FILE		*fpi, *fpo;
	static char	*fnames[] = {"mulist", "mulist", NULL};
	static char	ftypes[] = {1, 2, 0 };
	static MUREFL	*mudata;
#ifdef VMS
	static u_short	hklbuf[] = {3,
			DSC$K_DTYPE_B, SO_ASCENDING, 0, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 1, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 2, 1};
#endif

/* This switches h for l and leaves the signs in the proper order
	for a monoclinic cell.  Then it resorts on increasing order
	of signed h,k,l and writes the whole mess out. */

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	twiddle = 0;
	if (argc > 1 && (argv[1][0] == '-' && ((cflag = argv[1][1]) == 'm'||
		cflag == 'M')))	twiddle = 1;
	header(stdout, REDUCTION, fnames, ftypes);
	if (twiddle)
	  {
	  fprintf(stdout,"Reflections will be twiddled according to\n");
	  fprintf(stdout,"(h,k,l) -> (l,k,h) if l >=0 and\n");
	  fprintf(stdout,"(h,k,l) -> (-l,k,-h) if l < 0\n");
	  }
	if (qerror((AVOID *)(fpi = xgopen("mulist", "r")),
		"open input mulist")) return EOPENERR;
	for (nrefls = 0; NULL != fgets(line, 82, fpi); ) nrefls++;
	printf("%d lines detected in the input mulist\n", (int)nrefls);
	rewind(fpi);
	if (NULL == (AVOID *)(mudata = (MUREFL *)callop((size_t)(nrefls + 2),
	 sizeof (MUREFL), "mudata in absorder"))) return EALLOC;
	for (mup = mudata; NULL != fgets(line,82,fpi); mup++)
	   {
		if (11 > muinterp(line, mup)) break;
		if (twiddle)
		  {
			ts = mup->mu_hkl[0]; mup->mu_hkl[0] = mup->mu_hkl[2];
			mup->mu_hkl[2] = ts;
		  }
		if (mup->mu_hkl[2] < 0)
		  {
			mup->mu_hkl[0] = -mup->mu_hkl[0];
			mup->mu_hkl[2] = -mup->mu_hkl[2];
		  }
	   }
	(void)fclose(fpi);
	muend = mup;
	printf("%d reflections read in and fixed up\n", (int)(mup - mudata));
	nelts = mup - mudata;
	if (nelts > 1) qsort((AVOID *)mudata, nelts,
		sizeof (MUREFL), (QPE)murcmp);
	printf(" Sort completed\n");
	if (qerror((AVOID *)(fpo = xgopen("mulist", "w")),
		"open output mulist")) return EOPENERR;
	for (mup = mudata; mup < muend; mup++) (void)outmu(fpo, mup);
	(void)fclose(fpo);
	printf("%d sorted reflections written out\n", (int)(muend - mudata));
	memzap((AVOID **)&mudata, "mudata in absorder");
	globalkill();	return EOK;
}

/* end of abs_order functionality */

/*************************************
	The next few routines enable expansion of a MULIST out to its
	P1 equivalent.  It reads an ordinary mulist and writes one out
	that has been expanded so that a full hemisphere of data is given.
	The hemisphere it writes is h >= 0.
	The only runtime parameter required is the international tables'
	 spacegroup number.  By default this is 19, i.e. P2(1)2(1)2(1).
	Currently this program writes 'em out in place without resorting.
	Logical names:	MULIST	(input)
			MUOUT	(output)
*************************************/

static int putmu(FILE *fp, MUREFL *mup, int flip)
{ /*	This writes one reflection to the output file fp.  If flip & 01,
	we reverse the sense of Fp and Fm before writing.
	If flip & 02, we reverse the signs of l;
	If flip & 04, we reverse the signs of k;
	If flip & 010, we reverse the signs of h; This returns 1.*/
	double	tv;
	MUREFL	lmu;

	cpymem((AVOID *)&lmu, (AVOID *)mup, sizeof (MUREFL));
	if (flip & 01)
	  { /* swap F+ for F- and the corresponding sigmas */
		tv = lmu.mu_fminus;	lmu.mu_fminus = lmu.mu_fplus;
		lmu.mu_fplus = tv;	tv = lmu.mu_sigminus;
		lmu.mu_sigminus = lmu.mu_sigplus;	lmu.mu_sigplus = tv;
	  }
	if (flip & 02) lmu.mu_hkl[2] = -lmu.mu_hkl[2];
	if (flip & 04) lmu.mu_hkl[1] = -lmu.mu_hkl[1];
	if (flip & 010) lmu.mu_hkl[0] = -lmu.mu_hkl[0];
	(void)outmu(fp, &lmu);	return 1;
}

static int putall(FILE *fp, MUREFL *mup, short isys)
{ /*	This writes out as many P1 reflections as are required
	by the spacegroup */
	int	ih, ik, il, nput;

	nput = 0;
	ih = mup->mu_hkl[0];	ik = mup->mu_hkl[1];	il = mup->mu_hkl[2];
	switch (isys)
	      {
		case 1:	if (ih >= 0) nput += putmu(fp, mup, 0);
			else nput += putmu(fp, mup, 017);
			break;
		case 2:	if (ih >= 0)
			  {
				nput += putmu(fp, mup, 0);
				if (ik) nput += putmu(fp, mup, 05);
			  }
			else 
			  {
				nput += putmu(fp, mup, 012);
				if (ik) nput += putmu(fp, mup, 015);
			  }
			break;
		case 3:	if (ih >= 0)
			  {
				nput += putmu(fp, mup, 0);
				if ((ik) && (il)) nput += putmu(fp, mup, 06);
				if (il) nput += putmu(fp, mup, 03);
				if (ik) nput += putmu(fp, mup, 05);
			  }
			else
			  {
				nput += putmu(fp, mup, 012);
				if ((ik) && (il)) nput += putmu(fp, mup, 014);
				if (il) nput += putmu(fp, mup, 011);
				if (ik) nput += putmu(fp, mup, 015);
			  }
		   }
	return nput;
}

int expandmumain(int argc, char *argv[])
{ /* The mainline for converting a MULIST to the corresponding P1 MULIST. */
	register MUREFL	*mup;
	short		ispgp, isys;
	int		ih, ik, il, nout, spec_spg;
	char		line[90];
	FILE		*fpi, *fpo;
	static MUREFL	*mudata;
	static char	*fnames[] = {"mulist", "muout", NULL };
	static char	ftypes[] = {1, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if ((argc > 1) && 0 < (spec_spg = atoi(argv[1])) && spec_spg < 231)
		ispgp = spec_spg;
	else ispgp = DEF_SPACEG;
	if (ispgp < 3) isys = 1;
	else if (ispgp < 16) isys = 2;
	else if (ispgp < 74) isys = 3;
	else if (ispgp < 88) isys = 4;
	else if (ispgp < 142) isys = 5;
	else isys = 6;
	if (ispgp >= 88)
	  {
		fprintf(stderr, "Sorry! No code in that spgrp available.\n");
		return EOK;
	  }
	header(stdout, REDUCTION, fnames, ftypes);
	printf(
	 "Expansion will use elements appropriate to spacegroup %3d\n", ispgp);
	if (qerror((AVOID *)(fpi = xgopen("mulist", "r")),
		"open input mulist")) return EOPENERR;
	if (qerror((AVOID *)(fpo = xgopen("muout", "w")),
		"open output mulist")) return EOPENERR;
	if (NULL == (AVOID *)(mudata = (MUREFL *)callop(1, sizeof (MUREFL),
	 "mudata in expandmu"))) return EALLOC;
	for (mup = mudata, nout = 0; NULL != fgets(line,82,fpi); )
	   {
		if (11 > muinterp(line, mup)) break;
		ih = mup->mu_hkl[0], ik = mup->mu_hkl[1]; il = mup->mu_hkl[2];
		nout += putall(fpo, mup, isys);
	   }
	(void)fclose(fpi);	(void)fclose(fpo);
	printf("%d p1 reflections written out\n", nout);
	memzap((AVOID **)&mudata, "mudata in expandmu");
	globalkill();	return EOK;
}

/* end of expand_mu functionality */

/*************************************
 These next routine enables conversion of a MULIST containing
	h k l stol F sig F+ sig(F+) F- sig(F-) flag
 into a DORGBN file containing
	h k l D |F+ - F-| F+ F-
 for all reflections for which F+ and F- are both present.
 This is suitable for shoving into Richard Fisher's "ftob" and
 thence into EXPAND.
 It reads the standard input and writes the standard output!
**************************************/

/*ARGSUSED*/
int mubijmain(int argc, char *argv[])
{ /* This reformats the deltaFs in a MULIST into DORGBIN format */
	int	flag, ncentric, nin, nnobij, nnopri, ntoobig, nout;
	long	tl0;
	double	cutdi, deli, dspac, fbij, fdif, fpri, fsum, stol;
	char	mbline[256], timebuf[26];
	MUREFL	lmu;

	nin = ncentric = nnobij = nnopri = ntoobig = nout = 0;
	tl0 = NOW;	cutdi = -1;
	if ((argc > 1) && (argv[1][0] == '-') && (argv[1][1] == 'b' ||
		argv[1][1] == 'B')) cutdi = atof(&(argv[1][2]));
	if (cutdi > DFUZZ) printf(
" Data will be output only if 2 * |I+ - I-| / (I+ + I-) < %7.3f\n", cutdi);
	(void)strcpy(timebuf, ctime(&tl0));
	printf("%6d\n Anomalous Difference Data assembled %s", 3, timebuf);
	puts(" |Fbij - Fpri| for observed acentric reflections");
	puts(" Fpri for observed acentric reflections");
	puts(" Fbij for observed acentric reflections");
	while (NULL != fgets(mbline, 255, stdin))
	     {
		if (11 > muinterp(mbline, &lmu)) break;
		flag = lmu.mu_flag; fpri = lmu.mu_fplus; fbij = lmu.mu_fminus;
		nin++;	stol = lmu.mu_stl; fsum = fpri * fpri + fbij * fbij;
		deli = (fsum < DFUZZ) ?
			0. : 2. * fabs(fpri * fpri - fbij * fbij) / fsum;
		if (flag & 01) ncentric++;
		else if (flag & 022) nnopri++;
		else if (flag & 044) nnobij++;
		else if ((cutdi > DFUZZ) && (deli > cutdi)) ntoobig++;
		else {
			fdif = fabs(fpri - fbij);
			dspac = (stol > DFUZZ) ? 1. / (2. * stol) : 999.;
			printf(" %4hd%4hd%4hd%10.4f%10.4f%10.4f%10.4f\n",
			 lmu.mu_hkl[0], lmu.mu_hkl[1], lmu.mu_hkl[2],
			 dspac, fdif, fpri, fbij);
			nout++;
		     }
	     }
	printf(" %6d reflections read;        %6d written\n",
		nin, nout);
	if (ncentric) printf(" %6d reflections were centric\n", ncentric);
	if (nnopri) printf(
	 " %6d reflections lacked an F+ observation or had F+ <= 0\n", nnopri);
	if (nnobij) printf(
	 " %6d reflections lacked an F- observation or had F- <= 0\n", nnobij);
	if (ntoobig) printf(
	 " %6d reflections had 2 * |deli| / Isum > %7.2f\n", ntoobig, cutdi);
	return EOK;
}

/* end of mubij_2fmt functionality */

/*************************************
 These next two routines enable conversion of a DORBBIN-style file output
 by `mukb' into a format appropriate to FTOB and related functionalities
 It converts a mukb-produced DORGBN output file containing
	h k l stol Fnat signat dFnat Fder sigder dFder [...]
 into a DORGBN file containing
	h k l D |Fder - Fnat| Fnat Fder
 for all reflections for which Fnat and Fder are both present.
 This is suitable for shoving into Richard Fisher's "ftob" and
 thence into EXPAND.
 optional arguments:
	a numeric argument specifies the DORGBIN column where the
	 derivative of interest begins.  Thus "mukb2fmt 7"
	 would mean to use the derivative whose "Fder" value appears
	 in column 7 (rather than column 4, which is default).
	up to three flag arguments may be provided.  In the
	 definitions below J = sqrt(sigma(nat)^2 + sigma(der)^2):
	 -u<val>	Use only reflections for which |Fder-Fnat|/J < <val>,
	 -l<val>	Use only reflections for which Fder/sigma(der) > <val>
			 AND Fnat/sigma(nat) > <val>.
	 -r<val>	Use only reflections for which |Fder-Fnat|/J < R*<val>,
			 where R = rms value of |Fder-Fnat|/J.
	The maximum value of the numeric argument is 21.
 It reads the standard input and writes the standard output!
****************************************/

static int baddorg(char *str)
{ /*	This prints a meaningful error message and exits */
	fprintf(stderr, "Input DORGBN file has a faulty %s\n", str);
	return EREADERR;
}

int mu2fmtmain(int argc, char *argv[])
{ /*	This converts a MULIST iinto a formatted DORGBIN file */
	int	h, i, k, l, ncols, ndfos, nfdos, nfnos, nin;
	int	nnonat, nnoder, nout, nrms, speccol, startcol, useit;
	time_t	tl0;
	float	fr[18];
	double	dfnat, dfos, dspac, fder, fdif, fnat, lowfos;
	double	rmsdf, rmslim, sder, snat, sumdf, updel, varf;
	char	*cp, *valp, muline[256], timebuf[26];

	tl0 = time(0L);	startcol = 4;
	lowfos = -1.;	updel = rmslim = 1.e8;
	for (argc--, argv++; (argc) && **argv == '-'; argc--, argv++)
	   {
		cp = &(argv[0][1]);	valp = cp + 1;
		if ('a' <= *cp && *cp <= 'z') *cp -= ('a' - 'A');
		if ('U' == *cp) updel = atof(valp);
		else if ('L' == *cp) lowfos = atof(valp);
		else if ('R' == *cp) rmslim = atof(valp);
	   }
	if ((argc) && (4 <= (speccol = atoi(argv[1]))) &&
		(speccol < 21)) startcol = speccol;
	printf(
	 "Native data taken from column  1; derivative from column %2d\n",
		startcol);
	if (lowfos > -0.9999) printf(
	 "Reflections will be output only if delF/sigma < %7.2f\n", updel);
	if (updel < 0.99e8) printf(
 "Reflections will be output only if Fnat/sigma and Fder/sigma > %7.2f\n",
		lowfos);
	if (rmslim < 0.99e8) printf(
 "Reflections will be output only if delF/sigma < RMS(delF/sigma) * %7.2f\n",
		rmslim);
	updel *= updel;	rmslim *= rmslim;
	(void)strcpy(timebuf, ctime(&tl0));
	printf("%6d\n Isomorphous Difference Data assembled %s", 3, timebuf);
	puts(" |Fder - Fnat| for observed reflections");
	if (NULL == fgets(muline, 255, stdin))
		return baddorg("number of columns");
	ncols = atoi(muline);
	if (NULL == fgets(muline, 255, stdin)) baddorg("overall title");
	for (useit = 0; useit < 3; useit++)
	   {
		if (NULL == fgets(muline, 255, stdin))
			baddorg("native-data title");
		if (!useit) puts(muline); /* copy input title to output */
	   }
	for (useit = 0; useit < ncols-3; useit++)
	   {
		if (NULL == fgets(muline, 255, stdin))
			baddorg("derivative data title");
		if (useit == startcol-4) puts(muline);
	   }
	useit = startcol - 4;
/* calculate the RMS value of delF/sigma */
	if (rmslim < 0.99e8)
	  {
		for (sumdf = nin = 0;
		 (NULL != fgets(muline, 255, stdin)); nin++)
		   {
			snat = dfnat = 0.;
			for (i = 0; i < 18; i++) fr[i] = 0.;
			if (4 > sscanf(muline,
 "%d %d %d %lf %lf %lf %lf %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
			 &h, &k, &l, &dspac, &fnat, &snat, &dfnat,
			 &(fr[0]), &(fr[1]), &(fr[2]), &(fr[3]), &(fr[4]),
			 &(fr[5]), &(fr[6]), &(fr[7]), &(fr[8]), &(fr[9]),
			 &(fr[10]), &(fr[11]), &(fr[12]),&(fr[13]),&(fr[14]),
			 &(fr[15]),&(fr[16]),&(fr[17]))) break;
			fder = fr[useit];	sder = fr[useit+1];
			fdif = fder - fnat; varf = snat * snat + sder * sder;
			dfos = (fdif / varf) * fdif; sumdf += dfos;
		   }
		rmsdf = sumdf / (double)nin;
		printf("RMS(delF/sigma) = %7.2f\n", sqrt(rmsdf));
		rmsdf *= rmslim;
		if (-1 == lseek(0, (off_t)0L, 0))
		  {
			fprintf(stderr, "Unable to rewind input\n");
			return ESEEKERR;
		  }
		for (i = 0; i < ncols + 2; i++) if (NULL ==
		 fgets(muline, 255, stdin)) baddorg("header on 2nd pass");
	  }
	else rmsdf = -1;
	nin = nout = nnonat = nnoder = ndfos = nfnos = nfdos = nrms = 0;
	while (NULL != fgets(muline, 255, stdin))
	     {
		fnat = snat = dfnat = 0.; for (i = 0; i < 18; i++) fr[i] = 0.;
		if (4 > sscanf(muline,
 "%d %d %d %lf %lf %lf %lf %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
		 &h, &k, &l, &dspac, &fnat, &snat, &dfnat,
		 &(fr[0]), &(fr[1]), &(fr[2]), &(fr[3]), &(fr[4]), &(fr[5]),
		 &(fr[6]), &(fr[7]), &(fr[8]), &(fr[9]), &(fr[10]), &(fr[11]),
		 &(fr[12]),&(fr[13]),&(fr[14]),&(fr[15]),&(fr[16]),&(fr[17])))
			break;
		nin++;	fder = fr[useit];	sder = fr[useit + 1];
		if (fnat < -0.9999) nnonat++;
		else if (fder < -0.9999) nnoder++;
		else if (fnat < snat * lowfos) nfnos++;
		else if (fder < sder * lowfos) nfdos++;
		else {
			fdif = fabs(fder - fnat);	sder = fr[useit+1];
			varf = snat * snat + sder * sder;
			dfos = (fdif / varf) * fdif;
			if (dfos > updel) ndfos++;
			else if (dfos > rmsdf) nrms++;
			else {
				nout++;
				printf(" %4d%4d%4d%10.4f%10.4f%10.4f%10.4f\n",
				 h, k, l, dspac, fdif, fnat, fder);
			     }
		     }
	     }
	printf(" %6d reflection%sread;        %6d written\n",
		nin, (nin == 1 ? " " : "s "), nout);
	if (nnonat) printf(
		" %6d reflection%slacked a native observation\n",
		nnonat, (nnonat == 1 ? " " : "s "));
	if (nnoder) printf(
		" %6d reflection%slacked a derivative observation\n",
		nnoder, (nnoder == 1 ? " " : "s "));
	if (ndfos) printf(" %6d reflection%shad delF/sigma > %7.2f\n",
		ndfos, (ndfos == 1 ? " " : "s "), sqrt(updel));
	if (nfnos) printf(
	 " %6d reflection%shad Fnat/sig(nat) < %7.2f\n",
		nfnos, (nfnos == 1 ? " " : "s "), lowfos);
	if (nfdos) printf(
	 " %6d reflection%shad Fder/sig(der) < %7.2f\n",
		nfdos, (nfdos == 1 ? " " : "s "), lowfos);
	if (nrms) printf(
	 " %6d reflection%shad delF/sigma > %7.2f * RMS(delF/sigma)\n",
		nrms, (nrms == 1 ? " " : "s "), sqrt(rmslim));
	return EOK;
}

/* end of mukb_2fmt functionality */
/* end of mutools.c */
