/* scrio.c -- reflection input/output routines for XGEN's scaling function

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h rather than scdefs.h
	23 Jun 2006: move put_shifts here so it can be static
	 6 Jun 2006: statsh,sum_wilson,wilscale,sumscalekrun moved here;
		sccycles created; many routines are now static
	 9 Dec 2005: sum_input and scheads inserted here from elsewhere
	20 Nov 2005: inline argument declarations and strict prototyping
	18 Oct 2005: replace static globals with SCALERUN structure elements
	10 Oct 2005: housekeeping
	30 Jun 2005: make_shift_list(): skip correct # of DATARUN bytes
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	27 Aug 2004: routines used only here are declared as static; cosmetics
	21 Jan 2004: license message inserted and initial comment reordered
	 7 Jan 2004: SCALERUN structure elements in function args
		replace extern globals; args updated accordingly
	14 Nov 2003: replaced all globals from scmain.c with components of
		the SCALERUN structure
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in scext.h
	10 Nov 2002: compute variance values carefully for zeroth cycle
	 1 Nov 2002: vpq->v_coef[0] initialized to 1 in scheads()
	15 Apr 2002: (?) in several places, modify uses of shlp[]
	25 Mar 2002: removed MSI attachments
	15 Jun 2001: in scheads, deliberately overestimate # of shifts per run
	31 May 1995: "register" -> "register int" throughout
	12 Apr 1995: replace s_char with sichar;
		set nscales=0 before scaling so multiple runs work
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread to input CRYSTAL and DATARUN data
	 5 Oct 1994: use xgread to input CRYSTAL, DATARUN, MERGEDREF data
	 5 May 1994: source-file header cleaned up;
	30 Sep 1994: make sure $SCALEIN isn't empty before using it
		names changed from POLYDET to XGEN
	16 Mar 1994: global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: input and output steps of `scalek' functionality; mostly at.
	the beginning and end of the program, since the reflection list
	itself sits in memory during ordinary scaling cycles.
   Cross references:

*************************************************************************/
#define		NEED_TIME	1
#include	"xgpostint.h"
#define		FLOOR		(0.1)	/* weight floor for resol. */

static int scheads(SCALERUN *spi)
{ /*	This routine processes the input multiref header blocks
	and saves the worthwhile info from them.
	It reserves memory for the reflection and observation tanks.
	It also reserves memory for the POINTERS to the shift structures. */
	register int	j;
	unsigned	nshifts,reflestim, shad;
	xgulong		jj;
	short		*dvb, *dvp;
	CRYSTAL		*xhp;
	DATARUN		*rhp;
	RUNWORK		*rwp;
	QOB		*qp, *qe;
	VECPAR		*vpp, *vpq;

	spi->sc_xhdp = (CRYSTAL *)NULL;
	spi->sc_rhdp = (DATARUN *)NULL;
	if (EOK != (j = headmem(&(spi->sc_xhdp), &(spi->sc_rhdp), 1)))
		return j;
	xhp = spi->sc_xhdp;	rhp = spi->sc_rhdp;
	if (verror(0 < xgread((void *)xhp, XG_CRYSTAL, spi->sc_fpmi),
	 "read crystal header from the input multiref")) return EREADERR;
	nshifts = 0;			/* clear count of shifts */
	if (NULL == (AVOID *)(spi->sc_rub = (RUNWORK *)callop((unsigned)
	 xhp->xt_nruns, sizeof (RUNWORK), "rub in scheads"))) return EALLOC;
	/* read the run-header structures and save what we need from them */
	for (j = shad =0, rwp = spi->sc_rub; j < xhp->xt_nruns;
	 j++, rwp++, shad += 200)
	   {
		if (verror(0 < xgread((void *)rhp, XG_DATARUN, spi->sc_fpmi),
			"read a datarun header from the input multiref"))
				return EREADERR;
		(void)keep_dimen(rhp);
		workrun(rhp, rwp, xhp, (int)shad); /*save header info */
		/* provide some slop because rhd.run_nshifts is
		 often an underestimate */
		nshifts += rhp->run_nshifts + 4;
	   }
	if (nshifts > 1000) nshifts = xhp->xt_nshifts;
/* allocate memory for reflections, observations, and shift-scale pointers */
	reflestim = (unsigned)((spi->sc_flags & SC_MERGEBJ) ?
		xhp->xt_nmerged : xhp->xt_nrefls);
	if (NULL == (AVOID *)(spi->sc_refbase = (QREF *)callop(reflestim +
	 100, sizeof (QREF), "refbase in scheads"))) return EALLOC;
	if (NULL == (AVOID *)(spi->sc_obbase = (QOB *)callop((unsigned)
	 xhp->xt_ninteg, sizeof (QOB), "spi->sc_obbase in scheads")))
		return EALLOC;
	if (NULL == (spi->sc_shlp = init_shift(nshifts * 2)))
		return EALLOC;
	if (NULL == (AVOID *)(spi->sc_shwl = (SHSTAT *)callop(nshifts,
	 sizeof (SHSTAT), "spi->sc_shwl in scheads"))) return EALLOC;
	if (spi->sc_flags & SC_VECREFINED)
	  {
		qe = spi->sc_obbase + xhp->xt_ninteg;
		if (NULL == (AVOID *)(dvb = (short *)callop(
		 (unsigned)(spi->sc_noutpar * xhp->xt_ninteg), sizeof (short),
		 "dvb in scheads"))) return EALLOC;
		for (qp = spi->sc_obbase, dvp = dvb; qp < qe;
		 qp++, dvp += spi->sc_noutpar) qp->qo_sc = dvp;
		spi->sc_ndvgroups = xhp->xt_nruns;
		if (NULL == (AVOID *)(spi->sc_vshp = (VECPAR *)callop(2 *
		 spi->sc_ndvgroups, sizeof (VECPAR),
		 "spi->sc_vshp in scheads"))) return EALLOC;
		for (jj = shad = 0, vpp = spi->sc_vshp,
		 vpq = spi->sc_vshp + spi->sc_ndvgroups;
		  jj < spi->sc_ndvgroups; jj++, shad += 200, vpp++, vpq++)
		   {
			if (spi->sc_flags & SC_DIFFRACTION)
				vpp->v_ptype = vpq->v_ptype = 'd';
			else	vpp->v_ptype = vpq->v_ptype = 'e';
			vpp->v_ms = shad;	vpq->v_ms = shad;
			/* start with coeffs of 1, 0, 0, 0, ... */
			vpp->v_coef[0] = 1.;	vpq->v_coef[0] = 1.;
		   }
	  }
	else spi->sc_vshp = NULL;
	return EOK;
}

static void statsh(SHSTAT *sp)
{ /*	compute the actual least-square values of the temperature factor b
	and the scaling k value ksc based on an array of accumulated sum
	values from the data.  The I000 value and the standard deviations
	in b and k are also computed.  */
	float	*rp;
	double	b, chisqdf, disc, ks, var;
  
	sp->s_r[6] = sp->s_nob - 2L;
	if (sp->s_r[6] < 1) return;	/* no calcs if <= 2 obs */
	rp = sp->s_r;
	rp[1] /= 2.;	rp[2] /= 4.;	rp[4] /= 2.;
	disc = rp[0] * rp[2] - rp[1] * rp[1];
	if (fabs(disc) < DFUZZ) /* usually means no obs */
	  {
		sp->s_b = 0;	sp->s_i000 = sp->s_sigi0 = sp->s_sigb = 1.;
		return;
	  }
	ks = (rp[2] * rp[3] - rp[1] * rp[4]) / disc;
	sp->s_b = b = (rp[0] * rp[4] - rp[1] * rp[3]) / disc;
	sp->s_i000 = exp(ks);
	var = (rp[0] * ks * ks + 2. * rp[1] * ks * b +
	   rp[2] * b * b - 2. * (ks * rp[3] + b * rp[4]) + rp[5]) / disc;
	chisqdf = var / rp[6];
	sp->s_sigi0 = sqrt(rp[2] * chisqdf) * sp->s_i000;
	sp->s_sigb = sqrt(rp[0] * chisqdf);
	if (sp->s_nob > 1L) sp->s_res /= sp->s_nob;
}

static void sum_wilson(FILE *fp, SCALERUN *spi)
{ /*	This summarizes the Wilson-statistic calcs to file pointer fp */
	register int	n;
	SHIFT		*tsh;
	SHSTAT		*thisp;

	fprintf(fp, "\tInitial scale factors:\n");
	switch (spi->sc_nparpshift)
	  {
	   case	1:
		fprintf(fp,
" Shift Number  Initial    Shift Number  Initial    Shift Number  Initial\n");
		fprintf(fp,
"Number of Obs    Scale   Number of Obs    Scale   Number of Obs    Scale\n");
		for (n = 0, tsh = spi->sc_shlp, thisp = spi->sc_shwl;
		 n < spi->sc_nscales; n++, tsh++, thisp++)
		   fprintf(fp," %6d %6d %9.4f%s", tsh->shval, thisp->s_nob,
			tsh->shg, (n + 1) % 3 ? "  " : "\n");
		break;
	   case 2:
		fprintf(fp,
" Shift   Number  Initial  Initial    Shift   Number  Initial  Initial\n");
		fprintf(fp,
"Number   of obs  g value  B value   Number   of obs  g value  B value\n");
		for (n = 0, tsh = spi->sc_shlp, thisp = spi->sc_shwl;
		 n < spi->sc_nscales; n++, tsh++, thisp++)
		   fprintf(fp," %6d %8d %9.4f %9.4f%s",
			tsh->shval, thisp->s_nob, tsh->shg,
			 tsh->shb, (n+1)%2 ? "  " : "\n");
		break;
	   case 3:
		fprintf(fp, " Shift Number Initial Initial Initial   ");
		fprintf(fp, "Shift Number Initial Initial Initial\n");
		fprintf(fp, "Number of obs g value A value B value  ");
		fprintf(fp, "Number of obs g value A value B value\n");
		for (n = 0, tsh = spi->sc_shlp, thisp = spi->sc_shwl;
		 n < spi->sc_nscales; n++, tsh++, thisp++)
		   fprintf(fp," %6d %6d %7.4f %7.3f %7.3f%s",
			tsh->shval, thisp->s_nob, tsh->shg, tsh->sha,
			tsh->shb, (n+1)%2 ? "  " : "\n");
		break;
	  }
}

static void wilscale(SCALERUN *spi)
{ /*	This directs the computation of I(000) and B values for all the shifts
	which appear in a multiref.  An overall I(000) and B are also computed.
	Each shift's delta B and delta V (V = ln(I000) value is used in 
	computing starting values for the shift's scale factors, unless the
	user has requested 1-parameter-per shift refinement, in which case the
	mean value of I000*exp(-Bs*s) is used as the starting scale factor. */
	register int	i, n;
	int		odd;
	xgulong		ulsh;
	double		averes, oversc, sumb, sumscales, sumv, sumw, weight;
	SHSTAT		*sp, *allp;
	SHIFT		*tsh, *shq;

	spi->sc_tl0 = NOW;	allp = &(spi->sc_alls);
	for (n = 0, sp = spi->sc_shwl; n < spi->sc_nscales; n++, sp++)
	   {
		for (i = 0; i < 7; i++)	allp->s_r[i] += sp->s_r[i];
		allp->s_nob += sp->s_nob;	allp->s_res += sp->s_res;
		statsh(sp);
	   }
	statsh(allp);	averes = allp->s_res;
	spi->sc_barsq = averes / 4.;	spi->sc_bar = sqrt(spi->sc_barsq);
	oversc = allp->s_i000 * exp(-0.5 * allp->s_b * averes);
	sumscales = sumv = sumb = sumw = 0.;
	for (n = 0, sp = spi->sc_shwl; n < spi->sc_nscales; n++, sp++)
	   {
		sp->s_sk = sp->s_i000 * exp(-0.5 * sp->s_b * averes) / oversc;
		if (sp->s_i000 > FUZZ && sp->s_nob > 1)
		  {
			weight = log((double)(sp->s_nob));
			sumw += weight;	sumv += weight * sp->s_i000;
			sumb += weight * sp->s_b;
		  }
		sumscales += (sp->s_sk * sp->s_nob);
	   }
	if (sumw)
	  {
		sumv /= sumw;	sumb /= sumw;
	  }
	printf("Overall I000, B = %9.3f, %9.5f\n", sumv, sumb);
	if(allp->s_nob > 0L) sumscales /= allp->s_nob;
	if (spi->sc_flags & SC_INPUTSCALE) return;
	if (allp->s_nob == 0L || sumscales < DFUZZ)
	  {
		fprintf(spi->sc_fpl,
		 "Wilson-statistics failed to give initial scales\n");
		for (n = 0, tsh = spi->sc_shlp; n < spi->sc_nscales;
		 n++, tsh++)
		   {
			tsh->shg = 1.; tsh->shb = 0.;
		   }
		return;
	  }
	if (spi->sc_nparpshift < 2)
	  {
		for (sumscales = n = 0, tsh = spi->sc_shlp,
		 sp = spi->sc_shwl; n < spi->sc_nscales; n++, tsh++, sp++)
		   {
			tsh->shg = sp->s_sk; tsh->sha = tsh->shb = 0.;
			sumscales += tsh->shg;
		   }
	  }
	else {
		for (sumscales = n = 0, tsh = spi->sc_shlp, sp =spi->sc_shwl;
		 n < spi->sc_nscales; n++, tsh++, sp++)
		   {
			tsh->shg = sp->s_i000 / sumv;
			tsh->shb = -2. * (sp->s_b - sumb);
			tsh->sha = 0.;
			sumscales += tsh->shg;
		   }
	     }
	if (sumscales > DFUZZ)
	  {
		sumscales /= (double)(spi->sc_nscales);
		for (n = 0, tsh = spi->sc_shlp; n < spi->sc_nscales;
		 n++, tsh++) tsh->shg /= sumscales;
	  }
	/* if we are supposed to be merging top & bottom, do so */
	if (spi->sc_flags & SC_ALLOFDET)
	 {
	  for (n = 0, tsh = spi->sc_shlp; n < spi->sc_nscales; n++, tsh++)
	    if (!(tsh->shval % 2)) /* is this an even-numbered shift? */
	     { /* yes.  See if we can find the corresponding odd shift */
		ulsh = LSH(tsh->runno, 24) | LSH(tsh->chamno, 16) | tsh->shval;
		odd = find_shift(spi->sc_shlp, ulsh + 1L,
		 (int)(spi->sc_nscales));
		/* if odd >= 0 we found it.  Average the two. */
		if (0 <= odd)
		  {
			shq = spi->sc_shlp + odd;
			tsh->shg = 0.5 * (tsh->shg + shq->shg);
			tsh->sha = 0.5 * (tsh->sha + shq->sha);
			tsh->shb = 0.5 * (tsh->shb + shq->shb);
			shq->shg = tsh->shg;	shq->sha = tsh->sha;
			shq->shb = tsh->shb;
		  }
	      }
	 }
	sum_wilson(spi->sc_fpl, spi);
	spi->sc_tl1 = NOW;	printf(
	 " Completing Wilson-statistics analysis required %6d seconds\n",
		(int)(spi->sc_tl1 - spi->sc_tl0));
}

static void sum_input(FILE *fp, SCALERUN *spi)
{ /*	Print scale-factor inputs to stream fp */
	register int	i;
	register SHIFT	*tsh;
	SHSTAT		*allp, *thisp;

	fputs("Scaling input summary\n",fp);
	fprintf(fp,"Number of scales used:%6d\n", (int)(spi->sc_nscales));
	fprintf(fp,"%7d observations of %7d reflections were input\n",
		(int)spi->sc_nobin, (int)spi->sc_nrefin);
	fprintf(fp,"%7d observations of %7d reflections were used\n",
		(int)spi->sc_nobput, (int)spi->sc_nrefput);
	if ((spi->sc_flags & SC_VECREFINED) && (spi->sc_freq > 1))
	  {
		fprintf(fp,
	 " 1 out of %d refls will be sampled for diffraction-vector sums\n",
		spi->sc_freq);
	  }
	if (fp == stdout) return;
	allp = &(spi->sc_alls);
	if (!(spi->sc_flags & SC_INPUTSCALE))
	  {
		if (spi->sc_flags & SC_WILSON)
		  {
		  fprintf(fp,"Initial scales set with Wilson statistics\n");
		  fprintf(fp,"Overall I000, B values: %10.2f, %8.4f\n",
			allp->s_i000, allp->s_b);
		  }
		else
		  fprintf(fp,"All initial scale factors set to unity\n");
	  }
	if (spi->sc_flags & SC_VERBOSE)
	  {
		fprintf(fp,
		" Initial input scale constants:\n\n");
		fprintf(fp,
 "   Shift Initial Value   # Obs      I000 sig(I0)       B  sig(B)\n");
		for (i = 0, thisp = spi->sc_shwl, tsh = spi->sc_shlp;
		 i < spi->sc_nscales; i++, tsh++, thisp++)
		   fprintf(fp, "%8d%14.6f%8d%10.0f%8.1f%8.2f%8.3f\n",
			tsh->shval, tsh->shg, thisp->s_nob,
			thisp->s_i000, thisp->s_sigi0,
			thisp->s_b, thisp->s_sigb);
		fprintf(fp,"OVERALL VALUES:       ");
		fprintf(fp,"%8d%10.0f%8.2f%8.3f%8.3f\n", allp->s_nob,
		 allp->s_i000, allp->s_sigi0, allp->s_b, allp->s_sigb);
	  }
}

static int make_shift_list(SCALERUN *spi)
{ /*	This constructs the list of input shifts in the case where
	we aren't going to construct it from the "scalein" file.
	Note that we stop going through the data when we reach the number of
	scales listed in the header. */
	register int	shind, nob;
	register ABBROB	*aobp;
	register SHIFT	*tsh;
	xgulong		ulsh;
	CRYSTAL		*xhp;
	MERGEDREF	mr;
	off_t		seekpos;
	short		felloutearly, thisshift;
	static size_t	sizdatarun = 256;

	spi->sc_tl0 = NOW;	felloutearly = spi->sc_nscales = 0;
	xhp = spi->sc_xhdp;
	while (0 < xgread((void *)&mr, XG_MERGEDREF, spi->sc_fpmi))
	  if (mr.mr_nob > 1) /* refl MUST be multiply observed */
	    { /* loop thru observations, finding unencountered shifts */
		for (nob = 0; nob < (int)(mr.mr_nob); nob++)
		   {
			aobp = &(mr.mr_ob[nob]);
			if ((thisshift = aobp->abo_shift) >= 0)
			  {
				ulsh = LSH(aobp->abo_runno, 24) |
				 LSH(aobp->abo_chamno, 16) | thisshift;
				shind = (spi->sc_nscales ?
				 find_shift(spi->sc_shlp, ulsh,
				  (int)(spi->sc_nscales)): -1);
				if (0 > shind)
				  { /* this is a new shift: add it to list*/
					shind =((shind ==
					 -(++spi->sc_nscales)) ?  0 : -shind);
					add_shift(spi->sc_shlp, shind,
					 (int)(spi->sc_nscales));
					tsh = spi->sc_shlp + shind;
					tsh->runno = aobp->abo_runno;
					tsh->chamno = aobp->abo_chamno;
					tsh->shval = thisshift;
				  } /* end of "if (0 > shind) clause" */
			  } /* end of "if ((sigm ..." clause */
		   } /* end of "for (nob = 0"... clause */
		if (spi->sc_nscales >= xhp->xt_nshifts)
		  {
			felloutearly = 1;	break;
		  }
	  } /* end of (if mr.mr_nob > 1) clause */
	if (felloutearly)
	  {
		if (qerror((AVOID *)(spi->sc_fpmi =
		 freopen(trname("multiref"), "r", spi->sc_fpmi)),
		  "reopen multiref for continued reading")) return EOPENERR;
	  }
	seekpos = (off_t)(sizeof (CRYSTAL) + xhp->xt_nruns * (int)sizdatarun);
	if (-1 == fseek(spi->sc_fpmi, (long)seekpos, 0)) return ESEEKERR;
	spi->sc_tl1 = NOW;
	printf(
	 "Determining the initial shift assignments required %6d seconds\n",
		(int)(spi->sc_tl1 - spi->sc_tl0));
	if (stdout != spi->sc_fpl) fprintf(spi->sc_fpl,
	 "Determining the initial shift assignments required %6d seconds\n",
		(int)(spi->sc_tl1 - spi->sc_tl0));
	return EOK;
}

static int setup_ob(ABBROB *aobp, SCALERUN *spi)
{ /*	this sets up inclusion of an observation in the list.
	It returns 1 if the observation is being included */
	register int	shind, dvx;
	register SHSTAT	*sp;		/* ptr to shift-statistics structure */
	xgulong		ulsh;
	double		bi, inten, intenfu, sigm, wi, rve[10];
	RUNWORK		*rup;

	if ((sigm = aobp->abo_sigma) <= 0) return 0; /*sigma cut*/
	if (sigm < 1) sigm = 1;	/* so when we integerize it doesn't crap out */
	ulsh = LSH(aobp->abo_runno, 24) | LSH(aobp->abo_chamno, 16) |
		aobp->abo_shift; /* 32-bit version of shift value */
	if (0 > (shind = find_shift(spi->sc_shlp, ulsh,
	 (int)(spi->sc_nscales)))) return 0;
	inten = aobp->abo_ii;
	if (spi->sc_flags & SC_WILSON) /* do the Wilson calcs if requested*/
	  {
		sp = &(spi->sc_shwl[shind]);	/* Wilson-stat structure ptr*/
		intenfu = (inten < FUZZ) ? 0. : inten;
		wi = sqrt(intenfu / sigm) * (FLOOR + spi->sc_resq4);
		bi = (inten >= 1 ? log(inten) : 0.);
		sp->s_r[0] += wi;	sp->s_r[1] -= wi * spi->sc_resq4;
		sp->s_r[2] += wi * spi->sc_resq4 * spi->sc_resq4;
		sp->s_r[3] += wi * bi;
		sp->s_r[4] -= wi * bi * spi->sc_resq4;
		sp->s_r[5] += wi * bi * bi;
		sp->s_nob++;		sp->s_res += spi->sc_resq4;
	  }
	if (spi->sc_flags & SC_VECREFINED)
	  {
		/* determine the exit-beam or diffraction vector associated
		with this observation and get the spherical-harmonic
		coefficients associated with it.*/
		rup = spi->sc_rub + (int)(aobp->abo_shift / 200);
		refvec(aobp, rup, (int)spi->sc_flags,
		 (int)(spi->sc_noutpar), rve);
		for (dvx = 0; dvx < (int)(spi->sc_noutpar); dvx++)
		 (spi->sc_oblist)->qo_sc[dvx] = (short)(rve[dvx] / DVCSCALE);
	  }
	(spi->sc_oblist)->qo_inten = inten;
	(spi->sc_oblist)->qo_sigma = sigm;
	/* temporary calc */
	(spi->sc_oblist)->qo_dvgrp = (int)(aobp->abo_shift / 200);
	/* remember the omega value for interpolations */
	(spi->sc_oblist)->qo_omob = aobp->abo_omob;
	(spi->sc_oblist)->qo_shift = shind;	(spi->sc_oblist)++;
	if (inten > 0) (spi->sc_ngt0)++;	return 1;
}

static void del_omegas(SCALERUN *spi)
{ /*	This finds the omega centroid of each scaling group and saves
	it in the appropriate place in the SHIFT structure */
	int	nob, shind;
	double	rmn, *mnob, *meanom;
	QOB	*obend, *obp;
	SHIFT	*shp;

		/* get neighbor relationships */
	neighbors(spi->sc_shlp, (unsigned)(spi->sc_nscales), 0);
	mnob = (double *)callop((unsigned)(2 * spi->sc_nscales),
	 sizeof (double), "mnob in del_omegas");
	meanom = mnob + spi->sc_nscales;
	for (spi->sc_reflist = spi->sc_refbase,
	 spi->sc_oblist = spi->sc_obbase;
	 (nob = (spi->sc_reflist)->qr_nob);
	 (spi->sc_reflist)++, spi->sc_oblist += nob)
	   {
		obend = spi->sc_oblist + nob;
		for (obp = spi->sc_oblist; obp < obend; obp++)
		   {
			shind = obp->qo_shift;	mnob[shind] += 1;
			meanom[shind] += obp->qo_omob;
		   }
	   }
	for (shind = 0; shind < spi->sc_nscales; shind++)
	  	if ((rmn = mnob[shind])) meanom[shind] /= rmn;
	for (shind = 0, shp = spi->sc_shlp; shind < spi->sc_nscales;
	 shind++, shp++) shp->shomc = meanom[shind];
	memzap((AVOID **)&mnob, "mnob in del_omegas");
	/* if we're merging top & bottom, use the same mean-omega
	values for the odd scaling groups as for the even ones */
	if (spi->sc_flags & SC_ALLOFDET)
	  for (shind = 0, shp = spi->sc_shlp; shind < spi->sc_nscales;
	    shind++, shp++) if (shind % 2) shp->shomc = (shp - 1)->shomc;
}

static void putref(MERGEDREF *mp, SCALERUN *spi)
{ /*	This routine handles the insertion of a reflection into memory. */
	register int	i, nuseful;
	unsigned short	nobin;		/* # of observations input */
	CRYSTAL		*xhp;
	QOB		*obhold;	/* place holder in observation list */

	xhp = spi->sc_xhdp;	nobin = mp->mr_nob;
	(spi->sc_nrefin)++;	/* increment count of input refls */
	spi->sc_nobin += nobin;	/* ditto for input obs */
	if (nobin < (u_short)spi->sc_nrepmin) return;	/* only if >1 obs */
	/* Loop through the observations of the reflection, accumulating
	the Wilson-statistics values and sticking the intensity, sigma,
	and shift-index values into the temporary bins or files.  */
		/*4 (sintheta/lambda)squared*/
	spi->sc_resq4 = 4. * mp->mr_stl * mp->mr_stl;
	obhold = spi->sc_oblist;
	for (i = nuseful = spi->sc_ngt0 = 0; i < (int)nobin; i++)
	   nuseful += setup_ob(&(mp->mr_ob[i]), spi);
		/* use only if we have enough good obs */
	if (spi->sc_ngt0 >= spi->sc_nrepmin)
	  {
		(spi->sc_nrefput)++;	/* increment count of temp-file refs */
		spi->sc_nobput += nuseful; /* incr count of temp-file obs */
		(spi->sc_reflist)->qr_stl = mp->mr_stl;
		(spi->sc_reflist)->qr_nob =  nuseful; /* # of obs for refl */
		(spi->sc_reflist)++;	/* increment reflist pointer */
	  }
 		/* rewind list of observations if not enough good ones found */
	else	spi->sc_oblist = obhold;
}

static void refinp(SCALERUN *spi)
{ /*	This controls the reading-in of the reflections from the input multiref.
	It reads in each reflection, splits it if we're not merging Bijvoets,
	and sticks either the original reflection or the two unmerged refls
	into the memory tanks via a call to "putref".  */
	MERGEDREF	mr, mrbij, mrpri;
	
	spi->sc_tl0 = NOW;
	for (spi->sc_reflist = spi->sc_refbase,
	 spi->sc_oblist = spi->sc_obbase;
	 0 < xgread((void *)&mr, XG_MERGEDREF, spi->sc_fpmi); )
	   {
		if (!(spi->sc_flags & SC_MERGEBJ) &&
		 mr.mr_nobpri && mr.mr_nobbij)
		  {
			splitmr(&mr,&mrpri,&mrbij);
			putref(&mrpri, spi);
			putref(&mrbij, spi);
		  }
		else	putref(&mr, spi);
	   }
	/* compute the mean redundancy.  This will be used in the
	 diffraction-vector stuff if it is high enough */
	spi->sc_overredun = (spi->sc_nrefput > 0) ?
		((short)((spi->sc_nobput) / spi->sc_nrefput)) : 0;
	spi->sc_freq = spi->sc_overredun / SELECTIVITY;
		/* marker for end of list*/
	(spi->sc_reflist)->qr_stl = (spi->sc_reflist)->qr_nob = 0;
	spi->sc_tl1 = NOW;
	printf("Loading reflections into memory required %6d seconds\n",
		(int)(spi->sc_tl1 - spi->sc_tl0));
}

static int redo_headers(SCALERUN *spi)
{ /*	This re-reads the multiref headers */
	register int	i;
	CRYSTAL		*xhp;
	DATARUN		*rhp;

	if (verror(1 + fseek(spi->sc_fpmi, 0L, 0),
	 "rewind multiref to correct <I>")) return ESEEKERR;
	if (EOK != (i = headmem(&(spi->sc_xhdp), &(spi->sc_rhdp), 1)))
		return i;
	xhp = spi->sc_xhdp;	rhp = spi->sc_rhdp;
	if (verror(0 < xgread((void *)xhp, XG_CRYSTAL, spi->sc_fpmi),
		"reread crystal header from input multiref")) return EREADERR;
	for (i = 0; i < xhp->xt_nruns; i++)
	   if (verror(0 < xgread((void *)rhp, XG_DATARUN, spi->sc_fpmi),
		"reread datarun header from input multiref")) return EREADERR;
	return EOK;
}

static void put_means(MERGEDREF *mrp, double yyv[])
{ /*	This does the appropriate integer arithmetic for the y values */
	yyv[0] += 0.49999; if (yyv[0] > 65534) yyv[0] = 65534;
	yyv[1] = 1. / sqrt(yyv[1]) + 0.49999;
	mrp->mr_ibar = yyv[0];	mrp->mr_isig = yyv[0];	return;
}

static void finalrcalcs(MERGEDREF *mrp, SCALERUN *spi)
{ /*	This performs the final calculations that enable us
	to calculate several R-factors and write out the final multiref. */
	register int	i, nused, shind;
	register ABBROB	*aobp;
	int		j, this_is_centric;
	double		fbar, fden, fdenelt, fdif, felt, fmden;
	double		fminus, fplus, fpden, gisq, idif, qf, res, skf, ski;
	double		thisf, thisi, wf, wi, *fp2;
	double		av[10], yyv[2];
	double		sk[MAXOB], sff[MAXOB], swt[MAXOB], tf[MAXOB];
	float		*fp1;
	xgulong		ulsh;
	CRYSTAL		*xhp;
	SHIFT		*tsh;
	RUNWORK		*rup;
	VECPAR		*vpp;

	/* make sure we belong here */
	if (!mrp->mr_nob) return;
	xhp = spi->sc_xhdp;
	this_is_centric = iscentric(mrp->mr_hkl, xhp->xt_intl);
	if (spi->sc_flags & SC_VECREFINED)
		for (j = 0; j < (int)(spi->sc_noutpar); j++) av[j] = 0.;
	yyv[0] = yyv[1] = nused = 0;		/* zero out the sums */
	fbar = fden = 0;
	fplus = fminus = fpden = fmden = 0.;
	res = mrp->mr_stl;
	for (i = 0; i < (int)(mrp->mr_nob); i++) /* compute mean intensity */
	   {
		aobp = &(mrp->mr_ob[i]);
		ulsh = LSH(aobp->abo_runno, 24) |
			LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
		sk[i] = 0;
		if (0 > (shind = find_shift(spi->sc_shlp, ulsh,
		 (int)(spi->sc_nscales)))) continue;
		thisi = aobp->abo_ii;
		tsh = spi->sc_shlp + shind;
		ski = terpscale(aobp->abo_omob, tsh, res);
		if (ski < DFUZZ) continue;
		wi = aobp->abo_sigma;
/* Specific to this version: multiply by the previous cycle's calc. of
	the diffraction-vector stuff */
		if (spi->sc_flags & SC_VECREFINED)
		  {
			rup = spi->sc_rub + (int)(aobp->abo_shift / 200);
			vpp = spi->sc_vshp + (int)(aobp->abo_shift / 200);
			refvec(aobp, rup, (int)spi->sc_flags,
				(int)(spi->sc_noutpar), av);
			for (j = spi->sc_noutpar, qf = 0,
			 fp1 = &(vpp->v_coef[0]), fp2 = &(av[0]);
				j--; fp1++, fp2++) qf += *fp1 * *fp2;
			ski *= qf;
		  }
		if (wi < FUZZ) continue;
		sk[i] = ski;	gisq = ski * ski;
		swt[i] = wi = 1. / (wi * wi +
			thisi * thisi * tsh->shvar / gisq);
		yyv[1] += gisq * wi;
		yyv[0] += ski * thisi * wi;	nused++;
		if (thisi >= 0. && ski >= 0.) /*scaled I must be >0 to get F*/
		  {
			sff[i] = skf = sqrt(ski);	wf = 4. * thisi * wi;
			tf[i] = thisf = sqrt(thisi);
			felt = skf * thisf * wf; fdenelt = skf * skf * wf;
			fbar += felt;		fden += fdenelt;
			if (i < (int)(mrp->mr_nobpri))
			  {
				fplus += felt;	fpden += fdenelt;
			  }
			else {
				fminus += felt;	fmden += fdenelt;
			     }
		  }
	   }
	/* we do the rescaling and the R-factor sums only if the statistics
	are okay, at least 2 observations are ok, and <I> is nonnegative */
	if (nused < 1)	return;	/* can't do anything at all in this case */
	if (yyv[1] < DFUZZ) /* this is a perverse case too */
	  {
		mrp->mr_ibar = 0.;	mrp->mr_isig = 1.;	return;
	  }
	yyv[0] /= yyv[1];	/* compute the mean intensity */
	if (nused < 2) /* exactly one ob: just scale the mean intensity */
	  {
		put_means(mrp, yyv);	return;
	  }
	if (yyv[0] < 0.)
	  {
		yyv[0] = 0.;	put_means(mrp, yyv);	return;
	  }
	/* anytime we get here, we have yy >= 0 and at least two obs,
	 so we can do the R-factor calculations. */
	if (fden > DFUZZ) fbar /= fden;	else fbar = 0.;
	if (fpden > DFUZZ) fplus /= fpden;	else fplus = 0.;
	if (fmden > DFUZZ) fminus /= fmden;	else fminus = 0.;
	if (this_is_centric) (spi->sc_cenrefu)++;
	else
	  {
		if (spi->sc_flags & SC_MERGEBJ)	/* Bijvoet R factor */
		  {
			if (fpden > DFUZZ && fmden > DFUZZ)
			  {
				spi->sc_rfn[4] += fabs(fplus - fminus);
				spi->sc_rfd[4] += fbar;
			  }
		  }
		else
		  {
			mrp->mr_hkl[3] &= 6;
			if ((fden > DFUZZ) && (spi->sc_lastf > -1.) &&
				!hklcmp(spi->sc_lasthkl, mrp->mr_hkl))
			  {
				spi->sc_rfn[4] += fabs(fbar - spi->sc_lastf);
				spi->sc_rfd[4] += 0.5 * (fbar + spi->sc_lastf);
			  }
		  }
		(spi->sc_acenrefu)++;
	  }
	for (i = 0; i < (int)(mrp->mr_nob); i++)
	   if ((ski = sk[i]) > DFUZZ) /* obs must be scaled to be included */
	     {
		aobp = &(mrp->mr_ob[i]);
		wi = swt[i];	thisi = aobp->abo_ii;
		idif = yyv[0] * ski - thisi;
		spi->sc_rfn[0] += wi * idif * idif;
		spi->sc_rfd[0] += wi * thisi * thisi;
		spi->sc_rfn[2] += fabs(yyv[0] - thisi/ski);
		spi->sc_rfd[2] += yyv[0];
		if (this_is_centric)
		  {
			spi->sc_rfn[6] += wi * idif * idif;
			spi->sc_rfd[6] += wi * thisi *thisi;
			(spi->sc_cenobu)++;
		  }
		else {
			spi->sc_rfn[5] += wi * idif * idif;
			spi->sc_rfd[5] += wi * thisi *thisi;
			(spi->sc_acenobu)++;
		  }
		if (thisi < 0) continue; /* skip F calcs if I < 0 */
		wf = 4. * thisi * wi;	thisf = tf[i];
		skf = sff[i];	fdif = fbar * skf - thisf;
		spi->sc_rfn[1] += wf * fdif * fdif;
		spi->sc_rfd[1] += wf * thisf * thisf;
		spi->sc_rfn[3] += fabs(fbar - thisf/skf);
		spi->sc_rfd[3] += fbar;
	     }
	for (i = 4; i--; ) spi->sc_lasthkl[i] = mrp->mr_hkl[i];
	spi->sc_lasthkl[3] &= 06;
	spi->sc_lastf = (fden > DFUZZ) ? fbar : -2.;
	put_means(mrp, yyv);	/* put <I>, sigma(<I>) in the right place */
}

static void sum_rfactor(FILE *fp, double *rf, SCALERUN *spi)
{ /*	This writes a summary of the final rfactors to file pointer fp */
	fprintf(fp,"\n Final statistics on entire dataset:\n");
	fprintf(fp,"              Weighted RMS R-factor on I = %10.7lf\n",
		rf[0]);
	fprintf(fp,"              Weighted RMS R-factor on F = %10.7lf\n",
		rf[1]);
	fprintf(fp," Unweighted Absolute Value R-factor on I = %10.7lf\n",
		rf[2]);
	fprintf(fp," Unweighted Absolute Value R-factor on F = %10.7lf\n",
		rf[3]);
	fprintf(fp," Unweighted Bijvoet Absval R-factor on F = %10.7lf\n",
		rf[4]);
	fprintf(fp,
	 "   Weighted RMS R on I for %5d obs of %4d acentric refls = %10.7lf\n",
		(int)(spi->sc_acenobu), (int)(spi->sc_acenrefu), rf[5]);
	fprintf(fp,
	 "   Weighted RMS R on I for %5d obs of %4d  centric refls = %10.7lf\n",
		(int)(spi->sc_cenobu), (int)(spi->sc_cenrefu), rf[6]);
	fprintf(fp," Final calculations and output required %6d seconds\n",
		(int)(spi->sc_tl1 - spi->sc_tl0));
}

static int final_rfacs(SCALERUN *spi)
{ /*	Do the computations required for the final R-factor determinations. */
	register int	i;
	double		meanobs;
	MERGEDREF	mr, mrbij, mrpri;
	SHIFT		*tsh;
	SHSTAT		*wlp;

	spi->sc_tl0 = NOW;
	printf("Now recalculating <I>, etc. and writing to output\n");
	if (EOK != (i = redo_headers(spi))) /* deal with multiref headers */
	  {
		(void)fclose(spi->sc_fpmi);	spi->sc_fpmi = NULL;
		return i;
	  }
	for (i = NRFACS; i--; ) spi->sc_rfn[i] = spi->sc_rfd[i] = 0.;
	spi->sc_acenobu = spi->sc_cenobu = spi->sc_acenrefu =
		spi->sc_cenrefu = 0;
	if (!(spi->sc_flags & SC_MERGEBJ))
	  {
		spi->sc_lastf = 0.;
		for (i = 4; i--; ) spi->sc_lasthkl[i] = 0;
		while (0 < xgread((void *)&mr, XG_MERGEDREF, spi->sc_fpmi))
		     {
			splitmr(&mr, &mrpri, &mrbij); /* split reflection */
			/* do the R-factor calcs for primaries and Bijvoets */
			finalrcalcs(&mrpri, spi);
			finalrcalcs(&mrbij, spi);
		     }
	  }
	else while (0 < xgread((void *)&mr, XG_MERGEDREF, spi->sc_fpmi))
	 finalrcalcs(&mr, spi); /* final R-factor calcs */
	for (i = 0; i < NRFACS; i++)	/* compute all the R-factors */
	   {
		if (spi->sc_rfd[i] > DFUZZ) spi->sc_rfn[i] /= spi->sc_rfd[i];
		else spi->sc_rfn[i] = 1.;
	   }
	spi->sc_rfn[0] = sqrt(spi->sc_rfn[0]);
	spi->sc_rfn[1] = sqrt(spi->sc_rfn[1]);
	spi->sc_rfn[5] = sqrt(spi->sc_rfn[5]);
	spi->sc_rfn[6] = sqrt(spi->sc_rfn[6]);
	spi->sc_tl1 = NOW;
	/* summarize the R-factor calculations to stdout and the log file */
	sum_rfactor(stdout, spi->sc_rfn, spi);
	sum_rfactor(spi->sc_fpl, spi->sc_rfn, spi);
	/* special case: if spi->sc_maxcycles < 0, we need to put something
	 reasonable into the variance values. We can do this pretty
	 precisely when we've done Wilson scaling; otherwise, we're guessing */
	if (spi->sc_maxcycles < 0)
	  {
		if (spi->sc_flags & SC_WILSON)
		  {
			for (i = 0, tsh = spi->sc_shlp, wlp = spi->sc_shwl;
			 i < spi->sc_nscales; i++, tsh++, wlp++)
			   {
				if (wlp->s_nob > spi->sc_nparpshift)
				  tsh->shvar = spi->sc_rfn[0] /
				   ((double)(wlp->s_nob - spi->sc_nparpshift));
				else tsh->shvar = spi->sc_rfn[0];
			   }
		  }
		else {
			meanobs = ((double)(spi->sc_nobput)) /
				((double)spi->sc_nscales);
			if (meanobs > spi->sc_nparpshift)
				meanobs -= spi->sc_nparpshift;
			else meanobs = 1.;
			for (i = 0, tsh = spi->sc_shlp; i < spi->sc_nscales;
			 i++, tsh++) tsh->shvar = spi->sc_rfn[0] / meanobs;
		     }
	  }
	return EOK;
}

static void mrg_tb(SCALERUN *spi)
{ /*	Merge the scale factors for the top & bottom of the detector */
	int		i, nob, odd;
	xgulong		ulsh;
	QOB		*obp, *obend;
	SHIFT		*shv, *shq;

/* first, round all the scaling group numbers to the nearest
	even value. */
	for (spi->sc_reflist = spi->sc_refbase,
	 spi->sc_oblist = spi->sc_obbase;
	 (nob = spi->sc_reflist->qr_nob);
	 spi->sc_reflist++, spi->sc_oblist += nob)
	   {
		obend = spi->sc_oblist + nob;
		for (obp = spi->sc_oblist; obp < obend; obp++)
			obp->qo_shift &= (u_short)(65534);
	   }
	/* next, take the existing scaling group scale values
	 and average the top and bottom values.  */
	for (i = 0, shv = spi->sc_shlp; i < spi->sc_nscales; i++, shv++)
	   if (!(shv->shval % 2)) /* is this an even-numbered shift? */
	     {	/* yes.  See if we can find the corresponding odd shift */
		ulsh = LSH(shv->runno, 24) | LSH(shv->chamno, 16) | shv->shval;
		odd = find_shift(spi->sc_shlp, ulsh + 1L,
		 (int)(spi->sc_nscales));
		/* if odd >= 0 we found it.  Average the two. */
		if (0 <= odd)
		  {
			shq = spi->sc_shlp + odd;
			shv->shg = 0.5 * (shv->shg + shq->shg);
			shv->sha = 0.5 * (shv->sha + shq->sha);
			shv->shb = 0.5 * (shv->shb + shq->shb);
			shv->shvar = 0.3 * (shv->shvar + shq->shvar);
		  }
	     }
}

static void dup_shifts(SCALERUN *spi)
{ /*	Write out values in the even-numbered shift in the odd groups as well*/
	int	i, odd;
	xgulong	ulsh;
	SHIFT	*shq, *shv;

	for (i = 0, shq = spi->sc_shlp; i < spi->sc_nscales; i++, shq++)
	   if (!(shq->shval % 2)) /* is this an even-numbered shift? */
	     { /* yes.  See if we can find the corresponding odd shift */
		ulsh = LSH(shq->runno, 24) | LSH(shq->chamno, 16) | shq->shval;
		odd = find_shift(spi->sc_shlp, ulsh + 1L,
		 (int)(spi->sc_nscales));
		/* if odd >= 0 we found it.  Set the odd one == to the even */
		if (0 <= odd)
		  {
			shv = spi->sc_shlp + odd;
			shv->shg = shq->shg;	shv->sha = shq->sha;
			shv->shb = shq->shb;
			shv->shvar = shq->shvar = 2. * shq->shvar;
		  }
	     }
}

static void sumscalekrun(FILE *fp, SCALERUN *spi)
{ /*	This writes a summary of a run params for scalek to the stream fp */
	static char	*flist[] =
		{"multiref", "scaleout", "xlog", "scalein", "uspline", NULL};
	static char	ftypes[] = {1, 2, 2, 1, 1, 0 };

	ftypes[3] = spi->sc_flags & SC_INPUTSCALE;
	ftypes[4] = ((spi->sc_flags & SC_VECREFINED) ? 1 : 0);
	header(fp, REDUCTION, flist, ftypes);
	fputs("\tSummary of program defaults and flags:\n",fp);
	fprintf(fp,     "%s Number of Cycles to Perform:\t\t%5d\n",
		(spi->sc_flags & SC_KEEPCYC ? "       " : "Maximum"),
		spi->sc_maxcycles);
	fprintf(fp,"         Min number of replications:\t\t%5d\n",
		spi->sc_nrepmin);
	fprintf(fp,"Min f-squared value used in scaling:\t\t%5d\n",
		spi->sc_imin);
	fprintf(fp,"Number of parameters/shift to calculate:\t%5d\n",
		spi->sc_nparpshift);
	if (!(spi->sc_flags & SC_KEEPCYC))
	  fprintf(fp,
	    "Cycling will terminate if largest scale shift\t< %9.5f\n",
		(CONV_RESC * spi->sc_convcrit));
	fputs("Initial scale factors will be ",fp);
	if (spi->sc_flags & SC_INPUTSCALE) fprintf(fp,
	 "taken from file %s\n",trname("scalein"));
	else if (spi->sc_flags & SC_WILSON)
		fprintf(fp,"computed based on Wilson statistics\n");
	else	fprintf(fp,"set to 1.\n");
	if ((!(spi->sc_flags & SC_INPUTSCALE)) ||
	 (spi->sc_flags & SC_ALLOBS)) fprintf(fp,
 "All observations of multiply-observed reflections will be included\n");
	fprintf(fp,"Bijvoets will%sbe merged\n", (spi->sc_flags & SC_MERGEBJ ?
		" " : " NOT "));
	if (spi->sc_flags & SC_ALLOFDET) fprintf(fp,
		 "Top and bottom of detector will be scaled together\n");
	fprintf(fp,"Logfile output will be %s\n",
	 (spi->sc_flags & SC_VERBOSE) ?  "verbose" : "brief");
	if (spi->sc_flags & SC_DIFFRACTION) fprintf(fp,
	  "Scaling will be based on diffraction-vector parameters%s",
	   (spi->sc_flags & SC_DIFF_NOS2) ?
		" with the s^2 term == zero\n" : "\n");
	else if (spi->sc_flags & SC_EXITBEAM) fprintf(fp,
	  "Scaling will be based on exit-beam parameters\n");
	if (spi->sc_appshift != 1.) fprintf(fp,
	  "At each cycle %6.3f * the calculated shift will be applied\n",
		spi->sc_appshift);
	fprintf(fp,
	 "Variances used in scaling WILL %sbe augmented by (I*sigma(g))^2\n",
		(spi->sc_flags & SC_FLATVARIANCE) ? "NOT " : " ");
	if (spi->sc_flags & SC_NONLINEAR) fprintf(fp,
	 "Nonlinear refinement will be used rather than linear\n");
	else if (spi->sc_flags & SC_SWITCH) fprintf(fp,
	 "Linear refinements used until Rvalue increases; then nonlinear\n");
	else fprintf(fp,
	 "Linear least-squares refinements to be used throughout\n");
}

static void put_shifts(FILE *fp, SHIFT *ish, unsigned nsca,
 VECPAR *vsp, unsigned nru)
{ /*	This writes the current list of scale factors to the output file
	pointed to by fp.  In the current version this is a printable file. */
	register SHIFT	*tsh;
	unsigned int	i;
	VECPAR		*vpp, *vpe;

	for (i = 0, tsh = ish; i < nsca; i++, tsh++)
	   fprintf(fp,"%7d %2d %2d %9.5f %9.5f %9.5f %12.5e\n",
		tsh->shval, tsh->chamno, tsh->runno,
		tsh->shg, tsh->sha, tsh->shb, tsh->shvar);
	if (NULL != vsp)
	  for (vpp = vsp, vpe = vsp + nru; vpp < vpe; vpp++)
	     {
		fprintf(fp, "%c%6d", (sichar)vpp->v_ptype, vpp->v_ms);
		for (i = 0; i < MAXVECPARS; i++) fprintf(fp, "%6.3f",
			vpp->v_coef[i]);
		fprintf(fp, "\n");
	     }
	return;
}

int sccycles(SCALERUN *spi)
{ /* This goes through some scaling cycles */
	int		i, norders, nsca;
	FILE		*fpso;
	SHIFT		*shp;

	norders = 3;	/* 3 orders (0, 1, 2) for now */
	spi->sc_noutpar = (spi->sc_flags & SC_DIFFRACTION ? 
		(norders * (norders+1) * (norders + 2) / 6) :
		(spi->sc_flags & SC_EXITBEAM ? (norders * norders) : 0));
	if (spi->sc_flags & SC_DIFF_NOS2) (spi->sc_noutpar)--;
	/*process multiref file headers*/
	if (EOK != (i = scheads(spi))) return scend(i, spi);
		/* read spline if needed to get pix->cm conversions */
	if ((spi->sc_flags & SC_VECREFINED) && (NULL == gspline(5, &i)))
		return scend(i, spi);
	if (NULL == (spi->sc_fpl = propen("xlog"))) spi->sc_fpl = stdout;
	sumscalekrun(stdout, spi);	/* summarize run conditions */
	if (stdout != spi->sc_fpl) sumscalekrun(spi->sc_fpl, spi);
	if (spi->sc_flags & SC_INPUTSCALE)
	  {
		shp = get_shifts(spi->sc_shlp, spi->sc_vshp, &nsca);
		spi->sc_nscales = nsca;	spi->sc_shlp = shp;
	  }
	else spi->sc_nscales = 0;
	if (!(spi->sc_nscales)) spi->sc_flags &= ~SC_INPUTSCALE;
	spi->sc_nrefin = spi->sc_nobin = spi->sc_nrefput = spi->sc_nobput =
	 spi->sc_acenobu = spi->sc_acenrefu = spi->sc_cenobu =
	 spi->sc_cenrefu = 0;
	if ((!(spi->sc_flags & SC_INPUTSCALE) || (spi->sc_flags & SC_ALLOBS))
	 && (EOK != (i = make_shift_list(spi)))) scend(i, spi);
	refinp(spi);	/* read input refls & stick in temp storage */
	if (spi->sc_flags & SC_ALLOFDET) mrg_tb(spi);
	/* establish smoothing values to allow for linearly splined
	scaling parameters within data runs */
	del_omegas(spi);
	/* get initial scale factors from Wilson stats if no .si file */
	if (!(spi->sc_flags & SC_INPUTSCALE))
	  { /* get initial scale factors from Wilson stats or set 'em to 1.*/
	    if (spi->sc_flags & SC_WILSON)	wilscale(spi);
	    else for (i = 0, shp = spi->sc_shlp;
		i < spi->sc_nscales; i++, shp++)
		    {
			shp->shg = 1.;	shp->sha = shp->shb = 0.;
		    }
	  }
	sum_input(spi->sc_fpl, spi);  /* summarize inputs */
	sum_input(stdout, spi);
	xeflush(spi->sc_fpl);		/* make sure contents are flushed */
	if (spi->sc_flags & SC_NONLINEAR) scminim(0, spi);
	else {
		for (spi->sc_cycno = 0; spi->sc_cycno <= spi->sc_maxcycles;
		 (spi->sc_cycno)++) if (do_cycle(spi)) break;
	     }
	if (spi->sc_flags & SC_ALLOFDET) dup_shifts(spi);
	/* we calculate the final R-values before opening SCALEOUT so that in
	 the peculiar case of spi->sc_maxcycles == -1 it comes out right */
	i = final_rfacs(spi);
	/* only open SCALEOUT at this point; thus we can set SCALEIN and
	SCALEOUT to the same filename without getting into trouble.
	Also we avoid writing out empty files.  */
	if (qerror((AVOID *)(fpso = xgopen("scaleout", "w")),
		"open output shift scale file")) return scend(EOPENERR, spi);
	 /*output scale factors to SCALEOUT file*/
	if (spi->sc_flags & SC_VECREFINED) put_shifts(fpso, spi->sc_shlp,
	 (unsigned)(spi->sc_nscales), spi->sc_vshp, spi->sc_ndvgroups);
	else	put_shifts(fpso, spi->sc_shlp,
	 (unsigned)(spi->sc_nscales), (VECPAR *)NULL, 0);
	(void)fclose(fpso);	fpso = NULL;	return EOK;
}

/* end of scrio.c */
