/* cursestub.c -- stubs for Curses applications for non-Curses environments

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	12 Oct 2006: initial code, based trivially on utilities/mrefutil.c

   Purpose: providing replacement stubs for edcenmain, ewmmain, refmain,
	rejmain, and setupmain.c in non-curses environments.
   Cross references: all sources in ../curses

*************************************************************************/
#include	"xgpostint.h"

 /* declarations for routines found elsewhere */
extern int	filtercmain(int argc, char *argv[]);
extern int	cleanmmain(int argc, char *argv[]);
extern int	ref1main(int argc, char *argv[]);
extern int	rejswitch(int argc, char *argv[]);
 /* declarations for routines located here */
extern int	edcenmain(int argc, char *argv[]);
extern int	ewmmain(int argc, char *argv[]);
extern int	refmain(int argc, char *argv[]);
extern int	rejmain(int argc, char *argv[]);
extern int	setupmain(int argc, char *argv[]);

int edcenmain(int argc, char *argv[])
{ /* Mainline for editcen stub */

	printf(
	 "\nWarning: Curses unavailable; editcen is replaced by filtercen\n");
	return filtercmain(argc, argv);
}

int ewmmain(int argc, char *argv[])
{ /* Mainline for ewmmain stub */

	printf(
	 "\nCurses unavailable; ewmask is replaced by cleanmask\n");
	return cleanmmain(argc, argv);
}

int refmain(int argc, char *argv[])
{ /* Mainline for refmain stub */

	printf(
	 "\nCurses unavailable; refine is replaced by refine1\n");
	return ref1main(argc, argv);
}

int rejmain(int argc, char *argv[])
{ /*	rejmain() in this version does non-Curses rejections.
	The only arguments to the program in this version are the
	Booleans and switches. */
	int		retv;
	REJECTRUN	*rjp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort*/
	/* initialize the REJECTRUN structure */
	rjp = managerjp(02467);
	rjp->rj_flags = RJ_2_MAXIOS | RJ_BATCH; /* default is all flag bits
	 off, batch mode turned on; two-ob refls: keep larger I/sigma */
	/* Examine arguments.  Those beginning with a '-' are interpreted
	 as switches and processed with 'rejswitch'.
	Any other arguments are ignored for the moment.  */
	if (EOK != (retv = rejswitch(argc, argv))) return retv;
	/* we cannot do RJ_HUMANANY and we must run in batch mode: */
	rjp->rj_flags |= RJ_BATCH; rjp->rj_flags &= ~(RJ_HUMANANY);
	if (NULL == (rjp->rj_mrp = (MERGEDREF *)callop(1, sizeof (MERGEDREF),
		"rjp->rj_mrp in scmargs"))) return EALLOC;
	return autoreject();
}

int setupmain(int argc, char *argv[])
{ /* Mainline for setupmain stub */

	printf(
	 "\nCurses unavailable on this system, so setup can't run; sorry.\n");
	return 0; 
}

/* end of cursestub.c */
