/* edit_bord.c -- Editor for XGEN's list of active pixels

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	12 Oct 2005: eliminate static globals
	11 May 2005: check return status from xgjournal
	28 Aug 2004: routines used only here are declared as static; cosmetics
	 2 Aug 2004: add new arg to find_dimension()
	26 Jan 2004: convert a few static globals to local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: replace extern global variables with static globals
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: eboperate() is now a void.
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	14 Jun 1994: minor reorganization to accomodate GUI calls
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: minor bugfixes, reorganized libraries
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	 1 Jun 1989: add "shrink" option (like "-s" in BORDER)
	29 Mar 1988: reorganize and clean up.
	 1 Oct 1987: add "OCIR" and "OBOX" commands
	30 Apr 1987: original code.
	
   Purpose: primitive line-oriented database editor for the `amask' file,
	which the bitmap that defines which pixels of the detector are active.
   Cross references: ../prerefine/bordmain.c, ../xglib/actmask.c

 For the moment "editbord" will just take arguments from the terminal;
 we'll make it a window-based software package eventually.
 The routines performing the modifications to the database are actually
 in ../lib/actmask.c so that the static globals can do their thing.
*************************************************************************/
#define		NEED_ERRNO	1
#include	"xgen.h"
#define		ON		1
#define		OFF		0
#define		TOGGLE		2
#define		DOSHRINK	4
/* Values of the "shape" variable: */
#define		COLUMN		2
#define		ROW		4
#define		BOX		010
#define		OBOX		020
#define		CIRCLE		040
#define		OCIR		0100
#define		TRI		0200
#define		OTRI		0400
#define		QUAD		01000
#define		OQUAD		02000

#define		NO_ACTION	4
#define		MAX_ARGS	9
#define		CHARS_PER_ARG	20

 typedef struct {
		int		eb_ht, eb_wid;
		unsigned	eb_ncargs;
		char		**eb_cargs;
		u_char		**eb_amask;
		FILE		*eb_fpl;
		} EBWORK;

extern int ebordmain(int argc, char *argv[]);

static char *get_ebcom(EBWORK *ebw)
{ /*	This prompts the user for a command and returns a lowercase version of
	the command given by the user.  It also stuffs any other strings
	given by the user into ebw->eb_cargs.  */
	register int	i, j;
	register char	**dcp;
	int		lnar;
	static char	cmd[CHARS_PER_ARG], ebl[80];

	fprintf(stderr, "eb> "); for (j = CHARS_PER_ARG; j--; ) cmd[j] = '\0';
	dcp = ebw->eb_cargs;
	for (i = MAX_ARGS; i--; ) for (j = CHARS_PER_ARG; j--; )
		dcp[i][j] = '\0';
	if (NULL == fgets(ebl, 78, stdin))
	  {
		strcpy(cmd, "ex");	return &(cmd[0]);
	  }
	(void)sscanf(ebl, "%s %s %s %s %s %s %s %s %s %s",
		cmd, *dcp, *(dcp + 1), *(dcp + 2), *(dcp + 3), *(dcp + 4),
		*(dcp + 5), *(dcp + 6), *(dcp + 7), *(dcp + 8));
	lowercaseify(cmd);
	for (lnar = 0; lnar < MAX_ARGS && **dcp != '\0'; lnar++, dcp++)
		lowercaseify(*dcp);
	ebw->eb_ncargs = lnar;	return &(cmd[0]);
}

static char *describe(char actval)
{ /* return a string indicating a change in variable `actval' */
	register int	retindex;
	static char	*retstring[] = { "added", "deleted", "toggled" } ;

	retindex = (actval == ON ? 0 : (actval == OFF ? 1 : 2));
	return	retstring[retindex];
}

static void log_change(char action, int shape, int ia[], FILE *fp)
{ /*	This summarizes a change in the mask to the stream fp. */
	switch (shape)
	      {
		case	BOX:	fprintf(fp,
		 " box from [x,y]=[%4d,%4d] to [%4d,%4d] %s\n", ia[0],
			ia[1], ia[2], ia[3], describe(action)); break;
		case	CIRCLE:	fprintf(fp,
		 "Circle of radius %4d centered at [%4d,%4d] %s\n", ia[2],
			ia[0], ia[1], describe(action));	break;
		case	COLUMN:	fprintf(fp,
		 " %3d columns centered on column %4d %s\n",
			ia[1], ia[0], describe(action));	break;
		case	ROW:	fprintf(fp,
		 " %3d rows centered on row %4d %s\n",
			ia[1], ia[0], describe(action));	break;
		case	TRI:	fprintf(fp,
		 " triangle bounded by [%4d,%4d],[%4d,%4d],[%4d,%4d] %s\n",
			ia[0], ia[1], ia[2], ia[3], ia[4], ia[5],
			describe(action)); break;
		case	QUAD:	fprintf(fp,
 " quadrilateral bounded by [%4d,%4d],[%4d,%4d],[%4d,%4d],[%4d,%4d] %s\n",
			ia[0], ia[1], ia[2], ia[3], ia[4], ia[5], ia[6], ia[7],
			describe(action)); break;
		case	OBOX:	fprintf(fp,
 " Points outside of box from [x,y]=[%4d,%4d] to [%4d,%4d] %s\n", ia[0],
			ia[1], ia[2], ia[3], describe(action));	break;
		case	OCIR:	fprintf(fp,
 "Outside of circle of radius %4d centered at [%4d,%4d] %s\n", ia[2],
			ia[0], ia[1], describe(action));	break;
		case	OTRI:	fprintf(fp,
 "Outside of triangle bounded by [%4d,%4d],[%4d,%4d],[%4d,%4d] %s\n",
			ia[0], ia[1], ia[2], ia[3], ia[4], ia[5],
			describe(action)); break;
		case	OQUAD:	fprintf(fp,
 "Outside of quadril. bounded by [%4d,%4d],[%4d,%4d],[%4d,%4d],[%4d,%4d] %s\n",
			ia[0], ia[1], ia[2], ia[3], ia[4], ia[5], ia[6], ia[7],
			describe(action)); break;
	     }

}

static void chg_pixel(int x, int y, char changecode, EBWORK *ebw)
{ /*	This changes the activation status of the pixel (x, y).
	If changecode == OFF it turns the pixel off; if changecode == ON
	it turns the pixel on; if changecode == TOGGLE it toggles it.	*/
	u_char		usebit, *cp, **dcp;

	usebit = 1 << (x % 8);	dcp = ebw->eb_amask;
	cp = *(dcp + y) + x / 8;
	switch (changecode)
	  {
		case	OFF:	*cp &= ~usebit;	break;
		case	ON:	*cp |= usebit;	break;
		case	TOGGLE:	if ((*cp) & usebit) *cp &= ~usebit;
				else *cp |= usebit;	break;
	  }
}

static void chg_box(int xmn, int ymn, int xmx, int ymx,
	char changecode, EBWORK *ebw)
{ /*	This changes pixels in a rectangular box from [xmn,ymn]
	to [xmx, ymx]. */
	register int	x, y;

	if (xmx < xmn)
	  {
		x = xmn; xmn = xmx; xmx = x;
	  }
	if (ymx < ymn)
	  {
		y = ymn; ymn = ymx; ymx = y;
	  }
	for (x = xmn; x <= xmx; x++)
		for (y = ymn; y <= ymx; y++) chg_pixel(x, y, changecode, ebw);
}

static void chg_obox(int xmn, int ymn, int xmx, int ymx,
	char changecode, EBWORK *ebw)
{ /*	This changes pixels outside a rectangular box from [xmn,ymn]
	to [xmx, ymx]. */
	register int	x, y;

	if (xmx < xmn)
		{
			x = xmn; xmn = xmx; xmx = x;
		}
	if (xmx >= ebw->eb_wid) xmx = ebw->eb_wid - 1;
	if (xmn < 0) xmn = 0;
	if (ymx < ymn)
		{
			y = ymn; ymn = ymx; ymx = y;
		}
	if (ymx >= ebw->eb_ht) ymx = ebw->eb_ht - 1;
	if (ymn < 0) ymn = 0;
	for (x = 0; x < xmn; x++) for (y = 0; y < ebw->eb_ht; y++)
		chg_pixel(x, y, changecode, ebw);
	for (x = xmx + 1; x < ebw->eb_wid; x++)
	  for (y = 0; y < ebw->eb_ht; y++) chg_pixel(x, y, changecode, ebw);
	for (x = xmn; x <= xmx; x++)
	   {
		for (y = 0; y < ymn; y++) chg_pixel(x, y, changecode, ebw);
		for (y = ymx + 1; y < ebw->eb_ht; y++)
			chg_pixel(x, y, changecode, ebw);
	   }
}

static void chg_circle(int xcent, int ycent, int radius,
	char changecode, EBWORK *ebw)
{ /*	This changes pixels in a circle of radius `radius' about
	the pixel position (xcent, ycent).  It turns off pixels if
	`changecode' == OFF; it turns them on if `changecode' == ON;
	it toggles them if `changecode' == TOGGLE */
	register int	x, y;
	int		dx, dxsq, dy, maxx, maxy, minx, miny, radsq;

	minx = xcent - radius;	if (minx < 1) minx = 1;
	maxx = xcent + radius;	radsq = radius * radius;
	if (maxx >= ebw->eb_ht-1) maxx = ebw->eb_ht-2;
	miny = ycent - radius;	if (miny < 1) miny = 1;
	maxy = ycent + radius;
	if (maxy >= ebw->eb_ht - 1) maxy = ebw->eb_ht - 2;
	for (x = minx, dx = minx - xcent; x <= maxx; x++, dx++)
	   {
		dxsq = dx * dx;
		for (y = miny, dy = miny - ycent; y <= maxy; y++, dy++)
		  if (dxsq + dy * dy < radsq) chg_pixel(x, y, changecode, ebw);
	   }
}

static void chg_ocircle(int xcent, int ycent, int radius,
 char changecode, EBWORK *ebw)
{ /*	This changes pixels outside a circle of radius `radius' about
	the pixel position (xcent, ycent).  It turns off pixels if
	`changecode' == OFF; it turns them on if `changecode' == ON;
	it toggles them if `changecode' == TOGGLE */
	register int	x, y;
	int		dx, dxsq, dy, maxx, maxy, minx, miny, radsq;

	minx = xcent - radius;	if (minx < 1) minx = 1;
	maxx = xcent + radius;	radsq = radius * radius;
	if (maxx >= ebw->eb_ht - 1) maxx = ebw->eb_ht - 2;
	miny = ycent - radius;	if (miny < 1) miny = 1;
	maxy = ycent + radius;
	if (maxy >= ebw->eb_ht - 1) maxy = ebw->eb_ht - 2;
	/* first, wipe out the box outside the circle */
	chg_obox(minx, miny, maxx, maxy, changecode, ebw);
	for (x = minx, dx = minx - xcent; x <= maxx; x++, dx++)
	   {
		dxsq = dx * dx;
		for (y = miny, dy = miny - ycent; y <= maxy; y++, dy++)
		   if (dxsq + dy * dy > radsq)
			chg_pixel(x, y, changecode, ebw);
	   }
}

static int isintri(int xi, int yi, int xx0, int yy0,
 int xx1, int yy1, int xx2, int yy2)
{ /*	Returns 1 iff the point (xi,yi) is inside the triangle bounded by
	(xx0, yy0), (xx1, yy1), (xx2, yy2). It does this by calculating
	three cross products and seeing if they have the same sign. */
	int	dx0, dy0, dx1, dy1, dx2, dy2, r01, r12, r20;

	dx0 = xx0 - xi;	dy0 = yy0 - yi;
	dx1 = xx1 - xi;	dy1 = yy1 - yi;
	dx2 = xx2 - xi;	dy2 = yy2 - yi;
	r01 = dx0 * dy1 - dx1 * dy0;
	r12 = dx1 * dy2 - dx2 * dy1;
	r20 = dx2 * dy0 - dx0 * dy2;
	if (r01 > 0) return ((r12 > 0) && (r20 > 0)) ? 1 : 0;
	else if (r01 < 0) return ((r12 < 0) && (r20 < 0)) ? 1 : 0;
	return 0;
}

static int isconvex(int x0, int yy0, int x1, int yy1,
	int x2, int y2, int x3, int y3)
{ /*	This returns 1 iff. the quadrilateral bounded by
	(x0,yy0), (x1,yy1), (x2,y2), (x3,y3) is convex. A quadrilateral is
	convex iff. none of its vertices lies within the triangle bounded
	by the other three vertices */
	if (isintri(x0, yy0, x1, yy1, x2, y2, x3, y3)) return 0;
	if (isintri(x1, yy1, x2, y2, x3, y3, x0, yy0)) return 0;
	if (isintri(x2, y2, x3, y3, x0, yy0, x1, yy1)) return 0;
	if (isintri(x3, y3, x0, yy0, x1, yy1, x2, y2)) return 0;
	return 1;
}

static int farfrom3(int xx0, int yy0, int xx1, int yy1,
	int xx2, int yy2, int xx3, int yy3)
{ /*	This returns 0, 1, or 2 depending on which of the three points
	(xx0,yy0), (xx1,yy1), or (xx2,yy2) is farthest from (xx3, yy3). */
	int	mx, my, miss, misst, far;

	far = 0;
	mx = xx3 - xx0;	mx *= mx;
	my = yy3 - yy0;	my *= my;	miss = mx * mx + my * my;
	mx = xx3 - xx1; mx *= mx;
	my = yy3 - yy1; my *= my;	misst = mx * mx + my * my;
	if (misst > miss)
	  {
		miss = misst;	far = 1;
	  }
	mx = xx3 - xx2; mx *= mx;
	my = yy3 - yy2; my *= my;	misst = mx * mx + my * my;
	if (misst > miss)
	  {
		miss = misst;	far = 2;
	  }
	return far;
}

static int isinquad(int xi, int yi, int xx0, int yy0, int xx1, int yy1,
	int xx2, int yy2, int xx3, int yy3, int far)
{ /*	Returns 1 iff the point (xi,yi) is inside the quadrilateral
	bounded by (xx0, yy0), (xx1, yy1), (xx2, yy2), (xx3, yy3).
	If the quadrilateral is not convex it returns 0 by definition. */

	/* first make sure that the quadrilateral is convex */
	if (!isconvex(xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3)) return 0;
	/* next, define a triangle bounded by the first three vertices and
	 see if our point lies inside it. */
	if (isintri(xi, yi, xx0, yy0, xx1, yy1, xx2, yy2)) return 1;
	/* next, define a triangle bounded by the vertex we haven't used
	 yet, together with the two vertices closest to it. See if our
	 point lies inside it. */
	switch (far)
	      {
		case	0:
			return isintri(xi, yi, xx1, yy1, xx2, yy2, xx3, yy3);
		case	1:
			return isintri(xi, yi, xx2, yy2, xx3, yy3, xx0, yy0);
		case	2:	default:
			return isintri(xi, yi, xx3, yy3, xx0, yy0, xx1, yy1);
	      }
}

static void chg_tri(int x0, int yy0, int x1, int yy1, int x2, int y2,
	char changecode, EBWORK *ebw)
{ /*	This changes pixels in a triangle bounded by [x0,yy0], [x1,yy1],
	[x2, y2]. */
	register int	x, y;
	int		xmi, xma, ymi, yma;

	/* find the limits of the rectangle that contains the triangle */
	xmi = x0;	if (x1 < xmi) xmi = x1;	if (x2 < xmi) xmi = x2;
	if (xmi < 1) xmi = 1;
	if (xmi > ebw->eb_wid - 1) xmi = ebw->eb_wid - 1;
	ymi = yy0;	if (yy1 < ymi) ymi = yy1;	if (y2 < ymi) ymi = y2;
	if (ymi < 1) ymi = 1;	if (ymi > ebw->eb_ht - 1) ymi = ebw->eb_ht - 1;
	xma = x0;	if (x1 > xma) xma = x1;	if (x2 > xma) xma = x2;
	if (xma < 1) xma = 1;
	if (xma > ebw->eb_wid - 1) xma = ebw->eb_wid - 1;
	yma = yy0;	if (yy1 > yma) yma = yy1;	if (y2 > yma) yma = y2;
	if (yma < 1) yma = 1;	if (yma > ebw->eb_ht - 1) yma = ebw->eb_ht - 1;
	/* step through the points in the rectangle. If a point is
	 in the triangle, it gets modified; if not, it doesn't */
	for (x = xmi; x <= xma; x++) for (y = ymi; y <= yma; y++)
		if (isintri(x, y, x0, yy0, x1, yy1, x2, y2))
				chg_pixel(x, y, changecode, ebw);
}

static void chg_otri(int x0, int yy0, int x1, int yy1, int x2, int y2,
	char changecode, EBWORK *ebw)
{ /*	This changes pixels outside a triangular box bounded by
	[x0,yy0], [x1,yy1], [x2,y2] */
	register int	x, y;
	int		xmi, xma, ymi, yma;

	/* find the limits of the rectangle that contains the triangle */
	xmi = x0;	if (x1 < xmi) xmi = x1;	if (x2 < xmi) xmi = x2;
	if (xmi < 1) xmi = 1;
	if (xmi > ebw->eb_wid - 1) xmi = ebw->eb_wid - 1;
	ymi = yy0;	if (yy1 < ymi) ymi = yy1;	if (y2 < ymi) ymi = y2;
	if (ymi < 1) ymi = 1;	if (ymi > ebw->eb_ht - 1) ymi = ebw->eb_ht - 1;
	xma = x0;	if (x1 > xma) xma = x1;	if (x2 > xma) xma = x2;
	if (xma < 1) xma = 1;
	if (xma > ebw->eb_wid - 1) xma = ebw->eb_wid - 1;
	yma = yy0;	if (yy1 > yma) yma = yy1;	if (y2 > yma) yma = y2;
	if (yma < 1) yma = 1;	if (yma > ebw->eb_ht - 1) yma = ebw->eb_ht - 1;
	/* change pixels that lie to the left of the bounding rectangle */
	for (x = 0; x < xmi; x++) for (y = 0; y < ebw->eb_ht; y++)
		chg_pixel(x, y, changecode, ebw);
	/* change pixels that lie to the right of the bounding rectangle */
	for (x = xma + 1; x < ebw->eb_wid; x++)
	   for (y = 0; y < ebw->eb_ht; y++) chg_pixel(x, y, changecode, ebw);
	/* change pixels that lie below and above the bounding rectangle */
	for (x = xmi; x <= xma; x++)
	   {
		for (y = yma + 1; y < ebw->eb_ht; y++)
			chg_pixel(x, y, changecode, ebw);
		for (y = 0; y < ymi; y++) chg_pixel(x, y, changecode, ebw);
	   }
	/* change pixels that lie inside the bounding rectangle but
	 outside the triangle itself */
	for (x = xmi; x <= xma; x++) for (y = ymi; y <= yma; y++)
		if (!isintri(x, y, x0, yy0, x1, yy1, x2, y2))
			chg_pixel(x, y, changecode, ebw);
}

static void log_nonconvex(FILE *fp)
{ /*	Reports to stream fp that a quadrilateral isn't convex */
	fprintf(fp,
	 "The specified quadrilateral is not convex. No action taken.\n");
}

static int chg_quad(int x0, int yy0, int x1, int yy1, int x2, int y2,
	int x3, int y3, char changecode, EBWORK *ebw)
{ /*	This changes pixels in a quadrilateral bounded by
	[x0,yy0], [x1,yy1], [x2, y2], [x3,y3]. */
	register int	x, y;
	int		xmi, xma, ymi, yma, far;

	/* first make sure that the quadrilateral is a legal one.
	 If not, we don't change anything. */
	if (!isconvex(x0, yy0, x1, yy1, x2, y2, x3, y3))
	  {
		if (ebw->eb_fpl != stdout) log_nonconvex(ebw->eb_fpl);
		log_nonconvex(stdout);	return 1;
	  }
	/* find the vertex of the quadrilateral that is farthest from
	 the one labeled as vertex 3 */
	far = farfrom3(x0, yy0, x1, yy1, x2, y2, x3, y3);
	/* find the limits of the rectangle that contains the quadirlateral */
	xmi = x0;	if (x1 < xmi) xmi = x1;	if (x2 < xmi) xmi = x2;
	if (x3 < xmi) xmi = x3;
	if (xmi < 1) xmi = 1;
	if (xmi > ebw->eb_wid - 1) xmi = ebw->eb_wid - 1;
	ymi = yy0;	if (yy1 < ymi) ymi = yy1;	if (y2 < ymi) ymi = y2;
	if (y3 < ymi) ymi = y3;
	if (ymi < 1) ymi = 1;	if (ymi > ebw->eb_ht - 1) ymi = ebw->eb_ht - 1;
	xma = x0;	if (x1 > xma) xma = x1;	if (x2 > xma) xma = x2;
	if (xma < 1) xma = 1;
	if (xma > ebw->eb_wid - 1) xma = ebw->eb_wid - 1;
	if (x3 > xma) xma = x3;
	yma = yy0;	if (yy1 > yma) yma = yy1;	if (y2 > yma) yma = y2;
	if (y3 > yma) yma = y3;
	if (yma < 1) yma = 1;	if (yma > ebw->eb_ht - 1) yma = ebw->eb_ht - 1;
	/* go through the points in the bounding rectangle.
	 Those that lie inside the quadrilateral have their state
	 changed; those that do not are left unchanged. */
	for (x = xmi; x <= xma; x++) for (y = ymi; y <= yma; y++)
		if (isinquad(x, y, x0, yy0, x1, yy1, x2, y2, x3, y3, far))
				chg_pixel(x, y, changecode, ebw);
	return 0;
}

static int chg_oquad(int x0, int yy0, int x1, int yy1, int x2, int y2,
	int x3, int y3, char changecode, EBWORK *ebw)
{ /*	This changes pixels outside a triangular box bounded by
	[x0,yy0], [x1,yy1], [x2,y2] */
	register int	x, y;
	int		xmi, xma, ymi, yma, far;

	/* first make sure that the quadrilateral is a legal one.
	 If not, we don't change anything. */
	if (!isconvex(x0, yy0, x1, yy1, x2, y2, x3, y3))
	  {
		if (ebw->eb_fpl != stdout) log_nonconvex(ebw->eb_fpl);
		log_nonconvex(stdout);	return 1;
	  }
	/* find the vertex of the quadrilateral that is farthest from
	 the one labeled as vertex 3 */
	far = farfrom3(x0, yy0, x1, yy1, x2, y2, x3, y3);
	xmi = x0;	if (x1 < xmi) xmi = x1;	if (x2 < xmi) xmi = x2;
	if (x3 < xmi) xmi = x3;
	if (xmi < 1) xmi = 1;
	if (xmi > ebw->eb_wid - 1) xmi = ebw->eb_wid - 1;
	ymi = yy0;	if (yy1 < ymi) ymi = yy1;	if (y2 < ymi) ymi = y2;
	if (y3 < ymi) ymi = y3;
	if (ymi < 1) ymi = 1;	if (ymi > ebw->eb_ht - 1) ymi = ebw->eb_ht - 1;
	xma = x0;	if (x1 > xma) xma = x1;	if (x2 > xma) xma = x2;
	if (xma < 1) xma = 1;
	if (xma > ebw->eb_wid - 1) xma = ebw->eb_wid - 1;
	if (x3 > xma) xma = x3;
	yma = yy0;	if (yy1 > yma) yma = yy1;	if (y2 > yma) yma = y2;
	if (y3 > yma) yma = y3;
	if (yma < 1) yma = 1;	if (yma > ebw->eb_ht - 1) yma = ebw->eb_ht - 1;
	/* change the state of points that lie to the left of the
	 bounding rectangle */
	for (x = 0; x < xmi; x++) for (y = 0; y < ebw->eb_ht; y++)
		chg_pixel(x, y, changecode, ebw);
	/* change the state of points that lie to the right of the
	 bounding rectangle */
	for (x = xma + 1; x < ebw->eb_wid; x++)
	   for (y = 0; y < ebw->eb_ht; y++) chg_pixel(x, y, changecode, ebw);
	/* change the state of points that lie above and below the
	 bounding rectangle */
	for (x = xmi; x <= xma; x++)
	   {
		for (y = yma + 1; y < ebw->eb_ht; y++)
			chg_pixel(x, y, changecode, ebw);
		for (y = 0; y < ymi; y++) chg_pixel(x, y, changecode, ebw);
	   }
	/* go through the points that lie within the bounding
	 rectangle. Those that are inside the quadrilateral have their
	 state left along; those that lie outside have their state changed */
	for (x = xmi; x <= xma; x++) for (y = ymi; y <= yma; y++)
		if (!isinquad(x, y, x0, yy0, x1, yy1, x2, y2, x3, y3, far))
			chg_pixel(x, y, changecode, ebw);
	return 0;
}

static void chg_rows(int ycent, int yheight, char changecode, EBWORK *ebw)
{ /*	This changes the activation status of `yheight' rows centered
	on row `ycent'. */
	register int	x, y;
	int		ymn, ymx;

	ymn = ycent - yheight / 2;	ymx = ycent + yheight / 2;
	if (ymn < 1) ymn = 1;
	if (ymx > ebw->eb_ht - 2) ymx = ebw->eb_ht - 2;
	for (y = ymn; y <= ymx; y++)
	   for (x = 1; x < ebw->eb_wid- 1; x++)
		chg_pixel(x, y, changecode, ebw);
}

static void chg_columns(int xcent, int xwidth, char changecode, EBWORK *ebw)
{ /*	This changes the activation status of `xwidth' columns centered
	on column `xcent'. */
	register int	x, y;
	int		xmn, xmx;

	xmn = xcent - xwidth / 2;	xmx = xcent + xwidth / 2;
	if (xmn < 1) xmn = 1;
	if (xmx > ebw->eb_wid - 2) xmx = ebw->eb_wid - 2;
	for (x = xmn; x <= xmx; x++)
	   for (y = 1; y < ebw->eb_ht - 1; y++)
		chg_pixel(x, y, changecode, ebw);
}

static void proc_command(char action, int shape, int ia[], EBWORK *ebw)
{ /*	This calls the appropriate routine to change a particular part of
	the active mask.  It then scribbles meaningful stuff to the
	standard error and the logfile. */
	int	diag;

	diag = 0;
	switch (shape)
	      {
		case	BOX: chg_box(ia[0], ia[1], ia[2], ia[3], action, ebw);
			break;
		case	CIRCLE: chg_circle(ia[0], ia[1], ia[2], action, ebw);
			break;
		case	COLUMN:
			chg_columns(ia[0], ia[1], action, ebw);		break;
		case	OBOX:
			chg_obox(ia[0], ia[1], ia[2], ia[3], action, ebw);
			break;
		case	OCIR:
			chg_ocircle(ia[0], ia[1], ia[2], action, ebw);	break;
		case	ROW:
			chg_rows(ia[0], ia[1], action, ebw);	break;
		case	TRI:
			chg_tri(ia[0], ia[1], ia[2], ia[3], ia[4], ia[5],
			 action, ebw);	break;
		case	OTRI:
			chg_otri(ia[0], ia[1], ia[2], ia[3], ia[4], ia[5],
			 action, ebw);	break;
		case	QUAD:
			diag = chg_quad(ia[0], ia[1], ia[2], ia[3],
			 ia[4], ia[5], ia[6], ia[7], action, ebw); break;
		case	OQUAD:
			diag = chg_oquad(ia[0], ia[1], ia[2], ia[3],
			 ia[4], ia[5], ia[6], ia[7], action, ebw);	break;
	     }
	if (diag) return;
	log_change(action, shape, ia, stdout);
	if (ebw->eb_fpl != stdout) log_change(action, shape, ia, ebw->eb_fpl);
}

static void eboperate(char *str, EBWORK *ebw)
{ /*	This operates upon a single `editbord' command.  It checks it for
	validity and passes off the appropriate action to `proc_command' */
	int	argno, iano, shrinkval, iargs[MAX_ARGS - 1];
	char	action;

	if (*str == 'b') action = OFF;
	else if (*str == 'a') action = ON;
	else if (*str == 't') action = TOGGLE;
	else if (*str == 's')
	  {
		action = DOSHRINK;	shrinkval = atoi(*(ebw->eb_cargs));
		shrink(shrinkval, ebw->eb_wid, ebw->eb_ht);
		printf(
		 "All borders moved inward by %d pixels\n", shrinkval);
		return;
	  }
	else {
		fprintf(stderr, "Unrecognized command: %s\n", str);
		return;
	     }
	for (argno = 1, iano = 0; argno < (int)(ebw->eb_ncargs);
	 argno++, iano++)
	   {
		if ((iargs[iano] = atoi(ebw->eb_cargs[argno])) < 1)
		  {
			fprintf(stderr, "Each numeric argument must be > 0\n");
			fprintf(stderr, "Here iargs[%d] = %d from (%s)\n",
				iano, iargs[iano], ebw->eb_cargs[argno]);
			return;
		  }
	  }
	if (**ebw->eb_cargs == 'b')
	  {
		if (ebw->eb_ncargs >= 4) proc_command(action, BOX, iargs, ebw);
		else fprintf(stderr,
		 "Syntax: %s xmin ymin xmax ymax\n", str);
	  }
	else if (!strncmp(*ebw->eb_cargs, "ci", 2))
	  {
		if (ebw->eb_ncargs >= 3)
			proc_command(action, CIRCLE, iargs, ebw);
		else fprintf(stderr, "Syntax: %s xcent ycent radius\n", str);
	  }
	else if (!strncmp(*ebw->eb_cargs, "co", 2))
	  {
		if (ebw->eb_ncargs >= 2)
			proc_command(action, COLUMN, iargs, ebw);
		else fprintf(stderr, "Syntax: %s xcent width\n", str);
	  }
	else if (!strncmp(*ebw->eb_cargs, "tr", 2))
	  {
		if (ebw->eb_ncargs >= 6) proc_command(action, TRI, iargs, ebw);
		else fprintf(stderr, "Syntax: %s %s x0 y0 x1 y1 x2 y2\n",
			str, *ebw->eb_cargs);
	  }
	else if (!strncmp(*ebw->eb_cargs, "qu", 2))
	  {
		if (ebw->eb_ncargs >= 8) proc_command(action, QUAD, iargs, ebw);
		else {
			fprintf(stderr, "ncargs = %d; ", ebw->eb_ncargs);
			fprintf(stderr,
	 "Syntax: %s %s x0 y0 x1 y1 x2 y2 x3 y3\n", str, *ebw->eb_cargs);
		     }
	  }
	else if (!strncmp(*ebw->eb_cargs, "ob", 2))
	  {
		if (ebw->eb_ncargs >= 4) proc_command(action, OBOX, iargs,ebw);
		else fprintf(stderr, "Syntax: %s xmin ymin xmax ymax\n", str);
	  }
	else if (!strncmp(*ebw->eb_cargs, "oc", 2))
	  {
		if (ebw->eb_ncargs >= 3) proc_command(action, OCIR, iargs, ebw);
		else fprintf(stderr, "Syntax: %s xcent ycent radius\n", str);
	  }
	else if (!strncmp(*ebw->eb_cargs, "ot", 2))
	  {
		if (ebw->eb_ncargs >= 6) proc_command(action, OTRI, iargs,ebw);
		else fprintf(stderr, "Syntax: %s x0 y0 x1 y1 x2 y2\n", str);
	  }
	else if (!strncmp(*ebw->eb_cargs, "oq", 2))
	  {
		if (ebw->eb_ncargs >= 8) proc_command(action, OQUAD,iargs,ebw);
		else fprintf(stderr,
		 "Syntax: %s x0 y0 x1 y1 x2 y2 x3 y3\n", str);
	  }
	else if (**ebw->eb_cargs == 'r')
	  { 
		if (ebw->eb_ncargs >= 2) proc_command(action, ROW, iargs, ebw);
		else fprintf(stderr, "Syntax: %s row ycent width\n", str);
	  }
}

/*ARGSUSED*/
int ebordmain(int argc, char *argv[])
{ /*	Mainline for the border-editing functionality */
	register int	i, res;
	char		*chbu, *command, *cp, **dcp;
	EBWORK		ebws;
	static char	*fnames[] = {"amask", "xlog", NULL};
	static char	ftypes[] = {3, 2, 0};

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* open logfile and put a summary of the files there and on stdout */
	if (NULL == (ebws.eb_fpl = propen("xlog"))) ebws.eb_fpl = stdout;
	if (NULL == (chbu = (char *)mallop(CHARS_PER_ARG * MAX_ARGS,
		"chbu in ebordmain"))) return EALLOC;
	if (NULL == (ebws.eb_cargs = (char **)callop(MAX_ARGS, sizeof (char *),
		"ebws.eb_cargs in ebordmain"))) return EALLOC;
	for (dcp = ebws.eb_cargs, cp = chbu, i = 0; i < MAX_ARGS;
	 i++, dcp++, cp += CHARS_PER_ARG) *dcp = cp;
	ftypes[1] = (ebws.eb_fpl == stdout ? 0 : 2);
	header(stdout, PROCESSING, fnames, ftypes);
	if (ebws.eb_fpl != stdout)
		header(ebws.eb_fpl, PROCESSING, fnames, ftypes);
		/* get dimensions of old mask*/
	find_dimension(&(ebws.eb_wid), &(ebws.eb_ht), 0, (char *)NULL);
	if (NULL == (ebws.eb_amask = activmask(ebws.eb_wid, ebws.eb_ht, 5)))
	 return (errno == ENOMEM) ? EALLOC : EREADERR; /* read old mask */
	printf("Now analyzing %d -by- %d pixel active mask\n",
		ebws.eb_wid, ebws.eb_ht);
	printf(
	 " Legal commands are A[ctivate],B[lank],T[oggle],E[xit],S[hrink]\n");
	printf(" Legal subcommands for A, B, or T are ");
	printf(
	 "CI[rcle], R[ow], CO[lumn], B[ox], OC[ircle], OB[ox]\n");
	printf(" CIrcle requires 3 numeric args; Row and COlumn 2\n");
	printf(" Box requires 4 numeric args: xmn ymn xmx ymx\n");
	printf(" Tri requires 6 numeric args: x0,y0,x1,y1,x2,y2\n");
	printf(" Quad requires 8 numeric args: x0,y0,x1,y1,x2,y2,x3,y3\n");
	printf(" OC and OB affect the area outside a circle or box\n");
	printf(" Thus typical commands are:\n");
	printf("\tblank circle 255 260 4\n\tactivate row 214 2\n");
	printf("\t\tCommand `S' takes one argument: the shrinkval\n");
	while (*(command = get_ebcom(&ebws)) != 'e')
	     {
		if (!strncmp(command, "st", 2) || *command == 'q') break;
		eboperate(command, &ebws);
	     }
	if ((NULL != ebws.eb_fpl) && (stdout != ebws.eb_fpl))
		(void)fclose(ebws.eb_fpl);
	if (NULL == activmask(ebws.eb_wid, ebws.eb_ht, 011)) /* output masks */
	  {
		fprintf(stderr, "editbord: Cannot write out active-pixel mask");
		res = EWRITERR;
	  }
	else res = EOK;
	globalkill();	ebws.eb_fpl = NULL;	return res;
}

/* end of edit_bord.c */
