/* progmenu.c -- managing program menus for `setup' and `pmotif'

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Nov 2005: inline argument declarations and strict prototyping
	24 Oct 2005: move static globals inside routine
	25 Oct 2004: "newr", not "newrun", is the name of the app we can call
	21 Oct 2004: all labels shortened to <= 18 characters for simplicity
	20 Oct 2004: -s added to refamenu; utosmenu and stratmenu added
	15 Oct 2004: add 5 more programs; dial-down-the-center implemented;
	 XSCA, XASCA, etc. added to makemu menu; `cenmer' spelling fixed
	28 Aug 2004: cosmetic changes
	28 Feb 2004: bring border and integrate menus up to date (more options)
	31 Jan 2004: spaceg becomes another program we can run
	21 Jan 2004: license message inserted
	13 Dec 2003: manage the #-of-programs argument properly in pmall()
	25 Nov 2003: rename file to progmenu.c; replace globals with static
	 globals and set up callable interface to fish those out
	18 Oct 2003: add 3 more Booleans to integrate menu, correcting a bug
	19 Sep 2003: initialize arrays of structs according to ansi C rules
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 2 Mar 1995: add "purpose" comment.
	21 Apr 1994: renaming polydet.h->xgen.h, adding vsort.h
	16 Mar 1994: initial X-GEN version

   Purpose: provides displayable information for both the
	"curses" and non-MSI motif interfaces to the code
   Cross References: setupmain.c, ../motif/pmotif.c

********************************************************************/
#include	"xgen.h"
#define		NMENUS	33	/* number of currently defined menus */
 
PMENUSTUFF **pmall(int *nprogp)
{ /*	This returns a pointer to the entire list of program menus */
	static int		lnp;
	static PMENUSTUFF	**locpm;
	PMENUSTUFF		**lpdp;
	static PMENUSTUFF pmmenu = { 0, 0, 0, 1, 010, "pmotif",
 "X/Motif-based operating system for XGEN", "--X/Motif bookkeeping",
{{ "Run ID/NamePrefix", "newxtal" }} } ;

	static PMENUSTUFF pdispmenu = { 9, 0, 1, 2, 051, "pdisplay",
 "X/Motif detector image & stats displays", "--X/Motif displays",
{{" Input Frame Index",  "FRAMES"   }, {" Input Active Pix",  "AMASK"   },
 {" Input PixCm Map",    "USPLINE"  }, {" Input Xtal Params", "UPARAMS" },
 {" Input Integ.Refls",  "UREFLS"   }, {" Input Merged Refs", "MULTIREF"},
 {" Input Brass Image",  "CALDATA"  }, {" Input Bgnd Image",  "BGIMAGE" },
 {" Input DarkCurrent",  "DARKIMAGE"},
 {"Dark-current subtr", "s"        },
 {"Frame # displayed",   "1"        }, { "Frame Name displ.", "" }} } ;

	static PMENUSTUFF calibmenu = { 4, 8, 7, 2, 0, "calibrate",
 "Establish pixel-to-cm and cm-to-pixel conversions", " detector geometry",
{{ "Input Brass Image",  "CALDATA"  }, { "Output PixelCm Map", "USPLINE"   },
 { "Output Logfile",     "XLOG"     }, { " Input DarkCurrent", "DARKIMAGE" },
 { "ignore exclud.area", "bn"       }, { "hexagonal grid",     "hn"      },
 { "read calib ctroids", "cn"       }, { "write calib troids", "on"      },
 { "verbose output",     "vn"       }, { "skip interpolation", "nn"      },
 { "preindexed linear",  "in"       }, { "use header dimens.", "jn"      },
 { "Xspacing bw. holes", "x0.508"   }, { "Yspacing bw. holes", "y0.508"  },
 { "min #pixels/spot",   "m3"       }, { "overall cutoff I",   "f0"      },
 { "linear map: #cols",  "l"        }, { "Dark-current subtr", "s"      },
 { "I/SigMin of holes",  "a"        },
 { "Beam CenterX@2th=0", "255.00"   }, { "Beam CenterY@2th=0",  "255.00" }} } ;

	static PMENUSTUFF bordmenu = { 4, 1, 4, 3, 043, "border",
 "Determine active region of the detector", " of detector active area", 
{{ " Input Frame Index",  "FRAMES"   }, { "Output Pixel List", "AMASK"     },
 { "Output Logfile",      "XLOG"     }, { " Input DarkCurrent", "DARKIMAGE"},
 { "verbose output",      "vn"       },
 { "Dark current subtr", "d"        }, { "#Pix.shrunk inward", "s5.00"   },
 { "Min Ci/<Ci> value ",  "q0.42"   },	{ "Framerange:init->", "a6"    },
 { "First frame used",    "1"        }, { "Last frame used",     "999"     },
 { "Min Ci/<Ci> value ",  "0.420"    }} };

	static PMENUSTUFF spotsmenu = { 5, 3,12, 3, 023, "spots",
 "obtain model profiles and centroids for refinement",
 "--get centroids & profiles",
{{ " Input Frame Index", "FRAMES"    }, { "Output Centroids", "CENTROIDS" },
 { "Output 3D Profiles", "WMASK"     }, { "Output Logfile",   "XLOG"       },
 { " Input DarkCurrent", "DARKIMAGE" },
 { "ignore excl.region", "bn"        }, { "verbose logfile",    "vn"   },
 { "semi-verbose log",   "qn"        }, { "min bright pixels",  "h5"   },
 { "min inten.allowed",  "i50"       }, { "#pix drift bw pass", "m2"   },
 { "2nd-brightest frac", "n"         }, { "IntenFrac.ctr 3 fm", "p0.7" },
 { "Dark-current subtr", "s"         }, { "nominal refl. diam", "w5.5" },
 { "Beam Center Xpix",   "x255"      }, { "rocking halfwidth",  "f3" },
 { "phi-to-X ratio",     "k"         }, { "max frac pix < bg",  "l0.4" },
 { "adu per photon",     "e1"        },
 { "First frame used",   "1"         }, { "Last frame used",    "999"  },
 { "Nsigma cutoff",      "16.00"     }} } ;

	static PMENUSTUFF cmermenu = { 2, 0, 0, 5, 0, "cenmer",
 "Merge 2 or more (X,Y,om) centroids files", "--merge CENTROIDS files",
{{ "Output Centroids",     "CENOUT"  }, { "Output Logfile",       "XLOG"  },
 { "Input Centroids #1", ""	},	{ "InputCentroids #2",	"" },
 { "Input Centroids #3", ""	},	{ "InputCentroids #4",	"" },
 { "Input Centroids #5", ""	}} } ;

	static PMENUSTUFF ebordmenu = { 2, 0, 0, 0, 050, "editbord",
 "command-line editor for the active region", "--edit active-pixel list",
{{ "In/Out Active Pix",  "AMASK"     }, { "Output Logfile", "XLOG" }} } ;

	static PMENUSTUFF ecenmenu = { 3, 3, 0, 0, 010, "editcen",
 "Curses-based editor for a CENTROIDS list", "--edit CENTROIDS list",
{{ " Input Centroids",     "CENTROIDS"}, { "Output Centroids", "CENOUT"  },
 { "Output Logfile",       "XLOG"     },
 { "automatic deletes",     "an"      }, { "complement kept",  "cn"  },
 { "check each centr.", "en"      }} };

	static PMENUSTUFF filcmenu = { 3, 2, 6, 0, 050, "filterc",
 "One-pass editor for a CENTROIDS list", "--edit CENTROIDS list",
{{ " Input Centroids",     "CENTROIDS"}, { "Output Centroids", "CENOUT"  },
 { "Output Logfile",       "XLOG"     },
 { "complement kept",  "cn"  }, { "skip (hXYom) check", "sn" },
 { "max deltaD kept", "d10000." }, { "min Intensity", "i0" },
 { "min resol kept", "l10000."} , { "max nonmonoton", "1" },
 { "min sec.peak", "n-1000." }, { "max resol kept", "u0" }
 } };

	static PMENUSTUFF ewmenu = { 2, 0, 0, 0, 010, "ewmask",
 "CURSES editor for 3-D model profiles", "--edit 3-D model profiles",
{{ "In/Out 3D profiles", "WMASK"     }, { "Output Logfile", "XLOG" }} };

	static PMENUSTUFF mermmenu = { 2, 0, 0, 5, 040, "mermask",
 "Merge 2 or more WMASK files", " merge 2 or more WMASKs",
{{ "Output WMASK file",    "WMASK"    }, { "Output Logfile",       "XLOG"  },
 { "Input 3DProfile #1", ""         }, { "Input 3DProfile #2", ""      },
 { "Input 3DProfile #3", ""         }, { "Input 3DProfile #4", ""      },
 { "Input 3DProfile #5", ""         }} };

	static PMENUSTUFF refamenu = { 5, 4, 2, 0, 030, "refall",
 "Hands-off parameter refinement", " crystal & detector params",
{{ " Input PixelCm Map", "USPLINE" }, { " Input Centroids",   "CENTROIDS" },
 { " Input Active-Pix",  "AMASK"   }, { "Output Params",      "UPARAMS"   },
 { "Output Logfile",     "XLOG"    },
 { "Confirm BravChoice", "an"	   }, { "Flip sign(stepsiz)",  "fn" },
 { "Skip (XY) recenter", "nn"	   }, { "Ask for spacegroup",  "sn" },
 { "Specify spacegroup", "i0"	   }, { "Min RMSE(autoind)",	 "m0.1"   }} };

	static PMENUSTUFF refmenu = { 5, 0, 0, 0, 010, "refine",
 "Crystal-and-detector parameter refinement", " crystal & detector params",
{{ " Input PixelCm Map", "USPLINE"    }, { " Input Centroids",   "CENTROIDS" },
 { " Input Active-Pix",  "AMASK"      }, { "Output Params",      "UPARAMS"   },
 { "Output Logfile",     "XLOG"       }} };

	static PMENUSTUFF ref1menu = { 5, 0, 0, 0, 050, "refine1",
 "Line-oriented parameter refinement", " Line-based refinement",
{{ " Input PixelCm Map", "USPLINE"    }, { " Input Centroids",   "CENTROIDS" },
 { " Input Active-Pix",  "AMASK"      }, { "Output Params",      "UPARAMS"   },
 { "Output Logfile",     "XLOG"       }} };

	static PMENUSTUFF intmenu = { 9,13,12, 4, 023, "integrate",
 "measure intensities & sigmas of Bragg refls", " the Bragg reflections",
{{ " Input Frame Index", "FRAMES"  },	{ " Input Parameters",  "UPARAMS"   },
 { " Input PixelCm map", "USPLINE" },	{ "In/Out Bgnd Image",  "BGIMAGE"   },
 { "In/Out 3D Profiles", "WMASK"   },	{ "Output Integr.Refl", "UREFLS"    },
 { "Output Logfile",     "XLOG"    },	{ "Input DarkCurrent",  "DARKIMAGE" },
 { "Ice-ring List",      "ICERING" },
 { "compress images",		"dn" }, { "jagged backgrounds",  "jn"	},
 { "freeze 3D profiles",	"kn" },	{ "skip deadtime corr",  "ln"	},
 { "synchr:horiz omega",	"qn" },	{ "synchr:vert.omega",   "xn"	},
 { "ignore old backgnd",	"nn" },	{ "ignore expos. corr",  "gn"	},
 { "ultraVerbose outpt",	"vn" },	{ "semiverbose output",	"sn"	},
 { "iceRing correction",	"hn" },	{ "plane backgrounds",	"pn"	},
 { "unpolar. radiation",	"un" },
 { "#fms initial bgnd", "b16"	},	{ "refinement IoSMin",	"c6.0"	},
 { "bgnd frac.repl.", "i0.08"	},	{ "Saturation count",	"m62000"},
 { "refinement freq.", "r50"	},	{ "DarkCurrent subtr.",	"t"	},
 { "#adu per photon",  "e1"	},	{ "min|sy| Lorentz",	"y0.025"},
 { "polarization angle", "a90"	},	{ "FracOutside CenFms",	"f0.05"	},
 { "polarization frac",	 "o"	},	{ "bg output frequenc",	"w"	},
 { "First frame used",   "1"	},	{ "Last frame used",	"999" },
 { "Resol.min",		""	},	{ "Resol.max",	"" }} };

	static PMENUSTUFF stratmenu = { 4, 1, 1, 2, 0, "strategy",
 "Compute and plot completeness strategy", "--for acquisition",
{{ " Input PixelCm Map", "USPLINE"   },	{ " Input Active-Pix",  "AMASK"	},
 { "Output Params",      "UPARAMS"   },	{ "Output Logfile",     "XLOG"	},
 { "Min Lorentz==FUZZ",  "an"	},	{ "Min Lorentz value", "y0.03"	},
 { "Resol. minimum",	 "999." },	{ "Resol. maximum",	"1.5"	} }} ;

	static PMENUSTUFF utosmenu = { 2, 2, 0, 0, 040, "utos",
 "Convert UREFLS to SHELX HKL", "--urefls to SHELX hkl",
 {{ " Input Integ.Refls", "UREFLS"  }, { "Output SHELX HKL", "XHKL" },
  { "Batchno==1 for all", "an" }, { "Summed,not fit I's", "sn" } }} ;

	static PMENUSTUFF pickmenu = { 7, 2, 1, 0, 0, "pickuref",
 "Selective dump of integrated reflections", "--dump selected refls",
{{ " Input Frame Index",  "FRAMES"   }, { " Input Integ.Refls", "UREFLS"    },
 { " Input Refl-choice", "REFLIST"   }, { " Input Bgnd Image",  "BGIMAGE"   },
 { " Input Parameters",  "UPARAMS"   }, { " Input DarkCurrent", "DARKIMAGE" },
 { "Output Logfile",     "XLOG"      },
 { "subtract bgnd",	 "bn"     }, { "15x15x7 ImageDumps",   "in"  },
 { "subtr. DarkCurrent", "s"      }} } ;
 
	static PMENUSTUFF analmenu = { 2, 2, 0, 1, 040, "anal_u",
 "Summarize integrated-refls results", "--stats on integration",
{{ " Input Integ.Refls", "UREFLS"    }, { "Output Logfile",     "XLOG"  },
 { " Summed,unfit Int.", "sn"        }, { "Kabsch Inten",       "kn"    },
 { "#frames/refl group", "   20"     }} };

	static PMENUSTUFF filumenu = { 3, 0, 0, 0, 03, "filteru",
 "Filter nonsense from a UREFLS file", "--clean up UREFLS file",
{{ " Input IntegrRefls",   "UREFLS" }, { "Output IntegrRefls",   "UREFOUT"  },
 { "Output Logfile",       "XLOG"   } } };

	static PMENUSTUFF cutumenu = { 3, 0, 0, 2, 043, "cuturef",
 "Eliminate some frames from a UREFLS file", "-limit UREFLS frm range",
{{ " Input IntegrRefls",   "UREFLS" }, { "Output IntegrRefls",   "UREFOUT"  },
 { "Output Logfile",       "XLOG"   },
 { "First Frame# Out", "1"      }, { " Last Frame# Out", "999" }} };

	static PMENUSTUFF redmenu = { 8, 10, 7, 4, 024, "reduce",
 "Reformat and filter integrated intensity list", " the integrated data",
{{ " Input Active-Pix",  "AMASK"     }, { " Input Parameters",  "UPARAMS"    },
 { " Input PixelCm map", "USPLINE"   }, { " Input IntegrRefls", "UREFLS"     },
 { " Input Reindex mtx", "REINDEX"   }, { "Output MergedRefls", "MULTIREF"   },
 { "Output Centroids",   "CENTROIDS" }, { "Output Logfile",     "XLOG"       },
 { "XformedCoord.I&sig", "ey"        }, { "Reindex:3x4 matrix", "in"     },
 { "Reindex by UPARAMS", "kn"        }, { "Don't sort data",    "nn"     },
 { "Print allDeletions", "qn"        }, { "Summed,not fit,I's", "sn"     },
 { "Undo Monochr.corr",  "un"        }, { "elim.big del(XYOm)", "xy"     },
 { "reindex&elim.d(h)",  "zn"        }, { "Use bigger I/sig",   "bn"     },
 { "output frame range", "a1,999"    }, { "CentroidsOut /reg",  "c"      },
 { "Max |If-Is|/sigma",  "d3"        }, { "Max |If-k*Is|/sig",  "g"      },
 { "Min I/sig output",   "f-3"       }, { "ScalingGroup width", "w5"     },
 { "Print N del/type",   "p10"       },
 { "Run Number",         "0" ,       }, { "min sigma/I",        "0.000"  },
 { "Resol. minimum",     "999."      }, { "Resol. maximum",     "0.000"  }} };

	static PMENUSTUFF mrmenu = { 2, 1, 0, 5, 0, "mrmerge",
 "Merge 2 or more MULTIREF files", "--merge MULTIREFs together",
{{ "Output MergedRefls",  "MREFOUT"  }, { "Output Logfile",      "XLOG"      },
 { "keep duplic. obs.", "kn"       },
 { "1st Input MULTIREF",  ""         }, { "2nd Input MULTIREF",  ""      },  
 { "3rd Input MULTIREF",  ""         }, { "4th Input MULTIREF",  ""      },  
 { "5th Input MULTIREF",  "" }} };

	static PMENUSTUFF cutomenu = { 3, 1, 0, 2, 040, "cutoff",
 "Restrict MULTIREF to a sin(theta) range", " multiref on sin(theta)",
{{ " Input MergedRefls", "MULTIREF"  }, { "Output MergedRefls",  "MREFOUT" },
 { "Output Logfile",     "XLOG"      },
 { "Flag,don't delete",  "fn"        },
 { "Dmax allowed",       "999.00"    }, { "Dmin allowed",        "3.00" }} }; 

	static PMENUSTUFF scmmenu = { 12,14, 6, 0, 020, "scalem",
 "Hands-off postprocessing of MULTIREF data", "--all postprocessing",
{{" Input MergedRefls",	"MULTIREF" },	{ " Input Scale File",	"SCALEIN" },
 {"Output Logfile",	"XLOG"	},	{ "Output MissingList",	"MISSING" },
 {"Output MergedRefls",	"MREFOUT" },	{ " Input ReindexFile",	"REINDEX" },
 {"Output ascii hkls",	"MULIST" },	{ "Output SHELX hkls",	"XSHEL"   },
 {"Output CNS   hkls",	"XCNS"	},	{ "Output XPLOR hkls",	"XXPLOR"  },
 {"Output Merged .sca",	"XSCA"	},	{ "Output Unmerg.sca",	"XASCA" },
 {"CHECKREFL from rej",	"cn"	},	{ "Write F's, not I's",	"fn" },
 { "Reindex reflectns" , "in"	},	{ "2ob refs:kill both",	"kn" },
 { "Output MultOp refs", "mn"	},	{ "Set I< 0 Refs to 0",	"nn" },
 { "Omit MISSING outpt", "on"	},	{ "Keep scalfac(sinq)",	"pn" },
 { "Scale top&bot toge", "rn"	},	{ "Output resol order",	"tn" },
 { "Alt.variances vals", "un"	},	{ "Verbose log output",	"vn" },
 { "Wilson initScaling", "wn"	},	{ "MachineErr sigmas" ,	"zn" },
 { "MinRedun(Bijvoets)", "a4"	},	{ "#Stats resol bins" ,	"b"  },
 { "FracDeleted in rej", "g0.015"},	{ "min <I> in scaling",	"j0" },
 { "|I-<I>|/s RjCutoff", "q"	},	{ "Output OverallScal",	"s1" } } };
 
	static PMENUSTUFF scmenu = { 4,10, 6, 0, 0, "scalek",
 "Multiparameter scaling of MULTIREF data", "--scale intensities",
{{ " Input MergedRefls", "MULTIREF"  }, { " Input Scale File",   "SCALEIN"  },
 { "Output Logfile",     "XLOG"      }, { "Output Scale File",  "SCALEOUT"  },
 { "Use ALL observat.",  "an",       }, { " 9 DiffVector Pars", "gn" },
 { "10 DiffVector Pars", "dn",       }, { "10 ExitVector pars", "en" },
 { "keep cycling",       "kn",       }, { "Separate Bijvoets",  "nn" },
 { "Merge top&bottom",   "tn",       }, { "Verbose output",     "vn" },
 { "nonlin.scaling",     "bn",       }, { "Nonlin.ifR incr",    "hn" },
 { "# cycles",           "c10",      }, { "Min <I> used",       "i0" },
 { "#params/scalegrp",   "p1",       }, { "Frac.shift appl",    "s1.0" },
 { "Convergence*100000", "q100"      }, { "Min # obs/refl",     "o2" }} };

	static PMENUSTUFF rejmenu = { 4,10, 5, 0, 040, "reject",
 "Outlier Deletion from a Multiref", " outliers from data      ",
{{ " Input MergedRefls",  "MULTIREF" }, { " Input Scale File",   "SCALEIN"  },
 { "Output Logfile",     "XLOG"      }, { "Output MergedRefls",  "MREFOUT"  },
 { "Auto (batch) mode",  "bn"        }, { "Check-refl list",     "cn"  },
 { "Separate Bijvoets",  "nn"        }, { "Verbose output",      "vn"  },
 { "2ObRefs:skip Iosma", "xn"        }, { "2ObRefs:keep Imax",   "ln"  },
 { "2ObRefs:keep miSig", "gn"        }, { "2ObRefs:keep both",   "in"  },
 { "2ObRefs:kill both",  "kn"        }, { "2ObRefs:displ both",  "sn"  },
 { "Machine-error val",  "m"         }, { "(delI/sig) cutoff",   "q"   },
 { "Fraction to delete", "f0.03"     }, { "R-value to achieve",  "t"   },
 { "Start @ (hkl) = ",   "r"         }} } ;

	static PMENUSTUFF statmenu = { 4, 2, 0, 1, 0, "stats",
 "Determine statistics for merged-refl list", " obtained from MULTIREF",
{{ " Input MergedRefls", "MULTIREF"  }, { " Input Scale File",  "SCALEIN" },
 { "Output Logfile",     "XLOG"      }, { "Output MissingList", "MISSING" },
 { "Separate Bijvoets",  "nn",       }, { "Write MissingList",  "oN"      },
 { "Resolution Limit",   ""          }} } ;

	static PMENUSTUFF makemenu = {12, 8, 1, 1, 040, "makemu",
 "Produce ascii hkl list from MULTIREF", "--make ascii merged file ",
{{" Input MergedRefls",	"MULTIREF" },	{ " Input Scale File",	"SCALEIN" },
 {"Output Logfile",	"XLOG"	},	{ "Output MissingList",	"MISSING" },
 {"Output MergedRefls",	"MREFOUT" },	{ " Input ReindexFile",	"REINDEX" },
 {"Output ascii hkls",	"MULIST" },	{ "Output SHELX hkls",	"XSHEL"   },
 {"Output CNS   hkls",	"XCNS"	},	{ "Output XPLOR hkls",	"XXPLOR"  },
 {"Output Merged .sca",	"XSCA"	},	{ "Output Unmerg.sca",	"XASCA" },
 { "Absol (hkl) order",  "an",       }, { "Output Fs, not Is",  "fy"  },
 { "Reindex:3x4 matrix", "in",       }, { "Set I < 0 to 0",     "mn"  },
 { "Only MultiOb Refls", "nn",       }, { "Keep 2th. depend.",  "pn"  },
 { "2-Theta order",      "tn" ,      }, { "Alternate sigmas",   "vn"  },
 { "OverallScaleFactor", "s1.0"      }, { "Resolution limit",   ""    }} } ;

	static PMENUSTUFF spacegmenu = { 4, 3, 0, 1, 0, "spaceg",
 "Verify spacegroup from scaled data", " --verify spacegroup",
{{ " Input MergedRefls", "MULTIREF",  }, { " Input Scale File",  "SCALEIN" },
 { "Output Logfile",     "XLOG",      }, { "Output MergedRefls", "MREFOUT" },
 { "Output definitely",  "on"         }, { "Output conditional", "cn"    },
 { "Verbose logfile",    "vn"         }, { "Resolution limit",     ""   }} };

	static PMENUSTUFF mukbmenu = { 3, 4, 1, 8, 040, "mukb",
 "Get k,B, R-factor between native & derivs",  "--scale&merge MULISTs      ",
{{ "Output DORGBN file", ""         }, { "Output Rel. Mulist",  ""      },
 { "Output Patterson",    ""         },
 { "DORGBIN output",     "dn"        }, { "Force k&B output",   "in"     },
 { "Relative Mulist",    "on"        }, { "Patterson Output",   "pn"     },
 { "F/sig cutoff,R2&R4", "s2.0"      }, { "Resolution min",     "999.0"  },
 { "Resolution max",     "0.00"      },
 { "Parent MULIST",      ""          }, { "Deriv #1",           ""       },
 { "Deriv #2",           ""          }, { "Deriv #3",           ""       },
 { "Deriv #4",           ""          }, { "Deriv #5",           ""       }} } ;

	static PMENUSTUFF newrmenu = { 1, 0, 0, 6, 0, "newr",
 "Adjust parameters for subsequent dataruns", "--pseudogoniostat angles ",
{{ "In/Out Parameters",  "UPARAMS"  },
 { "Old Run:OmegaStart", " 0.00"    }, { "Old Run:  ChiStart", "45.00"   },
 { "Old Run:  ChiStart", "45.00"    }, { "New Run:OmegaStart", " 0.00"   },
 { "New Run:  ChiStart", "45.00"    }, { "New Run:  PhiStart", "90.00"   }} } ;

	static PMENUSTUFF xdumenu = { 0, 0, 0, 2, 040, "xdump",
 "Print ascii dump of XGEN binary file(s)", "--dump of XGEN file",
{{ "XGEN file type",      ""        }, { "Name of file",       ""        }} } ;

	if (NULL != locpm) /* see if we've already done */
	  { /* if the old number of menus is reasonable, return quickly */
		if ((0 < lnp) && (lnp <= NMENUS))
		  {
			*nprogp = lnp;	return locpm;
		  }
		/* otherwise, we don't trust this stuff.
		 Throw away locpm and start over */
		else memzap((AVOID **)&locpm, "locpm in pmall");
	  }
	locpm = (PMENUSTUFF **)callop(NMENUS, sizeof (PMENUSTUFF *),
	 "locpm in pmall");
	lpdp = locpm;
	*lpdp++ = &pmmenu;				*lpdp++ = &pdispmenu;
	*lpdp++ = &calibmenu;				*lpdp++ = &bordmenu;
				*lpdp++ = &spotsmenu;
	*lpdp++ = &cmermenu;				*lpdp++ = &ebordmenu;
	*lpdp++ = &ecenmenu;				*lpdp++ = &filcmenu;
	*lpdp++ = &ewmenu;				*lpdp++ = &mermmenu;
				*lpdp++ = &refamenu;
	*lpdp++ = &refmenu;				*lpdp++ = &ref1menu;
				*lpdp++ = &intmenu;
	*lpdp++ = &stratmenu;				*lpdp++ = &utosmenu;
	*lpdp++ = &pickmenu;				*lpdp++ = &analmenu;
	*lpdp++ = &filumenu;				*lpdp++ = &cutumenu;
				*lpdp++ = &redmenu;
	*lpdp++ = &mrmenu;				*lpdp++ = &cutomenu;
				*lpdp++ = &scmmenu;
	*lpdp++ = &scmenu;				*lpdp++ = &rejmenu;
	*lpdp++ = &statmenu;				*lpdp++ = &makemenu;
	*lpdp++ = &spacegmenu;				*lpdp++ = &mukbmenu;
	*lpdp++ = &newrmenu;				*lpdp++ = &xdumenu;
	*nprogp = lnp = lpdp - locpm;	return locpm;
}

/* end of progmenu.c */
