/* setupfork.c -- routines for calling X-GEN functions from outside

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	 3 Apr 2006: realxgcomd and isxgcommand are now static
	20 Nov 2005: inline argument declarations and strict prototyping
	28 Jun 2005: locsysn(): compare comdv to NULL, not 0.
	11 May 2005: move xgfork() from here to xgenstr.c
	25 Oct 2004: "newr", not "newrun", is the name of the app we can call
	20 Oct 2004: 34 programs called from here, consistent with progmenu.c
	15 Oct 2004: 32 programs called from here, consistent with progmenu.c
	28 Aug 2004: routines used only here are declared as static; cosmetics
	28 Jun 2004: plotting capability facilitated with plotnow()
	 5 Mar 2004: print diagnostic when fork() fails
	29 Feb 2004: cosmetic changes, including a \r\n in locsysn
	31 Jan 2004: spaceg becomes another program we can run
	21 Jan 2004: license message inserted
	 6 Jan 2004: globals progname, keepglobals replaced by calls to
		currpname; setkillglobs() called appropriately
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 6 Aug 2003: move this file from curses to xglib
	 5 Jun 2003: setkillglobs(1) just before forking
		so the program doesn't free the wrong batches of memory
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: arranged as part of the core code

   Purpose: enables calling X-GEN executables from `setup', `pmotif',
	or future user interfaces, by forking and direct function calls.
   Cross references: setupprog.c, ../motif/pmnames.c

*************************************************************************/
#define		NEED_ERRNO	1
#define		NEED_UNISTD	1
#include	"xgen.h"
#include	<sys/wait.h>
#define		ENOTXGEN	(-27354)

/*	A set of routines associated with calling X-GEN executables
	from the outside world.
	This doesn't do any CURSES or Motif stuff, so it's flexible */
extern int		analumain(int ac, char *av[]);
extern int		bordmain(int ac, char *av[]);
extern int		calibmain(int ac, char *av[]);
extern int		cenmmain(int ac, char *av[]);
extern int		cutoffmain(int ac, char *av[]);
extern int		cutumain(int ac, char *av[]);
extern int		ebordmain(int ac, char *av[]);
extern int		edcenmain(int ac, char *av[]);
extern int		ewmmain(int ac, char *av[]);
extern int		filtercmain(int ac, char *av[]);
extern int		filterurfmain(int ac, char *av[]);
extern int		integmain(int ac, char *av[]);
extern int		makemumain(int ac, char *av[]);
extern int		mergemmain(int ac, char *av[]);
extern int		mrmmain(int ac, char *av[]);
extern int		mukbmain(int ac, char *av[]);
extern int		newrunmain(int ac, char *av[]);
extern int		pickumain(int ac, char *av[]);
extern int		reducemain(int ac, char *av[]);
extern int		ref1main(int ac, char *av[]);
extern int		refallmain(int ac, char *av[]);
extern int		refmain(int ac, char *av[]);
extern int		rejmain(int ac, char *av[]);
extern int		scalemmain(int ac, char *av[]);
extern int		scmain(int ac, char *av[]);
extern int		setupmain(int ac, char *av[]);
extern int		spacegmain(int ac, char *av[]);
extern int		spotmain(int ac, char *av[]);
extern int		statmain(int ac, char *av[]);
extern int		stratmain(int ac, char *av[]);
extern int		utoshelmain(int ac, char *av[]);
extern int		xdumpmain(int ac, char *av[]);

static int isxgcommand(char *verb)
{ /*	This returns an integer associated with a particular X-GEN
	command. It returns -1 if the verb isn't an X-GEN command */
	int	pno;
	pno = 1;	if (!strcmp(verb, "setup")) return pno;
	pno++;	if (!strcmp(verb, "pmotif")) return pno;
	pno++;	if (!strcmp(verb, "pdisplay")) return pno;
	pno++;	if (!strcmp(verb, "calibrate")) return pno;
	pno++;	if (!strcmp(verb, "border")) return pno;
	pno++;	if (!strcmp(verb, "spots")) return pno;
	pno++;	if (!strcmp(verb, "cenmer")) return pno;
	pno++;	if (!strcmp(verb, "editbord")) return pno;
	pno++;	if (!strcmp(verb, "editcen")) return pno;
	pno++;	if (!strcmp(verb, "filterc")) return pno;
	pno++;	if (!strcmp(verb, "ewmask")) return pno;
	pno++;	if (!strcmp(verb, "mermask")) return pno;
	pno++;	if (!strcmp(verb, "refall")) return pno;
	pno++;	if (!strcmp(verb, "refine")) return pno;
	pno++;	if (!strcmp(verb, "refine1")) return pno;
	pno++;	if (!strcmp(verb, "integrate")) return pno;
	pno++;	if (!strcmp(verb, "strategy")) return pno;
	pno++;	if (!strcmp(verb, "utos")) return pno;
	pno++;	if (!strcmp(verb, "pickuref")) return pno;
	pno++;	if (!strcmp(verb, "anal_u")) return pno;
	pno++;	if (!strcmp(verb, "filteru")) return pno;
	pno++;	if (!strcmp(verb, "cuturef")) return pno;
	pno++;	if (!strcmp(verb, "reduce")) return pno;
	pno++;	if (!strcmp(verb, "mrmerge")) return pno;
	pno++;	if (!strcmp(verb, "cutoff")) return pno;
	pno++;	if (!strcmp(verb, "scalem")) return pno;
	pno++;	if (!strcmp(verb, "scalek")) return pno;
	pno++;	if (!strcmp(verb, "reject")) return pno;
	pno++;	if (!strcmp(verb, "stats")) return pno;
	pno++;	if (!strcmp(verb, "makemu")) return pno;
	pno++;	if (!strcmp(verb, "spaceg")) return pno;
	pno++;	if (!strcmp(verb, "mukb")) return pno;
	pno++;	if (!strcmp(verb, "newr")) return pno;
	pno++;	if (!strcmp(verb, "xdump")) return pno;
	return -1;
}

static int realxgcomd(int comdno, int ncom, char *comdv[])
{ /*	This actually calls an X-GEN command.
	If `comdno' is out of range, it returns ENOTXGEN.
	For the moment, it doesn't try to do motif calls.
	May 2003: setkillglobs(1) so that the program doesn't
	die right afterward */
	(void)managekill(1);
	switch (comdno)
	      {
		case	 1: return setupmain(ncom, comdv);
		case	 2: return ENOTXGEN;
		case	 3: return ENOTXGEN;
		case	 4: return calibmain(ncom, comdv);
		case	 5: return bordmain(ncom, comdv);
		case	 6: return spotmain(ncom, comdv);
		case	 7: return cenmmain(ncom, comdv);
		case	 8: return ebordmain(ncom, comdv);
		case	 9: return edcenmain(ncom, comdv);
		case	10: return filtercmain(ncom, comdv);
		case	11: return ewmmain(ncom, comdv);
		case	12: return mergemmain(ncom, comdv);
		case	13: return refallmain(ncom, comdv);
		case	14: return refmain(ncom, comdv);
		case	15: return ref1main(ncom, comdv);
		case	16: return integmain(ncom, comdv);
		case	17: return stratmain(ncom, comdv);
		case	18: return utoshelmain(ncom, comdv);
		case	19: return pickumain(ncom, comdv);
		case	20: return analumain(ncom, comdv);
		case	21: return filterurfmain(ncom, comdv);
		case	22: return cutumain(ncom, comdv);
		case	23: return reducemain(ncom, comdv);
		case	24: return mrmmain(ncom, comdv);
		case	25: return cutoffmain(ncom, comdv);
		case	26: return scalemmain(ncom, comdv);
		case	27: return scmain(ncom, comdv);
		case	28: return rejmain(ncom, comdv);
		case	29: return statmain(ncom, comdv);
		case	30: return makemumain(ncom, comdv);
		case	31: return spacegmain(ncom, comdv);
		case	32: return mukbmain(ncom, comdv);
		case	33: return newrunmain(ncom, comdv);
		case	34: return xdumpmain(ncom, comdv);
		default:	break;
	     }
	return ENOTXGEN;
}

static int locsysn(int ncom, char *comdv[], int shouldwait)
{ /*	A kinder, gentler `system' routine that expects an `argc, argv' type
	of argument system. It is interruptable, unlike the real system(),
	and it can be invoked via setuid.  Borrowed liberally from the
	official system(3) documentation. The third argument is used to specify
	whether the routine should wait for the execv'd routine to return. */
	int		i, pid, status;
	char		*comd, savestr[1024];
	static FILE	*fplo, *fple;

	/* make sure the command is really present */
	if ((NULL == comdv) || ('\0' == **comdv)) return 1;
	/* if this is an X-GEN command, we don't fork at all;
	 instead, we call the appropriate X-GEN command. There's
	 probably an elegant way to do this, but I'll skip that for now.
	 To identify the command, we look for filepart(*comdv, 2); */
	comd = filepart(*comdv, 2);
	/* we can't guarantee that comdv won't get altered by the fork&exec
	 or the xgen command itself. Therefore we save a copy of the
	 command string for diagnostic purposes */
	zeromem((AVOID *)savestr, 1024);
	strcpy(savestr, *comdv);
	for (i = 1; i < ncom; i++)
	   {
		strcat(savestr, " "); strcat(savestr, *(comdv + i));
	   }
	if (0 <= (i = isxgcommand(comd)))
	  {
		if (!shouldwait)
		  {
			if (-1 == (pid = fork()))
			  {
				printf("Attempt to fork new process failed\n");
				return pid;
			  }
			if (0 == pid) /* we're forked: */
			  {
				fplo = freopen("xgen.out", "a", stdout);
				fple = freopen("xgen.out", "a", stderr);
				if (ENOTXGEN == (i = realxgcomd(i,ncom,comdv)))
				  {
					execvp(comd, comdv);
					exit(127); /* only when execvp fails */
				  }
				else {
					fclose(fplo);	fclose(fple);
					return i;
				     }
			  }
			else return 0;
		  }
		else { /* don't fork unless we have to */
			if (ENOTXGEN == (status = realxgcomd(i, ncom, comdv)))
				return xgfork(comdv, 1);
			else {
				fprintf(stderr, "%s returned %d\r\n",
					savestr, status);
				return status;
			     }
		    }
	  }
	/* If we get here, we're running a non-X-GEN command. */
	return xgfork(comdv, shouldwait);
}

int locsysc(char *stri, int shouldwait)
{ /*	This is an analog of the traditional "system" command,
	calling a single string as an argument. */
	char		*cpname;
	int		resu, azc;
	char		**azv;

	if ((NULL == stri) || ('\0' == *stri)) return -1;
	cpname = xgprogname(NULL, 0);	(void)xgprogname(NULL, 2);
	/* now convert the command string to an argc-argv combination. */
	resu = (NULL == (azv = str2acav(stri, &azc))) ? EALLOC : EOK;
	(void)xgprogname(NULL, 3);	if (EOK != resu) return resu;
	return locsysn(azc, azv, shouldwait); /* fork & exec */
}

int plotnow(char *pfn)
{ /*	This forks to run the actual plotting functionality */
	int		resu, narg;
	static char	plcom[164];
	static char	curplot[] = "mxplot.tcl"; /* current plotting program*/
	char		**arv;

	zeromem(plcom, 164);
	sprintf(plcom, "%s < %s", curplot, pfn); /* construct plot command*/
	(void)xgprogname(NULL, 2); /* so str2acav doesn't prepend the name */
	if (NULL != (arv = str2acav(plcom, &narg))) /* create args */
		resu = locsysn(narg, arv, 0);	/* asynchronous call */
	else	resu = EALLOC;
	(void)xgprogname(NULL, 3); /* put program name back where it belongs */
	(void)str2acav(NULL, &narg);	/* kill memory for str2acav */
	return resu;	/* done! */
}

/* end of setupfork.c */
