/* xgenstr.c -- string-handling routines for X-GEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: get_calibparams: support for Aviex detectors
	20 Oct 2006: don't necessarily undefine NEED_SETENV in non-SGI environs
	 7 Sep 2006: add setenv, unsetenv calls for sgi
	 7 Jul 2006: helpapp bugfix: read into dotxt = doname, not doname+256
	24 Jun 2006: include xgimage.h; need some typedefs and #defines there
	11 Jun 2006: move filetonv from here to xgenviron so it can be static;
	 move get_calibparams here--almost static! ; move zeromem,trname up
	31 May 2006: ftkill gone;journaled,smartdefa,findbool,findsw moved here
	 3 Jan 2006: generally, set_logical can return >= 1;
	 globalkill() is busier; use lineend() to format diagnostics
	 5 Dec 2005: handle fork and execvp declarations properly
	20 Nov 2005: inline argument declarations and strict prototyping
	 8 Nov 2005: remove all but one of the static global variables
	22 Jun 2005: unsetenv() in SunOS version now declared static
	27 May 2005: str2acav() returns something reasonable if fed "\n"
	19 May 2005: helpapp() prints diagnostics on failure but doesn't exit
	17 May 2005: flagproc() looks for --help, --version before alloc'ing
	13 May 2005: set_logical, unset_logical rewritten so they
		don't explicitly look at _environ or environ
	11 May 2005: move xgfork() from setupfork.c to here
	 2 Apr 2005: don't actually capitalize args in flagproc();
	 	instead, compare the flag values with a capitalized version
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	24 Nov 2004: handle --help and --version args in flagproc()
	20 Oct 2004: make_upper available under Unix and externally
	28 Aug 2004: routines used only here are declared as static; cosmetics
	14 Aug 2004: verror,qerror,format-conversion routines moved here
	18 Jun 2004: get_calibparams(): add Mar 225 detector
	 8 Jun 2004: move shorten() to xgenfile.c
	13 Apr 2004: in journaled(), get rid of memory for temp string cpz
	29 Mar 2004: fix bug in str2acav() that caused overruns on reads
	28 Mar 2004: in get_calibparams(), set *lincolp = 0 for
		detectors that require brass-plate images
	28 Mar 2004: routine journaled() added and called from smartdefa();
	 fromcur set in smartdefa(); some smartdefa() defaults changed
	21 Jan 2004: license message inserted
	 6 Jan 2004: remaining globals removed, primarily by hiding how we kill
		memory for particular kinds of structures in globalkill();
		kill_mail() written to do just that;
		setkillglobs handles use of globalkill();
		callop(), memzap() calls replace calls to shcalloc(), shfree()
	10 Dec 2003: several extern variables declared at top of file
	 7 Dec 2003: eliminate memzap(iz0): it's now static in image.c
	 8 Nov 2003: memzap on lca in globalkill()
	20 Sep 2003: move zeromem() and cpymem() earlier in the file
	 so they don't need to be declared externally; lpunct is initilized
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	26 Aug 2003: lswawb() moved to this file, where it belongs
	22 Aug 2003: eliminate references to listname; it's handled differently
	18 Aug 2003: created from xgenio.c
 Relevant changes to xgenio.c:
	 6 Jun 2003: better handling of `keepglob' concept:
		globalkill() does all its initial memzap() calls but quits
		before it calls memzap() on the variables that need to
		persist across calls from user interfaces like pmotif
	26 Apr 2003: add `spaceg' to list of programs reached from smartdefa()
	22 Mar 2003: update names and args of some programs in smartdefa()
	 5 Mar 2003: add unset_logical(), xgstrend()
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: manipulate model-profile array sizes as variables,
		not as constants; save memory for profile ops with nabuf
	15 Apr 2002: declare shlp as SHIFT *, not SHIFT **
	25 Mar 2002: removed MSI attachments
	15 Aug 1998: in `filepart', retname is an auto variable;
		destroying its memory should be done externally
	15 Sep 1997: in filepart, always allocate memory for and return a
		fresh copy of the string, even if it's identical to the input
	26 Oct 1995: fix trivial error in call to lswawb
	13 Oct 1995: give lswawb a 2nd arg to handle both directions correctly
	31 Aug 1995: use `ewidth' as a local variable
	25 Aug 1995: `globalkill' is more careful
		about not freeing unallocated memory
	21 Aug 1995: rename temporary header in globalkill
	 7 Jul 1995: use pointer-equivalent integers in printouts
	31 May 1995: "register" -> "register int" throughout
		re-cast several memory-handling routines with size_t's
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		input corrected for reading old-style USPLINEs;
		SDMS headers modified
	 1 Dec 1994: don't do byteswaps & float-conversions in VMS
	16 Nov 1994: correct pluralization in diagnostic from callop
	 4 Nov 1994; Major rewrite of set_logical() to handle memory alloc
	20 Oct 1994: remove declaration of mkdir() from `create_direc()'
	11 Oct 1994: bugfix in `allnumeric' for negative arguments
	30 Sep 1994: `shorten' returns # of chars saved by abbreviation
		`allshorten' routine added to shorten names in systematic way
	23 Sep 1994: clean up `myenviron' stuff in set_logical
	 7 Jul 1994: memzap(&vshp,""), not memzap(vshp,"") in globalkill
	 5 May 1994: initial version of xgenio.c, cobbled from other files
	15 Feb 1994: in dir_access under VMS, ignore request for write access,
		since under VMS that causes more trouble than it solves.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of older files
	15 May 1992: create and use dosetenv() routine
	xx xxx 1987: A single trname() works under Unix and VMS.
	 4 Oct 1986: in trname(), don't bother to look if there's a $ or a /
		in the name.
   Purpose: dealing with strings in X-GEN
   Cross references: setupio.c, xgenfile.c, xgenmisc.c

*************************************************************************/
#define		NEED_SSDEF	1
#define		NEED_DESCRIP	1
#define		NEED_STAT	1
#define		NEED_ERRNO	1
#define		NEED_UNISTD	1
#define		NSHAREDPTRS	30	/* max # pointers shared among apps */
#define		CAPITALIZE(x)	(x &= 0137)
#include	"xgimage.h"
#include	<sys/types.h>
#include	<sys/wait.h>
#include	<stdlib.h>

int xgfork(char **comdve, int shouldw)
{ /* This does the fork-and-exec of the command `comdve'.
	since we're using the execvp version, we can just treat comdve[0]
	as the first arg to the execv command. */
	int		pid, status;
	char		loccom[256], *cpi, *cpo;
	static FILE	*fplo, *fple;

	for (cpo = &(loccom[0]), cpi = *comdve; '\0' != *cpi; ) *cpo++ = *cpi++;
	*cpo = '\0';
	if (-1 == (pid = fork())) return pid;
	if (0 == pid) /* we're forked here */
	  {
		if (!shouldw)
		  {
			fplo = freopen("xgen.out", "a", stdout);
			fple = freopen("xgen.out", "a", stderr);
		  }
		errno = 0;	execvp(loccom, comdve);
		exit(errno ? errno : 2); /* get here only if execvp fails*/
	  }
	else if (shouldw) /* we're here in the parent process when waiting */
	 {
		do {
			if (-1 == waitpid(pid, &status, 0))
			  {
				if (-1 == errno) return -1;
			  }
			else if (WIFEXITED(status)) return WEXITSTATUS(status);
			else return status;
	           } while (1);
	 }
	return 0;	/* parent process when we don't wait */
}

int verror(int value, char *errstr)
{ /*	This examines `value' and, if it's 0, prints an error string and
	returns EBAD.
	Otherwise, it returns silently (like a thief in the night) */
	if (0 == value)
	  {
		fprintf(stderr,
			"%s: Cannot %s\n", xgprogname(NULL, 0), errstr);
		return EBAD;
	  }
	return 0;
}

int qerror(AVOID *pvalue, char *errstr)
{ /*	This examines `pvalue' and, if it's NULL, prints an error string and
	returns EBAD.
	Otherwise, it returns silently (like a thief in the night). */
	if (NULL == pvalue)
	  {
		fprintf(stderr,
			"%s: Cannot %s\n", xgprogname(NULL, 0), errstr);
		return EBAD;
	  }
	return 0;
}

void *callop(size_t nitems, size_t sizitem, char *str)
{ /* a wrapper for `calloc' that under appropriate circumstances
	prints a diagnostic to stdout or stderr while it does its thing. */
	void	*res;
	PAU	pru;

	errno = 0;
	res = calloc(nitems, sizitem);
	if ((NULL != str) && ('\0' != *str))
	  {
		if (NULL == res) fprintf(stderr,
 "Cannot allocate memory for %6d items of %6d bytes each for %s; errno = %d\n",
				(int)nitems, (int)sizitem, str, errno);
		else if (getenv("ALLOC_VERBOSE"))
		  {
			pru.vpt = res;
			if (nitems == 1) printf(
			 " %6d item of %6d bytes allocated @ 0x%x for %s\n",
			 1, (int)sizitem, pru.ui, str);
			else printf(
		 "%6d items of %6d bytes each allocated @ 0x%x for %s\n",
		  (int)nitems, (int)sizitem, pru.ui, str);
		  }
	  }
	return res;
}

void *reallop(AVOID *oldptr, size_t nbytes, char *str)
{ /* a wrapper for `realloc' that under appropriate circumstances
	prints a diagnostic to stdout or stderr while it does its thing. */
	void	*res;
	PAU	pru;

	errno = 0;
	res = realloc(oldptr, nbytes);
	if ((NULL != str) && ('\0' != *str))
	  {
		if (NULL == res) fprintf(stderr,
 "Cannot reallocate %6d bytes for %s; errno = %d\n", (int)nbytes, str, errno);
		else {
			pru.vpt = res;
			if (getenv("ALLOC_VERBOSE")) printf(
		 "%6d bytes reallocated @ 0x%x for %s\n",
			(int)nbytes, pru.ui, str);
		     }
	  }
	return res;
}

void *mallop(size_t nbytes, char *str)
{ /* a wrapper for `malloc' that under appropriate circumstances
	prints a diagnostic to stderr while it does its thing. */
	void	*res;
	PAU	pru;

	errno = 0;
	res = malloc(nbytes);
	if ((NULL != str) && ('\0' != *str))
	  {
		if (NULL == res) fprintf(stderr,
	 "Cannot allocate %6d bytes for %s; errno = %d\n",
			(int)nbytes, str,errno);
		else {
			pru.vpt = res;
			if (getenv("ALLOC_VERBOSE")) printf(
			 "%6d bytes allocated @ 0x%x for %s\n",
			 (int)nbytes, pru.ui, str);
		     }
	  }
	return res;
}

void memzap(AVOID **base, char *str)
{ /* This frees memory for `*base' with an appropriate message,
	 and sets *base to NULL */
	PAU		pru1, pru2;

	if ((NULL == base) || (NULL == *base)) return;
	pru1.vdpt = base;	pru2.vpt = *base;
	if ((NULL != str) && ('\0' != *str) &&
	  (getenv("ALLOC_VERBOSE"))) printf(
		"Memory for %s at 0x%x freed\n", str, pru2.ui);
	free((void *)(*base));	*base = NULL;
}

static int isxgspace(char c)
{ /* returns 1 iff. c is one of the characters on the list
	of space-like characters or is '\0'. */
	register char	*cp;
	static char	spacer[] = " \t\n\r";

	if ('\0' == c) return 1;
	for (cp = &(spacer[0]); '\0' != *cp; cp++) if (*cp == c) return 1;
	return 0;
}

void zeromem(AVOID *base, size_t nbytes)
{ /* This sets `nbytes' of memory to zero starting at `cp' */
	register char	*cpp, *cpend;

	cpend = (char *)base + nbytes;
	for (cpp = (char *)base; cpp < cpend; cpp++) *cpp = 0;
}

char **str2acav(char *st, int *acp)
{ /*	This converts a string `st' into an (argv, argc)-type system in
	which the strings are arranged sequentially, one after another.
	By convention we prepend the program name to the string `st'
	if the program name exists */
	char		*cpname, *cp, *stu, **dvp, *ocp;
	int		i, nw, nch;
	static size_t	newnw, oldnw, newlen, oldlen;
	static char	*out, **odp;
	static char	odpmsg[] = "odp in str2acav";
	static char	outmsg[] = "out in str2acav";

	if (NULL == st) /* signal to eliminate memory */
	  {
		memzap((AVOID **)&out, outmsg);
		memzap((AVOID **)&odp, odpmsg);	return odp;
	  }
	*acp = 0;
	if (('\0' == *st) || (!strcmp("\n", st))) return EOK;
	/* skip over initial spaces, if any */
	for (stu = st; '\0' != *stu; stu++) if (!isxgspace(*stu)) break;
	if ('\0' == *stu) return EOK;	/* just blanks: no strings */
	cpname = xgprogname(NULL, 0);
	/* First count the number of words and chars in `st'.
	 We specifically ignore multiple spaces */
	for (nch = 0, nw = 1, cp = stu; '\0' != *cp; cp++)
	   {
		if (isxgspace(*cp))
		  {
			nw++;
			while (isxgspace(*(cp + 1)))
			     {
				cp++;
				if ((*cp == '\n') || (*cp == '\0')) break;
			     }
		  }
		else nch++;
	   }
	*acp = nw;	nch += nw;
	newnw = nw + 2; newlen = nch + 3;	/* to count spaces */
	if (newnw < 10) newnw = 10;	if (newlen < 256) newlen = 256;
	/* reserve memory for the correct number of words */
	if ((NULL == odp) || (newnw > oldnw))
	  {
		memzap((AVOID **)&odp, odpmsg);	oldnw = newnw;
		if (NULL == (odp = (char **)callop(newnw, sizeof (char *),
		 odpmsg)))
		  {
			memzap((AVOID **)&out, outmsg);	return NULL;
		  }
	  }
	if ((NULL != cpname) && ('\0' != *cpname))
	  {
		nch += strlen(cpname);	nw++;	(*acp)++;
	  }
	/* also reserve memory for the correct number of chars */
	if ((NULL == out) || (newlen > oldlen))
	  {
		memzap((AVOID **)&out, outmsg);	oldlen = newlen;
		if (NULL == (out = (char *)mallop(newlen, outmsg)))
		  {
			memzap((AVOID **)&odp, odpmsg); return NULL;
		  }
	  }
	zeromem((AVOID *)out, newlen);
	/* copy the input string to the output string, prepending the
	 program name if appropriate */
	if ((NULL != cpname) && ('\0' != *cpname))
	  {
		for (cp = cpname, ocp = out; '\0' != *cp; cp++, ocp++)
			*ocp = *cp;
		*ocp++ = ' ';
	  }
	else {
		ocp = out;	stu = st;
	     }
	for (cp = stu; '\0' != *cp; cp++)
	   {
		if (isxgspace(*cp))
		  {
			while (isxgspace(*(cp + 1)))
			     {
				cp++;
				if ((*cp == '\n') || (*cp == '\0')) break;
			     }
			*ocp++ = ' ';
		  }
		else *ocp++ = *cp;
	   }
	*ocp = '\0';
	/* go back through the entire string, picking out the words and
	 null-terminating each word */
	for (i = 0, ocp = out, dvp = odp; (i < *acp) && ('\0' != *ocp);
	  dvp++, i++)
	   {
		*dvp = ocp;
		while (!isxgspace(*ocp)) ocp++;
		*ocp++ = '\0';
	    }
	*dvp = NULL;	return odp;
}

void make_upper(char *name)
{ /* this capitalizes the string `name' */
	register char	*cp;

	for (cp = name; *cp != '\0'; cp++)
	   if ('a' <= *cp && *cp <= 'z') CAPITALIZE(*cp);
}

char *trname(char *name)
{ /*	This looks in the environment variables for the name `name' and
	for an uppercase version thereof.  It returns the translation
	thereby obtained if one exists; otherwise it returns the name itself.
	Note that memory for the returned name is grabbed here if necessary,
	so the calling routine doesn't need to.
	We don't bother to look if there's a $ or a / in
	the name. This version works under Unix and VMS. */
	int	low;
	char	uname[256], *nn, *un;

	/* look to see if there's a dollar sign or a slash somewhere.
	If there is, we don't try to do any further translation. */
	for (nn = name; '\0' != *nn; nn++)
#ifdef VMS
		if ((*nn == ']') || (*nn == '.')) return name;
#else
		if ((*nn == '$') || (*nn == '/')) return name;
#endif
	if (NULL != (nn = getenv(name)) && ('\0' != *nn)) return nn;
	/*no environment vbl so far -- try uppercase*/
	for (nn = name, un = &(uname[0]), low = 0; '\0' != *nn; nn++, un++)
	   {
		if ('a' <= *nn && *nn <= 'z')
		  {
			low = 1;	*un = (*nn) & 0137;
		  }
		else *un = *nn;
	   }
	if (!low) return name;	/* no lower-give up.*/
	*un = '\0';	nn = getenv(uname);
	if (nn && *nn) return nn;
	return name;	/* if we've still failed, return input */
}

#ifdef VMS		/* VMS-specific routines */
#define		VMS_DIR_PROT	(0xA000)

int dir_access(char *name, int mode)
{ /* this performs an "access" check on a directory.
	For VMS we ignore a request for write access,
	since under VMS that causes more trouble than it solves. */
	register int	retval;
	int		i, dirlen, istop;
	char		*cp, *dirname, *ocp;

	cp = trname(name);	/* get logical transl. if needed */
	dirname = mallop(strlen(cp) + 17, "dirname in dir_access");
	strcpy(dirname, cp);
	dirlen = strlen(dirname);
	if (dirlen < 2) return -2;
	if (dirname[dirlen-1] != ']') return -3;
	dirname[dirlen-1] = '\0';	/* get rid of final ']' */
	/* back up to previous left-bracket or dot.  We need to convert, say,
	DUA0:[XGEN.STUFF] to DUA0:[XGEN]STUFF.DIR and
	XMAIN:[XGDOC] to XMAIN:[000000]XGDOC.DIR. */
	for (cp = dirname + (dirlen-2), i = dirlen-2; i--; cp--)
	  if (NULL == cp || *cp == '\0' || *cp == '[' || *cp == '.') break;
	if (*cp == '[') istop = 1;	else if (*cp == '.') istop = 0;
	else return -4;
	ocp = (char *)mallop(strlen(cp + 1) + 2, "ocp in dir_access");
	strcpy(ocp, cp+1);	*cp = '\0';
	sprintf(dirname, "%s%s%s.DIR",
		dirname, (istop ? "[000000]" : "]"), ocp);
	memzap((AVOID **)&ocp, "ocp in dir_access");
	mode &= ~2;	/* VMS write permission is a pain in the ass! */
	retval = access(dirname, mode);
	memzap((AVOID **)&dirname, "dirname in dir_access"); return retval;
}

int create_direc(char *dirname)
{ /* This creates a directory under either VMS
	and returns the error code thus generated. In particular,
	it returns 0 if it creates the directory,
	1 if the directory already exists, and -1 if the creation fails. */
	static $DESCRIPTOR(vms_desc," ");
	static u_short	enable_mask = 0xffff;
	u_short		mode;
	int		retval;

	mode = VMS_DIR_PROT;
	vms_desc.dsc$a_pointer = dirname;
	vms_desc.dsc$w_length = strlen(dirname);
	retval = lib$create_dir(&vms_desc,0,&enable_mask, &mode);
	return (retval == SS$_CREATED ? 0 : (retval == SS$_NORMAL ? 1 : -1));
}

int set_logical(char *logical_name, char *value)
{ /* this mimics the VMS call
		$ DEFINE inname value
	 with appropriate capitalizations.
	It returns the error code from lib$set_logical, with the exception
	that the return values SS$_NORMAL and SS$_SUPERSEDE both return 1. */
	register char	*cp;
	char		*curval;
	int		retval;
	static $DESCRIPTOR(logdesc," ");
	static $DESCRIPTOR(valuedesc," ");

	make_upper(logical_name);	make_upper(value);
	/* if it's already defined as given, return success code */
	curval = getenv(logical_name);
	if ((curval) && ('\0' != *curval) &&
		!strcmp(curval, value)) return 1;
	logdesc.dsc$a_pointer = logical_name;
	logdesc.dsc$w_length = strlen(logical_name);
	valuedesc.dsc$a_pointer = value;
	valuedesc.dsc$w_length = strlen(value);
	retval = lib$set_logical(&logdesc, &valuedesc);
	if (SS$_NORMAL == retval || SS$_SUPERSEDE == retval) return 1;
	else return retval;
}

#else	/* Unix versions of the above... */

int dir_access(char *name, int mode)
{ /* under Unix: equiv. to a regular access */
	return access(name, mode);
}

int create_direc(char *dirname)
{ /* wrapper for 'mkdir(dirname, 0775)' */
	return mkdir(dirname, 0775);
}

/* Next are some general routines associated with environment variables.
 The `dosetenv' command will (I hope) work under ANSI C, K&R C,
 in the presence of existing "putenv" commands.
 It does require a working external variable called either "environ"
 (under K&R C) or "_environ" (under ANSI C).
	The routine returns 0 on success, various negative integers on failure.
	See comments below for failure codes.
 It appears that recent versions of SunOS do know about setenv() and unsetenv()
 but Irix 6.5 does not.
*/

#ifdef sgi
#define	NEED_SETENV	1
#endif

#ifdef NEED_SETENV
extern char	**environ;

static void unsetenv(const char *lonam)
{
 /* my attempt at implementing unsetenv */
	char		*cp, **dcp, **next;
	size_t		lolen;

	/* make sure the alleged environment variable is plausible */
	if ((NULL == lonam) || ('\0' == *lonam)) return;
	lolen = strlen(lonam);	if (lolen < 1) return;
	/* step through the current environment and look to see if the
	 environment variable lonam is actually there */
	for (dcp = environ; NULL != *dcp; dcp++)
	   {
		cp = *dcp;
		if ((!strncmp(lonam, cp, lolen)) &&
		 *(cp + lolen) == '=')
		  {
			/* blank out this element */
			for ( ; '\0' != *cp; cp++) *cp = '\0';
			/* if this is the last element in the
			 environment, set it to NULL and return */
			next = dcp + 1;
			if (NULL == *next)
			  {
				*dcp = NULL; return;
			  }
			/* otherwise, move everything down */
			for ( ; NULL != *next; next++)
				*(next - 1) = *next;
		  }
	   }
}

static int setenv(const char *lonam, const char *newval, int overwrite)
{
 /* My implementation of Version 7 AT&T setenv where it's missing.
  The assumption here is that `putenv' DOES exist here. */
	char		*curval;
	size_t		lolen, vallen;
	static char	*locstr;

	/* see if the name is already in the environment */
	lolen = strlen(lonam);	vallen = strlen(newval);
	if ((lolen < 1) || (vallen < 1)) return 0;
	curval = getenv(lonam);
	if ((NULL != curval) && ('\0' != *curval))
	  {
		/* special cases: no-overwrite or already correct */
		if (!overwrite) return 0;
		if (!strcmp(curval, newval)) return 0;
	  }
	/* otherwise, reserve memory and put it in the environment */
	if (NULL == (locstr = mallop(lolen + vallen + 4,
	 "locstr in setenv"))) return EALLOC;
	strcpy(locstr, lonam);	strcat(locstr, "=");
	strcat(locstr, newval);	return putenv(locstr);
}
#endif

int unset_logical(char *lname)
{ /* un-set an environment variable and return success */
	if (NULL == getenv(lname)) return 1;
	unsetenv(lname);	return 2;
}	

int set_logical(char *lname, char *newvalue)
{ /*	This sets the environment variable `lname' equal to `newvalue'.
	Third time around: now we're using the setenv() function for this. */
	int		res;
	char		*oldval;

	/* are the arguments legal ? */
	if ((NULL == newvalue) || (NULL == lname) ||
	  ('\0' == *newvalue) || ('\0' == *lname)) return -1;
	/* don't do anything special if the name is already set correctly */
	if ((NULL != (oldval = getenv(lname))) && ('\0' != *oldval) &&
	 !strcmp(oldval, newvalue)) return 2;
	res = setenv(lname, newvalue, 1);
	return (res) ? -5 : 1;
}

void cpymem(AVOID *mo, AVOID *mi, unsigned nb)
{	/* This copies exactly `nb' bytes from mi to m0. */
	register char	*a,*b;

	a = (char *)mo; b = (char *)mi; while (nb--) *a++ = *b++;
}

int envtoname(char *namein, char *nameout)
{ /*	This reads a string like $WORK/fe10cm.brs and returns the full
	name translated out from that.  It sticks the result in `nameout',
	which must have memory allocated for it externally.  The routine
	returns 1 if an environment variable DOES get translated;
	it returns 0 otherwise. */
	char	*cp, logbuf[40], *val, *ocp;

	strcpy(nameout, namein);
	 /* name must begin with a dollar sign */
	if (*namein != '$') return 0;
	 /* copy the environment-variable part (thru the first slash) */
	for (cp = namein + 1, ocp = &(logbuf[0]); '\0' != *cp; cp++, ocp++)
	   {
		*ocp = *cp; if (*cp == '/') break;
	   }
	*ocp = '\0';	/* null terminate the putative environment variable */
	val = getenv(logbuf);	/* try to translate the environment variable*/
	if ((NULL == val) || ('\0' == *val)) return 0; /* failure: give up */
	strcpy(nameout, val);	/* success: copy translation to output name */
	if (*cp) strcat(nameout, cp);	/* tack rest of input onto output*/
	return 1;
}
#endif		/* close the great long "else" to "ifdef VMS" */

void lowercaseify(char *str)
{
	register char	*cp;

	for (cp = str; '\0' != *cp; cp++)
	   if ('A' <= *cp && *cp <= 'Z') *cp += 'a' - 'A';
}

char *filepart(char *name, int extension_code)
{ /*	This returns the "file part" of a file name, i.e. the part following
	the last slash (Unix filespec) or the last bracket or colon (VMS
	filespec).  If "extension_code" is 0, the entire file part is returned;
	if "extension_code" is 1, everything except stuff starting
	with the second dot or the first semicolon is returned;
	if "extension_code" is 2, only the main body of the file name
	(up to the first dot or semicolon) is returned.
	if "extension_code" is 3, the main body only is returned,
	even if there's no front-matter (slashes, etc.).
	In the Unix versions, colons, semicolons, and brackets don't count.
	We allow Unix versions understand VMS filename conventions
	so that we can deal across networks nicely.
	We now deliver this output from a static area, so we need to
	copy the output to a safe area if we want to keep it.
	Special case: if name is NULL, we delete memory for the static area. */
	register char	*cp, *string_end;
	char		*start, *lastdelim, *lastdot;
	static char	*retname;
	static char	retmsg[] = "retname in filepart";
	static size_t	aclen, lenstr, oldlen;

	if (NULL == name) /* special case: kill off remaining memory */
	  {
		memzap((AVOID **)&retname, retmsg);	return NULL;
	  }
	lastdelim = lastdot = name - 1;
	for (cp = name; '\0' != *cp; cp++)
	   {
		if (*cp == ':' || *cp == ']' || *cp == '/') lastdelim = cp;
		if (*cp == '.' || *cp == ';')
		  {
			if ((lastdelim >= name) || (extension_code >= 3))
				lastdot = cp;
		  }
	   }
	/* if no delimiters or dots have been found, return
	 a copy of the input string */
	if ((lastdelim < name) && (lastdot < name || !extension_code))
	  {
		aclen = strlen(name);	start = name;
	  }
	/* otherwise, we return somewhat different strings depending
	on the value of extension_code */
	else {
		if (lastdot <= lastdelim || lastdot < name) lastdot = cp;
		string_end = (extension_code ? lastdot : cp);
		if (extension_code >= 2 && lastdot > name) /* just the name */
		  {
			for (cp = lastdot - 1; cp > lastdelim; cp--)
				if (*cp == '.') break;
			string_end = (cp > lastdelim ? cp : lastdot);
		  }
		aclen = string_end - lastdelim;	start = lastdelim + 1;
	     }
	lenstr = aclen + 2;	if (lenstr < 40) lenstr = 40;
	if ((NULL == retname) || (lenstr > oldlen))
	  {
		memzap((AVOID **)&retname, retmsg);	oldlen = lenstr;
		if (NULL == (retname = (char *)mallop(lenstr, retmsg)))
			return start;
	  }
	zeromem((AVOID *)retname, lenstr);
	strncpy(retname, start, aclen);	return retname;
}

void swap(AVOID *b0, AVOID *b1, unsigned dwidth)
{ /* this interchanges `dwidth' bytes between location b0 and location b1*/
	register char	*c0, *c1;
	char		ct;

	for (c0 = (char *)b0, c1 = (char *)b1; dwidth--; c0++, c1++)
	   {
		ct = *c0;	*c0 = *c1;	*c1 = ct;
	   }
}

int allnumeric(char *s)
{ /* This returns true iff s contains only digits, one decimal, and an
	initial minus sign. A bug for negative numbers was fixed Oct 94. */
	char	*cp;
	int	alreadydot;

	alreadydot = 0;
	for (cp = s; '\0' != *cp; cp++)
	   {
		if (*cp == '.')
		  {
			if (alreadydot) return 0;
			alreadydot = 1;
		  }
		else if (*cp == '-')
		  {
			if (cp != s) return 0;
		  }
		else if ((*cp < '0') || (*cp > '9')) return 0;
	   }
	return 1;
}

void kill_mail(MAIL **headdp, int *nmboxp)
{ /* This deletes memory for the MAIL structures. */
	int	i;
	MAIL	*thd;

	if ((NULL != headdp) && (NULL != *headdp))
	  {
		if (*nmboxp < 2) *nmboxp = 1;
		for (i = 0, thd = *headdp; i < *nmboxp; i++, thd++)
		   if (NULL != thd)
		     {
			memzap((AVOID **)&(thd->data8),
				"thd->data8 in kill_mail");
			memzap((AVOID **)&(thd->data16),
				"thd->data16 in kill_mail");
			memzap((AVOID **)&(thd->overflow),
				"thd->overflow in kill_mail");
			memzap((AVOID **)&(thd->runheader),
				"thd->runheader in kill_mail");
		      }
		memzap((AVOID **)headdp, "*headdp in kill_mail");
		*nmboxp = 1;	/* so we don't delete too many on next call */
	  }
}

int managekill(int inglob)
{ /* if inglob is negative, we return the current value of keepglob
 without changing it. If inglob is non-negative, we set keepglob to it
 and return it */
	static int	keepglob;

	if (inglob >= 0) keepglob = inglob;	return keepglob;
}

int xgstrend(char *bigstr, char *suffix, int cutit)
{ /*	This returns 1 iff `bigstr' ends with the suffix `suffix'.
	By convention we return 1 if the two strings are identical, too.
	If `cutit' is nonzero, we terminate `bigstr' just before the beginning
	of suffix. This one is case-sensitive. */
	char	*cpi, *cpt, *cend, *tend;

	if ((NULL == bigstr) || (NULL == suffix) || ('\0' == *bigstr)
	 || ('\0' == *suffix)) return 0;
	for (cpi = bigstr; '\0' != *cpi; ) cpi++;	cend = cpi - 1;
	for (cpt = suffix; '\0' != *cpt; ) cpt++;	tend = cpt - 1;
	for (cpi = cend, cpt = tend; (cpi >= bigstr) && (cpt >= suffix);
	 cpi--, cpt--) if (*cpi != *cpt) return 0;
	if ((cutit) && (cpi > bigstr)) *(cpi + 1) = '\0';
	return 1;
}

void helpapp(char *pname)
{ /*	This prints an appropriate help string for the application `pname'.
	As of 19 May 05, if errors arise we complain but don't exit. */
	char		*docd, *dotxt, *pnbody;
	static char	*doname;
	static size_t	dolen, olddolen;
	static FILE	*fph;
	static char	hsuff[] = ".txt";
	static char	donmsg[] = "doname in helpapp";

	if (NULL == pname) /* special case: eliminate memory for help */
	  {
		memzap((AVOID **)&doname, donmsg);	return;
	  }
	if ('\0' == *pname)
	  {
		printf("X-GEN help unavailable: no application specified\n");
		return;
	  }
	if (NULL == (docd = get_xgdoc(2)))
	  {
		printf("Cannot find directory containing X-GEN help\n");
		return;
	  }
	if ((NULL == (pnbody = filepart(pname, 2))) || ('\0' == *pnbody))
	  {
		printf("Application's name cannot be determined\n");
		return;
	  }
	dolen = strlen(docd) + strlen(pnbody) + 8;
	if (dolen < 256) dolen = 256;
	if ((NULL == doname) || (dolen > olddolen))
	  {
		memzap((AVOID **)&doname, donmsg);	olddolen = dolen;
		if (NULL == (doname = (char *)mallop(dolen, donmsg)))
		  {
			fprintf(stderr,
			 "Can't provide help: memory allocation problem\n");
			olddolen = 0;	return;
		  }
	  }
	sprintf(doname, "%s/%s%s", docd, pnbody, hsuff);
	fph = NULL;
	if (NULL == (fph = fopen(doname, "r")))
	  {
		printf("Documentation for X-GEN program %s unavailable\n",
			pnbody);	return;
	  }
	dotxt = doname;
	while (NULL != fgets(dotxt, 256, fph)) fputs(dotxt, stdout);
	if (NULL != fph) (void)fclose(fph);
}
	
void versionapp(char *pname)
{ /*	This prints an appropriate version string */
	char		*pnbody;
	static char	copy1[] =
	"Copyright (c) 2002, Illinois Institute of Technology.\n";
	static char	copy2[] = 
 	"See the file 'LICENSE' for information on usage and redistribution\n";
	static char	copy3[] =
	"of this file, and for a DISCLAIMER OF ALL WARRANTIES\n";

	if ((NULL != pname) && ('\0' != *pname))
	  {
		pnbody = filepart(pname, 2);
		if ((NULL == pnbody) || ('\0' == *pnbody)) pnbody = pname;
		printf("X-GEN application `%s', Version %s\n",
			pnbody, XGEN_VERSION);
	  }
	else	printf("X-GEN Version %s\n", XGEN_VERSION);
	fputs(copy1, stdout);	fputs(copy2, stdout);	fputs(copy3, stdout);
}

char **flagproc(int ac, char *av[], char *choices)
{ /* general-purpose flag-argument processor. We handle --help and --version
   correctly here, but they're usually intercepted in xgjournal() anyway.
   If we _do_ end up handling --help or --version we return av as a code. */
	char		*cp, *qp, tc;
	char		**dcp, **workdp, **wdp, **tadp, **rdp, **zadp;
	size_t		nchoices, maxlen, thislen;
	static char	*flagwb, **flagrdb;

	/* special case: if ac==0 and av == NULL, we're calling this routine
	 simply to get rid of memory for the flagwb and flagrdb vectors */
	if ((!ac) && (NULL == av))
	  {
		memzap((AVOID **)&flagrdb, "flagrdb in flagproc");
		memzap((AVOID **)&flagwb, "flagwb in flagproc");
	  }
	/* first, see if there are any arguments at all */
	if ((ac < 2) || (NULL == av) || (NULL == *av) || (NULL == *(av+1)))
		return NULL;
	/* new feature as of May 05: we screen for --help and --version before
	 creating the various internal vectors, so that we don't waste memory */
	for (dcp = av + 1; dcp < av + ac; dcp++)
	   {
		if ((!strncmp(*dcp, "--help", 6)) ||
		 (!strncmp(*dcp, "--HELP", 6)))
		  {
			helpapp(*av);	return av;
		  }
		else if ((!strncmp(*dcp, "--version", 9)) ||
		 (!strncmp(*dcp, "--VERSION", 9)))
		  {
			versionapp(*av);	return av;
		  }
	   }
	if ((NULL != choices) && ('\0' != *choices))
	  {
		for (cp = choices; '\0' != *cp; cp++) /* capitalize string */
			if ('a' <= *cp && *cp <= 'z') *cp -= ('a' - 'A');
		nchoices = strlen(choices) + 1;
	  }
	else	nchoices = 1;
	/* reserve memory for as many strings as we have
	choices of potential arguments */
	if (NULL == (workdp = (char **)callop(nchoices, sizeof (char *),
		"workdp in flagproc"))) return NULL;
	/* reserve memory for the individual strings */
	for (dcp = av + (ac - 1), maxlen = 0; dcp > av; dcp--)
		if ((thislen = strlen(*dcp)) > maxlen) maxlen = thislen;
	if (NULL == (flagwb = (char *)callop(nchoices * maxlen,
	 sizeof (char), "flagwb in flagproc")))
	  {
		memzap((AVOID **)&workdp, "workdp in flagproc"); return NULL;
	  }
	zeromem((AVOID *)flagwb, (size_t)(nchoices * maxlen)); /* blank it */
	for (wdp = workdp, qp = flagwb; wdp < (workdp + (nchoices - 1));
	 wdp++, qp += maxlen) *wdp = qp;
	/* Now go through the arguments.  We go through them backward
	so that if the user invokes the same argument more than once
	the LAST one will apply. */
	for (dcp = av + (ac - 1), wdp = workdp; dcp > av; dcp--)
	/* if an argument begins with a minus, it contains flags.
	Examine the remainder of the argument string and see if any
	of the characters is found in the list of choices.
	*/
	   if (**dcp == '-') /* don't touch non-flag arguments */
	     {
		cp = (*dcp) + 1;
		if (nchoices < 2) return NULL;
		for (tadp = wdp; '\0' != *cp; cp++)
		   { /* capitalize the argument first */
			tc = *cp;
			if ('a' <= tc && tc <= 'z') tc -= ('a' - 'A');
			/* go through the choices of possible flags and see
			if this character is one of them. */
			for (qp = choices; '\0' != *qp; qp++)
				if (*qp == tc) break;
			if ('\0' == *qp) /* unrecognized character. */
			  { /* if we've found one or more flags within
				the current argument, we attach the remainder
				of this argument to the last flag. */
				if (wdp > tadp) strcpy((*(wdp - 1)) + 1, cp);
				/* otherwise we ignore the character */
				break;	/* skip to next argument */
			  }
			else { /* a recognized flag.  First, see if
				we've already encountered it. */
				if (wdp > workdp)
				  {
					for (zadp = workdp;
					 NULL != zadp && NULL != *zadp; zadp++)
					   if (**zadp == tc) break;
				  }
				else zadp = NULL;
				/* if we've seen this flag, ignore it.
				Otherwise, save it & advance the flag pointer*/
				if (NULL == zadp || NULL == *zadp)
					**wdp++ = tc;
			     }
		   } /* end of loop thru chars of this argument */
	     } /* end of loop through arguments */
	*wdp = NULL;	/* we put an NULL string out on the end */
	if (wdp > workdp) /* set up an array of strings of the right length */
	  {
		if (NULL != (flagrdb = (char **)callop(
		 (size_t)((wdp - workdp) + 1), sizeof (char *),
		 "flagrdb in flagproc")))
		  {
			for (wdp = workdp, rdp = flagrdb; NULL != *wdp;
			  wdp++, rdp++) /* put the data into those strings */
				*rdp = (**wdp ? *wdp : NULL);
		  }
	  }
	else flagrdb = NULL;
	memzap((AVOID **)&workdp, "workdp in flagproc"); return	flagrdb;
}

void globalkill(void)
{ /* This frees memory for various global quantities. In each case the
	 object is set to zero so that multiple calls won't hurt hurt anything.
	 This routine calls `memzap' (indirectly) to do the freeing.
	Apr 03: we refrain from deleting memory for some of these
	 items if managekill returns 1. We'll get more selective
	 about that later. */
	int	ier;

	(void)activmask(0, 0, 0);	(void)managelca(0);
	(void)str2acav(NULL, &ier);
	nomoredv();	/* eliminates memory for stuff in xgenmask */
	(void)xgaccess(NULL, 0, -1);	/* zaps mem for augname in xgaccess */
	(void)imagename(NULL);
	(void)manageice(0);	kill3d();
	(void)mymedian(NULL, 0);
	(void)imedian(NULL, 0);
	(void)read_image(NULL, 0, NULL);
	(void)ftables(0);
	(void)helpapp(NULL);
	(void)expandit(NULL, 0);
	(void)gspline(0, &ier);	/* includes foffset(0), in effect */
	(void)flagproc(0, NULL, NULL);	/* get rid of flag-processing memory */
	(void)iminfo(0);	/* eliminate image information memory */
	(void)filepart(NULL, 0);	/* eliminate filepart memory */
	(void)startspline(-1);	/* get rid of memory for nameinpath */
	if (managekill(-1)) return; /* remaining steps skipped in this case */
	(void)sharename(0);	(void)xgprogname(NULL, 1);
}

/* the next several routines are used to convert various data types
 between one format and another.  One class of machines stores 16-bit
 integers with the high byte as the most significant and another with
 it as least significant, and so on.  These routines swap back and forth. */

void float_cvt(u_char *z)
{ /*	This converts a DEC float to a non-DEC Unix float or vice-versa */
	u_char	t;

#ifdef LITTLEND
	if ((*z) || (*(z+1))) *z += 1;
	t = *(z+1);	*(z+1) = *z;	*z = t;
#else
	t = *(z+1);	*(z+1) = *z;	*z = t;
	if (*z) *z -= 1;
#endif
	t = *(z+3);	*(z+3) = *(z+2);	*(z+2) = t;
}

void flcvt(u_char *z, int direction)
{ /*	As above, but capable of going either direction */
	u_char	t;

	direction = 1 - direction;
#ifdef VMS	/* under VMS we have a tougher conversion to do */
	if (direction)
	  {
		t = *(z+1);	*(z+1) = *z;	*z = t;
		if (*z) *z -= 1;
	  }
	else {
		t = *(z + 1);	*(z+1) = *z;	*z = t;
	     }
	t = *(z + 3);	*(z + 3) = *(z + 2);	*(z + 2) = t;
#else	/* between OSF/1 and Big-end Unix we just swap words & bytes */
	t = *(z + 3);	*(z + 3) = *z;		*z = t;
	t = *(z + 1);	*(z + 1) = *(z + 2);	*(z + 2) = t;
#endif
}

u_short uswab(u_short ush)
{ /* this swaps the bytes in a 16-bit quantity and returns the result*/
	return ((ush & 0xFF) << 8) | (ush >> 8);
}

xgulong swawb(xgulong ulv)
{ /* this swaps both words and bytes within a xgulong and returns the result*/
	return ((ulv & 0xFF)<<24) |((ulv & 0xFF00)<< 8) |
		((ulv >> 8) & 0xFF00) | (ulv >> 24);
}

short sswab(short sh)
{ /*	This converts a SIGNED short.  This is comparatively tricky. */
	u_short	ush;

	ush = (u_short)sh;
	if (ush & 0x80)	return -1 - ((short)
		(0xffff - (((ush & 0xff) << 8) | (ush >> 8))));
	else		return (((ush & 0x7f) << 8) | (ush >> 8));
}

xglong lswawb(xglong lv, int direc)
{ /*	This converts a signed xglong.  This is also relatively tricky. */
	xgulong	uv, ev;

	uv = (xgulong)lv;
	if (direc)
	  {
		ev = ((uv & 0xff) << 24) | ((uv & 0xff00) << 8) |
		 ((uv >> 8) & 0xff00) | ((uv >> 24) & 0x7f);
		return (lv < 0) ? ev | 0x80 : ev;
	  }
	else {
		ev = ((uv & 0x7f) << 24) | ((uv & 0xff00) << 8) |
		 ((uv >> 8) & 0xff00) | (uv >> 24);
		return (lv & 0x80) ? -ev : ev;
	     }
}

static char *journaled(char *titl)
{ /*	If we've used the command before during the current session
	(as measured by the presence of this command in the journal file)
	we return the most recent string found in the journal file,
	suitably modified to account for the current situation */
	int		lineno, titlen, lenret;
	XGNAME		*losh;
	static FILE	*fpj;
	static char	*jname, *jnlentry, *cp, *cpz, *retstr;

	if ((NULL == titl) || ('\0' == *titl)) return NULL; /*need good title*/
	if (NULL == (losh = sharename(7))) return NULL;
	jname = journalname(); /* determine name of journal file */
	if (NULL == (fpj = xgopen(jname, "r"))) return NULL; /* open journal */
	jnlentry = &(losh->xmm_miscc[1536]);	/* convenient memory */
	retstr = &(losh->xmm_miscc[1792]);	/* ditto */
	*retstr = '\0';	/* signal that we haven't found this command yet */
	/* go through the lines of the journal file looking to see if we've
	 already run this application */
	titlen = strlen(titl);
	for (lineno = 0; NULL != fgets(jnlentry, 254, fpj); lineno++)
	   {
		if (*jnlentry != '#') /* ignore comment lines */
		  { /* find first useful character in line */
			for (cp = jnlentry; '\0' != *cp; cp++)
			   if (('\n' == *cp) || (('.' <= *cp) && (*cp <= 'z')))
				break;
			if (('\n' != *cp) && ('\0' != *cp))
			  {
				cpz = filepart(cp, 0);	/* name of appl. */
				if (!strncmp(cpz, titl, titlen))
					strcpy(retstr, cp);
			  }
		  }
	   }
	(void)fclose(fpj);	/* we need to close the journal regardless */
	if ('\0' == *retstr) return NULL;	/* no instances found */
	lenret = strlen(retstr);
	if ('\n' == *(cp = retstr + (lenret - 1))) *cp = '\0';
	return retstr; /* return latest instance of this command */
}

void get_calibparams(int dtype, double *xspap, double *yspap, int *lincolp)
{ /*	This determines the default settings for calibration
	based on the detector type and dimension.
	As of March '04 we set *lincolp to zero for detectors for
	which we expect a real calibration image to be available. */
	int	stype, ssize;

	stype = dtype & 0377;
	ssize = detdimen(dtype >> 8);
	/* start with non-stupid defaults */
	*xspap = *yspap = 0.4125;	*lincolp = 45;
	switch(stype)
	 {
		case	1:	/* SDMS detectors */
			*xspap = 1.5681;	*yspap = 2.8;
			*lincolp = 20;	return;
		case	2: case 3:	/* Siemens, old Argonne */
			*xspap = *yspap = 0.508;	*lincolp = 0;	return;
		case	4:	/* FAST and post-1990 Argonne */
			if (ssize == 512) /* actual FAST detectors */
			  {
				*xspap = 0.144;	*yspap = 0.083333;
			  }
			else	*xspap = *yspap = 0.2261;
			*lincolp = 0;	return;
		case	5:	/* R-Axis II  or R-Axis IV */
			if (ssize == 1900) /* R-Axis II */
			  {
				*xspap = 0.40468; *yspap = 0.41781;
				*lincolp = 49;
			  }
			else {
				*xspap = *yspap = 0.7; /*R-Axis IV */
				*lincolp = 44;
			     }
			return;
		case	6:	/* Mar image plate detectors */
			if (ssize == 1200)
			  {
				*xspap = *yspap = 0.4125;
				*lincolp = 45;	return;
			  }
			else {
				*xspap = *yspap =0.45;
				*lincolp = 68;	return;
			     }
		case	10:	/* Siemens CCD */
			/* need something here to signal
			 special handling of Mosaic...
					SOMETHING CLEVER!
			*/
			*xspap = *yspap = 0.254;	*lincolp = 41;
			return;
		case	11:	/* Mar CCD */
			if (ssize == 4096)	/* Mar 300mm CCD */
			  {
				*xspap = *yspap = 0.46989322;	*lincolp = 65;
			  }
			else if (ssize == 3072) /* Mar 225 */
			  {
				*xspap = *yspap = 0.4907214;	*lincolp = 47;
			  }
			else {	/* Mar 165 */
				*xspap = *yspap = 0.33222; *lincolp = 50;
			     }
			return;
		case	12:	/* Brandeis CCD */
			*xspap = *yspap = 0.47; *lincolp = 23; return;
		case	13:	/* ADSC Quantum 4 */
			if (ssize == 1024)
			  {
				*xspap = *yspap = 0.38874;
				*lincolp = 50;
			  }
			else  {
				*xspap = *yspap = 0.37086;
				*lincolp = 48;
			      }
			return;
		case	14:	/* ADSC Quantum 210 */
			if (ssize == 2048)
				*xspap = *yspap = 0.43007;
			else	*xspap = *yspap = 0.42903;
			*lincolp = 50;	return;
		case	15:	/* ADSC Quantum 315 */
			if (ssize == 3072)
			  {
				*xspap = *yspap = 0.47104;
				*lincolp = 68;
			  }
			else {
				*xspap = *yspap = 0.46336;
				*lincolp = 69;
			     }
			return;
		case	17:	case	18: /* Aviex detectors (2x1 or 2x2) */
			*xspap = *yspap = 0.1;	*lincolp = 0; return;
		case	7: default:	/* generic 1024x1024 */
			*xspap = *yspap = 0.47;
			*lincolp = 48;	return;
	  }
}

char *smartdefa(char *title, int killem, int *ctypep)
{ /*	This generates an initial command string for the X-GEN command `title'.
	The string generated is returned. `killem' determines whether or not
	memory for a bunch of other X-GEN variables should be blotted out.
	*ctypep contains the detector type currently in use. */
	char		termi, *jver;
	int		linearcols, locflags, mval, nin, otype, retv, ftarg;
	int		ssize, stype, uplim, sbgcut, locwid, locht, detdim[2];
	double		fq, xc, yc, xspace, yspace;
	static char	two[258], varflags[40];
	static int	nfmhere[] = {
 65, 65, 65, 40, 65, 10, 10, 10,  8, 10, 18, 10, 12, 10, 12, 10, 10, 10, 10 } ;
	static int	announce_old = 0;
	static int	prevtype;
	static MAIL	lhd;
	static VSPLINE	*lspl;
	static FRAMEINDEX	*lfi;
	static FRAMEENTRY	*lfe;

	(void)xgprogname(NULL, 1);	(void)xgprogname(title, 4);
	prevtype = *ctypep;
		/* remember what kind of system this is */
	lfi = ftables(3);	/* try to get current frame index */
	*ctypep = dettype(detdim);
	stype = *ctypep & 0377;	ssize = *ctypep >> 8;
	if ((stype < 0) || (stype > 18)) stype = 0;
	else {
		nin = !strncmp(title, "calibrate", 9);
		lspl = gspline(3, &retv);
		if (detdim[0] > 1)
		  {
			lspl->v_detw = detdim[0]; lspl->v_maxx = detdim[0] +10;
		  }
		else {
			lspl->v_detw = 1024;	lspl->v_deth = 1034;
		     }
		if (detdim[1] > 1)
		  {
			lspl->v_deth = detdim[1]; lspl->v_maxy = detdim[1] +10;
		  }
		else {
			lspl->v_deth = 1024;	lspl->v_maxy = 1034;
		     }
	      }
	if (killem)
	  {
		termi = '\n';	ftarg = 7;
	  }
	else {
		termi = '\r';	ftarg = 15;
	     }
	/* if there's a mismatch between old & new types,
	 erase memory of the old one and, if possible, load up new info */
	if (prevtype != *ctypep)
	  {
		otype = prevtype & 0377;
		if ((otype < 0) || (otype > 15)) otype = 0;
		printf("Previous detector-type was %s; current type is %s%c",
			imtypestr(otype), imtypestr(stype), termi);
		xeflush(stdout);
		if (killem) globalkill(); /* erase old reserved memory */
		lfi = ftables(ftarg); /* get new frame limits*/
		lspl = gspline(5, &retv); /* try to get spline info */
		announce_old = 0;
	  }
	else if (!announce_old)
	  {
		printf("Detector type is %s%c", imtypestr(stype), termi);
		xeflush(stdout);	announce_old = 1;
	  }
	/* if we're in the middle of a project and we've run this
	 application before within the project, we return the string
	 associated with the most recent run of that application */
	if ((prevtype == *ctypep) && (strcmp(title, "refine1")) &&
	 (NULL != (jver = journaled(title)))) return jver;
	if (!strncmp(title, "border", 6))
	  {
		if ((NULL == lfi) || (lfi->fi_firstfm >= lfi->fi_lastfm)
		 || (lfi->fi_lastfm == 999)) lfi = ftables(ftarg);
		uplim = lfi->fi_firstfm + 30;
		if (stype >= 5) uplim = lfi->fi_firstfm + 8;
		if (lfi->fi_lastfm < uplim) uplim = lfi->fi_lastfm;
		nin = (ssize < 3) ? 4 : ((ssize < 4) ? 7 : 16);
		sprintf(two, "border -s%d -q0.42 %d %d",
			nin, (int)(lfi->fi_firstfm), uplim);
	  }
	else if (!strncmp(title, "calibrate", 9))
	 {
	   linearcols = (stype == 1) || (4 < stype && stype < 17);
	   mval = (stype == 4) ? 4 : 5;
	   if ((stype == 4) && (ssize == 3)) strcpy(varflags, "-hf0");
	   else strcpy(varflags, "-f0");
		/* we look for a current spline file, hoping to get detector
		dimensions from it */
	   if ((NULL == lspl) && (NULL == (lspl = gspline(5, &retv))))
	     {	/* try to get a frame index and look for dimensions that way */
	   	if ((NULL != lfi) || (NULL != (lfi = ftables(3))))
		  {
			locflags = JUST_HEADER;	lfe = lfi->fi_fmi;
			if (FRAME_OK == read_image(lfe->fmname,locflags, &lhd))
			  {
				locwid = lhd.xdimen;	locht = lhd.ydimen;
			  }
			else	locwid = locht = 1024;
		  }
		else locwid = locht = 1024;	/* desperation */
		xc = locwid / 2 - 1;	yc = locht / 2 - 1;
		get_calibparams(*ctypep, &xspace, &yspace, &linearcols);
	     }
	   else	{
		xspace = fabs(lspl->v_xrcm);
		yspace = fabs(lspl->v_yrcm);
		locwid = lspl->v_detw;	locht = lspl->v_deth;
		xc = lspl->v_axc;	yc = lspl->v_ayc;
		if (linearcols) linearcols = lspl->v_ncol;
		}
	   if (linearcols)
	   	sprintf(two, "calibrate -j %6.1f %6.1f", xc, yc);
/* OLD VERSION
		sprintf(two,
		"calibrate %s -m%d -l%d -x%6.4f -y%6.4f %6.1f %6.1f",
		varflags, mval, linearcols, xspace, yspace, xc, yc);
*/
	   else sprintf(two,
		"calibrate %s -m%d -x%6.4f -y%6.4f %6.1f %6.1f",
		varflags, mval, xspace, yspace, xc, yc);
	  }
	else if (!strncmp(title, "spots", 5))
	  {
		if ((NULL == lfi) || (lfi->fi_firstfm >= lfi->fi_lastfm) ||
		 (lfi->fi_lastfm == 999)) lfi = ftables(ftarg);
		if (stype <= 16) uplim = lfi->fi_firstfm + nfmhere[stype];
		else uplim = lfi->fi_firstfm + 15;
		if (lfi->fi_lastfm < uplim) uplim = lfi->fi_lastfm;
		switch (ssize)
		      {
			case	1: case	2:
				strcpy(varflags, "-w4.5");	break;
			case	3:	   strcpy(varflags, "-w5.7");	break;
			case	4: case	5: strcpy(varflags, "-w6");	break;
			default:	   strcpy(varflags, "-w6");	break;
		      }
		if (stype < 5)
		  {
			nin = 3; fq = 0.3; sbgcut = 20;
		  }
		else if (stype== 10)
		  {
			nin = 2; fq = 0.75; sbgcut = 80; }
		else {
			nin = 2; fq = 0.5; sbgcut = 80;
		     }
		sprintf(two,
		 "spots -f%d -l%3.1f -m2 -p0.7 %s %d %d %d",
		 nin, fq, varflags, (int)(lfi->fi_firstfm), uplim, sbgcut);
	  }
	else if (!strncmp(title, "cleanm", 9))
		strcpy(two, "cleanm -p2.5 -g 0.008");
	else if (!strncmp(title, "integrate", 9))
	  {
		if ((NULL == lfi) || (lfi->fi_firstfm >= lfi->fi_lastfm) ||
		 (lfi->fi_lastfm == 999)) lfi = ftables(ftarg);
		switch (stype)
		  {
			case	0 : case	1:
			 strcpy(varflags, "-i0.08 -b14 -r20");	break;
			case	2:
			 strcpy(varflags, "-i0.08 -zb14 -r20");	break;
			case	3:	case	4:
			 strcpy(varflags, "-i0.08 -b14 -lr20");	break;
			default:
			 strcpy(varflags, "-i0.09 -b9 -lpr10");	break;
		  }
		sprintf(two, "integrate -nc10 %s -m0 -y0.03 %d %d",
		 varflags, (int)(lfi->fi_firstfm), (int)(lfi->fi_lastfm));
	  }
	else if (!strncmp(title, "anal_u", 6))
	  {
		if (stype <= 1)		strcpy(two, "anal_u -f0.05 -i3 -s 6");
		else if (stype <= 4)	strcpy(two, "anal_u -f0.05 -i3 4");
		else			strcpy(two, "anal_u -f0.05 -i3 1");
	  }
	else if (!strncmp(title, "filterc", 9))
	  {
		if ((NULL == lfi) || (lfi->fi_firstfm >= lfi->fi_lastfm) ||
		 (lfi->fi_lastfm == 999)) lfi = ftables(ftarg);
		if (stype <= 4) strcpy(varflags, "-i500. -n0.1");
		else strcpy(varflags, "-i5000. -n0.05");
		sprintf(two, "filterc -hd50 %s -m0.05 %d %d",
		 varflags, (int)(lfi->fi_firstfm), (int)(lfi->fi_lastfm));
	  }
	else if (!strncmp(title, "reduce", 6))
	  {
		if ((NULL == lfi) || (lfi->fi_firstfm >= lfi->fi_lastfm) ||
		 (lfi->fi_lastfm == 999)) lfi = ftables(ftarg);
		if (stype < 3) sprintf(two,"reduce -exd3 -f-3 -w5 -a%d,%d 1 0",
		 (int)(lfi->fi_firstfm), (int)(lfi->fi_lastfm));
		else sprintf(two, "reduce -exd4 -f-3 -w1 -a%d,%d 1 0",
		 (int)(lfi->fi_firstfm), (int)(lfi->fi_lastfm));
	  }
	else if (!strncmp(title, "recycle", 7))
	  {
		if ((NULL == lfi) || (lfi->fi_firstfm >= lfi->fi_lastfm) ||
		 (lfi->fi_lastfm == 999)) lfi = ftables(ftarg);
		sprintf(two, "recycle -exd3 -f9 -w5 -c9 -a%d,%d 1 0",
		 (int)(lfi->fi_firstfm), (int)(lfi->fi_lastfm));
	  }
	else if (!strncmp(title, "qec", 5))
		sprintf(two, "qec -c999 -n15 -p1 -r0.01 14");
	else if (!strncmp(title, "analmr", 7)) strcpy(two, "analmr -s");
	else if (!strncmp(title, "applycor", 9)) strcpy(two, "applycor -s0");
	else if (!strncmp(title, "scalek", 6))
		strcpy(two, "scalek -c9 -i0 -o2 -thp1 -q100 -s1");
	else if (!strncmp(title, "reject", 6)) sprintf(two, "reject -bf0.02");
	else if (!strncmp(title, "spaceg", 6)) strcpy(two, "spaceg");
	else if (!strncmp(title, "makemu", 6)) strcpy(two, "makemu -afs1.");
	else if (!strncmp(title, "showover", 8)) strcpy(two, "showover");
	else if (!strncmp(title, "stats", 6)) strcpy(two, "stats -o");
	else if (!strncmp(title, "xdump", 5)) strcpy(two, "xdump amask");
	else if (!strncmp(title, "cutoff", 6)) strcpy(two, "cutoff 99 3.1");
	else if (!strncmp(title, "mukb", 4)) strcpy(two, "mukb -s2 99 3.1");
	else strcpy(two, title); return two;
}

int findbool(int aqc, char *aqv[], char fla)
{ /*	This returns 1 iff the flag `fla' appears as one of the Boolean-style
	arguments within the strings aqv[1], . . . aqv[acq-1]. */
	int	i;
	char	flacomp, **dcp, *cp;

	/* look for both upper and lowercase */
	if (('a' <= fla) && (fla <= 'z')) flacomp = fla - ('a' - 'A');
	else if (('A' <= fla) && (fla <= 'Z')) flacomp = fla + ('a' - 'A');
	else return 0;
	for (i = 1, dcp = aqv + 1; i < aqc; i++, dcp++)
	   {
		if (**dcp == '-') /* only examine flag arguments */
		  {
		    for (cp = (*dcp) + 1; '\0' != *cp; cp++)
		      {
			if ((*cp == fla) || (*cp == flacomp)) return 1;
		      }
		  }
	   }
	return 0;
}

char *findsw(int aqc, char *aqv[], char swf)
{ /*	This returns a string containing a numeric (possibly floating-point)
	value. This will be the string following a flag matching the
	switch-flag `swf'. The string may have a form like "47,238". */
	int		i;
	char		swfcomp, **dcp, *cp, *ret;
	static char	zer[] = "0";

	/* look for both upper and lowercase */
	if (('a' <= swf) && (swf <= 'z')) swfcomp = swf - ('a' - 'A');
	else if (('A' <= swf) && (swf <= 'Z')) swfcomp = swf + ('a' - 'A');
	else return NULL;
	for (i = 1, dcp = aqv + 1; i < aqc; i++, dcp++)
	   {
		if (**dcp == '-') /* only examine flag arguments */
		  {
		    for (cp = (*dcp) + 1; '\0' != *cp; cp++)
		      {
			if ((*cp == swf) || (*cp == swfcomp))
			  {
				/* by convention, if the flag is NOT
				 followed by a negative sign or a digit,
				 return the string "0". Otherwise we
				 return the string we find. */
				ret = cp + 1;
				if (('-' == *ret) ||
				 (('0' <= *ret) && (*ret <= '9'))) return ret;
				else return zer;
			  }
		       }
		   }
	   }
	return NULL;
}

/* end of xgenstr.c */
