/* xgenviron.c -- XGEN utilities for setting up environment

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: startspline: handle Aviex detectors
	11 Aug 2006: startspline: manage Mar IP formats in a more abstract way
	11 Jun 2006: move filetoenv from xgenstr to here so it can be static
	 1 Jun 2006: glistname static; substr, countimages moved in;
		journaled, smartdefa, findbool, findsw moved out
	 3 Jan 2006: eliminate record_detxy calls; use lineend() for formatting
	20 Nov 2005: inline argument declarations and strict prototyping
	17 Nov 2005: sharename(1) replaces kill_share()
	28 Oct 2005: get rid of most static globals by brute force
	28 Sep 2005: get rid of duplicated REINDEX environment variable
	17 May 2005: documentation search looks for xgen executable, not spots
	23 Apr 2005: get_xgdoc(): search in xgdoc, not xgdoc/help (phooey)
	25 Mar 2005: get_xgdoc(): search in xgdoc/help, not xgdoc
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	10 Jan 2005: marccd300 spline file made available
	18 Dec 2004: get_xgdoc() can be invoked quietly if arg is 2 or 3.
	22 Oct 2004: ready() works with additional applications;
	 get_xgdoc() knows that document suffix is .txt, not .doc
	20 Oct 2004: default frame-range for spots changed and extended;
	 default calibrate params for pre-corrected detectors simplified;
	 params for reduce and recycle improved
	15 Oct 2004: ready(): file-existence tests for refall, refine1,
		scalem, mukb
	28 Aug 2004: routines used only here are declared as static; cosmetics
	13 Aug 2004: transcription bug in check_directories corrected
	 8 Aug 2004: add SCA and XSCA file types (SCALEPACK-style); cosmetics
	 2 Aug 2004: calls to record_detxy take arg:
		nonzero for all cases except calibrate()
	27 Jun 2004: cosmetic changes; SHELX output included in knowntype()
	18 Jun 2004: startspline() knows about Mar 225mm CCDs
	30 May 2004: add XXPLOR, XSHEL structure types, delete XHKL type;
		changes in newidname() so that picking a sample image works 
	20 Mar 2004: add XCNS as another X-GEN environment variable;
		points to CNS-style output from `makemu'
	19 Mar 2004: change hardwired search path for executables from
		/usr/lib/xgen to /usr/share/xgen
	29 Feb 2004: getproject() takes "fromcu" as a second argument
	28 Feb 2004: get_xgdoc() returns the environment variable it finds;
		getproject() does project-name search all in one place
	26 Jan 2004: ugly bug in sharename fixed: we don't make multiple copies
	24 Jan 2004: various static globals made into static locals
	21 Jan 2004: license message inserted
	 8 Jan 2004: global variables replaced by calls to routines, args;
		kill_share() written to handle static global name buffer;
		execlognames() returns # of logicals it finds
	13 Dec 2003: eliminate unnecessary variables in readytest()
	27 Nov 2003: eliminate pname, ptitle as globals
	10 Oct 2003: look for resources in /usr/lib/xgen as well as others
	19 Sep 2003: typecast size_t insize, insize+4 for printouts
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace some access() with xgaccess() calls
	25 Aug 2003: move some initialized global variables elsewhere
	22 Aug 2003: modify glistname(): listname now is part of nabuf,
		and glistname simply puts stuff into it; cosmetic changes
	20 Aug 2003: combine old header.c and environment stuff from setupio.c
 Relevant changes to setupio.c and header.c:
	10 Jul 2003: get_xgdoc() figures out where the X-GEN documents are.
	 9 Jul 2003: ready() routine figures out if we can run a given app.
		Create bash-command file as well as (t)csh command file.
	 5 Jan 2003: reorganized source-file header
	 4 Jan 2003: copyright notice changes to IIT
	 2 Jan 2003: moved `knowntype()' here from header.c along with
		static globals lnames, lsuffix, lattrs;
		separate static global variables dirdmm and dirci for the
		two applications of DIR concepts
	28 Dec 2002: in countimages, call maybe_image before read_image
	15 Nov 2002: mnbuf() created: provides memory for general-purpose buf;
		used in setting up set_logical() calls and elsewhere
	15 Apr 2002: in check_directories, return dirno + 1, not dirno
	25 Mar 2002: removed MSI attachments
	 1 Jun 1999: in `newidname', make WORK name point at current directory,
		and look around for $DATA as either the current directory or
		its parent--whichever has things that resemble images in them.
		Don't append `idname' to workname.
	15 Oct 1998: writelognames() doesn't construct the environment-var
		strings at all; instead, it calls `more_logicals' to get them.
	10 Oct 1995: larger static char arrays lname,loname,liname,fname
	10 Oct 1995: filetoenv(): use counter on various errors so messages
		appear only once each.
	20 May 1995: correct return codes from routines
	 2 Mar 1995: XGEN replaced by X-GEN in titles
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	16 Nov 1994: mydircreate() no longer static;
		major rewrite of more_logicals: more powerful, flexible
		execlogicals now calls more_logicals(0)
	 4 Nov 1994: flush buffer for xgensave.lst after writing into it
	27 Oct 1994: default name of WORK directory
		built from current directory, not $HOME
		bookkeeping of current working directory improved
	 5 Oct 1994: check for readability of input files before listing 'em
	30 Sep 1994: revise default list of file types; write to xgensave.lst
		writelognames:write out current environment,
		not defaults for RUN_ID
	28 Sep 1994: in `writelognames' we use the current environment
		variable settings rather than overriding them.
	 5 May 1994: source-file headers, POLYDET->XGEN,
		EOK (gspline call) find_dimension
		"xgulong" and "xglong" constructions
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: eliminating unnecessary diagnostics
		after `polyutil' bugfix
	
   Purpose: manipulations of environment variables in X-GEN
		and displaying appropriate header information
   Cross references: ftable.c, xgenio.c
*************************************************************************/
#define		NEED_UNISTD	1
#define		NEED_ERRNO	1
#define		NEED_STAT	1
#define		NEED_TIME	1
#include	"xgen.h"
#include	<dirent.h>
#define		NWORKVAR	27	/* # of environ. vars in $WORK */
#define		NSCRTVAR	8	/* # of environ. vars in $SCRT */
#define		NENVVARS	(NWORKVAR + NSCRTVAR)
#define		ABSOLUTE	"/usr/share/xgen"
#define		MAX_SPACE	17

#ifdef VMS
#define		FILEFORMAT1	"%s%s.%s"	/* VMS format */
#define		DIVIDER		':'
#define		DELIM		']'
#else
#define		FILEFORMAT1	"%s/%s.%s"	/* Unix format */
#define		DIVIDER		'/'
#define		DELIM		'/'
#endif

static char *indtolog(int indx, char **sufxdp)
{ /*	This returns the logical name associated with the index value `indx'
	and returns the appropriate suffix in *sufxdp. */

	static char	*lnames[NENVVARS] = {
	 "AMASK", "BGIMAGE", "CALDATA", "DARKIMAGE", /* written to WORK: */
	 "DELETIONS", "FLOOD", "FRAMES", "MISSING", "MREFOUT", "MULIST",
	 "MULTIREF", "REFERENCE", "SCALEIN", "SCALEOUT", "UPARAMS",
	 "USPLINE", "UREFLS", "XLOG", "CALCENTROIDS",
	 "ICERING", "XGENCMD", "XGENCOM",
	 "XXPLOR", "XCNS", "XSHEL", "XSCA", "XASCA",
	 "CENOUT", "CENTROIDS", "CHECKREFL",	/* written to SCRT: */
	 "REINDEX", "UDUMP", "USPOTLIST", "WMASK", "REFLIST" } ;
#ifdef VMS
	static char	*lsuffix[NENVVARS] = {
	 "AMS", "BGI", "CAI", "DKI", "DEL", "FLD", "FRM", "MIS", "MRF", "MU",
	 "MRF", "RFR", "SI", "SI", "UPR", "UCA", "URF", "XLG", "CCI",
	 "ICE", "CMD", "COM", "XXP", "XCN", "HKL", "SCA", "ASC",
	 "CEO", "CEN", "CKR", "REI", "UDU", "USP", "WMS", "RLI" } ;
#else
	static char	*lsuffix[NENVVARS] = {
	 "ams", "bgi", "cai", "dki", "del", "fld", "frm", "mis", "mrf", "mu",
	 "mrf", "rfr", "si", "si", "upr", "uca", "urf", "xlg", "cci",
	 "ice", "cmd", "com", "xxp", "xcn", "hkl", "sca", "asc",
	 "ceo", "cen", "ckr", "rei", "udu", "usp", "wms", "rli" } ;
#endif
	*sufxdp = NULL;	if ((indx < 0) || (indx >= NENVVARS)) return NULL;
	*sufxdp = lsuffix[indx];	return lnames[indx];
}

int xgsuffix(char *stri)
{ /*	Returns 1 iff. `stri' ends with one of the standard X-GEN suffixes */
	int	j;
	char	*cp, *sufx;

	for (j = 0; j < NENVVARS; j++)
	   {
		(void)indtolog(j, &sufx);
		if (xgstrend(stri, sufx, 0))
		  {
			cp = stri + strlen(stri) - strlen(sufx) - 1;
			if (*cp == '.') return 1;
		  }
	   }
	return 0;
}

static int readytest(char *envname, int minlen)
{ /*	This returns 1 iff `envname' is a filename or environment variable
	that points to a file that exists, is readable, and is larger than a
	minimum length given by minlen. We allow gunzipping on the fly here. */
	char		*tnam;
	int		retv;
	struct stat	sloc;

	retv = 0; /* assume that the test will fail (how pessimistic!) */
	tnam = trname(envname); /* remember translation of env var */
	/* see if it's readable, stat-able, and >= minimum length */
	if (0 == xgaccess(tnam, 04, 1))
	  {
		if (0 == stat(tnam, &sloc))
		  {
			if (((int)sloc.st_size) >= minlen) retv = 1;
		  }
	  }
	return retv;
}

int ready(char *app)
{ /*	This returns 3, 2, 1, or 0 depending on the application name 'app'
	and the contents of certain files, as follows.
	Type of		Status of	Status of	Return	Name of
	application	needed inputs	outputs		value	return status
        -=-=-=-=-=-	-=-=-=-=-=-=-	-=-=-=-=	-=-=-	-=-=-=-=-=-=-
	Required	absent		absent		 0	CANNOT
	Optional	any		any		 1	CAN
	Required	present		absent		 2	SHOULD
	Required	present		present		 3	NEEDNOT */
	if ((!strcmp(app, "setup")) || (!strcmp(app, "pmotif")))
	  {
		if (readytest("XGENCMD", 400) && readytest("FRAMES", 6))
			return NEEDNOT; else return SHOULD;
	  }
	else if (!strcmp(app, "border")) /* border: FRAMES needed */
	  {
		if (readytest("FRAMES", 6)) /* AMASK is output */
			return (readytest("AMASK", 2048) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if (!strcmp(app, "calibrate")) /* either FRAMES or CALDATA*/
	  {
		if ((readytest("FRAMES", 6)) || (readytest("CALDATA", 16384)))
			return (readytest("USPLINE", 544) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if (!strcmp(app, "spots")) /* need FRAMES and USPLINE */
	  {
		if (readytest("FRAMES", 6) && readytest("USPLINE", 544))
			return ((readytest("CENTROIDS", 100) &&
			 readytest("WMASK", 19908)) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if ((!strcmp(app, "cenmerge")) || (!strcmp(app, "cenmer")))
	  {
		if (readytest("CENOUT", 100)) return NEEDNOT;
		else return CAN;
	  }
	else if (!strcmp(app, "editbord"))
		return (readytest("AMASK", 2048) ? CAN : CANNOT);
	else if ((!strcmp(app, "editcen")) || (!strcmp(app, "filterc")))
		return (readytest("CENTROIDS", 100) ? CAN : CANNOT);
	/* refinement: need active-area mask, centroids, spline */
	else if ((!strcmp(app, "refine")) || (!strcmp(app, "refine1")) ||
	 (!strcmp(app, "refall")))
	  {
		if (readytest("AMASK", 2048) && readytest("CENTROIDS", 300)
		 && readytest("USPLINE", 544))
			return (readytest("UPARAMS", 232) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if ((!strcmp(app, "newrun")) || (!strcmp(app, "newr")))
		return (readytest("UPARAMS", 232) ? CAN : CANNOT);
	else if (!strcmp(app, "pdisplay")) /* for now we need FRAMES */
	  {
		if (readytest("FRAMES", 6)) return CAN;
		return CANNOT;
	  }
	else if ((!strcmp(app, "ewmask")) || !(strcmp(app, "mermask")))
		return (readytest("WMASK", 19908) ? CAN : CANNOT);
	else if (!strcmp(app, "integrate"))
	  {
		if (readytest("UPARAMS", 232) && readytest("AMASK", 2048) &&
		 readytest("USPLINE", 544))
			return (readytest("UREFLS", 80) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if (!strcmp(app, "strategy")) /* yellow only if urefls empty */
	  {
		if (readytest("UPARAMS", 232) && readytest("AMASK", 2048) &&
		 readytest("USPLINE", 544))
		  {
			if (readytest("UREFLS", 80)) return NEEDNOT;
			else return SHOULD;
		  }
		else return CANNOT;
	  }
	else if (!strcmp(app, "utos"))
	  {
		return (readytest("UREFLS", 80)) ? CAN : CANNOT;
	  }
	else if ((!strcmp(app, "filteru")) || (!strcmp(app, "cuturef")))
		return (readytest("UREFLS", 232) ? CAN : CANNOT);
	else if (!strcmp(app, "reduce"))
	  {
		if (readytest("UPARAMS", 232) && readytest("AMASK", 2048) &&
		 readytest("USPLINE", 544) && readytest("UREFLS", 80))
			return (readytest("MULTIREF", 544) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if ((!strcmp(app, "pickuref")) || (!strcmp(app, "anal_u")))
		return (readytest("UREFLS", 80) ? CAN : CANNOT);
	else if ((!strcmp(app, "mrmerge")) || (!strcmp(app, "cutoff")))
		return (readytest("MULTIREF", 544) ? CAN : CANNOT);
	else if ((!strcmp(app, "scalek")) || (!strcmp(app, "scalem")))
	  {
		if (readytest("MULTIREF", 544))
		  {
			return (readytest("SCALEIN", 54) ? NEEDNOT : SHOULD);
		  }
		else return CANNOT;
	  }
	/* note that the logic for these three apps is identical to the
	 logic for makemu. But it may not remain that way in the future. */
	else if ((!strcmp(app, "reject")) || (!strcmp(app, "spaceg")) ||
	 (!strcmp(app, "stats")))
	  {
		if (readytest("MULTIREF", 544) && readytest("SCALEIN", 54))
			return (readytest("MULIST", 60) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if (!strcmp(app, "makemu"))
	  {
		if (readytest("MULTIREF", 544) && readytest("SCALEIN", 54))
			return (readytest("MULIST", 60) ? NEEDNOT : SHOULD);
		else return CANNOT;
	  }
	else if (!strcmp(app, "mukb"))
	  {
		return (readytest("MULIST", 544) ? CAN : CANNOT);
	  }
	return 2;
}

XGNAME *sharename(int whichop)
{ /*	This creates or destroys memory for names that can be shared
	among routines. It returns a pointer to the shared memory. */
	static XGNAME	*sgsh;	/* shared space */
	static char	sgshmsg[] = "sgsh in sharename";

	if (0 == whichop)
	  {
		memzap((AVOID **)&sgsh, sgshmsg);	return NULL;
	  }
	if (1 == whichop) /* unconditionally kill the old copy */
	  {
		if (NULL != sgsh) memzap((AVOID **)&sgsh, sgshmsg);
	  }
	if (3 == whichop) return sgsh;	/* unconditional return */
	if (NULL == sgsh)
	  sgsh = (XGNAME *)callop(1, sizeof (XGNAME), sgshmsg);
	if (0x8 & whichop) /* zero out memory selectively */
	  {
		if (0x10 & whichop) zeromem((AVOID *)sgsh->xmm_misc, 512);
		if (0x20 & whichop) zeromem((AVOID *)sgsh->xmm_idna, 128);
		if (0x40 & whichop) zeromem((AVOID *)sgsh->xmm_workna, 128);
		if (0x80 & whichop) zeromem((AVOID *)sgsh->xmm_datana, 128);
		if (0x100 & whichop) zeromem((AVOID *)sgsh->xmm_scrtna, 128);
		if (0x200 & whichop) zeromem((AVOID *)sgsh->xmm_listna, 128);
		if (0x400 & whichop) zeromem((AVOID *)sgsh->xmm_pname, 128);
		if (0x800 & whichop) zeromem((AVOID *)sgsh->xmm_ptitle, 256);
		if (0x1000 & whichop) zeromem((AVOID *)sgsh->xmm_miscb, 2560);
		if (0x2000 & whichop) zeromem((AVOID *)sgsh->xmm_miscc, 2048);
	  }
	return sgsh;
}

static int copy_old(char *new, char *old, int fromcu)
{ /* This copies the contents of the file `old' into the file `new'.
	If they're already exactly the same file, we don't.  The routine
	returns the number of bytes written to the output. */
	int			result;
	size_t			insize;
	static char		*cbuf, *le;
	static FILE		*fpli, *fplo;
	static struct stat	bcurr, bdest;
	static char		cbmsg[] = "cbuf in copy_old";

	le = lineend(stderr, fromcu);
	if (!strcmp(new, old))
	  {
		fprintf(stderr,
			"It's pointless to copy %s to itself%s", old, le);
		return -1;
	  }
	if (0 != xgaccess(old, 04, 1))
	  {
		fprintf(stderr, "Old file (%s) is not readable%s", old, le);
		return -2;
	  }
	if (0 != (result = stat(old, &bcurr)))
	  {
		fprintf(stderr, "Can't get attributes of old file %s%s",
			old, le); return -4;
	  }
	if (0 == (result = access(new, 0)))
	  {
		if (0 != (result = stat(new, &bdest)))
		  {
			fprintf(stderr,
			 "Can't get attributes of new file %s%s", old, le);
			return -6;
		  }
		if (bcurr.st_ino == bdest.st_ino)
		  {
			fprintf(stderr,
		 "Input %s and output %s are same file: no copy made%s",
				old, new, le);
			return -8;
		  }
	  }
	insize = bcurr.st_size;
	if (NULL != fpli)
	  {
		(void)fclose(fpli);	fpli = NULL;
	  }
	if (NULL == (fpli = xgopen(old, "r")))
	  {
		fprintf(stderr, "Can't open %s for input%s", old, le);
		return -12;
	  }
	if (NULL == (cbuf = (char *)mallop(insize + 4,
	 "copy-buffer in copy_old")))
	  {
		fprintf(stderr, "Can't reserve %d bytes of memory for copy%s",
			(int)(insize + 4), le); return -14;
	  }
	if (1 != (result = fread(cbuf, insize, 1, fpli)))
	  {
		fprintf(stderr,
 "Trying to read %d bytes from %s: fread returned %d%s",
			(int)insize, old, result, le);
		memzap((AVOID **)&cbuf, cbmsg);	return -16;
	  }
	(void)fclose(fpli);	fpli = NULL;
	if (NULL != fplo)
	  {
		(void)fclose(fplo);	fplo = NULL;
	  }
	if (NULL == (fplo = xgopen(new, "w")))
	  {
		fprintf(stderr, "Can't open %s as output%s", new, le);
		memzap((AVOID **)&cbuf, cbmsg);	return -18;
	  }
	if (1 != (result = fwrite(cbuf, insize, 1, fplo)))
	  {
		fprintf(stderr,
		 "fwrite returned %d while writing %d bytes to %s%s",
			result, (int)insize, new, le);
		memzap((AVOID **)&cbuf, cbmsg);	return -20;
	  }
	(void)fclose(fplo);	fplo = NULL;
	printf("Copied %d bytes from %s to %s%s", (int)insize, old, new, le);
	memzap((AVOID **)&cbuf, cbmsg);	return insize;
}

static char *nameinpath(char *str, int fromcu)
{ /*	This looks for a file called 'str' in some specified places.
	It does so by finding which directory "spots" lives in and then
	looking for 'str' in the directory where "xgen" is and then in
	the directory one level up and then two levels up from there.
	The routine returns the string containing the resulting pathname */
	size_t		elen, stlen;
	char		*execdir, *cp, *end1, *le;
	static char	*wfname;
	static char	wfmsg[] = "wfname in nameinpath";
	static size_t	newlen, oldlen;

	if (NULL == str) /* special case: kill memory for wfname */
	  {
		memzap((AVOID **)&wfname, wfmsg);	oldlen = 0;
		return NULL;
	  }
	le = lineend(stdout, fromcu);
	/* figure out what directory the X-GEN executables are in */
	if (NULL == (execdir = findpath("xgen")))
	  {
		if (NULL == (execdir = findpath("xgen.exe")))
		  {
			fprintf(stderr,
			 "Cannot find path for X-GEN executables%s", le);
			return NULL;
		  }
	  }
	/* measure how long the directory string is */
	if (1 > (elen = strlen(execdir)))
	  {
		fprintf(stderr,
			"NULL Directory name for X-GEN executables%s", le);
		return NULL;
	  }
	/* create memory for relevant directories if necessary */
	stlen = strlen(str);	newlen = elen + stlen + 2;
	if (newlen < 80) newlen = 80;
	if ((NULL == wfname) || (newlen > oldlen))
	  {
		memzap((AVOID **)&wfname, wfmsg);	oldlen = newlen;
		if (NULL == (wfname = (char *)mallop(newlen, wfmsg)))
		  {
			fprintf(stderr, "allocation error in nameinpath%s", le);
			return NULL;
		  }
	  }
	zeromem((AVOID *)wfname, newlen);
	strcpy(wfname, execdir);
	strcat(wfname, "/");	strcat(wfname, str);
	if (0 == access(wfname, 04)) return wfname;
	cp = wfname + elen;	*cp-- = '\0';
	for ( ; ('\0' != *cp) && ('/' != *cp); cp--) ;
	if ('\0' == *cp) return NULL; /* rewound all the way and still failed*/
	end1 = cp;	*++cp = '\0';	strcat(wfname, str);
	if (0 == access(wfname, 04)) return wfname;
	cp = end1 - 1;
	if (*cp == '\0') return NULL; /* backed up all the way on 2nd try */
	for ( ; ('\0' != *cp) && ('/' != *cp); cp--) ;
	if ('\0' == *cp) return NULL; /* backed up all the way on 3rd try (!)*/
	*++cp = '\0';	strcat(wfname, str);
	return (0 == access(wfname, 04)) ? wfname : NULL;
}

int startspline(int fromcur)
{ /*	This first makes sure that the environment variable
	USPLINE points at a filename. If it doesn't, it returns.
	It then looks to see if the file with environment variable name
	USPLINE already exists. If it does, the routine returns.
	If not, we attempt to figure out which detector type we have.
	If we can't figure it out, or if the type is one for which we must
	have an explicit spline file, the routine returns.
	If we do figure it out and the detector type is one for which a
	starting USPLINE file can be envisioned, we attempt to find the file
	in a resources directory and copy it over. We then load its contents
	and return. */
	char			*usname, *srcend, srcname[128];
	int			dtype, nout, ssize, stype, detdim[2];
	static char		*le, *resdir;

	if (fromcur == -1) /* flag: get rid of memory for "nameinpath" */
	  {
		(void)nameinpath(NULL, fromcur);	return EOK;
	  }
	le = lineend(stderr, fromcur);
	if ((NULL == (usname = trname("USPLINE"))) || ('\0' == *usname))
	  {
		fprintf(stderr, "environment variable USPLINE undefined%s", le);
		return -1; /* environment variable must exist */
	  }
	if (0 == xgaccess(usname, 04, 1)) return 0;
		/* figure out what kind of detector we have */
	dtype = dettype(detdim);
	stype = dtype & 0377;	ssize = detdimen(dtype >> 8);
	if ((0 >= stype) || (stype > 18)) return -2;
	if ((stype < 5) || ((7 <= stype) && (stype <= 9)) ||
	 ((17 <= stype) && (stype <= 18))) return 0;
	/* construct name of template file */
	zeromem((AVOID *)srcname, 128);
	if (NULL == (resdir = nameinpath("res", fromcur)))
	  { /* if absent from ~/blahblah/xgen/res, look in /usr/share/xgen */
		resdir = ABSOLUTE;
		if (0 != access(resdir, 4))
		  {
			fprintf(stderr, "Cannot locate resources directory%s",
				le); return -3;
		  }
	  }
	/* give up if this directory doesn't exist or isn't readable */
	if (strcmp(resdir, ABSOLUTE))
	  {
		if (0 != access(resdir, 4))
		  {
			fprintf(stderr, "Resource directory unreadable%s", le);
			return -4;
		  }
		/* append /XGEN to this directory if such a thing exists.
		 If that doesn't work, revert to version without the /XGEN.*/
		strcpy(srcname, resdir);	strcat(srcname, "/XGEN");
		if (0 != access(srcname, 4)) strcpy(srcname, resdir);
	  }
	else strcpy(srcname, resdir);
	/* now put the appropriate suffix on */
	switch (stype)
	      {
		case	5:	/* R-Axis */
			strcat(srcname, "/raxis");
			if (ssize != 1900) strcat(srcname, "4"); break;
		case	6:	/* Mar, 18, 24, and 30 cm */
			srcend = srcname + strlen(srcname);
			sprintf(srcend, "/mar345_%4d", ssize);
			break;
		case	10:	/* Bruker ccd */
			if (ssize == 1024) strcat(srcname, "/bruker1k");
			else	strcat(srcname, "/bruker2k");
			break;
		case	11:	/* Mar CCD */
			if (ssize == 4096) strcat(srcname, "/marccd300");
			else if (ssize == 3072) strcat(srcname, "/marccd225");
			else strcat(srcname, "/marccd");	break;
		case	13:	/* ADSC Quantum 4 */
			/* we'll need to fix this for the Q4 eventually */
			strcat(srcname, "/adscunbinned");	break;
		case	14:	/* ADSC Quantum 210 */
			if (ssize == 4096) strcat(srcname, "/q2104k");
			else strcat(srcname, "/q2102k");	break;
		case	15:	/* ADSC Quantum 315 */
			if (ssize == 6144) strcat(srcname, "/q3156k");
			else strcat(srcname, "/q3153k");	break;
		case	16:	/* SBC CCDs (APS1, its relatives) */
			if (ssize == 1536) strcat(srcname, "/aps1b");
			else strcat(srcname, "/aps1u");	break;
	     }
	strcat(srcname, ".uca");
	if (0 != xgaccess(srcname, 04, 1)) /* see if it exists & is readable*/
	  {
		fprintf(stderr, "spline file %s unavailable%s", srcname, le);
		return -5;
	  }
	if (100 > (nout = copy_old(usname, srcname, fromcur))) /* copy it */
	  {
		fprintf(stderr, "failure in copying %s to %s: %d bytes%s",
		 srcname, usname, nout, le);	return -6;
	  }
	return 1;	/* once we're ready */
}

int basic_names(int fromcu)
{ /*	This returns a count of the number of basic logical names that have
	been successfully defined. `fromcu' should be 0 if we seek output
	on stdout without a Curses-style overlay; 1 if we do want the
	Curses overlay; -1 if we don't want printout at all. */
	int		slr;
	char		*result, *le;
	XGNAME		*losh;	

	if (NULL == (losh = sharename(7))) return 0;
	result = getenv("RUN_ID");
	if ('\0' != *(losh->xmm_idna))
	  {
		le = lineend(stdout, fromcu);
		if ((NULL == result) || ('\0' == *result) ||
		 strcmp(result, losh->xmm_idna)) /* set it only if necessary */
		  {
			if (1 > (slr = set_logical("RUN_ID", losh->xmm_idna)))
				return 0;
			else if (2 == slr)
			 printf("RUN_ID was already %s; unchanged%s",
				losh->xmm_idna, le);
			else if (fromcu >= 0)
			 printf("RUN_ID set to %s%s", losh->xmm_idna, le);
		  }
	  }
	else {
		if (NULL != result && '\0' != *result)
			strcpy(losh->xmm_idna, result);
		else return 0;
	     }
	if (NULL == (result = getenv("WORK")) || '\0' == *result) return 1;
	if (NULL == (result = getenv("DATA")) || '\0' == *result) return 2;
	if (NULL == (result = getenv("SCRT")) || '\0' == *result) return 3;
	return 4;
}

int mydircreate(char *name)
{ /* This attempts to create the directory `name'.  If
	it succeeds, it returns 0.  If it doesn't, it looks to see
	if the parent of this directory exists. If it does, then it returns 0.
	If not, it tries to create the parent; if THAT operation succeeds,
	it again tries to create `name'.  Any error results in a return of 0
	unless it's caused by an error in the directory-creation step,
	in which case the return will be the value of errno. */
	char	loname[140], *ecp, *cp;

	errno = 0;	if (0 == create_direc(name)) return 0;	/* success */
#ifndef VMS	/* I haven't figured out how to mimic this under VMS... */
	/* if the problem is something other than a missing
		 path element, we give up */
	if (ENOENT != errno) return errno;
#endif
	/* construct the name missing its last directory component */
	ecp = name + strlen(name);
	for (cp = ecp - 1; cp >= name; cp--) if ('/' == *cp) break;
	if ((cp <= name) || (cp == ecp - 1)) return -1;
	strncpy(loname, name, (size_t)(cp - name)); /* construct 1 level up */
	loname[cp-name] = '\0';
	/* if the parent already exists, then it won't do any good
	to try to create it, will it? */
	if (0 == dir_access(loname, 4)) return -2;
	/* try to create the parent.  If we fail there, we give up.
	Note that this is implicitly recursive */
	errno = 0;	if (0 != mydircreate(loname)) return errno;
	/* ... if we succeeded on the parent, we try again on the child */
	errno = 0;	if (0 == create_direc(name)) return 0;	return errno;
}

int check_directories(int ntocheck, int maketoo)
{ /*	This checks for the existence of various directories and, if maketoo is 
	true, creates them if necessary.  Success is noted by a return of the
	same number of directories as were being checked in the first place.*/
	int		dirno;
	char		*result;
	XGNAME		*losh;
	static char	*dirn[] = { "WORK", "DATA", "SCRT" };
	 /* WORK and SCRT must be writable; DATA only needs to be readable */
	static int	accreq[] = { 6, 4, 6 };

	if (ntocheck > 3) ntocheck = 3;	if (ntocheck < 1) ntocheck = 1;
	if (NULL == (losh = sharename(7))) return 0;
	for (dirno = 0; dirno < ntocheck; dirno++)
	   { /* see if the environment variable for the directory exists */
		if ((NULL != (result = getenv(dirn[dirno]))) &&'\0' !=*result)
		  {
			errno = 0;
			switch (dirno)
			      {
				case	0: strcpy(losh->xmm_workna, result);
					break;
				case	1: strcpy(losh->xmm_datana, result);
					break;
				case	2: strcpy(losh->xmm_scrtna, result);
					break;
			      }
			if (0 != dir_access(result, accreq[dirno]))
			  {
				if (maketoo)
				  { /* if it's absent, try to create it */
#ifndef VMS		/* I haven't figured out how to do this under VMS */
					if (ENOENT != errno) return dirno;
					else
#endif
					if (0 != mydircreate(result))
						return dirno;
				  }
				else return dirno; /* failure: return index */
			  }
			/* the implied `else' here goes on to next dirno */
		  }
		else return dirno; /* if env var is absent, return its index */
	   }
	return dirno + 1; /* we should get a positive result here */
}

static void glistname(void)
{ /*	This puts either SYS$LOGIN:XGENSAVE.LST (under VMS) or
	$HOME/xgensave.lst (under Unix) into the variable losh->xmm_listna,
	assuming this is defined.
	Otherwise it puts it either in XGENSAVE.LST and ./xgensave.lst. */
	char		*cp;
	XGNAME		*losh;

	if (NULL == (losh = sharename(7))) return;
#ifdef VMS
	if ((NULL == (cp = getenv("SYS$LOGIN"))) || ('\0' == *cp))
		strcpy(losh->xmm_listna, "XGENSAVE.LST");
	else strcpy(losh->xmm_listna, "SYS$LOGIN:XGENSAVE.LST");
#else
 	if ((NULL == (cp = getenv("HOME"))) || ('\0' == *cp))
		strcpy(losh->xmm_listna, "xgensave.lst");
	else strcpy(losh->xmm_listna, cp);
	strcat(losh->xmm_listna, "/xgensave.lst");
#endif
}

static void nametolist(char *destname, int fromcu)
{ /* This writes the command-file name `destname' into the file
	$HOME/xgensave.lst (on Unix) or SYS$LOGIN:XGENSAVE.LST (VMS). */
	char		*le;
	XGNAME		*losh;
	static FILE	*fpsav;

	if (NULL == (losh = sharename(7))) return;
	le = lineend(stdout, fromcu);
	if ('\0' == *losh->xmm_listna) glistname();
	if ('\0' == *losh->xmm_listna) return;
	if (NULL == (fpsav = fopen(losh->xmm_listna, "w")))
	 fprintf(stderr,"Warning: %s not updated%s", losh->xmm_listna, le);
	else {
		fprintf(fpsav, "%s\n", destname);
		(void)xeflush(fpsav);	(void)fclose(fpsav);	fpsav = NULL;
		printf("%s now contains %s%s", losh->xmm_listna, destname, le);
	     }
}

int more_logicals(int newonly)
{ /*	This defines environment variables.  If `newonly' is nonzero, it will
	only create an environment variable if it is currently unassigned;
	otherwise it leaves it alone. If `newonly' is zero, it assigns the new
	environment variables even if there's something already there. It
	returns the number of new environment variables it creates. If the
	prefix for any variable is the template prefix "newxtal", we
	override the "newonly" check and reassign the name automatically.
	Aug 95: we get rid of the frame index if we've reassigned the
	logical name `frames'. */
	int	i, j, nlogs;
	char	*dirtok, *lnam, *oldv, *sufx, *tnb;
	XGNAME	*losh;

	nlogs = check_directories(3, 0);
	if (NULL == (losh = sharename(7))) return 0;
	tnb = losh->xmm_misc;
	/* set up environment variables for appropriate directories */
	for (j = 0; j < NENVVARS; j++)
	   {
		lnam = indtolog(j, &sufx);	oldv = getenv(lnam);
#ifdef VMS
		if (j < NWORKVAR) dirtok = "WORK";
		else dirtok = "SCRT";
#else
		if (j < NWORKVAR) dirtok = losh->xmm_workna;
		else dirtok = losh->xmm_scrtna;
#endif
		if ((!newonly) || (NULL == oldv) || ('\0' == *oldv) ||
		 strstr(oldv, "newxtal"))
		  {
			sprintf(tnb, "%s%c%s.%s", dirtok, DIVIDER,
			 losh->xmm_idna, sufx);
			nlogs += (i = (set_logical(lnam, tnb) >= 1));
		  }
	   }
	return nlogs;
}

int writelognames(int fromcu)
{ /* this writes out the ".cmd" or ".com" file. As of 28 Sep 94
	we use the current environment variable settings rather than
	overriding them.  It also sets the environment variable XGENCMD.
	This constructs the environment-variable strings via a call to
	`more_logicals' and then writes the assignments out. Any variable
	not successfully defined doesn't get written out at all. */
	int	j, nlogs;
	char	*docname, *le, *lnam, *lfname, *lshname, *ona, *sufx;
	XGNAME	*losh;
	static char	dq = '"';
	static FILE	*fped, *fpsh;
#ifdef VMS
	static char	setword[] = "$ define";
#else
	static char	setword[] = "setenv";
#endif

	le = lineend(stderr, fromcu);
	if (NULL == (losh = sharename(7)))
	  {
		fprintf(stderr, "Difficulty getting name-buffer memory\n");
		return 0;
	  }
	if (4 > (nlogs = basic_names(fromcu)))
	  {
		fprintf(stderr,
		 "Only %d basic names found;can't writelognames%s", nlogs, le);
		return nlogs;
	  }
	if (1 > check_directories(1,1)) return nlogs;
	lfname = losh->xmm_miscb;	lshname = losh->xmm_miscc;
	*lfname = *lshname = '\0';
	strcpy(losh->xmm_workna, getenv("WORK"));
	strcpy(losh->xmm_datana, getenv("DATA"));
	strcpy(losh->xmm_scrtna, getenv("SCRT"));
	(void)indtolog(20, &sufx);
	sprintf(lfname, FILEFORMAT1, losh->xmm_workna, losh->xmm_idna, sufx);
	(void)set_logical("XGENCMD", lfname);
	/* there's not much point in writing these out unless we get
	 most of the logicals written out correctly. */
	if (4 >= (nlogs = more_logicals(0))) return nlogs;
	if (NULL == (fped = fopen(lfname, "w")))
	  {
		fprintf(stderr, "Can't open %s as output XGENCMD file%s",
			lfname, le);
		return nlogs;
	  }
	(void)indtolog(21, &sufx);
	sprintf(lshname, FILEFORMAT1, losh->xmm_workna, losh->xmm_idna, sufx);
	(void)set_logical("XGENCOM", lshname);
	if (NULL == (fpsh = fopen(lshname, "w")))
	  {
		fprintf(stderr, "Can't open %s as output XGENCOM file%s",
			lshname, le);
	  }
		/* directory names */
	fprintf(fped, "%s WORK %s\n", setword, losh->xmm_workna);
	fprintf(fped, "%s DATA %s\n", setword, losh->xmm_datana);
	fprintf(fped, "%s SCRT %s\n", setword, losh->xmm_scrtna);
	fprintf(fped, "%s PARM %s\n", setword, losh->xmm_workna);
	fprintf(fpsh, "export WORK=%s\n", losh->xmm_workna);
	fprintf(fpsh, "export DATA=%s\n", losh->xmm_datana);
	fprintf(fpsh, "export SCRT=%s\n", losh->xmm_scrtna);
	fprintf(fpsh, "export PARM=%s\n", losh->xmm_workna);
	if (NULL != (docname = getenv("XGDOC")))
	  {
		fprintf(fped, "%s XGDOC %s\n", setword, docname);
		fprintf(fpsh, "export XGDOC=%s\n", docname);
	  }
	/* set up lines for files defined in WORK directory */
	for (j = 0; j < NENVVARS; j++)
	   {
		lnam = indtolog(j, &sufx);	ona = getenv(lnam);
		if ((NULL != ona) && ('\0' != *ona))
		  {
			fprintf(fped, "%s %s %s\n", setword, lnam, ona);
			fprintf(fpsh, "export %s=%s\n", lnam, ona);
		  }
	   }
	glistname();
	if ('\0' != *losh->xmm_listna)
#ifdef VMS
	   fprintf(fped,"$ xgensave 'f$environment(%cPROCEDURE%c)\n", dq, dq);
#else
	   fprintf(fped, "echo %c%s%c > %s\n",
		dq, lfname, dq, losh->xmm_listna);
#endif
	(void)fclose(fped);	/* close the output file */
	fprintf(fpsh, "echo %c%s%c > $HOME/xgensave.sh\n", dq, lshname, dq);
	(void)fclose(fpsh);	/* close the (ba)sh output file too */
	(void)startspline(fromcu); /* attempt to get starting USPLINE */
	/* we open SYS$LOGIN:XGENSAVE.LST or $HOME/xgensave.lst
	 and write the name of the current .COM or .cmd file into it */
	nametolist(lfname, fromcu);
	printf("File %s now contains logical assignments for crystal ID %s%s",
		lfname, losh->xmm_idna, le);
	return nlogs;
}

int execlognames(int fromcu)
{ /*	Assign the actual environment variables */
	int		nlogs, nval, nfirst;
	char		*le;

	nlogs = 0;	le = lineend(stderr, fromcu);
		/* are names in place?*/
	if (4 > (nlogs = nfirst = basic_names(fromcu)))
	  {
		fprintf(stderr,
	"Only %d basic names found; can't assign environment variables%s",
			nfirst, le); return nlogs;
	  }
	if (3 > (nval = check_directories(3,1)))
	  {
		fprintf(stderr,
 "check_directories returned %d; can't assign environment variables%s",
			nval, le); return nval; /*be sure directories exist */
	  }
	nfirst = nlogs;	/* save the number we've already got */
	(void)more_logicals(0);	/* get the other environment variables */
	nlogs += nfirst;	/* keep track of both lists */
	(void)startspline(fromcu); /* get initial pixel-to-cm mapping */
	printf("%d successful logical name assignments%s", nlogs, le);
	return nlogs;
}

static int isdelim(char c)
{ /* returns 1 iff. c is one of the characters on the list
	of delimiters or c is '\0'. */
	register char	*cp;
	static char	delim[] = "/[].:";

	if ('\0' == c) return 1;
	for (cp = &(delim[0]); '\0' != *cp; cp++) if (*cp == c) return 1;
	return 0;
}

static char *substr(char *s1, char *s2)
{ /* This locates the first occurrence in the string `s1' of the
	sequence of characters (excluding the terminating '\0') in
	the string `s2', returning a pointer to it (or NULL if s2 does
	not occur in s1).  If s2 points to a string with zero length,
	strstr returns s1.  The difference between this routine and
	`strstr' on a typical C library is that it requires that the
	substring within s1 either be at the end of s1 or that it be
	followed by a delimiter character (/, [, ], '.', or ':'); thus
	substr("USERDISK:[HOWARD.JUNKSTUF]", "JUNKSTUF") is "JUNKSTUF]", but
	substr("USERDISK:[HOWARD.JUNKSTUF]", "JUNK") is NULL. Further, it
	requires that s2 begin either at the beginning of s1 or right
	after one of these delimiters; thus
	substr("USERDISK:[HOWARD.JUNKSTUF]", "WARD") is NULL, and
	substr("USERDISK:[HOWARD.JUNKSTUF.STUF]", "STUF") is "STUF]". */
	int	len1, len2;
	char	*cp1, *cp2, *cp0, *cpe2, *cpe1;

	if ((NULL == s2) || ('\0' == *s2)) return s1;	/* degenerate case */
	/* by definition s2 must not be longer than s1 */
	for (cpe1 = s1; '\0' != *cpe1; ) cpe1++;	len1 = cpe1 - s1;
	for (cpe2 = s2; '\0' != *cpe2; ) cpe2++;	len2 = cpe2 - s2;
	if (len2 > len1) return NULL;
	for (cp1 = s1; cp1 < cpe1; cp1++, len1--)
	   {
		if (((cp1 == s1) || isdelim(*(cp1-1))) && (*cp1 == *s2))
		  {
			for (cp0 = cp1, cp2 = s2; cp2 < cpe2 && cp0 < cpe1;
			  cp0++, cp2++) if (*cp0 != *cp2) break;
			if ((*cp2 == '\0') && isdelim(*cp0)) return cp1;
		  }
		if (len2 > len1) return NULL;
	   }
	return NULL;
}

static int countimages(char *dname, int quitafter1)
{ /*	This goes into the directory "dname" and attempts to count all
	the legitimate detector images in it. It returns the number it finds.
	If "quitafter1" is nonzero, it returns as soon as it finds one.  */
	char			lfname[258], *icp, *ocp;
	int			chkflags, nexami, nimages, j;
	static MAIL		locm;
	static DIR		*dirci;
	static struct dirent	*mdirc;

	/* directory must be openable */
	chkflags = JUST_HEADER;
	if (NULL == (dirci = opendir(dname))) return 0;
	zeromem((AVOID *)&locm, sizeof (MAIL));
	for (nexami = nimages = 0; NULL != (mdirc = readdir(dirci)); )
	   {
		nexami++;
		/* 28 Dec 2002: don't try to read_image unless the file
		 comes back positively from imagename() and imagelike() */
		if ((0 != mdirc->d_ino) && (1 == imagename(mdirc->d_name)) &&
		 (1 == imagelike(dname, mdirc->d_name)))
		  {
			for (icp = dname, ocp = lfname; '\0' != *icp; )
				*ocp++ = *icp++;
			*ocp++ = '/';
			for (j = 0, icp = mdirc->d_name; '\0' != *icp &&
			 j < mdirc->d_reclen; j++) *ocp++ = *icp++;
			*ocp = '\0'; 
			if (FRAME_OK == read_image(lfname, chkflags, &locm))
			  {
				nimages++;
				if (quitafter1) break;
			  }
		  }
	   }
	(void)closedir(dirci);
	return nimages;
}

static int filetoenv(char *finame, int lineno)
{ /*	This reads the `lineno'th line of the file `finame' as
	containing the name of a file from which logical name assignments
	can be obtained.  It then performs those logical name assignments
	and assigns the crystal id name based on the filename we discover.
	We count lines from 1, not 0. The routine returns the number of
	assignments it performs. */

	int	thisline, nassignments, nitems;
	size_t	namewidth;
	char	assbuf[120], ename[40], evalue[120];
	char	*cp, *lpunct, *abp, *xname;
	FILE	*fpsrc, *fpcmd;
	static int	erval = 0;

		/* try to open ~/xensave.lst or its equivalent */
	if (NULL == (fpsrc = fopen(finame, "r")))
	  {
		if (!(erval & 1)) fprintf(stderr,
		 "Cannot open %s to look up current datarun command file\n",
			finame);
		erval |= 1;	return 0;
	  }
	erval &= ~01;
		/* read up through the appropriate line to get the filename */
	if (lineno < 1) lineno = 1;
	for (thisline = 0; thisline < lineno; thisline++)
	   if (NULL == fgets(assbuf, 116, fpsrc))
	     {
		(void)fclose(fpsrc);
		if (!(erval & 2)) fprintf(stderr,
		 "%s has fewer than %d lines of datarun names\n",
			finame, lineno);
		erval |= 2;	return 0;
	     }
	erval &= ~2;
	(void)fclose(fpsrc);
		/* try to open the file named on that line */
	abp = &(assbuf[0]);	namewidth = strlen(abp);
	if (*(abp + (namewidth-1)) == '\n')
	  {
		namewidth--;	*(abp + namewidth) = '\0';
	  }
	if (NULL == (fpcmd = fopen(assbuf, "r")))
	  {
		if (!(erval & 4)) fprintf(stderr,
			"Can't open datarun file %s\n", assbuf);
		erval |= 4;	return 0;
	  }
	erval &= ~4;
	/* extract from that line the crystal ID.  For generality
	 it is taken to be all the stuff between the last and
	 next-to-last punctuation marks, so if the named file
	 is /mnt/xengen/myor1/2/myor1_2.cmd, it'll be myor1_2.  If
	 there are no punctuation marks, we use the whole name */
	for (cp = abp + (namewidth - 1), lpunct = abp; cp >= abp; cp--)
		if (*cp == '/' || *cp == '.')
		  {
			lpunct = cp; break;
		  }
	if (cp > abp)
	  {
		for (cp--; cp >= abp; cp--) if (*cp == '/' ||*cp == ']') break;
		if ('/' == *cp || ']' == *cp) cp++; namewidth = lpunct - cp;
	  }
	else if ('/' == *cp)
	  {
		cp++;	namewidth--;
	  }
	if (namewidth < 1) return 0;	/* we should have a name by now */
	if (NULL == (xname = (char *)mallop((size_t)(namewidth + 2),
		"xname in filetoenv"))) return 0;
	strncpy(xname, cp, namewidth);
	*(xname + namewidth) = '\0';
		/* now make that assignment */
	nassignments = (set_logical("RUN_ID", xname) >= 1);
		/* now read through the file and parse its assignments */
	printf("Reading assignments for run_id = %s from %s\n", xname, assbuf);
	for ( ; NULL != fgets(assbuf, 116, fpcmd); )
	   {
#ifdef VMS
		nitems = sscanf(assbuf, "$ define %s %s", ename, evalue);
#else
		nitems = sscanf(assbuf, "setenv %s %s", ename, evalue);
#endif
		if ((2 <= nitems) && (NULL == getenv(ename)))
		  {
#ifdef VMS
			strcpy(assbuf, evalue);
#else
			(void)envtoname(evalue, assbuf); /* new name! */
#endif
			nassignments += (set_logical(ename, assbuf) >= 1);
		  }
	   }
	(void)fclose(fpcmd);	/* we're done with this file */
	printf("%d assignment%sobtained\n",
		nassignments, (nassignments == 1 ? " " : "s "));
	return nassignments;	/* return the number of assignments */
}

int newidname(char *idnam, int idnflags)
{ /*	This takes appropriate action based on a new name for the run ID. It
	returns 1 if successful and various return codes from set_logical()
	if not. The default WORK name is built from the current directory.
	`idnaflgs' is an or-ing of at least these bits:
	1: on means use CURSES output conventions
	2: on means assume the data are in the $WORK directory (often ".")
	3: on means assume the images are in ".." */
	int		idcode, neednew, slr, worklen;
	char		*cp, *ddir, *le, *subst, *sufx, *tdir;
	XGNAME		*losh;

	le = lineend(stdout, idnflags);
	if (NULL == (losh = sharename(7))) return -1;
		/* quit immediately if the new name is the same as the old */
	if (('\0' != *losh->xmm_idna) && (!strcmp(losh->xmm_idna, idnam)))
		return 1;
		/* try to set the logical name RUN_ID to the argument */
	if (1 > (idcode = set_logical("RUN_ID", idnam))) return idcode;
		/* store this id name in the global `losh->xmm_idna' */
	strcpy(losh->xmm_idna, idnam);
	/* also: redefine WORK, etc. based on this name unless they already
	 end with that suffix or are a subdirectory of a valid DATA directory*/
	neednew = 1;
	if (NULL != (tdir = getenv("WORK")) && ('\0' != *tdir))
	  {
		if ((NULL != (subst = substr(tdir, losh->xmm_idna))) &&
		 (strlen(subst) <= strlen(losh->xmm_idna) + 1)) neednew = 0;
		else if (((NULL != (ddir = getenv("DATA"))) && ('\0' != *ddir))
		  && (NULL != (subst = substr(tdir, ddir)))) neednew = 0;
	  }
	if (neednew) /* here we define an initial WORK directory name */
	  {
		mygetwd(losh->xmm_misc); /* determine the current directory */
		/* if we have failed to find a current working directory,
		construct a name based on the user's HOME directory */
		if ('\0' == *(losh->xmm_misc))
		  {
#ifdef VMS
			if ((NULL != (cp = getenv("SYS$LOGIN"))) &&'\0' != *cp)
			  {
				strcpy(losh->xmm_workna, cp);
				losh->xmm_workna[strlen(cp) - 1] = '.';
			  }
			else strcpy(losh->xmm_workna, "SYS$SYSDEVICE:[USERS.");
#else
			if (NULL != (cp = getenv("HOME")) && '\0' != *cp)
			  {
				strcpy(losh->xmm_workna, cp);
				strcat(losh->xmm_workna, "/");
			  }
			else	strcpy(losh->xmm_workna, "/usr/people/");
#endif
		  }
		else	strcpy(losh->xmm_workna, losh->xmm_misc);
#ifdef VMS
		if (*(losh->xmm_workna + strlen(losh->xmm_workna) - 1) != ']')
			strcat(losh->xmm_workna, "]");
#endif
		if (1 == (slr = set_logical("WORK", losh->xmm_workna)))
		  {
			printf("Setting WORK to %s%s", losh->xmm_workna, le);
		  }
		else if (2 == slr) printf(
		 "WORK was already %s; unchanged%s", losh->xmm_workna, le);
		else fprintf(stderr, "Attempt to set WORK to %s failed%s",
				losh->xmm_workna, le);
	  }
	else	strcpy(losh->xmm_workna, tdir);
	/* now if WORK:<idname>.COM (or $WORK/<idname>.cmd) already
	exists, we use it as input--i.e., we write it to XGENSAVE.LST
	and we set the logicals from it. */
	(void)indtolog(20, &sufx);
	sprintf(losh->xmm_misc, FILEFORMAT1, losh->xmm_workna,
		losh->xmm_idna, sufx);
		/* look for <losh->xmm_idna>.cmd or .COM */
	if (0 == xgaccess(losh->xmm_misc, 4, 1))
	  {
		/* stick that filename in xgensave.lst */
		nametolist(losh->xmm_misc, idnflags & 01);
		(void)filetoenv(losh->xmm_listna, 1); /* read it & use it */
	  }
	worklen = strlen(losh->xmm_workna);
	neednew = 1;	/* flag: see if DATA needs redefining */
	/* there used to be a rule here that said, in effect,
	 that the name of the DATA directory had to resemble the ID name.
	 That particular feature no longer appears useful. */
	if (NULL != (tdir = getenv("DATA")) && ('\0' != *tdir)) neednew = 0;
	if (neednew)
	  {
		if (NULL != (tdir = getenv("PERM_DATA")) && '\0' != *tdir)
			strcpy(losh->xmm_datana, tdir);
		else if (idnflags & 02) /* data should be in $WORK */
			strcpy(losh->xmm_datana, losh->xmm_workna);
		else if (idnflags & 04) /* data should be in $WORK/.. */
		  {
			strcpy(losh->xmm_datana, losh->xmm_workna);
			for (cp = losh->xmm_datana + strlen(losh->xmm_datana)
			 - 1; (cp > losh->xmm_datana); cp--)
				if ('/' == *cp) break;
			if (cp > losh->xmm_datana) *cp = '\0';
		  }
		else { /* see if the current directory contains images */
			strcpy(losh->xmm_datana, losh->xmm_workna);
			if (1 <= countimages(losh->xmm_datana, 1)) neednew = 0;
			else {	/* try the directory just above this */
				for (cp = losh->xmm_datana +
				 strlen(losh->xmm_datana) - 1; (cp >
				 losh->xmm_datana); cp--) if ('/' == *cp) break;
				if (cp > losh->xmm_datana)
				  {
					*cp = '\0';
					if (1 <= countimages(losh->xmm_datana,
					 1)) neednew = 0;
				  }
			     }
			if (neednew)
			  {
				strcpy(losh->xmm_datana, losh->xmm_workna);
#ifdef VMS
				strcpy(losh->xmm_datana + (worklen - 1),
				 ".DATA]");
#else
				strcpy(losh->xmm_datana + worklen, "/data");
#endif
			  }
		  }
		if (1 == (slr = set_logical("DATA", losh->xmm_datana)))
			printf("Setting DATA to %s%s", losh->xmm_datana, le);
		else if (2 == slr) printf(
		 "DATA was already %s; unchanged%s", losh->xmm_datana, le);
		else fprintf(stderr, "Attempt to set DATA to %s failed%s",
			losh->xmm_datana, le);
	  }
	else strcpy(losh->xmm_datana, tdir);
	neednew = 1;	/* flag: see if SCRT needs redefining */
	if (NULL != (tdir = getenv("SCRT")) && ('\0' != *tdir))
	  {
		if (NULL != substr(tdir, losh->xmm_idna)) neednew = 0;
		else if (((NULL != (ddir = getenv("DATA"))) && ('\0' != *ddir))
		  && (NULL != (subst = substr(tdir, ddir)))) neednew = 0;
	  }
	if (neednew)
	  {
		if (NULL != (tdir = getenv("PERM_SCRT")) && '\0' != *tdir)
			strcpy(losh->xmm_scrtna, tdir);
		else	strcpy(losh->xmm_scrtna, losh->xmm_workna);
		if (1 == (slr = set_logical("SCRT", losh->xmm_scrtna)))
			printf("Setting SCRT to %s%s", losh->xmm_scrtna, le);
		else if (2 == slr) printf(
		 "SCRT was already %s; unchanged%s", losh->xmm_scrtna, le);
		else fprintf(stderr, "Attempt to set SCRT to %s failed%s",
			losh->xmm_scrtna, le);
	  }
	else strcpy(losh->xmm_scrtna, tdir);
	return 1;	/* indicates success */
}

static char *knowntype(char *cp)
{ /* this looks at the name `cp' and sees if it's on the list
	of known file types.  If it is, the routine returns a string
	describing that file type; if it isn't, the routine returns NULL.  */
	int		j;
	char		*lnam, *sufx;

	static char *lattrs[NENVVARS] = {
	"Active-pixel", "Background-Image", "Fiducial-Image",
	"Dark-Image", "Deletion-list", "Flood-Image",
	"Frame Index", "Missing-List", "Multiref", "Mulist",
	"Multiref", "Reference", "Scale-Factor", "Scale-Factor", "Parameters",
	"Calibration", "Integrated Refls", "Log", "Calib-Centroids",
	"Ice-Ring", "csh Environ", "sh Environ",
	"X-PLOR HKL", "CNS HKL", "SHELX hkl", "Merged SCA", "Unmerged SCA",
	"Centroids", "Centroids", "Checked-Refls", "Reindexing",
	"Dump", "Predictions", "Weighted-Mask", "Reflection-List" } ;

	if ((NULL == cp) || ('\0' == *cp)) return NULL;
	for (j = 0; j < NENVVARS; j++)
	   {
		lnam = indtolog(j, &sufx);
		if ((NULL != lnam) && ('\0' != *lnam) &&
		 !strncmp(lnam, cp, strlen(lnam))) return lattrs[j];
	   }
	return NULL;
}

char *get_xgdoc(int gxflags)
{ /* This looks to see if the environment variable XGDOC is already set.
   If it is, it returns its value; if it isn't, it tries to find a plausible
   set of documentation and sets XGDOC to the directory where it finds that.
   It will then return the value of that environment variable */
	int		announce, fromcur, slr;
	char		*cp, *cpe;
	static char	*le;

	/* the help files live in XGDOC.  If XGDOC is undefined, we look
	 for the document files in the path and define it appropriately */
	/* if XGDOC is already defined and is pointed at something
	 that is readable, we're okay */
	if ((NULL != (cp = getenv("XGDOC"))) && ('\0' != *cp) &&
		(0 == access(cp, 04))) return cp;
	fromcur = gxflags & 01;	announce = !(gxflags & 02);
	le = lineend(stdout, fromcur);
	if (((NULL != (cp = nameinpath("spots.txt", fromcur))) &&
	 ('\0' != *cp)) ||
	 ((NULL != (cp = nameinpath("xgdoc/spots.txt", fromcur))) &&
	 ('\0' != *cp)) ||
	 ((NULL != (cp = nameinpath("doc/xgen/spots.txt", fromcur))) &&
	 ('\0' != *cp)) ||
	 ((NULL != (cp = nameinpath("share/doc/xgen/spots.txt", fromcur))) &&
	 ('\0' != *cp)))
	  {
		for (cpe = cp; '\0' != *cpe; cpe++) ;
		cpe -= 10;
		if (cpe <= cp)
		  {
			fprintf(stdout, "Unable to define XGDOC%s", le);
			return NULL;
		  }
		*cpe-- = '\0'; if ('/' == *cpe) *cpe = '\0';
		if (1 <= (slr = set_logical("XGDOC", cp)))
		  {
			if (announce)
			  {
				if (1 == slr)
				 printf("Setting XGDOC to %s%s", cp, le);
				else printf("XGDOC was already %s; unchanged%s",
				 cp, le);
			  }
			return cp;
		  }
		else {
			if (announce) printf(
		  "Unable to set environment variable for XGDOC%s", le);
			return NULL;
		     }
	  }
	if (announce) printf("Unable to define XGDOC%s", le);
	return NULL;
}

static void disp_filetype(FILE *fp, char *name, char ftype)
{ /* this does a one-line summary of the contents of a file. */
	char		*kt, wlog[13], *xpname;
	int		j, nblanks;
	static char 	blanks[] = "                  ";

	if (NULL == name || '\0' == *name) return;
	if (!ftype) return; /* print nothing if file-type variable is zero */
	/* Oct 1994: if the file-type variable is nonzero we require
	that we be able to read the file (makes sense, no?) */
	xpname = trname(name);	/* expand environment variable */
	if (0 != access(xpname, 4)) return;
	/* copy, pad, and null-terminate the logical name itself */
	strncpy(wlog, name, 12);	wlog[12] = '\0';
	if ((nblanks = 12 - strlen(wlog)) > 0)
	  for (j = nblanks, kt = &(wlog[11]); j--; kt--) *kt = ' ';
	for (kt = &(wlog[0]); '\0' != *kt; kt++) /* capitalize name if needed*/
	   if ('a' <= *kt && *kt <= 'z') *kt &= 0137;
	/* see if the filename-type is one of the known types.
	If it is, keep a summary of its attributes; if not, keep it.*/
	if (NULL == (kt = knowntype(wlog))) kt = name;
	/* null-terminate a group of blanks to put into the line */
	blanks[(nblanks = MAX_SPACE - strlen(kt))] = '\0';
	/* now print out the file type or attributes and its name */
	fprintf(fp, " %s %s%s %s %s\n",
	 (ftype == 1 ? " Input" : (ftype == 2 ? "Output" :
	 "In/Out")), kt, blanks, wlog, xpname);
	blanks[nblanks] = ' '; /* now get rid of the '\0' we wrote earlier */
}

void header(FILE *fp, char progtype, char **iolist, char *ftypelist)
{ /* this produces a generalized program-run header for a XGEN
	system program and writes it to the stream `fp', which will usually
	be either stderr, stdout, or the crystallographic logfile `fplog'.
	It will contain the following:
		(1) if the file is not a tty and is non-empty, it
			starts by emitting a form-feed.
		(2) it lists the name of the program and the date of the run.
		(3) it lists the input files and output files to be used
		in the program run.
	The arguments to this function are as follows:
		fp		the stream to which data will be written.
		progtype	type of program (PROCESSING or REDUCTION)
		iolist		logical names of the input/output files.
		ftypelist	array of file types (0 (read), 1 (write),
					or other (=> read&write)
	If `iolist' is NULL, no file type announcements will be made.
*/
	char		*cpname, **cdp, lrdate[26], *ftp;
	time_t		ltl0;
	int		fd;

	ltl0 = NOW;	strcpy(lrdate, ctime(&ltl0));	fd = fileno(fp);
	cpname = xgprogname(NULL, 0);
	if ((!isatty(fd)) && 0 < ftell(fp)) fprintf(fp,"\f");
	fprintf(fp, "X-GEN V.%s Data %s: Run of %s on %s",
	 XGEN_VERSION,(progtype == PROCESSING ? "Processing" : "Reduction"),
	 cpname, lrdate);
	fprintf(fp,
"Software Copyright Illinois Institute of Technology,2002. All rights reserved\n");
	if (NULL != iolist)
	  {
		fprintf(fp,"\tFiles to be Used in the Run:\n");
		fprintf(fp,
		 "   File Type              Logical Name   File Name\n");
		for (cdp = iolist, ftp = ftypelist; NULL != cdp && NULL != *cdp;
		  cdp++, ftp++) disp_filetype(fp, *cdp, *ftp);
	  }
}

XGNAME *getproject(char *projnam, int fromcu)
{ /*	This returns a pointer to a properly constructed XGNAME structure.
	If the argument is NULL it derives the run-id from various already
	extant sources; if argument is non-NULL it derives it from the name */
	int		nlogs;
	static char	*cp, *oldid;
	static XGNAME	*locna;

	/* reserve memory for some working buffers */
	if (NULL == (locna = sharename(7)))
	  {
		fprintf(stderr, "Cannot find name buffers\n");
		return NULL;
	  }
	 /* Before looking at the arguments we check the logical name RUN_ID
	 and, if it's missing, try to glean it from FRAMES and USPLINE. */
	cp = getenv("RUN_ID");
	if ((NULL != cp) && ('\0' == *cp)) cp = NULL;
	if (NULL == cp)
	  {
		if ((NULL == (cp = getenv("FRAMES"))) || ('\0' == *cp))
			cp = getenv("USPLINE");
		if ((NULL != cp) && ('\0' != *cp)) oldid = filepart(cp, 2);
		if ((NULL != oldid) && ('\0' == *oldid)) oldid = NULL;
	  }
	else {
		oldid = (char *)mallop(strlen(cp) + 2, "oldid in getproject");
		strcpy(oldid, cp);
	     }
	/* if we still don't have a RUN_ID, use xgensave.lst to get it */
	if (NULL == oldid)
	  {
		glistname();
		if ('\0' != *(locna->xmm_listna))
		  {
			if (1 <= (nlogs = filetoenv(locna->xmm_listna, 1)))
				cp = getenv("RUN_ID");
			else	cp = "newxtal";
		  }
		else {
			nlogs = 0;	cp = "newxtal";
		     }
		oldid = (char *)mallop(strlen(cp) + 2, "oldid in pmmain");
		strcpy(oldid, cp);
	  }
	else
	  {
		(void)set_logical("RUN_ID", oldid);
		nlogs = basic_names(fromcu); /* RUN_ID,WORK,DATA,SCRT*/
	  }
	/* if we still don't have a DATA or SCRT name, we try PERM_DATA
	and PERM_SCRT */
	if (((NULL == (cp = getenv("DATA"))) || ('\0' == *cp)) &&
	  (NULL != (cp = getenv("PERM_DATA")) && ('\0' != *cp)))
		nlogs += (set_logical("DATA", cp) >= 1);
	if (((NULL == (cp = getenv("SCRT"))) || ('\0' == *cp)) &&
	  (NULL != (cp = getenv("PERM_SCRT")) && ('\0' != *cp)))
		nlogs += (set_logical("SCRT", cp) >= 1);
	/* now we override this with the argument if one has been supplied */
	if ((NULL != projnam) && ('\0' != *projnam))
	  {
		oldid = (char *)mallop(strlen(projnam) + 2,
			"oldid in getproject");	strcpy(oldid, projnam);
		/* use the arg to generate names for WORK, DATA, & SCRT.*/
		if (1 <= newidname(oldid, fromcu))
			nlogs = basic_names(fromcu);
	  }
	/* now add in whatever remaining logicals are still missing */
	if (nlogs == 4) nlogs = more_logicals(1);
	/* after all that pain and travail, we can destroy oldid */
	oldid = filepart(NULL, 0);	return locna;
}

/* end of xgenviron.c */
