/* refio.c -- input/output routines used in XGEN's refine

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: check legald in showpeaks; cosmetics
	21 Dec 2006: rpolymap(): polynomial remapping on 'H" command
	28 Nov 2006: mat2vecs replaced by rp2vecs;
	 calls to ucell, show_bravchoice, showbravais, artocel updated
	 9 Jun 2006: ref1log moved to refdriver.c
	30 May 2006: move defin2,ref3p,domainb,get_hkl in;
		move end_refinement,updateall,ref1cmd,keepinter_ref in;
		list_commands no longer static; blank-line inputs handled
	18 May 2006: merge construct.c into here; elim lrp arg to geth
	 5 Mar 2006: various routines: REFREF element is framem, not frame
	16 Oct 2005: eliminate static globals; housekeeping;
		incorporate gpeaks.c and refremap.c
	13 Oct 2005: housekeeping
	18 Aug 2005: increase MAX_FOM from 0.015 to 0.040
	12 Aug 2005: dowrite in remap_process() is now 3, not 1
	 5 Aug 2005: remap_process() now takes "dowrite" argument
	28 Jun 2005: refco() call to remap_distortion():
		much of remap_distortion functionality -> xglib/spline.c;
		variables rmsshift, maxshift subsumed into VSPLINE structure
	24 Jan 2005: in impose_symmetry(), if the spacegroup _has_ been defined
	 and it's consistent with the lattice character, don't override it
	17 Jun 2005: remap_distortion(): pix2cm() call takes frame-number arg
	24 Jan 2005: handle externally-declared spacegroup & crystal system;
		don't force spacegroup if it's grossly wrong
	18 Jan 2005: choosebravais() handles -50, -51 cases --
	 pick centering&lattice type based on hints from user
	10 Jan 2005: default value of gamma2 changed from 0.0026 to 0.0010
	 6 Sep 2004: cosmetic changes to remap_distortion()
	28 Aug 2004: routines used only here are declared as static; cosmetics
	15 Jun 2004: fifth arg to I command means turn off auto-centering
		after auto-indexing
	 4 Jun 2004: eliminate RR_CHANGEDDET concept;
		fmtoupr() now called with a flags argument
	31 May 2004: most routines from ../nogui/ref1main.c moved here;
		added "b q" functionality; minor mods to some other commands
	 9 Feb 2004: xp->xversion set properly in globtxs based on
		the defined string XGEN_VERSION
	26 Jan 2004: a few static globals converted to local variables
	21 Jan 2004: license message inserted
	 7 Jan 2004: changes to calls to show_bravchoice(),
		proc_bravais(), impose_symmetry(), etc.; some routines renamed
	 2 Jan 2004: REFPARS and REFINERUN structure elements in function calls
		and one static global replace remaining global variables;
		refparin() and refparout() used for UPARAMS I/O;
		scmp() moved to vectors.c;  function calls updated
	10 Dec 2003: remove references to framdeg
	 8 Nov 2003: convert lca from a vector to a pointer
	26 Nov 2003: replace global variables with elements of REFINERUN struct
	 8 Nov 2003: convert lca from vector to pointer
	 5 Nov 2003: cosmetic changes in choosebravais()
	19 Sep 2003: always initialize 'failu' in incr1p
	19 Sep 2003: set up a default in the switch in impose_symmetry()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	24 Aug 2003: display lattice type & centering
		using routines glattice, gcentering
	16 Aug 2003: handle exclusion code for geth differently
	14 Aug 2003: eliminate global REFPARS structure in globtxs
	13 Aug 2003: fmtoupr() moved to refxtal.c so that
		routines that need it don't need to bind refio.o
	 5 Jun 2003: rescaling gam[0],gam[1] based on fabs(framrad)
		rather than on framrad itself
	24 Mar 2003: rocking_ref() calls from "R A" command will only
			produce one line of logfile output, rather than many
	18 Jan 2003: slight reformatting of logfile output
	16 Jan 2003: allow R command to have a "# cycles" argument
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h
	 1 Nov 2002: establish "F Q" (pick the best) command;
		modify diagnostics from updateall() so it's easier to call;
		create list_transforms(); trivial reformatting
	 1 Oct 2002: Command "F P" permitted: it means "pick the best lattice"
	15 Jul 2002: failure in rocking-curve refinement is non-fatal
	 1 Apr 2002: "announce_change" takes an extra diagnostic arg
		and provides some additional output to stdout
	25 Mar 2002: removed MSI attachments
	18 May 2001: initial versions of refco(), etc.
	10 Oct 1995: output spline size is explicitly 512 + CALPOINT sizes.
	 6 Jul 1995: fmtoupr(): add diagnostic argument in call to memzap();
		don't try to seek in CENTROIDS file if it isn't open
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: remove slash/star from comment somewhere in refremap.c
	12 May 1995: make_ftab takes no argument; totalframes is a global
	 2 Mar 1995: add frame-number arg to pix2cm calls;
		diagnostics to stdout; better diagnostics in remap_distortion 
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 4 Nov 1994: determine detector type by call to dettype() in globtxs();
		save detector type and X-GEN version #
		to REFPARS structure in UPARAMS
	 5 Oct 1994: use xgread and xgwrite for REFPARS, TROID, VSPLINE
	30 Sep 1994: read wavelength from frame header if possible;
		read framesize as float if possible
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: reading and writing binary discfiles and log-file results
	within the refinement functionality in XGEN; detector remapping too.
   Cross references: ref*.c, vectors.c,
	../prerefine/cal*.c, ../xglib/{calinv,spline}.c
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_UNISTD	1
#include	"refidefs.h"
#define		GOODDATA	01
#define		INPUTREFERROR	02
#define		INPUTOBERROR	04
#define		NOPIXTOCM	010
#define		NLABVARS	17
#define		NLABPARS	13
#define		NSYNONYMS	16
#define		NLABELS		(NLABVARS + NLABPARS + NSYNONYMS)
#define		MAX_FOM		0.040	/* max FOM for accepting lattice char*/
#define		PHIFLOOR	(0.25)

static int defin2(REFINERUN *rfp)
{ /*	This takes two indexed reflections and the unit cell dimensions
	the user has already input and uses them to compute an orientation
	matrix for the crystal.
	It returns 0 if successful, various negative integers if not. */
	register int	i, j;
	float		zmi[3][3];
	double		sm[3][3], zm[3][3], dzmi[3][3];
	REFPARS		*xp;
	REFREF		*rp, *rp0;

	xp = rfp->rfr_xp;
	rp0 = rfp->rfr_rpp0;	while (rp0->wt < 0.1) rp0++;
	rp = rp0 + 1;	while (rp->wt < 0.1) rp++;
	/* first make sure the two hkl's aren't parallel */
	if (parallel(rp0->rh, rp->rh, 3)) return -3;
	/* Now get the unit cell parameters */
	if (ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys))
		return -4;
	if (vecmat(xp->xabc, xp->xang, xp->xareal)) return -5;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xp->xareal[i][j] /= xp->xlambda;
	if (inv3(xp->xareal, xp->xrec)) return -6;
	/* now construct two matrices Z and S, where the first two
	 columns of S are the two scattering vectors and the 1st 2 cols
	 of Z are zi = Aohi, where Ao= the recip.sp.unit cell matrix
	 referenced to some nominal position, and hi is the index vector.
	 The last column is obtained from the cross product of the first two.
	 Meanwhile, calculate the norms of the two s vectors and the two z
	 vectors and make sure |si| ~~ |zi|, i=1,2 */
	fprintf(rfp->rfr_fpl, "Reflections used:\n");
	fprintf(rfp->rfr_fpl, "   h   k   l      s0      s1      s2\n");
	fprintf(rfp->rfr_fpl, "%4.0f%4.0f%4.0f %7.4f %7.4f %7.4f\n",
	 rp0->rh[0], rp0->rh[1], rp0->rh[2], rp0->rs[0],rp0->rs[1],rp0->rs[2]);
	fprintf(rfp->rfr_fpl, "%4.0f%4.0f%4.0f %7.4f %7.4f %7.4f\n",
	 rp->rh[0], rp->rh[1], rp->rh[2], rp->rs[0], rp->rs[1], rp->rs[2]);
	for (i = 0; i < 3; i++)
	   {
		sm[i][0] = rp0->rs[i];	sm[i][1] = rp->rs[i];
		zm[i][0] = xp->xrec[0][i] * rp0->rh[0] +
		 xp->xrec[1][i] * rp0->rh[1] + xp->xrec[2][i] * rp0->rh[2];
		zm[i][1] = xp->xrec[0][i] * rp->rh[0] +
		 xp->xrec[1][i] * rp->rh[1] + xp->xrec[2][i] * rp->rh[2];
	   }
	if (getorthog(sm)) return -7;	/* construct a 3rd scattering vector */
	if (getorthog(zm)) return -7;	/* construct a 3rd hkl vector */
	/* we want the transpose of Q, where Q is the solution to S=QZ;
	thus what we really want is (SZ~)', where "'" means "transpose".
	But S and Z are orthogonal or close to it, so that's ZS'. */
	dmm3t(dzmi, zm, sm);	/* get solution Q' */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   {
		zmi[i][j] = dzmi[i][j];
	   }
	cleanorthog(dzmi);	/* make the orthogonal matrix more precise */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) zmi[i][j] = dzmi[i][j];
	/* rather than constructing the angles directly, we put together
	the actual real-space unit cell matrix here: */
	mm3(xp->xrec, xp->xareal, zmi); /* a convenient place to stick it */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xp->xareal[i][j] = xp->xrec[i][j];
	return rp2vecs(xp); /* clean it up */
}

int ref3p(REFINERUN *rfp)
{ /*	this computes the unit cell params & matrix from >= 3 hkls.
	It returns 0 if successful, various negative integers on failure. */
	register int	i, j, n;
	double		dewt, thiss, thish, thats;
	double		hst[3][3], sst[3][3], ssti[3][3];
	REFPARS		*xp;

	for (i = 3; i--; ) for (j = 3; j--; ) sst[i][j] = hst[i][j] = 0.;
	/* The matrix problem we will solve is min(W(H-RS))(W(H-RS))',
	where ' denotes transpose and W is a diagonal weight matrix.
	Solution is R = HDS'(SDS')~, where ~ denotes inverse and D=WW'.*/
	if (rfp->rfr_inbounds > 3)
	 {
	   rfp->rfr_rpp = rfp->rfr_rpp0 + rfp->rfr_firstok;
	   for (n = rfp->rfr_inbounds; n--; (rfp->rfr_rpp)++)
	     if ((rfp->rfr_rpp)->wt > FUZZ)
	       {
		 dewt = rfp->rfr_rpp->wt * rfp->rfr_rpp->wt;
		 for (i = 3; i--; )
		   {
			thish = rfp->rfr_rpp->rhi[i] * dewt;
			thiss = rfp->rfr_rpp->rs[i] * dewt;
			for (j = 3; j--; )
			   {
				hst[i][j] += thish * (thats =
					rfp->rfr_rpp->rs[j]);
				sst[i][j] += thiss * thats;
			   }
		   }
	       }
	 }
	else	 /* simpler matrix problem with only 3 refls: R = HS~ */
	 {
	   rfp->rfr_rpp = rfp->rfr_rpp0 + rfp->rfr_firstok;
	   for (n = rfp->rfr_inbounds; n--; rfp->rfr_rpp++)
		for (j = 3; j--; )
		   {
			hst[j][n] = rfp->rfr_rpp->rhi[j];
			sst[j][n] = rfp->rfr_rpp->rs[j];
		   }
	 }
	if (dinv3(sst, ssti)) return -10;	/* invert l.h.s. matrix */
	dmm3(sst, hst, ssti);	/* solve for real-space matrix */
	xp = rfp->rfr_xp;
	for (i = 0; i < 3; i++)
		for (j = 0; j < 3; j++) xp->xareal[i][j] = sst[i][j];
	return rp2vecs(xp);
}

static int domainb(int nmain, float coords[2][3], REFPARS *xp)
{ /* this computes ax0 and ay0 from the main-beam coordinate values.
		If two designations have been given, it computes
		plated and tilt as well. */ 

	int	jfr;
	double	deltan, tanrsw;
	float	xz0, xz1, yz0, yz1;

	jfr = coords[0][2] + 0.49999;
	if (!pix2cm(coords[0][0], coords[0][1], &xz0, &yz0, jfr)) return -20;
	tanrsw = tan(coords[0][2] * RADC);
	if (nmain > 1 && fabs(coords[1][2] - coords[0][2]) > 1)
	  { /* 2 mainbeam pos'ns given: get plateD and tilt */
		jfr = coords[1][2] + 0.49999;
		if (!pix2cm(coords[1][0], coords[1][1], &xz1, &yz1, jfr))
			return -21;
		deltan = tan(coords[1][2] * RADC) - tanrsw;
		xp->xtilt = atan((yz1 - yz0) / (xz1 - xz0));
		xp->xctilt = cos(xp->xtilt);
		xp->xstilt = sin(xp->xtilt);
		xp->xplated = ((xz1 - xz0) * xp->xctilt +
			(yz1 - yz0) * xp->xstilt) / deltan;
		xp->xpltd2 = xp->xplated * xp->xplated;
		xp->xisref[9] = xp->xisref[12] = 'N'; /* don't refine them */
	  } /* ... in any case, get ax0 and ay0 */
	/* compute x,y center and turn off refinements of x0, y0, swing */
	xp->xx0 = xz0 - xp->xplated * tanrsw * xp->xctilt;
	xp->xy0 = yz0 - xp->xplated * tanrsw * xp->xstilt;
	xp->xisref[10] = xp->xisref[11] = xp->xisref[13] = 'N';
	return 0;
}

int get_hkl(REFINERUN *rfp)
{ /*	This does a calculation of the unit cell matrix based on the reference
	reflections.  It opens a reference-reflection file and reads the
	reflections in it in format "h,k,l,X,Y,phi".  It gripes if it gets
	fewer than 3 refls; it computes the rotation matrix from 2 refls;
	and it calculates the entire unit cell matrix from >= 3 refls.
	An observation with h=k=l=0 is treated as info for a computation
	of the (X,Y) center of the reflection.  Return values and meanings:
	0		operation successful
	-1		REFERENCE file cannot be opened.
	-2		not enough spots provided to permit calculation.
	-3 - -9		errors from "defin2" prevented calculation
	-10 - -11	errors from "ref3" prevented calculation
	-20 - -21	errors in determining mainbeam position from "do_main"
*/
	register int	i, j;
	short		excluco;
	int		jfr, ndirbeam, nlost, retv;
	char		*cp, cline[96];
	float		cen[2][3], oldh[3];
	FILE		*fpref;
	REFPARS		*xp;
	REFREF		*lrp;

	if (NULL == (fpref = xgopen("reference", "r"))) return -1;
	lrp = rfp->rfr_rpp;
	for (rfp->rfr_npeaks = ndirbeam = 0, lrp = rfp->rfr_rpp0;
	   (NULL != fgets(cline, 90, fpref)); )
	   { /* compatibility: get rid of commas if there are any */
		for (cp = &(cline[0]); '\0' != *cp && '\n' != *cp; cp++)
			if (*cp == ',') *cp = ' ';
		if (6 > sscanf(cline,"%f %f %f %f %f %f",
		 &(lrp->rhi[0]), &(lrp->rhi[1]), &(lrp->rhi[2]),
		 &lrp->xto, &lrp->yto, &lrp->frame0)) continue;
		if (lrp->rhi[0] == 0 && lrp->rhi[1] == 0 && lrp->rhi[2] == 0)
		  {
			if (ndirbeam < 2) 
			  {
				cen[ndirbeam][0] = lrp->xto;
				cen[ndirbeam][1] = lrp->yto;
				cen[ndirbeam++][2] = lrp->frame0;
			  }
		  }
		else {
			lrp->framem = lrp->frame0; lrp++; (rfp->rfr_npeaks)++;
		     }
	   }
	(void)fclose(fpref);	/* don't need reference-refl file any more*/
	/* if we provided a 000 value, use its position to define mainbeam */
	xp = rfp->rfr_xp;
	if ((ndirbeam) && (retv = domainb(ndirbeam, cen, xp))) return retv;
	/* go through the remaining points and remember them */
	for (i = rfp->rfr_npeaks, nlost = 0, lrp = rfp->rfr_rpp0; i--; lrp++)
	   {
		lrp->wt = 1;	/* flat weighting for this purpose */
		lrp->maxdevh = 0; /* we're SETTING the hkls, right? */
		jfr = lrp->frame0 + 0.49999;
		if (!pix2cm(lrp->xto, lrp->yto,&(lrp->xtcm),&(lrp->ytcm), jfr))
		  {
			fprintf(rfp->rfr_fpl,
  "Reflection %6.1f%6.1f%6.1f%9.3f%9.3f%9.3f:cannot convert coords--ignored\n",
			lrp->rhi[0],lrp->rhi[1],lrp->rhi[2],
			lrp->xto,lrp->yto,lrp->frame0);
			lrp->wt = 0;	nlost++;
		  }
		else {
			rfp->rfr_rpp = lrp; scattv(rfp); refgonio(rfp);
			for (j = 3; j--; ) lrp->rh[j] = lrp->rhi[j];
		     }
	   }
	rfp->rfr_inbounds = rfp->rfr_npeaks;	rfp->rfr_firstok = 0;
	switch (rfp->rfr_npeaks - nlost)
	      {
		case	0:	case	1:	return -2;
		case	2:
		 if ((retv = defin2(rfp)))
		   {
			fprintf(rfp->rfr_fpl, "defin2:code %d\n", retv);
			return retv;
		   }
		break;
		case	3:	default:
		 if ((retv = ref3p(rfp)))
		   {
			fprintf(rfp->rfr_fpl, "ref3p:code %d\n", retv);
			return retv;
		   }
		break;
	      }
	fprintf(rfp->rfr_fpl,
	 "Observed and computed reference reflection data:\n");
	fprintf(rfp->rfr_fpl,
"   h   k   l      X      Y    phi     sx     sy     sz    h    k    l\n");
	/* produce globals that we'll use in geth's call of getinthkl */
	if ((xp->xisys < 3) || (xp->xisys == 7)) gethfacs(rfp);
	excluco = exclucode(xp->xispgp, xp->xisys);
	for (i = rfp->rfr_npeaks, lrp = rfp->rfr_rpp0; i--; lrp++) 
	   {
		rfp->rfr_rpp = lrp;
		for (j = 3; j--; ) oldh[j] = lrp->rhi[j];
		j = geth(rfp, excluco);
		fprintf(rfp->rfr_fpl,
	 "%4.0f%4.0f%4.0f%7.2f%7.2f%7.2f%7.4f%7.4f%7.4f%5.1f%5.1f%5.1f\n",
		 oldh[0], oldh[1], oldh[2], lrp->xto, lrp->yto, lrp->frame0, 
		 lrp->rs[0], lrp->rs[1], lrp->rs[2],
		 lrp->rh[0], lrp->rh[1], lrp->rh[2]);
	   }
	rfp->rfr_npeaks = 0;	/* signal: we need to read real data now */
	return 0;
}

static int g1upeak(FILE *fp, REFREF *rp)
{ /*	This reads data for a single peak into the REFREF reflection pointed
	to by `rp' from the unix-style "centroids" file.
	It returns GOODATA if it's successful; it returns 0 if the first
	part of the reflection is absent; it returns other codes in the event
	of read errors within the record or when the (x,y) of the
	reflection can't be mapped from pixels to centimeters */
	int	jfr;

	if (0 >= xgread((void *)rp, XG_TROID, fp)) return 0;
	jfr = rp->frame0 + 0.49999;
	if (1 != pix2cm(rp->xto,rp->yto,&rp->xtcm,&rp->ytcm, jfr))
		return NOPIXTOCM;
	rp->framem = rp->frame0;	/* save the unrefined frame number */
	/* the log of the intensity is used in weighting the observations */
	rp->totinten = (rp->totinten < 3. ? 1 : log(rp->totinten));
	return GOODDATA;
}

static int g1vpeak(FILE *fp, REFREF *rp)
{ /* This reads data for a single peak from a VMS-style "troidinf" file
	into the REFREF reflection pointed to by `rp'.
	It returns GOODATA if it's successful; it returns 0
	if the record counter is absent; it returns other codes in the
	event of read errors within the record or when the (x,y) of the
	reflection can't be mapped from pixels to centimeters */
	int	jfr;
	short	unfsp;

	if (1 != fread((void *)&unfsp, sizeof (short), 1, fp)) return 0;
	if (0 >= xgread((void *)&rp, XG_TROID, fp)) return INPUTREFERROR;
	jfr = rp->frame0 + 0.49999;
	if (1 != pix2cm(rp->xto, rp->yto, &rp->xtcm, &rp->ytcm, jfr))
		return NOPIXTOCM;
	rp->framem = rp->frame0;	/* save the unrefined frame number */
	rp->totinten = 1;	/* no intensity in this type of file*/
	return GOODDATA;
}

int gpeaks(REFINERUN *rfp)
{ /*	This routine reads the [u]troidinf file to obtain the omega and
	(X,Y) centroids of some bright reflections.  `gpeaks' generates a
	set of scattering vectors and non-integer indices from these values.
	It returns the number of reflections NOT used if successful;
	it returns -1 if it can't open the file. */
	register int	nnotused, readerr;
	FILE		*fpi;

	nnotused = 0;
	if (rfp->rfr_flags & RR_ISOLDTROID)
	  {
		rfp->rfr_cname = trname("troidinf");
		if (NULL == (fpi = xgopen(rfp->rfr_cname, "r"))) return -1;
		 for (rfp->rfr_rpp = rfp->rfr_rpp0;
		 (readerr = g1vpeak(fpi, rfp->rfr_rpp)); )
		     {
			if (readerr == GOODDATA) (rfp->rfr_rpp)++;
			else	nnotused++;
		     }
		(void)fclose(fpi);	fpi = NULL;
	  }
	else
	  {
		rfp->rfr_cname = trname("centroids");
		if (NULL == (fpi = xgopen(rfp->rfr_cname, "r"))) return -1;
		for (rfp->rfr_rpp = rfp->rfr_rpp0;
		 (readerr = g1upeak(fpi, rfp->rfr_rpp)); )
		     {
			if (readerr == GOODDATA) rfp->rfr_rpp++;
			else	nnotused++;
		     }
		(void)fclose(fpi);	fpi = NULL;
	  }
	rfp->rfr_npeaks = rfp->rfr_rpp - rfp->rfr_rpp0;
	rfp->rfr_tl0 = NOW;
	strcpy(rfp->rfr_rundate, ctime(&(rfp->rfr_tl0)));
	fprintf(rfp->rfr_fpl,
		"%6d reflections read in", (int)(rfp->rfr_npeaks));
	if (strlen(rfp->rfr_cname) > 17) fputs("\n  ", rfp->rfr_fpl);
	fprintf(rfp->rfr_fpl, " from %s at %s", rfp->rfr_cname,
		rfp->rfr_rundate);
	xeflush(rfp->rfr_fpl);
	get_rmserrs("", rfp);/* compute rms errors in index,X,Y,phi */
	return nnotused;
}

int remap_distortion(REFINERUN *rfp)
{ /*	This computes the average deviation of predicted from calculated
	X and Y for all usable reflections in the vicinity of each of
	the fiducial points.  It then applies this mean deviation to
	appropriate values in pixels. Then it does the pair of hermite
	splines to get the pixel-to-cm conversion, and applies them
	back to render everything self-consistent.  The routine returns
	pointers to the maximum and rms shifts in its arguments.
	Return values and their meanings:
	0	successful run of `remap_distortion'
	-1	failure to reserve memory for remapping
	1	failure to open new USPLINE file
	2	failure to write out new USPLINE file header
	3	failure to write main body of USPLINE file
	4	failure to determine main-beam position */
	int		i;
	double		rweight, *adel, lov[11];
	VSPLINE		*lsp;
	REFREF		*rp;

	get_rmserrs("before remapping", rfp); /* get initial stats */
	rfp->rfr_tl0 = NOW;
	lsp = rfp->rfr_spl;
	if (NULL == (adel = remap_setup(lsp, &(lov[8])))) return -1;
	scalars(rfp);	matrices(rfp);	/* so calls to "refl_coords" work */
	/* now go through the observational data, computing the shifts and
	the weights for each reflection relative to each fiducial point */
	for (rfp->rfr_rpp = rfp->rfr_rpp0, i = rfp->rfr_npeaks;
	 i--; (rfp->rfr_rpp)++)
	   if (refl_coords(rfp) == 3)
	     { /* refl must be well-behaved */
		rp = rfp->rfr_rpp;
		lov[0] = rp->xto - rp->xtc;	lov[1] = rp->yto - rp->ytc;
		lov[2] = rp->frame0 - rp->framc;
		lov[3] = rp->xtcm;	lov[4] = rp->ytcm;
		rweight = rp->framc - rp->framem;
		 /* weight each refl by log(I)/(A + rmserr(phi)) */
		lov[5] = rp->totinten / (PHIFLOOR + rweight * rweight);
		lov[6] = rp->xto;	lov[7] = rp->yto;
		if (EOK != remap_sums(lsp, lov, adel)) return -1;
	     } /* close "for (rfp... " loop and "if (refl_coords" clause */
	return remap_process(lsp, adel, rfp->rfr_fpl, 3);
}

static void announce_change(FILE *fp, int isf, int thisi,
	double thisv, char *thisstr, char *fldstr)
{ /*	This records a recent change in a parameter */
	if ((0 <= isf) && (isf < NLABVARS)) fprintf(fp,
	 "new value of variable %s is %10.4f based on string input (%s)\n",
		fldstr, thisv, thisstr);
	else if ((isf < 25) || (isf > 26)) fprintf(fp,
	 "new value of the parameter %s is %10.4f based on string input (%s)\n",
		fldstr, thisv, thisstr);
	else fprintf(fp,
	 "new value of parameter %s is %2d based on string input (%s)\n",
		fldstr, thisi, thisstr);
}

static int ref1modify(int ac, char *av[], float *mainb, REFINERUN *rfp,
	int realv[], char *flds[])
{ /*	This enables a modification to one of the variables or parameters.
	We return the main-beam coordinates in mainb if appropriate */
	register int	ic, lac;
	int		isfound, parmod;
	short		thisiv;
	char		*cp, **lav, loclabel[20], locvalue[40];
	double		g0, g1, thisvalue;
	REFPARS		*xp;

	if (ac < 2) return EBADVALUE;	/* parameters come in pairs */
	parmod = 0;	g0 = g1 = -HUGE;	isfound = EOK;
	xp = rfp->rfr_xp;
	for (lac = 2, lav = av + 2; lac < ac - 1; lac += 2, lav += 2)
	   {
		strcpy(loclabel, "");	strcpy(locvalue, "");
		strcpy(loclabel, *lav);	strcpy(locvalue, *(lav+1));
		for (cp = loclabel; '\0' != *cp; cp++)
			if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		for (cp = locvalue; '\0' != *cp; cp++)
			if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		thisvalue = atof(locvalue); thisiv = (short)(atoi(locvalue));
		isfound = -1;
		for (ic = 0; ic < NLABELS; ic++)
		   {
			if (!strcmp(loclabel, flds[ic]))
			  {
				isfound = realv[ic];	break;
			  }
		   }
		if ((0 <= isfound) && (isfound <= 8)) parmod |= 1;
		else if ((9 <= isfound) && (isfound <= 13)) parmod |= 2;
		else if ((14 <= isfound) && (isfound <= 16)) parmod |= 4;
		switch (isfound)
		      {
			case	0:	xp->xabc[0] = thisvalue;	break;
			case	1:	xp->xabc[1] = thisvalue;	break;
			case	2:	xp->xabc[2] = thisvalue;	break;
			case	3:	xp->xang[0] = thisvalue;	break;
			case	4:	xp->xang[1] = thisvalue;	break;
			case	5:	xp->xang[2] = thisvalue;	break;
			case	6:	xp->xdelom = thisvalue;		break;
			case	7:	xp->xdelchi = thisvalue;	break;
			case	8:	xp->xdelphi = thisvalue;	break;
			case	9:	xp->xplated = thisvalue;	break;
			case	10:	xp->xx0 = thisvalue;	break;
			case	11:	xp->xy0 = thisvalue;	break;
			case	12:	xp->xtilt = thisvalue * RADC;	break;
			case	13:	xp->xswing = thisvalue * RADC;	break;
			case	14:	g0 = thisvalue;		break;
			case	15:	g1 = thisvalue;		break;
			case	16:	xp->xgam[2] = thisvalue;	break;
			case	17:	xp->xerrind = thisvalue;	break;
			case	18:	xp->xerrx = thisvalue;	break;
			case	19:	xp->xerry = thisvalue;	break;
			case	20:	xp->xerrphi = thisvalue;	break;
			case	21:	xp->xresmin = thisvalue;	break;
			case	22:	xp->xresmax = thisvalue;	break;
			case	23:	xp->xlambda = thisvalue;	break;
			case	24:	xp->xfrmsize = thisvalue * RADC; break;
			case	25:	xp->xisys = rfp->rfr_spsys = thisiv;
				if (0 == rfp->rfr_spspgrp) rfp->rfr_spcen = 1;
				else rfp->rfr_spcen =
					spgtocen(rfp->rfr_spspgrp);
				break;
			case	26:	xp->xispgp = rfp->rfr_spspgrp = thisiv;
				rfp->rfr_spsys = spgtosys(rfp->rfr_spspgrp);
				rfp->rfr_spcen = spgtocen(rfp->rfr_spspgrp);
				break;
			case	27:	xp->xstart[0] = thisvalue;	break;
			case	28:	xp->xstart[1] = thisvalue;	break;
			case	29:	xp->xstart[2] = thisvalue;	break;
			default: break;
		     }
		if ((0 <= isfound) && (isfound <= 29))
		  {
			announce_change(stdout, isfound, thisiv,
			 thisvalue, locvalue, flds[isfound]);
			announce_change(rfp->rfr_fpl, isfound, thisiv,
			 thisvalue, locvalue, flds[isfound]);
		  }
	   }
	if ((lac < 4) && (isfound < 0)) return EBADVALUE;
	isfound = EOK;	if (parmod & 1) isfound = unitcellx(rfp);
	/* gamma0 and gamma1 are special cases because they need to
	 be rescaled by the (possibly) new value of the framesize */
	if (g0 > -1.e6) xp->xgam[0] = g0 * fabs(xp->xfrmsize);
	if (g1 > -1.e6) xp->xgam[1] = g1 * fabs(xp->xfrmsize);
	get_rmserrs("", rfp);	if (parmod & 2) getmainbeam(xp, mainb);
	return isfound;
}

static void announce_permitchange(FILE *fp, char newperm, char *fldstr)
{ /*	Announce a change of parameter `str' to state `newperm' on stream fp */
	fprintf(fp, "refinement permission for parameter %s is now %c\n",
		fldstr, newperm);
}

static int permit1modify(int ac, char *av[], REFINERUN *rfp,
	int realv[], char *flds[])
{
	register int	ic, lac;
	int		isfound;
	char		newpermit, *cp, **lav, loclabel[20], locvalue[40];
	REFPARS		*xp;

	/* parameters come in pairs */
	if (ac < 2) return EBADVALUE;	isfound = -1;
	xp = rfp->rfr_xp;
	for (lac = 2, lav = av + 2; lac < ac - 1; lac += 2, lav += 2)
	   {
		strcpy(loclabel, "");	strcpy(locvalue, "");
		strcpy(loclabel, *lav);	strcpy(locvalue, *(lav+1));
		for (cp = loclabel; '\0' != *cp; cp++)
			if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		for (cp = locvalue; '\0' != *cp; cp++)
			if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		newpermit = locvalue[0];	isfound = -1;
		for (ic = 0; ic < NLABELS; ic++)
		   {
			if (!strcmp(loclabel, flds[ic]))
			  {
				isfound = realv[ic];	break;
			  }
		   }
		if ((0 <= isfound) && (isfound < NLABVARS) &&
		 (xp->xisref[isfound] != newpermit))
		  {
			announce_permitchange(stdout,
			 newpermit, flds[isfound]);
			announce_permitchange(rfp->rfr_fpl,
			 newpermit, flds[isfound]);
			xp->xisref[isfound] = newpermit;
		  }
	   }
	if ((lac < 4) && (!isfound)) return EBADVALUE;
	return EOK;
}

void showpeaks(FILE *fp, char *s, REFINERUN *rfp)
{ /* This displays the reflections under consideration in a refinement. */
	register int	i;
	REFREF		*lrp;

	fprintf(fp," Reflections under consideration %s\n",s);
	fprintf(fp, "       Indices         X             Y                ");
	fprintf(fp, "frame          inten flg\n");
	fprintf(fp, "     h    k    l  obs'd calc'd  obs'd calc'd");
	fprintf(fp, "   input modif'd  calc'd  sity\n");
	for (i = rfp->rfr_npeaks, lrp = rfp->rfr_rpp0; i--; lrp++)
	   {
		if (!legald(rfp, lrp)) lrp->rflags |= INICERING;
		else lrp->rflags &= ~INICERING;
		fprintf(fp,
	 " %5.1f%5.1f%5.1f%7.2f%7.2f%7.2f%7.2f%8.3f%8.3f%8.3f%7.0f%3o\n",
		 lrp->rh[0], lrp->rh[1], lrp->rh[2], lrp->xto, lrp->xtc, 
		 lrp->yto, lrp->ytc, lrp->frame0, lrp->framem, lrp->framc, 
		 exp(lrp->totinten), lrp->rflags);
	   }
	rfp->rfr_rpp = lrp;
	fprintf(fp, "\n\n");
	rfp->rfr_flags |= RR_NEEDHEADER; /* to print headers on later cycles */
	return;
}

int globtxs(int op, REFINERUN *rfp)
{ /*	This does the complementary operation to `xstclean', sort of.
	It then writes the parameter structure to disk.
	If (op & 01) is nonzero, we set up the structure for writing.
	If (op & 02) is nonzero, we write the structure out.
	If (op & 04) is nonzero, we announce the writing out. */
	register int		i;
	int			nitems, maj, mid, min, detdim[2];
	char			*verstr, *uparnam;
	REFPARS			*xp, outp;

	xp = rfp->rfr_xp;
	/* record the detector dimensions in the xp structure */
	xp->xdetwid = (rfp->rfr_spl)->v_detw;
	xp->xdetht = (rfp->rfr_spl)->v_deth;
	verstr = XGEN_VERSION;
	maj = mid = min = 0;
	nitems = sscanf(verstr, "%d.%d.%d\n", &maj, &mid, &min);
	if (maj > 320) maj = 320;	else if (maj < 0) maj = 0;
	if (mid > 9) mid = 9;		else if (mid < 0) mid = 0;
	if (min > 9) min = 9;		else if (min < 0) min = 0;
	xp->xversion = 100 * maj + 10 * mid + min;
	if (op & 01) /* this is almost pointless at this stage */
	  {
		for (i = ALLPARS; i < 20; i++) xp->xisref[i] = 'N';
		xp->xdtype = dettype(detdim);	/* detector type */
		if (op & 04) fprintf(stdout,
	   "data structure for output has been established and updated\n");
		if (!(op & 02)) return 0;	/* in this case we're done */
	 }
	/* since we now explicitly call refparout(), we're covered:
	 if user has changed what UPARAMS points to, we'll use the new one */
	uparnam = trname("uparams");
	/* we're going to mess up the REFPARS structure, so we
	 scribble on a temporary copy thereof */
	cpymem((AVOID *)&outp, (AVOID *)xp, sizeof (REFPARS));
	if (EOK != (i = refparout(&outp)))
	  {
		fprintf(rfp->rfr_fpl,
		 "Warning: unable to write parameters to %s\n", uparnam);
		return 3;
	  }
	if (op & 04) fprintf(stdout, "Parameters have been written to %s\n",
		uparnam);
	return 0;
}

static void fixisref(REFPARS *xp)
{ /* this turns on and off the refinement of angles according to xtal system*/
	register int	i;

	switch (xp->xisys)
	      {
		case	1:	/* triclinic: refine 'em all */ 
			for (i = 0; i < 6; i++) xp->xisref[i] = 'Y'; return;
		case	2:	/* monoclinic: a,b,c,beta */
			for (i = 0; i < 3; i++) xp->xisref[i] = 'Y';
			xp->xisref[3] = xp->xisref[5] = 'N';
			xp->xisref[4] = 'Y'; return;
		case	3:	/* orthorhombic: a,b,c */
			for (i = 0; i < 3; i++) xp->xisref[i] = 'Y';
			for (i = 3; i < 6; i++) xp->xisref[i] = 'N'; return;
		case	4:	case	6:	/* tetragonal,hexagonal: a,c */
			xp->xisref[0] = xp->xisref[2] = 'Y';
			xp->xisref[1] = 'N';
			for (i = 3; i < 6; i++) xp->xisref[i] = 'N'; return;
		case	5:	/* cubic: a only */
			xp->xisref[0] = 'Y';
			for (i = 1; i < 6; i++) xp->xisref[i] = 'N'; return;
		case	7:	/* rhombohedral: a,gamma only */
			xp->xisref[0] = xp->xisref[5] = 'Y';
			for (i = 1; i < 5; i++) xp->xisref[i] = 'N'; return;
	      }
}

static int xstclean(REFPARS *xp, int nbread, int fixis, REFINERUN *rfp)
{ /* This cleans up the REFPARS structure `xp' and puts a few
	useful values into *rfp and creates a few static globals.
	It differs from the older xstglob in that it doesn't produce
	or rely on any global variables whatsoever. */
	register int	i, j;
	int		isnewfmt;
	float		lare[3][3];
	REFPARS		*lpp;

	xp->xpltd2 = xp->xplated * xp->xplated;
	xp->xssw = sin(xp->xswing);	xp->xcsw = cos(xp->xswing);
	xp->xstilt = sin(xp->xtilt);	xp->xctilt = cos(xp->xtilt);
	if (xp->xlambda < FUZZ) xp->xlambda = DEF_LAMBDA;
	/* always work with the rocking-curve parameters as their moduli */
	for (i = 3; i--; ) xp->xgam[i] = fabs(xp->xgam[i]);
	/* require real-space unit cell to have non-negative determinant */
	if (deter(xp->xareal) < -DFUZZ)	/* negative determinant: */
	  {
	   for (i = 3; i--; )
	      {
		xp->xareal[2][i] = -xp->xareal[2][i];	/* flip c */
		xp->xrec[i][2] = -xp->xrec[i][2];	/* flip c* */
	      }
	  }
	if ((unsigned)nbread >= sizeof (REFPARS) - 24)	/* longer file format*/
	  {
		if ((xp->xisys < 1) || (xp->xisys > 7)) xp->xisys = 1;
		xp->xispgp = check_sys(xp->xispgp, xp->xisys);
		if (fixis) fixisref(xp); /* refine only some pars */
		if (xp->xerrind < FUZZ || xp->xerrind > 0.5)
			xp->xerrind = 0.2;
		/* read goniostat start angles from file if they're there.
		If this is an old-format file with strings containing the
		axis directions in those slots instead, just set zeroes. */
		if (nbread == sizeof (REFPARS)) isnewfmt = 1;
		else {
			for (i = 0; i < 3; i++) xp->xstart[i] = 0.;
			isnewfmt = 0;
		     }
		xp->xcmd = 'R';	/* start with 'R' no matter what */
	  }
	else	/* shorter file format: get some default values */
	  {
		isnewfmt = 0;	/* by definition this is an old format */
		for (i = 0; i < ALLPARS; i++) xp->xisref[i] = 'Y';
		xp->xisref[16] = 'N';	xp->xcmd = 'R';
		xp->xisys = 1;	xp->xispgp = 1;
		if (fixis) fixisref(xp);
		xp->xerrind = 0.2;
		xp->xerrx = xp->xerry = xp->xerrphi = 5.;
		xp->xresmin = 80.;	xp->xresmax = 2.;
		for (i = 0; i < 3; i++) xp->xstart[i] = 0.;
		xp->xdelom = xp->xdelchi = xp->xdelphi = 0.;
	  }
	/* components of rfp that mirror things in xp */
	/* if the starting omega,chi,phi values appear to be zero
	at this stage,  use the values that appear in the headers
	of the raw data themselves (assuming those data are available). */
	for (i = j = 0; i < 3; i++) if (fabs(xp->xstart[i]) > FUZZ) j++;
	if ((!j) && NULL != (lpp = rfp->rfr_fmprp))
		for (i = 0; i < 3; i++) xp->xstart[i] = lpp->xstart[i];
	/* If we've read in a new-format file, then we have been storing
	 the matrix and pseudo-goniostat angles appropriate to the reference
	 position being (0,0,0) rather than frame 0 of the current run.
	 We work within the program, though, with the values appropriate
	 to frame zero of the current run. */
	ztos(xp->xstart, rfp->rfr_zsa);
	if (isnewfmt)
	  {
		for (i = 3; i--; ) for (j = 3; j--; )
			lare[i][j] = xp->xareal[i][j];
		mm3(xp->xareal, lare, rfp->rfr_zsa);	/* rotate with it */
		if (inv3(xp->xareal, xp->xrec)) return -6;
	  }
	/* To prevent all this newfangled stuff from screwing up old
	 uparams files, we throw away the input values of xdelchi, xdelphi,
	 and xdelom and recalculate them from the unit cell matrix. */
	return rp2vecs(xp);
}

int readpars(REFINERUN *rfp)
{ /* See if the parameter file exists.  If it does, read the parameters in.
	If it doesn't, get the appropriate values from dataframe headers.
	If that doesn't work either, use some some innocuous default values.
	In any case, copy the structure elements to some global variables. */
	register int	i;
	int		ier, nbread;
	REFPARS		*xp;

	if (NULL == (rfp->rfr_fpl = propen("xlog"))) rfp->rfr_fpl = stdout;
	/* explicitly determine the name of the UPARAMS file in case
	 the user has changed it from within REFINE since we invoked it.*/
	xp = rfp->rfr_xp;
	if (NULL == rfp->rfr_spl) /* need spline to get detector dimensions */
	  {
		if (NULL == (rfp->rfr_spl = gspline(5, &ier)))
			return ier;
	  }
	xp->xdetwid = (rfp->rfr_spl)->v_detw;	/* get detector dimensions */
	xp->xdetht = (rfp->rfr_spl)->v_deth;
	if (EOK == (i = refparin(xp, &nbread)))
	  {
		xp->xcmd = 'R';	return unitcellx(rfp);
	  }
	/* we try to read this stuff from the raw data frames */
	if (NULL == rfp->rfr_fmprp) rfp->rfr_fmprp = rfp->rfr_xp + 1;
	if (EOK == fmtoupr(rfp->rfr_flags, rfp->rfr_fmprp))
	  {
		if (EOK == (i = xstclean(rfp->rfr_fmprp,
			(int)(sizeof (REFPARS)), 0, rfp)))
		  {
			printf(
 "No input parameter file available: values read from dataframe headers\n");
			sleep(2);	/* so message is readable */
			/* success: use those parameters */
			cpymem((AVOID *)xp, rfp->rfr_fmprp, sizeof (REFPARS));
			return unitcellx(rfp);
		  }
		else printf(
	 "Failure %d to read parameters from dataframes: defaults used\n", i);
	  }
	else printf(
 "No input parameter file OR dataframe header available: defaults used\n");
	for (i = 0; i < 3; i++)
	   {
		xp->xabc[i] = 100.;	xp->xang[i] = 90.; xp->xstart[i] = 0.;
	   }
	xp->xisys = 3;	/* default == orthorhombic	*/
	xp->xispgp = 19; /* default = P212121 (intl tables # 19) */
	xp->xfrmsize = 0.25 * RADC;	/* default = 0.25 degrees */
	xp->xgam[0] = xp->xgam[1] = fabs(xp->xfrmsize);
	xp->xgam[2] = 0.0010;
	xp->xlambda = DEF_LAMBDA;	/* default wavelength */
	xp->xerrind = 0.2;	/* accept nonintegerness < 0.2 units */
	xp->xerrx = 5.;		/* accept X   error < 5 pixels */
	xp->xerry = 5.;		/* accept Y   error < 5 pixels */
	xp->xerrphi = 5.;	/* accept phi error < 5 images */
	xp->xresmin = 80.;	xp->xresmax = 1.2;
	/* default: refine all parameters except cell angles, gamma2 */
	for (i = 0; i < ALLPARS; i++) xp->xisref[i] = 'Y';
	xp->xisref[3] = xp->xisref[4] = xp->xisref[5] =
		xp->xisref[16] = 'N';
	xp->xplated = 12.;
	xp->xx0 = xp->xy0 = xp->xtilt = xp->xswing = 0.;
	xp->xcmd = 'R';	/* default is to start with rocking-curve ref*/
	return unitcellx(rfp);
}

void impose_symmetry(REFPARS *xp, int lsym)
{ /*	This imposes a symmetry on a cell based on the
	lattice character number */
	int	tig, tis;
	short	newcen, newsys, oldcen, oldsys;

	switch (lsym)
	      {
		case	1:	/* Face-centered cubic */
			tis = 5;	tig = 196;	break;
		case	2: case 4: case 9: case 24: /* rhombohedral*/
			tis = 6;	tig = 146;	break;
		case	45:	/* rhombo. indexed rhombohedrally */
			tis = 7;	tig = 146;	break;
		case	3:	/* primitive cubic */
			tis = 5; tig = 195;	break;
		case	5:	/* body-centered cubic */
			tis = 5; tig = 197;	break;
		case	6: case	7: case 15: case 18: /* tetragonal I*/
			 	tis = 4; tig = 79;	break;
		case	8: case	19:	case	42: /* ortho. I */
				tis = 3; tig = 23;	break;
		case	10: case 14: case 17: case 20: case 25:
		case	27: case 28: case 29: case 30: case 37:
		case 39:	case 41:
			tis = 2; tig = 5;	break; /* mono. C */
		case	11:	case 21:	/* tetragonal P */
			tis = 4;	tig = 75;	break;
		case	12:	case 22:	/* hexag./trigonal P*/
			tis = 6;	tig = 143;	break;
		case	13: case 23: case 36: case 38: case 40:
			tis = 3; tig = 21; break; /* ortho. C */
		case	16: case 26:	/* orthorhombic F */
			tis = 3; tig = 22; break;
		case	32:	/* orthorhombic P */
			tis = 3;	tig = 16;	break;
		case	33: case 34: case 35:	/* monoclinic P */
			tis = 2;	tig = 3;	break;
		case	43: /* monoclinic I; convert to C later */
			tis = 2;	tig = 4;	break;	
		case	31: case 44:	default:	/* triclinic P */
			tis = tig = 1;	break;
	      }
	xp->xisys = tis;	/* always impose the point group */
	/* if the spacegroup stored locally is consistent with the
	 symmetry, don't replace it */
	oldsys = spgtosys(xp->xispgp);	oldcen = spgtocen(xp->xispgp);
	newsys = tis;	newcen = spgtocen((short)tig);
	if ((newsys != oldsys) || (newcen != oldcen)) xp->xispgp = tig;
	fixisref(xp);
}

static int proc_bravais(int lcind, LATTICESCORE *lc0, REFINERUN *rfp)
{ /*	This processes the Bravais lattice at lc0 + lcind.
	It does so as a manipulation of the REFINERUN structure. */
	int		j, jj;
	float		cel[9];
	REFPARS		*xp;
	LATTICESCORE	*lcap;

	lcap = lc0 + lcind;
	headbravais(rfp->rfr_fpl);	headbravais(stdout); xp = rfp->rfr_xp;	
	show_bravchoice(rfp->rfr_fpl, xp->xstart, lcap);
	show_bravchoice(stdout, xp->xstart, lcap);
	artocel(xp->xstart, lcap->lc_cel, cel);
	for (j = 0; j < 3; j++)
	   {
		xp->xabc[j] = cel[j];	xp->xang[j] = cel[j+3];
	   }
	xp->xdelom = cel[6]; xp->xdelchi = cel[7]; xp->xdelphi = cel[8];
	for (jj = 0; jj < 3; jj++) for (j = 0; j < 3; j++)
		xp->xareal[j][jj] = lcap->lc_cel[j][jj] / xp->xlambda;
	impose_symmetry(xp, lcap->lc_latchar);
	if (EOK != (j = rp2vecs(xp))) return j;
	if (EOK != impose_special(rfp)) return -7;
	return lcind;
}

static int choosebravais(int bravchoice, REFINERUN *rfp)
{ /*	This actually chooses a particular Bravais choice.
	It returns its argument if successful, negative integers if not */
	int		besti, i, j, latcode, locchar, nbv;
	short		tcen, tsys;
	double		bestfom, lar[3][3];
	REFPARS		*xp;
	LATTICESCORE	*lc0, *lcap;

	if (NULL == (lc0 = managelca(3))) return -62;
	for (xp = rfp->rfr_xp, i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		lar[i][j] = xp->xareal[i][j] * xp->xlambda;
	/* The best solution can be designated either 'a' or 'q' */
	if (bravchoice == 0) bravchoice = 16;
	if (bravchoice < 0)
	  {
		if (49 > (nbv = -bravchoice))
		 {
		  for (i = 0, lcap = lc0; i < 45; i++, lcap++)
		    {
			if (lcap->lc_choice >= 0)
			  { /* first look for an exact numeric match
				for the lattice character. */
				if (nbv == lcap->lc_latchar)
				  {
					bravchoice = i; break;
				  } /* end of "if (nbv ==" clause */
			  } /* end if "if (lcap->lc_choice >= 0)" clause */
		    } /* end of loop through i values */
		 } /* end of "if (49 > (nbv = -bravchoice))" clause */
		/* if we haven't found a match that way, we look for a
		 match in the lattice code and the centering code */
		if (bravchoice < 0)
		 {
		   if (nbv <= 49) /* define lattice system and centering */
		     {
			tsys = lattcode(nbv);	tcen = centercode(nbv);
		     } /* ... based on this choice */
		   else if (nbv == 50) /* here we have specified spgrp */
		    {
			tsys = spgtosys(rfp->rfr_spspgrp);
		   	tcen = spgtocen(rfp->rfr_spspgrp);
		    }
		  else {
			tsys = rfp->rfr_spsys; tcen = rfp->rfr_spcen;
		       } /* here we've specified system & centering */
		  for (i = 0, lcap = lc0; i < 45; i++, lcap++)
		   { /* step through the choices and look for a match
			that has an acceptable figure of merit */
			if ((lcap->lc_choice >= 0) &&
			 ((nbv <= 49) || (lcap->lc_fom < MAX_FOM)))
			  {
				locchar = lcap->lc_latchar;
				if ((tsys == lattcode(locchar)) &&
				 (tcen == centercode(locchar)))
				  {
					bravchoice = i; break;
				  }
			  }
		   }
		/* if we get to here with nbv == 50 or 51, it's because the
		 recommended spacegroup or system-and-centering had a bad FOM.
		 In that case we should just pick the one with the best FOM.*/
		  if ((nbv >= 50) && (bravchoice < 0))
		    {
			locchar = lc0->lc_latchar;
			rfp->rfr_spsys = lattcode(locchar);
			rfp->rfr_spcen = centercode(locchar);
			impose_symmetry(rfp->rfr_xp, locchar);
			rfp->rfr_spspgrp = (rfp->rfr_xp)->xispgp;
			bravchoice = 0;
		    } /* end of "if ((nbv >= 50)" clause */
		 } /* end if "if (bravchoice < 0)" clause */
		if (bravchoice < 0) return -11;
		lcap = lc0 + bravchoice;
		return proc_bravais(bravchoice, lc0, rfp);
	  }
	/* we get here with a positive value of `bravchoice'. */
	if (bravchoice == 16)	/* 'Q' means "pick the best" */
	  { /* by definition the best solution is the one with
		 the smallest figure of merit */
		for (bestfom = 1.e9, besti = -1, i = 0, lcap = lc0;
		 i < 45; i++, lcap++)
		   if (lcap->lc_choice >= 0)
		     {
			latcode = lcap->lc_latchar;
			if (lcap->lc_fom < bestfom)
			  {
				besti = i;	bestfom = lcap->lc_fom;
			  }
		     }
		if (besti >= 0)
		  {
			lcap = lc0 + besti;	latcode = lcap->lc_latchar;
			fprintf(rfp->rfr_fpl,
 "Choosing the %s %s lattice: raw score = %9.4f, fom = %9.4f\n",
			 glattice(latcode), gcentering(latcode),
			 100. * lcap->lc_score, lcap->lc_fom);
			return proc_bravais(besti, lc0, rfp);
		  }
		else return proc_bravais(0, lc0, rfp);
	  }
	for (i = 0, lcap = lc0; i < 45; i++, lcap++)
	   if (lcap->lc_choice == bravchoice)
		return proc_bravais(i, lc0, rfp);
	return -bravchoice; /* shouldn't get here */
}

void list_commands(void)
{ /* issue a list of commands to the screen */
	int		i;
	static char	*helprows[] = {
 " A:nonlinear,all resids B:Bravais calculation  C:closed-form ref'ment",
 " D:modify a parameter   E:exit                 F:Bravais choice",
 " G:re-read params< disk H:nonlinear,delta H    I:auto-index",
 " J:alter refinabilities K:linear ref on phi    L:linear ref on H",
 " M:reMap pix-to-cm corr N:interactive          O:print refs to disk",
 " P:nonlinear, delta phi Q:quit without output  R:RockingCurve ref'mnt",
 " S:index from 3 refls   T:transform matrix     U:update params&print",
 " V:re-print this list   W:write params > disk  X:linear ref on XY",
 " Y:linear ref,XYomega   Z:linear ref,omega   0-9:nonlinear,phi+H", "" };

	printf("Commands available interactively:\n");
	for (i = 0; (NULL != helprows[i]) && ('\0' != helprows[i][0]); i++)
		puts(helprows[i]);
}

static void list_transforms(void)
{
	int		i;
	static char	*chelp[] = {
 "a: primitive a <b <c   ; b: permute (abc)->(bca); c: permute (abc)->(cab);",
 "d: swap (abc)->(bac)   ; e: swap (abc)->(cba)   ; f: swap (abc)->(acb)   ;",
 "g: supplement alfa,beta; h: supplement alfa,gama; i: supplement beta,gama;",
 "j: use $REINDEX matrix ;", "" };

	printf("Transforms available interactively:\n");
	for (i = 0; (NULL != chelp[i]) && ('\0' != chelp[i][0]); i++)
		puts(chelp[i]);
}

static int query_bravais(REFINERUN *rfp)
{ /*	This prompts the user to choose a Bravais lattice */
	int		platt, retv;
	char		cplatt, inc[3];
	REFPARS		*xp;

 	xp = rfp->rfr_xp;
	printf("Indicate choice (a-p):\n");
	fgets(inc, 2, stdin);
	if (('A' <= inc[0]) && (inc[0] <= 'Z')) inc[0] += 'a' - 'A';
	cplatt = inc[0] - 'a';	platt = (int)cplatt;
	if ((0 > (retv = choosebravais(platt, rfp))) || (retv > 16))
	  {
		fprintf(stderr, "Error %d from choosebravais\n", retv);
		return retv;
	  }
	return EOK;
}

static int rpolymap(REFINERUN *rfp)
{ /* This does a detector remapping via polynomial fits. */
	int	i, nusable;
	double	delp, delpden, wvec[5];
	VSPLINE	*losp;
	REFREF	*lrp;

	get_rmserrs("", rfp); /* make sure we know the xcalc,ycalc values*/
	losp = rfp->rfr_spl;	nusable = 0;
	inv_init(losp->v_lax, losp->v_ldx, 6); /* zero the vectors */
	inv_init(losp->v_lay, losp->v_ldy, 6);
	if ((delpden = (rfp->rfr_xp)->xerrphi) < 0.1) delpden = 0.1;
	for (lrp = rfp->rfr_rpp0, i = rfp->rfr_npeaks; i--; lrp++)
	   { /* go through the refls and add to the polynomial sums */
		if (lrp->rflags == OKREFL)
		  {
			delp = (lrp->frame0 - lrp->framc) / delpden;
			wvec[4] = (fabs(delp) < 1.) ? cos(PIO2 * delp) : 0.;
			if (wvec[4] > DFUZZ)
			  {
				wvec[0] = lrp->xto;	wvec[1] = lrp->yto;
				wvec[2] = lrp->xtc;	wvec[3] = lrp->ytc;
				polyadd(losp, wvec);	nusable++;
			  }
		  }
	   }
	if (nusable < 12) return EBAD; /* need >= 12 refls */
	/* now do the actual polynomial remapping and recalc errors */
	if (EOK == (i = polymap(losp, rfp->rfr_fpl, 032)))
		get_rmserrs("after polymap", rfp);
	return i;
}

int refco(int refcom, int ac, char *av[], int inta,double *flar,REFINERUN *rfp)
{ /*	This handles a single command,
	either directly to the shell or as an input from stdin  */
	register int	i;
	int		failure, j;
	double		madif, tdif, lar[3][3], prevg[3];
	float		mbv[2], oldga[3];
	REFPARS		*xp;
	LATTICESCORE	*lc0;
	ICEWORK		*lic;
	static char *fields[] = {	"A", "B", "C", "ALPHA", "BETA", "GAMMA",
				"OMEGA", "CHI", "PHI",
				"PLATED", "XCEN", "YCEN", "TILT", "SWING",
				"GAMMA0", "GAMMA1", "GAMMA2",
				"ERRORH", "ERRORX", "ERRORY", "ERRORPHI",
				"RESMIN", "RESMAX", "LAMBDA", "STEPSIZE",
				"SYSTEM", "SPACEGROUP",
				"STARTOMEGA", "STARTCHI", "STARTPHI",
				"DISTANCE", "AX0", "AY0", "2THETA",
				"GAM0", "GAM1", "GAM2",
				"GAMA0", "GAMA1", "GAMA2",
				"MAXERH", "MAXERX", "MAXERY",
				"MAXERP", "MAXERZ", "ISYS" };

	static int realval[] = {	 0, 1, 2, 3, 4, 5, /* index in array */
				 6, 7, 8,		/* of vbls and params*/
				 9,10,11,12,13,
				14,15,16,
				17,18,19,20,
				21,22,23,24,
				25,26,
				27,28,29,
				 9,10,11,13,	/* synonyms after this */
				14,15,16,
				14,15,16,
				17,18,19,
				20,20,25 };

	xp = rfp->rfr_xp;
	switch (refcom)
	      {
		case	'A':	/* nonlinear refinement, type A */
		/* we now use P and H for something else.
		 Use 0 to do what P did, and 9 to do what H did. */
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			minpar1(refcom, rfp);	break;
		case	'B': /* calculate & display Bravais solutions*/
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				lar[i][j] = xp->xareal[i][j] * xp->xlambda;
			if (NULL == (lc0 = allbravais(lar, &j))) return EALLOC;
			if (14 > j) return EBAD;
			showbravais(rfp->rfr_fpl, xp->xareal, xp->xstart);
			showbravais(stdout, xp->xareal, xp->xstart);
			fprintf(rfp->rfr_fpl, "inta = %d\n", inta);
			if (inta) (void)query_bravais(rfp);
			break;
		case	'C':	/* choose an autoindexing solution.
			This command supersedes the old closed-form refinement,
			which tended not to work very well anyway */
			if (rfp->rfr_ndecent < 1) fprintf(stderr,
			 "You must run the I command first\n");
			else (void)impose_solution(inta, rfp);
			break;
		/* case	'C': */	/* closed-form linear refinement */
			/* get_rmserrs("before closed-form", rfp);
			if (ref3p(rfp)) return EBAD;
			else get_rmserrs(" after closed-form", rfp);
			break; */
		case	'D':	/* modify a parameter or setting */
			getmainbeam(rfp->rfr_xp, mbv);
			if (EOK != (failure = ref1modify(ac, av,
			 mbv, rfp, realval, fields))) return failure;
			break;
		case	'F':	/* Choose a Bravais lattice */
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				lar[i][j] = xp->xareal[i][j] * xp->xlambda;
			if (NULL == (lc0 = allbravais(lar, &j))) return EALLOC;
			if (14 > j) return EBAD;
			if (0 > (failure = choosebravais(inta, rfp)))
			  {
				fprintf(stderr, "Failure %d in choosebravais\n",
					failure);
				return EBAD;
			  }
			fprintf(rfp->rfr_fpl,
	 "imposing symmetry associated with xtal system %1d, spacegroup %3d\n",
				(int)(xp->xisys), (int)(xp->xispgp));
			fprintf(stdout,
	 "imposing symmetry associated with xtal system %1d, spacegroup %3d\n",
				(int)(xp->xisys), (int)(xp->xispgp));
			break;
		case	'H':	/* polynomial remapping.
			 H used to have a different job. */
			if (EOK != (failure = rpolymap(rfp)))
			  {
				fprintf(stdout, "rpolymap error %d\n",failure);
				return EBAD;
			  }
			break;
		case	'I':	/* automatic indexing */
			if (ac > 5) rfp->rfr_flags |= RR_NORECENTER;
			else rfp->rfr_flags &= ~RR_NORECENTER;
			if (EOK != (failure = autoind1(flar, rfp)))
			  {
				fprintf(stdout,
				 "autoindexing failed with error %d\n",
					failure);
				return EBAD;
			  }
			get_rmserrs("  after autoindex", rfp);	break;
		case	'J':	/* alter the permission on parameters */
			if (EOK != (failure = permit1modify(ac, av,
			 rfp, realval, fields))) return failure;
			break;
		case	'K':	/* linear refinement on scanning angle */
		case	'L':	/* incremental linear refinement */
		case	'X': case 'Y': case 'Z': /* X,Y,omega linear */
			if ((EOK != (failure = incr1p(refcom, inta, rfp))) &&
			 (-1 != failure)) return failure;
			break;
		case	'M':	/*modify mapping from pix to cm & cm to pix*/
			get_rmserrs("before remapping", rfp); /*initial stats*/
			if ((failure = remap_distortion(rfp))) return EBAD;
			break;
		case	'N':	case	' ': break; /* shouldn't happen */
		case	'O':	/* output refl list to logfile */
			get_rmserrs("", rfp);
			showpeaks(rfp->rfr_fpl, "during refinement", rfp);
			j = rfp->rfr_nreftot - rfp->rfr_nrefok;
			if (ac >= 0) fprintf(stdout,
			 " %5d good, %5d bad reflections printed to logfile\n",
				(int)(rfp->rfr_nrefok), j);
			return EOK;
		case	'P':	/* control refinement on refls in icerings */
			if (ac < 3) /* no args: toggle icering status */
			  {
				if (rfp->rfr_flags & RR_EXCLUDEICE)
					rfp->rfr_flags &= ~RR_EXCLUDEICE;
				else	rfp->rfr_flags |= RR_EXCLUDEICE;
			  }
			else if ((inta > 0) || (av[2][1] == 'n') ||
			 (av[2][1] == 'N')) rfp->rfr_flags |= RR_EXCLUDEICE;
			else if ((inta <= 0) || (av[2][1] == 'f') ||
			 (av[2][1] == 'F')) rfp->rfr_flags &= ~RR_EXCLUDEICE;
			fprintf(stdout, "Reflections in icerings will be %s\n",
			 rfp->rfr_flags & RR_EXCLUDEICE ?
			 "excluded" : "included");
			if (rfp->rfr_flags & RR_EXCLUDEICE)
			  {
				if (NULL ==(lic = manageice(013)))
				 fprintf(stderr,
				 "Failure to build icering list\n");
				else if (lic->icw_nranges < 1) fprintf(stderr,
				 "No icering ranges stored\n");
				else describe_rings(stdout);
			  }
			break;
		case	'Q':	/* signal to quit */
			return EOK;
		case	'R':	/* rocking-curve refinement */
			oldga[0] = xp->xgam[0] / fabs(xp->xfrmsize);
			oldga[1] = xp->xgam[1] / fabs(xp->xfrmsize);
			oldga[2] = xp->xgam[2];	 j = 0;
		/* failure in rocking-curve refinement is non-fatal */
			if ((inta < 2) || (inta > 100))
				(void)rocking_ref(oldga, 1, rfp);
			else
			 {
			   for (madif = j = 0; j < inta; j++)
			    {
			      for (i = 0; i < 3; i++) prevg[i] = xp->xgam[i];
				if (1 == rocking_ref(oldga,
				 (j >= inta - 1), rfp))
				  {
					printf(
	 "Failure from gamref on cycle %d\n", j);	break;
				  }
				oldga[0] = xp->xgam[0] / fabs(xp->xfrmsize);
				oldga[1] = xp->xgam[1] / fabs(xp->xfrmsize);
				oldga[2] = xp->xgam[2];
				madif = fabs(xp->xgam[0] - prevg[0]);
				if (madif < (tdif = fabs(xp->xgam[1] -
					prevg[1]))) madif = tdif;
				if (madif < (tdif = fabs(xp->xgam[2] -
					prevg[2]))) madif = tdif;
				if (madif < DFUZZ)
				  {
					printf(
 "Rocking-curve refinement converged to %12.5e in %d cycles\n", madif, j + 1);
					break;
				  }
			    } /* end of loop through j values */
			   if (j >= inta) printf(
 "Rocking-curve refinement converged to %12.5e in %d cycles\n", madif, j + 1);
			 } /* end of else of "if (inta < 2"... clause */
			break;
		case	'S':	/* specify refls by h,k,l,X,Y,phi */
			if (get_hkl(rfp)) /* failure: ignore results */
				return EBAD;
			if (-1 == gpeaks(rfp)) return EREADERR; /*read 'em */
			break;
		case	'T':	/* transform unit cell */
			if (0 >= inta) list_transforms();
			else if (EOK != (failure =
			 cellconvert(rfp->rfr_xp, inta))) return failure;
			break;
		case	'V':	/* list available commands */
			list_commands();	break;
		case	'W':	/* might as well report on write */
		case	'E':	/* exit involves reporting */
			j = 0;	if (EOK != (j = globtxs(7, rfp)))
			  {
				if (NULL != rfp->rfr_fpl) fprintf(rfp->rfr_fpl,
 "Error %d from globtxs while attempting to write out parameters\n", j);
				return j;
			  }
			break;
		case	'G':	/* reading in requires reporting */
			j = 0;	if (EOK != (j == readpars(rfp))) return j;
			break;
		case	'U':	/* report */
			break;
		default:	break;
	      }
	return EOK;
}

/* end of refio.c */
